// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CBitmap K-3D object, which encapsulates bitmap images
		\author Tim Shead (tshead@k-3d.com)
*/

#include <bitmap.h>

#include <k3dmisc.h>
#include <k3dfilefilters.h>
#include <k3dcom.h>

#include <k3diapplication.h>
#include <k3difileformat.h>
#include <k3dibitmapreadformat.h>
#include <k3dibitmapwriteformat.h>
#include <k3dibitmapserver.h>
#include <k3diobjectcontext.h>
#include <k3direnderjob.h>

#include <sdpfile.h>
#include <sdpgtkevent.h>
#include <sdpgtkfileselector.h>

/////////////////////////////////////////////////////////
// CBitmap

CBitmap::CBitmap()
{
	AddVariable(&m_ImagePath("imagepath", ""));
	AddVariable(&m_XRepeat("xrepeat", TILE));
	AddVariable(&m_YRepeat("yrepeat", TILE));
	AddVariable(&m_UseAlpha("usealpha", true));
	AddVariable(&m_UseLuma("useluma", false));
	AddVariable(&m_FullSize("fullsize", true));
	AddVariable(&m_Perspective("perspective", true));
	AddVariable(&m_Antialias("antialias", false));

	AddObject(&m_X1("x1"));
	AddObject(&m_X2("x2"));
	AddObject(&m_Y1("y1"));
	AddObject(&m_Y2("y2"));
}

CBitmap::~CBitmap()
{
}

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CBitmap::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIAnimate>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIBitmap>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIEditorTexture>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIObject>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIPersist>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIRiTexture>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dITexture>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dITimeServer>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIObject*>(this), InterfaceID, Interface)) return SDP_S_OK;										

	return SDP_E_NO_INTERFACE;
}

gulong CBitmap::AddRef()
{
	return sdpComAddRef(m_ReferenceCount);
}

gulong CBitmap::Release()
{
	return sdpComRelease(this, m_ReferenceCount);
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CBitmap::EditObject()
{
	g_return_val_if_fail(m_UserInterface.Create(this), SDP_E_FAIL);
	return SDP_S_OK;
}

sdpResult CBitmap::ObjectRenamed(SDPCSTRING NewName)
{
	m_UserInterface.ObjectRenamed(NewName);
	return k3dPersistObject::ObjectRenamed(NewName);
}

sdpResult CBitmap::ObjectDeleted()
{
	m_UserInterface.ObjectDeleted();
	return k3dPersistObject::ObjectDeleted();
}

///////////////////////////////////////////////////////////////
// k3dISelectorClient overrides

///////////////////////////////////////////////////////////////
// k3dIBitmap overrides

sdpResult CBitmap::Allocate(gulong Width, gulong Height)
{
	// Sanity checks ...
	g_return_val_if_fail(Width, SDP_E_FAIL);
	g_return_val_if_fail(Height, SDP_E_FAIL);

	// Get rid of existing data ...
	m_OpenGLData.Free();
	
	bool result = m_Image.Allocate(Width, Height);
	
	Notify("bitmap-changed");
	
	return result ? SDP_S_OK : SDP_E_FAIL;
}

sdpResult CBitmap::Free()
{
	m_ImagePath() = "";
	m_Image.Free();

	Notify("bitmap-changed");

	return SDP_S_OK;
}

k3dPixel* CBitmap::GetData()
{
	return m_Image.Data();
}

gulong CBitmap::GetWidth()
{
	return m_Image.Width();
}

gulong CBitmap::GetHeight()
{
	return m_Image.Height();
}

sdpResult CBitmap::LoadFile(SDPCSTRING FilePath, SDPCSTRING Format, k3dIBitmapReadFormat* FormatFilter)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);
	g_return_val_if_string_fail(Format, SDP_E_FAIL);

	// Our result ...
	sdpResult result;

	// Free existing data ...
	Free();

	// Store the filepath ...
	m_ImagePath() = FilePath;

	// Try using the provided format ...
	k3dIBitmapReadFormat* formatfilter = FormatFilter;

	// A format was provided, so use it ...
	if(formatfilter)
		{
			// Read the file ...
			gulong width = 0;
			gulong height = 0;
			result = formatfilter->PreRead(this, FilePath, Format, &width, &height);
			if(sdpSucceeded(result))
				{
					result = Allocate(width, height);
					if(sdpSucceeded(result))
						result = formatfilter->ReadFile(this, FilePath, Format);
				}
		}
	// A format wasn't provided ...
	else
		{
			// ... so come up with one of our own ...
			formatfilter = k3dFileFormat<k3dIBitmapReadFormat>(Document(), FilePath, Format);

			// If we couldn't find a bitmap format, we're done ...
			g_return_val_if_fail(formatfilter, SDP_E_FAIL);

			// Read the file ...
			gulong width = 0;
			gulong height = 0;
			result = formatfilter->PreRead(this, FilePath, Format, &width, &height);
			if(sdpSucceeded(result))
				{
					result = Allocate(width, height);
					if(sdpSucceeded(result))
						{
							result = formatfilter->ReadOptions(this, FilePath, Format);
							if(sdpSucceeded(result))
								result = formatfilter->ReadFile(this, FilePath, Format);
						}
				}

			// Since we created this object ourselves, zap it ...
			k3dDeleteObject(formatfilter);
		}

	Notify("bitmap-changed");

	return result;
}

sdpResult CBitmap::SaveFile(SDPCSTRING FilePath, SDPCSTRING Format, k3dIBitmapWriteFormat* FormatFilter)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);
	g_return_val_if_string_fail(Format, SDP_E_FAIL);

	// Our return value ...
	sdpResult result;

	// Try using the provided format ...
	k3dIBitmapWriteFormat* formatfilter = FormatFilter;

	// A format was provided, so use it ...
	if(formatfilter)
		{
			result = formatfilter->PreWrite(this, FilePath, Format);
			
			if(sdpSucceeded(result))
				result = formatfilter->WriteFile(this, FilePath, Format);
		}
	// A format wasn't provided ...
	else
		{
			// ... so come up with one of our own ...
			formatfilter = k3dFileFormat<k3dIBitmapWriteFormat>(Document(), FilePath, Format);

			// If we couldn't find a bitmap format, we're done ...
			g_return_val_if_fail(formatfilter, SDP_E_FAIL);

			result = formatfilter->PreWrite(this, FilePath, Format);
			if(sdpSucceeded(result))
				{
					result = formatfilter->WriteOptions(this, FilePath, Format);
					if(sdpSucceeded(result))
						result = formatfilter->WriteFile(this, FilePath, Format);
				}

			// Since we created this object ourselves, zap it ...
			k3dDeleteObject(formatfilter);
		}

	return result;
}

sdpResult CBitmap::FlipVertical()
{
	// Sanity checks ...
	g_return_val_if_fail(GetData() && GetWidth() && GetHeight(), SDP_E_FAIL);

	m_Image.FlipVertical();

	Notify("bitmap-changed");
	return SDP_S_OK;
}

sdpResult CBitmap::FlipHorizontal()
{
	// Sanity checks ...
	g_return_val_if_fail(GetData() && GetWidth() && GetHeight(), SDP_E_FAIL);

	m_Image.FlipHorizontal();

	Notify("bitmap-changed");
	return SDP_E_FAIL;
}

sdpResult CBitmap::Scale(gulong NewWidth, gulong NewHeight)
{
	// Sanity checks ...
	g_return_val_if_fail(GetData() && GetWidth() && GetHeight(), SDP_E_FAIL);

	m_Image.Scale(NewWidth, NewHeight, k3dImageBuffer::NEAREST);

	Notify("bitmap-changed");
	return SDP_E_FAIL;
}

////////////////////////////////////////////////////////////////////////////
// k3dIEditorTexture overrides

sdpResult CBitmap::EditorTexturePreRender(k3dIEditorEngine* Engine)
{
	// If we don't have any data, we're done ...
	if(!GetData())
		return SDP_S_FALSE;

	// Create our OpenGL data as needed ...
	if(!m_OpenGLData.Data() || !m_OpenGLData.Width() || !m_OpenGLData.Height())
		{
			m_OpenGLData.Copy(GetData(), GetWidth(), GetHeight());
			m_OpenGLData.ConvertForOpenGL(k3dImageBuffer::NEAREST);
		}

	return SDP_S_OK;
}

sdpResult CBitmap::EditorTextureSetup(k3dIEditorEngine* Engine)
{
	// If we don't have any data, we're done ...
	if(!GetData())
		{
			glDisable(GL_TEXTURE_2D);
			return SDP_S_OK;
		}

	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	GLint texturecomponents = m_UseLuma() ? 1 : 3;
	texturecomponents += m_UseAlpha() ? 1 : 0;

	glTexImage2D(GL_TEXTURE_2D,	// Reserved
		0,												// Detail level
		texturecomponents,				// Components used for modulating
		m_OpenGLData.Width(),			// Bitmap width
		m_OpenGLData.Height(),		// Bitmap height
		0,												// Border size
		GL_RGBA,									// GL_RGB or GL_RGBA
		GL_UNSIGNED_BYTE,					// Data type of pixel values
		m_OpenGLData.Data()				// Pixel values
		);

	if(m_XRepeat() == CLAMP)
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	else
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);

	if(m_YRepeat() == CLAMP)
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	else
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

	if(m_Antialias())
		{
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		}
	else
		{
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		}

	if(m_Perspective())
		glHint(GL_PERSPECTIVE_CORRECTION_HINT,GL_NICEST);
	else
		glHint(GL_PERSPECTIVE_CORRECTION_HINT,GL_FASTEST);

	glEnable(GL_TEXTURE_2D);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glDisable(GL_TEXTURE_GEN_S);
	glDisable(GL_TEXTURE_GEN_T);

	return SDP_S_OK;
}

////////////////////////////////////////////////////////////////////
// k3dIRiTexture overrides

sdpResult CBitmap::RiTextureStartJob(k3dIRenderJob* Job, k3dIRiEngine* Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Job, SDP_E_FAIL);
	g_return_val_if_fail(Engine, SDP_E_FAIL);
	
	gulong buffersize = 1024;
	Job->GetUniqueFile("texture", m_RiTexturePath.buffer(buffersize), &buffersize);
	
	m_RiTexturePath.resize(strlen(m_RiTexturePath));
	m_RiTexturePath += ".tif";

	return SaveFile(m_RiTexturePath, k3dIFileFormat::Automatic, 0);
}

sdpResult CBitmap::RiTexturePreRender(k3dIRenderFrame* Frame, k3dIRiEngine* Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Frame, SDP_E_FAIL);
	g_return_val_if_fail(Engine, SDP_E_FAIL);
	
	return SDP_S_OK;
}

sdpResult CBitmap::RiTextureFinishJob(k3dIRenderJob* Job, k3dIRiEngine* Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Job, SDP_E_FAIL);
	g_return_val_if_fail(Engine, SDP_E_FAIL);
	
	return SDP_S_OK;
}

SDPCSTRING CBitmap::RiTexturePath(k3dIRenderFrame* Frame, k3dIRiEngine* Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Frame, 0);
	g_return_val_if_fail(Engine, 0);
	
	return m_RiTexturePath;
}

///////////////////////////////////////////////////////////////
// k3dIAnimate overrides

sdpResult CBitmap::PostAnimate()
{
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIPersist overrides

sdpResult CBitmap::InitNew()
{
	g_return_val_if_fail(sdpSucceeded(k3dPersistObject::InitNew()), SDP_E_FAIL);

	g_return_val_if_fail(sdpSucceeded(m_X1.Create(Document(), "Bitmap X1", 0.0, true, -ULONG_MAX, ULONG_MAX, "Time", "X1", k3dIChannel::DATA_SCALAR, true, this)), SDP_E_FAIL);
	g_return_val_if_fail(sdpSucceeded(m_X2.Create(Document(), "Bitmap X2", 1.0, true, -ULONG_MAX, ULONG_MAX, "Time", "X2", k3dIChannel::DATA_SCALAR, true, this)), SDP_E_FAIL);
	g_return_val_if_fail(sdpSucceeded(m_Y1.Create(Document(), "Bitmap Y1", 0.0, true, -ULONG_MAX, ULONG_MAX, "Time", "Y1", k3dIChannel::DATA_SCALAR, true, this)), SDP_E_FAIL);
	g_return_val_if_fail(sdpSucceeded(m_Y2.Create(Document(), "Bitmap Y2", 1.0, true, -ULONG_MAX, ULONG_MAX, "Time", "Y2", k3dIChannel::DATA_SCALAR, true, this)), SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CBitmap::LoadComplete()
{
	g_return_val_if_fail(sdpSucceeded(k3dPersistObject::LoadComplete()), SDP_E_FAIL);
	
	// Reload the bitmap image ...
	if(m_ImagePath().size())
		LoadFile(m_ImagePath(), k3dIFileFormat::Automatic, 0);
	
	return SDP_S_OK;
}

/////////////////////////////////////////////////////////////////////////////
// CBitmapDialog

CBitmapDialog::CBitmapDialog()
{
	m_Object = 0;
}

bool CBitmapDialog::Create(CBitmap* Object)
{
	// Store our parent object ...
	m_Object = Object;

	// Load our template ...
	g_return_val_if_fail(k3dDialog::Create(Object, false, true, "bitmap.gtkml", "bitmap.html"), false);

	return true;
}

bool CBitmapDialog::InitializeControls()
{
	// Setup OpenGL for rendering previews ...
	sdpGtkContainer rgbcontainer(Container("rgbpreview"));
	g_return_val_if_fail(m_RGBWidget.Create(rgbcontainer, true, true, 5, 6, 5, 0), false);
	
	MapEvent("configure-event", "onconfigurergb", m_RGBWidget, true);
	MapEvent("expose-event", "onexposergb", m_RGBWidget, true);
	
	sdpGtkContainer alphacontainer(Container("alphapreview"));
	g_return_val_if_fail(m_AlphaWidget.Create(alphacontainer, true, true, 5, 6, 5, 0), false);

	MapEvent("configure-event", "onconfigurealpha", m_AlphaWidget, true);
	MapEvent("expose-event", "onexposealpha", m_AlphaWidget, true);
	
	return true;
}

void CBitmapDialog::UpdateControls()
{
	if(m_Object->m_ImagePath().empty())
		Label("pathlabel").SetText("<Choose Bitmap>");
	else
		Label("pathlabel").SetText(sdpFile(m_Object->m_ImagePath()));
		
	// Get image data ...
	gulong imagewidth = m_Object->GetWidth();
	gulong imageheight = m_Object->GetHeight();

	gdouble imageratio = 1.0;
	if(imageheight)
		imageratio = gdouble(imagewidth) / gdouble(imageheight);

	m_PreviewImage.Copy(m_Object->m_Image);
	m_PreviewImage.ConvertForOpenGL(128, 128, k3dImageBuffer::NEAREST);

	// Update our frame aspect ratios to match the image ...
	AspectFrame("rgbframe").Set(0.5, 0.5, imageratio, false);
	AspectFrame("alphaframe").Set(0.5, 0.5, imageratio, false);

	// Update our information fields ...
	Label("width").SetText(sdpFormatString("Width: %u", imagewidth));
	Label("height").SetText(sdpFormatString("Height: %u", imageheight));
}

void CBitmapDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "onconfigurergb" || Event->Name() == "onexposergb")
		DrawRGB();
	else if(Event->Name() == "onconfigurealpha" || Event->Name() == "onexposealpha")
		DrawAlpha();
	else if(Event->Name() == "onload")
		OnLoad();
	else if(Event->Name() == "onsave")
		OnSave();
	else if(Event->Name() == "onview")
		OnView();
	else
		k3dDialog::OnEvent(Event);
}

void CBitmapDialog::FinalizeControls()
{
	// Get rid of our OpenGL contexts ...
	if(m_RGBWidget.Initialized())
		m_RGBWidget.Destroy();
		
	if(m_AlphaWidget.Initialized())
		m_AlphaWidget.Destroy();

	k3dDialog::FinalizeControls();
}

void CBitmapDialog::DrawRGB()
{
	// If we haven't been initialized yet, we're done ...
	if(!m_RGBWidget.Initialized())
		return;

	// Get the preview image data ...
	k3dPixel* imagedata = m_PreviewImage.Data();
	gulong imagewidth = m_PreviewImage.Width();
	gulong imageheight = m_PreviewImage.Height();

	// If we don't have any data, we're done ...
	if(!imagedata || !imagewidth || !imageheight)
		return;

	// Start OpenGL rendering ...
	m_RGBWidget.Begin();

	// Get the widget sizes ...
	gulong width = m_RGBWidget.Width();
	gulong height = m_RGBWidget.Height();

	glViewport(0, 0, width, height);

	// Setup the viewing projection ...
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(-1.0, 1.0, -1.0, 1.0, -1.0, 1.0);

	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
	glShadeModel(GL_FLAT);

	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, imagewidth, imageheight, 0, GL_RGBA, GL_UNSIGNED_BYTE, imagedata);

	glEnable(GL_TEXTURE_2D);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glBegin(GL_POLYGON);
		glTexCoord2d(0.0, 0.0); glVertex2d(-1.0, 1.0);
		glTexCoord2d(1.0, 0.0); glVertex2d(1.0, 1.0);
		glTexCoord2d(1.0, 1.0); glVertex2d(1.0, -1.0);
		glTexCoord2d(0.0, 1.0); glVertex2d(-1.0, -1.0);
	glEnd();

	glFlush();

	m_RGBWidget.SwapBuffers();
	m_RGBWidget.End();
}

void CBitmapDialog::DrawAlpha()
{
	// If we haven't been initialized yet, we're done ...
	if(!m_AlphaWidget.Initialized())
		return;

	// Get image data ...
	k3dPixel* imagedata = m_PreviewImage.Data();
	gulong imagewidth = m_PreviewImage.Width();
	gulong imageheight = m_PreviewImage.Height();

	// If we don't have any data, we're done ...
	if(!imagedata || !imagewidth || !imageheight)
		return;

	// Start OpenGL rendering ...
	m_AlphaWidget.Begin();

	// Get the widget sizes ...
	gulong width = m_AlphaWidget.Width();
	gulong height = m_AlphaWidget.Height();

	glViewport(0, 0, width, height);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(-1.0, 1.0, -1.0, 1.0, -1.0, 1.0);

	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
	glShadeModel(GL_FLAT);

	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	glTexImage2D(GL_TEXTURE_2D, 0, GL_ALPHA, imagewidth, imageheight, 0, GL_RGBA, GL_UNSIGNED_BYTE, imagedata);

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_TEXTURE_2D);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glBegin(GL_POLYGON);
		glTexCoord2d(0.0, 0.0); glVertex2d(-1.0, 1.0);
		glTexCoord2d(1.0, 0.0); glVertex2d(1.0, 1.0);
		glTexCoord2d(1.0, 1.0); glVertex2d(1.0, -1.0);
		glTexCoord2d(0.0, 1.0); glVertex2d(-1.0, -1.0);
	glEnd();

	glFlush();

	m_AlphaWidget.SwapBuffers();
	m_AlphaWidget.End();
}

void CBitmapDialog::OnLoad()
{
	// Get our starting path - start with the user's home directory ...
	sdpString startpath = sdpGetHomeDirectory();
	
	// Next, override it with the user's options ...
	sdpXmlIElement* startpathelement = sdpXmlGetXPathElement(m_Object->Application()->Options()->GetRootElement(), "application/paths/path[@type=\"bitmap\"]");
	if(startpathelement)
		startpath = startpathelement->Text();
	
	// Last-but-not-least, see if the bitmap object already has a path ...
	if(m_Object->m_ImagePath().size())
		startpath = sdpDirectory(m_Object->m_ImagePath());

	// Append a directory separator (GtkFileSelector seems to want it) ...
	startpath += G_DIR_SEPARATOR;

	// Put up a file selection dialog ...
	sdpGtkFileSelector selector("Choose bitmap file:", startpath);
	selector.DoModal();
	if(selector.OK())
		{
			// Store the new path ...
			if(startpathelement)
				startpathelement->SetText(sdpDirectory(selector.FilePath()));

			m_Object->StartQueueEvents();
			m_Object->LoadFile(selector.FilePath(), k3dIFileFormat::Automatic, 0);
			m_Object->EndQueueEvents();

			UpdateControls();

			// Update the display ...
			k3dRedrawAll(m_Object->Document(), false);		
		}
}

void CBitmapDialog::OnSave()
{
	// Get our starting path - start with the user's home directory ...
	sdpString startpath = sdpGetHomeDirectory();
	
	// Next, override it with the user's options ...
	sdpXmlIElement* startpathelement = sdpXmlGetXPathElement(m_Object->Application()->Options()->GetRootElement(), "application/paths/path[@type=\"bitmap\"]");
	if(startpathelement)
		startpath = startpathelement->Text();
	
	// Last-but-not-least, see if the bitmap object already has a path ...
	if(m_Object->m_ImagePath().size())
		startpath = sdpDirectory(m_Object->m_ImagePath());

	// Append a directory separator (GtkFileSelector seems to want it) ...
	startpath += G_DIR_SEPARATOR;

	// Put up a file selection dialog ...
	sdpGtkFileSelector selector("Save bitmap file:", startpath);
	selector.DoModal();
	if(selector.OK())
		{
			// Store the new path ...
			if(startpathelement)
				startpathelement->SetText(sdpDirectory(selector.FilePath()));

			m_Object->StartQueueEvents();
			m_Object->SaveFile(selector.FilePath(), k3dIFileFormat::Automatic, 0);
			m_Object->EndQueueEvents();
		}
}

void CBitmapDialog::OnView()
{
	// Figure out their preferred bitmap viewer ...
	sdpXmlElementArray viewers = sdpXmlXPathSelect(m_Object->Application()->Options()->GetRootElement(), "application/helpers/command[@type=\"bitmapviewer\"]");
	if(viewers.size())
		{
			// View the image ...
			sdpString viewcommand = viewers[0]->Text();
			viewcommand.formattedreplace('%', "p", m_Object->m_ImagePath());
			viewcommand += " &";
			
			system(viewcommand);
		}
}


