// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CNetPBMReader K-3D object, which reads and writes JPEG images
		\author Tim Shead (tshead@k-3d.com)
*/

#include "netpbmreader.h"

#include <k3dibitmap.h>
#include <k3dmisc.h>

#include <sdpxmlutility.h>
#include <sdpregex.h>
#include <sdpfile.h>

#include <stdio.h>
#include <fstream>
#include <strstream>

CNetPBMReader::~CNetPBMReader()
{
}

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CNetPBMReader::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIFileFormat>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIBitmapReadFormat>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIObject>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIObject*>(this), InterfaceID, Interface)) return SDP_S_OK;

	return SDP_E_NO_INTERFACE;
}

gulong CNetPBMReader::AddRef()
{
	return sdpComAddRef(m_ReferenceCount);
}

gulong CNetPBMReader::Release()
{
	return sdpComRelease(this, m_ReferenceCount);
}

///////////////////////////////////////////////////////////////
// k3dIFileFormat overrides

K3DFILEFORMATHANDLE CNetPBMReader::GetFirstFileFormat()
{
	return 0;
}

K3DFILEFORMATHANDLE CNetPBMReader::GetNextFileFormat(K3DFILEFORMATHANDLE Handle)
{
	return 0;
}

SDPCSTRING CNetPBMReader::GetFileFormat(K3DFILEFORMATHANDLE Handle)
{
	return "";
}

sdpResult CNetPBMReader::QueryCanHandle(SDPCSTRING FilePath, SDPCSTRING Format)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);
	g_return_val_if_string_fail(Format, SDP_E_FAIL);

	sdpXmlIElement* reader = GetReader(FilePath, Format);

	if(reader)
		{
			sdpString format;
			sdpXmlGetAttribute(reader, "format", format);
			std::cout << "Identified " << format << " bitmap image" << std::endl;
			
			return SDP_S_OK;
		}	
	
	return SDP_S_FALSE;
}

///////////////////////////////////////////////////////////////
// k3dIBitmapReadFormat overrides

k3dIFileFormat* CNetPBMReader::FileFormat()
{
	return this;
}

sdpResult CNetPBMReader::PreRead(k3dIBitmap* Bitmap, SDPCSTRING FilePath, SDPCSTRING Format, gulong* Width, gulong* Height)
{
	// Sanity checks ...
	g_return_val_if_fail(Bitmap, SDP_E_FAIL);
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);
	g_return_val_if_string_fail(Format, SDP_E_FAIL);
	g_return_val_if_fail(Width, SDP_E_FAIL);
	g_return_val_if_fail(Height, SDP_E_FAIL);

	// Convert the source file to a ppm file ...
	g_return_val_if_fail(ConvertToPNM(FilePath, Format, m_PNMFile), SDP_E_FAIL);

	// Read the PNM file header to get its dimensions ...
	std::ifstream file(m_PNMFile);
	g_return_val_if_fail(file.good(), SDP_E_FAIL);

	gulong width = 0;
	gulong height = 0;
	ImageType type;

	g_return_val_if_fail(ReadPNMHeader(file, width, height, type), SDP_E_FAIL);
	g_return_val_if_fail(width, SDP_E_FAIL);
	g_return_val_if_fail(height, SDP_E_FAIL);

	// Return 'em ...
	*Width = width;
	*Height = height;

	return SDP_S_OK;
}

sdpResult CNetPBMReader::ReadOptions(k3dIBitmap* Bitmap, SDPCSTRING FilePath, SDPCSTRING Format)
{
	// Sanity checks ...
	g_return_val_if_fail(Bitmap, SDP_E_FAIL);
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);
	g_return_val_if_string_fail(Format, SDP_E_FAIL);
	
	return SDP_S_OK;
}

sdpResult CNetPBMReader::ReadFile(k3dIBitmap* Bitmap, SDPCSTRING FilePath, SDPCSTRING Format)
{
	// Sanity checks ...
	g_return_val_if_fail(Bitmap, SDP_E_FAIL);
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);
	g_return_val_if_string_fail(Format, SDP_E_FAIL);
	g_return_val_if_fail(m_PNMFile.size(), SDP_E_FAIL);

	// Open the file ...
	std::ifstream file(m_PNMFile);
	g_return_val_if_fail(file.good(), SDP_E_FAIL);

	// Load it up ...
	g_return_val_if_fail(ReadPNM(file, Bitmap), SDP_E_FAIL);

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Helper functions

void CNetPBMReader::LoadFormats()
{
	if(m_Formats.GetRootElement())
		return;
	
	sdpString formatfile = sdpAppendPath(k3dLibraryPath(this), "netpbm.conf");
	g_return_if_fail(m_Formats.LoadFile(formatfile));
}

sdpXmlIElement* CNetPBMReader::GetReader(SDPCSTRING FilePath, SDPCSTRING Format)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, 0);
	g_return_val_if_string_fail(Format, 0);

	// Load our external format description file ...
	LoadFormats();

	// Make sure we've got something ...
	g_return_val_if_fail(m_Formats.GetRootElement(), 0);
	g_return_val_if_fail(m_Formats.GetRootElement()->Name() == sdpString("netpbm"), 0);
	
	sdpXmlIElement* readers = sdpXmlFindElement(m_Formats.GetRootElement(), "readers");
	g_return_val_if_fail(readers, 0);

	// First, see if the filetype can be identified by /etc/magic through the "file" command ...
	sdpString commandline = sdpString("file -b ") + FilePath;
	std::ostrstream filetypebuffer;
	if(sdpRunProcess(commandline, filetypebuffer))
		{
			sdpString filetype = filetypebuffer.str();
			filetypebuffer.freeze(false);

			// For each netpbm reader ...
			for(SDPXMLELEMENTHANDLE e = readers->GetFirstElement(); e; e = readers->GetNextElement(e))
				{
					sdpXmlIElement* reader = readers->GetElement(e);
				
					// See if it has a "magic" regular expression ...
					sdpString magic;
					sdpXmlGetAttribute(reader, "magic", magic);
					if(0 == magic.size())
						continue;
						
					sdpRegEx expression;
					if(!expression.Create(magic))
						{
							std::cerr << "Error compiling magic regular expression " << magic << std::endl;
							continue;
						}
	
					// See if the "magic" regular expression has any matches against the output of the "file" command ...
					sdpRegExMatchArray matches;	
					expression.Search(filetype, matches);
	
					if(matches.size() > 0)
						return reader;
				}
		}
		
	// Second try ... look at filename for standard file extensions ...
	for(SDPXMLELEMENTHANDLE e = readers->GetFirstElement(); e; e = readers->GetNextElement(e))
		{
			sdpXmlIElement* reader = readers->GetElement(e);
		
			// See if it has a "magic" regular expression ...
			sdpString filename;
			sdpXmlGetAttribute(reader, "filename", filename);
			if(0 == filename.size())
				continue;

			sdpRegEx expression;
			if(!expression.Create(filename, false))
				{
					std::cerr << "Error compiling filename regular expression " << filename << std::endl;
					continue;
				}

			// See if the filename regular expression has any matches against the input filename ...
			sdpRegExMatchArray matches;
			expression.Search(FilePath, matches);

			if(matches.size() > 0)
				return reader;
		}

	return 0;
}

bool CNetPBMReader::ConvertToPNM(SDPCSTRING FilePath, SDPCSTRING Format, sdpString& PNMFilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, false);
	g_return_val_if_string_fail(Format, false);
	
	PNMFilePath.clear();
	
	sdpXmlIElement* reader = GetReader(FilePath, Format);
	g_return_val_if_fail(reader, false);
	
	sdpString command;
	sdpXmlGetAttribute(reader, "command", command);
	g_return_val_if_fail(command.size(), false);
 	
	command.formattedreplace('%', "p", FilePath);
	  
	sdpString pnmfile = "/tmp/netpbmreader.pnm";
	sdpString commandline = command + " > " + pnmfile;
	
	g_return_val_if_fail(sdpRunProcess(commandline, std::cout), false);

	PNMFilePath = pnmfile;
	return true;
}

bool CNetPBMReader::ReadPNMHeader(std::istream& Stream, gulong& Width, gulong& Height, ImageType& Type)
{
	// Sanity checks ...
	g_return_val_if_fail(Stream.good(), false);

	// Check for magic characters at the beginning of the file ...
	char magic1, magic2;
	Stream >> magic1 >> magic2;
	if(Stream.eof())
		return false;
	if(magic1 != 'P')
		return false;

	switch(magic2)
		{
			case '4':
				Type = PBM;
				break;
			case '5':
				Type = PGM;
				break;
			case '6':
				Type = PPM;
				break;
			default:
				return false;
		}

	// Get the file dimensions ...
	Stream >> Width >> Height;
	if(Stream.eof())
		return false;

	// If it's a bitmap image, we're done ...
	if(PBM == Type)
		{
			Stream >> std::ws;
			return true;
		}

	// Get the maximum per-channel value for the image ...
	gulong maxvalue = 0;
	Stream >> maxvalue >> std::ws;
	if(Stream.eof())
		return false;
		
	return true;
}

bool CNetPBMReader::ReadPNM(std::istream& Stream, k3dIBitmap* Bitmap)
{
	// Sanity checks ...
	g_assert(Stream.good());
	g_assert(Bitmap);
	
	gulong sourcewidth = 0;
	gulong sourceheight = 0;
	ImageType sourcetype;

	g_return_val_if_fail(ReadPNMHeader(Stream, sourcewidth, sourceheight, sourcetype), false);
	g_return_val_if_fail(sourcewidth, false);
	g_return_val_if_fail(sourceheight, false);
	
	// Get the destination bitmap buffer ...
	k3dPixel* destination = Bitmap->GetData();
	gulong destinationwidth = Bitmap->GetWidth();
	gulong destinationheight = Bitmap->GetHeight();

	// Sanity checks ...
	g_return_val_if_fail(destination, false);
	g_return_val_if_fail(sourcewidth == destinationwidth, false);
	g_return_val_if_fail(sourceheight == destinationheight, false);

	switch(sourcetype)
		{
			case PBM:
				return ReadPBM(Stream, destination, sourcewidth, sourceheight);
			case PGM:
				return ReadPGM(Stream, destination, sourcewidth, sourceheight);
			case PPM:
				return ReadPPM(Stream, destination, sourcewidth, sourceheight);
			default:
				g_assert_not_reached();
		}

	return false;
}

bool CNetPBMReader::ReadPBM(std::istream& Stream, k3dPixel* Destination, gulong Width, gulong Height)
{
	// Sanity checks ...
	g_assert(Stream.good());
	g_assert(Destination);
	g_assert(Width);
	g_assert(Height);
	
	gchar packed, pixel;

	// For each source row ...
	for(gulong y = 0; y < Height; y++)
		{
			for(gulong x = 0; x < Width; x++)
				{
					if(0 == x%8)
						{
							Stream.get(packed);
							g_return_val_if_fail(!Stream.eof(), false);
						}

					pixel = (packed >> (7 - (x%8))) & 0x01;

					Destination->Red = pixel * 255;
					Destination->Green = pixel * 255;
					Destination->Blue = pixel * 255;
					Destination->Alpha = 255;
			
					Destination++;
				}
		}
	return true;
} 

bool CNetPBMReader::ReadPGM(std::istream& Stream, k3dPixel* Destination, gulong Width, gulong Height)
{
	// Sanity checks ...
	g_assert(Stream.good());
	g_assert(Destination);
	g_assert(Width);
	g_assert(Height);
	
	gchar grey;

	// For each source row ...
	for(gulong y = 0; y < Height; y++)
		{
			for(gulong x = 0; x < Width; x++)
				{
					Stream.get(grey);
					g_return_val_if_fail(!Stream.eof(), false);

					Destination->Red = grey;
					Destination->Green = grey;
					Destination->Blue = grey;
					Destination->Alpha = 255;
					
					Destination++;
				}
		}
	return true;
} 

bool CNetPBMReader::ReadPPM(std::istream& Stream, k3dPixel* Destination, gulong Width, gulong Height)
{
	// Sanity checks ...
	g_assert(Stream.good());
	g_assert(Destination);
	g_assert(Width);
	g_assert(Height);
	
	gchar red, green, blue;

	// For each source row ...
	for(gulong y = 0; y < Height; y++)
		{
			for(gulong x = 0; x < Width; x++)
				{
					Stream.get(red);
					Stream.get(green);
					Stream.get(blue);
					g_return_val_if_fail(!Stream.eof(), false);

					Destination->Red = red;
					Destination->Green = green;
					Destination->Blue = blue;
					Destination->Alpha = 255;
					
					Destination++;
				}
		}
	return true;
}


