// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CPBMWriter K-3D object, which writes bitmap images using the NetPBM tools
		\author Tim Shead (tshead@k-3d.com)
*/

#include "netpbmwriter.h"

#include <k3dibitmap.h>
#include <k3dmisc.h>

#include <sdpxmlutility.h>
#include <sdpregex.h>
#include <sdpfile.h>

#include <stdio.h>
#include <fstream>
#include <strstream>

CNetPBMWriter::~CNetPBMWriter()
{
}

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CNetPBMWriter::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIFileFormat>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIBitmapWriteFormat>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIObject>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIObject*>(this), InterfaceID, Interface)) return SDP_S_OK;

	return SDP_E_NO_INTERFACE;
}

gulong CNetPBMWriter::AddRef()
{
	return sdpComAddRef(m_ReferenceCount);
}

gulong CNetPBMWriter::Release()
{
	return sdpComRelease(this, m_ReferenceCount);
}

///////////////////////////////////////////////////////////////
// k3dIFileFormat overrides

K3DFILEFORMATHANDLE CNetPBMWriter::GetFirstFileFormat()
{
	return 0;
}

K3DFILEFORMATHANDLE CNetPBMWriter::GetNextFileFormat(K3DFILEFORMATHANDLE Handle)
{
	return 0;
}

SDPCSTRING CNetPBMWriter::GetFileFormat(K3DFILEFORMATHANDLE Handle)
{
	return "";
}

sdpResult CNetPBMWriter::QueryCanHandle(SDPCSTRING FilePath, SDPCSTRING Format)
{
	if(GetWriter(FilePath, Format))
		return SDP_S_OK;
		
	return SDP_S_FALSE;
}

/////////////////////////////////////////////////////////////////////////////////////////////
// k3dIBitmapWriteFormat overrides ...

k3dIFileFormat* CNetPBMWriter::FileFormat()
{
	return this;
}

sdpResult CNetPBMWriter::PreWrite(k3dIBitmap* Bitmap, SDPCSTRING FilePath, SDPCSTRING Format)
{
	// Sanity checks ...
	g_return_val_if_fail(Bitmap, SDP_E_FAIL);
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);
	g_return_val_if_string_fail(Format, SDP_E_FAIL);
	
	g_return_val_if_fail(GetWriter(FilePath, Format), SDP_E_FAIL);
	
	return SDP_S_OK;
}

sdpResult CNetPBMWriter::WriteOptions(k3dIBitmap* Bitmap, SDPCSTRING FilePath, SDPCSTRING Format)
{
	// Sanity checks ...
	g_return_val_if_fail(Bitmap, SDP_E_FAIL);
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);
	g_return_val_if_string_fail(Format, SDP_E_FAIL);

	g_return_val_if_fail(GetWriter(FilePath, Format), SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CNetPBMWriter::WriteFile(k3dIBitmap* Bitmap, SDPCSTRING FilePath, SDPCSTRING Format)
{
	// Sanity checks ...
	g_return_val_if_fail(Bitmap, SDP_E_FAIL);
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);
	g_return_val_if_string_fail(Format, SDP_E_FAIL);
	
	sdpXmlIElement* writer = GetWriter(FilePath, Format);
	g_return_val_if_fail(writer, SDP_E_FAIL);

	// Save the image as a PPM file ...
	sdpString ppmfilepath = "/tmp/netpbmwriter.ppm";
	g_return_val_if_fail(SaveAsPPM(Bitmap, ppmfilepath), SDP_E_FAIL);

	// Convert the PPM file to the desired end file format ...
	sdpString command;
	sdpXmlGetAttribute(writer, "command", command);
	g_return_val_if_fail(command.size(), false);
 	
	command.formattedreplace('%', "p", ppmfilepath);
	sdpString commandline = command + " > " + sdpString(FilePath);
	
	g_return_val_if_fail(sdpRunProcess(commandline, std::cout), SDP_E_FAIL);

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Helper functions

void CNetPBMWriter::LoadFormats()
{
	if(m_Formats.GetRootElement())
		return;
	
	sdpString formatfile = sdpAppendPath(k3dLibraryPath(this), "netpbm.conf");
	g_return_if_fail(m_Formats.LoadFile(formatfile));
}

sdpXmlIElement* CNetPBMWriter::GetWriter(SDPCSTRING FilePath, SDPCSTRING Format)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, 0);
	g_return_val_if_string_fail(Format, 0);

	// Load our external format description file ...
	LoadFormats();

	// Make sure we've got something ...
	g_return_val_if_fail(m_Formats.GetRootElement(), 0);
	g_return_val_if_fail(m_Formats.GetRootElement()->Name() == sdpString("netpbm"), 0);
	
	sdpXmlIElement* writers = sdpXmlFindElement(m_Formats.GetRootElement(), "writers");
	g_return_val_if_fail(writers, 0);

	// If the requested format is "automatic", look at the filename for standard file extensions ...
	if(k3dIFileFormat::Automatic == sdpString(Format))
		{
			for(SDPXMLELEMENTHANDLE e = writers->GetFirstElement(); e; e = writers->GetNextElement(e))
				{
					sdpXmlIElement* writer = writers->GetElement(e);
		
					// See if it has a "filename" regular expression ...
					sdpString filename;
					sdpXmlGetAttribute(writer, "filename", filename);
					if(0 == filename.size())
						continue;

					sdpRegEx expression;
					if(!expression.Create(filename, false))
						{
							std::cerr << "Error compiling filename regular expression " << filename << std::endl;
							continue;
						}

					// See if the filename regular expression has any matches against the input filename ...
					sdpRegExMatchArray matches;
					expression.Search(FilePath, matches);

					if(matches.size() > 0)
						return writer;
				}
		}

	return 0;
}

bool CNetPBMWriter::SaveAsPPM(k3dIBitmap* Bitmap, SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_fail(Bitmap, false);
	g_return_val_if_string_fail(FilePath, false);

	// Get data from the bitmap ...
	k3dPixel* data = Bitmap->GetData();
	gulong width = Bitmap->GetWidth();
	gulong height = Bitmap->GetHeight();

	g_return_val_if_fail(data, false);

	// Create the output stream ...
	std::ofstream file(FilePath);
	g_return_val_if_fail(file.good(), false);
	
	return SaveAsPPM(data, width, height, file);
}

bool CNetPBMWriter::SaveAsPPM(k3dPixel* Data, gulong Width, gulong Height, std::ostream& Stream)
{
	// Write our magic headers ...
	Stream << "P6" << std::endl;
	Stream << Width << " " << Height << std::endl;
	Stream << "255" << std::endl;
	
	// Write data ...	
	for(gulong y = 0; y < Height; y++)
		{
			for(gulong x = 0; x < Width; x++)
				{
					Stream << Data->Red << Data->Green << Data->Blue;
					Data++;
				}
		}

	return true;
}

