#ifndef LIGHT_H
#define LIGHT_H

// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Defines the CLight K-3D object, which encapsulates a light source within the 3D workspace
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include <k3dtransformobject.h>
#include <k3ddialog.h>
#include <k3dchannelbutton.h>

#include <k3dilight.h>
#include <k3dieditorlight.h>
#include <k3dirilight.h>
#include <k3dlocationserver.h>
#include <k3dshaderinstance.h>
#include <k3dshaderselector.h>
#include <k3dshaderargumentspage.h>


#include <k3dvariable.h>

/////////////////////////////////////////////////////////////////////////////
// CLightDialog

/// Encapsulates the Light UI
class CLightDialog : public k3dDialog
{
friend class CLight;
public:
	CLightDialog();

	/// Called to initialize the UI
	bool Create(CLight* Object);

protected:
	/// Stores a pointer to the owning object
	CLight* m_Object;

	/// Handles our custom events
	void OnEvent(sdpGtkEvent* Event);

	/// Called when the user toggles the "cast shadows" button
	void OnCastShadows();

	/// Called to initialize controls
	bool InitializeControls();
	/// Called to update the state of any UI controls
	void UpdateControls();

	/// Provides the UI for selecting shaders
	k3dShaderSelector m_ShaderSelector;
	/// Provides the UI for controlling shader arguments
	k3dShaderArgumentsPage m_ShaderPage;
};

///////////////////////////////////////////////////////////////////////////////
// CLight

class CLight : public k3dTransformObject,
							 public k3dILight,
							 public k3dIEditorLight,
							 public k3dIRiLight,
							 public k3dLocationServer
{
friend class CLightDialog;
public:
	CLight();
	~CLight();

	// k3dIUnknown overrides
	sdpResult QueryInterface(const sdpIID& InterfaceID, void** Interface);
	gulong AddRef();
	gulong Release();

	// k3dIObject overrides
	sdpResult SetObjectContext(k3dIObjectContext* Context);
	sdpResult EditObject();
	sdpResult ObjectRenamed(SDPCSTRING NewName);
	sdpResult ObjectDeleted();

	// k3dIEventClient overrides
	sdpResult OnPosition(const k3dID ObjectID);
	sdpResult OnTrack(const k3dID ObjectID);
	sdpResult OnSelectionChanged(const k3dID ObjectID);

	// k3dIEditorRender overrides
	sdpResult EditorPushRender(k3dIEditorEngine* Engine);
	sdpResult EditorRender(k3dIEditorEngine* Engine);

	// k3dIEditorLight overrides
	sdpResult EditorLightPreRender(k3dIEditorEngine* Engine);
	sdpResult EditorLightSetup(k3dIEditorEngine* Engine, gulong Light);
	sdpResult Setup(gulong Light);

	// k3dIRiLight overrides
	sdpResult RiLightStartJob(k3dIRenderJob* Job, k3dIRiEngine* Engine);
	sdpResult RiLightPreRender(k3dIRenderFrame* Frame, k3dIRiEngine* Engine);
	sdpResult RiLightSetup(k3dIRenderFrame* Frame, k3dIRiEngine* Engine);
	sdpResult RiLightFinishJob(k3dIRenderJob* Job, k3dIRiEngine* Engine);

	// k3dIPersist overrides
	sdpResult InitNew();
	sdpResult LoadComplete();
	sdpResult Save(sdpXmlIElement* Element, k3dIDependencies* Dependencies);
	sdpResult Load(sdpXmlIElement* Element);

protected:
	void GetShaderLightType();
	void GetShaderLightIntensity();
	void GetShaderLightColor();

	void EditorRenderAmbient();
	void EditorRenderPoint();
	void EditorRenderDistant();
	void EditorRenderSpotlight();

	typedef enum
	{
		AMBIENT = 1,
		POINT = 2,
		DISTANT = 3,
		SPOTLIGHT = 4,
	} LightType;

	static SDPCSTRING const Key_LightType;
	static SDPCSTRING const Key_Ambient;
	static SDPCSTRING const Key_Point;
	static SDPCSTRING const Key_Distant;
	static SDPCSTRING const Key_Spotlight;
	static SDPCSTRING const Key_LightIntensityParameter;
	static SDPCSTRING const Key_LightColorParameter;

	class VariableLightType : public k3dCustomVariable<LightType>
	{
	protected:
		sdpString ConvertToString() { return sdpFormatString("%d", m_Value); }
		bool ConvertFromString(SDPCSTRING String) { m_Value = LightType(atoi(String)); return true; }
	};

	VariableLightType m_Type;
	k3dVariable<bool> m_CastShadows;
//	k3dScalarChannel m_Target;
	k3dScalarChannel m_Intensity;
	k3dColorChannel m_Color;
	k3dShaderInstance m_Shader;
	
	CLightDialog m_UserInterface;
};


#endif // LIGHT_H
