#ifndef MATERIAL_H
#define MATERIAL_H

// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the Material K-3D object, which controls the surface appearance of rendered geometry
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dpersistobject.h>

#include <k3dimaterial.h>
#include <k3dieditormaterial.h>
#include <k3dirimaterial.h>

#include <k3ditexture.h>
#include <k3dieditortexture.h>
#include <k3diritexture.h>

#include <k3dshaderinstance.h>
#include <k3dshaderselector.h>
#include <k3dshaderargumentspage.h>

#include <k3ddialog.h>
#include <sdpgtkevent.h>

#include <sdpgl.h>

// Forward class declarations
class CMaterial;

/////////////////////////////////////////////////////////////////////////////
// CMaterialDialog

class CMaterialDialog : public k3dDialog
{
public:
	CMaterialDialog();
	~CMaterialDialog();

	bool Create(CMaterial* Object);

protected:
	CMaterial* m_Object;

	/// Provides the UI for selecting  the surface shader
	k3dShaderSelector m_SurfaceSelector;
	/// Provides the UI for controlling the surface shader parameters
	k3dShaderArgumentsPage m_SurfacePage;
	/// Provides the UI for selecting a displacement shader
	k3dShaderSelector m_DisplacementSelector;
	/// Provides the UI for controlling the displacement shader parameters
	k3dShaderArgumentsPage m_DisplacementPage;

	// Event handling ...
	void OnEvent(sdpGtkEvent* Event);

	void OnShadowType();
	void OnTrueDisplacement();
	void OnMatte();
	void OnCameraRays();
	void OnReflectionRays();
	void OnApply();
	void OnRemove();
	void OnSelect();

	// Update controls ...
	bool InitializeControls();
	void UpdateControls();
};


//////////////////////////////////////////////////////////////////////
// CMaterial

class CMaterial :	public k3dPersistObject,
									public k3dIMaterial,
									public k3dIEditorMaterial,
									public k3dIRiMaterial
{
friend class CMaterialDialog;
public:
	CMaterial();
	~CMaterial();

	CMaterialDialog m_UserInterface;

	// k3dIUnknown overrides
	sdpResult QueryInterface(const sdpIID& InterfaceID, void** Interface);
	gulong AddRef();
	gulong Release();

	// k3dIObject overrides
	sdpResult SetObjectContext(k3dIObjectContext* Context);
	sdpResult EditObject();
	sdpResult ObjectRenamed(SDPCSTRING NewName);
	sdpResult ObjectDeleted();

	// k3dIMaterial overrides
	sdpResult SetAmbient(double Color[3]);
	sdpResult SetDiffuse(double Color[3]);
	sdpResult SetSpecular(double Color[3]);
	sdpResult SetEmissive(double Color[3]);
	sdpResult SetSpecularSize(double Size); 
	sdpResult SetTransparency(double Transparency);

	// k3dIEditorMaterial overrides
	sdpResult EditorMaterialPreRender(k3dIEditorEngine* Engine);
	sdpResult EditorMaterialSetup(k3dIEditorEngine* Engine);

	// k3dIRiMaterial overrides
	sdpResult RiMaterialStartJob(k3dIRenderJob* Job, k3dIRiEngine* Engine);
	sdpResult RiMaterialPreRender(k3dIRenderFrame* Frame, k3dIRiEngine* Engine);
	sdpResult RiMaterialSetup(k3dIRenderFrame* frame, k3dIRiEngine* Engine);
	sdpResult RiMaterialFinishJob(k3dIRenderJob* Job, k3dIRiEngine* Engine);

	// k3dIAnimate overrides
	sdpResult PostAnimate();

	// k3dIPersist overrides
	sdpResult InitNew();
	sdpResult Load(sdpXmlIElement* Element);
	sdpResult LoadComplete();
	sdpResult Save(sdpXmlIElement* Element, k3dIDependencies* Dependencies);

	// k3dISelectorClient overrides
	sdpResult OnSelectionChanged(const k3dID ObjectID);

protected:
	typedef enum
	{
		ST_NONE = 1,
		ST_OPAQUE = 2,
		ST_OS = 3,
		ST_SHADE = 4,
	} ShadowType;

	class VariableShadowType : public k3dCustomVariable<ShadowType>
	{
	protected:
		sdpString ConvertToString() { return sdpFormatString("%d", m_Value); }
		bool ConvertFromString(SDPCSTRING String) { m_Value = ShadowType(atoi(String)); return true; }
	};

	k3dShaderInstance m_SurfaceShader;
	k3dShaderInstance m_DisplacementShader;

	VariableShadowType m_ShadowType;
	k3dVariable<bool> m_TrueDisplacement;
	k3dVariable<bool> m_Matte;
	k3dVariable<bool> m_CameraRays;
	k3dVariable<bool> m_ReflectionRays;

	k3dColorChannel m_Color;
	k3dScalarChannel m_Opacity;
	k3dColorChannel m_LuminousColor;
	k3dScalarChannel m_Luminosity;

	GLfloat m_AmbientData[4];
	GLfloat m_DiffuseData[4];
	GLfloat m_SpecularData[4];
	GLfloat m_EmissiveData[4];
	static GLfloat m_BlackData[4];
	static GLfloat m_WhiteData[4];

	k3dIEditorTexture* m_EditorTexture;
	k3dIRiTexture* m_RiTexture;

	sdpResult RiMaterialDefaultsSetup(k3dIRiEngine* Engine);
};

#endif // MATERIAL_H
