// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CCamera K-3D object, which controls the position, orientation, and display of a CCameraFrame within the 3D workspace
		\author Tim Shead (tshead@k-3d.com)
*/

#include "camera.h"

#include <k3dipoint.h>
#include <k3dipolygon.h>
#include <k3dipolypoint.h>
#include <k3dipolypointserver.h>
#include <k3direnderfarm.h>
#include <k3direnderjob.h>
#include <k3direnderframe.h>
#include <k3diactions.h>

#include <k3dmisc.h>
#include <k3dboundingbox.h>

#include <sdpgtkevent.h>

#include <sdpfile.h>

#include <gdk/gdkprivate.h>

#if defined SDPWIN32
#include <win32/gdkwin32.h>
#else // SDPWIN32
#include <GL/glx.h>
#endif // !SDPWIN32

///////////////////////////////////////////////////////////
// CFrustum

CFrustum::CFrustum()
{
}

CFrustum::CFrustum(gulong ViewportWidth, gulong ViewportHeight, double FocalLength)
{
	SetParameters(ViewportWidth, ViewportHeight, FocalLength);
}

void CFrustum::SetParameters(gulong ViewportWidth, gulong ViewportHeight, double FocalLength)
{
	// Stored parameters ...
	m_ViewportWidth = ViewportWidth;
	m_ViewportHeight = ViewportHeight;
	m_FocalLength = FocalLength;

	// Calculated parameters ...
	m_FrustumNear = m_FocalLength / 22.0;
	m_FrustumFar = 10000.0;
	m_FrustumLeft = -1.0;
	m_FrustumRight = 1.0;
	m_FrustumTop = -(double(ViewportHeight) / double(ViewportWidth));
	m_FrustumBottom = double(ViewportHeight) / double(ViewportWidth);

	m_HalfWidth = 0.5 * double(m_ViewportWidth);
	m_HalfHeight = 0.5 * double(m_ViewportHeight);

	m_InverseWidth = 1.0 / m_FrustumRight;
	m_InverseHeight = 1.0 / m_FrustumBottom;
}

void CFrustum::CameraToScreen(sdpVector3& Coords)
{
	// Point is at the camera position, so avoid divide-by-zero ...
	if(!Coords[2])
		return;

	double screenx = m_InverseWidth * ((Coords[0] * m_FrustumNear) / Coords[2]);
	double screeny = m_InverseHeight * ((Coords[1] * m_FrustumNear) / Coords[2]);

	Coords[0] = (screenx * m_HalfWidth) + m_HalfWidth;
	Coords[1] = m_HalfHeight - (screeny * m_HalfHeight);
	Coords[2] = 0.0;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////
// CCamera

gdouble CCamera::m_EditorBackgroundColor[4] = {0.8, 0.8, 0.8, 1.0};
gdouble CCamera::m_EditorNormalColor[4] = {0.3, 0.3, 0.1, 1.0};
gdouble CCamera::m_EditorSelectorColor[4] = {0.215, 0.215, 0.8, 1.0};
gdouble CCamera::m_EditorSelectedColor[4] = {1.0, 1.0, 1.0, 1.0};
gdouble CCamera::m_EditorUnselectedColor[4] = {0.0, 0.0, 0.0, 1.0};
gdouble CCamera::m_EditorControlPointColor[4] = {0.0, 0.0, 1.0, 1.0};
gdouble CCamera::m_EditorFirstColor[4] = {1.0, 0.0, 1.0, 1.0};
gdouble CCamera::m_EditorActiveColor[4] = {1.0, 0.0, 0.0, 1.0};

GLfloat CCamera::m_PointSize = 4.0f;
gulong CCamera::m_ClickSensitivity = 4;
gdouble CCamera::m_PositionSensitivity = -1.0;
gdouble CCamera::m_OrientationSensitivity = 1.0;
gdouble CCamera::m_ZoomSign = -1.0;

bool CCamera::m_BoxSelectPartial = false;
bool CCamera::m_LassoSelectPartial = false;

CCamera::CCamera()
{
	m_TimeBase = 0;

	// Camera options ...
	AddVariable(&m_DisplayMode("displaymode", EDITOR));
	AddVariable(&m_RenderMode("rendermode", RI));
	AddVariable(&m_CameraAutoUpdate("cameraautoupdate", true));
	AddVariable(&m_CameraLockSize("cameralocksize", false));
	AddVariable(&m_CameraEnableInteractive("cameraenableinteractive", true));
	AddVariable(&m_CameraIsometric("cameraisometric", false));
	AddVariable(&m_CameraField("camerafield", EVENFIELDFIRST));

	// Editor options ...
	AddVariable(&m_EditorSolid("editorsolid", true));
	AddVariable(&m_EditorFog("editorfog", false));
	AddVariable(&m_EditorSmoothing("editorsmoothing", true));
	AddVariable(&m_EditorTexture("editortexture", false));
	AddVariable(&m_EditorTwoSided("editortwosided", false));
	AddVariable(&m_EditorOrientation("editororientation", true));
	AddVariable(&m_EditorNormals("editornormals", false));
	AddVariable(&m_EditorWireframePatchDivisions("editorwireframepatchdivisions", 4));
	AddVariable(&m_EditorSolidPatchDivisions("editorsolidpatchdivisions", 12));

	// Ri-renderer options ...
	AddVariable(&m_RiRenderEngine("riengine", "bmrt"));
	AddVariable(&m_RiXSamples("rixsamples", 1));
	AddVariable(&m_RiYSamples("riysamples", 1));
	AddVariable(&m_RiPixelFilterName("ripixelfiltername", k3dIRiEngine::RI_GAUSSIAN));
	AddVariable(&m_RiPixelFilterWidth("ripixelfilterwidth", 2.0));
	AddVariable(&m_RiPixelFilterHeight("ripixelfilterheight", 2.0));
	AddVariable(&m_RiAtmosphere("riatmosphere", false));
	AddVariable(&m_RiDepthOfField("ridepthoffield", false));
	AddVariable(&m_RiMotionBlur("rimotionblur", false));
	AddVariable(&m_RiFieldRender("rifieldrender", false));
	AddVariable(&m_RiRadiosity("riradiosity", false));
	AddVariable(&m_RiRadiositySteps("riradiositysteps", 100));
	AddVariable(&m_RiRadiosityPatchSamples("riradiositypatchsamples", 1));
	AddVariable(&m_RiRadiosityPatchSize("riradiositypatchsize", 10));
	AddVariable(&m_RiRadiosityElementSize("riradiosityelementsize", 5));
	AddVariable(&m_RiRadiosityMinimumSize("riradiosityminimumsize", 0.5));

	// Owned objects ...
	AddObject(&m_Focallength("focallength"));
	AddObject(&m_Focalplane("focalplane"));
	AddObject(&m_Fstop("fstop"));
	AddObject(&m_Shutteropen("shutteropen"));
	AddObject(&m_Shutterclose("shutterclose"));
	AddObject(&m_Exposure("exposure"));
	AddObject(&m_Gamma("gamma"));
	AddObject(&m_Xdirection("xdirection"));
	AddObject(&m_Ydirection("ydirection"));
	AddObject(&m_Zdirection("zdirection"));
//	AddObject(&m_Fogstart("fogstart"));
//	AddObject(&m_Fogend("fogend"));
	AddObject(&m_Farplane("farplane"));
//	AddObject(&m_Fogcolor("fogcolor"));
//	AddObject(&m_BackgroundColor("backgroundcolor"));
//	AddObject(&m_Target("target", sdpObjectSelector, IID_k3dIObjectSelector));
//	AddObject(&m_Atmosphere("atmosphere", sdpShaderSelector, IID_k3dIShaderSelector));

/*
	SetNumLocations(1);

	// Initialize cursors ...
	if(!m_TranslateCursor) m_TranslateCursor = ::LoadCursor(theApp.m_hInstance, MAKEINTRESOURCE(IDC_TRANSLATECURSOR));
	if(!m_SelectCursor) m_SelectCursor = ::LoadCursor(theApp.m_hInstance, MAKEINTRESOURCE(IDC_SELECTCURSOR));
	if(!m_PanCursor) m_PanCursor = ::LoadCursor(theApp.m_hInstance, MAKEINTRESOURCE(IDC_PANCURSOR));
	if(!m_DollyCursor) m_DollyCursor = ::LoadCursor(theApp.m_hInstance, MAKEINTRESOURCE(IDC_DOLLYCURSOR));
	if(!m_ZoomCursor) m_ZoomCursor = ::LoadCursor(theApp.m_hInstance, MAKEINTRESOURCE(IDC_ZOOMCURSOR));
	if(!m_TiltCursor) m_TiltCursor = ::LoadCursor(theApp.m_hInstance, MAKEINTRESOURCE(IDC_TILTCURSOR));
*/
}

CCamera::~CCamera()
{
//	std::cout << "Destroying CCamera" << std::endl;

	// Release interfaces ...
	if(m_TimeBase)
		m_TimeBase->Release();
}

bool CCamera::Create()
{
	g_return_val_if_fail(k3dTransformObject::Create(), false);

	m_TimeBase = sdpQueryInterface<k3dITimeBase>(Document());
	g_return_val_if_fail(m_TimeBase, false);

	return true;
}

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CCamera::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIAnimate>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dICamera>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dIEditorEngine>(&m_EditorEngine, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dIEditorRender>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dIEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dILocationServer>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dIObject>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dIPersist>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dIRiEngine>(&m_RiEngine, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dITarget>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dITimeServer>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dITransform>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dIUserEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIObject*>(this), InterfaceID, Interface)) return SDP_S_OK;											

	return SDP_E_NO_INTERFACE;
}

gulong CCamera::AddRef()
{
	return sdpComAddRef(m_ReferenceCount);
}

gulong CCamera::Release()
{
	return sdpComRelease(this, m_ReferenceCount);
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CCamera::SetObjectContext(k3dIObjectContext* Context)
{
	g_return_val_if_fail(sdpSucceeded(k3dTransformObject::SetObjectContext(Context)), SDP_E_FAIL);
	
	g_return_val_if_fail(m_RiAtmosphereShader.Create(Document()), SDP_E_FAIL);
	m_RiAtmosphereShader.SetShader("k3d_null");

	return SDP_S_OK;
}

sdpResult CCamera::EditObject()
{
	g_return_val_if_fail(m_UserInterface.Create(this), SDP_E_FAIL);
	return SDP_S_OK;
}

sdpResult CCamera::QueryDeleteObject()
{
	return SDP_S_FALSE;
}

sdpResult CCamera::ObjectRenamed(SDPCSTRING NewName)
{
	m_UserInterface.ObjectRenamed(NewName);
	return k3dTransformObject::ObjectRenamed(NewName);
}

sdpResult CCamera::ObjectDeleted()
{
	m_UserInterface.ObjectDeleted();
	m_Frame.ObjectDeleted();
	
	return k3dTransformObject::ObjectDeleted();
}

///////////////////////////////////////////////////////////////
// k3dIAnimate overrides

sdpResult CCamera::PostAnimate()
{
	SetupViewVolume();
	
	return k3dTransformObject::PostAnimate();
}

///////////////////////////////////////////////////////////////
// k3dIPersist overrides

sdpResult CCamera::InitNew()
{
	g_return_val_if_fail(sdpSucceeded(k3dTransformObject::InitNew()), SDP_E_FAIL);

	g_return_val_if_fail(m_Focallength.Create(Document(), "Camera focal length", 50.0, true, 0.0, 5000.0, "Time", "Lens Focal Length", k3dIChannel::DATA_SCALAR, true, this), SDP_E_FAIL);
	g_return_val_if_fail(m_Focalplane.Create(Document(), "Camera focus distance", 35.0, true, 0, ULONG_MAX, "Time", "Lens Focal Plane", k3dIChannel::DATA_DISTANCE, true, this), SDP_E_FAIL);
	g_return_val_if_fail(m_Fstop.Create(Document(), "Camera f-stop", 0.3, true, 0.01, DBL_MAX, "Time", "Lens f-stop", k3dIChannel::DATA_SCALAR, true, this), SDP_E_FAIL);
	g_return_val_if_fail(m_Shutteropen.Create(Document(), "Camera shutter-opened", 0.0, true, 0.0, 1.0, "Time", "Shutter Open", k3dIChannel::DATA_SCALAR, true, this), SDP_E_FAIL);
	g_return_val_if_fail(m_Shutterclose.Create(Document(), "Camera shutter-closed", 0.0, true, 0.0, 1.0, "Time", "Shutter Closed", k3dIChannel::DATA_SCALAR, true, this), SDP_E_FAIL);
	g_return_val_if_fail(m_Exposure.Create(Document(), "Camera exposure", 1.0, true, -DBL_MAX, DBL_MAX, "Time", "Exposure", k3dIChannel::DATA_SCALAR, true, this), SDP_E_FAIL);
	g_return_val_if_fail(m_Gamma.Create(Document(), "Camera gamma", 1.0, true, -DBL_MAX, DBL_MAX, "Time", "Gamma", k3dIChannel::DATA_SCALAR, true, this), SDP_E_FAIL);
	g_return_val_if_fail(m_Xdirection.Create(Document(), "Camera X direction", 0.0, true, -1.0 * ULONG_MAX, ULONG_MAX, "Time", "Camera X Direction", k3dIChannel::DATA_ANGLE, true, this), SDP_E_FAIL);
	g_return_val_if_fail(m_Ydirection.Create(Document(), "Camera Y direction", 0.0, true, -1.0 * ULONG_MAX, ULONG_MAX, "Time", "Camera Y Direction", k3dIChannel::DATA_ANGLE, true, this), SDP_E_FAIL);
	g_return_val_if_fail(m_Zdirection.Create(Document(), "Camera Z direction", 0.0, true, -1.0 * ULONG_MAX, ULONG_MAX, "Time", "Camera Z Direction", k3dIChannel::DATA_ANGLE, true, this), SDP_E_FAIL);
	g_return_val_if_fail(m_Farplane.Create(Document(), "Camera farplane", 1000000.0, true, -1.0 * ULONG_MAX, ULONG_MAX, "Time", "Camera Farplane Distance", k3dIChannel::DATA_DISTANCE, true, this), SDP_E_FAIL);

	sdpXmlElementArray defaultengines = sdpXmlXPathSelect(Application()->Options()->GetRootElement(), "application/defaultengine");
	g_assert(1 == defaultengines.size());
	sdpString defaultengine;
	sdpXmlGetAttribute(defaultengines[0], "engine", defaultengine);
	m_RiRenderEngine() = defaultengine;

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dICamera overrides

sdpResult CCamera::ShowViewport()
{
	// If it's already open, we're done ...
	if(m_Frame.Root())
		return SDP_S_OK;

	// Setup our rendering engines ...
	g_return_val_if_fail(sdpSucceeded(m_EditorEngine.SetCamera(this)), SDP_E_FAIL);
	g_return_val_if_fail(sdpSucceeded(m_RiEngine.SetCamera(this)), SDP_E_FAIL);
	
	g_return_val_if_fail(m_Frame.Create(this), SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CCamera::HideViewport()
{
	// If it's not already open, we're done ...
	if(!m_Frame.Root())
		return SDP_S_OK;
		
	m_Frame.RootWidget().Destroy();	
		
	return SDP_S_OK;
}

gulong CCamera::Width()
{
	sdpGtkOpenGLDrawingArea& viewport = Viewport();
	if(viewport.Attached())
		return viewport.Width();

	return 0;
}

gulong CCamera::Height()
{
	sdpGtkOpenGLDrawingArea& viewport = Viewport();
	if(viewport.Attached())
		return viewport.Height();

	return 0;
}

sdpResult CCamera::Redraw(bool Synchronous)
{
	if(Synchronous)
		return Redraw();
	
	if(Viewport().Attached())
		Viewport().QueueDraw();	
	
	return SDP_S_OK;
}

sdpResult CCamera::Redraw()
{
	switch(m_DisplayMode())
		{
			case EDITOR:
				return m_EditorEngine.Redraw();
			case RI:
				return m_RiEngine.Redraw();
		}
		
	// Unknown render mode!
	g_return_val_if_fail(0, SDP_E_FAIL);
}

sdpGtkOpenGLDrawingArea& CCamera::Viewport()
{
	return m_Frame.Viewport();
}

sdpResult CCamera::StartJob(k3dIRenderJob* Job)
{
	// Sanity checks ...
	g_return_val_if_fail(Job, SDP_E_FAIL);

	switch(m_RenderMode())
		{
			case EDITOR:
				return m_EditorEngine.StartJob(Job);
			case RI:
				return m_RiEngine.StartJob(Job);
			default:
				g_assert_not_reached();	// Unknown render mode!
		}

	return SDP_E_FAIL;
}

sdpResult CCamera::QueueFrame(k3dIRenderFrame* Frame)
{
	// Sanity checks ...
	g_return_val_if_fail(Frame, SDP_E_FAIL);

	switch(m_RenderMode())
		{
			case EDITOR:
				return m_EditorEngine.QueueFrame(Frame);
			case RI:
				return m_RiEngine.QueueFrame(Frame);
			default:
				g_assert_not_reached();	// Unknown render mode!
		}

	return SDP_E_FAIL;
}

sdpResult CCamera::FinishJob(k3dIRenderJob* Job)
{
	// Sanity checks ...
	g_return_val_if_fail(Job, SDP_E_FAIL);

	switch(m_RenderMode())
		{
			case EDITOR:
				return m_EditorEngine.FinishJob(Job);
			case RI:
				return m_RiEngine.FinishJob(Job);
			default:
				g_assert_not_reached();	// Unknown render mode!
		}

	return SDP_E_FAIL;
}

void CCamera::SetupViewVolume()
{
	// Store camera position ...
	m_CurrentPosition = m_Position[k3dITransform::ANIMATED_FRAME];
	m_CurrentOrientation = m_Orientation[k3dITransform::ANIMATED_FRAME];

	// Calculate camera focal length / near / far / focal plane / depth-of-field amplitude
	m_CurrentFocalLength = m_Focallength->GetValue1(m_CurrentTime, m_DeltaTime);
	m_CurrentNearPlane = m_CurrentFocalLength / 22.0;
	m_CurrentFarPlane = m_Farplane->GetValue1(m_CurrentTime, m_DeltaTime);
	m_CurrentFocalPlane = m_Focalplane->GetValue1(m_CurrentTime, m_DeltaTime);
	m_CurrentFStop = m_Fstop->GetValue1(m_CurrentTime, m_DeltaTime);

	// Calculate the width & height of the view frustum / paralleleped
	gulong width = Width();
	gulong height = Height();

	if(m_CameraIsometric())
		{
			m_CurrentWidthPlane = 1250.0 / m_CurrentFocalLength;
			m_CurrentHeightPlane = m_CurrentWidthPlane * ((double)height / (double)width);
		}
	else
		{
			m_CurrentWidthPlane = 1.0;
			m_CurrentHeightPlane = (double)height / (double)width;
		}

	// Store our world-to-camera transform matrix
	sdpQuaternion q(m_CurrentOrientation);
	sdpEulerAngles a(q, sdpEulerAngles::ZXYstatic);

//std::cout << "position: " << sdpString(m_CurrentPosition) << std::endl;
//std::cout << "orientation: " << sdpString(a) << std::endl << std::endl;

	m_CurrentWorldToCameraMatrix = sdpIdentity3D();
	m_CurrentWorldToCameraMatrix = m_CurrentWorldToCameraMatrix * sdpRotation3D(-a[0], sdpVector3(0.0, 0.0, 1.0));
	m_CurrentWorldToCameraMatrix = m_CurrentWorldToCameraMatrix * sdpRotation3D(-a[1], sdpVector3(1.0, 0.0, 0.0));
	m_CurrentWorldToCameraMatrix = m_CurrentWorldToCameraMatrix * sdpRotation3D(-a[2], sdpVector3(0.0, 1.0, 0.0));
	m_CurrentWorldToCameraMatrix = m_CurrentWorldToCameraMatrix * sdpTranslation3D(sdpVector3(-m_CurrentPosition[0], -m_CurrentPosition[1], -m_CurrentPosition[2]));

//	m_CurrentWorldToCameraMatrix = k3dGetWorldToCameraMatrix(this);
}


// Functor for calculating camera coordinates ...
class CCameraSnapshotCameraCoordinates
{
public:
	CCameraSnapshotCameraCoordinates(const sdpMatrix4& CameraMatrix) { Matrix = CameraMatrix; }

	bool HandleObject(k3dIObjectContext* Context)
	{
		// Skip objects that don't support locations ...
		if(SDP_S_OK != Context->Object()->QueryInterface(k3dILocationServer::iid, (void**)&LocationServer))
			return true;

		// For each location ...
		for(LocationHandle = LocationServer->GetFirstLocation(); LocationHandle; LocationHandle = LocationServer->GetNextLocation(LocationHandle))
			{
				Location = LocationServer->GetLocation(LocationHandle);
				Location->GetWorldXYZ(Coords);
				Coords = Matrix * Coords;
				Location->SetCameraXYZ(Coords);
			}

		// Cleanup ...
		LocationServer->Release();

		return true;
	}

	k3dILocationServer* LocationServer;
	K3DLOCATIONHANDLE LocationHandle;
	k3dILocation* Location;
	sdpMatrix4 Matrix;
	sdpVector3 Coords;
};

// Calculate camera coordinates ...
sdpResult CCamera::SnapshotCameraCoordinates()
{
	// Calculate world coordinates ...
	Document()->SnapshotWorldCoordinates();

	// Transform objects ...
	CCameraSnapshotCameraCoordinates functor(m_CurrentWorldToCameraMatrix);
	k3dForEachObject(Document(), 0, functor);

	return SDP_S_OK;
}

// Functor for calculating screen coordinates ...
class CCameraSnapshotScreenCoordinates
{
public:
	CCameraSnapshotScreenCoordinates(const CFrustum& CameraFrustum)	{ Frustum = CameraFrustum; }

	bool HandleObject(k3dIObjectContext* Context)
	{
		// Skip objects that don't support locations ...
		if(SDP_S_OK != Context->Object()->QueryInterface(k3dILocationServer::iid, (void**)&LocationServer))
			return true;

		// For each location ...
		for(LocationHandle = LocationServer->GetFirstLocation(); LocationHandle; LocationHandle = LocationServer->GetNextLocation(LocationHandle))
			{
				Location = LocationServer->GetLocation(LocationHandle);

				Location->GetCameraXYZ(Coords);
				Frustum.CameraToScreen(Coords);
				Location->SetScreenXYZ(Coords);
			}

		// Cleanup ...
		LocationServer->Release();

		return true;
	}

	k3dILocationServer* LocationServer;
	K3DLOCATIONHANDLE LocationHandle;
	k3dILocation* Location;
	sdpVector3 Coords;
	double ScreenX;
	double ScreenY;
//	double ScreenZ;

	CFrustum Frustum;
};

// Calculate screen coordinates ...
sdpResult CCamera::SnapshotScreenCoordinates()
{
	// Calculate camera coordinates
	SnapshotCameraCoordinates();

	// Housekeeping ...
	CFrustum frustum(Width(), Height(), m_CurrentFocalLength);
	CCameraSnapshotScreenCoordinates functor(frustum);
	k3dForEachObject(Document(), 0, functor);

	return SDP_S_OK;
}

sdpResult CCamera::GetViewingFrustum(gdouble* Left, gdouble* Right, gdouble* Top, gdouble* Bottom, gdouble* Near, gdouble* Far)
{
	// Make sure the user requested at least one parameter ...
	g_return_val_if_fail(Left || Right || Top || Bottom || Near || Far, SDP_E_FAIL);

	// Return values ...
	if(Left)
		*Left = -m_CurrentWidthPlane;
	if(Right)
		*Right = m_CurrentWidthPlane;
	if(Top)
		*Top = m_CurrentHeightPlane;
	if(Bottom)
		*Bottom = -m_CurrentHeightPlane;
	if(Near)
		*Near = m_CurrentNearPlane;
	if(Far)
		*Far = m_CurrentFarPlane;

	return SDP_S_OK;
}

sdpResult CCamera::GetWorldToCameraMatrix(gdouble Matrix[16])
{
	m_CurrentWorldToCameraMatrix.CopyArray(Matrix);
	return SDP_S_OK;
}

sdpResult CCamera::GetWorldPosition(gdouble Coords[3])
{
	m_CurrentPosition.CopyArray(Coords);

	return SDP_S_OK;
}

sdpResult CCamera::GetWorldOrientation(gdouble Angles[3])
{
	m_CurrentOrientation.CopyArray(Angles);

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIUserEventClient overrides

/// Functor for selecting points
class CCameraSelectSinglePoint
{
public:
	CCameraSelectSinglePoint(sdpVector2 Mouse, bool SelectMode)
	{
		m_Mouse = Mouse;
		m_SelectMode = SelectMode;
		m_Selected = false;
	}

	bool Selected() { return m_Selected; }

	bool HandlePoint(k3dIPoint* Point)
	{
		sdpVector3 screencoords;
		Point->Location()->GetScreenXYZ(screencoords);

		if((fabs(screencoords[0] - m_Mouse[0]) < 4.0) &&
			 (fabs(screencoords[1] - m_Mouse[1]) < 4.0))
			{
				Point->State()->SetState(k3dIState::SELECTED, m_SelectMode);
				m_Selected = true;

				return false;
			}
/*
				if(selectmode)
					Document()->SetActivePoint(point);
				else
					if(point == activepoint)
						Document()->SetActivePoint(NULL);
*/

		return true;
	}

	sdpVector2 m_Mouse;
	bool m_SelectMode;
	bool m_Selected;
};

bool CCamera::SelectSinglePoint(GdkModifierType Modifiers, sdpVector2 CurrentMouse)
{
	Document()->Actions()->StartAction("Select Point");

	bool selectmode = ControlDown(Modifiers) ? false : true;
	CCameraSelectSinglePoint functor(CurrentMouse, selectmode);
	k3dForEachPoint(Document(), k3dIState::VISIBLESELECTABLE, functor);

	Document()->Actions()->FinishAction();

	return functor.Selected();
}

/// Functor for selecting polygons
class CCameraSelectSinglePolygon
{
public:
	CCameraSelectSinglePolygon(sdpVector2 Mouse, bool SelectMode)
	{
		m_Mouse = Mouse;
		m_SelectMode = SelectMode;
		m_Selected = false;
	}

	bool Selected() { return m_Selected; }

	bool HandlePolygon(k3dIPolygon* Polygon)
	{
		// If the mouse isn't inside the polygon, skip it ...
		if(!k3dIsInside(m_Mouse, Polygon))
			return true;

		// Select the polygon and make it active ...
		Polygon->State()->SetState(k3dIState::SELECTED, m_SelectMode);
//		Document()->SetActivePolygon(selectmode ? polygon : 0);

		// Decide which point to make active ...
//		Document()->SetActivePoint(selectmode ? CurrentMouse.Closest(polygon)->Point() : 0);

		m_Selected = true;

		return false;
	}

	sdpVector2 m_Mouse;
	bool m_SelectMode;
	bool m_Selected;
};

bool CCamera::SelectSinglePolygon(GdkModifierType Modifiers, sdpVector2 CurrentMouse)
{
	Document()->Actions()->StartAction("Select Polygon");

	bool selectmode = ControlDown(Modifiers) ? false : true;
	CCameraSelectSinglePolygon functor(CurrentMouse, selectmode);
	k3dForEachPolygon(Document(), k3dIState::VISIBLESELECTABLE, functor);

	Document()->Actions()->FinishAction();

	return functor.Selected();
}

/// Functor for selecting meshes
class CCameraSelectSingleMesh
{
public:
	CCameraSelectSingleMesh(k3dIDocument* Document, sdpVector2 Mouse, bool SelectMode)
	{
		m_Document = Document;
		m_Mouse = Mouse;
		m_SelectMode = SelectMode;
		m_Selected = false;
	}

	bool Selected() { return m_Selected; }

	bool HandleMesh(k3dIMesh* Mesh)
	{
		// Calculate the mesh's screen bounding box ...
		k3dBoundingBox2 boundingbox;
		boundingbox.Init();

		sdpVector3 screencoords;
		k3dIPointServer* points = Mesh->Points();
		for(K3DPOINTHANDLE pointhandle = points->GetFirstPoint(); pointhandle; pointhandle = points->GetNextPoint(pointhandle))
			{
				points->GetPoint(pointhandle)->Location()->GetScreenXYZ(screencoords);
				boundingbox.Expand(sdpVector2(screencoords, 2));
			}

		// If the mouse click is outside the screen bounding box we're done ...
		if(!boundingbox.Contains(m_Mouse))
			return true;

		// For each polygon in the mesh ...
		k3dIPolygonServer* polygons = Mesh->Polygons();
		for(K3DPOLYGONHANDLE polyhandle = polygons->GetFirstPolygon(); polyhandle; polyhandle = polygons->GetNextPolygon(polyhandle))
			{
				k3dIPolygon* polygon = polygons->GetPolygon(polyhandle);
				if(SDP_S_OK != polygon->State()->QueryState(k3dIState::VISIBLESELECTABLE))
					continue;

				// If the mouse click is outside the polygon, skip it ...
				if(!k3dIsInside(m_Mouse, polygon))
					continue;

				// Select/deselect the mesh ...
				k3dIObject* object = sdpQueryInterface<k3dIObject>(Mesh);
				g_assert(object);
				object->ObjectContext()->State()->SetState(k3dIState::SELECTED, m_SelectMode);

				// If selecting, make the mesh the active object ...
				if(m_SelectMode)
					m_Document->SetActiveObjectContext(object->ObjectContext());

				// We're done ...
				m_Selected = true;
				return false;
			}

		return true;
	}

	k3dIDocument* m_Document;
	sdpVector2 m_Mouse;
	bool m_SelectMode;
	bool m_Selected;
};

bool CCamera::SelectSingleMesh(GdkModifierType Modifiers, sdpVector2 CurrentMouse)
{
	Document()->Actions()->StartAction("Select Object");

	bool selectmode = ControlDown(Modifiers) ? false : true;
	CCameraSelectSingleMesh functor(Document(), CurrentMouse, selectmode);
	k3dForEachMesh(Document(), k3dIState::VISIBLESELECTABLE, functor);

	Document()->Actions()->FinishAction();

	return functor.Selected();
}

sdpResult CCamera::OnLButtonClick(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY)
{
	SnapshotScreenCoordinates();

	switch(Document()->GetActiveEditMode())
		{
			case k3dIDocument::POINT:
				ShiftDown(Modifiers) ? SelectSinglePolygon(Modifiers, sdpVector2(CurrentX, CurrentY)) : SelectSinglePoint(Modifiers, sdpVector2(CurrentX, CurrentY));
				break;
			case k3dIDocument::POLYGON:
				ShiftDown(Modifiers) ? SelectSingleMesh(Modifiers, sdpVector2(CurrentX, CurrentY)) : SelectSinglePolygon(Modifiers, sdpVector2(CurrentX, CurrentY));
				break;
			default:
				g_assert(0);	// Unknown edit mode!!
		}

	k3dRedrawAll(Document(), false);

	return SDP_S_OK;
}

sdpResult CCamera::OnLButtonDoubleClick(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY)
{
	EditObject();

	return SDP_S_OK;
}

GdkGC* CCamera::SelectionGC(sdpGtkWidget& Widget)
{
	// Sanity checks ...
	g_return_val_if_fail(Widget.Attached(), 0);
	
	// Create the rubber-band color ...
	GdkColor color;
	color.red = gushort(m_EditorSelectorColor[0] * 0xffff);
	color.green = gushort(m_EditorSelectorColor[1] * 0xffff);
	color.blue = gushort(m_EditorSelectorColor[2] * 0xffff);
	gdk_color_alloc(gdk_colormap_get_system(), &color);

	// Create an XOR gc for drawing ...
	GdkGC* gc = gdk_gc_new(((GtkWidget*)Widget)->window);
	gdk_gc_set_foreground(gc, &color);
	gdk_gc_set_function(gc, GDK_XOR);

	return gc;
}

void CCamera::RubberBand(sdpGtkWidget& Widget, sdpRectangle Rectangle)
{
	// Sanity checks ...
	g_return_if_fail(Widget.Attached());

	// Get our GC ...
	GdkGC* gc = SelectionGC(Widget);

	// Normalize coordinates ...
	Rectangle.Normalize();

	// Draw the rectangle ...
	gdk_draw_rectangle(((GtkWidget*)Widget)->window, gc, 0, gint(Rectangle.Left()), gint(Rectangle.Top()), gint(Rectangle.Width()), gint(Rectangle.Height()));

	// Cleanup ...
	gdk_gc_destroy(gc);
}

sdpResult CCamera::OnLButtonStartDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY)
{
	m_LassoPoints.clear();
	m_LassoPoints.push_back(sdpVector2(CurrentX, CurrentY));

	if(!ShiftDown(Modifiers))
		{
			m_RubberBand = sdpRectangle(CurrentX, CurrentX, CurrentY, CurrentY);
			RubberBand(Viewport(), m_RubberBand);
		}

	return SDP_S_OK;
}

sdpResult CCamera::OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	// Erase the existing rubber-band box ...
	if(!ShiftDown(Modifiers))
		RubberBand(Viewport(), m_RubberBand);

	if(ShiftDown(Modifiers))	// Lasso selection
		{
			if((sdpVector2(CurrentX, CurrentY) - m_LassoPoints[m_LassoPoints.size()-1]).Length2() > 10)
				{
					m_LassoPoints.push_back(sdpVector2(CurrentX, CurrentY));

					GdkGC* gc = SelectionGC(Viewport());

					gdk_draw_line(((GtkWidget*)Viewport())->window,
												gc,
												gint(m_LassoPoints[m_LassoPoints.size() - 2][0]),
												gint(m_LassoPoints[m_LassoPoints.size() - 2][1]),
												gint(m_LassoPoints[m_LassoPoints.size() - 1][0]),
												gint(m_LassoPoints[m_LassoPoints.size() - 1][1]));

					gdk_gc_destroy(gc);
				}
		}
	else	// Rubber-band selection ...
		{
			m_RubberBand = sdpRectangle(StartX, LastX, StartY, LastY);
			RubberBand(Viewport(), m_RubberBand);
		}

	return SDP_S_OK;
}

// Functor object for selecting points with a lasso ...
class CCameraSelectLassoPoints
{
public:
	CCameraSelectLassoPoints(sdpVector2Array& LassoPoints, bool SelectMode) : m_LassoPoints(LassoPoints)
	{
		m_SelectMode = SelectMode;
		m_Selected = false;
	}

	bool Selected() { return m_Selected; }

	bool HandlePoint(k3dIPoint* Point)
	{
		Point->Location()->GetScreenXYZ(m_ScreenCoords);
		sdpVector2 point(m_ScreenCoords, 2);
		if(sdpIsInside(point, m_LassoPoints))
			{
				m_Selected = true;
				Point->State()->SetState(k3dIState::SELECTED, m_SelectMode);
			}

		return true;
	}

	sdpVector2Array& m_LassoPoints;
	bool m_SelectMode;
	bool m_Selected;
	sdpVector3 m_ScreenCoords;
};

// Select points with a lasso ...
bool CCamera::SelectLassoPoints(GdkModifierType Modifiers)
{
	bool selectmode = ControlDown(Modifiers) ? false : true;
	CCameraSelectLassoPoints functor(m_LassoPoints, selectmode);
	k3dForEachPoint(Document(), k3dIState::VISIBLESELECTABLE, functor);

	// Get rid of the active point ...
	// Document()->SetActivePoint(0);

	return functor.Selected();
}

// Functor object for selecting polygons with a rubber-band box ...
class CCameraSelectLassoPolygons
{
public:
	CCameraSelectLassoPolygons(sdpVector2Array& LassoPoints, bool SelectMode, bool SelectPartial) : m_LassoPoints(LassoPoints)
	{
		m_SelectMode = SelectMode;
		m_SelectPartial = SelectPartial;
		m_Selected = false;
	}

	bool Selected() { return m_Selected; }

	bool HandlePolygon(k3dIPolygon* Polygon)
	{
		m_SomeInside = false;
		m_AllInside = true;

		// For each point in the poly ...
		k3dIPolyPointServer* points = Polygon->Points();
		for(K3DPOLYPOINTHANDLE pointhandle = points->GetFirstPolyPoint(); pointhandle; pointhandle = points->GetNextPolyPoint(pointhandle))
			{
				points->GetPolyPoint(pointhandle)->Point()->Location()->GetScreenXYZ(m_ScreenCoords);
				sdpVector2 point(m_ScreenCoords, 2);
				if(!sdpIsInside(point, m_LassoPoints))
					{
						m_AllInside = false;
						if(!m_SelectPartial)
							return true;
					}
				else
					{
						m_SomeInside = true;
					}
			}

		// If none of the points were inside the bounding-box, we're done ...
		if(!m_SomeInside)
			return true;

		// If select partial is off, and we didn't get every point, we're done ...
		if(!m_AllInside && !m_SelectPartial)
			return true;

		m_Selected = true;
		Polygon->State()->SetState(k3dIState::SELECTED, m_SelectMode);

		return true;
	}

	sdpVector2Array m_LassoPoints;
	bool m_SelectMode;
	bool m_SelectPartial;
	bool m_Selected;
	sdpVector3 m_ScreenCoords;
	bool m_SomeInside;
	bool m_AllInside;
};

// Select polygons with a lasso ...
bool CCamera::SelectLassoPolygons(GdkModifierType Modifiers)
{
	bool selectmode = ControlDown(Modifiers) ? false : true;
	CCameraSelectLassoPolygons functor(m_LassoPoints, selectmode, m_LassoSelectPartial);
	k3dForEachPolygon(Document(), k3dIState::VISIBLESELECTABLE, functor);
	
	return functor.Selected();
}

// Select geometry with a lasso ...
bool CCamera::SelectLasso(GdkModifierType Modifiers)
{
	Document()->Actions()->StartAction("Lasso Select");

	SnapshotScreenCoordinates();

	bool result;
	switch(Document()->GetActiveEditMode())
		{
			case k3dIDocument::POINT:
				result = SelectLassoPoints(Modifiers);
				break;
			case k3dIDocument::POLYGON:
				result = SelectLassoPolygons(Modifiers);
				break;
			default:
				g_assert_not_reached();	// Unknown edit mode!!!
		}

	Document()->Actions()->FinishAction();

	return result;
}

// Functor object for selecting points with a rubber-band box ...
class CCameraSelectBoxPoints
{
public:
	CCameraSelectBoxPoints(sdpVector2& UpperLeft, sdpVector2& LowerRight, bool SelectMode)
	{
		m_BoundingBox = k3dBoundingBox2(UpperLeft, LowerRight);
		m_SelectMode = SelectMode;
		m_Selected = false;
	}

	bool Selected() { return m_Selected; }

	bool HandlePoint(k3dIPoint* Point)
	{
		Point->Location()->GetScreenXYZ(m_ScreenCoords);
		sdpVector2 screencoords(m_ScreenCoords, 2);

		if(m_BoundingBox.Contains(screencoords))
			{
				m_Selected = true;
				Point->State()->SetState(k3dIState::SELECTED, m_SelectMode);
			}

		return true;
	}

	k3dBoundingBox2 m_BoundingBox;
	bool m_SelectMode;
	bool m_Selected;
	sdpVector3 m_ScreenCoords;
};

// Select points with a rubber-band box ...
bool CCamera::SelectBoxPoints(GdkModifierType Modifiers, sdpVector2 CurrentMouse, sdpVector2 StartMouse)
{
	bool selectmode = ControlDown(Modifiers) ? false : true;
	CCameraSelectBoxPoints functor(StartMouse, CurrentMouse, selectmode);
	k3dForEachPoint(Document(), k3dIState::VISIBLESELECTABLE, functor);

	// Clear the active point ...
	// Document()->SetActivePoint(0);

	return functor.Selected();
}

// Functor object for selecting polygons with a rubber-band box ...
class CCameraSelectBoxPolygons
{
public:
	CCameraSelectBoxPolygons(sdpVector2& UpperLeft, sdpVector2& LowerRight, bool SelectMode, bool SelectPartial)
	{
		m_BoundingBox = k3dBoundingBox2(UpperLeft, LowerRight);
		m_SelectMode = SelectMode;
		m_SelectPartial = SelectPartial;
		m_Selected = false;
	}

	bool Selected() { return m_Selected; }

	bool HandlePolygon(k3dIPolygon* Polygon)
	{
		m_SomeInside = false;
		m_AllInside = true;

		// For each point in the poly ...
		k3dIPolyPointServer* points = Polygon->Points();
		for(K3DPOLYPOINTHANDLE pointhandle = points->GetFirstPolyPoint(); pointhandle; pointhandle = points->GetNextPolyPoint(pointhandle))
			{
				points->GetPolyPoint(pointhandle)->Point()->Location()->GetScreenXYZ(m_ScreenCoords);

				sdpVector2 screencoords(m_ScreenCoords, 2);
				if(!m_BoundingBox.Contains(screencoords))
					{
						m_AllInside = false;
						if(!m_SelectPartial)
							break;
					}
				else
					{
						m_SomeInside = true;
					}
			}

		// If none of the points were inside the bounding-box, we're done ...
		if(!m_SomeInside)
			return true;

		// If select partial is off, and we didn't get every point, we're done ...
		if(!m_AllInside && !m_SelectPartial)
			return true;

		m_Selected = true;
		Polygon->State()->SetState(k3dIState::SELECTED, m_SelectMode);

		return true;
	}

	k3dBoundingBox2 m_BoundingBox;
	bool m_SelectMode;
	bool m_SelectPartial;
	bool m_Selected;
	sdpVector3 m_ScreenCoords;
	bool m_SomeInside;
	bool m_AllInside;
};

// Select polygons with a rubber-band box ...
bool CCamera::SelectBoxPolygons(GdkModifierType Modifiers, sdpVector2 CurrentMouse, sdpVector2 StartMouse)
{
	bool selectmode = ControlDown(Modifiers) ? false : true;
	bool selectpartial = false;

	CCameraSelectBoxPolygons functor(StartMouse, CurrentMouse, selectmode, selectpartial);
	k3dForEachPolygon(Document(), k3dIState::VISIBLESELECTABLE, functor);

	// Clear the active point / poly ...
	// Document()->SetActivePoint(0);
	// Document()->SetActivePolygon(0);

	return functor.Selected();
}

// Select geometry with a rubber-band box ...
bool CCamera::SelectBox(GdkModifierType Modifiers, sdpVector2 CurrentMouse, sdpVector2 StartMouse)
{
	Document()->Actions()->StartAction("Box Select");

	SnapshotScreenCoordinates();

	bool result;
	switch(Document()->GetActiveEditMode())
		{
			case k3dIDocument::POINT:
				result = SelectBoxPoints(Modifiers, CurrentMouse, StartMouse);
				break;
			case k3dIDocument::POLYGON:
				result = SelectBoxPolygons(Modifiers, CurrentMouse, StartMouse);
				break;
			default:
				g_assert_not_reached();	// Unknown edit mode!!!
		}

	Document()->Actions()->FinishAction();

	return result;
}

sdpResult CCamera::OnLButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	if(ShiftDown(Modifiers))
		{
			SelectLasso(Modifiers);
		}
	else
		{
			SelectBox(Modifiers, sdpVector2(CurrentX, CurrentY), sdpVector2(StartX, StartY));
		}

	k3dRedrawAll(Document(), false);

	return SDP_S_OK;
}

// Functor object for deselecting points ...
class CCameraDeselectPoint
{
public:
	bool HandlePoint(k3dIPoint* Point) { Point->State()->SetState(k3dIState::SELECTED, false); return true; }
};

// Functor object for deselecting polygons ...
class CCameraDeselectPolygon
{
public:
	bool HandlePolygon(k3dIPolygon* Polygon) { Polygon->State()->SetState(k3dIState::SELECTED, false); return true; }
};

// Functor object for deselecting objects ...
class CCameraDeselectObject
{
public:
	bool HandleObject(k3dIObjectContext* Context) { Context->State()->SetState(k3dIState::SELECTED, false); return true; }
};

// User clicked the right mouse button, so deselect everything ...
sdpResult CCamera::OnRButtonClick(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY)
{
	Document()->Actions()->StartAction("Deselect All");

	// Deselect points ...
	CCameraDeselectPoint pointfunctor;
	k3dForEachPoint(Document(), k3dIState::VISIBLESELECTED, pointfunctor);

	// Deselect polygons ...
	CCameraDeselectPolygon polygonfunctor;
	k3dForEachPolygon(Document(), k3dIState::VISIBLESELECTED, polygonfunctor);

	// Deselect objects ...
	CCameraDeselectObject objectfunctor;
	k3dForEachObject(Document(), k3dIState::VISIBLESELECTED, objectfunctor);

	Document()->Actions()->FinishAction();

	k3dRedrawAll(Document(), false);

	return SDP_S_OK;
}

sdpResult CCamera::OnRButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
//	::SetCursor(NULL);

	return SDP_S_OK;
}

sdpResult CCamera::OnRButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	// If the camera isn't interactive, we're done ...
	if(!m_CameraEnableInteractive())
		return SDP_S_OK;

	if(ShiftDown(Modifiers) && ControlDown(Modifiers))
		{
//					::SetCursor(m_DollyCursor);

			gdouble xfactor = gdouble(CurrentX - LastX);
			gdouble yfactor = gdouble(CurrentY - LastY);

			sdpVector3 offset;

			switch(Document()->GetActiveAxis())
				{
					case k3dIDocument::XAXIS:
						offset[0] = (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor) * (4.0 / m_CurrentFocalLength);
						offset[1] = 0.0;
						offset[2] = 0.0;
						break;

					case k3dIDocument::YAXIS:
						offset[0] = 0.0;
						offset[1] = (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor) * (-4.0 / m_CurrentFocalLength);
						offset[2] = 0.0;
						break;

					case k3dIDocument::ZAXIS:
						offset[0] = 0.0;
						offset[1] = 0.0;
						offset[2] = (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor) * (4.0 / m_CurrentFocalLength);
						break;

					default:
						g_assert(0);
						break;
				}

			offset *= m_PositionSensitivity;
			m_Position[k3dITransform::CURRENT_FRAME] += offset;
		}
	else if(ShiftDown(Modifiers))
		{
//					::SetCursor(m_DollyCursor);

			sdpVector3 offset;

			switch(Document()->GetActiveAxis())
				{
					case k3dIDocument::XAXIS:
						offset[0] = 0.0;
						offset[1] = double(CurrentY - LastY) * (4.0 / m_CurrentFocalLength);
						offset[2] = double(CurrentX - LastX) * (4.0 / m_CurrentFocalLength);
						break;

					case k3dIDocument::YAXIS:
						offset[0] = double(CurrentX - LastX) * (-4.0 / m_CurrentFocalLength);
						offset[1] = 0.0;
						offset[2] = double(CurrentY - LastY) * (4.0 / m_CurrentFocalLength);
						break;

					case k3dIDocument::ZAXIS:
						offset[0] = double(CurrentX - LastX) * (-4.0 / m_CurrentFocalLength);
						offset[1] = double(CurrentY - LastY) * (4.0 / m_CurrentFocalLength);
						offset[2] = 0.0;
						break;

					default:
						g_assert(0);
						break;
				}

			offset *= m_PositionSensitivity;
			m_Position[k3dITransform::CURRENT_FRAME] += offset;
		}
	else if(ControlDown(Modifiers))
		{
//					::SetCursor(m_TiltCursor);

			gdouble xfactor = gdouble(CurrentX - LastX);
			gdouble yfactor = gdouble(CurrentY - LastY);
			gdouble factor = fabs(xfactor) > fabs(yfactor) ? xfactor : yfactor;

			sdpQuaternion q(m_Orientation[k3dITransform::CURRENT_FRAME]);
			sdpEulerAngles a(q, sdpEulerAngles::ZXYstatic);
			a[0] -= sdpRadians(0.25 * m_OrientationSensitivity * factor);
			m_Orientation[k3dITransform::CURRENT_FRAME] = a;

// Don't ask me why the following version of the above code causes bizarre compiler errors!
//			sdpEulerAngles euler(sdpQuaternion(m_Orientation), sdpEulerAngles::ZXYstatic);
//			euler[0] += sdpRadians(0.25 * m_OrientationSensitivity * factor);
//			m_Orientation = euler;
		}
	else
		{
//					::SetCursor(m_PanCursor);

			sdpQuaternion q(m_Orientation[k3dITransform::CURRENT_FRAME]);
			sdpEulerAngles a(q, sdpEulerAngles::ZXYstatic);
			a[2] -= sdpRadians(0.25 * m_OrientationSensitivity * gdouble(LastX - CurrentX));
			a[1] -= sdpRadians(0.25 * m_OrientationSensitivity * gdouble(LastY - CurrentY));
			m_Orientation[k3dITransform::CURRENT_FRAME] = a;

// Don't ask me why the following version of the above code causes bizarre compiler errors!
//			sdpEulerAngles euler(sdpQuaternion(m_Orientation), sdpEulerAngles::ZXYstatic);
//			euler[2] += sdpRadians(0.25 * m_OrientationSensitivity * gdouble(LastX - CurrentX));
//			euler[1] += sdpRadians(0.25 * m_OrientationSensitivity * gdouble(LastY - CurrentY));
//			m_Orientation = euler;
		}

	// Update the view ...
	k3dRedrawAll(Document(), false);

	return SDP_S_OK;
}

sdpResult CCamera::OnLRButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{

//	::SetCursor(NULL);

	return SDP_S_OK;
}

sdpResult CCamera::OnLRButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	// If the camera isn't interactive, we're done ...
	if(!m_CameraEnableInteractive())
		return SDP_S_OK;

//	::SetCursor(m_ZoomCursor);

	double zoomfactor;
	double focallength = m_Focallength->GetValue1(0.0, 0.1);

	if(CurrentY > LastY)
		{
			if(m_ZoomSign == -1.0)
				zoomfactor = pow(.995, (double)(CurrentY - LastY));
			else
				zoomfactor = pow(1.005, (double)(CurrentY - LastY));

			focallength *= zoomfactor;
		}
	else
		{
			if(m_ZoomSign == -1.0)
				zoomfactor = pow(1.005, (double)(LastY - CurrentY));
			else
				zoomfactor = pow(.995, (double)(LastY - CurrentY));

			focallength *= zoomfactor;
		}

	k3dChannelSetConstantValue(m_Focallength.Interface(), focallength);
	k3dRedrawAll(Document(), false);

	return SDP_S_OK;
}

/////////////////////////////////////////////////////////////////////////////
// CCameraDialog

CCameraDialog::CCameraDialog()
{
	m_Object = 0;
}

bool CCameraDialog::Create(CCamera* Object)
{
	// Sanity checks ...
	g_assert(Object);

	// Store our parent object ...
	m_Object = Object;

	// Load our template ...
	g_return_val_if_fail(k3dDialog::Create(Object, false, true, "camera.gtkml", "camera.html"), false);

	return true;
}

bool CCameraDialog::InitializeControls()
{
	// Setup channel buttons ...
	ConnectChannelButton(m_Object->m_Farplane.Interface(), "farplane");
	ConnectChannelButton(m_Object->m_Focalplane.Interface(), "focalplane");
	ConnectChannelButton(m_Object->m_Fstop.Interface(), "fstop");
	ConnectChannelButton(m_Object->m_Exposure.Interface(), "riexposure");
	ConnectChannelButton(m_Object->m_Gamma.Interface(), "rigamma");

	// Setup pixel filter options ...
	sdpStringArray pixelfilters;
	pixelfilters.push_back(k3dIRiEngine::RI_BOX);
	pixelfilters.push_back(k3dIRiEngine::RI_CATMULL_ROM);
	pixelfilters.push_back(k3dIRiEngine::RI_GAUSSIAN);
	pixelfilters.push_back(k3dIRiEngine::RI_SINC);
	pixelfilters.push_back(k3dIRiEngine::RI_TRIANGLE);
	Combo("ripixelfiltername").SetList(pixelfilters);
	MapEvent("changed", "onripixelfiltername", GTK_OBJECT(((GtkCombo*)Combo("ripixelfiltername"))->entry), true);

	// Setup render engine options ...
	sdpString selectedengine = m_Object->m_RiRenderEngine();

	sdpXmlIElement* root = m_Object->Document()->Application()->Options()->GetRootElement();
	sdpXmlElementArray engines = sdpXmlXPathSelect(root, "application/helpers/command[@type=\"render\" and @enginetype=\"ri\" and @viewduringrender=\"false\"]");

	sdpStringArray enginelist;
	sdpString selectedenginename;
	for(gulong i = 0; i < engines.size(); i++)
		{
			sdpString engine, enginename;
			sdpXmlGetAttribute(engines[i], "engine", engine);
			sdpXmlGetAttribute(engines[i], "name", enginename);

			enginelist.push_back(enginename);
			
			if(selectedengine == engine)
				selectedenginename = enginename;
		}
	Combo("riengine").SetList(enginelist);
	Combo("riengine").SetEntryText(selectedenginename);
	
	MapEvent("changed", "onriengine", GTK_OBJECT(((GtkCombo*)Combo("riengine"))->entry), true);

	// Setup atmosphere shader selector ...
	m_RiAtmosphereShaderSelector.Create(m_Object->Application(), &m_Object->m_RiAtmosphereShader, &m_RiAtmosphereShaderPage, true, k3dIShaderInfo::VOLUME, Container("riatmosphereshaderselector"));
	// Setup the atmosphere shader arguments ...
	m_RiAtmosphereShaderPage.Create(m_Object->Document(), &m_Object->m_RiAtmosphereShader, Container("riatmosphereshadercontainer"));

	return true;
}

void CCameraDialog::UpdateControls()
{
	// Setup editor options ...
	ToggleButton("editorsolid").SetState(m_Object->m_EditorSolid());
	ToggleButton("editorfog").SetState(m_Object->m_EditorFog());
	ToggleButton("editortexture").SetState(m_Object->m_EditorTexture());
	ToggleButton("editortwosided").SetState(m_Object->m_EditorTwoSided());
	ToggleButton("editororientation").SetState(m_Object->m_EditorOrientation());
	ToggleButton("editornormals").SetState(m_Object->m_EditorNormals());

	Entry("rixsamples").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_RiXSamples(), 0));
	Entry("riysamples").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_RiYSamples(), 0));
	Combo("ripixelfiltername").SetEntryText(m_Object->m_RiPixelFilterName());
	Entry("ripixelfilterwidth").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_RiPixelFilterWidth(), 3));
	Entry("ripixelfilterheight").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_RiPixelFilterHeight(), 3));
	ToggleButton("riatmosphere").SetState(m_Object->m_RiAtmosphere());
	ToggleButton("ridepthoffield").SetState(m_Object->m_RiDepthOfField());
	ToggleButton("riradiosity").SetState(m_Object->m_RiRadiosity());
	Entry("riradiositysteps").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_RiRadiositySteps(), 0));
}

void CCameraDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "oneditorsolid")
		OnEditorSolid();
	else if(Event->Name() == "oneditorfog")
		OnEditorFog();
	else if(Event->Name() == "oneditortexture")
		OnEditorTexture();
	else if(Event->Name() == "oneditortwosided")
		OnEditorTwoSided();
	else if(Event->Name() == "oneditororientation")
		OnEditorOrientation();
	else if(Event->Name() == "oneditornormals")
		OnEditorNormals();
	else if(Event->Name() == "onriengine")
		OnRiEngine();
	else if(Event->Name() == "onrixsamples")
		OnRiXSamples();
	else if(Event->Name() == "onriysamples")
		OnRiYSamples();
	else if(Event->Name() == "onripixelfiltername")
		OnRiPixelFilterName();
	else if(Event->Name() == "onripixelfilterwidth")
		OnRiPixelFilterWidth();
	else if(Event->Name() == "onripixelfilterheight")
		OnRiPixelFilterHeight();
	else if(Event->Name() == "onriatmosphere")
		OnRiAtmosphere();
	else if(Event->Name() == "onridepthoffield")
		OnRiDepthOfField();
	else if(Event->Name() == "onriradiosity")
		OnRiRadiosity();
	else if(Event->Name() == "onriradiositysteps")
		OnRiRadiositySteps();
	else if(Event->Name() == "onshowviewport")
		OnShowViewport();
	else if(Event->Name() == "onhideviewport")
		OnHideViewport();
	else
		k3dDialog::OnEvent(Event);
}

void CCameraDialog::OnEditorSolid()
{
	m_Object->m_EditorSolid() = ToggleButton("editorsolid").GetState();
	m_Object->OnModify(false);
}

void CCameraDialog::OnEditorFog()
{
	m_Object->m_EditorFog() = ToggleButton("editorfog").GetState();
	m_Object->OnModify(false);
}

void CCameraDialog::OnEditorTexture()
{
	m_Object->m_EditorTexture() = ToggleButton("editortexture").GetState();
	m_Object->OnModify(false);
}

void CCameraDialog::OnEditorTwoSided()
{
	m_Object->m_EditorTwoSided() = ToggleButton("editortwosided").GetState();
	m_Object->OnModify(false);
}

void CCameraDialog::OnEditorOrientation()
{
	m_Object->m_EditorOrientation() = ToggleButton("editororientation").GetState();
	m_Object->OnModify(false);
}

void CCameraDialog::OnEditorNormals()
{
	m_Object->m_EditorNormals() = ToggleButton("editornormals").GetState();
	m_Object->OnModify(false);
}

void CCameraDialog::OnRiEngine()
{
	sdpString selectedenginename = Combo("riengine").GetEntryText();

	sdpXmlIElement* root = m_Object->Document()->Application()->Options()->GetRootElement();
	sdpXmlElementArray engines = sdpXmlXPathSelect(root, "application/helpers/command[@type=\"render\" and @enginetype=\"ri\" and @viewduringrender=\"false\"]");
	
	for(gulong i = 0; i < engines.size(); i++)
		{
			sdpString engine, enginename;
			sdpXmlGetAttribute(engines[i], "engine", engine);
			sdpXmlGetAttribute(engines[i], "name", enginename);

			if(selectedenginename == enginename)
				{
					m_Object->m_RiRenderEngine() = engine;
					return;
				}
		}
		
	// Shouldn't ever happen!
	g_return_if_fail(0);
}

void CCameraDialog::OnRiXSamples()
{
	UpdateControl(Entry("rixsamples"), m_Object->m_RiXSamples(), k3dIChannel::DATA_SCALAR, 1, ULONG_MAX);
}

void CCameraDialog::OnRiYSamples()
{
	UpdateControl(Entry("riysamples"), m_Object->m_RiYSamples(), k3dIChannel::DATA_SCALAR, 1, ULONG_MAX);
}

void CCameraDialog::OnRiPixelFilterName()
{
	m_Object->m_RiPixelFilterName() = Combo("ripixelfiltername").GetEntryText();
}

void CCameraDialog::OnRiPixelFilterWidth()
{
	UpdateControl(Entry("ripixelfilterwidth"), m_Object->m_RiPixelFilterWidth(), k3dIChannel::DATA_SCALAR, 0, ULONG_MAX, 3);
}

void CCameraDialog::OnRiPixelFilterHeight()
{
	UpdateControl(Entry("ripixelfilterheight"), m_Object->m_RiPixelFilterHeight(), k3dIChannel::DATA_SCALAR, 0, ULONG_MAX, 3);
}

void CCameraDialog::OnRiAtmosphere()
{
	m_Object->m_RiAtmosphere() = ToggleButton("riatmosphere").GetState();
}

void CCameraDialog::OnRiDepthOfField()
{
	m_Object->m_RiDepthOfField() = ToggleButton("ridepthoffield").GetState();
}

void CCameraDialog::OnRiRadiosity()
{
	m_Object->m_RiRadiosity() = ToggleButton("riradiosity").GetState();
}

void CCameraDialog::OnRiRadiositySteps()
{
	UpdateControl(Entry("riradiositysteps"), m_Object->m_RiRadiositySteps(), k3dIChannel::DATA_SCALAR, 1, ULONG_MAX);
}

void CCameraDialog::OnShowViewport()
{
	m_Object->ShowViewport();
}

void CCameraDialog::OnHideViewport()
{
	m_Object->HideViewport();
}

///////////////////////////////////////////////////////////////////////////////
// CCameraFrame

CCameraFrame::CCameraFrame()
{
	m_Object = 0;
}

bool CCameraFrame::Create(CCamera* Camera)
{
	// Sanity checks ...
	g_assert(Camera);
	
	// Store our parent object ...
	m_Object = Camera;

	// Load the frame ...
	g_return_val_if_fail(LoadFile(sdpAppendPath(k3dLibraryPath(m_Object), "cameraframe.gtkml")), false);

	// Create the OpenGL drawing area ...
	sdpGtkContainer container(Container("viewport"));
	g_return_val_if_fail(m_Viewport.Create(container, true, true, 5, 6, 5, 16), false);

	// Setup some events ...
	MapEvent("configure-event", "onconfigure", m_Viewport, true);
	MapEvent("expose-event", "onexpose", m_Viewport, true);
	MapEvent("motion-notify-event", "onmousemove", m_Viewport, true);
	MapEvent("button-press-event", "onbuttondown", m_Viewport, true);
	MapEvent("button-release-event", "onbuttonup", m_Viewport, true);

	// Update our titlebar ...
	UpdateTitlebar();

	// Make ourselves visible ...
	RootWidget().Show();

	return true;
}

void CCameraFrame::ObjectDeleted()
{
	// No more events from this point forward ...
	DisconnectAllEvents();

	// Cleanup our drawing area ...
	Viewport().Destroy();

	// Let the document know that we're no longer active ...	
	if(m_Object->Document()->GetActiveCamera() == m_Object)
		m_Object->Document()->SetActiveCamera(0);
	
	// Get rid of any leftover GTK+ objects ...
	if(Root())
		RootWidget().Destroy();
		
	// Cleanup memory ...
	Clear();
}

void CCameraFrame::OnEvent(sdpGtkEvent* Event)
{
	if(Event->Name() == "onconfigure")
		OnConfigureEvent(Event);
	else if(Event->Name() == "onexpose")
		OnExposeEvent(Event);
	else if(Event->Name() == "onmousemove")
		k3dMouseInput::OnMouseMove(Event);
	else if(Event->Name() == "onbuttondown")
		k3dMouseInput::OnButtonDown(Event);
	else if(Event->Name() == "onbuttonup")
		k3dMouseInput::OnButtonUp(Event);
	else if(Event->Name() == "onclose")
		OnClose();
	else if(Event->Name() == "ondelete")
		OnDelete(Event);
	else if(Event->Name() == "ondestroy")
		OnDestroy();
	else if(Event->Name() == "onsetfocus")
		OnSetFocus();
	else if(Event->Name() == "onviewrender")
		OnRender();
	else if(Event->Name() == "ondefaultview")
		OnDefaultView();
	else if(Event->Name() == "on-zview")
		OnAxisView("-z");
	else if(Event->Name() == "on+zview")
		OnAxisView("+z");
	else if(Event->Name() == "on-xview")
		OnAxisView("-x");
	else if(Event->Name() == "on+xview")
		OnAxisView("+x");
	else if(Event->Name() == "on-yview")
		OnAxisView("-y");
	else if(Event->Name() == "on+yview")
		OnAxisView("+y");
	else
		sdpGtkObjectContainer::OnEvent(Event);
}

void CCameraFrame::OnClose()
{
	RootWidget().Destroy();
}

void CCameraFrame::OnDelete(sdpGtkEvent* Event)
{
	RootWidget().Destroy();
}

void CCameraFrame::OnDestroy()
{
	// Block any subsequent events ...
	DisconnectAllEvents();

	// No need to keep track of events anymore ...
	Clear();

	// Cleanup our drawing area ...
	Viewport().Destroy();

	// Let the document know that we're no longer active ...	
	if(m_Object->Document()->GetActiveCamera() == m_Object)
		m_Object->Document()->SetActiveCamera(0);
}

void CCameraFrame::OnSetFocus()
{
/*
	// If the document is in the process of being closed, do nothing ...
	if(m_Object->Document()->IsDying())
		return true;
*/

	// Activate the document ...
	m_Object->Document()->SetActiveCamera(m_Object);
}

void CCameraFrame::OnRender()
{
	// Get an interface to the render farm ...
	k3dIRenderFarm* renderfarm = sdpQueryInterface<k3dIRenderFarm>(m_Object->Document()->Application());
	g_return_if_fail(renderfarm);

	// Start a new render job ...
	k3dIRenderJob* job = renderfarm->CreateJob("k3d-preview");
	g_return_if_fail(job);

	// Add a single render frame to the job ...
	k3dITimeBase* time = sdpQueryInterface<k3dITimeBase>(m_Object->Document());
	g_return_if_fail(time);
	
	k3dIRenderFrame* frame = job->CreateFrame(sdpFormatString("frame%.06u", time->GetViewFrame()));
	g_return_if_fail(frame);

	frame->SetViewDuringRender(true);
	frame->SetViewAfterRender(false);

	// Setup the view for rendering ...
	m_Object->StartJob(job);
	m_Object->QueueFrame(frame);
	m_Object->FinishJob(job);

	// Start the job running ...
	renderfarm->StartJob(job);
}

#ifdef RootWindow
#undef RootWindow
#endif // RootWindow

void CCameraFrame::UpdateTitlebar()
{
	// Get our title and stuff it in the titlebar ...
	RootWindow().SetTitle(Title());
}

sdpString CCameraFrame::Title()
{
	return "Camera Frame";
}

void CCameraFrame::OnDefaultView()
{
	k3dSetPosition(m_Object, sdpVector3(-20.0, 20.0, -20.0));
	k3dSetOrientation(m_Object, sdpEulerAngles(sdpRadians(0.0), sdpRadians(37.0), sdpRadians(45.0), sdpEulerAngles::ZXYstatic));
	k3dRedrawAll(m_Object->Document(), false);
}

void CCameraFrame::OnAxisView(SDPCSTRING Axis)
{
	if(Axis == sdpString("-z"))
		{
			k3dSetPosition(m_Object, sdpVector3(0.0, 0.0, -34.0));
			k3dSetOrientation(m_Object, sdpEulerAngles(sdpRadians(0.0), sdpRadians(0.0), sdpRadians(0.0), sdpEulerAngles::XYZstatic));
		}
	else if(Axis == sdpString("+z"))
		{
			k3dSetPosition(m_Object, sdpVector3(0.0, 0.0, 34.0));
			k3dSetOrientation(m_Object, sdpEulerAngles(sdpRadians(sdpVector3(0.0, 180.0, 0.0)), sdpEulerAngles::XYZstatic));
		}
	else if(Axis == sdpString("-x"))
		{
			k3dSetPosition(m_Object, sdpVector3(-34.0, 0.0, 0.0));
			k3dSetOrientation(m_Object, sdpEulerAngles(sdpRadians(0.0), sdpRadians(90.0), sdpRadians(0.0), sdpEulerAngles::XYZstatic));
		}
	else if(Axis == sdpString("+x"))
		{
			k3dSetPosition(m_Object, sdpVector3(34.0, 0.0, 0.0));
			k3dSetOrientation(m_Object, sdpEulerAngles(sdpRadians(0.0), sdpRadians(-90.0), sdpRadians(0.0), sdpEulerAngles::XYZstatic));
		}
	else if(Axis == sdpString("-y"))
		{
			k3dSetPosition(m_Object, sdpVector3(0.0, -34.0, 0.0));
			k3dSetOrientation(m_Object, sdpEulerAngles(sdpRadians(-90.0), sdpRadians(0.0), sdpRadians(0.0), sdpEulerAngles::XYZstatic));
		}
	else if(Axis == sdpString("+y"))
		{
			k3dSetPosition(m_Object, sdpVector3(0.0, 34.0, 0.0));
			k3dSetOrientation(m_Object, sdpEulerAngles(sdpRadians(90.0), sdpRadians(0.0), sdpRadians(0.0), sdpEulerAngles::XYZstatic));
		}
	k3dRedrawAll(m_Object->Document(), false);
}

// Called when the view is resized ...
void CCameraFrame::OnConfigureEvent(sdpGtkEvent* Event)
{
	m_Object->Redraw();
}

// Called when the view gets uncovered / needs to be redrawn ...
void CCameraFrame::OnExposeEvent(sdpGtkEvent* Event)
{
	m_Object->Redraw();
}



#define HANDLEEVENT(Event)	\
	k3dIObjectContext* context = m_Object->Document()->GetActiveObjectContext();	\
	if(context)	\
		{	\
			k3dIUserEventClient* usereventclient = sdpQueryInterface<k3dIUserEventClient>(context->Object());	\
			if(usereventclient)	\
				{	\
					sdpResult result = usereventclient->Event(m_Object, Modifiers, CurrentMouse[0], CurrentMouse[1]);	\
					usereventclient->Release();	\
					if(SDP_S_OK == result)	\
						return;	\
				}	\
		}	\
	m_Object->Event(m_Object, Modifiers, CurrentMouse[0], CurrentMouse[1]);



#define HANDLEDRAGEVENT(Event) \
	k3dIObjectContext* context = m_Object->Document()->GetActiveObjectContext();	\
	if(context)	\
		{	\
			k3dIUserEventClient* usereventclient = sdpQueryInterface<k3dIUserEventClient>(context->Object());	\
			if(usereventclient)	\
				{	\
					sdpResult result = usereventclient->Event(m_Object, Modifiers, CurrentMouse[0], CurrentMouse[1], LastMouse[0], LastMouse[1], StartMouse[0], StartMouse[1]);	\
					usereventclient->Release();	\
					if(SDP_S_OK == result)	\
						return;	\
				}	\
		}	\
	m_Object->Event(m_Object, Modifiers, CurrentMouse[0], CurrentMouse[1], LastMouse[0], LastMouse[1], StartMouse[0], StartMouse[1]);

	

void CCameraFrame::OnMouseMove(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	HANDLEEVENT(OnMouseMove);
}

void CCameraFrame::OnLButtonClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	HANDLEEVENT(OnLButtonClick);
}

void CCameraFrame::OnLButtonDoubleClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	HANDLEEVENT(OnLButtonDoubleClick);
}

void CCameraFrame::OnLButtonStartDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	HANDLEEVENT(OnLButtonStartDrag);
}

void CCameraFrame::OnLButtonDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
	HANDLEDRAGEVENT(OnLButtonDrag);
}

void CCameraFrame::OnLButtonEndDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
	HANDLEDRAGEVENT(OnLButtonEndDrag);
}

void CCameraFrame::OnMButtonClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	HANDLEEVENT(OnMButtonClick);
}

void CCameraFrame::OnMButtonDoubleClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	HANDLEEVENT(OnMButtonDoubleClick);
}

void CCameraFrame::OnMButtonStartDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	HANDLEEVENT(OnMButtonStartDrag);
}

void CCameraFrame::OnMButtonDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
	HANDLEDRAGEVENT(OnMButtonDrag);
}

void CCameraFrame::OnMButtonEndDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
	HANDLEDRAGEVENT(OnMButtonEndDrag);
}

void CCameraFrame::OnRButtonClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	HANDLEEVENT(OnRButtonClick);
}

void CCameraFrame::OnRButtonDoubleClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	HANDLEEVENT(OnRButtonDoubleClick);
}

void CCameraFrame::OnRButtonStartDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	HANDLEEVENT(OnRButtonStartDrag);
}

void CCameraFrame::OnRButtonDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
	HANDLEDRAGEVENT(OnRButtonDrag);
}

void CCameraFrame::OnRButtonEndDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
	HANDLEDRAGEVENT(OnRButtonEndDrag);
}

void CCameraFrame::OnLRButtonStartDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	HANDLEEVENT(OnLRButtonStartDrag);
}

void CCameraFrame::OnLRButtonDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
	HANDLEDRAGEVENT(OnLRButtonDrag);
}

void CCameraFrame::OnLRButtonEndDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
	HANDLEDRAGEVENT(OnLRButtonEndDrag);
}


