// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CChannel class which implements a data object that supports k3dIChannel
		\author Tim Shead (tshead@k-3d.com)
*/

#include <channel.h>

#include <k3distate.h>
#include <k3dmisc.h>

#include <sdpgtkcolorselector.h>
#include <sdpgtkevents.h>
#include <sdpxmlutility.h>
#include <sdpfile.h>

#include <iostream>
#include <algorithm>

////////////////////////////////////////////////////////////////////////
// CChannel

gulong CChannel::m_MaxIterations = 5000;
bool CChannel::m_AutoUpdate = true;

CChannel::CChannel()
{
	AddVariable(&m_DataType("datatype", k3dIChannel::DATA_SCALAR));
//	AddVariable(&m_ShowOnButton("showonbutton", false));
//	AddVariable(&m_ShowTime("showtime", true));
	AddVariable(&m_MinValue("minvalue", -DBL_MAX));
	AddVariable(&m_MaxValue("maxvalue", DBL_MAX));
	AddVariable(&m_MinCoords("mincoords", sdpVector2(0.0, -DBL_MAX)));
	AddVariable(&m_MaxCoords("maxcoords", sdpVector2(1.0, DBL_MAX)));
//	AddVariable(&m_GridSegmentsX("gridsegmentsx", 10));
//	AddVariable(&m_GridSegmentsY("gridsegmentsy", 10));
//	AddVariable(&m_GridSnapX("gridsnapx", false));
//	AddVariable(&m_GridSnapY("gridsnapy", false));
//	AddVariable(&m_Zoom("zoom", 1.0));
//	AddVariable(&m_Pan("pan", sdpVector2(0.5, 0.0)));
	AddVariable(&m_Cycles("cycles", 1.0));
	AddVariable(&m_Phase("phase", 0.0));
//	AddVariable(&m_LabelX("labelx", "Time"));
//	AddVariable(&m_LabelY("labely", "Value"));

	// We always have at least one segment ...
	m_Segments.push_back(new CChannelSegment);
}

CChannel::~CChannel()
{
	Clear();
}

void CChannel::Clear()
{
	for(SegmentIterator segment = m_Segments.begin(); segment != m_Segments.end(); segment++)
		delete (*segment);

	m_Segments.clear();
}

gulong CChannel::GetElementsPerNode()
{
	switch(m_DataType())
		{
			case k3dIChannel::DATA_SCALAR:
			case k3dIChannel::DATA_TIME:
			case k3dIChannel::DATA_ANGLE:
			case k3dIChannel::DATA_DISTANCE:
			case k3dIChannel::DATA_AREA:
			case k3dIChannel::DATA_VOLUME:
			case k3dIChannel::DATA_MASS:
			case k3dIChannel::DATA_FORCE:
			case k3dIChannel::DATA_PRESSURE:
				return 1;
			case k3dIChannel::DATA_COLOR:
			case k3dIChannel::DATA_ROTATIONXYZ:
			case k3dIChannel::DATA_SCALEXYZ:
				return 3;
			default:
				g_assert_not_reached(); // Did you add to the enumeration without telling me?
		}

	return 0;
}

void CChannel::NotifyChanged()
{
	Notify("channel-changed");
}

void CChannel::NotifyTracking()
{
	Notify("channel-tracking");
}

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CChannel::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIAnimate>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dIChannel>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dISplineChannel>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dIObject>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dIPersist>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dITimeServer>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dIUserEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIObject*>(this), InterfaceID, Interface)) return SDP_S_OK;											

	return SDP_E_NO_INTERFACE;
}

gulong CChannel::AddRef()
{
	return sdpComAddRef(m_ReferenceCount);
}

gulong CChannel::Release()
{
	return sdpComRelease(this, m_ReferenceCount);
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

/*
sdpResult CChannel::SetObjectContext(k3dIObjectContext* Context)
{
	g_return_val_if_fail(sdpSucceeded(k3dPersistObject::SetObjectContext(Context)), SDP_E_FAIL);

	// Make sure we don't show up in the hierarchy ...
	Context->State()->SetState(k3dIState::USERVIEWABLE, false);
	return SDP_S_OK;
}
*/

sdpResult CChannel::EditObject()
{
	g_return_val_if_fail(m_UserInterface.Create(this), SDP_E_FAIL);
	return SDP_S_OK;
}

sdpResult CChannel::ObjectRenamed(SDPCSTRING NewName)
{
	m_UserInterface.ObjectRenamed(NewName);
	return k3dPersistObject::ObjectRenamed(NewName);
}

sdpResult CChannel::ObjectDeleted()
{
	m_UserInterface.ObjectDeleted();
	return k3dPersistObject::ObjectDeleted();
}

//////////////////////////////////////////////////////////////
// k3dITimeServer overrides

sdpResult CChannel::SetAnimTime(double Time, double DeltaTime)
{
	k3dTimeServer::SetAnimTime(Time, DeltaTime);

//	if(m_Sheet)
//		m_Sheet->UpdateTime();

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIChannel overrides

sdpResult CChannel::SetDatatype(k3dIChannel::DataType Data)
{
	m_DataType() = Data;

	if(k3dIChannel::DATA_COLOR == Data)
		SetRange3(sdpVector3(0.0), sdpVector3(1.0));

	return SDP_S_OK;
}

k3dIChannel::DataType CChannel::GetDatatype()
{
	return m_DataType();
}

sdpResult CChannel::GetCurveRange(gdouble* Min, gdouble* Max)
{
	// Sanity checks ...
	g_return_val_if_fail(Min, SDP_E_FAIL);
	g_return_val_if_fail(Max, SDP_E_FAIL);
	g_return_val_if_fail(1 == GetElementsPerNode(), SDP_E_FAIL);
	
	*Min = DBL_MAX;
	*Max = -DBL_MAX;

	// For each segment ...
	for(SegmentIterator segment = m_Segments.begin(); segment != m_Segments.end(); segment++)
		{
			// For each node ...
			for(int i = 0; i < NODESPERSEGMENT; i++)
				{
					if((*segment)->Node(i).m_Coords[1] > *Max) *Max = (*segment)->Node(i).m_Coords[1];
					if((*segment)->Node(i).m_Coords[1] < *Min) *Min = (*segment)->Node(i).m_Coords[1];
				}
		}

	return SDP_S_OK;
}

gdouble CChannel::GetValue1(const gdouble Time, const gdouble MaxError)
{
	// We skip checking the data type here for efficiency ...
	gdouble time = Time;

	// Clamp time to 0.0 - 1.0 ...
	if(time < 0.0)
		time = 0.0;
	else if(time > 1.0)
		time = 1.0;

	// Offset our time frequency and phase ...
	time *= m_Cycles();
	time += m_Phase();
	if(time > 1.0)
		time -= floor(time);

	// Get the segment that contains our time ...
	SegmentIterator segment;
	SegmentIterator begin = m_Segments.begin();
	SegmentIterator end = m_Segments.end();
	for(segment = begin; segment != end; segment++)
		if(time <= (*segment)->Node(1)[0]) break;

	// If we're past the end of the last segment, that's the value ...
	if(segment == end)
		return m_Segments.back()->Node(1)[1];

	// If we're before the beginning of the first segment, that's the value ...
	if(segment == begin)
		return (*segment)->Node(1)[1];

	// Somewhere between two segments ...
	SegmentIterator last = segment;
	last--;

	// If the two segments are adjacent, we're done ...
	double dx = (*segment)->Node(1)[0] - (*last)->Node(1)[0];
	if(0.0 == dx)
		return (*segment)->Node(1)[1];

	// Find the bezier interpolation between the two segments ...
	//
	// Basically, we choose a Bezier curve length u that we think might be close,
	// then calculate the x value at that u.  We compare x to our desired
	// time dt, and, if it exceeds the maximum error, adjust u a portion of the
	// distance and repeat.  We also keep track of the number of iterations and,
	// if they exceed a threshold, abort.
	//
	// If anyone has a better idea, I'd love to hear it!
	double dt = (time - (*last)->Node(1)[0]) / dx;
	double u = dt;
	double x, u2, omu, omu2;

	gulong count = 0;
	while(1)
		{
			u2 = u * u;
			omu = 1.0 - u;
			omu2 = omu * omu;

			x = omu * omu2 * (*last)->Node(1)[0] + 3 * u * omu2 * (*last)->Node(2)[0] + 3 * u2 * omu * (*segment)->Node(0)[0] +	u * u2 * (*segment)->Node(1)[0];
			x = (x - (*last)->Node(1)[0]) / dx;

			if(fabs(x - dt) < MaxError) break;

			u += .4 * (dt - x);

			if(++count > m_MaxIterations)
				{
					std::cout << "CChannel::GetValue1() couldn't converge on a solution in " << m_MaxIterations << "iterations!" << std::endl;
					break;
				}
		}

	return omu * omu2 * (*last)->Node(1)[1] + 3 * u * omu2 * (*last)->Node(2)[1] +	3 * u2 * omu * (*segment)->Node(0)[1] + u * u2 * (*segment)->Node(1)[1];
}

sdpResult CChannel::GetValue2(const gdouble Time, const gdouble MaxError, double Value[2])
{
	g_return_val_if_fail(0, SDP_E_FAIL);
}

sdpResult CChannel::GetValue3(const gdouble Time, const gdouble MaxError, double Value[3])
{
	gdouble time = Time;

	if(time < 0.0)
		time = 0.0;
	else if(time > 1.0)
		time = 1.0;

	time *= m_Cycles();
	time += m_Phase();
	if(time > 1.0)
		time -= floor(time);

	SegmentIterator segment;
	SegmentIterator begin = m_Segments.begin();
	SegmentIterator end = m_Segments.end();
	for(segment = begin; segment != end; segment++)
		{
			if(false == (*segment)->Node(1).m_ValueNode)
				continue;
			if(time <= (*segment)->Node(1)[0])
				break;
		}

	// Past the end of the last segment
	if(segment == end)
		{
			m_Segments.back()->Node(1).m_Value.CopyArray(Value);
			return SDP_S_OK;
		}

	// Before the first segment
	if(segment == begin)
		{
			(*segment)->Node(1).m_Value.CopyArray(Value);
			return SDP_S_OK;
		}

	// Somewhere between two segments ...
	SegmentIterator last = segment;
	last--;
	for(; last != begin; last--)
		{
			if((*last)->Node(1).m_ValueNode) break;
		}

	// If the two segments have the same position ...
	if((*last)->Node(1)[1] == (*segment)->Node(1)[1])
		{
			(*last)->Node(1).m_Value.CopyArray(Value);
			return SDP_S_OK;
		}

	// Get the mix position ...
	gdouble mix = GetValue1(Time, MaxError);

	// Adjust the mix for our "sawtooth" pattern ...
	if((*last)->Node(1)[1] > (*segment)->Node(1)[1])
		mix = 1.0 - mix;

	// Linear interpolate the final value ...
	sdpVector3 result = sdpLerp<sdpVector3>((*last)->Node(1).m_Value, (*segment)->Node(1).m_Value, mix);
	result.CopyArray(Value);

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dISplineChannel overrides

sdpResult CChannel::SetRange1(const gdouble Min, const gdouble Max)
{
	// Make sure the data type really is one-dimensional ...
	g_return_val_if_fail(1 == GetElementsPerNode(), SDP_E_FAIL);

	m_MinValue() = Min;
	m_MaxValue() = Max;

	m_MinCoords() = sdpVector2(0.0, Min);
	m_MaxCoords() = sdpVector2(1.0, Max);

	return SDP_S_OK;
}

sdpResult CChannel::SetRange2(const gdouble Min[2], const gdouble Max[2])
{
	// Make sure the data type really is two-dimensional ...
	g_return_val_if_fail(2 == GetElementsPerNode(), SDP_E_FAIL);

	m_MinValue() = Min;
	m_MaxValue() = Max;

	m_MinCoords() = sdpVector2(0.0, 0.0);
	m_MaxCoords() = sdpVector2(1.0, 1.0);

	return SDP_S_OK;
}

sdpResult CChannel::SetRange3(const gdouble Min[3], const gdouble Max[3])
{
	// Make sure the data type really is three-dimensional ...
	g_return_val_if_fail(3 == GetElementsPerNode(), SDP_E_FAIL);

	m_MinValue() = Min;
	m_MaxValue() = Max;

	m_MinCoords() = sdpVector2(0.0, 0.0);
	m_MaxCoords() = sdpVector2(1.0, 1.0);

	return SDP_S_OK;
}

sdpResult CChannel::GetRange1(gdouble* Min, gdouble* Max)
{
	// Sanity checks ...
	g_return_val_if_fail(1 == GetElementsPerNode(), SDP_E_FAIL);
	g_return_val_if_fail(Min, SDP_E_FAIL);
	g_return_val_if_fail(Max, SDP_E_FAIL);
	
	*Min = m_MinValue()[0];
	*Max = m_MaxValue()[0];
	
	return SDP_S_OK;
}

sdpResult CChannel::GetRange2(gdouble Min[2], gdouble Max[2])
{
	// Sanity checks ...
	g_return_val_if_fail(1 == GetElementsPerNode(), SDP_E_FAIL);
	g_return_val_if_fail(Min, SDP_E_FAIL);
	g_return_val_if_fail(Max, SDP_E_FAIL);
	
	Min[0] = m_MinValue()[0];
	Min[1] = m_MinValue()[1];
	
	Max[0] = m_MaxValue()[0];
	Max[1] = m_MaxValue()[1];
	
	return SDP_S_OK;
}

sdpResult CChannel::GetRange3(gdouble Min[3], gdouble Max[3])
{
	// Sanity checks ...
	g_return_val_if_fail(1 == GetElementsPerNode(), SDP_E_FAIL);
	g_return_val_if_fail(Min, SDP_E_FAIL);
	g_return_val_if_fail(Max, SDP_E_FAIL);
	
	Min[0] = m_MinValue()[0];
	Min[1] = m_MinValue()[1];
	Min[2] = m_MinValue()[2];
	
	Max[0] = m_MaxValue()[0];
	Max[1] = m_MaxValue()[1];
	Max[2] = m_MaxValue()[2];
	
	return SDP_S_OK;
}

gulong CChannel::GetSegmentCount()
{
	return m_Segments.size() - 1;
}

K3DSPLINECHANNELSEGMENTHANDLE CChannel::GetFirstSegment()
{
	return K3DSPLINECHANNELSEGMENTHANDLE(m_Segments.size() > 1 ? 1 : 0);
}

K3DSPLINECHANNELSEGMENTHANDLE CChannel::GetNextSegment(const K3DSPLINECHANNELSEGMENTHANDLE Handle)
{
	// Sanity checks ...
	g_return_val_if_fail(Handle, 0);
	return K3DSPLINECHANNELSEGMENTHANDLE(gulong(Handle) < GetSegmentCount() ? gulong(Handle)+1 : 0);
}

k3dISplineChannelSegment* CChannel::GetSegment(const K3DSPLINECHANNELSEGMENTHANDLE Handle)
{
	// Sanity checks ...
	g_return_val_if_fail(Handle, 0);
	g_return_val_if_fail(gulong(Handle) <= GetSegmentCount(), 0);
	
	return m_Segments[gulong(Handle) - 1];
}

k3dISplineChannelSegment* CChannel::CreateSegment()
{
	CChannelSegment* newsegment = new CChannelSegment();
	g_return_val_if_fail(newsegment, 0);

	m_Segments.push_back(newsegment);

	return newsegment;
}

sdpResult CChannel::DestroySegment(const k3dISplineChannelSegment* Segment)
{
	// Sanity checks ...
	g_return_val_if_fail(Segment, SDP_E_FAIL);

	// Find the iterator for this segment ...
	SegmentIterator segment = std::find(m_Segments.begin(), m_Segments.end(), Segment);
	g_return_val_if_fail(segment != m_Segments.end(), SDP_E_FAIL);

	// Quietly fail if this is the last segment ...
	if(1 == m_Segments.size())
		return SDP_S_OK;

	// Delete it ...
	delete (*segment);
	m_Segments.erase(segment);

	return SDP_S_OK;
}

gulong CChannel::GetNodeCount()
{
	return ((m_Segments.size() - 1) * 3) + 1;
}

K3DSPLINECHANNELNODEHANDLE CChannel::GetFirstNode()
{
	return K3DSPLINECHANNELNODEHANDLE(1);
}

K3DSPLINECHANNELNODEHANDLE CChannel::GetNextNode(const K3DSPLINECHANNELNODEHANDLE Handle)
{
	// Sanity checks ...
	g_return_val_if_fail(Handle, 0);
	g_return_val_if_fail(gulong(Handle) < GetNodeCount(), 0);
	
	return K3DSPLINECHANNELNODEHANDLE(gulong(Handle) < GetNodeCount() ? gulong(Handle) + 1 : 0);
}

k3dISplineChannelNode* CChannel::GetNode(const K3DSPLINECHANNELNODEHANDLE Handle)
{
	// Sanity checks ...
	g_return_val_if_fail(Handle, 0);
	g_return_val_if_fail(gulong(Handle) <= GetNodeCount(), 0);

	gulong segment = gulong(Handle) / gulong(3);	
	gulong node = gulong(Handle) % 3;
	
	return &m_Segments[segment]->Node(node);
}

///////////////////////////////////////////////////////////////
// k3dIPersist overrides

sdpResult CChannel::Save(sdpXmlIElement* Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert(Dependencies);

	g_return_val_if_fail(sdpSucceeded(k3dPersistObject::Save(Element, Dependencies)), SDP_E_FAIL);

	sdpXmlIElement* segments = sdpXmlCreateElement(Element, "segments");

	for(SegmentIterator segment = m_Segments.begin(); segment != m_Segments.end(); segment++)
		(*segment)->Save(segments, Dependencies);

	return SDP_S_OK;
}

class CChannelLoadChannelSegment
{
public:
	CChannelLoadChannelSegment(CChannel* Channel) { m_Channel = Channel; m_Count = 0; }

	bool HandleElement(sdpXmlIElement* Element)
	{
		// Since CChannels always have at least one segment from the get-go,
		// avoid re-creating the first ...
		if(m_Count++)
			{
				CChannelSegment* newsegment = new CChannelSegment();
				m_Channel->m_Segments.push_back(newsegment);
			}

		m_Channel->m_Segments.back()->Load(Element);
		
		return true;
	}

	CChannel* m_Channel;
	gulong m_Count;
};

sdpResult CChannel::Load(sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_assert(Element);

	g_return_val_if_fail(sdpSucceeded(k3dPersistObject::Load(Element)), SDP_E_FAIL);

	sdpXmlIElement* segments = sdpXmlFindElement(Element, "segments");
	if(!segments)
		return SDP_S_OK;

	CChannelLoadChannelSegment functor(this);
	sdpXmlForEachElement(segments, "segment", functor);

	return SDP_S_OK;
}

////////////////////////////////////////////////////////////////////////
// CChannelNode

int CChannelNode::m_NodeSize = 3;
int CChannelNode::m_ValueNodeSize = 4;

CChannelNode::CChannelNode()
{
	m_Value = 0.0;
	m_Coords = 0.0;
	m_Selected = false;
	m_ValueNode = false;
}

CChannelNode::CChannelNode(const sdpVector3& Value, const sdpVector2& Coords)
{
	m_Value = Value;
	m_Coords = Coords;
	m_Selected = false;
	m_ValueNode = false;
}

CChannelNode::~CChannelNode()
{
}

///////////////////////////////////////////////////////
// k3dISplineChannelNode overrides ...

sdpResult CChannelNode::ValueNode()
{
	return m_ValueNode ? SDP_S_OK : SDP_S_FALSE;
}

sdpResult CChannelNode::SetPosition(gdouble Position[2])
{
	m_Coords = Position;
	return SDP_S_OK;
}

sdpResult CChannelNode::SetValue1(gdouble Value)
{
	m_Value = sdpVector3(Value, 0, 0);
	return SDP_S_OK;
}

sdpResult CChannelNode::SetValue2(gdouble Value[2])
{
	m_Value = sdpVector3(Value[0], Value[1], 0);
	return SDP_S_OK;
}

sdpResult CChannelNode::SetValue3(gdouble Value[3])
{
	m_Value = Value;
	return SDP_S_OK;
}

sdpResult CChannelNode::GetPosition(gdouble Position[2])
{
	m_Coords.CopyArray(Position);
	return SDP_S_OK;
}

sdpResult CChannelNode::GetValue1(gdouble* Value)
{
	// Sanity checks ...
	g_return_val_if_fail(Value, SDP_E_FAIL);

	*Value = m_Value[0];

	return SDP_S_OK;
}

sdpResult CChannelNode::GetValue2(gdouble Value[2])
{
	// Sanity checks ...
	g_return_val_if_fail(Value, SDP_E_FAIL);

	Value[0] = m_Value[0];
	Value[1] = m_Value[1];

	return SDP_S_OK;
}

sdpResult CChannelNode::GetValue3(gdouble Value[3])
{
	// Sanity checks ...
	g_return_val_if_fail(Value, SDP_E_FAIL);

	Value[0] = m_Value[0];
	Value[1] = m_Value[1];
	Value[2] = m_Value[2];

	return SDP_S_OK;
}

//////////////////////////////////////////////////////
// Odds-n-ends ...

sdpResult CChannelNode::Save(sdpXmlIElement* Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert(Dependencies);

	sdpXmlCreateElement(Element, "node", 0, "value", m_Value, "coords", m_Coords, "valuenode", m_ValueNode, "selected", m_Selected);

	return SDP_S_OK;
}

sdpResult CChannelNode::Load(sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_assert(Element);

	sdpXmlGetAttribute(Element, "value", m_Value);
	sdpXmlGetAttribute(Element, "coords", m_Coords);
	sdpXmlGetAttribute(Element, "valuenode", m_ValueNode);
	sdpXmlGetAttribute(Element, "selected", m_Selected);

	return SDP_S_OK;
}

////////////////////////////////////////////////////////////////////////
// CChannelSegment

CChannelSegment::CChannelSegment()
{
	m_Selected = false;
}

CChannelSegment::CChannelSegment(const sdpVector3& AValue, const sdpVector2& ACoords,
																 const sdpVector3& BValue, const sdpVector2& BCoords,
																 const sdpVector3& CValue, const sdpVector2& CCoords)
{
	m_Nodes[0] = CChannelNode(AValue, ACoords);
	m_Nodes[1] = CChannelNode(BValue, BCoords);
	m_Nodes[2] = CChannelNode(CValue, CCoords);

	m_Selected = false;
}

sdpResult CChannelSegment::Save(sdpXmlIElement* Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert(Dependencies);

	sdpXmlIElement* segment = sdpXmlCreateElement(Element, "segment", 0, "selected", m_Selected);

	m_Nodes[0].Save(segment, Dependencies);
	m_Nodes[1].Save(segment, Dependencies);
	m_Nodes[2].Save(segment, Dependencies);

	return SDP_S_OK;
}

sdpResult CChannelSegment::Load(sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_assert(Element);

	sdpXmlGetAttribute(Element, "selected", m_Selected);

	int index = 0;
	for(SDPXMLELEMENTHANDLE e = Element->GetFirstElement(); e; e = Element->GetNextElement(e))
		{
			sdpXmlIElement* node = Element->GetElement(e);
			if(node->Name() != sdpString("node"))
				continue;

			m_Nodes[index].Load(node);

			index++;
			if(index > 2)
				break;
		}

	return SDP_S_OK;
}

/////////////////////////////////////////////////////////////////////////////
// CChannelDialog

sdpVector3 CChannelDialog::m_BackgroundColor(0.8, 0.8, 0.8);
sdpVector3 CChannelDialog::m_NormalColor(0.0, 0.0, 0.0);
sdpVector3 CChannelDialog::m_ValueNodeColor(1.0, 1.0, 0.0);
sdpVector3 CChannelDialog::m_SelectedColor(1.0, 1.0, 1.0);
sdpVector3 CChannelDialog::m_TangentColor(0.0, 0.0, 1.0);
sdpVector3 CChannelDialog::m_SelectorColor(0.215, 0.215, 0.8);
sdpVector3 CChannelDialog::m_CursorColor(1.0, 0.0, 0.0);
sdpVector3 CChannelDialog::m_TimeColor(0.0, 1.0, 0.0);
sdpRectangle CChannelDialog::m_Frustum(-1, 1, 1, -1);

CChannelDialog::CChannelDialog()
{
	m_Object = 0;
}

bool CChannelDialog::Create(CChannel* Object)
{
	// Store our parent object ...
	m_Object = Object;

	// Load our template ...
	g_return_val_if_fail(k3dDialog::Create(Object, false, true, "channel.gtkml", "channel.html"), false);

	// Reset our view so it's properly scaled & centered ...
	ResetView();

	return true;
}

bool CChannelDialog::InitializeControls()
{
	// Create the OpenGL widgets ...
	sdpGtkContainer curvecontainer(Container("curve"));
	m_CurveWidget.Create(curvecontainer, true, true, 5, 6, 5, 0);

		// Setup some events ...
	MapEvent("configure-event", "onconfigurecurve", m_CurveWidget, true);
	MapEvent("expose-event", "onexposecurve", m_CurveWidget, true);
	MapEvent("motion-notify-event", "onmousemovecurve", m_CurveWidget, true);
	MapEvent("button-press-event", "onbuttondowncurve", m_CurveWidget, true);
	MapEvent("button-release-event", "onbuttonupcurve", m_CurveWidget, true);
	
	sdpGtkContainer colorcontainer(Container("color"));
	m_ColorWidget.Create(colorcontainer, true, true, 5, 6, 5, 0);

	// Setup some events ...
	MapEvent("configure-event", "onconfigurecolor", m_ColorWidget, true);
	MapEvent("expose-event", "onexposecolor", m_ColorWidget, true);
//	MapEvent("motion-notify-event", "onmousemovecolor", m_ColorWidget, true);
//	MapEvent("button-press-event", "onbuttondowncolor", m_ColorWidget, true);
//	MapEvent("button-release-event", "onbuttonupcolor", m_ColorWidget, true);

	// Only make the color preview bar visible if the channel is, indeed, a color channel ...
	if(k3dIChannel::DATA_COLOR == m_Object->GetDatatype())
		Widget("colorframe").Show();
	else
		Widget("colorframe").Hide();

	// Initialize the cursor ...
	SetCursorX(0.0);

	return k3dDialog::InitializeControls();
}

void CChannelDialog::UpdateControls()
{
	Entry("cycles").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_Cycles(), 3));
	Entry("phase").SetText(k3dConvertDoubleString(k3dIChannel::DATA_ANGLE, m_Object->m_Phase(), 2));
}

void CChannelDialog::FinalizeControls()
{
	// Cleanup our OpenGL controls ...
	m_CurveWidget.Destroy();
	m_ColorWidget.Destroy();	
	
	k3dDialog::FinalizeControls();
}

sdpVector2 CChannelDialog::World(sdpVector2 ScreenCoords)
{
	// Get the dimensions of the curve window ...
	gulong width = m_CurveWidget.Width();
	gulong height = m_CurveWidget.Height();

	sdpVector2 result;

	// Convert pixel coordinates to "screen" coordinates ...
	// (Make sure we handle (possible) minimized / hidden widget)
	if(width)
		result[0] = sdpLerp(m_Frustum.Left(), m_Frustum.Right(), ScreenCoords[0] / width);

	if(height)
		result[1] = sdpLerp(m_Frustum.Top(), m_Frustum.Bottom(), ScreenCoords[1] / height);

	// Scale the results to account for pan/zoom ...
	result = sdpProduct(result, m_Scale) + m_Origin;
	result[1] *= m_VScale;

	return result;
}

sdpVector2 CChannelDialog::Screen(sdpVector2 WorldCoords)
{
	// Get the dimensions of the curve window ...
	gulong width = m_CurveWidget.Width();
	gulong height = m_CurveWidget.Height();

	// Convert world coordinates to "screen" coordinates ...
	WorldCoords[1] *= (1.0 / m_VScale);
	sdpVector2 result = sdpProduct((WorldCoords - m_Origin), sdpVector2(1.0 / m_Scale[0], 1.0 / m_Scale[1]));

	// Convert to pixel coordinates ...
	result[0] = width * (result[0] - m_Frustum.Left()) / m_Frustum.Width();
	result[1] = height * (1 - ((result[1] - m_Frustum.Bottom()) / m_Frustum.Height()));

	return result;
}

sdpRectangle CChannelDialog::Extents()
{
	sdpRectangle result(0, 1, -DBL_MAX, DBL_MAX);

	// For each segment ...
	for(SegmentIterator segment = m_Object->m_Segments.begin(); segment != m_Object->m_Segments.end(); segment++)
		{
			// For each node ...
			for(int i = 0; i < NODESPERSEGMENT; i++)
				{
					if((*segment)->Node(i).m_Coords[1] > result.top)
						result.top = (*segment)->Node(i).m_Coords[1];
					if((*segment)->Node(i).m_Coords[1] < result.bottom)
						result.bottom = (*segment)->Node(i).m_Coords[1];
				}
		}

	return result;
}

sdpRectangle CChannelDialog::NodeExtents()
{
	sdpRectangle result(0, 1, -DBL_MAX, DBL_MAX);

	// For each segment ...
	for(SegmentIterator segment = m_Object->m_Segments.begin(); segment != m_Object->m_Segments.end(); segment++)
		{
			if((*segment)->Node(1).m_Coords[1] > result.top)
				result.top = (*segment)->Node(1).m_Coords[1];
			if((*segment)->Node(1).m_Coords[1] < result.bottom)
				result.bottom = (*segment)->Node(1).m_Coords[1];
		}

	return result;
}

bool CChannelDialog::HitTest(sdpVector2 Mouse, SegmentIterator& Segment, gulong& NodeIndex)
{
	const static gdouble tolerance = 6.0;

	// For each node ...
	for(SegmentIterator segment = m_Object->m_Segments.begin(); segment != m_Object->m_Segments.end(); segment++)
		{
			for(gulong i = 0; i < NODESPERSEGMENT; i++)
				{
					// Skip the very first and very last (unused) nodes ...
					if((segment == m_Object->m_Segments.begin()) && (0 == i))
						continue;
					if((segment == m_Object->m_Segments.end() - 1) && (2 == i))
						continue;

					// Check to see if we hit anything ...
					sdpVector2 coords = Screen((*segment)->Node(i).m_Coords);

					if((fabs(Mouse[0] - coords[0]) < tolerance) &&
						 (fabs(Mouse[1] - coords[1]) < tolerance))
						{
							Segment = segment;
							NodeIndex = i;
							return true;
						}
				}
		}

	return false;
}

void CChannelDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "onconfigurecurve" || Event->Name() == "onexposecurve")
		DrawCurve();
	else if(Event->Name() == "onconfigurecolor" || Event->Name() == "onexposecolor")
		DrawColor();
	else if(Event->Name() == "onmousemovecurve")
		OnMouseMove(Event);
	else if(Event->Name() == "onbuttondowncurve")
		OnButtonDown(Event);
	else if(Event->Name() == "onbuttonupcurve")
		OnButtonUp(Event);
	else if(Event->Name() == "onvscale")
		OnVScale();
	else if(Event->Name() == "onresetview")
		OnResetView();
	else if(Event->Name() == "onchannelstraighten")
		OnChannelStraighten();
	else if(Event->Name() == "onchannelmirrorx")
		OnChannelMirrorX();
	else if(Event->Name() == "onchannelmirrory")
		OnChannelMirrorY();
	else if(Event->Name() == "onsegmentstraighten")
		OnSegmentStraighten();
	else if(Event->Name() == "oncycles")
		OnCycles();
	else if(Event->Name() == "onphase")
		OnPhase();
	else
		k3dDialog::OnEvent(Event);
}

void CChannelDialog::OnLButtonClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	SegmentIterator segment;
	gulong nodeindex;
	
	if(HitTest(CurrentMouse, segment, nodeindex))
		{
			(*segment)->Node(nodeindex).m_Selected = true;
			(*segment)->m_Selected = true;

			SetCursorXY((*segment)->Node(nodeindex)[0], (*segment)->Node(nodeindex)[1]);
			RedrawAll();
		}
	else
		{
			sdpVector2 world = World(CurrentMouse);
			SetCursorX(world[0]);
			RedrawAll();
		}

//	RedrawAll();
}

void CChannelDialog::OnLButtonDoubleClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	sdpVector2 worldcoords = World(CurrentMouse);

	// See if the user clicked on an existing node, first ...
	SegmentIterator segment;
	gulong nodeindex;
	if(HitTest(CurrentMouse, segment, nodeindex))
		{
			SetCursorXY(worldcoords[0], worldcoords[1]);
			EditNode((*segment)->Node(nodeindex));
		}
	// Otherwise, create a new node ...
	else
		{
			SegmentIterator newsegment;
			CreateSegment(worldcoords, true, newsegment);

			SetCursorXY(worldcoords[0], worldcoords[1]);
			RedrawAll();
			
			EditNode((*newsegment)->Node(1));
	
			m_Object->NotifyChanged();
		}
}

void CChannelDialog::OnLButtonStartDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
}

void CChannelDialog::OnLButtonDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
	sdpVector2 delta = World(CurrentMouse) - World(LastMouse);

	SetCursorX(World(CurrentMouse)[0]);

	// For each node ...
	for(SegmentIterator segment = m_Object->m_Segments.begin(); segment != m_Object->m_Segments.end(); segment++)
		{
			for(int i = 0; i < NODESPERSEGMENT; i++)
				{
					if((*segment)->Node(i).m_Selected)
						MoveNode(segment, i, (*segment)->Node(i).m_Coords + delta);
				}
		}

	RedrawAll();
	m_Object->NotifyTracking();
}

void CChannelDialog::OnLButtonEndDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
	m_Object->NotifyChanged();
}

void CChannelDialog::OnRButtonClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	DeselectAll();
	RedrawAll();
}

void CChannelDialog::OnRButtonStartDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
}

void CChannelDialog::OnRButtonDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
	sdpVector2 delta = World(CurrentMouse) - World(LastMouse);
	m_Origin += delta;
	RedrawAll();
}

void CChannelDialog::OnRButtonEndDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
}

void CChannelDialog::OnLRButtonStartDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
}

void CChannelDialog::OnLRButtonDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
	gdouble zoomfactor = CurrentMouse[1] > LastMouse[1] ? pow(1.005, CurrentMouse[1] - LastMouse[1]) : pow(0.995, LastMouse[1] - CurrentMouse[1]);
	m_Scale *= zoomfactor;	
	RedrawAll();
}

void CChannelDialog::OnLRButtonEndDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
}

void CChannelDialog::OnVScale()
{
	m_VScale = pow(10.0, Adjustment("vscale").Value());
	RedrawAll();
}

void CChannelDialog::OnResetView()
{
	ResetView();
	RedrawAll();
}

void CChannelDialog::OnChannelStraighten()
{
	// If the curve only has one segment, we're done ...
	const gulong segmentcount = m_Object->m_Segments.size();
	if(1 == segmentcount)
		return;

	// Get the maximum extents of the curve ...
	sdpRectangle extents = NodeExtents();

	// Take a guess at whether the slope of the result should be positive or negative ...
	if(m_Object->m_Segments[0]->Node(1).m_Coords[1] > m_Object->m_Segments[segmentcount-1]->Node(1).m_Coords[1])
		std::swap(extents.top, extents.bottom);

	const sdpVector2 lowerleft = sdpVector2(extents.Left(), extents.Bottom());
	const sdpVector2 upperright = sdpVector2(extents.Right(), extents.Top());

	// For each segment ...
	gulong count = 0;
	for(SegmentIterator segment = m_Object->m_Segments.begin(); segment != m_Object->m_Segments.end(); segment++)
		{
			gdouble middledelta = gdouble(count) / gdouble(segmentcount-1);
			gdouble leftdelta = middledelta - (0.2 / gdouble(segmentcount-1));
			gdouble rightdelta = middledelta + (0.2 / gdouble(segmentcount-1));

			(*segment)->Node(0).m_Coords = sdpLerp(lowerleft, upperright, leftdelta);
			(*segment)->Node(1).m_Coords = sdpLerp(lowerleft, upperright, middledelta);
			(*segment)->Node(2).m_Coords = sdpLerp(lowerleft, upperright, rightdelta);

			count++;
		}

	RedrawAll();
}

void CChannelDialog::OnChannelMirrorX()
{
	// Reverse the segments ...
	std::reverse(m_Object->m_Segments.begin(), m_Object->m_Segments.end());

	// Reverse their nodes ...
	for(SegmentIterator segment = m_Object->m_Segments.begin(); segment != m_Object->m_Segments.end(); segment++)
		{
			std::swap((*segment)->Node(0), (*segment)->Node(2));

			for(int i = 0; i < NODESPERSEGMENT; i++)
				(*segment)->Node(i).m_Coords[0] = 1.0 - (*segment)->Node(i).m_Coords[0];
		}

	RedrawAll();
}

void CChannelDialog::OnChannelMirrorY()
{
	// Get the curve extents ...
	sdpRectangle extents = NodeExtents();
	gdouble reflection = extents.Center()[1];

	// For each segment ...
	for(SegmentIterator segment = m_Object->m_Segments.begin(); segment != m_Object->m_Segments.end(); segment++)
		{
			for(int i = 0; i < NODESPERSEGMENT; i++)
				(*segment)->Node(i).m_Coords[1] = (((*segment)->Node(i).m_Coords[1] - reflection) * -1.0) + reflection;
		}

	RedrawAll();
}

void CChannelDialog::OnSegmentStraighten()
{
	RedrawAll();
}

void CChannelDialog::OnCycles()
{
	if(UpdateControl(Entry("cycles"), m_Object->m_Cycles(), k3dIChannel::DATA_SCALAR, 3))
		RedrawAll();
}

void CChannelDialog::OnPhase()
{
	if(UpdateControl(Entry("phase"), m_Object->m_Phase(), k3dIChannel::DATA_ANGLE, 2))
		RedrawAll();
}

class CChannelColorChange
{
public:
	CChannelColorChange(CChannelDialog& Dialog, CChannelNode& Node) : m_Dialog(Dialog), m_Node(Node)
	{
		m_OriginalColor = m_Node.m_Value;
	}
	
	sdpVector3 OriginalColor() { return m_OriginalColor; }
	
	void HandleColor(sdpVector3 Color)
	{
		m_Node.m_Value = Color;
		m_Dialog.RedrawAll();
	}
	
protected:
	CChannelDialog& m_Dialog;
	CChannelNode& m_Node;
	sdpVector3 m_OriginalColor;
};

void CChannelDialog::EditNode(CChannelNode& Node)
{
	// If it isn't a value node, we're done ...
	if(!Node.m_ValueNode)
		return;
		
	// If it's a color ...
	if(k3dIChannel::DATA_COLOR == m_Object->GetDatatype())
		{
			// Prompt the user for a new color ...
			CChannelColorChange functor(*this, Node);
			sdpGtkFeedbackColorSelector<CChannelColorChange> selector("Select Color", Node.m_Value, functor);
			selector.DoModal();
			if(selector.Cancelled())
				{
					Node.m_Value = functor.OriginalColor();
					RedrawAll();
				}
		}
}

void CChannelDialog::ResetView()
{
	// Get the maximum extents of the curve ...
	sdpRectangle extents = Extents();

	// Make sure we can always see the X axis ...
	if(extents.bottom > 0)
		extents.bottom = 0;
		
	if(extents.top < 0)
		extents.top = 0;

	// Make sure we can handle curves that are straight across the origin ...
	if(extents.top == extents.bottom)
		extents.top += 1.0;

	// Initialize the scale & origin ...
	m_VScale = 1.0;
	m_Scale = sdpVector2(extents.Width(), extents.Height()) * 0.55;
	m_Origin = extents.Center();
}

void CChannelDialog::DrawCurve()
{
	// If we're not completely initialized, we're done ...
	if(!m_CurveWidget.Initialized())
		return;

	m_CurveWidget.Begin();

	// Get the widget sizes ...
	gulong width = m_CurveWidget.Width();
	gulong height = m_CurveWidget.Height();
	glViewport(0, 0, width, height);

	// Clear the background ...
	glClearColor(m_BackgroundColor[0], m_BackgroundColor[1], m_BackgroundColor[2], 1.0);
	glClear(GL_COLOR_BUFFER_BIT);

	// Setup the viewing projection ...
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(m_Frustum.Left(), m_Frustum.Right(), m_Frustum.Bottom(), m_Frustum.Top(), -1, 1);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glScaled(1.0 / m_Scale[0], 1.0 / m_Scale[1], 1.0);
	glTranslated(-m_Origin[0], -m_Origin[1], 0.0);
	glScaled(1, 1.0 / m_VScale, 1.0);

	// Setup drawing options ...
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
	glShadeModel(GL_FLAT);


	// Make it happen ...
	DrawGrid();
	DrawCursor();
	DrawTime();
	DrawPreview();
	DrawControlCurve();
	DrawTangents();
	DrawHandles();

	// Cleanup ...
	m_CurveWidget.SwapBuffers();
	m_CurveWidget.End();
}

void CChannelDialog::DrawGrid()
{
	glPushAttrib(GL_ALL_ATTRIB_BITS);

	glColor3dv(m_BackgroundColor * 0.8);

	// Draw the grid ...
	const static int xdivisions = 10;
	const static int ydivisions = 10;

	glBegin(GL_LINES);

	for(int i = 0; i <= xdivisions; i++)
		{
			glVertex2d(gdouble(i) / gdouble(xdivisions), 0.0);
			glVertex2d(gdouble(i) / gdouble(xdivisions), 1.0);
		}

	for(int j = 0; j <= ydivisions; j++)
		{
			glVertex2d(0.0, gdouble(j) / gdouble(ydivisions));
			glVertex2d(1.0, gdouble(j) / gdouble(ydivisions));
		}

	glEnd();

	gdouble left = World(sdpVector2(0, 0))[0];
	gdouble top = World(sdpVector2(0, 0))[1];
	gdouble right = World(sdpVector2(0, m_CurveWidget.Width()))[0];
	gdouble bottom = World(sdpVector2(0, m_CurveWidget.Height()))[1];

	// Draw the origin / axes ...
	glColor3dv(m_BackgroundColor * 0.6);
	glLineWidth(2.0f);

	glBegin(GL_LINES);

	glVertex2d(0.0, top);
	glVertex2d(0.0, bottom);

	glVertex2d(left, 0.0);
	glVertex2d(right, 0.0);

	glEnd();

	glPopAttrib();
}

void CChannelDialog::DrawPreview()
{
	// Don't do it if we don't need to!
	if((m_Object->m_Cycles() == 1.0) && (m_Object->m_Phase() == 0.0))
		return;

	glColor3dv(m_BackgroundColor * 0.7);

	glBegin(GL_LINE_STRIP);
						
	// Sample the curve ...
	const static gulong stepcount = 200;
	const static gdouble error = 0.5 / stepcount;
	
	for(gulong i = 0; i < stepcount; i++)
		{
			gdouble delta = gdouble(i) / gdouble(stepcount-1);
			glVertex2d(delta, m_Object->GetValue1(delta, error));
		}
		
	glEnd();
}

void CChannelDialog::DrawControlCurve()
{
	// Setup Bezier curve drawing
	const static int stepsize = 50;

	GLdouble controlpoints[4][3];
	memset(controlpoints, 0, sizeof(controlpoints));

	glEnable(GL_MAP1_VERTEX_3);

	// Draw curve
	SegmentIterator segment;
	for(segment = m_Object->m_Segments.begin(); segment != m_Object->m_Segments.end(); segment++)
		{
			// Draw initial straight line segment before first node
			if(m_Object->m_Segments.begin() == segment)
				{
					// Select a color based on whether the segment is selected or not ...
					glColor3dv((*segment)->m_Selected ? m_SelectedColor : m_NormalColor);

					glBegin(GL_LINES);
					glVertex2d(-10000.0, (*segment)->Node(1).m_Coords[1]);
					glVertex2d((*segment)->Node(1).m_Coords[0], (*segment)->Node(1).m_Coords[1]);
					glEnd();
				}
			// Draw each segment
			else
				{
					// Select a color based on whether the segment is selected or not ...
					glColor3dv((*segment)->m_Selected ? m_SelectedColor : m_NormalColor);

					controlpoints[0][0] = (*(segment-1))->Node(1).m_Coords[0];
					controlpoints[0][1] = (*(segment-1))->Node(1).m_Coords[1];
					controlpoints[1][0] = (*(segment-1))->Node(2).m_Coords[0];
					controlpoints[1][1] = (*(segment-1))->Node(2).m_Coords[1];
					controlpoints[2][0] = (*segment)->Node(0).m_Coords[0];
					controlpoints[2][1] = (*segment)->Node(0).m_Coords[1];
					controlpoints[3][0] = (*segment)->Node(1).m_Coords[0];
					controlpoints[3][1] = (*segment)->Node(1).m_Coords[1];

					glMap1d(GL_MAP1_VERTEX_3, 0.0, 1.0, 3, 4, &controlpoints[0][0]);

					glBegin(GL_LINE_STRIP);
					for(int j = 0; j <= stepsize; j++)
						glEvalCoord1d(GLdouble(j) / GLdouble(stepsize));
					glEnd();
				}
		}

	// Draw straight line segment after last node
	segment = m_Object->m_Segments.end() - 1;

	// Select a color based on whether the segment is selected or not ...
	glColor3dv((*segment)->m_Selected ? m_SelectedColor : m_NormalColor);

	glBegin(GL_LINES);
	glVertex2d((*segment)->Node(1).m_Coords[0], (*segment)->Node(1).m_Coords[1]);
	glVertex2d(10000.0, (*segment)->Node(1).m_Coords[1]);
	glEnd();
}

void CChannelDialog::DrawTangents()
{
	glColor3dv(m_TangentColor);
	glBegin(GL_LINES);

	for(SegmentIterator segment = m_Object->m_Segments.begin() + 1; segment != m_Object->m_Segments.end(); segment++)
		{
			glVertex2d((*(segment-1))->Node(1).m_Coords[0], (*(segment-1))->Node(1).m_Coords[1]);
			glVertex2d((*(segment-1))->Node(2).m_Coords[0], (*(segment-1))->Node(2).m_Coords[1]);

			glVertex2d((*segment)->Node(1).m_Coords[0], (*segment)->Node(1).m_Coords[1]);
			glVertex2d((*segment)->Node(0).m_Coords[0], (*segment)->Node(0).m_Coords[1]);
		}

	glEnd();
}

void CChannelDialog::DrawHandles()
{
	// Set ourselves up to draw some big round points ...
	glEnable(GL_POINT_SMOOTH);
	glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);

	// Display the first point ...
	SegmentIterator segment = m_Object->m_Segments.begin();
	DrawControlPoint((*segment)->Node(1));

	// Display the rest ...
	for(segment++; segment != m_Object->m_Segments.end(); segment++)
		{
			DrawControlPoint((*(segment-1))->Node(1));
			DrawTangentPoint((*(segment-1))->Node(2));
			DrawTangentPoint((*segment)->Node(0));
			DrawControlPoint((*segment)->Node(1));
		}

	glEnd();
}

void CChannelDialog::DrawControlPoint(CChannelNode& Node)
{
	bool valuenode = Node.m_ValueNode && (k3dIChannel::DATA_COLOR == m_Object->GetDatatype());
	
	glPointSize(valuenode ? 6.0 : 5.0);
	glColor3dv(Node.m_Selected ? m_SelectedColor : valuenode ? m_ValueNodeColor : m_NormalColor);

	glBegin(GL_POINTS);
	glVertex2dv(Node.m_Coords);
	glEnd();
}

void CChannelDialog::DrawTangentPoint(CChannelNode& Node)
{
	glPointSize(4.0);
	glColor3dv(Node.m_Selected ? m_SelectedColor : m_TangentColor);

	glBegin(GL_POINTS);
	glVertex2dv(Node.m_Coords);
	glEnd();
}

void CChannelDialog::DrawCursor()
{
	gdouble top = World(sdpVector2(0, 0))[1];
	gdouble bottom = World(sdpVector2(0, m_CurveWidget.Height()))[1];

	glColor3dv(m_CursorColor);
	glBegin(GL_LINES);

	// Vertical cursor ...
	glVertex2d(m_Cursor[0], top);
	glVertex2d(m_Cursor[0], bottom);

	glEnd();
}

void CChannelDialog::DrawTime()
{
	gdouble top = World(sdpVector2(0, 0))[1];
	gdouble bottom = World(sdpVector2(0, m_CurveWidget.Height()))[1];

	glColor3dv(m_TimeColor);
	glBegin(GL_LINES);

	// Vertical cursor ...
	glVertex2d(m_Object->m_CurrentTime, top);
	glVertex2d(m_Object->m_CurrentTime, bottom);

	glEnd();
}

void CChannelDialog::DrawColor()
{
	// If we're not completely initialized, we're done ...
	if(!m_ColorWidget.Initialized())
		return;

	m_ColorWidget.Begin();

	// Get the widget sizes ...
	gulong width = m_ColorWidget.Width();
	gulong height = m_ColorWidget.Height();
	glViewport(0, 0, width, height);

	// Setup the viewing projection ...
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(m_Frustum.Left(), m_Frustum.Right(), -1, 1, -1, 1);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glScaled(1.0 / m_Scale[0], 1.0, 1.0);
	glTranslated(-m_Origin[0], 0.0, 0.0);

	// Setup drawing options ...
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
	glShadeModel(GL_FLAT);

	// Make it happen ...
	glBegin(GL_QUADS);
	
	sdpVector3 color1;
	gdouble error = 0.01;
	gulong granularity = 2;

	gdouble x2;
	gdouble x1 = World(sdpVector2(0, 0))[0];		
	for(gulong i = 0; i < width+granularity; i += granularity)
		{
			x2 = World(sdpVector2(gdouble(i+1), 0.0))[0];		
		
			m_Object->GetValue3(x2, error, color1);

			glColor3dv(color1);
			glVertex2d(x1, -1);
			glVertex2d(x1, 1);
			glVertex2d(x2, 1);
			glVertex2d(x2, -1);
			
			x1 = x2;
		}

	glEnd();

	// Cleanup ...
	m_ColorWidget.SwapBuffers();
	m_ColorWidget.End();
}

void CChannelDialog::DeselectAll()
{
	for(SegmentIterator segment = m_Object->m_Segments.begin(); segment != m_Object->m_Segments.end(); segment++)
		{
			(*segment)->m_Selected = false;

			for(int i = 0; i < NODESPERSEGMENT; i++)
				(*segment)->Node(i).m_Selected = false;
		}
}

void CChannelDialog::RedrawAll()
{
	Widget("curve").QueueDraw();
	Widget("color").QueueDraw();
}

void CChannelDialog::SetCursorX(gdouble X)
{
	// Constrain the cursor position
	if(X < 0.0)
		X = 0.0;
	if(X > 1.0)
		X = 1.0;

	m_Cursor[0] = X;

	Entry("cursorx").SetText(sdpFormatString("%.4f", m_Cursor[0]));
	Entry("cursory").SetText("******");
	Entry("cursorvalue").SetText(sdpFormatString("%.4f", m_Object->GetValue1(m_Cursor[0], m_Object->m_DeltaTime)));
}

void CChannelDialog::SetCursorXY(gdouble X, gdouble Y)
{
	// Constrain the cursor position
	if(X < 0.0)
		X = 0.0;
	if(X > 1.0)
		X = 1.0;

	m_Cursor[0] = X;
	m_Cursor[1] = Y;

	Entry("cursorx").SetText(sdpFormatString("%.4f", m_Cursor[0]));
	Entry("cursory").SetText(sdpFormatString("%.4f", m_Cursor[1]));
	Entry("cursorvalue").SetText(sdpFormatString("%.4f", m_Object->GetValue1(m_Cursor[0], m_Object->m_DeltaTime)));
}

void CChannelDialog::MoveNode(const SegmentIterator& Segment, const int Node, gdouble X, gdouble Y)
{
	MoveNode(Segment, Node, sdpVector2(X, Y));
}

void CChannelDialog::MoveNode(const SegmentIterator& Segment, const int Node, sdpVector2 Coords)
{
	// Get previous and next segments ...
	SegmentIterator prev = Segment;
	SegmentIterator next = Segment;

	prev--;
	next++;

	// Constrain the new coordinates
	if(Coords[0] < m_Object->m_MinCoords()[0])
		Coords[0] = m_Object->m_MinCoords()[0];
	if(Coords[0] > m_Object->m_MaxCoords()[0])
		Coords[0] = m_Object->m_MaxCoords()[0];
	if(Coords[1] < m_Object->m_MinCoords()[1])
		Coords[1] = m_Object->m_MinCoords()[1];
	if(Coords[1] > m_Object->m_MaxCoords()[1])
		Coords[1] = m_Object->m_MaxCoords()[1];

	// If this is the node point, restrict its' neighboring tangent vectors ...
	if(Node == 1)
		{
			if(prev != m_Object->m_Segments.begin()-1)
				{
					if(Coords[0] < (*prev)->Node(1)[0])
						MoveNode(prev, 1, Coords[0], (*prev)->Node(1)[1]);

					if(Coords[0] < (*prev)->Node(2)[0])
						MoveNode(prev, 2, Coords[0], (*prev)->Node(2)[1]);
				}

			if(Coords[0] < (*Segment)->Node(0)[0])
				MoveNode(Segment, 0, Coords[0], (*Segment)->Node(0)[1]);

			if(Coords[0] > (*Segment)->Node(2)[0])
				MoveNode(Segment, 2, Coords[0], (*Segment)->Node(2)[1]);

			if(next != m_Object->m_Segments.end())
				{
					if(Coords[0] > (*next)->Node(1)[0])
						MoveNode(next, 1, Coords[0], (*next)->Node(1)[1]);

					if(Coords[0] > (*next)->Node(0)[0])
						MoveNode(next, 0, Coords[0], (*next)->Node(0)[1]);
				}
		}
	// If this is the lefthand control point, restrict its' x value
	else if(Node == 0)
		{
			if(prev != m_Object->m_Segments.begin()-1)
				if(Coords[0] < (*prev)->Node(1)[0])
					Coords[0] = (*prev)->Node(1)[0];

			if(Coords[0] > (*Segment)->Node(1)[0])
				Coords[0] = (*Segment)->Node(1)[0];
		}
	// If this is the righthand control point, restrict its' x value
	else if(Node == 2)
		{
			if(Coords[0] < (*Segment)->Node(1)[0])
				Coords[0] = (*Segment)->Node(1)[0];

			if(next != m_Object->m_Segments.end())
				if(Coords[0] > (*next)->Node(1)[0])
					Coords[0] = (*next)->Node(1)[0];
		}

	(*Segment)->Node(Node) = Coords;
}

void CChannelDialog::CreateSegment(sdpVector2 Coords, bool AutoSmooth, SegmentIterator& Result)
{
	// Constrain the new coordinates
	if(Coords[0] < m_Object->m_MinCoords()[0])
		Coords[0] = m_Object->m_MinCoords()[0];
	if(Coords[0] > m_Object->m_MaxCoords()[0])
		Coords[0] = m_Object->m_MaxCoords()[0];
	if(Coords[1] < m_Object->m_MinCoords()[1])
		Coords[1] = m_Object->m_MinCoords()[1];
	if(Coords[1] > m_Object->m_MaxCoords()[1])
		Coords[1] = m_Object->m_MaxCoords()[1];

	// Find the node to the right of our insertion spot ...
	SegmentIterator right;
	for(right = m_Object->m_Segments.begin(); right != m_Object->m_Segments.end(); right++)
		if((*right)->Node(1)[0] > Coords[0])
			break;

	// Insert the new node ...
	CChannelSegment* newsegment = new CChannelSegment;
	SegmentIterator item = m_Object->m_Segments.insert(right, newsegment);
	SegmentIterator left = right = item;
	left--;
	right++;

	// If this is a color, set values appropriately
	if(k3dIChannel::DATA_COLOR == m_Object->m_DataType())
		{
			Coords[1] = (Coords[1] >= 0.5) ? 1.0 : 0.0;

			(*item)->Node(1).m_Value = sdpVector3(1.0);
		}
	else
		{
			(*item)->Node(1).m_Value = sdpVector3(Coords[1]);
		}

	(*item)->Node(1).m_ValueNode = true;
	(*item)->Node(0).m_Coords = (*item)->Node(1).m_Coords = (*item)->Node(2).m_Coords = Coords;


	// Adjust neighbors and tangent vectors ...
	if(AutoSmooth)
		{
			if(left != m_Object->m_Segments.end())
				{
					(*left)->Node(2)[0] = sdpLerp<double>((*left)->Node(1)[0], (*item)->Node(1)[0], 0.2);
					(*item)->Node(0)[0] = sdpLerp<double>((*left)->Node(1)[0], (*item)->Node(1)[0], 0.8);
				}
			if(right != m_Object->m_Segments.end())
				{
					(*item)->Node(2)[0] = sdpLerp<double>((*item)->Node(1)[0], (*right)->Node(1)[0], 0.2);
					(*right)->Node(0)[0] = sdpLerp<double>((*item)->Node(1)[0], (*right)->Node(1)[0], 0.8);
				}
		}
	else
		{
			if(left != m_Object->m_Segments.end())
				{
					(*left)->Node(2) = sdpLerp<sdpVector2>((*left)->Node(1), (*item)->Node(1), 0.2);
					(*item)->Node(0) = sdpLerp<sdpVector2>((*left)->Node(1), (*item)->Node(1), 0.8);
				}
			if(right != m_Object->m_Segments.end())
				{
					(*item)->Node(2) = sdpLerp<sdpVector2>((*item)->Node(1), (*right)->Node(1), 0.2);
					(*right)->Node(0) = sdpLerp<sdpVector2>((*item)->Node(1), (*right)->Node(1), 0.8);
				}
		}
		
	Result = item;
}


