// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CEditorEngine class, which is responsible for rendering editor-mode graphics into a camera
		\author Tim Shead (tshead@k-3d.com)
*/

//#include "editorengine.h"
#include "camera.h"

#include <k3dichannel.h>
#include <k3didocument.h>
#include <k3dieditorlight.h>
#include <k3dieditormaterial.h>
#include <k3dieditortexture.h>

#include <k3dmisc.h>

//////////////////////////////////////////////////////////////////////////////////////
// CEditorEngine

CEditorEngine::CEditorEngine()
{
	m_Camera = 0;
}

CEditorEngine::~CEditorEngine()
{
}

/////////////////////////////////////////////////////////////////
// k3dIRenderEngine overrides ...

sdpResult CEditorEngine::SetCamera(k3dICamera* Camera)
{
	// Sanity checks ...
	g_return_val_if_fail(Camera, SDP_E_FAIL);

	m_Camera = (CCamera*)Camera;

	return SDP_S_OK;
}

sdpResult CEditorEngine::Redraw()
{
	// Sanity checks ...
	g_return_val_if_fail(m_Camera, SDP_E_FAIL);

	// Render ...
	m_Camera->Viewport().Begin();
	EditorRender();
	m_Camera->Viewport().SwapBuffers();
	m_Camera->Viewport().End();

	return SDP_S_OK;
}

sdpResult CEditorEngine::StartJob(k3dIRenderJob* Job)
{
	// Sanity checks ...
	g_return_val_if_fail(m_Camera, SDP_E_FAIL);
	g_return_val_if_fail(Job, SDP_E_FAIL);

	g_return_val_if_fail(0, SDP_E_FAIL);
}

sdpResult CEditorEngine::QueueFrame(k3dIRenderFrame* Frame)
{
	// Sanity checks ...
	g_return_val_if_fail(m_Camera, SDP_E_FAIL);
	g_return_val_if_fail(Frame, SDP_E_FAIL);

	g_return_val_if_fail(0, SDP_E_FAIL);
}

sdpResult CEditorEngine::FinishJob(k3dIRenderJob* Job)
{
	// Sanity checks ...
	g_return_val_if_fail(m_Camera, SDP_E_FAIL);
	g_return_val_if_fail(Job, SDP_E_FAIL);

	g_return_val_if_fail(0, SDP_E_FAIL);
}

///////////////////////////////////////////////////////////////
// k3dIEditorEngine overrides ...

k3dICamera* CEditorEngine::Camera()
{
	return m_Camera;
}

sdpResult CEditorEngine::EditorQuerySolid()
{
	return m_Camera->m_EditorSolid() ? SDP_S_OK : SDP_S_FALSE;
}

sdpResult CEditorEngine::EditorQueryShaded()
{
	return m_Camera->m_EditorSmoothing() ? SDP_S_OK : SDP_S_FALSE;
}

sdpResult CEditorEngine::EditorQueryTextured()
{
	return m_Camera->m_EditorTexture() ? SDP_S_OK : SDP_S_FALSE;
}

sdpResult CEditorEngine::EditorQueryOrientation()
{
	return m_Camera->m_EditorOrientation() ? SDP_S_OK : SDP_S_FALSE;
}

sdpResult CEditorEngine::EditorQueryNormals()
{
	return m_Camera->m_EditorNormals() ? SDP_S_OK : SDP_S_FALSE;
}

gulong CEditorEngine::EditorQueryWireframePatchDivisions()
{
	return m_Camera->m_EditorWireframePatchDivisions();
}

gulong CEditorEngine::EditorQuerySolidPatchDivisions()
{
	return m_Camera->m_EditorSolidPatchDivisions();
}

///////////////////////////////////////////////////////////////
// Helper functions ...

void CEditorEngine::EditorRender()
{
	sdpglTrace("CEditorEngine::EditorRender()");

	RenderOptions();
	PreRender();
	Render();
	PostRender();

	glFlush();
}


void CEditorEngine::RenderOptions()
{
	sdpglTrace("CEditorEngine::RenderOptions()");

	// Set point size ...
	glPointSize(m_Camera->m_PointSize);

	// Setup stippling for polygon normals / bezier control lines ...
	glLineStipple(2, 0xaaaa);
	glDisable(GL_LINE_STIPPLE);

	// Setup culling ...
	glFrontFace(GL_CW);
	if(m_Camera->m_EditorTwoSided())
		{
			glDisable(GL_CULL_FACE);
		}
	else
		{
			glCullFace(GL_FRONT);
			glEnable(GL_CULL_FACE);
		}

	// Setup dithering ...
	glDisable(GL_DITHER);

	// Disable 1D textures ...
	glDisable(GL_TEXTURE_1D);

	// Setup antialiasing ...
	glDisable(GL_LINE_SMOOTH);
	glDisable(GL_BLEND);
	glLineWidth(1.0f);

	// Disable stencil ...
	glDisable(GL_STENCIL_TEST);
	glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);
	glStencilMask(0x00);

	// Prepare bezier options ...
	glEnable(GL_MAP1_VERTEX_3);
	glEnable(GL_MAP2_VERTEX_3);

	// Calculate current time & animate the document ...
	m_Camera->m_TimeBase->SetRenderTime(m_Camera->m_TimeBase->GetViewTime());
	m_Camera->m_TimeBase->SetRenderDeltaTime(1.0 / m_Camera->m_TimeBase->GetDuration() * m_Camera->m_TimeBase->GetFrameRate());
	m_Camera->Document()->AnimateDocument();

	// Compute viewport variables ...
	SetupViewVolume(0.0, 0.0, 0.0, 0.0);
	glViewport(0, 0, m_Camera->Width(), m_Camera->Height());

	// Clear background ...
	glClearDepth(1.0);
	glClearColor(GLfloat(m_Camera->m_EditorBackgroundColor[0]), GLfloat(m_Camera->m_EditorBackgroundColor[1]), GLfloat(m_Camera->m_EditorBackgroundColor[2]), 0.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// Setup fog ...
	if(m_Camera->m_EditorFog())
		{
/*
			glEnable(GL_FOG);

			GLfloat fogdata[4];
			fogdata[0] = (GLfloat)m_Camera->m_EditorBackgroundColor[0];
			fogdata[1] = (GLfloat)m_Camera->m_EditorBackgroundColor[1];
			fogdata[2] = (GLfloat)m_Camera->m_EditorBackgroundColor[2];
			fogdata[3] = 1.0f;

			glFogfv(GL_FOG_COLOR, fogdata);
			glFogf(GL_FOG_START, (GLfloat)m_Camera->m_Fogstart->GetValue1(m_Camera->m_CurrentTime, m_Camera->m_DeltaTime));
			glFogf(GL_FOG_END, (GLfloat)m_Camera->m_Fogend->GetValue1(m_Camera->m_CurrentTime, m_Camera->m_DeltaTime));
			glHint(GL_FOG_HINT, GL_NICEST);

			glFogi(GL_FOG_MODE, GL_LINEAR);
*/
		}
	else
		{
			glDisable(GL_FOG);
		}

	if(m_Camera->m_EditorSolid())
		RenderSolidOptions();
	else
		RenderWireframeOptions();
}

void CEditorEngine::RenderWireframeOptions()
{
	sdpglTrace("CEditorEngine::RenderWireframeOptions()");

	// Set wireframe rendering ...
	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

	// Disable lighting ...
	glDisable(GL_LIGHTING);

	// Set polygon shading ...
	glShadeModel(GL_FLAT);

	// Set Z buffer options ...
	glDisable(GL_DEPTH_TEST);
	glDepthMask(GL_FALSE);

	// Normalization ...
	glDisable(GL_NORMALIZE);

	// If we're supposed to draw polygon normals, make sure they're up-to-date ...
	if(m_Camera->m_EditorNormals())
		k3dCalculateNormals(m_Camera->Document());

	// Disable textures ...
	glDisable(GL_TEXTURE_2D);
}

void CEditorEngine::RenderSolidOptions()
{
	sdpglTrace("CEditorEngine::RenderSolidOptions()");

	// Set solid rendering ...
	glDisable(GL_POLYGON_STIPPLE);
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

	// Enable polygon shading ...
	glShadeModel(GL_SMOOTH);

	// Set Z buffer options ...
	glDepthMask(GL_TRUE);
	glDepthFunc(GL_LESS);
	glEnable(GL_DEPTH_TEST);

	// Normalization ...
	glEnable(GL_NORMALIZE);

	// Prepare additional bezier options ...
	glEnable(GL_MAP2_TEXTURE_COORD_2);
	glEnable(GL_AUTO_NORMAL);

	// Calculate normal vectors ...
	k3dCalculateNormals(m_Camera->Document());

	// Prepare texture options ...
	if(m_Camera->m_EditorTexture())
		glEnable(GL_TEXTURE_2D);
	else
		glDisable(GL_TEXTURE_2D);

	LightSetup();
}

class CEditorEngineEditorPreRender
{
public:
	CEditorEngineEditorPreRender(k3dIEditorEngine* Engine) { engine = Engine; }

	inline bool HandleObject(k3dIObjectContext* ObjectContext)
	{
		if(SDP_S_OK != ObjectContext->State()->QueryState(k3dIState::VISIBLE))
			return true;

		render = sdpQueryInterface<k3dIEditorRender>(ObjectContext->Object());
		if(render)
			{
				render->EditorPreRender(engine);
				render->Release();
			}

		light = sdpQueryInterface<k3dIEditorLight>(ObjectContext->Object());
		if(light)
			{
				light->EditorLightPreRender(engine);
				light->Release();
			}

		texture = sdpQueryInterface<k3dIEditorTexture>(ObjectContext->Object());
		if(texture)
			{
				texture->EditorTexturePreRender(engine);
				texture->Release();
			}

		material = sdpQueryInterface<k3dIEditorMaterial>(ObjectContext->Object());
		if(material)
			{
				material->EditorMaterialPreRender(engine);
				material->Release();
			}

		return true;
	}

	k3dIEditorEngine* engine;
	k3dIEditorRender* render;
	k3dIEditorLight* light;
	k3dIEditorTexture* texture;
	k3dIEditorMaterial* material;
};

void CEditorEngine::PreRender()
{
	sdpglTrace("CEditorEngine::PreRender()");

	CEditorEngineEditorPreRender functor(this);
	k3dForEachObject(m_Camera->Document(), 0, functor);
}

class CEditorEngineEditorLightSetup
{
public:
	CEditorEngineEditorLightSetup(k3dIEditorEngine* Engine) { engine = Engine; light = 0; lightindex = 0; }

	inline bool HandleObject(k3dIObjectContext* ObjectContext)
	{
		if(SDP_S_OK != ObjectContext->State()->QueryState(k3dIState::VISIBLE))
			return true;

		light = sdpQueryInterface<k3dIEditorLight>(ObjectContext->Object());
		if(light)
			{
				light->EditorLightSetup(engine, lightindex++);
				light->Release();
			}

		return true;
	}

	k3dIEditorEngine* engine;
	k3dIEditorLight* light;
	gulong lightindex;
};

void CEditorEngine::LightSetup()
{
	sdpglTrace("CEditorEngine::LightSetup()");

	// Enable lighting ...
	glEnable(GL_LIGHTING);
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
	glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);

	// Make sure all lights are turned off initially ...
	GLint maxlights;
	glGetIntegerv(GL_MAX_LIGHTS, &maxlights);
	for(GLint i = 0; i < maxlights; i++)
		glDisable(GLenum(GL_LIGHT0 + i));

	CEditorEngineEditorLightSetup functor(this);
	k3dForEachObject(m_Camera->Document(), 0, functor);
}

class CEditorEngineEditorRender
{
public:
	CEditorEngineEditorRender(k3dIEditorEngine* Engine) { engine = Engine; }

	inline bool HandleObject(k3dIDocument::TreeNodeType Type, k3dIObjectContext* ObjectContext)
	{
		if(SDP_S_OK != ObjectContext->State()->QueryState(k3dIState::VISIBLE))
			return true;

		render = sdpQueryInterface<k3dIEditorRender>(ObjectContext->Object());
		if(render)
			{
				switch(Type)
					{
						case k3dIDocument::PUSH:
							render->EditorPushRender(engine);
							break;

						case k3dIDocument::NODE:
							render->EditorRender(engine);
							render->EditorPushChildRender(engine);
							break;

						case k3dIDocument::POP:
							render->EditorPopChildRender(engine);
							render->EditorPopRender(engine);
							break;

						default:
							g_assert_not_reached();
					}

				render->Release();
			}

		return true;
	}

	k3dIEditorEngine* engine;
	k3dIEditorRender* render;
};

void CEditorEngine::Render()
{
	sdpglTrace("CEditorEngine::Render()");

	CEditorEngineEditorRender functor(this);
	k3dForEachObjectTree(m_Camera->Document(), 0, functor);
}

class CEditorEngineEditorPostRender
{
public:
	CEditorEngineEditorPostRender(k3dIEditorEngine* Engine) { engine = Engine; }

	inline bool HandleObject(k3dIObjectContext* ObjectContext)
	{
		if(SDP_S_OK != ObjectContext->State()->QueryState(k3dIState::VISIBLE))
			return true;

		render = sdpQueryInterface<k3dIEditorRender>(ObjectContext->Object());
		if(render)
			{
				render->EditorPostRender(engine);
				render->Release();
			}

		return true;
	}

	k3dIEditorEngine* engine;
	k3dIEditorRender* render;
};

void CEditorEngine::PostRender()
{
	sdpglTrace("CEditorEngine::PostRender()");

	CEditorEngineEditorPostRender functor(this);
	k3dForEachObject(m_Camera->Document(), 0, functor);
}

void CEditorEngine::SetupViewVolume(GLdouble PixelDX, GLdouble PixelDY, GLdouble EyeDX, GLdouble EyeDY)
{
	sdpglTrace("CEditorEngine::SetupViewVolume()");

	GLdouble xwsize, ywsize;
	double dx, dy;

	// Adjust depth-of-field offsets for isometric / perspective
	if(m_Camera->m_CameraIsometric())
		{
			EyeDX = 0.0;
			EyeDY = 0.0;
		}
	else
		{
			EyeDX *= m_Camera->m_CurrentFStop;
			EyeDY *= m_Camera->m_CurrentFStop;
		}

	// Get viewport dimensions
	gulong windowwidth = m_Camera->Width();
	gulong windowheight = m_Camera->Height();

	// Setup texture matrix
	glMatrixMode(GL_TEXTURE);
	glLoadIdentity();

	// Setup projection matrix
	xwsize = 2.0 * m_Camera->m_CurrentWidthPlane;
	ywsize = 2.0 * m_Camera->m_CurrentHeightPlane;

	dx = -(PixelDX * xwsize / (GLdouble)windowwidth + EyeDX * m_Camera->m_CurrentNearPlane / m_Camera->m_CurrentFocalPlane);
	dy = -(PixelDY * ywsize / (GLdouble)windowheight + EyeDY * m_Camera->m_CurrentNearPlane / m_Camera->m_CurrentFocalPlane);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	if(m_Camera->m_CameraIsometric())
		{
			glOrtho(-m_Camera->m_CurrentWidthPlane + dx,
							m_Camera->m_CurrentWidthPlane + dx,
							-m_Camera->m_CurrentHeightPlane + dy,
							m_Camera->m_CurrentHeightPlane + dy,
							m_Camera->m_CurrentNearPlane,
							m_Camera->m_CurrentFarPlane);
		}
	else
		{
			glFrustum(-m_Camera->m_CurrentWidthPlane + dx,
								m_Camera->m_CurrentWidthPlane + dx,
								-m_Camera->m_CurrentHeightPlane + dy,
								m_Camera->m_CurrentHeightPlane + dy,
								m_Camera->m_CurrentNearPlane,
								m_Camera->m_CurrentFarPlane);
		}

	// Setup modelview matrix
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	glTranslated(EyeDX, EyeDY, 0.0);

	glScaled(1.0, 1.0, -1.0);

	glRotated(-sdpDegrees(m_Camera->m_CurrentOrientation.angle), m_Camera->m_CurrentOrientation.axis[0], m_Camera->m_CurrentOrientation.axis[1], m_Camera->m_CurrentOrientation.axis[2]);
	glTranslated(-m_Camera->m_CurrentPosition[0], -m_Camera->m_CurrentPosition[1], -m_Camera->m_CurrentPosition[2]);
}


