// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CRiEngine class, which can render the 3D workspace to a RIB file
		\author Tim Shead (tshead@k-3d.com)
*/

//#include "riengine.h"
#include "camera.h"

#include <k3diapplication.h>
#include <k3dirilight.h>
#include <k3diritexture.h>
#include <k3dirimaterial.h>
#include <k3dichannel.h>
#include <k3direnderjob.h>
#include <k3direnderframe.h>

#include <k3dmisc.h>

#include <sdpfile.h>

#include "../../k3d_version.h"

/////////////////////////////////////////////////////
// CRiEngine
 
CRiEngine::CRiEngine()
{
	m_Camera = 0;
}


CRiEngine::~CRiEngine()
{
}

///////////////////////////////////////////////////////////////
// Helper functions ...

// Functor for calling StartJob() ...
class CRiEngineRiStartJob
{
public:
	CRiEngineRiStartJob(k3dIRenderJob* Job, k3dIRiEngine* Engine) { m_Job = Job; m_Engine = Engine; }
	bool HandleObject(k3dIObjectContext* Context)
	{
		// Skip hidden objects ...
		if(SDP_S_OK != Context->State()->QueryState(k3dIState::VISIBLE))
			return true;

		k3dIRiRender* render = sdpQueryInterface<k3dIRiRender>(Context->Object());
		if(render)
			{
				render->RiStartJob(m_Job, m_Engine);
				render->Release();
			}

		k3dIRiLight* light = sdpQueryInterface<k3dIRiLight>(Context->Object());
		if(light)
			{
				light->RiLightStartJob(m_Job, m_Engine);
				light->Release();
			}

		k3dIRiTexture* texture = sdpQueryInterface<k3dIRiTexture>(Context->Object());
		if(texture)
			{
				texture->RiTextureStartJob(m_Job, m_Engine);
				texture->Release();
			}

		k3dIRiMaterial* material = sdpQueryInterface<k3dIRiMaterial>(Context->Object());
		if(material)
			{
				material->RiMaterialStartJob(m_Job, m_Engine);
				material->Release();
			}

		return true;
	}
	
	k3dIRenderJob* m_Job;
	k3dIRiEngine* m_Engine;
};


// Functor for calling FinishJob() ...
class CRiEngineRiFinishJob
{
public:
	CRiEngineRiFinishJob(k3dIRenderJob* Job, k3dIRiEngine* Engine) { m_Job = Job; m_Engine = Engine; }
	bool HandleObject(k3dIObjectContext* Context)
	{
		// Skip hidden objects ...
		if(SDP_S_OK != Context->State()->QueryState(k3dIState::VISIBLE))
			return true;

		k3dIRiRender* render = sdpQueryInterface<k3dIRiRender>(Context->Object());
		if(render)
			{
				render->RiFinishJob(m_Job, m_Engine);
				render->Release();
			}

		k3dIRiLight* light = sdpQueryInterface<k3dIRiLight>(Context->Object());
		if(light)
			{
				light->RiLightFinishJob(m_Job, m_Engine);
				light->Release();
			}

		k3dIRiTexture* texture = sdpQueryInterface<k3dIRiTexture>(Context->Object());
		if(texture)
			{
				texture->RiTextureFinishJob(m_Job, m_Engine);
				texture->Release();
			}

		k3dIRiMaterial* material = sdpQueryInterface<k3dIRiMaterial>(Context->Object());
		if(material)
			{
				material->RiMaterialFinishJob(m_Job, m_Engine);
				material->Release();
			}

		return true;
	}
	
	k3dIRenderJob* m_Job;
	k3dIRiEngine* m_Engine;
};

// Functor for pre-rendering objects ...
class CRiEngineRiPreRender
{
public:
	CRiEngineRiPreRender(k3dIRenderFrame* Frame, k3dIRiEngine* Engine) { m_Frame =Frame; m_Engine = Engine; }
	bool HandleObject(k3dIObjectContext* Context)
	{
		// Skip hidden objects ...
		if(SDP_S_OK != Context->State()->QueryState(k3dIState::VISIBLE))
			return true;

		k3dIRiRender* render = sdpQueryInterface<k3dIRiRender>(Context->Object());
		if(render)
			{
				render->RiSetSolid(false);
				render->RiPreRender(m_Frame, m_Engine);
				render->Release();
			}

		k3dIRiLight* light = sdpQueryInterface<k3dIRiLight>(Context->Object());
		if(light)
			{
				light->RiLightPreRender(m_Frame, m_Engine);
				light->Release();
			}

		k3dIRiTexture* texture = sdpQueryInterface<k3dIRiTexture>(Context->Object());
		if(texture)
			{
				texture->RiTexturePreRender(m_Frame, m_Engine);
				texture->Release();
			}

		k3dIRiMaterial* material = sdpQueryInterface<k3dIRiMaterial>(Context->Object());
		if(material)
			{
				material->RiMaterialPreRender(m_Frame, m_Engine);
				material->Release();
			}

		return true;
	}
	
	k3dIRenderFrame* m_Frame;
	k3dIRiEngine* m_Engine;
};


// Functor for setting-up lights ...
class CRiEngineRiLightSetup
{
public:
	CRiEngineRiLightSetup(k3dIRenderFrame* Frame, k3dIRiEngine* Engine) { m_Frame = Frame; m_Engine = Engine; }
	bool HandleObject(k3dIObjectContext* Context)
	{
		// Skip hidden objects ...
		if(SDP_S_OK != Context->State()->QueryState(k3dIState::VISIBLE))
			return true;

		k3dIRiLight* light = sdpQueryInterface<k3dIRiLight>(Context->Object());
		if(light)
			{
				light->RiLightSetup(m_Frame, m_Engine);
				light->Release();
			}

		return true;
	}
	
	k3dIRenderFrame* m_Frame;
	k3dIRiEngine* m_Engine;
};


// Render all objects ...
void CRiEngine::RenderAll(k3dIRenderFrame* Frame)
{
	K3DTREEHANDLE handle;
	for(k3dIDocument::TreeNodeType type = m_Camera->Document()->GetFirstObjectContextTree(&handle); k3dIDocument::COMPLETE != type; type = m_Camera->Document()->GetNextObjectContextTree(&handle))
		{
			k3dIObjectContext* context = m_Camera->Document()->GetObjectContextTree(handle);

			// Skip hidden objects ...
			if(SDP_S_OK != context->State()->QueryState(k3dIState::VISIBLE))
				continue;

			k3dIRiRender* render = sdpQueryInterface<k3dIRiRender>(context->Object());
			if(render)
				{
					switch(type)
						{
							case k3dIDocument::PUSH:
								render->RiPushRender(Frame, this);
								break;

							case k3dIDocument::NODE:
								render->RiRender(Frame, this);
								break;

							case k3dIDocument::POP:
								render->RiPopRender(Frame, this);
								break;
							default:
								g_assert_not_reached();	// Should never happen!
								break;
						}

					render->Release();
				}
		}
}

// Functor for post-rendering objects ...
class CRiEngineRiPostRender
{
public:
	CRiEngineRiPostRender(k3dIRenderFrame* Frame, k3dIRiEngine* Engine) { m_Frame = Frame; m_Engine = Engine; }
	bool HandleObject(k3dIObjectContext* Context)
	{
		// Skip hidden objects ...
		if(SDP_S_OK != Context->State()->QueryState(k3dIState::VISIBLE))
			return true;

		k3dIRiRender* render = sdpQueryInterface<k3dIRiRender>(Context->Object());
		if(render)
			{
				render->RiPostRender(m_Frame, m_Engine);
				render->Release();
			}

		return true;
	}
	
	k3dIRenderFrame* m_Frame;
	k3dIRiEngine* m_Engine;
};

// Setup RIB file ...
bool CRiEngine::SetupRIBFile(k3dIRenderFrame* Frame, std::ofstream& File)
{
	// Sanity checks ...
	g_return_val_if_fail(Frame, false);

	// Start our RIB file ...
	sdpString filename;
	gulong bufferlength = sdpMaxPathLength;
	Frame->GetUniqueFile("world.rib", filename.buffer(bufferlength), &bufferlength);
	
	// Open the RIB file stream ...
	File.open(sdpAppendPath(Frame->Path(), filename));
	g_return_val_if_fail(File.good(), false);

	// Start the RIB file ...
	SetOutputStream(&File);

	WriteRIBHeader();
	RiNewline();
	
	RiComment(sdpFormatString("Created with K-3D Version %s", VERSION_STRING));
	RiNewline();

	// Set path options ...
	RiComment("Setup file search paths");

	k3dRiParameterList searchpath;
	searchpath.AddString("shader", sdpString(m_Camera->Application()->ShadersPath()) + ":&");
	searchpath.DeclareParameters(this);
	searchpath.SetOptions("searchpath", this);
	RiNewline();

	return true;
}

// Setup display options ...
void CRiEngine::SetupDisplay(k3dIRenderFrame* Frame)
{
	// Sanity checks ...
	g_assert(Frame);

	// Set the display type ...
	RiComment("Setup Display");
	
	// Make sure those slashes point the correct direction! (Even under Win32)
	sdpString filepath = sdpAppendPath(Frame->Path(), k3dIRenderFrame::OutputImageFile);
	filepath = sdpMakePosixFilePath(filepath);

	if(Frame->GetViewDuringRender())
		RiDisplayV(filepath, k3dIRiEngine::RI_FRAMEBUFFER, k3dIRiEngine::RI_RGB, 0, 0, 0);
	else
		RiDisplayV(filepath, k3dIRiEngine::RI_FILE, k3dIRiEngine::RI_RGB, 0, 0, 0);
	
	RiFormat(m_Camera->Width(), m_Camera->Height(), 1.0);

	// Set pixel sampling rates ...
	RiPixelSamples(m_Camera->m_RiXSamples(), m_Camera->m_RiYSamples());

	// Set pixel filtering ...
	RiPixelFilter(m_Camera->m_RiPixelFilterName(), m_Camera->m_RiPixelFilterWidth(), m_Camera->m_RiPixelFilterHeight());

	// Default shading rate ...
	RiShadingRate(1.0);

	// Double-sided surfaces ...
	RiSides(2);

	// Set gain & gamma ...
	RiExposure(m_Camera->m_Exposure->GetValue1(m_Camera->m_CurrentTime, m_Camera->m_DeltaTime),
						m_Camera->m_Gamma->GetValue1(m_Camera->m_CurrentTime, m_Camera->m_DeltaTime));

	// Set depth-of-field options ...
	if(m_Camera->m_RiDepthOfField())
		RiDepthOfField(m_Camera->m_CurrentFStop, m_Camera->m_CurrentFocalLength, m_Camera->m_CurrentFocalPlane);
		
	RiNewline();
}

void CRiEngine::SetupView(k3dIRenderFrame* Frame)
{
	// Sanity checks ...
	g_assert(Frame);

	// Setup up viewing transformations
	RiComment("Setup viewing transformations");
	
	RiProjectionV("perspective", 0, 0, 0);
	RiScreenWindow(-m_Camera->m_CurrentWidthPlane / m_Camera->m_CurrentNearPlane,
								 m_Camera->m_CurrentWidthPlane / m_Camera->m_CurrentNearPlane,
								 -m_Camera->m_CurrentHeightPlane / m_Camera->m_CurrentNearPlane,
								 m_Camera->m_CurrentHeightPlane / m_Camera->m_CurrentNearPlane);
	RiClipping(m_Camera->m_CurrentNearPlane, m_Camera->m_CurrentFarPlane);

	sdpQuaternion q(m_Camera->m_CurrentOrientation);
	sdpEulerAngles a(q, sdpEulerAngles::ZXYstatic);
	RiRotate(-sdpDegrees(a[0]), 0.0f, 0.0f, 1.0f);
	RiRotate(-sdpDegrees(a[1]), 1.0f, 0.0f, 0.0f);
	RiRotate(-sdpDegrees(a[2]), 0.0f, 1.0f, 0.0f);
	RiTranslate(-m_Camera->m_CurrentPosition[0], -m_Camera->m_CurrentPosition[1], -m_Camera->m_CurrentPosition[2]);
	
	RiNewline();
}

void CRiEngine::SetupRadiosity(k3dIRenderFrame* Frame)
{
	// Sanity checks ...
	g_assert(Frame);

	// If radiosity is disabled, we're done ...
	if(!m_Camera->m_RiRadiosity())
		return;
		
	RiComment("Setup radiosity options");

	// Setup light attributes ...
	k3dRiParameterList parameters;
	parameters.AddInteger("steps", m_Camera->m_RiRadiositySteps());
	
	parameters.DeclareParameters(this);
	parameters.SetOptions("radiosity", this);

/*
	tokens[0] = "steps";
	intarg = m_Camera->m_RiRadiositySteps();
	pointers[0] = (RtPointer)&intarg;
	RiOptionV("radiosity", 1, tokens, pointers);

	intarg = m_Camera->m_RiRadiosityPatchSamples();
	RiOption("radiosity", "minpatchsamples", (RtInt*)&intarg, k3dIRiEngine::RI_NULL);

	RiDeclare("patchsize", "float");
	RiDeclare("elemsize", "float");
	RiDeclare("minsize", "float");

	floatarg = (RtFloat)m_Camera->m_RiRadiosityPatchSize();
	floatarg2 = (RtFloat)m_Camera->m_RiRadiosityElementSize();
	floatarg3 = (RtFloat)m_Camera->m_RiRadiosityMinimumSize();
	RiAttribute("radiosity", "patchsize", (RtFloat*)&floatarg, "elemsize", (RtFloat*)&floatarg2, "minsize", (RtFloat*)&floatarg3, k3dIRiEngine::RI_NULL);
*/

	RiNewline();
}

void CRiEngine::SetupAtmosphere(k3dIRenderFrame* Frame)
{
	// Sanity checks ...
	g_assert(Frame);

	// If atmospheric rendering is disabled, we're done ...
	if(!m_Camera->m_RiAtmosphere())
		return;

	// Setup the atmosphere shader ...
	RiComment("Setup atmosphere options");
	
	k3dRiVolumeShader volumeshader;
	m_Camera->m_RiAtmosphereShader.Initialize(Frame, this, volumeshader, m_Camera->m_CurrentTime, m_Camera->m_DeltaTime);
	volumeshader.SetupAtmosphere(this);

	RiNewline();
}

/////////////////////////////////////////////////////////////////
// k3dIRenderEngine overrides ...

sdpResult CRiEngine::SetCamera(k3dICamera* Camera)
{
	// Sanity checks ...
	g_return_val_if_fail(Camera, SDP_E_FAIL);

	m_Camera = (CCamera*)Camera;

	return SDP_S_OK;
}

sdpResult CRiEngine::Redraw()
{
	// Sanity checks ...
	g_return_val_if_fail(m_Camera, SDP_E_FAIL);

	// Not gonna do it!
	return SDP_S_FALSE;
}

sdpResult CRiEngine::StartJob(k3dIRenderJob* Job)
{
	// Sanity checks ...
	g_return_val_if_fail(m_Camera, SDP_E_FAIL);
	g_return_val_if_fail(Job, SDP_E_FAIL);

	CRiEngineRiStartJob functor(Job, this);
	k3dForEachObject(m_Camera->Document(), k3dIState::VISIBLE, functor);
	
	return SDP_S_OK;
}

sdpResult CRiEngine::QueueFrame(k3dIRenderFrame* Frame)
{
	// Sanity checks ...
	g_return_val_if_fail(m_Camera, SDP_E_FAIL);
	g_return_val_if_fail(Frame, SDP_E_FAIL);

	// Setup the frame for RI rendering with the user's preferred engine ...
	Frame->SetEngineType("ri");
	Frame->SetEngine(m_Camera->m_RiRenderEngine());
	
	// Calculate the current time
	gdouble frametime = m_Camera->m_TimeBase->GetViewTime();
	gdouble totalframes = m_Camera->m_TimeBase->GetDuration() * m_Camera->m_TimeBase->GetFrameRate();
	gdouble actualtime = frametime;
	m_Camera->m_TimeBase->SetRenderTime(actualtime);
	m_Camera->m_TimeBase->SetRenderDeltaTime(1.0 / totalframes);

	// Animate the document
	m_Camera->Document()->AnimateDocument();
	k3dCalculateNormals(m_Camera->Document());
	
	// Make everything happen ...
	std::ofstream ribfile;
	g_return_val_if_fail(SetupRIBFile(Frame, ribfile), SDP_E_FAIL);
	
	SetupDisplay(Frame);
	SetupView(Frame);
	SetupRadiosity(Frame);

	// Begin the world ...
	RiWorldBegin();

	// Setup our identity transformation matrix
	RiTransformBegin();
	RiIdentity();

	// Setup custom attributes for use by lights ...
	RiDeclare("shadows", "string");
	RiDeclare("casts_shadows", "string");

	// Setup atmospheric effects
	SetupAtmosphere(Frame);

	// Render objects ...
	CRiEngineRiPreRender prerender(Frame, this);
	k3dForEachObject(m_Camera->Document(), k3dIState::VISIBLE, prerender);

	CRiEngineRiLightSetup lightsetup(Frame, this);
	k3dForEachObject(m_Camera->Document(), k3dIState::VISIBLE, lightsetup);
		
	RenderAll(Frame);
	
	CRiEngineRiPostRender postrender(Frame, this);
	k3dForEachObject(m_Camera->Document(), k3dIState::VISIBLE, postrender);

	// Finished transforming ...
	RiTransformEnd();

	// Finish the world ...
	RiWorldEnd();

	// Finish the RIB file ...
	ribfile.close();
	
	return SDP_S_OK;
}

sdpResult CRiEngine::FinishJob(k3dIRenderJob* Job)
{
	// Sanity checks ...
	g_return_val_if_fail(m_Camera, SDP_E_FAIL);
	g_return_val_if_fail(Job, SDP_E_FAIL);

	CRiEngineRiFinishJob functor(Job, this);
	k3dForEachObject(m_Camera->Document(), k3dIState::VISIBLE, functor);
	
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIRiEngine overrides ...

k3dICamera* CRiEngine::Camera()
{
	return m_Camera;
}

