#ifndef POLYGONIZER_H
#define POLYGONIZER_H

// A C++ Implicit Surface Polygonizer
// Copyright 2000, Romain Behar <romainbehar@yahoo.com>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares Polygonizer, an implicit surface polygonizer
		\author Romain Behar (romainbehar@yahoo.com)
*/

// It is based on the following program :
//
// C code from the article
// "An Implicit Surface Polygonizer"
// by Jules Bloomenthal, jbloom@beauty.gmu.edu
// in "Graphics Gems IV", Academic Press, 1994
//
// implicit.c
//     an implicit surface polygonizer, translated from Mesa
//     applications should call polygonize()
//
// Authored by Jules Bloomenthal, Xerox PARC.
// Copyright (c) Xerox Corporation, 1991.  All rights reserved.
// Permission is granted to reproduce, use and distribute this code for
// any and all purposes, provided that this notice appears in all copies. */


#include <math.h>

#include <iostream>
#include <vector>
#include <stack>

#include <sdptypes.h>


// - - - - - - - - - - - - - - - - Data structures


inline bool AlmostEqual(const sdpVector3 p1, const sdpVector3 p2, gdouble range)
{ return (fabs(p1.n[0] - p2.n[0]) < range) && (fabs(p1.n[1] - p2.n[1]) < range) && (fabs(p1.n[2] - p2.n[2]) < range); }



class triangle
{
public:
	int u, v, w;

	triangle() { u = v = w = 0; }
	triangle(int tu, int tv, int tw) { u = tu, v = tv, w = tw; }
};


// Lattice index
class location
{
public:
	int i, j, k;

	location() {}
	location(int li, int lj, int lk) { i = li, j = lj, k = lk; }
};
inline bool operator == (const location& l1, const location& l2)
{ return (l1.i == l2.i) && (l1.j == l2.j) && (l1.k == l2.k); }


// Cube corner
class corner
{
public:
	location l;
	sdpVector3 p;
	gdouble value;

	corner() {}
};
inline bool operator == (const corner& c1, const corner& c2)
{ return (c1.l == c2.l) && (c1.p == c2.p) && (c1.value == c2.value); }


// Partitioning cell
class cube
{
public:
	location l;
	corner corners[8];

	cube() {}
};


class edge
{
public:
	location l1;
	location l2;

	// vertex id
	int vid;

	edge() {}
};



// - - - - - - - - - - - - - - - - Hash function and constants


// Make a hash from 3 integers, coding each value on 5 bits
const int HashBit = 5;
const int HashSize = 1 << (3 * HashBit);
const int HashMask = (1 << HashBit) - 1;
inline int HashFunc(const location& l)
{
	return ((((l.i & HashMask) << HashBit) | (l.j & HashMask)) << HashBit) | (l.k & HashMask);
}


// - - - - - - - - - - - - - - - - An Implicit Surface Polygonizer

class Polygonizer
{
private:
	// implicit surface function
	virtual gdouble FieldFunction(const sdpVector3&) = 0;

	// cube info
	gdouble voxel_size;

	// cube range within lattice
	int nxmin, nxmax, nymin, nymax, nzmin, nzmax;

	// start sdpVector3 on surface
	sdpVector3 start;

	// active cubes
	std::stack<cube> cubes;

	// surface
	std::vector<sdpVector3> *vertices;
	std::vector<triangle> *triangles;

	// hash tables
	std::vector< std::vector<location> > centers;
	std::vector< std::vector<corner> > corners;
	std::vector< std::vector<edge> > edges;


	// inline functions
	inline int Bit(int i, int n) { return (i>>n)&1; }
	inline int Flip(int i, int bit) { return i ^ (1 << bit); }

	// build fast Marching Cube tables
	std::vector <std::vector <std::vector <int> > > cubetable;
	int NextCWedge(int edge, int face);
	int OtherFace(int edge, int face);

	typedef enum
	{
		TETRAHEDRAL,
		MARCHING_CUBE
	} PolygonizerType;

	PolygonizerType m_PolygonizerDecomposition;

public:
	Polygonizer() {}
	virtual ~Polygonizer() {}

	void Init(gdouble size, int xmin, int xmax, int ymin, int ymax, int zmin, int zmax, bool tetrahedral);

	bool Polygonize(const sdpVector3& starting_sdpVector3, std::vector<sdpVector3> *save_vertices, std::vector<triangle> *save_triangles);

	void Converge(const sdpVector3& p1, const sdpVector3& p2, gdouble v, sdpVector3& p);
	bool FirstPoint(const sdpVector3& starting_sdpVector3);

	void TetrahedralDecomposition(const cube& c);
	void DoTet(const cube& cube1, int c1, int c2, int c3, int c4);

	void MakeCubeTable();
	void MarchingCube(const cube& cube1);

	void TestFace(int i, int j, int k, cube& old, int face, int c1, int c2, int c3, int c4);

	corner SetCorner(const location& l);
	bool SetCenter(const location& l);
	void SetEdge(const location& l1, const location& l2, int vid);
	int GetEdge(const location& l1, const location& l2);

	int VerticeId(const corner& c1, const corner& c2);
	void VNormal(sdpVector3& p, sdpVector3& v);

	bool IsNotPolygonized(const sdpVector3& test_point);

	void Triangle(int u, int v, int w);
};


#endif // POLYGONIZER_H
