// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CSoftObject K-3D object, which generates implicit surfaces
		\author Romain Behar (romainbehar@yahoo.com)
		\author Tim Shead (tshead@k-3d.com)
*/

#include "softobject.h"
#include "softobjectformulae.h"

#include <sdpfile.h>
#include <sdpgtkevents.h>

#include <k3diobjectcontext.h>
#include <k3diactions.h>
#include <k3dicamera.h>
#include <k3dmisc.h>

#include <iostream>

const sdpCLSID sdpSoftObject(0xF0000000, 0x00000000, 0x00000000, 0x00000002);

#include "polygonizer.h"

/////////////////////////////////////////////////////////////////////////////
// CSoftObject

CSoftObject::CSoftObject()
{
	m_VoxelSize = 1.0;
	m_Algorithm = CSoftObject::MARCHINGCUBE;

	m_Radius = 10.0;
	m_Hardness = 1.0;
	m_SkeletonType = CSoftObject::SETOFPOINTS;
	m_PotentialType = CSoftObject::BLOB;
}

CSoftObject::~CSoftObject()
{
}

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CSoftObject::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIAnimate>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIEditorRender>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dILocationServer>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIMaterialServer>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIObject>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIPersist>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIRiRender>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dITarget>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dITransform>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dITimeServer>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIUserEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIObject*>(this), InterfaceID, Interface)) return SDP_S_OK;

	return SDP_E_NO_INTERFACE;
}

gulong CSoftObject::AddRef()
{
	return sdpComAddRef(m_ReferenceCount);
}

gulong CSoftObject::Release()
{
	return sdpComRelease(this, m_ReferenceCount);
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

// Called to open an object's UI
sdpResult CSoftObject::EditObject()
{
	// Create a user interface ...
	g_return_val_if_fail(m_UserInterface.Create(this), SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CSoftObject::ObjectRenamed(SDPCSTRING NewName)
{
	m_UserInterface.ObjectRenamed(NewName);
	return k3dTriangleMesh::ObjectRenamed(NewName);
}

sdpResult CSoftObject::ObjectDeleted()
{
	m_UserInterface.ObjectDeleted();
	return k3dTriangleMesh::ObjectDeleted();
}

/////////////////////////////////////////////////////////////////////////////
// CSoftObject message handlers

class CSoftObjectCollectControlPoints
{
public:
	CSoftObjectCollectControlPoints(sdpVector3Array& ControlPoints) : m_ControlPoints(ControlPoints)
	{
	}

	bool HandleDescendantObject(k3dIObjectContext* Context)
	{
		k3dILocationServer* locationserver = sdpQueryInterface<k3dILocationServer>(Context->Object());
		if(0 == locationserver)
			return true;

std::cout << "count: " << locationserver->GetLocationCount() << std::endl;

		sdpMatrix4 toworld = k3dGetObjectToWorldMatrix(Context);

		sdpVector3 coords;
		for(K3DLOCATIONHANDLE handle = locationserver->GetFirstLocation(); handle; handle = locationserver->GetNextLocation(handle))
			{
				locationserver->GetLocation(handle)->GetAnimXYZ(coords);
				coords = toworld * coords;
				m_ControlPoints.push_back(coords);
			}

		locationserver->Release();
		return true;
	}

protected:
	sdpVector3Array& m_ControlPoints;
};


class CSoftObjectCollectControlSticks
{
public:
	CSoftObjectCollectControlSticks(std::vector<sdpVector3Array>& ControlSticks) : m_ControlSticks(ControlSticks)
	{
	}
protected:
	std::vector<sdpVector3Array>& m_ControlSticks;
};


void CSoftObject::OnModify(bool SaveAction)
{
	// Get control-point data from descendant locations ...
	m_ControlPoints.clear();
	CSoftObjectCollectControlPoints functor(m_ControlPoints);
	k3dForEachDescendant(Document(), ObjectContext(), functor);

std::cout << m_ControlPoints.size() << std::endl;

	// If none is specified, use a single point as default
	if(m_ControlPoints.size() == 0)
		{
			m_ControlPoints.push_back(sdpVector3(0.0, 0.0, 0.0));
		}

	sdpString statistics = "";

	// Polygonize skeletons
	PotentialField *polygonizer;

	// Get chosen potential formula
	gdouble (*potentialFunction)(gdouble, gdouble);
	switch(m_PotentialType)
		{
			case CSoftObject::BLOB :
				potentialFunction = BlobPotential;
			break;
			case CSoftObject::FINITESOFTOBJECT :
				potentialFunction = FiniteSoftObjectPotential;
			break;
			case CSoftObject::INFINITESOFTOBJECT :
				potentialFunction = InfiniteSoftObjectPotential;
			break;
			default:
				g_assert_not_reached();
		}

	//switch(m_Skeleton)
	//	{
	//		case CSoftObject::SETOFPOINTS :
				polygonizer = new PointSetPotentialField(m_ControlPoints, m_Radius, m_Hardness, potentialFunction);
	//		break;
	//		case CSoftObject::WIREFRAME :
				// Get control-stick data from descendant locations ...
				//m_ControlSticks.clear();
				//CSoftObjectCollectControlSticks functor(m_ControlPoints);
				//k3dForEachDescendant(Document(), ObjectContext(), functor);
	//		break;
	//		case CSoftObject::MESH :
	//		break;
	//		default:
	//			g_assert_not_reached();
	//	}

	bool myzer;
	if(m_Algorithm == CSoftObject::MARCHINGCUBE)
		myzer = false;
	else
		myzer = true;

	polygonizer->Init(m_VoxelSize, -50, 50, -50, 50, -50, 50, myzer);

	// Check polygonized skeletons
	std::vector<bool> polygonizedSkeletons;
	polygonizedSkeletons.reserve(m_ControlPoints.size());
	std::fill(polygonizedSkeletons.begin(), polygonizedSkeletons.begin() + m_ControlPoints.size(), false);

	// Polygonize unchecked polygons
	bool allPolygonized = false;
	guint polygonizeMe = 0;
	guint locationNb = 0;
	guint triangleNb = 0;
	do
		{
			guint n;
			std::vector<sdpVector3> vertices;
			std::vector<triangle> triangles;

			// Polygonize current skeleton's implicit surface
			polygonizer->Polygonize(m_ControlPoints[polygonizeMe], &vertices, &triangles);

			// Store surface
			ReserveLocations(locationNb + vertices.size());
			for(n = 0; n < vertices.size(); n++)
				Location(locationNb + n).m_LocalXYZ = vertices[n];

			ReserveTriangles(triangleNb + triangles.size());
			for(n = 0; n < triangles.size(); n++)
				Triangle(triangleNb + n).SetPoints(triangles[n].u + locationNb, triangles[n].w + locationNb, triangles[n].v + locationNb);

			locationNb += vertices.size();
			triangleNb += triangles.size();

			// Mark newly polygonized skeletons
			polygonizedSkeletons[polygonizeMe] = true;
			for(n = polygonizeMe + 1; n < m_ControlPoints.size(); n++)
				if(polygonizedSkeletons[n] == false)
					if(!polygonizer->IsNotPolygonized(m_ControlPoints[n]))
						polygonizedSkeletons[n] = true;

			// Free current surface
			vertices.clear();
			triangles.clear();

			// Look for another non-polygonized one
			allPolygonized = true;
			for(n = 0; n < m_ControlPoints.size(); n++)
				if(polygonizedSkeletons[n] == false)
					{
						polygonizeMe = n;
						allPolygonized = false;
						break;
					}
		}
	while(!allPolygonized);

	// free polygonizer data
	//polygonizer.freemem();
	polygonizedSkeletons.clear();

	// Update statistics
	statistics += sdpFormatString("%u points", locationNb);
	statistics += sdpFormatString(", %u triangles", triangleNb);

	if(m_UserInterface.RootObject().Attached())
		m_UserInterface.Entry("stats").SetText(statistics);

	// Render smooth surface
	SetSmoothShading();

	k3dTriangleMesh::OnModify(SaveAction);
}


///////////////////////////////////////////////////////////////
// k3dIPersist overrides

sdpResult CSoftObject::InitNew()
{
	g_return_val_if_fail(sdpSucceeded(k3dTriangleMesh::InitNew()), SDP_E_FAIL);

	OnModify(false);

	return SDP_S_OK;
}

sdpResult CSoftObject::LoadComplete()
{
	g_return_val_if_fail(sdpSucceeded(k3dTriangleMesh::LoadComplete()), SDP_E_FAIL);

	OnModify(false);

	return SDP_S_OK;
}

//////////////////////////////////////////////////////////////////////////////////////
// CSoftObjectDialog

CSoftObjectDialog::CSoftObjectDialog()
{
	m_Object = 0;
}

bool CSoftObjectDialog::Create(CSoftObject* Object)
{
	// Store our parent object ...
	m_Object = Object;

	// Load our template ...
	g_return_val_if_fail(k3dDialog::Create(Object, false, true, "softobject.gtkml", "softobject.html"), false);

	return true;
}

void CSoftObjectDialog::UpdateControls()
{
	Entry("voxelsize").SetText(k3dConvertDoubleString(k3dIChannel::DATA_DISTANCE, m_Object->m_VoxelSize, 2));
	Entry("radius").SetText(k3dConvertDoubleString(k3dIChannel::DATA_DISTANCE, m_Object->m_Radius, 2));
	Entry("hardness").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_Hardness, 1));

	switch(m_Object->m_Algorithm)
		{
			case CSoftObject::MARCHINGCUBE :
				ToggleButton("marchingcube").SetState(true);
				break;
			case CSoftObject::TETRAHEDRAL :
				ToggleButton("tetrahedral").SetState(true);
				break;
			default:
				g_assert_not_reached();
		}

	switch(m_Object->m_SkeletonType)
		{
			case CSoftObject::SETOFPOINTS :
				ToggleButton("pointskeleton").SetState(true);
				break;
			case CSoftObject::WIREFRAME :
				ToggleButton("wireframeskeleton").SetState(true);
				break;
			case CSoftObject::MESH :
				ToggleButton("meshskeleton").SetState(true);
				break;
			default:
				g_assert_not_reached();
		}

	switch(m_Object->m_PotentialType)
		{
			case CSoftObject::BLOB:
				ToggleButton("blobpotential").SetState(true);
				break;
			case CSoftObject::FINITESOFTOBJECT:
				ToggleButton("finitesopotential").SetState(true);
				break;
			case CSoftObject::INFINITESOFTOBJECT:
				ToggleButton("infinitesopotential").SetState(true);
				break;
			default:
				g_assert_not_reached();
		}
}

void CSoftObjectDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "onvoxelsize")
		OnVoxelSize();
	else if(Event->Name() == "onmarchingcube")
		OnMarchingCube();
	else if(Event->Name() == "ontetrahedral")
		OnTetrahedral();
	else if(Event->Name() == "onradius")
		OnRadius();
	else if(Event->Name() == "onhardness")
		OnHardness();
	else if(Event->Name() == "onpointskeleton")
		OnPointSkeleton();
	else if(Event->Name() == "onwireframeskeleton")
		OnWireFrameSkeleton();
	else if(Event->Name() == "onmeshskeleton")
		OnMeshSkeleton();
	else if(Event->Name() == "onblobpotential")
		OnBlobPotential();
	else if(Event->Name() == "onfinitesopotential")
		OnFiniteSOPotential();
	else if(Event->Name() == "oninfinitesopotential")
		OnInfiniteSOPotential();
	else
		k3dDialog::OnEvent(Event);
}

void CSoftObjectDialog::OnVoxelSize()
{
	if(UpdateControl(Entry("voxelsize"), m_Object->m_VoxelSize, k3dIChannel::DATA_DISTANCE, 2))
		m_Object->OnModify(false);
}

void CSoftObjectDialog::OnMarchingCube()
{
	if(ToggleButton("marchingcube").GetState())
		{
			m_Object->m_Algorithm = CSoftObject::MARCHINGCUBE;
			m_Object->OnModify(false);
		}
}

void CSoftObjectDialog::OnTetrahedral()
{
	if(ToggleButton("tetrahedral").GetState())
		{
			m_Object->m_Algorithm = CSoftObject::TETRAHEDRAL;
			m_Object->OnModify(false);
		}
}

void CSoftObjectDialog::OnRadius()
{
	if(UpdateControl(Entry("radius"), m_Object->m_Radius, k3dIChannel::DATA_DISTANCE, 2))
		m_Object->OnModify(false);
}

void CSoftObjectDialog::OnHardness()
{
	if(UpdateControl(Entry("hardness"), m_Object->m_Hardness, k3dIChannel::DATA_SCALAR, 1))
		m_Object->OnModify(false);
}

void CSoftObjectDialog::OnPointSkeleton()
{
	if(ToggleButton("pointskeleton").GetState())
		{
			m_Object->m_SkeletonType = CSoftObject::SETOFPOINTS;
			m_Object->OnModify(false);
		}
}

void CSoftObjectDialog::OnWireFrameSkeleton()
{
	if(ToggleButton("wireframeskeleton").GetState())
		{
			m_Object->m_SkeletonType = CSoftObject::WIREFRAME;
			m_Object->OnModify(false);
		}
}

void CSoftObjectDialog::OnMeshSkeleton()
{
	if(ToggleButton("meshskeleton").GetState())
		{
			m_Object->m_SkeletonType = CSoftObject::MESH;
			m_Object->OnModify(false);
		}
}

void CSoftObjectDialog::OnBlobPotential()
{
	if(ToggleButton("blobpotential").GetState())
		{
			m_Object->m_PotentialType = CSoftObject::BLOB;
			m_Object->OnModify(false);
		}

}

void CSoftObjectDialog::OnFiniteSOPotential()
{
	if(ToggleButton("finitesopotential").GetState())
		{
			m_Object->m_PotentialType = CSoftObject::FINITESOFTOBJECT;
			m_Object->OnModify(false);
		}

}

void CSoftObjectDialog::OnInfiniteSOPotential()
{
	if(ToggleButton("infinitesopotential").GetState())
		{
			m_Object->m_PotentialType = CSoftObject::INFINITESOFTOBJECT;
			m_Object->OnModify(false);
		}

}


