#ifndef ACTION_H
#define ACTION_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the CStep and CAction classes for storing undo/redo data within a document
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include <sdpstring.h>

#include <list>
#include <stack>
#include <map>

// Forward class declarations
class k3dIActionClient;

///////////////////////////////////////////////////////////////////////////////
// CStep

/// Stores a single "step" (change) within the document, in binary form
class CStep
{
public:
	CStep();
	CStep(k3dIActionClient* Client, gulong OldValueCode, gulong OldValue, gulong OldValueSize, gulong NewValueCode, gulong NewValue, gulong NewValueSize);

	/// Called to erase the step's data
	void Erase();
	/// Called to undo the step's effect
	void Undo();
	/// Called to redo the step's effect
	void Redo();

protected:
	/// Stores the originator of this step
	k3dIActionClient* m_Client;

	/// Stores the original property affected by this step
	gulong m_OldValueCode;
	gulong m_OldValue;
	gulong m_OldValueSize;

	/// Stores the new property set by this step
	gulong m_NewValueCode;
	gulong m_NewValue;
	gulong m_NewValueSize;
};

/// A collection of steps
typedef std::list<CStep> StepList;
typedef StepList::iterator StepIterator;

///////////////////////////////////////////////////////////////////////////////
// CAction

/// Stores a complete "action" (set of steps) performed on a document.  Represents a complete action that can be "undone" by the user.
class CAction
{
public:
	CAction();
	CAction(SDPCSTRING Name);

	SDPCSTRING Name() { return m_Name; }
	StepList& Steps() { return m_Steps; }

protected:
	/// Stores the action name (visible as the "Undo [name]" or "Redo [name]" menu items in the UI)
	sdpString m_Name;
	/// Stores all of the steps needed to undo / redo this action
	StepList m_Steps;
};

/// A collection of actions
typedef std::vector<CAction> ActionStack;
typedef ActionStack::iterator ActionIterator;

#endif // ACTION_H
