#ifndef APPLICATION_H
#define APPLICATION_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the CApplication class, which encapsulates the running K-3D program
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3diapplication.h>
#include <k3direnderfarm.h>

#include <k3ddocument.h>

#include "objectinfo.h"
#include "shaderinfo.h"
#include "renderjob.h"

// Forward class references ...
class CDocument;
class CApplicationFrame;
class k3dIScriptEngine;

/////////////////////////////////////////////////////////////////////////////
// CApplication

/// Encapsulates the running K-3D server application
class CApplication :	public k3dIApplication,
									public k3dIRenderFarm
{
friend class CApplicationFrame;
public:
	typedef std::list<CDocument*> DocumentList;
	typedef DocumentList::iterator DocumentIterator;

public:
	CApplication();
	virtual ~CApplication();

	/// Application initialization at program startup
	bool Create(int* argc, char*** argv);
	/// Starts the user interface loop
	int Run();

	// k3dIUnknown overrides ...
	sdpResult QueryInterface(const sdpIID& InterfaceID, void** Interface);
	gulong AddRef();
	gulong Release();

	// k3dIApplication overrides
	sdpResult Close();
	k3dIDocument* OpenDocument(SDPCSTRING FilePath);
	k3dIDocument* CreateDocument();
	k3dIDocument* GetActiveDocument();
	sdpResult CloseDocument(k3dIDocument* Document);
	K3DOBJECTINFOHANDLE GetFirstObjectInfo();
	K3DOBJECTINFOHANDLE GetNextObjectInfo(K3DOBJECTINFOHANDLE Handle);
	k3dIObjectInfo* GetObjectInfo(K3DOBJECTINFOHANDLE Handle);
	K3DSHADERINFOHANDLE GetFirstShaderInfo();
	K3DSHADERINFOHANDLE GetNextShaderInfo(K3DSHADERINFOHANDLE Handle);
	k3dIShaderInfo* GetShaderInfo(K3DSHADERINFOHANDLE Handle);
	SDPCSTRING BasePath();
	SDPCSTRING OptionsPath();
	SDPCSTRING ShadersPath();
	sdpResult BrowserNavigate(SDPCSTRING URL);
	sdpResult Message(SDPCSTRING Message, SDPCSTRING Title);
	sdpResult Query(SDPCSTRING Message, SDPCSTRING Title);
	sdpResult Error(SDPCSTRING Message, SDPCSTRING Title);
	sdpResult SetStatusBarMessage(SDPCSTRING Message);
	sdpXmlIDocument* Options();
	k3dIScriptEngine* ScriptEngine(sdpCLSID Language);

	// k3dIRenderFarm overrides ...
	SDPCSTRING Name();
	SDPCSTRING Path();
	k3dIRenderJob* CreateJob(SDPCSTRING JobName);
	sdpResult StartJob(k3dIRenderJob* Job);

	/// Return a pointer to the main UI frame
	CApplicationFrame* Frame() { return m_Frame; }
	/// Returns a pointer to the current active document
	CDocument* ActiveDocument();
	/// Sets the current active document
	void SetActiveDocument(CDocument* ActiveDocument);
	/// Called whenever a document is activated (receives the user focus)
	void OnActivate(CDocument* Document);
	/// Returns the collection of plugin libraries
	CObjectLibraries& Libraries() { return m_Libraries; }
	/// Returns the path to the application's GTKML dialog templates
	sdpString DialogPath();
	/// Returns the status of the Verbose Output flag ...
	bool Verbose() { return m_Verbose; }

protected:
	/// Tests for runtime capabilities that are needed in order for K-3D to function
	bool CheckDependencies();
	/// Parses command line options the first time and dispatches commands
	bool ParseCommandLine1(int argc, char* argv[]);
	/// Parses command line options a second time and dispatches commands
	bool ParseCommandLine2(int argc, char* argv[]);
	/// Closes the application UI frame
	void DeleteFrame(CApplicationFrame* Frame);
	/// Loads user preferences at startup
	void LoadOptions();
	/// Saves user preferences before shutdown
	void SaveOptions();
	/// Loads external plugin libraries
	void LoadPlugins();
	/// Sends version information to a stream
	void Version(std::ostream& Stream);

	/// Stores the K-3D base path (contains plugins, dialog templates, help files)
	sdpString m_BasePath;
	/// Stores the user options path
	sdpString m_OptionsPath;
	/// Stores the shader directories path(s)
	sdpString m_ShadersPath;
	/// Stores user options as an XML document
	k3dDocument m_Options;
	/// Stores the collection of dynamically-loaded object libraries
	CObjectLibraries m_Libraries;
	/// Stores the collection of dynamically-loaded Renderman Shaders
	CShaderInfoList m_Shaders;
	/// Stores the collection of open documents
	DocumentList m_Documents;
	/// Stores the collection of jobs waiting to be submitted to the render farm
	RenderJobList m_Jobs;
	/// Stores the collection of cached scripting engines
	typedef std::map<sdpCLSID, k3dIScriptEngine*> ScriptEngineMap;
	typedef ScriptEngineMap::iterator ScriptEngineIterator;
	ScriptEngineMap m_ScriptEngines;
	/// Used to signal the UI that the current document has changed
	gulong m_ActiveDocumentRevision;
	/// Set to true when the user wants verbose reporting to stdout
	bool m_Verbose;
	/// Controls whether the startup splash dialog will be shown
	bool m_ShowSplash;
	/// Stores a pointer to the application UI
	CApplicationFrame* m_Frame;
	/// Stores a pointer to the currently active document
	CDocument* m_ActiveDocument;
	/// Stores a reference count for sdpCOM
	gulong m_ReferenceCount;         
};

/////////////////////////////////////////////////////////////////////////////
// Program globals

extern CApplication Application;

#endif APPLICATION_H
