#ifndef DOCUMENTFRAME_H
#define DOCUMENTFRAME_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the CDocumentFrame class, which encapsulates an open K-3D document's UI
		\author Tim Shead (tshead@k-3d.com)
*/

#include "animationparameters.h"
#include "hierarchycontrol.h"

// Forward class declarations
class CDocument;

///////////////////////////////////////////////////////////
// CDocumentFrame

/// Implements a document's user interface
class CDocumentFrame : public sdpGtkObjectContainer
{
public:
	CDocumentFrame();
	virtual ~CDocumentFrame();

	/// Called to create the frame from scratch
	bool Create(CDocument* Document);
	/// Called to load the frame from an XML document
	bool Load(CDocument* Document, sdpXmlIElement* Element);

	/// Saves the frame to an XML document
	bool Save(sdpXmlIElement* Element, k3dIDependencies* Dependencies);

	/// Returns the owning document
	CDocument* Document() { return m_Document; }

protected:
	/// Overridden from sdpGtkObjectContainer to handle events
	void OnEvent(sdpGtkEvent* Event);

	/// Called to see if its OK to delete the frame
	void OnDelete(sdpGtkEvent* Event);
	/// Called when the frame is about to be destroyed
	void OnDestroy();
	/// Called when the user wants to create a new hierarchy view
	void OnFileNewHierarchy();
	/// Called when the user wants to create a new camera view
	void OnFileNewCamera();
	/// Called when the user wants to import geometry into the document
	void OnFileImport();
	/// Called when the user wants to export geometry from the document
	void OnFileExport();
	/// Called when the user wants to save the document
	void OnFileSave();
	/// Called when the user wants to save the document to a new file
	void OnFileSaveAs();
	/// Called when the user wants to close the document
	void OnFileClose();
	/// Called when the user wants to undo an action
	void OnEditUndo();
	/// Called when the user wants to redo an undone action
	void OnEditRedo();
	/// Called when the user wants to delete geometry
	void OnEditDelete();
	/// Called when the user wants to edit an object's properties
	void OnEditProperties();
	/// Called when the user wants to rename an object
	void OnEditRename();
	/// Called when the user toggles between point- and polygon-edit mode
	void OnEditMode();
	/// Called when the user makes the X axis active
	void OnXAxis();
	/// Called when the user makes the Y axis active
	void OnYAxis();
	/// Called when the user makes the Z axis active
	void OnZAxis();
	/// Called when the user wants to modify positions
	void OnPosition();
	/// Called when the user wants to modify orientations
	void OnOrientation();
	/// Called when the user wants to modify scales
	void OnScale(); 
	/// Called when the user wants to smooth-shade polygons
	void OnEditSmoothShade();
	/// Called when the user wants to unshade smooth-shade polygons
	void OnEditUnshade();
	/// Called when the user wants to edit animation parameters
	void OnAnimationParameters();
	/// Called to create a new layer
	void OnLayerAdd();
	/// Called to delete selected layers
	void OnLayerDelete();
	/// Called to rename selected layers
	void OnLayerRename();
	/// Called when the frame receives the input focus
	void OnSetFocus();
	/// Called when the user hits the "rewind" button
	void OnRewind();
	/// Called when the user hits the "reverse play" button
	void OnReversePlay();
	/// Called when the user hits the "stop" button
	void OnStop();
	/// Called when the user hits the "play" button
	void OnPlay();
	/// Called when the user hits the "continuous play" button
	void OnLoopPlay();
	/// Called when the user hits the "fast forward" button
	void OnFastForward();
	/// Called when the user hits the "record frame" button
	void OnRecordFrame();
	/// Called when the user hits the "record" button
	void OnRecord();
	/// Called when the user moves the animation scrollbar
	void OnAnimationScrollbar();
	/// Called when the user selects a new animation frame
	void OnAnimationFrame();
	/// Called when the user interface is idle
	void OnIdle();

	/// Returns the frame title
	sdpString Title();
	/// Updates the frame titlebar to display the frame title
	void UpdateTitlebar();

	/// Updates the layers tree as needed
	void UpdateLayers();
	/// Updates the undo/redo menu commands as needed
	void UpdateActions();
	/// Updates the edit-mode menu commands as needed
	void UpdateEditMode();
	/// Updates the active axis menu commands as needed
	void UpdateActiveAxis();
	/// Updates the active transformation menu commands as needed
	void UpdateActiveTransformation();
	/// Updates the animation playback controls as needed
	void UpdatePlaybackMode();
	/// Updates the animation frame controls as needed
	void UpdateTimebase();

	/// Prompts the user to see if it's safe to close the document
	bool SafeToClose();

	/// Stores the last action revision, so we know when it changes 
	gulong m_LastActionRevision;
	/// Stores the last layer revision, so we know when it changes 
	gulong m_LastLayerRevision;
	/// Stores the last edit mode revision, so we know when it changes 
	gulong m_LastEditModeRevision;
	/// Stores the last active axis revision, so we know when it changes 
	gulong m_LastActiveAxisRevision;
	/// Stores the last active transformation revision, so we know when it changes
	gulong m_LastActiveTransformationRevision;
	/// Stores the last playback revision, so we know when it changes 
	gulong m_LastPlaybackModeRevision;
	/// Stores the last timebase revision, so we know when it changes
	gulong m_LastTimebaseRevision;
	/// Stores the last animation frame, to help avoid unnecessary updates
	gulong m_LastAnimationFrame;

	/// Stores a pointer to the parent document
	CDocument* m_Document;

	/// Stores the hierarchy UI
	CHierarchyControl m_HierarchyControl;

	/// Stores a UI for changing animation parameters
	CAnimationParametersDialog m_AnimationParametersDialog;	
};

#endif DOCUMENTFRAME_H
