#ifndef SHADERINFO_H
#define SHADERINFO_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares classes for encapsulating information about a Pixar Renderman Shading Language shader
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dishaderinfo.h>
#include <k3dishaderinfoargument.h>

#include <sdpstring.h>

// Forward class declarations
class sdpXmlIElement;

////////////////////////////////////////////////////////////////////
// CShaderInfoArgument

/// Encapsulates a single Pixar Renderman Shading Language shader argument
class CShaderInfoArgument : public k3dIShaderInfoArgument
{
public:
	CShaderInfoArgument(sdpXmlIElement* Element);
	virtual ~CShaderInfoArgument() {}

	// k3dIUnknown overrides ...
	sdpResult QueryInterface(const sdpIID& InterfaceID, void** Interface) { return SDP_E_NO_INTERFACE; }
	gulong AddRef() { return 1; }
	gulong Release() { return 0; }

	// k3dIShaderInfoArgument overrides ...
	SDPCSTRING Name();
	k3dIShaderInfoArgument::ArgumentType Type();
	SDPCSTRING Default();

protected:
	/// Stores the argument name
	sdpString m_Name;
	/// Stores the argument type
	k3dIShaderInfoArgument::ArgumentType m_Type;
	/// Stores the argument default value
	sdpString m_Default;
	/// Stores the argument custom properties
	sdpStringMap m_Properties;
};

////////////////////////////////////////////////////////////////////
// CShaderInfoArgumentArray

typedef std::vector<CShaderInfoArgument*> ShaderInfoArgumentArray;
typedef ShaderInfoArgumentArray::iterator ShaderInfoArgumentIterator;

/// Stores a collection of shader arguments
class CShaderInfoArgumentArray : public ShaderInfoArgumentArray
{
public:
	~CShaderInfoArgumentArray();
};

////////////////////////////////////////////////////////////////////
// CShaderInfo

/// Encapsulates a single Pixar Renderman Shading Language shader
class CShaderInfo : public k3dIShaderInfo
{
public:
	CShaderInfo(SDPCSTRING FilePath);
	virtual ~CShaderInfo() {}

	// k3dIUnknown overrides ...
	sdpResult QueryInterface(const sdpIID& InterfaceID, void** Interface) { return SDP_E_NO_INTERFACE; }
	gulong AddRef() { return 1; }
	gulong Release() { return 0; }

	// k3dIShader overrides ...
	SDPCSTRING Name();
	k3dIShaderInfo::ShaderType Type();
	K3DSHADERINFOARGUMENTHANDLE GetFirstArgument();
	K3DSHADERINFOARGUMENTHANDLE GetNextArgument(K3DSHADERINFOARGUMENTHANDLE Handle);
	k3dIShaderInfoArgument* GetArgument(K3DSHADERINFOARGUMENTHANDLE Handle);

protected:
	/// Stores the shader name (could be totally unrelated to the filepath)
	sdpString m_Name;
	/// Stores the shader filepath
	sdpString m_FilePath;
	/// Stores the shader type
	k3dIShaderInfo::ShaderType m_Type;
	/// Stores the shader custom properties
	sdpStringMap m_Properties;
	/// Stores the shader arguments
	CShaderInfoArgumentArray m_Arguments;
};

//////////////////////////////////////////////////////////////////////////////////////
// CShaderInfoList

typedef std::vector<CShaderInfo*> ShaderInfoArray;
typedef ShaderInfoArray::iterator ShaderInfoIterator;

/// Stores a collection of shaders
class CShaderInfoList : public ShaderInfoArray
{
public:
	~CShaderInfoList();

	/// Loads shaders from the given directory
	bool LoadShaders(SDPCSTRING Directory);

	/// Loads shaders based on the user's environment
	bool LoadShaders();

	/// Functor operator for use with sdpForEachFile()
	bool HandleFile(SDPCSTRING FilePath, SDPCSTRING FileName);
};

#endif SHADERINFO_H
