// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CApplication class, which encapsulates the running K-3D program
		\author Tim Shead (tshead@k-3d.com)
*/

#include <sdpsimpletypes.h>

#include "application.h"
#include "applicationframe.h"
#include "document.h"
#include "dialogs.h"

#include <sdpxmlielement.h>
#include <sdpxmlutility.h>
#include <sdpfile.h>

#include <errno.h>
#include <time.h>
#include <fstream>
#include <algorithm>

#ifdef SDPWIN32
#define popen _popen
#endif // SDPWIN32

#ifdef SDPUNIX
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#endif // SDPUNUX

#include <sdpgl.h>

#include <k3discriptengine.h>

/////////////////////////////////////////////////////////////////////////////
// The one and only CApplication object

CApplication Application;

/////////////////////////////////////////////////////////////////////////////
// CApplication

CApplication::CApplication()
{
	// Setup our custom exception handler ...
//	KException::InstallStructuredExceptionHandler();

	m_ReferenceCount = 0;
	m_ActiveDocument = 0;
	m_Frame = 0;

	m_ActiveDocumentRevision = 1;

	// Disable verbose output ...
	m_Verbose = false;
	// Enable the splash dialog by default ...
	m_ShowSplash = true;

/*
	m_TipDialog = NULL;
	m_HelpSearchDialog = NULL;
*/

}

CApplication::~CApplication()
{
	if(Verbose())
		std::cout << "CApplication::~CApplication()" << std::endl;

	// For each cached scripting engine ...
	for(ScriptEngineIterator engine = m_ScriptEngines.begin(); engine != m_ScriptEngines.end(); engine++)
		{
			// Call Release() on it until it disappears ...
			while(1)
				if(0 == engine->second->Release())
					break;
		}
		
	// Get rid of any leftover jobs ...
	for(RenderJobIterator job = m_Jobs.begin(); job != m_Jobs.end(); job++)
		delete (*job);
}

#if defined SDPUNIX

SDPCSTRING ClassOf(int XVisualClass)
{
	switch(XVisualClass)
		{
			case StaticGray:
				return "sg";
			case GrayScale:
				return "gs";
			case StaticColor:
				return "sc";
			case PseudoColor:
				return "pc";
			case TrueColor:
				return "tc";
			case DirectColor:
				return "dc";
			default:
				return "??";
		}
}

#endif // SDPUNIX

// Check for things that we need in order to run ...
bool CApplication::CheckDependencies()
{
	// Make sure this platform supports dynamic-modules ...
	if(!g_module_supported())
		{
			std::cerr << "*** Host does not support dynamically-loaded modules.  You cannot run K-3D!!" << std::endl;
			return false;
		}

	// Check to be sure that the user has specified a basepath one-way-or-another ...
	if(0 == m_BasePath.size())
		{
			m_BasePath = sdpGetEnv("K3D_BASE_PATH");
			if(0 == m_BasePath.size())
				{
					std::cerr << "*** You MUST specify a path to the directory that" << std::endl;
					std::cerr << "*** contains the K-3D plugins & support files (usually" << std::endl;
					std::cerr << "*** the directory that contains the K-3D executable).  You can" << std::endl;
					std::cerr << "*** do this by setting the K3D_BASE_PATH environment variable, or" << std::endl;
					std::cerr << "*** running K-3D with the --basepath [path] command-line option." << std::endl;

					return false;
				}
		}

	if(Verbose())
		std::cout << "Base path: " << BasePath() << std::endl;

	// Nag the user if they haven't specified a shader path one-way-or-another ...
	if(0 == m_ShadersPath.size())
		{
			m_ShadersPath = sdpGetEnv("K3D_SHADERS_PATH");
			if(0 == m_ShadersPath.size())
				{
					std::cerr << "*** You should specify one-or-more paths to the directories K-3D" << std::endl;
					std::cerr << "*** should load shaders from (usually the K-3D shaders directory, plus" << std::endl;
					std::cerr << "*** the shaders directory provided by your render engine).  You can" << std::endl;
					std::cerr << "*** do this by setting the K3D_SHADERS_PATH environment variable, or" << std::endl;
					std::cerr << "*** running K-3D with the --shaderspath [path] command-line option." << std::endl;
					std::cerr << "*** Although you can continue to use K-3D, you will not be able to render" << std::endl;
					std::cerr << "*** using a Renderman Interface compliant render engine, unless you" << std::endl;
					std::cerr << "*** specify your shader paths." << std::endl;
				}
		}

	if(Verbose())
		std::cout << "Shaders path: " << ShadersPath() << std::endl;

	// Test for an OpenGL implementation ...
	if(!sdpglExists(std::cerr))
		{
			std::cerr << "*** Could not connect to an OpenGL server.  You cannot run K-3D!!" << std::endl;
			return false;
		}

	// Display details about the OpenGL implementation ...
	if(Verbose())
		sdpglConfiguration(std::cout);
		
	std::cout << std::endl;

	return true;
}

// Initialize the application ...
bool CApplication::Create(int* argc, char*** argv)
{
	// Sanity checks ...
	g_assert(argc);
	g_assert(argv);

	// Parse command-line options the first time ...
	if(!ParseCommandLine1(*argc, *argv))
		return false;

	if(!CheckDependencies())
		return false;

	// Initialize GTK+ library ...
	gtk_init(argc, argv);

	// Create a splash screen ...
	CSplashBox* splashscreen = 0;
	if(m_ShowSplash)
		{
			splashscreen = new CSplashBox();
			splashscreen->Create();
		}

	// Load user options ...
	LoadOptions();

	// Dynamically load external plugins ...
	LoadPlugins();

	// Dynamically load Ri Shaders ...
	m_Shaders.LoadShaders();

	// Create our main window ...
	m_Frame = new CApplicationFrame();
	g_return_val_if_fail(m_Frame, false);
	g_return_val_if_fail(m_Frame->Create(this), false);

	// Close the splash screen ...
	if(splashscreen)
		delete splashscreen;

	// Parse command-line options a second time...
	if(!ParseCommandLine2(*argc, *argv))
		return false;

	return true;
}

// Write program version info to a stream ...
void CApplication::Version(std::ostream& Stream)
{
#include "../../k3d_version.h"

	Stream << "K-3D Version " << VERSION_STRING << std::endl;
	Stream << "Copyright (c) 1995-2000, Timothy M. Shead.  See the AUTHORS file for contributors." << std::endl;
	Stream << "Licensed by the GNU General Public License.  See the COPYING file for details." << std::endl;
	Stream << "K-3D Home Page: http://www.k-3d.com" << std::endl;
	Stream << "Bug reports to tshead@k-3d.com" << std::endl;
}

// Parse initial command-line options at startup ...
bool CApplication::ParseCommandLine1(int argc, char* argv[])
{
	// For each command-line argument ...
	for(int i = 1; i < argc; i++)
		{
			const sdpString name = argv[i];

			if(name == "--help" || name == "-h")
				{
					std::cout << "usage: k3d [options] [file]" << std::endl;
					std::cout << std::endl;
					std::cout << "  -h, --help               prints this help information and exits" << std::endl;
					std::cout << "      --basepath [path]    specifies the path for loading plugins, templates, etc." << std::endl;
					std::cout << "      --optionspath [path] overrides the path for storing user options" << std::endl;
					std::cout << "      --shaderspath [path] specifies the path(s) from which shaders will be loaded" << std::endl; 
					std::cout << "  -n, --new                creates a new document after startup" << std::endl;
					std::cout << "  -o, --open [file]        opens the given file after startup" << std::endl;
					std::cout << "      --prompt             automatically prompt the user for a document to open after startup" << std::endl;
					std::cout << "  -v, --verbose            prints diagnostic messages while running" << std::endl;
					std::cout << "      --version            prints program version information and exits" << std::endl;
					std::cout << "      --nosplash           disables the startup splash screen" << std::endl;
					std::cout << std::endl;
					exit(0);
				}
			else if(name == "--version")
				{
					Version(std::cout);
					std::cout << std::endl;
					exit(0);
				}
			else if(name == "--verbose" || name == "-v")
				{
					Version(std::cout);
					std::cout << std::endl;
					m_Verbose = true;
				}
			else if(name == "--nosplash")
				{
					m_ShowSplash = false;
				}
			else if(name == "--basepath")
				{
					m_BasePath = argv[i+1];
				}
			else if(name == "--optionspath")
				{
					m_OptionsPath = argv[i+1];
				}
			else if(name == "--shaderspath")
				{
					m_ShadersPath = argv[i+1];
				}
		}

	return true;
}

// Parse final command-line options once initialization is complete ...
bool CApplication::ParseCommandLine2(int argc, char* argv[])
{
	// For each command-line argument ...
	for(int i = 1; i < argc; i++)
		{
			sdpString name = argv[i];

			if(name == "--new" || name == "-n")
				CreateDocument();
			else if(name == "--open" || name == "-o")
				{
					OpenDocument(argv[i+1]);
					i++;
				}
			else if(name == "--prompt")
				m_Frame->OnFileOpen();
		}

	return true;
}

// Delete our frame ...
void CApplication::DeleteFrame(CApplicationFrame* Frame)
{
	// Sanity checks ...
	g_assert(Frame == m_Frame);

	delete m_Frame;
	m_Frame = 0;
	
	Close();
}

// The main GTK event loop ...
int CApplication::Run()
{
	gtk_main();
	return 0;
}

//////////////////////////////////////////////////////////////////////////////////
// k3dIUnknown overrides ...

sdpResult CApplication::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIApplication>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIRenderFarm>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIApplication*>(this), InterfaceID, Interface)) return SDP_S_OK;

	g_return_val_if_fail(0, SDP_E_NO_INTERFACE);
}

gulong CApplication::AddRef()
{
	return 1;
}

gulong CApplication::Release()
{
	return 0;
}

////////////////////////////////////////////////////////////////////////////////////
// k3dIApplication overrides ...

sdpResult CApplication::Close()
{
	if(Verbose())
		std::cout << "CApplication::Close()" << std::endl;

	// Save user preferences ...
	SaveOptions();

	// Delete any remaining documents ...
	for(DocumentIterator doc = m_Documents.begin(); doc != m_Documents.end(); doc = m_Documents.begin())
		CloseDocument(*doc);
		
	// Get rid of our frame ...
	if(m_Frame)
		m_Frame->RootWidget().Destroy();

	gtk_main_quit();

	return SDP_S_OK;
}

k3dIDocument* CApplication::OpenDocument(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, 0);

	// Create a new doc ...
	CDocument* document = new CDocument;
	g_return_val_if_fail(document, 0);

	// Initialize it ...
	if(!document->Open(this, FilePath))
		{
			delete document;
			g_return_val_if_fail(0, 0);
		}

	// Add it to the doc list ...
	m_Documents.push_back(document);
	
	return document;
}

k3dIDocument* CApplication::CreateDocument()
{
	// Create a new doc ...
	CDocument* document = new CDocument;
	g_return_val_if_fail(document, 0);

	// Initialize it ...
	if(!document->Create(this))
		{
			delete document;
			g_return_val_if_fail(0, 0);
		}

	// Add it to the doc list ...
	m_Documents.push_back(document);

	return document;
}

k3dIDocument* CApplication::GetActiveDocument()
{
	return ActiveDocument();
}

sdpResult CApplication::CloseDocument(k3dIDocument* Document)
{
	// Sanity checks ...
	g_return_val_if_fail(Document, SDP_E_FAIL);

	// Get the private document structure ...	
	CDocument* document = static_cast<CDocument*>(Document);
	
	// Find the document iterator ...
	DocumentIterator iterator = std::find(m_Documents.begin(), m_Documents.end(), document);
	g_return_val_if_fail(iterator != m_Documents.end(), SDP_E_FAIL);

	// Cleanup links to this document ...
	if(m_ActiveDocument == document)
		{
			// Activate a new document ...
			SetActiveDocument(0);

			if(m_Documents.size())
				m_Documents.back()->OnActivate(this);
		}
	
	// Remove it from the document list ...
	m_Documents.erase(iterator);

	// Delete the document ...
	delete document;
	
	return SDP_S_OK;
}

CDocument* CApplication::ActiveDocument()
{
	return m_ActiveDocument;
}

void CApplication::SetActiveDocument(CDocument* Document)
{
	// See if anything really changed ...
	if(m_ActiveDocument != Document)
		{
			// Set the new active document ...
			m_ActiveDocument = Document;

			// Encourage the UI to update itself ...
			m_ActiveDocumentRevision++;
		}
}

K3DOBJECTINFOHANDLE CApplication::GetFirstObjectInfo()
{
	return K3DOBJECTINFOHANDLE(Libraries().Objects().size() ? 1 : 0);
}

K3DOBJECTINFOHANDLE CApplication::GetNextObjectInfo(K3DOBJECTINFOHANDLE Handle)
{
	// Sanity checks ...
	g_return_val_if_fail(Handle, 0);
	g_return_val_if_fail(gulong(Handle) <= Libraries().Objects().size(), 0);

	return K3DOBJECTINFOHANDLE(gulong(Handle) < Libraries().Objects().size() ? gulong(Handle)+1 : 0);
}

k3dIObjectInfo* CApplication::GetObjectInfo(K3DOBJECTINFOHANDLE Handle)
{
	// Sanity checks ...
	g_return_val_if_fail(Handle, 0);
	g_return_val_if_fail(gulong(Handle) <= Libraries().Objects().size(), 0);

	return Libraries().Objects()[gulong(Handle)-1];
}

K3DSHADERINFOHANDLE CApplication::GetFirstShaderInfo()
{
	return K3DSHADERINFOHANDLE(m_Shaders.size() ? 1 : 0);
}

K3DSHADERINFOHANDLE CApplication::GetNextShaderInfo(K3DSHADERINFOHANDLE Handle)
{
	// Sanity checks ...
	g_return_val_if_fail(gulong(Handle), 0);

	return K3DSHADERINFOHANDLE(gulong(Handle) < m_Shaders.size() ? gulong(Handle) + 1 : 0);
}

k3dIShaderInfo* CApplication::GetShaderInfo(K3DSHADERINFOHANDLE Handle)
{
	// Sanity checks ...
	g_return_val_if_fail(Handle, 0);
	g_return_val_if_fail(gulong(Handle) <= m_Shaders.size(), 0);

	return m_Shaders[gulong(Handle) - 1];
}

sdpXmlIDocument* CApplication::Options()
{
	return &m_Options;
}

k3dIScriptEngine* CApplication::ScriptEngine(sdpCLSID Language)
{
	// Sanity checks ...
	g_return_val_if_fail(Language != sdpCLSIDNull, 0);

	// Try to find an existing engine ...
	ScriptEngineIterator engineiterator = m_ScriptEngines.find(Language);
	if(engineiterator != m_ScriptEngines.end())
		return engineiterator->second;

	// Try to find the object info for this ClassID ...
	CObjectInfo* info = Libraries().ObjectInfo(Language);

	// If we don't support this object type, fail ...
	g_return_val_if_fail(info, 0);

	// Create an instance of the object ...
	k3dIScriptEngine* engine = 0;
	g_return_val_if_fail(sdpSucceeded(info->ClassProcedure()(Language, k3dIScriptEngine::iid, (void**)&engine)), 0);
	g_return_val_if_fail(engine, 0);

	// Initialize it ...
	g_return_val_if_fail(sdpSucceeded(engine->SetOwner(this)), 0);

	// Add it to our cache for future requests ...
	m_ScriptEngines[Language] = engine;

	// Some debugging output ...
	if(Verbose())
		std::cout << "CApplication::ScriptEngine() - created " << engine->Language() << " scripting engine" << std::endl;

	// Increment the reference count and return it ...
	engine->AddRef();
	
	return engine;
}

SDPCSTRING CApplication::BasePath()
{
	return m_BasePath;
}

SDPCSTRING CApplication::OptionsPath()
{
	return m_OptionsPath;
}

SDPCSTRING CApplication::ShadersPath()
{
	return m_ShadersPath;
}

sdpResult CApplication::Message(SDPCSTRING Message, SDPCSTRING Title)
{
	CMessageBox dialog;
	dialog.DoModal(Title, Message);

	return SDP_S_OK;
}

sdpResult CApplication::Query(SDPCSTRING Message, SDPCSTRING Title)
{
	CQueryBox dialog;
	switch(dialog.DoModal(Title, Message, "OK", "Cancel"))
		{
			case 0:
				return SDP_S_OK;
			case 1:
				return SDP_S_FALSE;
		}
		
	g_return_val_if_fail(0, SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CApplication::Error(SDPCSTRING Message, SDPCSTRING Title)
{
	CErrorBox dialog;
	dialog.DoModal(Title, Message);

	return SDP_S_OK;
}

sdpResult CApplication::SetStatusBarMessage(SDPCSTRING Message)
{
/*
	if(Message)
		{
			m_Frame->SetMessageText(Message);
		}
	else
		{
			m_Frame->SetMessageText("For Help, press F1");
		}
*/
	return SDP_S_OK;
}

#if defined SDPWIN32

#include <string.h>
#include <windows.h>

static glong GetRegKey(HKEY key, SDPCSTRING subkey, SDPSTRING retdata)
{
	HKEY hkey;
	LONG retval = RegOpenKeyEx(key, subkey, 0, KEY_QUERY_VALUE, &hkey);

	if(retval == ERROR_SUCCESS)
		{
			long datasize = MAX_PATH;
			TCHAR data[MAX_PATH];
			RegQueryValue(hkey, 0, data, &datasize);
			lstrcpy(retdata,data);
			RegCloseKey(hkey);
		}

	return retval;
}

#endif // SDPWIN32


sdpResult CApplication::BrowserNavigate(SDPCSTRING URL)
{
	// Sanity checks ...
	g_return_val_if_string_fail(URL, SDP_E_FAIL);
	
/*
#if defined SDPWIN32
	// "GotoURL" function by Stuart Patterson.  As seen in the August, 1997 Windows Developer's Journal.
	// First try ShellExecute() ...
	HINSTANCE result = ShellExecute(0, "open", URL, 0, 0, SW_SHOW);

	// If it failed, get the .htm regkey and lookup the program ...
	if((gulong)result <= HINSTANCE_ERROR)
		{		
			TCHAR key[MAX_PATH + MAX_PATH];	
			if(GetRegKey(HKEY_CLASSES_ROOT, ".htm", key) == ERROR_SUCCESS)
				{
					lstrcat(key, "\\shell\\open\\command");

					if(GetRegKey(HKEY_CLASSES_ROOT, key, key) == ERROR_SUCCESS)
						{
							TCHAR *pos;
							pos = strstr(key, "\"%1\"");
							if(pos == NULL)	// No quotes found
								{
									pos = strstr(key, "%1");	// Check for %1, without quotes
									if(pos == NULL)	// No parameter at all...
										pos = key+lstrlen(key)-1;
									else
										*pos = '\0';	// Remove the parameter
								}
							else
								*pos = '\0';	// Remove the parameter

							lstrcat(pos, " ");
							lstrcat(pos, URL);
							result = (HINSTANCE)WinExec(key, SW_SHOW);
						}
				}
		}

	return result ? SDP_S_OK : SDP_E_FAIL;
	
#else // SDPWIN32
*/

	sdpString commandline;
	
	// See if we can override it from the user options
	sdpXmlElementArray browsers = sdpXmlXPathSelect(Options(), "application/helpers/command[@type=\"htmlviewer\"]");
	if(browsers.size())
		{
			commandline = browsers.front()->Text();
			commandline.formattedreplace('%', "p", URL);
		}
	else
		{
			// Try the user's BROWSER path as a default ...
			std::cout << "Couldn't find htmlviewer helper in options ... defaulting to BROWSER environment variable instead ..." << std::endl;
			commandline = sdpGetEnv("BROWSER") + " " + URL;
		}

	// Fire-up the web browser ...
	if(Verbose())
		std::cout << "CApplication::BrowserNavigate(): " << commandline << std::endl;
	
	popen(commandline, "r");

	return SDP_S_OK;

//#endif // !SDPWIN32
}

/////////////////////////////////////////////////////
// k3dIRenderFarm overrides

SDPCSTRING CApplication::Name()
{
	return "K-3D Builtin Render Farm";
}

SDPCSTRING CApplication::Path()
{
	return sdpXmlGetXPathElement(Options()->GetRootElement(), "application/paths/path[@type=\"renderfarm\"]")->Text();
}

k3dIRenderJob* CApplication::CreateJob(SDPCSTRING JobName)
{
	// Sanity checks ...
	g_return_val_if_string_fail(JobName, 0);

	// Create a new job ...
	CRenderJob* newjob = new CRenderJob(this, JobName);
	g_assert(newjob);

	if(Verbose())
		std::cout << "CApplication::CreateJob(): Created job " << newjob->Name() << std::endl;
			
	// Store a reference and return it to the caller ...
	m_Jobs.push_back(newjob);
	return newjob;
}

sdpResult CApplication::StartJob(k3dIRenderJob* Job)
{
	// Sanity checks ...
	g_return_val_if_fail(Job, SDP_E_FAIL);

	// Make sure it really is one of our jobs ...
	CRenderJob* job = static_cast<CRenderJob*>(Job);
	g_return_val_if_fail(std::find(m_Jobs.begin(), m_Jobs.end(), job) != m_Jobs.end(), SDP_E_FAIL);

	// Update control files for the job and all its frames ...
	g_return_val_if_fail(job->WriteControlFiles(), false);
	
	// Mark the job ready-to-run ...
	g_return_val_if_fail(job->MarkReady(), false);

	// Start the local rendering process ...
	sdpString commandline = sdpFormatString("renderjob %s &", job->Path());
	
	errno=0;
	int result = system(commandline);
	if(-1 == result || (127 == result && errno))
		{
			std::cerr << "CApplication::StartJob(): Error starting render job " << commandline << std::endl;
			return SDP_E_FAIL;
		}
			
	return SDP_S_OK;	
}

sdpString CApplication::DialogPath()
{
	return sdpAppendPath(BasePath(), "dialogtemplates");
}

static void AddRenderEngine(sdpXmlIElement* Element, SDPCSTRING EngineType, SDPCSTRING Engine, SDPCSTRING Name, SDPCSTRING Command, SDPCSTRING VisibleCommand)
{
	sdpString xpath;
	sdpXmlIElement* engine;
	
	xpath = "command[@type=\"render\" and @enginetype=\"" + sdpString(EngineType) + "\" and @engine=\"" + sdpString(Engine) + "\" and @viewduringrender=\"false\"]";
	engine = sdpXmlGetSafeXPathElement(Element, xpath, "command", Command, "type", "render", "enginetype", EngineType, "engine", Engine, "name", Name, "viewduringrender", "false", 0);
	g_assert(engine);
	
	xpath = "command[@type=\"render\" and @enginetype=\"" + sdpString(EngineType) + "\" and @engine=\"" + sdpString(Engine) + "\" and @viewduringrender=\"true\"]";
	engine = sdpXmlGetSafeXPathElement(Element, xpath, "command", VisibleCommand, "type", "render", "enginetype", EngineType, "engine", Engine, "name", Name, "viewduringrender", "true", 0);
	g_assert(engine);
}

void CApplication::LoadOptions()
{
	// Figure out some default data for all our helper applications / paths ...
#ifdef SDPWIN32
	sdpString htmlviewercommand = "mozilla %p";// What's a sensible default for Windoze?
	sdpString bitmapviewercommand = "";// What's a sensible default for Windoze?
	sdpString slpreprocessorcommand = "cpp %p";
	sdpString slparsercommand = "sdpslparse";
#else // SDPWIN32
	sdpString htmlviewercommand = sdpGetEnv("BROWSER"); if(htmlviewercommand.size()) { htmlviewercommand += " %p";  } else { htmlviewercommand = "mozilla %p"; }
	sdpString bitmapviewercommand = "xview %p";
	sdpString slpreprocessorcommand = "gcc -E -P -x c %p";
	sdpString slparsercommand = "sdpslparse";
#endif // !SDPWIN32

	// Load the document (we ignore load errors because it might not exist, yet ... if so, create it)
	m_Options.LoadOptions(m_OptionsPath);
						
	// Make sure the document isn't empty ...
	sdpXmlIElement* root = m_Options.GetRootElement();
	if(!root)
		{
			root = m_Options.CreateRootElement("k3dml");
			g_assert(root);
		}

	// And make sure it's really a K-3D Markup Language document ...
	if(root->Name() != sdpString("k3dml"))
		{
			m_Options.DestroyRootElement();
			root = m_Options.CreateRootElement("k3dml");
			g_assert(root);
		}

	// Make sure it contains application data ...
	sdpXmlIElement* application = sdpXmlGetSafeXPathElement(root, "application", "application", "", 0);
	g_assert(application);

	// Make sure it contains helpers data ...
	sdpXmlIElement* helpers = sdpXmlGetSafeXPathElement(application, "helpers", "helpers", "", 0);
	g_assert(helpers);

	// Make sure we have an entry for the HTML viewer ... 
	sdpXmlIElement* htmlviewer = sdpXmlGetSafeXPathElement(helpers, "command[@type=\"htmlviewer\"]", "command", htmlviewercommand, "type", "htmlviewer", 0);
	g_assert(htmlviewer);

	// Make sure we have an entry for the bitmap viewer ...
	sdpXmlIElement* bitmapviewer = sdpXmlGetSafeXPathElement(helpers, "command[@type=\"bitmapviewer\"]", "command", bitmapviewercommand, "type", "bitmapviewer", 0);
	g_assert(bitmapviewer);

	// Make sure we have an entry for the shading language pre-processor ...
	sdpXmlIElement* slpreprocessor = sdpXmlGetSafeXPathElement(helpers, "command[@type=\"slpreprocessor\"]", "command", slpreprocessorcommand, "type", "slpreprocessor", 0);
	g_assert(slpreprocessor);

	// Make sure we have an entry for the shading language parser ...
	sdpXmlIElement* slparser = sdpXmlGetSafeXPathElement(helpers, "command[@type=\"slparser\"]", "command", slparsercommand, "type", "slparser", 0);
	g_assert(slparser);

	// Make sure we have render engine entries ...
	AddRenderEngine(helpers, "ri", "bmrt", "BMRT", "rendrib %p", "rendrib -d 8 %p");
	AddRenderEngine(helpers, "ri", "3delight", "3Delight", "renderdl %p", "renderdl %p");
	AddRenderEngine(helpers, "ri", "rdc", "Render Dot C", "renderdc %p", "renderdc %p");
	AddRenderEngine(helpers, "ri", "aqsis", "Aqsis", "aqsis %p", "aqsis %p");
	AddRenderEngine(helpers, "ri", "custom", "Custom Render Engine", "rendrib %p", "rendrib -d 8 %p");
	
	// BMRT is the default ... hey, we gotta pick something ... !
	sdpXmlIElement* defaultengine = sdpXmlGetSafeXPathElement(application, "defaultengine", "defaultengine", 0, "enginetype", "ri", "engine", "bmrt", 0);
	g_assert(defaultengine);
	
	// Make sure we have a <paths> section ...
	sdpXmlIElement* paths = sdpXmlGetSafeXPathElement(application, "paths", "paths", "", 0);
	g_assert(paths);

	// Make sure we have a default document path ...
	sdpXmlIElement* documentpath = sdpXmlGetSafeXPathElement(paths, "path[@type=\"document\"]", "path", sdpGetHomeDirectory(), "type", "document", 0);
	g_assert(documentpath);
	
	// Make sure we have a default bitmap path ...
	sdpXmlIElement* bitmappath = sdpXmlGetSafeXPathElement(paths, "path[@type=\"bitmap\"]", "path", sdpGetHomeDirectory(), "type", "bitmap", 0);
	g_assert(bitmappath);
	
	// Make sure we have a default render farm path ...
	sdpXmlIElement* renderfarmpath = sdpXmlGetSafeXPathElement(paths, "path[@type=\"renderfarm\"]", "path", sdpGetTempDirectory(), "type", "renderfarm", 0);
	g_assert(renderfarmpath);
	
	// Make sure we have a default script path ...
	sdpXmlIElement* scriptpath = sdpXmlGetSafeXPathElement(paths, "path[@type=\"script\"]", "path", sdpGetHomeDirectory(), "type", "script", 0);
	g_assert(scriptpath);
}

void CApplication::SaveOptions()
{
	if(Verbose())
		std::cout << "Saving options to: " << m_Options.OptionsPath(m_OptionsPath) << std::endl;

	m_Options.SaveOptions(m_OptionsPath);
}

void CApplication::LoadPlugins()
{
	sdpString pluginpath = sdpAppendPath(BasePath(), "plugins");

	if(Verbose())
		std::cout << "Plugin Path: " << pluginpath << std::endl;

	m_Libraries.LoadLibraries(pluginpath, true);
}

void CApplication::OnActivate(CDocument* Document)
{
	SetActiveDocument(Document);
}

