
// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CDocument class, which encapsulates an open K-3D document
		\author Tim Shead (tshead@k-3d.com)
*/

#include "application.h"
#include "document.h"
#include "objectcontext.h"
#include "layer.h"
#include "documentframe.h"
#include "hierarchyframe.h"

// We've got some type of unfortunate clash with k3dITimeServer ...
#ifdef CurrentTime
#undef CurrentTime
#endif // CurrentTime
#include <k3ditimeserver.h>

#include <k3ditransform.h>
#include <k3dianimate.h>
#include <k3difileformat.h>
#include <k3digeometryreadformat.h>
#include <k3digeometrywriteformat.h>
#include <k3dilocation.h>
#include <k3dimaterial.h>
#include <k3dipersist.h>

#include <k3ddocument.h>
#include <k3ddependencies.h>
#include <k3dfilefilters.h>
#include <k3dcom.h>
#include <k3dmisc.h>

#include <sdpxmlutility.h>

#include <sdpfile.h>

#include <stack>
#include <algorithm>
#include <fstream>

/////////////////////////////////////////////////////////////////////////////
// Utility functions ...

template<class Type> bool ForEachDocumentFrame(CDocument* Document, Type& Functor)
{
	for(DocumentFrameIterator frame = Document->m_DocumentFrames.begin(); frame != Document->m_DocumentFrames.end(); frame++)
		{
			if(!Functor(*frame))
				return false;
		}

	return true;
}

template<class Type> bool ForEachHierarchyFrame(CDocument* Document, Type& Functor)
{
	for(HierarchyFrameIterator frame = Document->m_HierarchyFrames.begin(); frame != Document->m_HierarchyFrames.end(); frame++)
		{
			if(!Functor(*frame))
				return false;
		}

	return true;
}

template<class Type> bool ForEachLayer(CDocument* Document, Type& Functor)
{
	for(CDocument::LayerIterator layer = Document->m_Layers.begin(); layer != Document->m_Layers.end(); layer++)
		{
			if(!Functor(*layer))
				return false;
		}

	return true;
}

/////////////////////////////////////////////////////////////////////////////
// CDocument

gulong CDocument::m_DocumentNumber = 1;

CDocument::CDocument()
{
	m_Dying = false;

	m_CurrentObjectID = 0;

	// Animation State ...
	m_PlaybackMode = STOP;
	m_AnimationTimer = 0;
	m_IgnoreAnimationTimer = false;

	m_Duration = 4.0;
	m_FrameRate = 30.0;
	m_StartTime = 0.0;
	m_ViewTime = 0.0;
	m_EndTime = 1.0 - GetFrameDelta();
	m_RenderTime = 0.0;
	m_RenderDeltaTime = GetFrameDelta();

	// Odds-and-ends ...
	m_ActiveAxis = k3dIDocument::ZAXIS;
	m_ActiveTransformType = k3dIDocument::POSITION;
	m_ActiveCoordinateSystem = k3dIDocument::WORLD;
	m_ActiveEditMode = k3dIDocument::POLYGON;

	// Setup geometry ...
	m_ActiveObjectContext = 0;
	m_ActiveLayer = 0;
	m_ActiveCamera = 0;

	// Default material ...
	m_DefaultMaterial = 0;

	// Actions / undos ...
	m_CurrentAction = 0;

	// State changes for the user interface ...
	m_ActionRevision = 1;
	m_HierarchyRevision = 1;
	m_LayerRevision = 1;
	m_EditModeRevision = 1;
	m_ActiveAxisRevision = 1;
	m_ActiveTransformationRevision = 1;
	m_PlaybackModeRevision = 1;
	m_TimebaseRevision = 1;
}

CDocument::~CDocument()
{
	if(m_Application->Verbose())
		std::cout << "CDocument::~CDocument() " << m_DocumentTitle << std::endl;
		
	// Update our "dying" flag ...
	m_Dying = true;

	// Kill the animation timer ...
	if(m_AnimationTimer)
		gtk_timeout_remove(m_AnimationTimer);

	// Get rid of any UI's ...
	DeleteAllFrames();

	// Dump the hierarchy for debugging ...
	if(m_Application->Verbose())
		DumpContextHierarchy();
		
	// Delete our layers ...
	DeleteAllLayers();
	
	// We don't need the context hierarchy any more ...
	m_Hierarchy.clear();
}

// Create a new, default document from scratch ...
bool CDocument::Create(CApplication* Application)
{
	// Sanity checks ...
	g_assert(Application);

	m_Application = Application;

	// Set our default title ...
	m_DocumentTitle = sdpFormatString("Untitled Document %u", m_DocumentNumber++);

	// Create a default layer ...
	g_return_val_if_fail(CreateLayer(this), false);

	// Create a default document frame ...
	g_return_val_if_fail(CreateDocumentFrame(), false);

	// Create some default objects ...
	k3dIObject* object;

	// Create an axis object ...
	object = CreateObject(sdpAxes, false, true, true);
	g_return_val_if_fail(object, false);

	// Create the default material object ...
	object = CreateObject(sdpMaterial, false, true, true);
	g_return_val_if_fail(object, false);
	m_DefaultMaterial = sdpQueryInterface<k3dIMaterial>(object);
	g_return_val_if_fail(m_DefaultMaterial, false);
	object->Release();

	// Create a default light object ...
	object = CreateObject(sdpLight, false, true, true);
	g_return_val_if_fail(object, false);
	k3dSetPosition(object, sdpVector3(-20.0, 20.0, -30.0));

	// Create a default camera object ...
	object = CreateObject(sdpCamera, false, true, true);
	g_return_val_if_fail(object, false);

	k3dICamera* camera = sdpQueryInterface<k3dICamera>(object);
	g_return_val_if_fail(camera, false);
	object->Release();

	k3dSetPosition(camera, sdpVector3(-20.0, 20.0, -20.0));
	k3dSetOrientation(camera, sdpEulerAngles(sdpRadians(0.0), sdpRadians(37.0), sdpRadians(45.0), sdpEulerAngles::ZXYstatic));

	camera->ShowViewport();

	// Create our animation timer ...
	m_AnimationTimer = gtk_timeout_add(33, RawAnimationTimer, gpointer(this));

	return true;
}

// Open a document from disk ...
bool CDocument::Open(CApplication* Application, SDPCSTRING DocumentPath)
{
	// Sanity checks ...
	g_assert(Application);
	g_assert_string(DocumentPath);

	m_Application = Application;

	// Try loading the file as XML ...
	k3dDocument doc;
	g_return_val_if_fail(doc.LoadFile(DocumentPath), false);

	// Look for the document element ...
	sdpXmlIElement* document = sdpXmlFindElement(doc.GetRootElement(), "document");
	if(0 == document)
		return true;

	// Load layers ...
	LoadLayers(document);

	// Load objects ...
	LoadObjects(document);
	LoadObjectHierarchy(document);

	// Load odds-and-ends ...
	LoadDefaultMaterial(document);
	LoadTimeBase(document);

	// Record our path & title ...
	m_DocumentPath = DocumentPath;
	m_DocumentTitle = sdpFile(DocumentPath);

	// Let all our objects know we're done loading ...
	LoadComplete(document);

	// Load frames ...
	LoadFrames(document);

	// Force a display update ...
	k3dRedrawAll(this, false);

	// Create our animation timer ...
	m_AnimationTimer = gtk_timeout_add(33, RawAnimationTimer, gpointer(this));

	return true;
}

// Load the object hierarchy ...
sdpResult CDocument::LoadObjectHierarchy(sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_assert(Element);

	sdpXmlIElement* hierarchy = sdpXmlFindElement(Element, "objecthierarchy");
	if(!hierarchy)
		return SDP_S_FALSE;

//	ReadHierarchy(hierarchy, NULL);

	return SDP_S_OK;
}

// Load objects ...
sdpResult CDocument::LoadObjects(sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_assert(Element);

	// Load objects
	sdpXmlIElement* objects = sdpXmlFindElement(Element, "objects");
	if(!objects)
		return SDP_S_FALSE;

	// For each object in the document ...
	for(SDPXMLELEMENTHANDLE e = objects->GetFirstElement(); e; e = objects->GetNextElement(e))
		{
			sdpXmlIElement* objectelement = objects->GetElement(e);
			CreateObject(objectelement);
		}

	return SDP_S_OK;
}

// Load time parameters from an XML document ...
sdpResult CDocument::LoadTimeBase(sdpXmlIElement* Element)
{
	// Sanity check ...
	g_assert(Element);

	sdpXmlIElement* timebase = sdpXmlFindElement(Element, "timebase");
	if(!timebase)
		return SDP_S_FALSE;

	sdpXmlGetAttribute(timebase, "starttime", m_StartTime);
	sdpXmlGetAttribute(timebase, "viewtime", m_ViewTime);
	sdpXmlGetAttribute(timebase, "endtime", m_EndTime);
	sdpXmlGetAttribute(timebase, "duration", m_Duration);
	sdpXmlGetAttribute(timebase, "framerate", m_FrameRate);

	return SDP_S_OK;
}

// Load the default material from an XML document ...
sdpResult CDocument::LoadDefaultMaterial(sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_assert(Element);

	sdpXmlIElement* defaultmaterial = sdpXmlFindElement(Element, "defaultmaterial");
	if(!defaultmaterial)
		return SDP_S_FALSE;

	k3dID materialid = 0;
	sdpXmlGetAttribute(defaultmaterial, "id", materialid);
	g_assert(materialid);

	// Connect the ID to the loaded object ...
	k3dIObject* materialobject = k3dGetObject(this, materialid);
	g_assert(materialobject);

	m_DefaultMaterial = sdpQueryInterface<k3dIMaterial>(materialobject);

	return SDP_S_OK;
}

// Load layers from an XML document ...
sdpResult CDocument::LoadLayers(sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_assert(Element);

	sdpXmlIElement* layers = sdpXmlFindElement(Element, "layers");
	if(!layers)
		return SDP_S_FALSE;

	for(SDPXMLELEMENTHANDLE e = layers->GetFirstElement(); e; e = layers->GetNextElement(e))
		{
//			sdpXmlIElement* layerelement = layers->GetElement(e);
/*			CLayer* layer =*/ CreateLayer(this);
		}

	return SDP_S_OK;
}

// Load all our frames from an XML document ...
sdpResult CDocument::LoadFrames(sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_assert(Element);

	sdpXmlIElement* frames = sdpXmlFindElement(Element, "frames");
	if(!frames)
		return SDP_S_FALSE;

	// For each frame ...
	for(SDPXMLELEMENTHANDLE e = frames->GetFirstElement(); e; e = frames->GetNextElement(e))
		{
			sdpXmlIElement* frame = frames->GetElement(e);
			g_assert(frame);

			if(frame->Name() == sdpString("documentframe"))
				LoadDocumentFrame(frame);
			else if(frame->Name() == sdpString("hierarchyframe"))
				LoadHierarchyFrame(frame);
			else
				std::cout << "Unknown frame type: " << frame->Name() << std::endl;
		}

	return SDP_S_OK;
}

class CDocumentLoadComplete
{
public:
	bool HandleObject(k3dIObjectContext* Context)
	{
		// Sanity checks ...
		g_assert(Context);

		CObjectContext* context = static_cast<CObjectContext*>(Context);
		context->LoadComplete();

		return true;
	}
};


sdpResult CDocument::LoadComplete(sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_assert(Element);

	CDocumentLoadComplete functor;
	k3dForEachObject(this, 0, functor);

	return SDP_S_OK;
}

bool CDocument::IsDying()
{
	return m_Dying;
}

void CDocument::OnActivate(CApplication* Application)
{
	if(m_Application->Verbose())
		std::cout << "CDocument::OnActivate(CApplication*)" << std::endl;
		
	// Activate ourselves ...
	OnActivate();
}

void CDocument::OnActivate(CDocumentFrame* Frame)
{
	if(m_Application->Verbose())
		std::cout << "CDocument::OnActivate(CDocumentFrame*)" << std::endl;

	// Update the application's active document ...
	m_Application->OnActivate(this);

	// Activate ourselves ...
	OnActivate();
}

void CDocument::OnActivate(CHierarchyFrame* Frame)
{
	if(m_Application->Verbose())
		std::cout << "CDocument::OnActivate(CHierarchyFrame*)" << std::endl;

	// Update the application's active document ...
	m_Application->OnActivate(this);

	// Activate ourselves ...
	OnActivate();
}

void CDocument::OnActivate()
{
	if(m_Application->Verbose())
		std::cout << "CDocument::OnActivate()" << std::endl;
}

///////////////////////////////////////////////////////////////////
// Everything you ever wanted to do with frames ...

CDocumentFrame* CDocument::CreateDocumentFrame()
{
	// Create the frame ...
	CDocumentFrame* frame = new CDocumentFrame();
	g_return_val_if_fail(frame, 0);

	// Initialize it ...
	if(!frame->Create(this))
		{
			delete frame;
			g_return_val_if_fail(0, 0);
		}

	// Store it in our list ...
	m_DocumentFrames.push_back(frame);

	return frame;
}

CHierarchyFrame* CDocument::CreateHierarchyFrame()
{
	// Create the frame ...
	CHierarchyFrame* frame = new CHierarchyFrame();
	g_return_val_if_fail(frame, 0);

	// Initialize it ...
	if(!frame->Create(this))
		{
			delete frame;
			g_return_val_if_fail(0, 0);
		}

	m_HierarchyFrames.push_back(frame);

	return frame;
}

CDocumentFrame* CDocument::LoadDocumentFrame(sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_assert(Element);

	// Create the frame ...
	CDocumentFrame* frame = new CDocumentFrame();
	g_return_val_if_fail(frame, 0);

	// Initialize it ...
	if(!frame->Load(this, Element))
		{
			delete frame;
			g_return_val_if_fail(0, 0);
		}

	// Store it in our list ...
	m_DocumentFrames.push_back(frame);

	return frame;
}

CHierarchyFrame* CDocument::LoadHierarchyFrame(sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_assert(Element);

	// Create the frame ...
	CHierarchyFrame* frame = new CHierarchyFrame();
	g_return_val_if_fail(frame, 0);

	// Initialize it ...
	if(!frame->Load(this, Element))
		{
			delete frame;
			g_return_val_if_fail(0, 0);
		}

	m_HierarchyFrames.push_back(frame);

	return frame;
}

void CDocument::DeleteFrame(CDocumentFrame* Frame)
{
	if(m_Application->Verbose())
		std::cout << "CDocument::DeleteFrame(CDocumentFrame*)" << std::endl;

	// Sanity checks ...
	g_assert(Frame);

	// Find the iterator for this frame ...
	DocumentFrameIterator i = std::find(m_DocumentFrames.begin(), m_DocumentFrames.end(), Frame);
	g_assert(i != m_DocumentFrames.end());

	DeleteFrame(i);
}

void CDocument::DeleteFrame(DocumentFrameIterator Frame)
{
	if(m_Application->Verbose())
		std::cout << "CDocument::DeleteFrame(CDocumentFrame*)" << std::endl;

	// Remove the frame ...
	delete (*Frame);
	m_DocumentFrames.erase(Frame);

	// If we're not already dying, see if this was the last frame ...
	if(!IsDying())
		if(!m_DocumentFrames.size())
			m_Application->CloseDocument(this);
}

void CDocument::DeleteFrame(CHierarchyFrame* Frame)
{
	if(m_Application->Verbose())
		std::cout << "CDocument::DeleteFrame(CHierarchyFrame*)" << std::endl;

	// Sanity checks ...
	g_assert(Frame);

	// Find the iterator for this frame ...
	HierarchyFrameIterator i = std::find(m_HierarchyFrames.begin(), m_HierarchyFrames.end(), Frame);
	g_assert(i != m_HierarchyFrames.end());

	DeleteFrame(i);
}


void CDocument::DeleteFrame(HierarchyFrameIterator Frame)
{
	if(m_Application->Verbose())
		std::cout << "CDocument::DeleteFrame(HierarchyFrameIterator)" << std::endl;

	// Remove the frame ...
	delete (*Frame);
	m_HierarchyFrames.erase(Frame);
}

void CDocument::DeleteAllFrames()
{
	if(m_Application->Verbose())
		std::cout << "CDocument::DeleteAllFrames()" << std::endl;

	// For each hierarchy frame ...
	for(HierarchyFrameIterator hierarchyframe = m_HierarchyFrames.begin(); hierarchyframe != m_HierarchyFrames.end(); hierarchyframe = m_HierarchyFrames.begin())
		(*hierarchyframe)->RootWidget().Destroy();

	// For each document frame ...
	for(DocumentFrameIterator docframe = m_DocumentFrames.begin(); docframe != m_DocumentFrames.end(); docframe = m_DocumentFrames.begin())
		(*docframe)->RootWidget().Destroy();
}

//////////////////////////////////////////////////////////////////////////////////////
// Layers

CLayer* CDocument::CreateLayer(CDocument* Document)
{
	// Create a new layer ...
	CLayer* layer = new CLayer();
	g_return_val_if_fail(layer, 0);

	// Initialize it ...
	if(!layer->Create(this))
		{
			delete layer;
			g_return_val_if_fail(0, 0);
		}

	m_Layers.push_back(layer);

	// Make this the active layer ...
	SetActiveLayer(layer);

	// Let the UI update itself ...
	m_LayerRevision++;

	return layer;
}

bool CDocument::CanDeleteLayer(CLayer* Layer)
{
	return m_Layers.size() > 1;
}

void CDocument::DeleteLayer(CLayer* Layer)
{
	// Sanity checks ...
	g_assert(Layer);

	// Find the iterator for this layer ...
	LayerIterator i = std::find(m_Layers.begin(), m_Layers.end(), Layer);
	g_assert(i != m_Layers.end());

	// Delete the layer ...
	DeleteLayer(*i);
}

void CDocument::DeleteLayer(LayerIterator Layer)
{
	// Delete the layer ...
	delete (*Layer);
	m_Layers.erase(Layer);

	// Let the UI update itself ...
	m_LayerRevision++;
}

void CDocument::DeleteAllLayers()
{
	// For each layer ...
	for(LayerIterator layer = m_Layers.begin(); layer != m_Layers.end(); layer = m_Layers.begin())
		DeleteLayer(layer);
}

void CDocument::SetActiveLayer(CLayer* Layer)
{
	m_ActiveLayer = Layer;
}

/////////////////////////////////////////////////////////////////////////////
// CDocument commands

k3dID CDocument::GetNextObjectID()
{
	return ++m_CurrentObjectID;
}

/////////////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CDocument::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIDocument>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dITimeBase>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIDocument*>(this), InterfaceID, Interface)) return SDP_S_OK;

	return SDP_E_NO_INTERFACE;
}


gulong CDocument::AddRef()
{
	return 1;
}


gulong CDocument::Release()
{
	return 1;
}




///////////////////////////////////////////////////////////////
// k3dIDocument overrides

sdpResult CDocument::FreezeChanges()
{
	// Calculate time delta
	m_RenderDeltaTime = GetFrameDelta();

	// Pre-animate the hierarchy ...
	K3DLOCATIONHANDLE locationhandle;
	k3dILocation* location;
	sdpVector3 position;
	sdpAngleAxis orientation;
	sdpVector3 scale;
	sdpVector3 coords;
	CObjectContext* object;

	// For each object ...
	HierarchyTreeIterator end = m_Hierarchy.end();
	for(HierarchyTreeIterator i = m_Hierarchy.begin(); i != end; i++)
		{
			// Don't call our objects twice ...
			if(i->m_Type != k3dIDocument::PUSH)
				continue;

			// Get the object ...
			object = i->m_ObjectContext;

			// Reset transformations ...
			if(object->m_Transform)
				{
					object->m_Transform->GetPosition(k3dITransform::CURRENT_FRAME, position);
					object->m_Transform->SetPosition(k3dITransform::OLD_FRAME, position);
					object->m_Transform->GetOrientation(k3dITransform::CURRENT_FRAME, orientation);
					object->m_Transform->SetOrientation(k3dITransform::OLD_FRAME, orientation);
					object->m_Transform->GetScale(k3dITransform::CURRENT_FRAME, scale);
					object->m_Transform->SetScale(k3dITransform::OLD_FRAME, scale);

					object->m_Transform->GetPosition(k3dITransform::CURRENT_CHILD_FRAME, position);
					object->m_Transform->SetPosition(k3dITransform::OLD_CHILD_FRAME, position);
					object->m_Transform->GetOrientation(k3dITransform::CURRENT_CHILD_FRAME, orientation);
					object->m_Transform->SetOrientation(k3dITransform::OLD_CHILD_FRAME, orientation);
					object->m_Transform->GetScale(k3dITransform::CURRENT_CHILD_FRAME, scale);
					object->m_Transform->SetScale(k3dITransform::OLD_CHILD_FRAME, scale);
				}

			// Reset locations ...
			if(object->m_LocationServer)
				for(locationhandle = object->m_LocationServer->GetFirstLocation(); locationhandle; locationhandle = object->m_LocationServer->GetNextLocation(locationhandle))
					{
						location = object->m_LocationServer->GetLocation(locationhandle);
						location->GetLocalXYZ(coords);
						location->SetOldXYZ(coords);
					}
		}

	return SDP_S_OK;
}


sdpResult CDocument::AnimateDocument()
{
	// Calculate time delta
	m_RenderDeltaTime = GetFrameDelta();

	// Pre-animate the hierarchy ...
	K3DLOCATIONHANDLE locationhandle;
	k3dILocation* location;
	HierarchyTreeIterator i;
	HierarchyTreeIterator end;
	sdpVector3 position;
	sdpAngleAxis orientation;
	sdpVector3 scale;
	sdpVector3 coords;
	CObjectContext* object;

	// For each object ...
	end = m_Hierarchy.end();
	for(i = m_Hierarchy.begin(); i != end; i++)
		{
			// Don't call our objects twice ...
			if(i->m_Type != k3dIDocument::PUSH)
				continue;

			// Get the object ...
			object = i->m_ObjectContext;

			// Reset times ...
			if(object->m_TimeServer)
				object->m_TimeServer->SetAnimTime(m_RenderTime, m_RenderDeltaTime);

			// Reset transformations ...
			if(object->m_Transform)
				{
					object->m_Transform->GetPosition(k3dITransform::CURRENT_FRAME, position);
					object->m_Transform->SetPosition(k3dITransform::ANIMATED_FRAME, position);
					object->m_Transform->GetOrientation(k3dITransform::CURRENT_FRAME, orientation);
					object->m_Transform->SetOrientation(k3dITransform::ANIMATED_FRAME, orientation);
					object->m_Transform->GetScale(k3dITransform::CURRENT_FRAME, scale);
					object->m_Transform->SetScale(k3dITransform::ANIMATED_FRAME, scale);

					object->m_Transform->GetPosition(k3dITransform::CURRENT_CHILD_FRAME, position);
					object->m_Transform->SetPosition(k3dITransform::ANIMATED_CHILD_FRAME, position);
					object->m_Transform->GetOrientation(k3dITransform::CURRENT_CHILD_FRAME, orientation);
					object->m_Transform->SetOrientation(k3dITransform::ANIMATED_CHILD_FRAME, orientation);
					object->m_Transform->GetScale(k3dITransform::CURRENT_CHILD_FRAME, scale);
					object->m_Transform->SetScale(k3dITransform::ANIMATED_CHILD_FRAME, scale);
				}

			// Reset locations ...
			if(object->m_LocationServer)
				for(locationhandle = object->m_LocationServer->GetFirstLocation(); locationhandle; locationhandle = object->m_LocationServer->GetNextLocation(locationhandle))
					{
						location = object->m_LocationServer->GetLocation(locationhandle);
						location->GetLocalXYZ(coords);
						location->SetAnimXYZ(coords);
					}

			// Pre-animate ...
			if(object->m_Animate)
				object->m_Animate->PreAnimate();
		}

	// Animate the document ...
	end = m_Hierarchy.end();
	for(i = m_Hierarchy.begin(); i != end; i++)
		{
			object = i->m_ObjectContext;
			if(!object->m_Animate)
				continue;

			switch(i->m_Type)
				{
					case k3dIDocument::PUSH:
						object->m_Animate->PushAnimate();
						break;
					case k3dIDocument::NODE:
						object->m_Animate->Animate();
						break;
					case k3dIDocument::POP:
						object->m_Animate->PopAnimate();
						break;
					default:
						g_assert_not_reached();
						break;
				}
		}

	// Post-animate pass for the hierarchy ...
	end = m_Hierarchy.end();
	for(i = m_Hierarchy.begin(); i != end; i++)
		{
			object = i->m_ObjectContext;
			if(!object->m_Animate)
				continue;

			object->m_Animate->PostAnimate();
		}

	return SDP_S_OK;
}


// Store world coordinates for each location in the document ...
sdpResult CDocument::SnapshotWorldCoordinates()
{
	// Setup a matrix stack ...
	sdpMatrix4 matrix = sdpIdentity3D();
	std::stack<sdpMatrix4> stack;

	// Setup local variables for speed ...
	CObjectContext* context;
	k3dILocationServer* locationserver;
	K3DLOCATIONHANDLE locationhandle;
	k3dILocation* location;
	sdpVector3 coords;

	// For each object ...
	HierarchyTreeIterator end = m_Hierarchy.end();
	for(HierarchyTreeIterator i = m_Hierarchy.begin(); i != end; i++)
		{
			context = i->m_ObjectContext;

			switch(i->m_Type)
				{
					case k3dIDocument::PUSH:
						// Push the object's matrix onto the stack ...
						if(context->m_Transform)
							{
								stack.push(matrix);
								matrix = matrix * k3dGetTransformMatrix(context->m_Transform, k3dITransform::ANIMATED_FRAME);
							}
						break;

					case k3dIDocument::NODE:
						// If this object contains locations ...
						if(context->m_LocationServer)
							{
								locationserver = context->m_LocationServer;
								for(locationhandle = locationserver->GetFirstLocation(); locationhandle; locationhandle = locationserver->GetNextLocation(locationhandle))
									{
										location = locationserver->GetLocation(locationhandle);
										location->GetAnimXYZ(coords);
										coords = matrix * coords;
										location->SetWorldXYZ(coords);
									}
									
								matrix = matrix * k3dGetTransformMatrix(context->m_Transform, k3dITransform::ANIMATED_CHILD_FRAME);
							}
						break;

					case k3dIDocument::POP:
						// Pop the object's matrix from the stack ...
						if(context->m_Transform)
							{
								matrix = stack.top();
								stack.pop();
							}
						break;
						
					default:
						g_assert_not_reached();
						break;
				}
		}

	return SDP_S_OK;
}

sdpResult CDocument::AttachObjectContext(CObjectInfo* Info, k3dIObject* Object, bool UserVisible, bool Saveable)
{
	// Sanity checks ...
	g_assert(Info);
	g_assert(Object);

	// We should ALWAYS have an active layer ...
	g_assert(m_ActiveLayer);

	return m_ActiveLayer->AttachObjectContext(Info, Object, UserVisible, Saveable);
}

sdpResult CDocument::AttachObjectContext(CObjectInfo* Info, k3dIObject* Object, sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_assert(Info);
	g_assert(Object);
	g_assert(Element);

	// We should ALWAYS have an active layer ...
	g_assert(m_ActiveLayer);

	return m_ActiveLayer->AttachObjectContext(Info, Object, Element);
}

///////////////////////////////////////////////////////////////////////////
// k3dIActions overrides ...

sdpResult CDocument::StartAction(SDPCSTRING Name)
{
	// Sanity checks ...
	g_return_val_if_string_fail(Name, SDP_E_FAIL);

	// Make sure we're not already in the middle of an action ...
	g_return_val_if_fail(!m_CurrentAction, SDP_E_FAIL);

	if(m_Application->Verbose())
		std::cout << "CDocument::StartAction(" << Name << ")" << std::endl;

	// Create the action ...
	m_ActionStack.push_back(CAction(Name));
	m_CurrentAction = &m_ActionStack.back();

	return SDP_S_OK;
}

sdpResult CDocument::IsRecording()
{
	// Let the caller know whether we're in the middle of an action or not ...
	return m_CurrentAction ? SDP_S_OK : SDP_S_FALSE;
}

sdpResult CDocument::RecordStep(k3dIActionClient* Client, gulong OldValueCode, gulong OldValue, gulong OldValueSize, gulong NewValueCode, gulong NewValue, gulong NewValueSize)
{
	// If we're not in the middle of an action, ignore it ...
	if(!m_CurrentAction)
		return SDP_S_FALSE;

	// Sanity checks ...
	g_return_val_if_fail(Client, SDP_E_FAIL);
	g_return_val_if_fail(m_ActionStack.size(), SDP_E_FAIL);

	// Store the step ...
	m_CurrentAction->Steps().push_back(CStep(Client, OldValueCode, OldValue, OldValueSize, NewValueCode, NewValue, NewValueSize));

	if(m_Application->Verbose())
		std::cout << "   CDocument::RecordStep()" << std::endl;

	return SDP_S_OK;
}

sdpResult CDocument::FinishAction()
{
	// Sanity checks ...
	g_return_val_if_fail(m_CurrentAction, SDP_E_FAIL);

	m_CurrentAction = 0;

	if(m_Application->Verbose())
		std::cout << "CDocument::FinishAction()" << std::endl;
			
	// Get rid of empty actions to keep the UI uncluttered ...
	if(m_ActionStack.back().Steps().size() == 0)
		{
			if(m_Application->Verbose())
				std::cout << "Pruning empty action \"" << m_ActionStack.back().Name() << "\"." << std::endl;
			
			m_ActionStack.pop_back();
			
			return SDP_S_FALSE;
		}

	// Get rid of any stale redos ...
	m_RedoStack.clear();

	// Let the UI update itself ...
	m_ActionRevision++;

	return SDP_S_OK;
}

gulong CDocument::ActionCount()
{
	return m_ActionStack.size();
}

gulong CDocument::RedoCount()
{
	return m_RedoStack.size();
}

sdpResult CDocument::UndoAction()
{
	// Should never happen in the middle of defining an action ...
	g_return_val_if_fail(!m_CurrentAction, SDP_E_FAIL);

	// Make sure we have actions to undo ...
	g_return_val_if_fail(m_ActionStack.size(), SDP_E_FAIL);

	if(m_Application->Verbose())
		std::cout << "CDocument::UndoAction(" << m_ActionStack.back().Name() << ")" << std::endl;

	// Undo the most recent action ...
	StepList& steps = m_ActionStack.back().Steps();

	// For each step ...
	for(StepIterator step = steps.begin(); step != steps.end(); step++)
		step->Undo();

	// Move the action to the redo stack ...
	m_RedoStack.push_back(m_ActionStack.back());
	m_ActionStack.pop_back();

	// Freeze the new state ...
	FreezeChanges();

	// Let the UI update itself ...
	m_ActionRevision++;

	return SDP_S_OK;
}


sdpResult CDocument::RedoAction()
{
	// Should never happen in the middle of defining an action ...
	g_return_val_if_fail(!m_CurrentAction, SDP_E_FAIL);

	// Make sure we have actions to undo ...
	g_return_val_if_fail(m_RedoStack.size(), SDP_E_FAIL);

	if(m_Application->Verbose())
		std::cout << "CDocument::RedoAction(" << m_RedoStack.back().Name() << ")" << std::endl;

	// Undo the most recent action ...
	StepList& steps = m_RedoStack.back().Steps();

	// For each step ...
	for(StepIterator step = steps.begin(); step != steps.end(); step++)
		step->Redo();

	// Move the action to the undo stack ...
	m_ActionStack.push_back(m_RedoStack.back());
	m_RedoStack.pop_back();

	// Freeze the new state ...
	FreezeChanges();

	// Let the UI update itself ...
	m_ActionRevision++;

	return SDP_S_OK;
}

////////////////////////////////////////////////////////////////////////////////
// k3dIActionClient overrides

sdpResult CDocument::HandleStep(gulong ValueCode, gulong Value, gulong ValueSize)
{
	switch(ValueCode)
		{
			case STEP_CREATE_OBJECT:
			case STEP_UNDELETE_OBJECT:
				{
					CObjectContext* context = reinterpret_cast<CObjectContext*>(Value);
			
					// Create some new hierarchy records ...
					m_Hierarchy.push_back(CHierarchyTreeRecord(k3dIDocument::PUSH, context));
					m_Hierarchy.push_back(CHierarchyTreeRecord(k3dIDocument::NODE, context));
					m_Hierarchy.push_back(CHierarchyTreeRecord(k3dIDocument::POP, context));
			
					// Encourage the UI to update itself ...
					m_HierarchyRevision++;
			
					return SDP_S_OK;
				}
			
			case STEP_UNCREATE_OBJECT:
			case STEP_DELETE_OBJECT:
				{
					CObjectContext* context = reinterpret_cast<CObjectContext*>(Value);

					// Get rid of the hierarchy records ...
					HierarchyTreeIterator push = std::find(m_Hierarchy.begin(), m_Hierarchy.end(), CHierarchyTreeRecord(k3dIDocument::PUSH, context));
					g_assert(push != m_Hierarchy.end());
					m_Hierarchy.erase(push);

					HierarchyTreeIterator node = std::find(m_Hierarchy.begin(), m_Hierarchy.end(), CHierarchyTreeRecord(k3dIDocument::NODE, context));
					g_assert(node != m_Hierarchy.end());
					m_Hierarchy.erase(node);
					
					HierarchyTreeIterator pop = std::find(m_Hierarchy.begin(), m_Hierarchy.end(), CHierarchyTreeRecord(k3dIDocument::POP, context));
					g_assert(pop != m_Hierarchy.end());
					m_Hierarchy.erase(pop);
									
					// Encourage the UI to update itself ...
					m_HierarchyRevision++;
			
					return SDP_S_OK;
				}
		}

	return k3dActionClient::HandleStep(ValueCode, Value, ValueSize);
}

sdpString CDocument::GetUniqueObjectName(sdpString Name)
{
	// Trim any leading / trailing whitespace from the name ...
	Name.trim();

	// See if the name already exists ...
	for(K3DOBJECTCONTEXTHANDLE handle = GetFirstObjectContext(); handle; handle = GetNextObjectContext(handle))
		{
			if(GetObjectContext(handle)->Name() == Name)
				{
					Name += "*";

					// Recursively see if the new name works ...
					Name = GetUniqueObjectName(Name);
					return Name;
				}
		}
		
	return Name;
}

// Called by the animation interval timer
gint CDocument::RawAnimationTimer(gpointer Data)
{
	// Call the object-instance version ...
	return ((CDocument*)Data)->AnimationTimer();
}

// Called by the animation interval timer
gint CDocument::AnimationTimer()
{
	// We only pay attention to every-other timer signal ...
	m_IgnoreAnimationTimer = !m_IgnoreAnimationTimer;
	if(m_IgnoreAnimationTimer)
		return true;

	switch(m_PlaybackMode)
		{
			case REVERSEPLAY:
				{
					// Calculate our new time ...
					gdouble newtime = GetViewTime() - GetFrameDelta();
					if(newtime < GetStartTime())
						{
							// We passed the beginning of the animation, so stop ...
							newtime = GetStartTime();
							SetPlaybackMode(STOP);
						}

					// Set the new time ...
					SetViewTime(newtime);

					// Redraw everything ...
					k3dRedrawAll(this, false);
				}
				break;

			case STOP:
				break;

			case PLAY:
				{
					// Calculate our new time ...
					gdouble newtime = GetViewTime() + GetFrameDelta();
					if(newtime > GetEndTime())
						{
							// We passed the end of the animation, so stop ...
							newtime = GetEndTime();
							SetPlaybackMode(STOP);
						}

					// Set the new time ...
					SetViewTime(newtime);

					// Redraw everything ...
					k3dRedrawAll(this, false);
				}
				break;

			case LOOPPLAY:
				{
					// Calculate our new time ...
					gdouble newtime = GetViewTime() + GetFrameDelta();
					if(newtime > GetEndTime())
						newtime = GetStartTime();

					// Set the new time ...
					SetViewTime(newtime);

					// Redraw everything ...
					k3dRedrawAll(this, false);
				}
				break;

			default:
				g_assert_not_reached();		// Did you add a new playback mode without updating this code?
		}

	// Continue indefinitely ...
	return true;
}

void CDocument::DumpContextHierarchy()
{
	std::cout << "CDocument::DumpContextHierarchy()" << std::endl;

	static SDPCSTRING const indent = "   ";
	sdpString indentation;

	for(HierarchyTreeIterator i = m_Hierarchy.begin(); i != m_Hierarchy.end(); i++)
		{
			if(k3dIDocument::POP == i->m_Type)
				indentation = indentation.left(indentation.size() - strlen(indent));

			std::cout << SDPCSTRING(indentation);

			switch(i->m_Type)
				{
					case k3dIDocument::COMPLETE:
						std::cout << "COMPLETE - ";
						break;
					case k3dIDocument::PUSH:
						std::cout << "PUSH - ";
						break;
					case k3dIDocument::NODE:
						std::cout << "NODE - ";
						break;
					case k3dIDocument::POP:
						std::cout << "POP - ";
						break;
					default:
						std::cout << "UNKNOWN: " << gint(i->m_Type) << " - ";
						break;
				}

			std::cout << ((void*)i->m_ObjectContext) << " - " << i->m_ObjectContext->Name() << std::endl;

			if(k3dIDocument::PUSH == i->m_Type)
				indentation += indent;
		}
}


void CDocument::ValidateContextHierarchy()
{
	if(m_Application->Verbose())
		std::cout << "CDocument::ValidateContextHierarchy()" << std::endl;

	// The hierarchy size MUST be a multiple of three ...
	if(0 != (m_Hierarchy.size() % 3))
		std::cout << "*** Illegal hierarchy size (not a multiple of 3): " << m_Hierarchy.size()  << std::endl;

	// Step through the hierarchy to make sure there are no over/under-flows or unknown record types ...
	int stack = 0;
	for(HierarchyTreeIterator i = m_Hierarchy.begin(); i != m_Hierarchy.end(); i++)
		{
			switch(i->m_Type)
				{
					case k3dIDocument::PUSH:
						stack++;
						break;
					case k3dIDocument::NODE:
						break;
					case k3dIDocument::POP:
						if(stack)
							{
								stack--;
							}
						else
							std::cout << "*** Hierarchy stack underflow" << std::endl;
						break;
					default:
						std::cout << "*** Unknown hierarchy record type: " << gint(i->m_Type) << std::endl;
						break;
				}
		}

	if(stack)
		std::cout << "*** Hierarchy stack overflow: " << stack << std::endl;
}

/////////////////////////////////////////////////////////////////////////////////////////////
// k3dIDocument overrides

k3dIApplication* CDocument::Application()
{
	return m_Application;
}

sdpResult CDocument::Import(SDPCSTRING FilePath, SDPCSTRING Format, k3dIGeometryReadFormat* FormatFilter)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);
	g_return_val_if_string_fail(Format, SDP_E_FAIL);

	// Our result ...
	sdpResult result;

	// If a format was provided, use it ...
	k3dIGeometryReadFormat* formatfilter = FormatFilter;
	if(formatfilter)
		{
			// Give the format a chance to look at the file, first ...
			result = formatfilter->PreRead(FilePath, Format);
			if(sdpSucceeded(result))
				{
					// Prompt the user for any options before reading ...
					result = formatfilter->ReadOptions(FilePath, Format);
					if(sdpSucceeded(result))
						{
							// Read the file ...
							result = formatfilter->ReadFile(FilePath, Format);
						}
				}
		}
	// A format wasn't provided ...
	else
		{
			// ... so come up with one of our own ...
			formatfilter = k3dFileFormat<k3dIGeometryReadFormat>(this, FilePath, Format);

			// If we couldn't find a format, we're done ...
			g_return_val_if_fail(formatfilter, SDP_E_FAIL);

			// Give the format a chance to look at the file, first ...
			result = formatfilter->PreRead(FilePath, Format);
			if(sdpSucceeded(result))
				{
					// Prompt the user for any options before reading ...
					result = formatfilter->ReadOptions(FilePath, Format);
					if(sdpSucceeded(result))
						{
							// Read the file ...
							result = formatfilter->ReadFile(FilePath, Format);
						}
				}

			// Since we created the format object ourselves, zap it ...
			k3dDeleteObject(formatfilter);
		}

	return result;
}

sdpResult CDocument::Export(SDPCSTRING FilePath, SDPCSTRING Format, k3dIGeometryWriteFormat* FormatFilter)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);
	g_return_val_if_string_fail(Format, SDP_E_FAIL);

	// Our result ...
	sdpResult result;

	// If a format was provided, use it ...
	k3dIGeometryWriteFormat* formatfilter = FormatFilter;
	if(formatfilter)
		{
			// Give the format a chance to look at the file, first ...
			result = formatfilter->PreWrite(FilePath, Format);
			if(sdpSucceeded(result))
				{
					// Prompt the user for any options before writing ...
					result = formatfilter->WriteOptions(FilePath, Format);
					if(sdpSucceeded(result))
						{
							// Read the file ...
							result = formatfilter->WriteFile(FilePath, Format);
						}
				}
		}
	// A format wasn't provided ...
	else
		{
			// ... so come up with one of our own ...
			formatfilter = k3dFileFormat<k3dIGeometryWriteFormat>(this, FilePath, Format);

			// If we couldn't find a format, we're done ...
			g_return_val_if_fail(formatfilter, SDP_E_FAIL);

			// Give the format a chance to look at the file, first ...
			result = formatfilter->PreWrite(FilePath, Format);
			if(sdpSucceeded(result))
				{
					// Prompt the user for any options before writing ...
					result = formatfilter->WriteOptions(FilePath, Format);
					if(sdpSucceeded(result))
						{
							// Read the file ...
							result = formatfilter->WriteFile(FilePath, Format);
						}
				}

			// Since we created the format object ourselves, zap it ...
			k3dDeleteObject(formatfilter);
		}

	return result;
}

// Save the document to an existing path (returns SDP_S_FALSE if none) ...
sdpResult CDocument::Save()
{
	// If we don't have a filepath, yet, we're done ...
	if(m_DocumentPath.empty())
		return SDP_S_FALSE;

	// Create our output document and dependencies objects ...
	sdpXmlDocument xmldocument;
	k3dDependencies dependencies;

	// Start the document ...
	sdpXmlIElement* root = xmldocument.CreateRootElement("k3dml");
	
	// Application-wide stuff ...
	sdpXmlCreateElement(root, "application");

	// The document itself ...
	sdpXmlIElement* document = sdpXmlCreateElement(root, "document");

	SaveTimeBase(document, &dependencies);
	SaveDefaultMaterial(document, &dependencies);

//	WriteMaterialLookupTable(&stream, &dependencies);	// This function MUST be called before objects are saved

	// Save our layers ...
	SaveLayers(document, &dependencies);

	// Save our various frames ...
	SaveFrames(document, &dependencies);

	// Objects ...
	SaveObjects(document, &dependencies);
	SaveObjectHierarchy(document, &dependencies);

	// Try to save the output file ...
	g_return_val_if_fail(xmldocument.SaveFile(m_DocumentPath), SDP_E_FAIL);

	return SDP_S_OK;
}

// Save the document to the given path ...
sdpResult CDocument::SaveAs(SDPCSTRING FilePath)
{
	// Record our new path & title ...
	m_DocumentPath = FilePath;
	m_DocumentTitle = sdpFile(FilePath);

	return Save();
}

// Store the current time state in our document stream ...
sdpResult CDocument::SaveTimeBase(sdpXmlIElement* Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert(Dependencies);

	sdpXmlCreateElement(Element, "timebase", 0, "starttime", m_StartTime, "viewtime", m_ViewTime, "endtime", m_EndTime, "duration", m_Duration, "framerate", m_FrameRate);

	return SDP_S_OK;
}

// Store the default material in our document stream ...
sdpResult CDocument::SaveDefaultMaterial(sdpXmlIElement* Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert(Dependencies);

	// Get the material object ...
	k3dIObject* object = sdpQueryInterface<k3dIObject>(m_DefaultMaterial);
	g_assert(object);

	sdpXmlCreateElement(Element, "defaultmaterial", 0, "id", object->ObjectContext()->ObjectID());

	return SDP_S_OK;
}

// Functor for saving information about a document layer ...
class CDocumentSaveLayer
{
public:
	CDocumentSaveLayer(sdpXmlIElement* Element, k3dIDependencies* Dependencies) { m_Element = Element; m_Dependencies = Dependencies; }

	bool operator()(CLayer* Layer)
	{
		// Sanity checks ...
		g_assert(Layer);

		sdpXmlCreateElement(m_Element, "layer");

		return true;
	}

	sdpXmlIElement* m_Element;
	k3dIDependencies* m_Dependencies;
};

// Save layers to our document stream ...
sdpResult CDocument::SaveLayers(sdpXmlIElement* Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert(Dependencies);

	sdpXmlIElement* layers = sdpXmlCreateElement(Element, "layers");

	CDocumentSaveLayer functor(layers, Dependencies);
	ForEachLayer(this, functor);

	return SDP_S_OK;
}

// Functor for saving information about a document frame ...
class CDocumentSaveDocumentFrame
{
public:
	CDocumentSaveDocumentFrame(sdpXmlIElement* Element, k3dIDependencies* Dependencies) { m_Element = Element; m_Dependencies = Dependencies; }
	bool operator()(CDocumentFrame* Frame)
	{
		// Sanity checks ...
		g_assert(Frame);
		Frame->Save(m_Element, m_Dependencies);

		return true;
	}

	sdpXmlIElement* m_Element;
	k3dIDependencies* m_Dependencies;
};

// Functor for saving information about a hierarchy frame ...
class CDocumentSaveHierarchyFrame
{
public:
	CDocumentSaveHierarchyFrame(sdpXmlIElement* Element, k3dIDependencies* Dependencies) { m_Element = Element; m_Dependencies = Dependencies; }
	bool operator()(CHierarchyFrame* Frame)
	{
		// Sanity checks ...
		g_assert(Frame);
		Frame->Save(m_Element, m_Dependencies);

		return true;
	}

	sdpXmlIElement* m_Element;
	k3dIDependencies* m_Dependencies;
};

// Store information about open windows ...
sdpResult CDocument::SaveFrames(sdpXmlIElement* Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert(Dependencies);

	sdpXmlIElement* frames = sdpXmlCreateElement(Element, "frames");

	// Save document frames ...
	CDocumentSaveDocumentFrame savedocumentframe(frames, Dependencies);
	ForEachDocumentFrame(this, savedocumentframe);

	CDocumentSaveHierarchyFrame savehierarchyframe(frames, Dependencies);
	ForEachHierarchyFrame(this, savehierarchyframe);

	return SDP_S_OK;
}

// Functor for notifying objects that they should save themselves ...
class CDocumentSave
{
public:
	CDocumentSave(sdpXmlIElement* Element, k3dIDependencies* Dependencies)
	{
		m_Element = Element;
		m_Dependencies = Dependencies;
		m_Count = 0;
	}

	bool HandleObject(k3dIObjectContext* Context)
	{
		// Get our object context ...
		CObjectContext* context = static_cast<CObjectContext*>(Context);

		// Save the object ...
		context->Save(m_Element, m_Dependencies);

		m_Count++;

		return true;
	}

	sdpXmlIElement* m_Element;
	k3dIDependencies* m_Dependencies;
	gulong m_Count;
};

// Functor for notifying objects that the save operation has been completed ...
class CDocumentSaveComplete
{
public:
	bool HandleObject(k3dIObjectContext* Context)
	{
		// Get our object context ...
		CObjectContext* context = static_cast<CObjectContext*>(Context);

		// Let the object know we're done saving the document ...
		context->SaveComplete();

		return true;
	}
};

// Save all our objects ...
sdpResult CDocument::SaveObjects(sdpXmlIElement* Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert(Dependencies);

	sdpXmlIElement* objects = sdpXmlCreateElement(Element, "objects");

	// Save every object ...
	CDocumentSave functorsave(objects, Dependencies);
	k3dForEachObject(this, 0, functorsave);

	// Let each object know that the save is complete ...
	CDocumentSaveComplete functorsavecomplete;
	k3dForEachObject(this, 0, functorsavecomplete);

	return SDP_S_OK;
}


// Store the object hierarchy ...
sdpResult CDocument::SaveObjectHierarchy(sdpXmlIElement* Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert(Dependencies);

	sdpXmlIElement* objecthierarchy = sdpXmlCreateElement(Element, "objecthierarchy");

	std::stack<sdpXmlIElement*> elementstack;
	elementstack.push(objecthierarchy);

	for(HierarchyTreeIterator i = m_Hierarchy.begin(); i != m_Hierarchy.end(); i++)
		{
			// Skip objects that don't support persistence ...
			if(!i->m_ObjectContext->m_Persist)
				continue;

			// Skip objects that aren't marked as saveable ...
			if(SDP_S_OK != i->m_ObjectContext->QueryState(k3dIState::SAVEABLE))
				continue;

			if(k3dIDocument::PUSH == i->m_Type)
				elementstack.push(sdpXmlCreateElement(elementstack.top(), "object", 0, "id", i->m_ObjectContext->ObjectID()));
			else if(k3dIDocument::POP == i->m_Type)
				elementstack.pop();
		}

	return SDP_S_OK;
}

// Return a pointer to our k3dIActions interface ...
k3dIActions* CDocument::Actions()
{
	return this;
}

// Canonical object instance creation ...
k3dIObject* CDocument::CreateObject(const sdpCLSID& ClassID, bool Edit, bool UserVisible, bool Saveable)
{
	// Try to find the object info for this ClassID ...
	CObjectInfo* info = m_Application->Libraries().ObjectInfo(ClassID);

	// If we don't support this object type, fail ...
	g_return_val_if_fail(info, 0);

	// Create an instance of the object ...
	k3dIObject* object = 0;
	g_return_val_if_fail(sdpSucceeded(info->ClassProcedure()(ClassID, k3dIObject::iid, (void**)&object)), 0);
	g_return_val_if_fail(object, 0);

	// Attach an object context to the object ...
	g_return_val_if_fail(sdpSucceeded(AttachObjectContext(info, object, UserVisible, Saveable)), 0);

	// Get the object context ...
	CObjectContext* context = static_cast<CObjectContext*>(object->ObjectContext());

	// Are we recording actions?
	if(SDP_S_OK == IsRecording())
		RecordStep(this, STEP_UNCREATE_OBJECT, reinterpret_cast<gulong>(context), 0,
										STEP_CREATE_OBJECT, reinterpret_cast<gulong>(context), 0);

	// Open the object for editing as requested ...
	if(Edit)
		object->EditObject();

	return object;
}

// Destroy an object ...
sdpResult CDocument::DeleteObject(k3dIObject* Object)
{
	if(m_Application->Verbose())
		std::cout << "CDocument::DeleteObject() " << Object->ObjectContext()->Name() << std::endl;

	// Sanity checks ...
	g_return_val_if_fail(Object, SDP_E_FAIL);

/*
	// See if the object will allow it ...
	if(SDP_S_OK != Object->QueryDeleteObject())
		return SDP_S_FALSE;
*/

	// Let the object know that it's getting deleted ...
	Object->ObjectDeleted();

	// Get the object context ...
	CObjectContext* context = static_cast<CObjectContext*>(Object->ObjectContext());

	// If this is the active object, de-activate it ...
	if(m_ActiveObjectContext == context)
		SetActiveObjectContext(0);

	// Tell the layer to hide it ...
	static_cast<CLayer*>(Object->ObjectContext()->Layer())->DeleteObjectContext(context);	

	// Get rid of the hierarchy records ...
	HierarchyTreeIterator push = std::find(m_Hierarchy.begin(), m_Hierarchy.end(), CHierarchyTreeRecord(k3dIDocument::PUSH, context));
	g_assert(push != m_Hierarchy.end());
	m_Hierarchy.erase(push);

	HierarchyTreeIterator node = std::find(m_Hierarchy.begin(), m_Hierarchy.end(), CHierarchyTreeRecord(k3dIDocument::NODE, context));
	g_assert(node != m_Hierarchy.end());
	m_Hierarchy.erase(node);
	
	HierarchyTreeIterator pop = std::find(m_Hierarchy.begin(), m_Hierarchy.end(), CHierarchyTreeRecord(k3dIDocument::POP, context));
	g_assert(pop != m_Hierarchy.end());
	m_Hierarchy.erase(pop);

	// Are we recording actions?
	if(SDP_S_OK == IsRecording())
		// If so, record our undo data ...
		RecordStep(this, STEP_UNDELETE_OBJECT, reinterpret_cast<gulong>(context), 0,
										STEP_DELETE_OBJECT, reinterpret_cast<gulong>(context), 0);

	// Encourage the UI to update itself ...
	m_HierarchyRevision++;
	
	return SDP_S_OK;
}

// Create an object from an XML document ...
sdpResult CDocument::CreateObject(sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert(Element->Name() == sdpString("object"));

	// Get the object's class ID ...
	sdpCLSID classid = sdpCLSIDNull;
	sdpXmlGetAttribute(Element, "class", classid);
	g_return_val_if_fail(classid != sdpCLSIDNull, SDP_E_FAIL);

	// Get the object's class info ...
	CObjectInfo* info = m_Application->Libraries().ObjectInfo(classid);

	// If we don't support this object type, fail ...
	g_return_val_if_fail(info, SDP_E_UNKNOWN_OBJECT);

	// Create an instance of the object ...
	k3dIObject* object = 0;
	g_return_val_if_fail(sdpSucceeded(info->ClassProcedure()(classid, k3dIObject::iid, (void**)&object)), SDP_E_FAIL);
	g_return_val_if_fail(object, SDP_E_FAIL);

	// Attach an object context to the object ...
	g_return_val_if_fail(sdpSucceeded(AttachObjectContext(info, object, Element)), SDP_E_FAIL);

	// Maintain object reference count ...
	object->Release();

	return SDP_S_OK;
}

k3dICamera* CDocument::GetActiveCamera()
{
	return m_ActiveCamera;
}

sdpResult CDocument::SetActiveCamera(k3dICamera* Camera)
{
	if(m_Application->Verbose())
		std::cout << "CDocument::SetActiveCamera()" << std::endl;
		
	m_ActiveCamera = Camera;
	return SDP_S_OK;
}

K3DLAYERHANDLE CDocument::GetFirstLayer()
{
	return m_Layers.size() ? K3DLAYERHANDLE(1) : K3DLAYERHANDLE(0);
}

K3DLAYERHANDLE CDocument::GetNextLayer(K3DLAYERHANDLE Layer)
{
	// Sanity checks ...
	g_return_val_if_fail(Layer, K3DLAYERHANDLE(0));

	return guint(Layer) < m_Layers.size() ? K3DLAYERHANDLE(Layer+1) : K3DLAYERHANDLE(0);
}

k3dILayer* CDocument::GetLayer(K3DLAYERHANDLE Layer)
{
	// Sanity checks ...
	g_return_val_if_fail(Layer, 0);

	return m_Layers[guint(Layer)-1];
}

sdpResult CDocument::SetActiveLayer(k3dILayer* Layer)
{
	// Sanity checks ...
	g_return_val_if_fail(Layer, SDP_E_FAIL);

	SetActiveLayer((CLayer*)Layer);
	return SDP_S_OK;
}

k3dILayer* CDocument::GetActiveLayer()
{
	return m_ActiveLayer;
}

k3dILayer* CDocument::CreateLayer()
{
	// Call our internal layer-creation function ...
	CLayer* layer = CreateLayer(this);
	g_return_val_if_fail(layer, 0);

	return layer;
}


sdpResult CDocument::DestroyLayer(k3dILayer* Layer)
{
	// Sanity checks ...
	g_return_val_if_fail(Layer, SDP_E_FAIL);

	// Call our internal layer-deletion function ...
	DeleteLayer((CLayer*)Layer);

	return SDP_S_OK;
}


K3DOBJECTCONTEXTHANDLE CDocument::GetFirstObjectContext()
{
	return m_Hierarchy.size() ? K3DOBJECTCONTEXTHANDLE(1) : K3DOBJECTCONTEXTHANDLE(0);
}

K3DOBJECTCONTEXTHANDLE CDocument::GetNextObjectContext(K3DOBJECTCONTEXTHANDLE ObjectContext)
{
	// Sanity checks ...
	g_return_val_if_fail(ObjectContext, K3DOBJECTCONTEXTHANDLE(0));

	// Iterate through the hierarchy looking for "PUSH" records ...
	guint size = m_Hierarchy.size();
	guint i;
	for(i = guint(ObjectContext); i < size; i++)
		{
			if(m_Hierarchy[i].m_Type == k3dIDocument::PUSH)
				break;
		}

	return i < size ? K3DOBJECTCONTEXTHANDLE(i+1) : K3DOBJECTCONTEXTHANDLE(0);
}

k3dIObjectContext* CDocument::GetObjectContext(K3DOBJECTCONTEXTHANDLE ObjectContext)
{
	// Sanity checks ...
	g_return_val_if_fail(ObjectContext, 0);

	return m_Hierarchy[guint(ObjectContext)-1].m_ObjectContext;
}

k3dIObjectContext* CDocument::GetActiveObjectContext()
{
	return m_ActiveObjectContext;
}

sdpResult CDocument::SetActiveObjectContext(k3dIObjectContext* ObjectContext)
{
	m_ActiveObjectContext = static_cast<CObjectContext*>(ObjectContext);
	return SDP_S_OK;
}

k3dIDocument::TreeNodeType CDocument::GetFirstObjectContextTree(K3DTREEHANDLE* Handle)
{
	if(0 == m_Hierarchy.size())
		return k3dIDocument::COMPLETE;

	*Handle = K3DTREEHANDLE(1);
	return m_Hierarchy[0].m_Type;
}

k3dIDocument::TreeNodeType CDocument::GetNextObjectContextTree(K3DTREEHANDLE* Handle)
{
	// Sanity checks ...
	g_return_val_if_fail(Handle, TREE_NODE_ERROR);

	if(guint(*Handle) >= m_Hierarchy.size())
		return k3dIDocument::COMPLETE;

	*Handle = K3DTREEHANDLE(guint(*Handle)+1);
	return m_Hierarchy[guint(*Handle)-1].m_Type;
}

k3dIDocument::TreeNodeType CDocument::GetContextHandleTree(k3dIObjectContext* Context, K3DTREEHANDLE* Handle)
{
	// Sanity checks ...
	g_return_val_if_fail(Context, TREE_NODE_ERROR);

	// Find the PUSH record for this object ...
	CObjectContext* context = static_cast<CObjectContext*>(Context);
	HierarchyTreeIterator push = std::find(m_Hierarchy.begin(), m_Hierarchy.end(), CHierarchyTreeRecord(k3dIDocument::PUSH, context));
	g_return_val_if_fail(push != m_Hierarchy.end(), TREE_NODE_ERROR);
	
	// Return the data ...
	*Handle = K3DTREEHANDLE((push - m_Hierarchy.begin()) + 1);
	return push->m_Type;
}

k3dIObjectContext* CDocument::GetObjectContextTree(K3DTREEHANDLE Handle)
{
	// Sanity checks ...
	g_return_val_if_fail(Handle, 0);

	return m_Hierarchy[guint(Handle)-1].m_ObjectContext;
}

k3dIDocument::Axis CDocument::GetActiveAxis()
{
	return m_ActiveAxis;
}

sdpResult CDocument::SetActiveAxis(Axis ActiveAxis)
{
	// Store the new axis ...
	m_ActiveAxis = ActiveAxis;

	// Encourage the UI to update itself ...
	m_ActiveAxisRevision++;

	return SDP_S_OK;
}

k3dIDocument::TransformType CDocument::GetActiveTransformType()
{
	return m_ActiveTransformType;
}

sdpResult CDocument::SetActiveTransformType(TransformType ActiveTransformType)
{
	// Store the new transform type ...
	m_ActiveTransformType = ActiveTransformType;

	// Encourage the UI to update itself ...
	m_ActiveTransformationRevision++;

	return SDP_S_OK;
}

k3dIDocument::CoordinateSystem CDocument::GetActiveCoordinateSystem()
{
	return m_ActiveCoordinateSystem;
}

sdpResult CDocument::SetActiveCoordinateSystem(CoordinateSystem System)
{
	m_ActiveCoordinateSystem = System;
	return SDP_S_OK;
}

k3dIDocument::EditMode CDocument::GetActiveEditMode()
{
	return m_ActiveEditMode;
}

sdpResult CDocument::SetActiveEditMode(EditMode Mode)
{
	// Store the new mode ...
	m_ActiveEditMode = Mode;

	// Encourage the UI to update itself ...
	m_EditModeRevision++;

	return SDP_S_OK;
}


k3dIMaterial* CDocument::GetDefaultMaterial()
{
	return m_DefaultMaterial;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// k3dITimeBase overrides

gdouble CDocument::GetStartTime()
{
	return m_StartTime;
}

gulong CDocument::GetStartFrame()
{
	return gulong(floor(m_StartTime * m_Duration * m_FrameRate + 1.5));
}

gdouble CDocument::GetViewTime()
{
	return m_ViewTime;
}

gulong CDocument::GetViewFrame()
{
	return gulong(floor(m_ViewTime * m_Duration * m_FrameRate + 1.5));
}

gdouble CDocument::GetEndTime()
{
	return m_EndTime;
}

gulong CDocument::GetEndFrame()
{
	return gulong(floor(m_EndTime * m_Duration * m_FrameRate + 1.5));
}

gdouble CDocument::GetDuration()
{
	return m_Duration;
}

gulong CDocument::GetTotalFrames()
{
	return gulong(m_Duration * m_FrameRate);
}

gdouble CDocument::GetFrameRate()
{
	return m_FrameRate;
}

gdouble CDocument::GetFrameDelta()
{
	return 1.0 / (m_Duration * m_FrameRate);
}

gdouble CDocument::GetRenderTime()
{
	return m_RenderTime;
}

gdouble CDocument::GetRenderDeltaTime()
{
	return m_RenderDeltaTime;
}

sdpResult CDocument::SetStartTime(gdouble Time)
{
	if(m_StartTime != Time)
		{
			m_StartTime = Time;
			
			// Reset the view time to make sure it stays within limits
			SetViewTime(m_ViewTime);

			// Encourage the UI to update itself ...
			m_TimebaseRevision++;
		}

	return SDP_S_OK;
}

sdpResult CDocument::SetStartFrame(gulong Frame)
{
	return SetStartTime((Frame - 1) / (m_FrameRate * m_Duration));
}

sdpResult CDocument::SetViewTime(gdouble Time)
{
	if(m_ViewTime != Time)
		{
			m_ViewTime = Time;
			
			// Clamp starttime <= viewtime <= endtime
			m_ViewTime = sdpMax(m_StartTime, m_ViewTime);
			m_ViewTime = sdpMin(m_ViewTime, m_EndTime);

			// Encourage the UI to update itself ...
			m_TimebaseRevision++;
		}

	return SDP_S_OK;
}

sdpResult CDocument::SetViewFrame(gulong Frame)
{
	return SetViewTime((Frame - 1) / (m_FrameRate * m_Duration));
}

sdpResult CDocument::SetEndTime(gdouble Time)
{
	if(m_EndTime != Time)
		{
			m_EndTime = Time;

			// Reset the view time to make sure it stays within limits
			SetViewTime(m_ViewTime);

			// Encourage the UI to update itself ...
			m_TimebaseRevision++;
		}

	return SDP_S_OK;
}

sdpResult CDocument::SetEndFrame(gulong Frame)
{
	return SetEndTime((Frame - 1) / (m_FrameRate * m_Duration));
}

sdpResult CDocument::SetDuration(gdouble Duration)
{
	if(m_Duration != Duration)
		{
			m_Duration = Duration;

			// Encourage the UI to update itself ...
			m_TimebaseRevision++;
		}

	return SDP_S_OK;
}

sdpResult CDocument::SetTotalFrames(gulong Frames)
{
	// Sanity checks ...
	g_return_val_if_fail(Frames, SDP_E_FAIL);
	
	// Set the new animation duration in seconds ...
	SetDuration(Frames / GetFrameRate());

	return SDP_S_OK;
}

sdpResult CDocument::SetFrameRate(gdouble FrameRate)
{
	// Sanity checks ...
	g_return_val_if_fail(0.0 != FrameRate, SDP_E_FAIL);

	if(m_FrameRate != FrameRate)
		{
			m_FrameRate = FrameRate;

			// Encourage the UI to update itself ...
			m_TimebaseRevision++;
		}

	return SDP_S_OK;
}

sdpResult CDocument::SetRenderTime(gdouble Time)
{
	m_RenderTime = Time;

	return SDP_S_OK;
}

sdpResult CDocument::SetRenderDeltaTime(gdouble DeltaTime)
{
	m_RenderDeltaTime = DeltaTime;

	return SDP_S_OK;
}

CDocument::PlaybackMode CDocument::GetPlaybackMode()
{
	return m_PlaybackMode;
}

void CDocument::SetPlaybackMode(PlaybackMode Mode)
{
	if(m_PlaybackMode != Mode)
		{
			// Store the new mode ...
			m_PlaybackMode = Mode;

			// Encourage the UI to update itself ...
			m_PlaybackModeRevision++;
		}
}

