
// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CDocumentFrame class, which encapsulates an open K-3D document's UI
		\author Tim Shead (tshead@k-3d.com)
*/

#include "application.h"
#include "document.h"
#include "objectcontext.h"
#include "documentframe.h"
#include "layer.h"

#include <k3dmisc.h>
#include <k3dfilefilters.h>

#include <k3difileformat.h>
#include <k3digeometryreadformat.h>
#include <k3digeometrywriteformat.h>

#include <sdpgtkfileselector.h>
#include <sdpgtkevents.h>
#include <sdpxmlutility.h>
#include <sdpfile.h>
#include <algorithm>

// We have an unfortunate clash with X ...
#ifdef RootWindow
#undef RootWindow
#endif // RootWindow

//////////////////////////////////////////////////////////////////////
// CDocumentFrame

CDocumentFrame::CDocumentFrame()
{
	m_Document = 0;
	
	m_LastActionRevision = 0;
	m_LastLayerRevision = 0;
	m_LastEditModeRevision = 0;
	m_LastActiveAxisRevision = 0;
	m_LastPlaybackModeRevision = 0;
	m_LastTimebaseRevision = 0;
	m_LastAnimationFrame = ULONG_MAX;
}


CDocumentFrame::~CDocumentFrame()
{
	if(Application.Verbose())
		std::cout << "CDocumentFrame::~CDocumentFrame()" << std::endl;
}


// Called by CDocument to create the frame ...
bool CDocumentFrame::Create(CDocument* Document)
{
	// Sanity checks ...
	g_assert(Document);
	m_Document = Document;

	// Load the frame ...
	g_return_val_if_fail(LoadFile(sdpAppendPath(Application.DialogPath(), "documentframe.gtkml")), false);

	// Create the hierarchy control ...
	g_return_val_if_fail(m_HierarchyControl.Create(m_Document, Widget("hierarchy")), false);

	// Create an event for our animation scrollbar ...
	sdpGtkAdjustment adjustment = Scrollbar("animationscrollbar").Adjustment();
	MapEvent("value-changed", "onanimationscrollbar", adjustment, true);

	// Update our titlebar ...
	UpdateTitlebar();

	// Make it all visible ...
	RootWidget().Show();

	return true;
}

// Called by CDocument to open the frame ...
bool CDocumentFrame::Load(CDocument* Document, sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_assert(Document);
	g_assert(Element);
	g_assert(Element->Name() == sdpString("documentframe"));

	return Create(Document);
}

// Persistence ...
bool CDocumentFrame::Save(sdpXmlIElement* Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert(Dependencies);

	sdpXmlCreateElement(Element, "documentframe");

	return true;
}

// Called by the framework to handle widget events ...
void CDocumentFrame::OnEvent(sdpGtkEvent* Event)
{
	if(Event->Name() == "ondelete")
		OnDelete(Event);
	else if(Event->Name() == "ondestroy")
		OnDestroy();
	else if(Event->Name() == "onfilenewhierarchy")
		OnFileNewHierarchy();
	else if(Event->Name() == "onfilenewcamera")
		OnFileNewCamera();
	else if(Event->Name() == "onfileimport")
		OnFileImport();
	else if(Event->Name() == "onfileexport")
		OnFileExport();
	else if(Event->Name() == "onfilesave")
		OnFileSave();
	else if(Event->Name() == "onfilesaveas")
		OnFileSaveAs();
	else if(Event->Name() == "onfileclose")
		OnFileClose();
	else if(Event->Name() == "oneditundo")
		OnEditUndo();
	else if(Event->Name() == "oneditredo")
		OnEditRedo();
	else if(Event->Name() == "oneditdelete")
		OnEditDelete();
	else if(Event->Name() == "oneditproperties")
		OnEditProperties();
	else if(Event->Name() == "oneditrename")
		OnEditRename();
	else if(Event->Name() == "oneditmode")
		OnEditMode();
	else if(Event->Name() == "onxaxis")
		OnXAxis();
	else if(Event->Name() == "onyaxis")
		OnYAxis();
	else if(Event->Name() == "onzaxis")
		OnZAxis();
	else if(Event->Name() == "onposition")
		OnPosition();
	else if(Event->Name() == "onorientation")
		OnOrientation();
	else if(Event->Name() == "onscale")
		OnScale();
	else if(Event->Name() == "oneditsmoothshade")
		OnEditSmoothShade();
	else if(Event->Name() == "oneditunshade")
		OnEditUnshade();
	else if(Event->Name() == "onanimationparameters")
		OnAnimationParameters();
	else if(Event->Name() == "onlayeradd")
		OnLayerAdd();
	else if(Event->Name() == "onlayerdelete")
		OnLayerDelete();
	else if(Event->Name() == "onlayerrename")
		OnLayerRename();
	else if(Event->Name() == "onsetfocus")
		OnSetFocus();
	else if(Event->Name() == "onrewind")
		OnRewind();
	else if(Event->Name() == "onreverseplay")
		OnReversePlay();
	else if(Event->Name() == "onstop")
		OnStop();
	else if(Event->Name() == "onplay")
		OnPlay();
	else if(Event->Name() == "onloopplay")
		OnLoopPlay();
	else if(Event->Name() == "onfastforward")
		OnFastForward();
	else if(Event->Name() == "onrecordframe")
		OnRecordFrame();
	else if(Event->Name() == "onrecord")
		OnRecord();
	else if(Event->Name() == "onanimationscrollbar")
		OnAnimationScrollbar();
	else if(Event->Name() == "onanimationframe")
		OnAnimationFrame();
	else if(Event->Name() == "onidle")
		OnIdle();
	else
		sdpGtkObjectContainer::OnEvent(Event);
}

// The user wants to close the window ...
void CDocumentFrame::OnDelete(sdpGtkEvent* Event)
{
	if(Application.Verbose())
		std::cout << "CDocumentFrame::OnDelete()" << std::endl;

	// Ask to see if it's OK to close ...
	((sdpGtkEventWidgetDeleteEvent*)Event)->SetResult(!SafeToClose());
}

// The window is about to be destroyed ...
void CDocumentFrame::OnDestroy()
{
	if(Application.Verbose())
		std::cout << "CDocumentFrame::OnDestroy()" << std::endl;

	// Block events from this point forward ...
	DisconnectAllEvents();

	// Don't need to keep track of widgets anymore ...
	Clear();

	// Zap the control, so it has a chance to clean-up BEFORE we delete ourselves...
//	m_HierarchyControl.DisconnectAllEvents();
	m_HierarchyControl.Destroy();

	// Remove ourselves from the document ...
	m_Document->DeleteFrame(this);
}

// Creates a new hierarchy view ...
void CDocumentFrame::OnFileNewHierarchy()
{
	m_Document->CreateHierarchyFrame();
}

// Creates a new camera view ...
void CDocumentFrame::OnFileNewCamera()
{
	// Create a camera and move it away from the origin ...
	k3dIObject* object = m_Document->CreateObject(sdpCamera, false, true, true);
	g_return_if_fail(object);

	k3dICamera* camera = sdpQueryInterface<k3dICamera>(object);
	g_return_if_fail(camera);
	object->Release();

	k3dSetPosition(camera, sdpVector3(-20.0, 20.0, -20.0));
	k3dSetOrientation(camera, sdpEulerAngles(sdpRadians(0.0), sdpRadians(37.0), sdpRadians(45.0), sdpEulerAngles::ZXYstatic));

	camera->ShowViewport();
}

// Called to import geometry into the document ...
void CDocumentFrame::OnFileImport()
{
	// Get the list of available import formats ...
	sdpStringMultiSet formats = k3dFileFormats(Document(), k3dIGeometryReadFormat::iid);
	if(0 == formats.size())
		{
			Document()->Application()->Error("No geometry import file filters available", "Import Geometry:");
			return;
		}

	// Prompt the user for a file to import ...
	sdpGtkFileSelector selector("Import geometry:", "");
	selector.DoModal(RootWindow());
	if(selector.OK())
		{
			m_Document->Import(selector.FilePath(), k3dIFileFormat::Automatic, 0);
			k3dRedrawAll(m_Document, false);
		}
}

// Called to export geometry from the document ...
void CDocumentFrame::OnFileExport()
{
	// Get the list of available export formats ...
	sdpStringMultiSet formats = k3dFileFormats(Document(), k3dIGeometryWriteFormat::iid);
	if(0 == formats.size())
		{
			Document()->Application()->Error("No geometry export file filters available", "Export Geometry:");
			return;
		}

	// Prompt the user for a file to export ...
	sdpGtkFileSelector selector("Export geometry:", "");
	selector.DoModal(RootWindow());
	if(selector.OK())
		{
			m_Document->Export(selector.FilePath(), k3dIFileFormat::Automatic, 0);
			k3dRedrawAll(m_Document, false);
		}
}

// Save the document ...
void CDocumentFrame::OnFileSave()
{
	// Try to save the doc ...
	sdpResult result = m_Document->Save();

	// If the user hasn't chosen a name, yet, turn it into a "Save As" ...
	if(SDP_S_FALSE == result)
		OnFileSaveAs();
}

// Save the document to a new filename ...
void CDocumentFrame::OnFileSaveAs()
{
	// Get our starting path - start with the user's home directory ...
	sdpString startpath = sdpGetHomeDirectory();
	
	// Next, override it with the user's options ...
	sdpXmlIElement* startpathelement = sdpXmlGetXPathElement(Document()->Application()->Options()->GetRootElement(), "application/paths/path[@type=\"document\"]");
	if(startpathelement)
		startpath = startpathelement->Text();
	
	// Append a directory separator (GtkFileSelector seems to want it) ...
	startpath += G_DIR_SEPARATOR;

	sdpGtkFileSelector selector("Save K-3D document as:", startpath);
	selector.DoModal(RootWindow());

	// If the user cancelled, we're done ...
	if(!selector.OK())
		return;

	if(startpathelement)
		startpathelement->SetText(sdpDirectory(selector.FilePath()));

	m_Document->SaveAs(selector.FilePath());
}

// Close the document ...
void CDocumentFrame::OnFileClose()
{
	if(Application.Verbose())
		std::cout << "CDocumentFrame::OnFileClose()" << std::endl;

	// Tell the document to close ...
	if(SafeToClose())
		RootWidget().Destroy();
//		m_Document->Application()->CloseDocument(m_Document);
}

// User wants to undo an action ...
void CDocumentFrame::OnEditUndo()
{
	// Ignore it if there aren't any actions to undo ...
	if(!m_Document->ActionCount())
		return;

	// Tell the document to undo the last action ...
	m_Document->UndoAction();
	k3dRedrawAll(m_Document, false);
}

// User wants to redo an action ...
void CDocumentFrame::OnEditRedo()
{
	// Ignore it if there aren't any actions to redo ...
	if(!m_Document->RedoCount())
		return;

	// Tell the document to redo the last action ...
	m_Document->RedoAction();
	k3dRedrawAll(m_Document, false);

}

// User wants to delete something ...
void CDocumentFrame::OnEditDelete()
{
	// Record this for action for Undo-ing ...
	m_Document->Actions()->StartAction("Delete");

	// Get the set of currently selected objects ...
	bool locked = false;
	k3dObjectSet objects;
	k3dGetObjects(m_Document, k3dIState::VISIBLESELECTED, objects);
	
	// For each object ...
	for(k3dObjectSet::iterator object = objects.begin(); object != objects.end(); object++)
		{
			// Make sure it's deletable ...
			if(SDP_S_OK == (*object)->QueryDeleteObject())
				(*object)->ObjectContext()->Layer()->Document()->DeleteObject(*object);	
			else
				locked = true;
		}

	// Get the (leftover) set of selected polygons ...
	k3dPolygonArray polygons;
	k3dGetPolygons(m_Document, k3dIState::VISIBLESELECTED, polygons);

	// Delete polygons (including any points owned "exclusively" by that polygon)
	for(gulong i = 0; i < polygons.size(); i++)
		polygons[i]->Mesh()->DeletePolygon(polygons[i], true);

	// If we're in point-edit mode ...
	if(k3dIDocument::POINT == m_Document->GetActiveEditMode())
		{
			// Get the (leftover) set of selected points ...
			k3dPointArray points;
			k3dGetPoints(m_Document, k3dIState::VISIBLESELECTED, points);
	
			// Delete *them* ...
			for(gulong j = 0; j < points.size(); j++)
				points[j]->Mesh()->DeletePoint(points[j]);
		}

	// Finish recording the action ...
	m_Document->Actions()->FinishAction();
		
	// Update the display ...
	k3dRedrawAll(m_Document, false);
	
	// Let the user know if there was anything that couldn't be deleted ...
	if(locked)
		m_Document->Application()->Message("Some objects couldn't be deleted!", "Delete:");
}

// User wants to edit an object's properties ...
void CDocumentFrame::OnEditProperties()
{
	// Get the set of currently selected objects ...
	HierarchyItems items = m_HierarchyControl.HighlightedItems();
	for(HierarchyItemIterator item = items.begin(); item != items.end(); item++)
		item->EditObject();
}

// User wants to rename an object ...
void CDocumentFrame::OnEditRename()
{
	// Record this for action for Undo-ing ...
	m_Document->Actions()->StartAction("Rename");

	// Get the set of currently selected objects ...
	HierarchyItems items = m_HierarchyControl.HighlightedItems();
	for(HierarchyItemIterator item = items.begin(); item != items.end(); item++)
		item->RenameObject();

	// Finish recording the action ...
	m_Document->Actions()->FinishAction();
}

// User picked a new edit mode ...
void CDocumentFrame::OnEditMode()
{
	// Get the menu item ...
	sdpGtkCheckMenuItem menuitem = CheckMenuItem("editmodemenuitem");

	// Set the new edit mode ...
	if(menuitem.GetState())
		m_Document->SetActiveEditMode(k3dIDocument::POINT);
	else
		m_Document->SetActiveEditMode(k3dIDocument::POLYGON);

	// Update the display ...
	k3dRedrawAll(m_Document, false);
}

// User picked a new active axis ...
void CDocumentFrame::OnXAxis()
{
	// Set the new active axis ...
	m_Document->SetActiveAxis(k3dIDocument::XAXIS);
}

// User picked a new active axis ...
void CDocumentFrame::OnYAxis()
{
	// Set the new active axis ...
	m_Document->SetActiveAxis(k3dIDocument::YAXIS);
}

// User picked a new active axis ...
void CDocumentFrame::OnZAxis()
{
	// Set the new active axis ...
	m_Document->SetActiveAxis(k3dIDocument::ZAXIS);
}

void CDocumentFrame::OnPosition()
{
	// Set the new active transformation ...
	m_Document->SetActiveTransformType(k3dIDocument::POSITION);
}

void CDocumentFrame::OnOrientation()
{
	// Set the new active transformation ...
	m_Document->SetActiveTransformType(k3dIDocument::ORIENTATION);
}

void CDocumentFrame::OnScale()
{
	// Set the new active transformation ...
	m_Document->SetActiveTransformType(k3dIDocument::SCALE);
}

class CDocumentFrameSmoothShade
{
public:
	CDocumentFrameSmoothShade(gulong State) { m_State = State; }

	bool HandleMesh(k3dIMesh* Mesh)
	{
		Mesh->ShadePolygons(m_State, 0.0);
		return true;
	}

	gulong m_State;
};

// Called when the user wants to smooth-shade polygons
void CDocumentFrame::OnEditSmoothShade()
{
	CDocumentFrameSmoothShade functor(k3dIState::VISIBLESELECTED);
	k3dForEachMesh(m_Document, k3dIState::VISIBLESELECTED, functor);
	k3dRedrawAll(m_Document, false);
}

class CDocumentFrameUnshade
{
public:
	CDocumentFrameUnshade(gulong State) { m_State = State; }

	bool HandleMesh(k3dIMesh* Mesh)
	{
		Mesh->UnshadePolygons(m_State);
		return true;
	}

	gulong m_State;
};

// Called when the user wants to unshade smooth-shade polygons
void CDocumentFrame::OnEditUnshade()
{
	CDocumentFrameUnshade functor(k3dIState::VISIBLESELECTED);
	k3dForEachMesh(m_Document, k3dIState::VISIBLESELECTED, functor);
	k3dRedrawAll(m_Document, false);
}

// User wants to alter animation parameters
void CDocumentFrame::OnAnimationParameters()
{
	m_AnimationParametersDialog.Create(m_Document);
}

// User wants to add a new layer ...
void CDocumentFrame::OnLayerAdd()
{	
	// Update ourselves and the display ...
	m_Document->CreateLayer(m_Document);

}

// User wants to delete a layer ...
void CDocumentFrame::OnLayerDelete()
{
}

// User wants to rename a layer ...
void CDocumentFrame::OnLayerRename()
{
}

// The window just received the input focus ...
void CDocumentFrame::OnSetFocus()
{
	// If the document is in the process of being closed, do nothing ...
	if(m_Document->IsDying())
		return;

	// Activate the document ...
	m_Document->OnActivate(this);
	
}

// Called when the user hits the "rewind" button
void CDocumentFrame::OnRewind()
{
	m_Document->SetViewFrame(m_Document->GetStartFrame());
	k3dRedrawAll(m_Document, false);

}

// Called when the user hits the "reverse play" button
void CDocumentFrame::OnReversePlay()
{
	m_Document->SetPlaybackMode(CDocument::REVERSEPLAY);
}

// Called when the user hits the "stop" button
void CDocumentFrame::OnStop()
{
	m_Document->SetPlaybackMode(CDocument::STOP);
}

// Called when the user hits the "play" button
void CDocumentFrame::OnPlay()
{
	m_Document->SetPlaybackMode(CDocument::PLAY);
}

// Called when the user hits the "continuous play" button
void CDocumentFrame::OnLoopPlay()
{
	m_Document->SetPlaybackMode(CDocument::LOOPPLAY);
}

// Called when the user hits the "fast forward" button
void CDocumentFrame::OnFastForward()
{
	m_Document->SetViewFrame(m_Document->GetEndFrame());
	k3dRedrawAll(m_Document, false);
}

// Called when the user hits the "record frame" button
void CDocumentFrame::OnRecordFrame()
{
	// Get the active view ...
	k3dICamera* activecamera = m_Document->GetActiveCamera();
	if(!activecamera)
		{
			Application.Error("You must have an open view to render!", "Record Frame:");
			return;
		}

	// Prompt the user for a file name ...
	sdpGtkFileSelector selector("Record Frame:", "");
	selector.DoModal(RootWindow());
	if(!selector.OK())
		return;

	// Start a new render job ...
	k3dIRenderJob* job = Application.CreateJob("k3d-render-frame");
	g_assert(job);

	// Add a single render frame to the job ...
	k3dIRenderFrame* renderframe = job->CreateFrame(sdpFormatString("frame%.06u", m_Document->GetViewFrame()));
	g_assert(renderframe);	
	
	// We want to see the results when they're complete ...
	renderframe->SetViewDuringRender(false);
	renderframe->SetViewAfterRender(true);	

	// Setup the view for rendering ...
	activecamera->StartJob(job);
	activecamera->QueueFrame(renderframe);
	activecamera->FinishJob(job);

	// Start the job running ...
	Application.StartJob(job);
}

// Called when the user hits the "record" button
void CDocumentFrame::OnRecord()
{
	// Get the active view ...
	k3dICamera* activecamera = m_Document->GetActiveCamera();
	if(!activecamera)
		{
			Application.Error("You must have an open camera to render!", "Record Animation:");
			return;
		}

	// Prompt the user for a file name ...
	sdpGtkFileSelector selector("Choose Base FilePath:", "");
	selector.DoModal(RootWindow());
	if(!selector.OK())
		return;

	// Start a new render job ...
	k3dIRenderJob* job = Application.CreateJob("k3d-render");
	g_assert(job);

	// For each frame to be rendered ...
	for(gulong frame = m_Document->GetStartFrame(); frame < m_Document->GetEndFrame(); frame++)
		{
			// Set the frame time ...
			m_Document->SetViewFrame(frame);

			// Redraw everything ...
			k3dRedrawAll(m_Document, true);

			// Add a render frame to the job ...
			k3dIRenderFrame* renderframe = job->CreateFrame(sdpFormatString("frame%.06u", frame));
			g_assert(renderframe);		

			renderframe->SetViewDuringRender(false);
			renderframe->SetViewAfterRender(false);

			// Setup the view for rendering ...
			activecamera->StartJob(job);
			activecamera->QueueFrame(renderframe);
			activecamera->FinishJob(job);
		}

	// Start the job running ...
	Application.StartJob(job);
}

// Called when the user moves the animation scrollbar ...
void CDocumentFrame::OnAnimationScrollbar()
{
	// Update our scrollbar ...
	sdpGtkAdjustment adjustment = Scrollbar("animationscrollbar").Adjustment();
	m_Document->SetViewFrame(adjustment.Value());
	k3dRedrawAll(m_Document, false);

}

// Called when the user selects a new animation frame ...
void CDocumentFrame::OnAnimationFrame()
{
	// Set the new frame ...
	m_Document->SetViewFrame(atoi(Entry("animationframe").GetText()));
	k3dRedrawAll(m_Document, false);

}

// Nothing much going on, so let's update our UI ...
void CDocumentFrame::OnIdle()
{
	UpdateLayers();
	UpdateActions();
	UpdateEditMode();
	UpdateActiveAxis();
	UpdateActiveTransformation();
	UpdatePlaybackMode();
	UpdateTimebase();
}

// Update our window titlebar ...
void CDocumentFrame::UpdateTitlebar()
{
	// Get our title and stuff it in the titlebar ...
	Window("documentframe").SetTitle(Title());
}

// Return the document title for our window titlebar ...
sdpString CDocumentFrame::Title()
{
	// Create the title string ...
	sdpString title = m_Document->Title();

	return title;
}

class CDocumentFrameAddLayer
{
public:
	CDocumentFrameAddLayer(sdpGtkCList& List) : m_List(List) {};

	void operator()(CLayer* Layer)
	{
		const char* data[1];
		data[0] = Layer->Name();
		m_List.Append(data);
	}

	sdpGtkCList& m_List;
};


// Called to update the list of layers ...
void CDocumentFrame::UpdateLayers()
{
	// If nothing's changed, we're done ...
	if(m_LastLayerRevision == m_Document->m_LayerRevision)
		return;
/*
	// Get our list widget ...
	sdpGtkCList list = CList("layers");

	// Clear out the old stuff ...
	list.Clear();

	// Add list entries ...
	std::for_each(m_Document->m_Layers.begin(), m_Document->m_Layers.end(), FunctorAddLayer(list));
*/

	// Record the last revision to prevent unnecessary updates ...
	m_LastLayerRevision = m_Document->m_LayerRevision;
}


// Called to update the "Undo" and "Redo" menu items ...
void CDocumentFrame::UpdateActions()
{
	// If nothing's changed, we're done ...
	if(m_LastActionRevision == m_Document->m_ActionRevision)
		return;

	// Setup the undo controls ...
	sdpGtkMenuItem undoitem = MenuItem("undomenuitem");
	GList* undolist = gtk_container_children(undoitem);
	sdpGtkLabel undolabel((GtkLabel*)undolist->data);

	if(m_Document->ActionCount())
		{
			undolabel.SetText(sdpString("Undo ") + m_Document->m_ActionStack.back().Name());
			undolabel.SetSensitive(true);
		}
	else
		{
			undolabel.SetText("Can't Undo");
			undolabel.SetSensitive(false);
		}

	// Setup the redo controls ...
	sdpGtkMenuItem redoitem = MenuItem("redomenuitem");
	GList* redolist = gtk_container_children(redoitem);
	sdpGtkLabel redolabel((GtkLabel*)redolist->data);

	if(m_Document->RedoCount())
		{
			redolabel.SetText(sdpString("Redo ") + m_Document->m_RedoStack.back().Name());
			redolabel.SetSensitive(true);
		}
	else
		{
			redolabel.SetText("Can't Redo");
			redolabel.SetSensitive(false);
		}
		
	// Setup the undo list ...
	sdpGtkCList list = CList("undolist");

	// Freeze updates ...
	list.Freeze();

	// Clear out the old stuff ...
	list.Clear();

	// Add action entries ...
	for(ActionIterator i = m_Document->m_ActionStack.begin(); i != m_Document->m_ActionStack.end(); i++)
		{
			const char* text[1];
			text[0] = i->Name();
			list.Append(text);
//			GtkCTreeNode* node = tree.InsertNode(parentstack.top(), 0, text);

//			gtk_ctree_node_set_row_data(tree, node, gpointer(context));
		}
		
	list.Thaw();
		
	// Record the last revision to prevent unnecessary updates ...
	m_LastActionRevision = m_Document->m_ActionRevision;
}


// Called to update the Edit/Point Edit Mode menu item ...
void CDocumentFrame::UpdateEditMode()
{
	// If nothing's changed, we're done ...
	if(m_LastEditModeRevision == m_Document->m_EditModeRevision)
		return;

	// Get the menu item ...
	sdpGtkCheckMenuItem menuitem = CheckMenuItem("editmodemenuitem");

	if(k3dIDocument::POINT == m_Document->GetActiveEditMode())
		menuitem.SetState(true);
	else if(k3dIDocument::POLYGON == m_Document->GetActiveEditMode())
		menuitem.SetState(false);
	else
		g_assert_not_reached();	// Did you add a new edit mode without updating this code?
		
	// Record the last revision to prevent unnecessary updates ...
	m_LastEditModeRevision = m_Document->m_EditModeRevision;
}


// Called to update the Edit/Axis menu items ...
void CDocumentFrame::UpdateActiveAxis()
{
	// If nothing's changed, we're done ...
	if(m_LastActiveAxisRevision == m_Document->m_ActiveAxisRevision)
		return;

	BlockAllEvents();

	CheckMenuItem("menuxaxis").SetState(k3dIDocument::XAXIS == m_Document->GetActiveAxis());
	CheckMenuItem("menuyaxis").SetState(k3dIDocument::YAXIS == m_Document->GetActiveAxis());
	CheckMenuItem("menuzaxis").SetState(k3dIDocument::ZAXIS == m_Document->GetActiveAxis());

	ToggleButton("toolbarxaxis").SetState(k3dIDocument::XAXIS == m_Document->GetActiveAxis());
	ToggleButton("toolbaryaxis").SetState(k3dIDocument::YAXIS == m_Document->GetActiveAxis());
	ToggleButton("toolbarzaxis").SetState(k3dIDocument::ZAXIS == m_Document->GetActiveAxis());

	UnblockAllEvents();

	// Record the last revision to prevent unnecessary updates ...
	m_LastActiveAxisRevision = m_Document->m_ActiveAxisRevision;
}

// Called to update the Edit/Axis menu items ...
void CDocumentFrame::UpdateActiveTransformation()
{
	// If nothing's changed, we're done ...
	if(m_LastActiveTransformationRevision == m_Document->m_ActiveTransformationRevision)
		return;
	
	BlockAllEvents();
	
	CheckMenuItem("menuposition").SetState(k3dIDocument::POSITION == m_Document->GetActiveTransformType());
	CheckMenuItem("menuorientation").SetState(k3dIDocument::ORIENTATION == m_Document->GetActiveTransformType());
	CheckMenuItem("menuscale").SetState(k3dIDocument::SCALE == m_Document->GetActiveTransformType());

	ToggleButton("toolbarposition").SetState(k3dIDocument::POSITION == m_Document->GetActiveTransformType());
	ToggleButton("toolbarorientation").SetState(k3dIDocument::ORIENTATION == m_Document->GetActiveTransformType());
	ToggleButton("toolbarscale").SetState(k3dIDocument::SCALE == m_Document->GetActiveTransformType());

	UnblockAllEvents();

	// Record the last revision to prevent unnecessary updates ...
	m_LastActiveTransformationRevision = m_Document->m_ActiveTransformationRevision;
}

// Called to update the animation playback controls ...
void CDocumentFrame::UpdatePlaybackMode()
{
	// If nothing's changed, we're done ...
	if(m_LastPlaybackModeRevision == m_Document->m_PlaybackModeRevision)
		return;

	// Record the last revision to prevent unnecessary updates ...
	m_LastPlaybackModeRevision = m_Document->m_PlaybackModeRevision;

/*
	// Set all our buttons to "un-toggled" ...
	ToggleButton("rewind").SetState(false);
	ToggleButton("reverseplay").SetState(false);
	ToggleButton("stop").SetState(false);
	ToggleButton("play").SetState(false);
	ToggleButton("fastforward").SetState(false);
	ToggleButton("recordframe").SetState(false);
	ToggleButton("record").SetState(false);

	// Toggle the new button ...
	switch(m_Document->m_PlaybackMode)
		{
			case CDocument::REVERSEPLAY:
				ToggleButton("reverseplay").SetState(true);
				break;
			case CDocument::STOP:
				ToggleButton("stop").SetState(true);
				break;
			case CDocument::PLAY:
				ToggleButton("play").SetState(true);
				break;
			case CDocument::RECORDFRAME:
				ToggleButton("recordframe").SetState(true);
				break;
			case CDocument::RECORD:
				ToggleButton("record").SetState(true);
				break;
			default:
				g_assert_not_reached();	// Did you add a new playback state without telling me?
		}
*/
}

// Called to update the animation playback controls ...
void CDocumentFrame::UpdateTimebase()
{
	// If nothing's changed, we're done ...
	if(m_LastTimebaseRevision == m_Document->m_TimebaseRevision)
		return;

	// Record the last revision to prevent unnecessary updates ...
	m_LastTimebaseRevision = m_Document->m_TimebaseRevision;

	// Display the frame number ...
	Entry("animationframe").SetText(sdpFormatString("%u", m_Document->GetViewFrame()));

	// Update our scrollbar ...
	gulong pagesize = sdpMin<gulong>(m_Document->GetTotalFrames() * 0.05, 1);

	sdpGtkAdjustment adjustment = Scrollbar("animationscrollbar").Adjustment();
	adjustment.SetLower(m_Document->GetStartFrame());
	adjustment.SetUpper(m_Document->GetEndFrame() + pagesize);
	adjustment.SetValue(m_Document->GetViewFrame());
	adjustment.SetStepIncrement(1);
	adjustment.SetPageIncrement(pagesize);
	adjustment.SetPageSize(pagesize);
	adjustment.Changed();
	adjustment.ValueChanged();
}

// Prompts the user to see if it's safe to close the document
bool CDocumentFrame::SafeToClose()
{
	if(SDP_S_OK == m_Document->Application()->Query("Close the document?  Unsaved changes will be lost (No Undo)", "Close Document:"))
		return true;
		
	return false;
}


