// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CObjectInfo class, which encapsulates information about a class of K-3D objects
		\author Tim Shead (tshead@k-3d.com)
*/

#include <strstream>
#include <iostream>
#include <algorithm>

#include "application.h"
#include "objectinfo.h"

#include <sdpxmlutility.h>
#include <sdpxmldocument.h>

#include <sdpfile.h>

/////////////////////////////////////////////////////////////////////////////
// CObjectInfo

CObjectInfo::CObjectInfo(k3dIObjectLibrary* Library, sdpXmlIElement* Element, sdpComClassProcedureProc ClassProcedure)
{
	// Sanity checks ...
	g_assert(Library);
	g_assert(Element);
	g_assert(ClassProcedure);

	// Store object information ...
	m_Library = Library;

	sdpXmlGetAttribute(Element, "clsid", m_ClassID);
	sdpXmlGetAttribute(Element, "name", m_Name);
	m_ClassProcedure = ClassProcedure;

	// Look for interfaces supported by this object ...
	sdpXmlForEachElement(Element, "interface", *this);

	// Look for bookmarks ...
	sdpXmlForEachElement(Element, "bookmark", *this);

	if(Application.Verbose())
		std::cout << "  Loading object " << m_Name << std::endl;
}

bool CObjectInfo::HandleElement(sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_assert(Element);
	
	if(Element->Name() == sdpString("interface"))
		{
			// Get the interface ID ...
			sdpIID iid = sdpIIDNull;
			sdpXmlGetAttribute(Element, "iid", iid);
			g_return_val_if_fail(iid != sdpIIDNull, true);

			// Push it onto our list of supported interfaces ...
			m_Interfaces.push_back(iid);
		}
	else if(Element->Name() == sdpString("bookmark"))
		{
			// Get the bookmark location ...
			sdpString location;
			sdpXmlGetAttribute(Element, "location", location);
			
			// Add it to our list of bookmarks ...
			m_Bookmarks.push_back(location);
		}

	return true;
}

///////////////////////////////////////////////////////////////////////////////////
// k3dIObjectInfo overrides

k3dIObjectLibrary* CObjectInfo::Library()
{
	return m_Library;
}

const sdpCLSID& CObjectInfo::ClassID()
{
	return m_ClassID;
}

SDPCSTRING CObjectInfo::Name()
{
	return m_Name;
}

sdpResult CObjectInfo::QueryObjectInterface(const sdpIID& Interface)
{
	// If the list is empty (likely) we're done ...
	if(!m_Interfaces.size())
		return SDP_S_FALSE;

	// Iterate through available interfaces ...
	InterfaceIterator i = std::find(m_Interfaces.begin(), m_Interfaces.end(), Interface);
	if(i != m_Interfaces.end())
		return SDP_S_OK;

	// Couldn't find it ...
	return SDP_S_FALSE;
}

K3DBOOKMARKHANDLE CObjectInfo::GetFirstBookmarkHandle()
{
	return K3DBOOKMARKHANDLE(m_Bookmarks.size() ? 1 : 0);
}

K3DBOOKMARKHANDLE CObjectInfo::GetNextBookmarkHandle(K3DBOOKMARKHANDLE Handle)
{
	// Sanity checks ...
	g_return_val_if_fail(Handle, 0);
	
	return K3DBOOKMARKHANDLE(gulong(Handle) < m_Bookmarks.size() ? gulong(Handle) + 1 : 0);
}

SDPCSTRING CObjectInfo::GetBookmark(K3DBOOKMARKHANDLE Handle)
{
	// Sanity checks ...
	g_return_val_if_fail(Handle, 0);
	g_return_val_if_fail(gulong(Handle) <= m_Bookmarks.size(), 0);
	
	return m_Bookmarks[gulong(Handle) - 1];
}

//////////////////////////////////////////////////////////////////////////////////////
// CObjectLibrary

CObjectLibrary::CObjectLibrary(SDPCSTRING FilePath, SDPCSTRING FileName, GModule* Module, sdpComLibraryInfoProc LibraryInfoProc, sdpComClassProcedureProc ClassProcedureProc)
{
	// Sanity checks ...
	g_assert_string(FilePath);
	g_assert_string(FileName);
	g_assert(Module);
	g_assert(LibraryInfoProc);
	g_assert(ClassProcedureProc);

	// Store the data ...
	m_Path = sdpDirectory(FilePath);
	m_Name = FileName;
	m_Module = Module;
	m_LibraryInfoProc = LibraryInfoProc;
	m_ClassProcedureProc = ClassProcedureProc;

	// Load objects within this library ...
	LoadObjects(LibraryInfoProc, ClassProcedureProc);
}

CObjectLibrary::~CObjectLibrary()
{
	// Free memory ...
	for(ObjectInfoIterator i = m_Objects.begin(); i != m_Objects.end(); i++)
		delete(*i);

	// Unload the module ...
	g_module_close(m_Module);
}

class CObjectLibraryLoadObject
{
public:
	CObjectLibraryLoadObject(CObjectLibrary* Library, ObjectInfos& Objects, sdpComClassProcedureProc ClassProcedureProc) : m_Library(Library), m_Objects(Objects), m_ClassProcedureProc(ClassProcedureProc) {}
	bool HandleElement(sdpXmlIElement* Element)
	{
		// Sanity checks ...
		g_assert(Element->Name() == sdpString("object"));

		m_Objects.push_back(new CObjectInfo(m_Library, Element, m_ClassProcedureProc));
		return true;
	}

	CObjectLibrary* m_Library;
	ObjectInfos& m_Objects;
	sdpComClassProcedureProc m_ClassProcedureProc;
};

// Load every object within the given library ...
void CObjectLibrary::LoadObjects(sdpComLibraryInfoProc LibraryInfoProc, sdpComClassProcedureProc ClassProcedureProc)
{
	// Sanity checks ...
	g_assert(LibraryInfoProc);
	g_assert(ClassProcedureProc);

	// Load the sdpComLibraryInfo document and parse it using XML ...
	sdpString libraryinfo = LibraryInfoProc();
	sdpXmlDocument doc;
	g_return_if_fail(doc.Load(libraryinfo, ""));

	// Make sure we've got a root element ...
	sdpXmlIElement* root = doc.GetRootElement();
	g_return_if_fail(root);
	
	// Make sure we've got an sdpComLibraryInfo document ...
	g_return_if_fail(root->Name() == sdpString("sdpcomlibraryinfo"));

	// Get information about the library ...
	sdpString name;
	sdpXmlGetAttribute(root, "name", name);

	sdpString version;
	sdpXmlGetAttribute(root, "version", version);

	if(Application.Verbose())
		std::cout << "Loading library " << name << " Version " << version << std::endl;

	// Iterate through its' children and load their info ...
	CObjectLibraryLoadObject functor(this, m_Objects, ClassProcedureProc);
	sdpXmlForEachElement(root, "object", functor);

	if(Application.Verbose())
		std::cout << std::endl;
}

////////////////////////////////////////////////
// k3dIObjectLibrary overrides ...

SDPCSTRING CObjectLibrary::Name()
{
	return m_Name;
}

SDPCSTRING CObjectLibrary::Path()
{
	return m_Path;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
// CObjectLibraries

CObjectLibraries::CObjectLibraries()
{
}

CObjectLibraries::~CObjectLibraries()
{
	// Free memory ...
	for(ObjectLibraryIterator i = m_Libraries.begin(); i != m_Libraries.end(); i++)
		delete(*i);
}

// Load libraries within the given directory ...
void CObjectLibraries::LoadLibraries(SDPCSTRING Directory, bool Recursive)
{
	sdpForEachFile(Directory, *this, Recursive);
}

// For each file within the original directory ...
bool CObjectLibraries::HandleFile(SDPCSTRING FilePath, SDPCSTRING FileName)
{
	// Skip files that don't end in ".k3dlib" ...
	sdpString filename(FileName);
	if(filename.right(7) != ".k3dlib")
		return true;

	// Try to open the file as a module ...
	GModule* module = g_module_open(FilePath, GModuleFlags(0));
	if(!module)
		{
			std::cout << "Library " << FileName << ": " << g_module_error() << std::endl;
			return true;
		}

	// Look for the required sdpCom entry points ...
	sdpComLibraryInfoProc libraryinfo = 0;
	g_module_symbol(module, "sdpComLibraryInfo", (gpointer*)(&libraryinfo));

	sdpComClassProcedureProc classprocedure = 0;
	g_module_symbol(module, "sdpComClassProcedure", (gpointer*)(&classprocedure));

	if(!libraryinfo)
		std::cout << "Library " << FileName << ": missing sdpComLibraryInfo entry point" << std::endl;

	if(!classprocedure)
		std::cout << "Library " << FileName << ": missing sdpComClassProcedure entry point" << std::endl;

	// If we're missing either of the sdpCom entry points, we're done ...
	if(!(libraryinfo && classprocedure))
		return true;

	// Store the library ...
	m_Libraries.push_back(new CObjectLibrary(FilePath, FileName, module, libraryinfo, classprocedure));

	return true;
}

ObjectInfoPointers CObjectLibraries::Objects()
{
	ObjectInfoPointers results;

	ObjectLibraryIterator libraryend = m_Libraries.end();
	for(ObjectLibraryIterator i = m_Libraries.begin(); i != libraryend; i++)
		{
			ObjectInfoIterator infoend = (*i)->Objects().end();
			for(ObjectInfoIterator j = (*i)->Objects().begin(); j != infoend; j++)
				{
					results.push_back(*j);
				}
		}

	return results;
}

// Return the object info for the given class ID ...
CObjectInfo* CObjectLibraries::ObjectInfo(const sdpCLSID& ClassID)
{
	ObjectLibraryIterator libraryend = m_Libraries.end();
	for(ObjectLibraryIterator i = m_Libraries.begin(); i != libraryend; i++)
		{
			ObjectInfoIterator infoend = (*i)->Objects().end();
			for(ObjectInfoIterator  j = (*i)->Objects().begin(); j != infoend; j++)
				{
					if((*j)->ClassID() == ClassID)
						return (*j);
				}
		}

	g_return_val_if_fail(0, 0);
}

