// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CRenderFrame class, which encapsulates a frame to be rendered as part of a job
		\author Tim Shead (tshead@k-3d.com)
*/

#include "renderframe.h"

#include <k3direnderjob.h>
#include <sdpxmldocument.h>
#include <sdpfile.h>

#include <fstream>

CRenderFrame::CRenderFrame(k3dIRenderJob* Job, SDPCSTRING FrameName)
{
	// Sanity checks ...
	g_assert(Job);
	g_assert_string(FrameName);
	
	// Store data ...
	m_Job = Job;

	// Ensure that we end up with a unique name ...
	for(gulong i = 0; ; i++)
		{
			if(i)
				m_Name = sdpFormatString("%s-%u", FrameName, i);
			else
				m_Name = FrameName;
			
			if(!sdpFileExists(sdpAppendPath(Job->Path(), m_Name)))
				break;
		}
		
	// Store our new path ...
	m_Path = sdpAppendPath(Job->Path(), m_Name);
	
	// Create the directory ...
	if(!sdpCreateDirectory(m_Path))
		std::cerr << "CRenderFrame::CRenderFrame(): Error creating frame directory \"" << m_Path << "\"" << std::endl;
	
	// We default to moderate priority ...
	m_Priority = k3dIRenderFrame::MODERATE_PRIORITY;
	
	// We default to hidden rendering ...
	m_ViewDuringRender = false;
	m_ViewAfterRender = false;
}

CRenderFrame::~CRenderFrame()
{
}

//////////////////////////////////////////////////////////////////////////////////////
// k3dIRenderFrame overrides ...
			
k3dIRenderJob* CRenderFrame::Job()
{
	return m_Job;
}
	
SDPCSTRING CRenderFrame::Name()
{
	return m_Name;
}

SDPCSTRING CRenderFrame::Path()
{
	return m_Path;
}

sdpResult CRenderFrame::GetUniqueFile(SDPCSTRING File, SDPSTRING Buffer, gulong* BufferLength)
{
	// Sanity checks ...
	g_return_val_if_string_fail(File, SDP_E_FAIL);
	g_return_val_if_fail(Buffer, SDP_E_FAIL);
	g_return_val_if_fail(BufferLength, SDP_E_FAIL);
	g_return_val_if_fail(*BufferLength, SDP_E_FAIL);

	// Get the current time ...
	for(gulong i = 0; ; i++)
		{
			// Generate the path ...
			sdpString path;
			if(i)
				path = sdpAppendPath(Path(), sdpFormatString("%s-%u", File, i));
			else
				path = sdpAppendPath(Path(), File);

			// Make sure the file doesn't already exist ...
			if(sdpFileExists(path))
				continue;

			// Make sure the given buffer is large enough ...
			g_return_val_if_fail(path.size() <= *BufferLength, SDP_E_FAIL);

			// Copy the resulting path into the buffer ...
			strcpy(Buffer, SDPCSTRING(path));
			*BufferLength = path.size();

			return SDP_S_OK;
		}

	return SDP_E_FAIL;
}

sdpResult CRenderFrame::SetPriority(guchar Priority)
{
	m_Priority = Priority;
	return SDP_S_OK;
}

guchar CRenderFrame::GetPriority()
{
	return m_Priority;
}

sdpResult CRenderFrame::SetViewDuringRender(bool View)
{
	m_ViewDuringRender = View;
	return SDP_S_OK;
}

bool CRenderFrame::GetViewDuringRender()
{
	return m_ViewDuringRender;
}

sdpResult CRenderFrame::SetViewAfterRender(bool View)
{
	m_ViewAfterRender = View;
	return SDP_S_OK;
}

bool CRenderFrame::GetViewAfterRender()
{
	return m_ViewAfterRender;
}

sdpResult CRenderFrame::SetEngineType(SDPCSTRING EngineType)
{
	// Sanity checks ...
	g_return_val_if_string_fail(EngineType, SDP_E_FAIL);
	
	m_EngineType = EngineType;
	return SDP_S_OK;
}

SDPCSTRING CRenderFrame::GetEngineType()
{
	return m_EngineType;
}

sdpResult CRenderFrame::SetEngine(SDPCSTRING Engine)
{
	g_return_val_if_string_fail(Engine, SDP_E_FAIL);
	
	m_Engine = Engine;
	return SDP_S_OK;
}

SDPCSTRING CRenderFrame::GetEngine()
{
	return m_Engine;
}

//////////////////////////////////////////////////////////////////////////////
// Helper methods ...

bool CRenderFrame::WriteControlFile()
{
	// Create the empty control file ...
	std::ofstream file(sdpAppendPath(Path(), ControlFile));
	g_return_val_if_fail(file.good(), false);
	
	// Create the XML document ...
	sdpXmlDocument document;
	sdpXmlIElement* root = document.CreateRootElement("k3dml");
	g_assert(root);
	
	sdpXmlIElement* frame = root->CreateElement("frame");
	g_assert(frame);
	
	frame->CreateAttribute("priority", sdpString(m_Priority));
	frame->CreateAttribute("viewduringrender", sdpString(m_ViewDuringRender));
	frame->CreateAttribute("viewafterrender", sdpString(m_ViewAfterRender));
	frame->CreateAttribute("enginetype", m_EngineType);
	frame->CreateAttribute("engine", m_Engine);
	
	// Write the document ...
	g_return_val_if_fail(document.Save(file), false);	
	
	return true;
}

bool CRenderFrame::MarkReady()
{
	// Create the empty "ready" status file ...
	std::ofstream file(sdpAppendPath(Path(), StatusReadyFile));
	g_return_val_if_fail(file.good(), false);
	
	return true;
}



