#ifndef K3DIDOCUMENT_H
#define K3DIDOCUMENT_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the k3dIDocument interface, which encapsulates an open K-3D document
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dilayer.h>
#include <k3diobjectcontext.h>

// Forward class references
class k3dIActions;
class k3dIApplication;
class k3dICamera;
class k3dIObject;
class k3dIMesh;
class k3dIPolygon;
class k3dIPoint;
class k3dIMaterial;
class k3dIGeometryReadFormat;
class k3dIGeometryWriteFormat;

SDP_DECLARE_HANDLE(K3DTREEHANDLE);

/// Interface encapsulating an open K-3D document
class k3dIDocument : public virtual sdpIUnknown
{
public:
	/// Stores the interface ID
	static const sdpIID iid;

	/// Enumerates the possible axes that can the "active"
	typedef enum
	{
		XAXIS,
		YAXIS,
		ZAXIS,
	} Axis;

	/// Enumerates the possible transformation types that can be "active"
	typedef enum
	{
		POSITION,
		ORIENTATION,
		SCALE
	} TransformType;

	/// Enumerates the possible coordinate systems that can be "active"
	typedef enum
	{
		MODEL,
		WORLD,
		SCREEN,
	} CoordinateSystem;

	/// Enumerates the possible edit modes than can be "active"
	typedef enum
	{
		POINT,
		POLYGON,
		MESH,
	} EditMode;

	/// Enumerates the tree-node values used to iterate through the object hierarchy
	typedef enum
	{
		COMPLETE,
		PUSH,
		NODE,
		POP,
		TREE_NODE_ERROR
	} TreeNodeType;

	/// Returns a k3dIApplication interface on the owning application
	virtual k3dIApplication* Application() = 0;

	/// Imports a geometry file into the document, using the given file format
	virtual sdpResult Import(SDPCSTRING FilePath, SDPCSTRING Format, k3dIGeometryReadFormat* FormatFilter) = 0;
	/// Exports the document to a geometry file, using the given file format
	virtual sdpResult Export(SDPCSTRING FilePath, SDPCSTRING Format, k3dIGeometryWriteFormat* FormatFilter) = 0;
	/// Saves the document
	virtual sdpResult Save() = 0;
	/// Saves the document to the given file
	virtual sdpResult SaveAs(SDPCSTRING FilePath) = 0;

	/// Returns the k3dIAction interface for this document
	virtual k3dIActions* Actions() = 0;

	/// Creates a K-3D object
	virtual k3dIObject* CreateObject(const sdpCLSID& ClassID, bool Edit, bool UserVisible, bool Saveable) = 0;
	/// Destroys a K-3D object
	virtual sdpResult DeleteObject(k3dIObject* Object) = 0;

	/// Returns the currently active camera
	virtual k3dICamera* GetActiveCamera() = 0;
	/// Sets the active camera
	virtual sdpResult SetActiveCamera(k3dICamera* Camera) = 0;

	/// Returns a handle to the first layer within the document
	virtual K3DLAYERHANDLE GetFirstLayer() = 0;
	/// Returns a handle to the next layer within the document
	virtual K3DLAYERHANDLE GetNextLayer(K3DLAYERHANDLE Layer) = 0;
	/// Converts a handle into a k3dILayer interface on the underlying layer
	virtual k3dILayer* GetLayer(K3DLAYERHANDLE Layer) = 0;
	/// Creates a new layer
	virtual k3dILayer* CreateLayer() = 0;
	/// Destroys the given layer
	virtual sdpResult DestroyLayer(k3dILayer* Layer) = 0;
	/// Sets the currently active layer
	virtual sdpResult SetActiveLayer(k3dILayer* Layer) = 0;
	/// Returns the currently active layer
	virtual k3dILayer* GetActiveLayer() = 0;

	/// Returns a handle to the first (flat) K-3D object (context) in the document
	virtual K3DOBJECTCONTEXTHANDLE GetFirstObjectContext() = 0;
	/// Returns a handle to the next (flat) K-3D object (context) in the document
	virtual K3DOBJECTCONTEXTHANDLE GetNextObjectContext(K3DOBJECTCONTEXTHANDLE ObjectContext) = 0;
	/// Converts a handle into a k3dIObjectContext interface on the underlying object context
	virtual k3dIObjectContext* GetObjectContext(K3DOBJECTCONTEXTHANDLE ObjectContext) = 0;
	/// Sets the currently active object (context)
	virtual sdpResult SetActiveObjectContext(k3dIObjectContext* ObjectContext) = 0;
	/// Returns the currently active object (context)
	virtual k3dIObjectContext* GetActiveObjectContext() = 0;

	/// Returns a handle to the root (hierarchical) K-3D object (context) in the document
	virtual TreeNodeType GetFirstObjectContextTree(K3DTREEHANDLE* Handle) = 0;
	/// Returns a handle to the next (hierarchical) K-3D object (context) in the document
	virtual TreeNodeType GetNextObjectContextTree(K3DTREEHANDLE* Handle) = 0;
	/// Returns a (hierarchical) handle to an arbitrary object within the document
	virtual TreeNodeType GetContextHandleTree(k3dIObjectContext* Context, K3DTREEHANDLE* Handle) = 0;
	/// Converts a handle into a k3dIObjectContext interface on the underlying object context
	virtual k3dIObjectContext* GetObjectContextTree(K3DTREEHANDLE Handle) = 0;

	/// Returns the currently active axis
	virtual Axis GetActiveAxis() = 0;
	/// Sets the current active axis
	virtual sdpResult SetActiveAxis(Axis ActiveAxis) = 0;

	/// Returns the currently active transformation type
	virtual TransformType GetActiveTransformType() = 0;
	/// Sets the current active transformation type
	virtual sdpResult SetActiveTransformType(TransformType ActiveTransformType) = 0;

	/// Returns the currently active coordinate system
	virtual CoordinateSystem GetActiveCoordinateSystem() = 0;
	/// Sets the current coordinate system
	virtual sdpResult SetActiveCoordinateSystem(CoordinateSystem System) = 0;

	/// Returns the currently active edit mode
	virtual EditMode GetActiveEditMode() = 0;
	/// Sets the current active edit mode
	virtual sdpResult SetActiveEditMode(EditMode Mode) = 0;

	/// Freezes changes to the document
	virtual sdpResult FreezeChanges() = 0;
	/// Makes k3dILocation WorldXYZ cooordinates current
	virtual sdpResult SnapshotWorldCoordinates() = 0;
	/// Updates the document to reflect the current time
	virtual sdpResult AnimateDocument() = 0;

	/// Returns the default material (guaranteed valid)
	virtual k3dIMaterial* GetDefaultMaterial() = 0;
};

#endif // K3DIDOCUMENT_H
