#ifndef K3DIOBJECT_H
#define K3DIOBJECT_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the k3dIObject interface, which MUST be implemented by ALL K-3D objects
		\author Tim Shead (tshead@k-3d.com)
*/

#include <sdpcom.h>
#include <vector>
#include <list>
#include <set>

// Forward class references
class k3dIApplication;
class k3dIDocument;
class k3dIObjectContext;
class k3dIEventClient;

/// Declares mandatory functionality for all K-3D objects
class k3dIObject : public virtual sdpIUnknown
{
public:
	/// Stores the interface ID
	static const sdpIID iid;

	/// Called when an object is instantiated to connect it to its object context
	virtual sdpResult SetObjectContext(k3dIObjectContext* Context) = 0;

	/// Returns the object's object context
	virtual k3dIObjectContext* ObjectContext() = 0;
	/// Called to request that the object open its user interface
	virtual sdpResult EditObject() = 0;

	/// Called to request that the given client be notified when the specified event occurs
	virtual sdpResult AddClient(k3dIEventClient* Client, gulong UserData, SDPCSTRING Event) = 0;
	/// Called to remove a client from future event notification
	virtual sdpResult RemoveClient(k3dIEventClient* Client, gulong UserData, SDPCSTRING Event) = 0;
	/// Instructs the object to signal the given event
	virtual sdpResult Notify(SDPCSTRING Event) = 0;
	/// Instructs the object to begin queueing events instead of signalling them
	virtual sdpResult StartQueueEvents() = 0;
	/// Instructs the object to signal all queued events and resume normal signalling
	virtual sdpResult EndQueueEvents() = 0;

	/// Called to see if the object may be selected
	virtual sdpResult QuerySelectObject() = 0;
	/// Called to notify the object that its selection state has changed
	virtual sdpResult ObjectSelected(bool State) = 0;
	/// Called to see if the object may be renamed
	virtual sdpResult QueryRenameObject() = 0;
	/// Called to notify the object that it has been renamed
	virtual sdpResult ObjectRenamed(SDPCSTRING NewName) = 0;
	/// Called to see if the object may be deleted
	virtual sdpResult QueryDeleteObject() = 0;
	/// Called to notify the object that it has been deleted
	virtual sdpResult ObjectDeleted() = 0;
};

/// Contains an array of objects
typedef std::vector<k3dIObject*> k3dObjectArray;
/// Contains a linked-list of objects
typedef std::list<k3dIObject*> k3dObjectList;
/// Contains a set of objects
typedef std::set<k3dIObject*> k3dObjectSet;

#endif // K3DIOBJECT_H
