#ifndef K3DBOUNDINGBOX_H
#define K3DBOUNDINGBOX_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the k3dBoundingBox2 and k3dBoundingBox3 classes, which encapsulate rectangular and cubic regions, respectively
		\author Tim Shead (tshead@k-3d.com)
*/

#include <sdpmath.h>

//////////////////////////////////////////////////////////////////////////////////
// k3dBoundingBox2

class k3dBoundingBox2
{
public:
	// Constructors
	k3dBoundingBox2();
	k3dBoundingBox2(const sdpVector2& Corner1, const sdpVector2& Corner2);

	// Assignment operators
	k3dBoundingBox2& operator	= ( const k3dBoundingBox2& b );		// assignment of a k3dBoundingBox2

	// Special functions
	void Init();													// Initialize a k3dBoundingBox2 to max bounds
	void Expand(const sdpVector2& Point);		// Expand current bounds if necessary to contain the given point
	bool IsDefined();											// Test to see if bounding box is defined (m_Min < m_Max) or not
	bool Contains(sdpVector2& Point);				// Test to see if the given point is within the current bounds

	sdpVector2 Min();												// Return the minimum bounds as a vector
	sdpVector2 Max();												// Return the maximum bounds as a vector
	sdpVector2 Center();										// Return the geometric center as a vector
	sdpVector2 Size();											// Return the dimensions of the box as a vector

protected:
	sdpVector2 m_Min;
	sdpVector2 m_Max;

	// Friends
	friend k3dBoundingBox2 k3dIntersect(const k3dBoundingBox2& A, const k3dBoundingBox2& B);
};

//////////////////////////////////////////////////////////////////////////////////
// k3dBoundingBox3

class k3dBoundingBox3
{
public:
	// Constructors
	k3dBoundingBox3();
	k3dBoundingBox3(const sdpVector3& Corner1, const sdpVector3& Corner2);

	// Assignment operators
	k3dBoundingBox3& operator	= ( const k3dBoundingBox3& b );		// assignment of a k3dBoundingBox3

	// Special functions
	void Init();													// Initialize a k3dBoundingBox3 to max bounds
	void Expand(const sdpVector3& Point);		// Expand current bounds if necessary to contain the given point
	bool IsDefined();											// Test to see if bounding box is defined (m_Min < m_Max) or not
	bool Contains(sdpVector3& Point);				// Test to see if the given point is within the current bounds

	sdpVector3 Min();												// Return the minimum bounds as a vector
	sdpVector3 Max();												// Return the maximum bounds as a vector
	sdpVector3 Center();										// Return the geometric center as a vector
	sdpVector3 Size();											// Return the dimensions of the box as a vector

protected:
	sdpVector3 m_Min;
	sdpVector3 m_Max;

	// Friends
	friend k3dBoundingBox3 k3dIntersect(const k3dBoundingBox3& A, const k3dBoundingBox3& B);
};


///////////////////////////////////////////////////////////////////
// k3dBoundingBox2 implementation

inline k3dBoundingBox2::k3dBoundingBox2()
{	Init(); }


inline k3dBoundingBox2::k3dBoundingBox2(const sdpVector2& Corner1, const sdpVector2& Corner2)
{	m_Min = sdpVectorMin(Corner1, Corner2); m_Max = sdpVectorMax(Corner1, Corner2); }


inline k3dBoundingBox2& k3dBoundingBox2::operator = (const k3dBoundingBox2& Box)
{	m_Min = Box.m_Min; m_Max = Box.m_Max;	return *this; }


inline void k3dBoundingBox2::Init()
{	m_Min = sdpVector2(DBL_MAX); m_Max = sdpVector2(-DBL_MAX); }


inline void k3dBoundingBox2::Expand(const sdpVector2& Point)
{	m_Min = sdpVectorMin(m_Min, Point); m_Max = sdpVectorMax(m_Max, Point); }


inline bool k3dBoundingBox2::IsDefined()
{	return (m_Min[0] <= m_Max[0]) && (m_Min[1] <= m_Max[1]) && (m_Min[2] <= m_Max[2]); }


inline bool k3dBoundingBox2::Contains(sdpVector2& Point)
{	return (m_Min[0] <= Point[0]) && (Point[0] <= m_Max[0]) && (m_Min[1] <= Point[1]) && (Point[1] <= m_Max[1]); }


inline sdpVector2 k3dBoundingBox2::Min()
{	return m_Min; }


inline sdpVector2 k3dBoundingBox2::Max()
{	return m_Max; }


inline sdpVector2 k3dBoundingBox2::Center()
{	return (m_Min + m_Max) * 0.5; }


inline sdpVector2 k3dBoundingBox2::Size()
{	return m_Max - m_Min; }


inline k3dBoundingBox2 k3dIntersect(const k3dBoundingBox2& a, const k3dBoundingBox2& b)
{
	sdpVector2 corner1(sdpVectorMax(a.m_Min, b.m_Min));
	sdpVector2 corner2(sdpVectorMin(a.m_Max, b.m_Max));
	return k3dBoundingBox2(corner1, corner2);
}


///////////////////////////////////////////////////////////////////
// k3dBoundingBox3

inline k3dBoundingBox3::k3dBoundingBox3()
{	Init(); }


inline k3dBoundingBox3::k3dBoundingBox3(const sdpVector3& Corner1, const sdpVector3& Corner2)
{	m_Min = sdpVectorMin(Corner1, Corner2); m_Max = sdpVectorMax(Corner1, Corner2); }


inline k3dBoundingBox3& k3dBoundingBox3::operator = (const k3dBoundingBox3& Box)
{	m_Min = Box.m_Min; m_Max = Box.m_Max;	return *this; }


inline void k3dBoundingBox3::Init()
{	m_Min = sdpVector3(DBL_MAX); m_Max = sdpVector3(-DBL_MAX); }


inline void k3dBoundingBox3::Expand(const sdpVector3& Point)
{	m_Min = sdpVectorMin(m_Min, Point); m_Max = sdpVectorMax(m_Max, Point); }


inline bool k3dBoundingBox3::IsDefined()
{ return (m_Min[0] <= m_Max[0]) && (m_Min[1] <= m_Max[1]) && (m_Min[2] <= m_Max[2]); }


inline bool k3dBoundingBox3::Contains(sdpVector3& Point)
{
	return (m_Min[0] <= Point[0]) && (Point[0] <= m_Max[0]) &&
				 (m_Min[1] <= Point[1]) && (Point[1] <= m_Max[1]) &&
				 (m_Min[2] <= Point[2]) && (Point[2] <= m_Max[2]);
}


inline sdpVector3 k3dBoundingBox3::Min()
{	return m_Min; }


inline sdpVector3 k3dBoundingBox3::Max()
{	return m_Max; }


inline sdpVector3 k3dBoundingBox3::Center()
{	return (m_Min + m_Max) * 0.5; }


inline sdpVector3 k3dBoundingBox3::Size()
{	return m_Max - m_Min; }


inline k3dBoundingBox3 k3dIntersect(const k3dBoundingBox3& a, const k3dBoundingBox3& b)
{
	sdpVector3 corner1(sdpVectorMax(a.m_Min, b.m_Min));
	sdpVector3 corner2(sdpVectorMin(a.m_Max, b.m_Max));
	return k3dBoundingBox3(corner1, corner2);
}

#endif // K3DBOUNDINGBOX_H
