#ifndef K3DFILEFILTERS_H
#define K3DFILEFILTERS_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares convenience functions for working with file filter objects
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3diapplication.h>
#include <k3didocument.h>
#include <k3dmisc.h>

/// Returns a new file format object that can read/write the given file
template <class Interface>
Interface* k3dFileFormat(k3dIDocument* Document, SDPCSTRING FilePath, SDPCSTRING Format)
{
	// Sanity checks ...
	g_assert(Document);
	g_assert_string(FilePath);
	g_assert_string(Format);

	// For each object info ...
	for(K3DOBJECTINFOHANDLE handle = Document->Application()->GetFirstObjectInfo(); handle; handle = Document->Application()->GetNextObjectInfo(handle))
		{
			// If it doesn't support our interface, skip it ...
			k3dIObjectInfo* info = Document->Application()->GetObjectInfo(handle);
			g_return_val_if_fail(info, 0);

			if(SDP_S_OK != info->QueryObjectInterface(Interface::iid))
				continue;

			// Create an object of this type ...
			k3dIObject* object = Document->CreateObject(info->ClassID(), false, false, false);
			g_return_val_if_fail(object, 0);

			// Get the requested interface pointer ...
			Interface* format = sdpQueryInterface<Interface>(object);
			g_return_val_if_fail(format, 0);

			// If the object supports this file type, we're done ...
			if(SDP_S_OK == format->FileFormat()->QueryCanHandle(FilePath, Format))
				{
					object->Release();
					return format;
				}

			// Otherwise, destroy the object and continue ...
			k3dDeleteObject(format);
		}

	// Couldn't find a format to fit this filetype ...
	return 0;
}

/// Returns all available file formats for the given file filter type
sdpStringMultiSet k3dFileFormats(k3dIDocument* Document, const sdpIID& InterfaceID);

#endif K3DFILEFILTERS_H
