#ifndef K3DMESH_H
#define K3DMESH_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the k3dMesh class, which implements k3dIMesh, and encapsulates a collection of points, polygons, NURBS curves, material, and shading information
		\author Timothy M. Shead (tshead@k-3d.com)
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include <k3dtransformobject.h>
#include <k3dvariable.h>

#include <k3dimesh.h>
#include <k3dipolygonserver.h>
#include <k3dinurbscurveserver.h>
#include <k3dipointserver.h>
#include <k3dilocationserver.h>

// Forward class declarations ...
class k3dPoint;
class k3dPolygon;
class k3dNURBSCurve;
class k3dPolyPoint;
class k3dShadingNode;
class k3dIRiEngine;
class k3dIRiMaterial;

/////////////////////////////////////////////////////////////////////////////////////
// k3dMesh

/// Encapsulates a collection of points, polygons, NURBS curves, material, and shading information
class k3dMesh : public k3dTransformObject,
							public k3dIMesh,
							public k3dIPolygonServer,
							public k3dINURBSCurveServer,
							public k3dIPointServer,
							public k3dILocationServer
{
friend class k3dMeshLoadShadingNode;
public:
	k3dMesh();
	~k3dMesh();

	// k3dIMesh overrides ...
	k3dIDocument* Document();
	k3dIPolygonServer* Polygons();
	k3dIPolygon* CreatePolygon();
	sdpResult DeletePolygon(k3dIPolygon* Polygon, bool DeletePoints);
	k3dINURBSCurveServer* NURBSCurves();
	k3dINURBSCurve* CreateNURBSCurve();
	sdpResult DeleteNURBSCurve(k3dINURBSCurve* Curve, bool DeletePoints);
	k3dIPointServer* Points();
	k3dIPoint* CreatePoint();
	sdpResult DeletePoint(k3dIPoint* Point);
	sdpResult ShadePolygons(gulong Criteria, gdouble Angle);
	sdpResult UnshadePolygons(gulong Criteria);
	sdpResult CalculateNormals();
	sdpResult GetDimensions(gulong Criteria, double Min[3], double Max[3]);
	sdpResult GetColor(double Color[3]);
	sdpResult SetColor(double Color[3]);

	// k3dIObject overrides ...
	sdpResult ObjectSelected(bool State);

	// k3dIActionClient overrides ...
	sdpResult HandleStep(gulong ValueCode, gulong Value, gulong ValueSize);

	// k3dIPolygonServer overrides ...
	gulong GetPolygonCount();
	K3DPOLYGONHANDLE GetFirstPolygon();
	K3DPOLYGONHANDLE GetNextPolygon(K3DPOLYGONHANDLE Polygon);
	k3dIPolygon* GetPolygon(K3DPOLYGONHANDLE Polygon);
	sdpResult AddPolygon(k3dIPolygon* Polygon);
	sdpResult RemovePolygon(k3dIPolygon* Polygon);

	// k3dINURBSCurveServer overrides ...
	gulong GetNURBSCurveCount();
	K3DNURBSCURVEHANDLE GetFirstNURBSCurve();
	K3DNURBSCURVEHANDLE GetNextNURBSCurve(K3DNURBSCURVEHANDLE Curve);
	k3dINURBSCurve* GetNURBSCurve(K3DNURBSCURVEHANDLE Curve);
	sdpResult AddNURBSCurve(k3dINURBSCurve* Curve);
	sdpResult RemoveNURBSCurve(k3dINURBSCurve* Curve);

	// k3dIPointServer overrides ...
	gulong GetPointCount();
	K3DPOINTHANDLE GetFirstPoint();
	K3DPOINTHANDLE GetNextPoint(K3DPOINTHANDLE Point);
	k3dIPoint* GetPoint(K3DPOINTHANDLE Point);
	sdpResult AddPoint(k3dIPoint* Point);
	sdpResult RemovePoint(k3dIPoint* Point);

	// k3dILocationServer overrides ...
	gulong GetLocationCount();
	K3DLOCATIONHANDLE GetFirstLocation();
	K3DLOCATIONHANDLE GetNextLocation(K3DLOCATIONHANDLE Location);
	k3dILocation* GetLocation(K3DLOCATIONHANDLE Location);

	// k3dIEditorRender overrides ...
	sdpResult EditorRender(k3dIEditorEngine* Engine);

	// k3dIRiRender overrides ...
	sdpResult RiPushRender(k3dIRenderFrame* Frame, k3dIRiEngine* Engine);
	sdpResult RiRender(k3dIRenderFrame* Frame, k3dIRiEngine* Engine);
	sdpResult RiPopRender(k3dIRenderFrame* Frame, k3dIRiEngine* Engine);

	// k3dIPersist overrides ...
	sdpResult Save(sdpXmlIElement* Element, k3dIDependencies* Dependencies);
	sdpResult Load(sdpXmlIElement* Element);
	sdpResult LoadComplete();

	/// Returns a point based on its point ID
	k3dPoint* FindPoint(k3dID PointID);
	/// Returns a polypoint based on its polypoint ID
	k3dPolyPoint* FindPolyPoint(k3dID PolyPointID);

	/// Returns the next unique point ID
	k3dID GetNextPointID();
	/// Returns the next unique polygon ID
	k3dID GetNextPolygonID();
	/// Returns the next unique NURBS curve ID
	k3dID GetNextNURBSCurveID();
	/// Returns the next unique polypoint ID
	k3dID GetNextPolyPointID();

protected:
	/// Stores the current point ID
	k3dID m_CurrentPointID;
	/// Stores the current polygon ID
	k3dID m_CurrentPolygonID;
	/// Stores the current NURBS curve ID
	k3dID m_CurrentNURBSCurveID;
	/// Stores the current polypoint ID
	k3dID m_CurrentPolyPointID;

	// Patch lookup ...
	static short m_RiBilinearPatchLookup[4];

	/// Stores the color for drawing selected geometry
	static sdpVector3 m_SelectedColor;
	/// Stores the default color for drawing unselected geometry
	static sdpVector3 m_DefaultColor;

	/// Stores the color for drawing unselected geometry
	k3dVariable<sdpVector3> m_Color;

	typedef std::vector<k3dPoint*> PointArray;
	typedef PointArray::iterator PointIterator;
	/// Stores the array of point objects
	PointArray m_Points;
	/// Stores points that have been deleted
	PointArray m_DeletedPoints;

	typedef std::vector<k3dPolygon*> PolygonArray;
	typedef PolygonArray::iterator PolygonIterator;
	/// Stores the array of polygon objects
	PolygonArray m_Polygons;
	/// Stores polygons that have been deleted
	PolygonArray m_DeletedPolygons;

	typedef std::vector<k3dNURBSCurve*> NURBSCurveArray;
	typedef NURBSCurveArray::iterator NURBSCurveIterator;
	/// Stores the array of NURBS curves
	NURBSCurveArray m_NURBSCurves;
	/// Stores NURBS curves that have been deleted
	NURBSCurveArray m_DeletedNURBSCurves;

	typedef std::list<k3dShadingNode*> ShadingNodeArray;
	typedef ShadingNodeArray::iterator ShadingNodeIterator;
	/// Stores the array of shading node objects
	ShadingNodeArray m_ShadingNodes;

	/// Saves geometry to a K-3D (XML) document
	sdpResult SaveGeometry(sdpXmlIElement* Element, k3dIDependencies* Dependencies);
	/// Saves points to a K-3D (XML) document
	sdpResult SavePoints(sdpXmlIElement* Element, k3dIDependencies* Dependencies);
	/// Saves polygons to a K-3D (XML) document
	sdpResult SavePolygons(sdpXmlIElement* Element, k3dIDependencies* Dependencies);
	/// Saves NURBS curves to a K-3D (XML) document
	sdpResult SaveNURBSCurves(sdpXmlIElement* Element, k3dIDependencies* Dependencies);
	/// Saves shading nodes to a K-3D (XML) document
	sdpResult SaveShading(sdpXmlIElement* Element, k3dIDependencies* Dependencies);

	/// Loads geometry from a K-3D (XML) document
	sdpResult LoadGeometry(sdpXmlIElement* Element);
	/// Loads points from a K-3D (XML) document
	sdpResult LoadPoints(sdpXmlIElement* Element);
	/// Loads polygons from a K-3D (XML) document
	sdpResult LoadPolygons(sdpXmlIElement* Element);
	/// Loads NURBS curves from a K-3D (XML) document
	sdpResult LoadNURBSCurves(sdpXmlIElement* Element);
	/// Loads shading nodes from a K-3D (XML) document
	sdpResult LoadShading(sdpXmlIElement* Element);

	/// Editor rendering in wireframe mode
	inline void EditorRenderWireframe(k3dIEditorEngine* Engine);
	/// Editor rendering in solid mode
	inline void EditorRenderSolid(k3dIEditorEngine* Engine);

	/// Draws polygons as wireframes
	inline void DrawWireframePolygons(k3dIEditorEngine* Engine, gulong Criteria);
	/// Draws all points
	inline void DrawPoints();
	/// Draws normals for all selected polygons
	inline void DrawSelectedNormals();

	/// Draws all visible polygons solid
	inline void EditorSolidPolygons(k3dIEditorEngine* Engine);
	/// Draws all visible polygons solid & textured
	inline void EditorSolidTexturedPolygons(k3dIEditorEngine* Engine);

	/// Stores an array of points for Renderman rendering
	typedef std::vector<sdpVector3> RtPointArray;
	/// Stores an array of scalars for Renderman rendering
	typedef std::vector<gdouble> RtFloatArray;

	/// Draws a polygon
	inline void RiPolygon(k3dIRiEngine* Engine, k3dIRiMaterial* Material, k3dPolygon* Polygon);
	/// Draws a 3-sided polygon
	inline void RiPolygon3(k3dIRiEngine* Engine, k3dIRiMaterial* Material, k3dPolygon* Polygon);
	/// Draws a 4-sided polygon
	inline void RiPolygon4(k3dIRiEngine* Engine, k3dIRiMaterial* Material, k3dPolygon* Polygon);
	/// Draws a greater-than-4 sided polygon
	inline void RiPolygonN(k3dIRiEngine* Engine, k3dIRiMaterial* Material, k3dPolygon* Polygon);

	/// Enumerates undo-able actions
	typedef enum
	{
		STEP_CREATE_POLYGON,
		STEP_UNCREATE_POLYGON,
		STEP_UNDELETE_POLYGON,
		STEP_DELETE_POLYGON,
		STEP_CREATE_NURBSCURVE,
		STEP_UNCREATE_NURBSCURVE,
		STEP_UNDELETE_NURBSCURVE,
		STEP_DELETE_NURBSCURVE,
		STEP_CREATE_POINT,
		STEP_UNCREATE_POINT,
		STEP_UNDELETE_POINT,
		STEP_DELETE_POINT,
	} StepType;
};

#endif K3DMESH_H
