#ifndef K3DMISC_H
#define K3DMISC_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares a kitchen-junk-drawer-ful of convenience functions, templates, and classes
		\author Tim Shead (tshead@k-3d.com)
*/

// Forward class declarations
class sdpXmlIElement;
class sdpXmlIDocument;

#include <k3dcom.h>
#include <k3diapplication.h>
#include <k3didocument.h>
#include <k3dichannel.h>
#include <k3diobjectinfo.h>
#include <k3diobject.h>
#include <k3dipoint.h>
#include <k3dipointserver.h>
#include <k3dipolypoint.h>
#include <k3dipolypointserver.h>
#include <k3dipolygon.h>
#include <k3dipolygonserver.h>
#include <k3dilocationserver.h>
#include <k3distate.h>
#include <k3diriengine.h>
#include <k3ditransform.h>
#include <k3diactions.h>
#include <k3diactionclient.h>
#include <k3dicamera.h>

#include <sdpsimpletypes.h>
#include <sdpstring.h>
#include <sdpmath.h>

/// Function for creating a standard sdpcomlibraryinfo document
sdpXmlIElement* k3dStartLibraryInfo(sdpXmlIDocument* Document, SDPCSTRING Name, SDPCSTRING Version);
/// Function for creating a standardized object entry in an sdpcomlibraryinfo document
sdpXmlIElement* k3dAddObject(sdpXmlIElement* Library, const sdpCLSID& ClassID, SDPCSTRING Name);
/// Function for adding a standardized object interface entry in an sdpcomlibraryinfo document
void k3dAddInterface(sdpXmlIElement* Object, const sdpIID& Interface);
/// Function for adding a K-3D object bookmark entry in an sdpcomlibraryinfo document
void k3dAddBookmark(sdpXmlIElement* Object, SDPCSTRING Location);

/// Checks to see if a screen coordinate is contained within a polygon's screen coordinates
bool k3dIsInside(sdpVector2 Point, k3dIPolygon* Polygon);

/// Called to update polygon normals throughout a document
void k3dCalculateNormals(k3dIDocument* Document);
/// Calculates a polygon normal
sdpVector3 k3dGetNormal(const sdpVector3Array& Points);
/// Calculates a polygon normal
sdpVector3 k3dGetNormal(const k3dPointArray& Points);
/// Calculates a polygon normal
sdpVector3 k3dGetNormal(const k3dPolyPointArray& PolyPoints);
/// Calculates a polygon normal
sdpVector3 k3dGetNormal(k3dIPolygon* Polygon);

/// Parses a scalar value from a string, converting units as needed
double k3dConvertStringDouble(k3dIChannel::DataType Data, SDPCSTRING Input);
/// Converts a double to a string, displaying units-of-measure
sdpString k3dConvertDoubleString(k3dIChannel::DataType Data, double InputValue, gulong Precision);
/// Parsers a scalar value from a string, converting units and returning a string with the converted value and units
sdpString k3dUpdateDoubleString(k3dIChannel::DataType Data, SDPCSTRING Input, double& OutputValue, gulong Precision);

/// Returns the absolute path to an object's library
SDPCSTRING k3dLibraryPath(k3dIObject* Object);
/// Returns the absolute path to the global dialog template collection
sdpString k3dGlobalDialogTemplatePath(k3dIApplication* Application);

/// Deletes an object (this is undo-able)
sdpResult k3dDeleteObject(sdpIUnknown* Object);

/// Sets an object's position (if possible)
bool k3dSetPosition(sdpIUnknown* Object, sdpVector3 Position);
/// Sets an object's orientation (if possible)
bool k3dSetOrientation(sdpIUnknown* Object, sdpEulerAngles Orientation);
/// Sets an object's scale (if possible)
bool k3dSetScale(sdpIUnknown* Object, sdpVector3 Scale);

/// Sets an object's name (if possible)
bool k3dSetName(sdpIUnknown* Object, SDPCSTRING ObjectName);

/// Sets a channel's maximum range of values
bool k3dChannelSetValueRange(sdpIUnknown* Channel, const gdouble MinValue, const gdouble MaxValue);
/// Sets a channel's maximum range of values
bool k3dChannelSetValueRange(sdpIUnknown* Channel, const sdpVector3 MinValue, const sdpVector3 MaxValue);
/// Sets a channel to a constant value (if possible)
bool k3dChannelSetConstantValue(sdpIUnknown* Channel, const gdouble Value);
/// Sets a channel to a constant value (if possible)
bool k3dChannelSetConstantValue(sdpIUnknown* Channel, const sdpVector3 Value);
/// Sets a channel to a linear range of values (if possible)
bool k3dChannelSetLinearValue(sdpIUnknown* Channel, const gdouble Value1, const gdouble Value2);
/// Sets a channel to a linear range of values (if possible)
bool k3dChannelSetLinearValue(sdpIUnknown* Channel, const sdpVector3 Value1, const gdouble Value2);

/// Return the min/max world coordinates of all points within the given document that fit the given criteria
bool k3dWorldDimensions(k3dIDocument* Document, gulong Criteria, sdpVector3& Min, sdpVector3& Max);
/// Returns the dimensions of an arbitrary object whose points/locations meet the given criteria
bool k3dObjectDimensions(k3dIObject* Object, gulong Criteria, sdpVector3& Min, sdpVector3& Max);
/// Returns the min/max coordinates of all points within the given location server that fit the given criteria
bool k3dObjectDimensions(k3dILocationServer* LocationServer, gulong Criteria, sdpVector3& Min, sdpVector3& Max);

/// Returns an object by ID
k3dIObject* k3dGetObject(k3dIDocument* Document, k3dID ObjectID);
/// Returns an object by name
k3dIObject* k3dGetObject(k3dIDocument* Document, SDPCSTRING Name);

/// Returns a set of object infos (zero or one) that implement the requested class ID
void k3dGetObjectInfo(k3dIApplication* Application, const sdpCLSID& ClassID, k3dObjectInfoSet& Results);
/// Returns a set of object infos that support the requested interface ID
void k3dGetObjectInfo(k3dIApplication* Application, const sdpIID& InterfaceID, k3dObjectInfoSet& Results);
/// Returns a set of object infos that support an object with the requested name
void k3dGetObjectInfo(k3dIApplication* Application, SDPCSTRING Name, k3dObjectInfoSet& Results);
/// Returns a set of objects that are of the requested class type
void k3dGetObjects(k3dIDocument* Document, const sdpCLSID& ClassID, k3dObjectSet& Results);
/// Returns a set of objects that support the requested interface
void k3dGetObjects(k3dIDocument* Document, const sdpIID& InterfaceID, k3dObjectSet& Results);
/// Returns a set of objects that fit the given state criteria
void k3dGetObjects(k3dIDocument* Document, const gulong Criteria, k3dObjectSet& Results);
/// Returns an array of meshes that fit the given state criteria
void k3dGetMeshes(k3dIDocument* Document, const gulong Criteria, k3dMeshArray& Results);
/// Returns an array of points that fit the given state criteria
void k3dGetPoints(k3dIDocument* Document, const gulong Criteria, k3dPointArray& Results);
/// Returns an array of points from a polygon
void k3dGetPoints(k3dIPolygon* Polygon, k3dPointArray& Results);
/// Returns an array of polypoints from a polygon
void k3dGetPolyPoints(k3dIPolygon* Polygon, k3dPolyPointArray& Results);
/// Returns an array of polygons that fit the given state criteria
void k3dGetPolygons(k3dIDocument* Document, const gulong Criteria, k3dPolygonArray& Results);

/// Returns a transformation matrix for the requested object and reference frame
sdpMatrix4 k3dGetTransformMatrix(k3dITransform* Transform, const k3dITransform::ReferenceFrame Frame);
/// Returns a complete, concatenated, object-to-world transformation matrix for an object (includes all its' parents)
sdpMatrix4 k3dGetObjectToWorldMatrix(k3dIObjectContext* Object);
/// Returns a complete, concatenated, world-to-object transformation matrix for an object (includes all its' parents)
sdpMatrix4 k3dGetWorldToObjectMatrix(k3dIObjectContext* Object);
/// Returns a complete, concatenated transformation matrix to transform from one object space to another
sdpMatrix4 k3dGetObjectToObjectMatrix(k3dIObjectContext* From, k3dIObjectContext* To);
/// Returns a complete, concatenated world-to-camera transformation matrix for the given camera
sdpMatrix4 k3dGetWorldToCameraMatrix(k3dICamera* Camera);
/// Returns a complete, concatenated camera-to-world transformation matrix for the given camera
sdpMatrix4 k3dGetCameraToWorldMatrix(k3dICamera* Camera);
/// Returns a complete, concatenated transformation matrix to transform from an object's space to the given camera's space
sdpMatrix4 k3dGetObjectToCameraMatrix(k3dIObjectContext* Object, k3dICamera* Camera);
/// Returns a complete, concatenated transformation matrix to transform from a camera's space to the given object's space
sdpMatrix4 k3dGetCameraToObjectMatrix(k3dICamera* Camera, k3dIObjectContext* Object);

/// Returns the position of an object in world coordinates
sdpVector3 k3dGetWorldPosition(k3dIObjectContext* Context);

/// Pushes a matrix onto the OpenGL matrix stack
void k3dPushOpenGLMatrix(const sdpMatrix4& Matrix);
/// Pushes a matrix onto the Renderman Interface matrix stack
void k3dPushRiMatrix(const sdpMatrix4& Matrix, k3dIRiEngine* Engine);

/// Transform mouse coordinates into some other reference frame
sdpVector2 k3dTransformMouseCoords(k3dICamera* Camera, const sdpVector2 MouseCoords, gdouble Left, gdouble Right, gdouble Top, gdouble Bottom);
/// Transform mouse coordinates to a normalized ray originating at the camera, within the camera's reference frame
sdpVector3 k3dMouseToCameraRay(k3dICamera* Camera, const sdpVector2 MouseCoords);
/// Transform mouse coordinates to a normalized ray originating at the camera, within the world reference frame
void k3dMouseToWorldRay(k3dICamera* Camera, const sdpVector2 MouseCoords, sdpVector3& RayOrigin, sdpVector3& RayDirection);

/// Redraws all cameras
void k3dRedrawAll(k3dIDocument* Document, bool Synchronous);

/// Applies a functor to every object in the given document ...
template<class Functor>
bool k3dForEachObject(k3dIDocument* Document, gulong Criteria, Functor& ObjectFunctor)
{
	// Sanity checks ...
	g_assert(Document);

	// For each object in the document ...
	k3dIObjectContext* context;
	for(K3DOBJECTCONTEXTHANDLE contexthandle = Document->GetFirstObjectContext(); contexthandle; contexthandle = Document->GetNextObjectContext(contexthandle))
		{
			context = Document->GetObjectContext(contexthandle);
		
			// Skip the object if it doesn't fit our criteria ...
			if(SDP_S_OK != context->State()->QueryState(Criteria))
				continue;
		
			// Apply the functor ...
			if(!ObjectFunctor.HandleObject(context))
				return false;
		}

	return true;
}

/// Applies a functor to every object in the given document, in hierarchy order
template<class Functor>
bool k3dForEachObjectTree(k3dIDocument* Document, gulong Criteria, Functor& ObjectFunctor)
{
	// Sanity checks ...
	g_assert(Document);

	// For each node in the hierarchy tree ...
	K3DTREEHANDLE handle;
	k3dIObjectContext* context;
	for(k3dIDocument::TreeNodeType type = Document->GetFirstObjectContextTree(&handle); k3dIDocument::COMPLETE != type; type = Document->GetNextObjectContextTree(&handle))
		{
			context = Document->GetObjectContextTree(handle);
			
			// Skip the object if it doesn't fit our criteria ...
			if(SDP_S_OK != context->State()->QueryState(Criteria))
				continue;
			
			// Apply the functor ...
			if(!ObjectFunctor.HandleObject(type, context))
				return false;
		}

	return true;
}

/// Applies a functor to every child of the given object
template<class Functor>
bool k3dForEachChild(k3dIDocument* Document, k3dIObjectContext* Parent, Functor& ObjectFunctor)
{
	// Sanity checks ...
	g_assert(Document);
	g_assert(Parent);

	// Move through the parent's child records ...
	gulong depth = 0;
	K3DTREEHANDLE handle;
	for(k3dIDocument::TreeNodeType type = Document->GetContextHandleTree(Parent, &handle); ; type = Document->GetNextObjectContextTree(&handle))
		{
			switch(type)
				{
					case k3dIDocument::PUSH:
						depth++;
						break;
						
					case k3dIDocument::NODE:
						if(2 == depth)
							if(!ObjectFunctor.HandleDescendantObject(Document->GetObjectContextTree(handle)))
								return false;
						break;
						
					case k3dIDocument::POP:
						depth--;
						if(0 == depth)
							return true;
						break;
						
					case k3dIDocument::COMPLETE:
					case k3dIDocument::TREE_NODE_ERROR:
						g_return_val_if_fail(0, false);
				}
		}
	
	g_return_val_if_fail(0, false);
}

/// Applies a functor to every descendant of the given object
template<class Functor>
bool k3dForEachDescendant(k3dIDocument* Document, k3dIObjectContext* Parent, Functor& ObjectFunctor)
{
	// Sanity checks ...
	g_assert(Document);
	g_assert(Parent);

	// Move through the parent's child records ...
	gulong depth = 0;
	K3DTREEHANDLE handle;
	for(k3dIDocument::TreeNodeType type = Document->GetContextHandleTree(Parent, &handle); ; type = Document->GetNextObjectContextTree(&handle))
		{
			switch(type)
				{
					case k3dIDocument::PUSH:
						depth++;
						break;
						
					case k3dIDocument::NODE:
						if(depth > 1)
							if(!ObjectFunctor.HandleDescendantObject(Document->GetObjectContextTree(handle)))
								return false;
						break;
						
					case k3dIDocument::POP:
						depth--;
						if(0 == depth)
							return true;
						break;
						
					case k3dIDocument::COMPLETE:
					case k3dIDocument::TREE_NODE_ERROR:
						g_return_val_if_fail(0, false);
				}
		}
	
	g_return_val_if_fail(0, false);
}

/// Applies a functor to every mesh in the given document
template<class Functor>
bool k3dForEachMesh(k3dIDocument* Document, gulong Criteria, Functor& MeshFunctor)
{
	// Sanity checks ...
	g_assert(Document);

	// For each object in the document ...
	k3dIMesh* mesh;
	k3dIObjectContext* context;
	for(K3DOBJECTCONTEXTHANDLE contexthandle = Document->GetFirstObjectContext(); contexthandle; contexthandle = Document->GetNextObjectContext(contexthandle))
		{
			// Get the context ...
			context = Document->GetObjectContext(contexthandle);

			// Skip the object if it doesn't fit our criteria ...
			if(SDP_S_OK != context->State()->QueryState(Criteria))
				continue;

			// See if the object is a mesh ...
			mesh = sdpQueryInterface<k3dIMesh>(context->Object());
			if(mesh)
				{
					if(!MeshFunctor.HandleMesh(mesh))
						return false;
				}
		}

	return true;
}

/// Applies a functor to every mesh/point in the given document
template<class Functor>
bool k3dForEachMeshPoint(k3dIDocument* Document, gulong Criteria, Functor& MeshFunctor)
{
	// Sanity checks ...
	g_assert(Document);

	k3dIObjectContext* context;
	k3dIMesh* mesh;
	k3dIPoint* point;
	k3dIPointServer* points;
	K3DPOINTHANDLE pointhandle;
	K3DPOINTHANDLE nextpointhandle;

	// For each object in the document ...
	for(K3DOBJECTCONTEXTHANDLE contexthandle = Document->GetFirstObjectContext(); contexthandle; contexthandle = Document->GetNextObjectContext(contexthandle))
		{
			// Get the context ...
			context = Document->GetObjectContext(contexthandle);

			// See if the object is a mesh ...
			mesh = sdpQueryInterface<k3dIMesh>(context->Object());
			if(mesh)
				{
					if(!MeshFunctor.HandleMesh(mesh))
						return false;

					// For each point ...
					points = mesh->Points();
					for(pointhandle = points->GetFirstPoint(); pointhandle; pointhandle = nextpointhandle)
						{
							nextpointhandle = points->GetNextPoint(pointhandle);

							point = points->GetPoint(pointhandle);

							// Skip the point if it doesn't fit our criteria ...
							if(SDP_S_OK != point->State()->QueryState(Criteria))
								continue;

							if(!MeshFunctor.HandleMeshPoint(mesh, point))
								return false;
						}
				}
		}

	return true;
}

/// Applies a functor to every mesh/polygon in the given document
template<class Functor>
bool k3dForEachMeshPolygon(k3dIDocument* Document, gulong Criteria, Functor& MeshFunctor)
{
	// Sanity checks ...
	g_assert(Document);

	k3dIObjectContext* context;
	k3dIMesh* mesh;
	k3dIPolygon* polygon;
	k3dIPolygonServer* polygons;
	K3DPOLYGONHANDLE polygonhandle;
	K3DPOLYGONHANDLE nextpolygonhandle;

	// For each object in the document ...
	for(K3DOBJECTCONTEXTHANDLE contexthandle = Document->GetFirstObjectContext(); contexthandle; contexthandle = Document->GetNextObjectContext(contexthandle))
		{
			// Get the context ...
			context = Document->GetObjectContext(contexthandle);

			// See if the object is a mesh ...
			mesh = sdpQueryInterface<k3dIMesh>(context->Object());
			if(mesh)
				{
					if(!MeshFunctor.HandleMesh(mesh))
						return false;

					// For each polygon ...
					polygons = mesh->Polygons();
					for(polygonhandle = polygons->GetFirstPolygon(); polygonhandle; polygonhandle = nextpolygonhandle)
						{
							nextpolygonhandle = polygons->GetNextPolygon(polygonhandle);

							polygon = polygons->GetPolygon(polygonhandle);

							// Skip the polygon if it doesn't fit our criteria ...
							if(SDP_S_OK != polygon->State()->QueryState(Criteria))
								continue;

							if(!MeshFunctor.HandleMeshPolygon(mesh, polygon))
								return false;
						}
				}
		}

	return true;
}

/// Applies a functor to every point in the given document
template<class Functor>
bool k3dForEachPoint(k3dIDocument* Document, gulong Criteria, Functor& PointFunctor)
{
	// Sanity checks ...
	g_assert(Document);

	// Declare some stuff up front for efficiency ...
	k3dIObjectContext* context;
	k3dIPointServer* pointserver;
	K3DPOINTHANDLE pointhandle;
	k3dIPoint* point;

	// For each object in the document ...
	for(K3DOBJECTCONTEXTHANDLE objecthandle = Document->GetFirstObjectContext(); objecthandle; objecthandle = Document->GetNextObjectContext(objecthandle))
		{
			context = Document->GetObjectContext(objecthandle);

			// Skip it if it doesn't contain points ...
			pointserver = sdpQueryInterface<k3dIPointServer>(context->Object());
			if(pointserver)
				{
					// For each point in the object ...
					for(pointhandle = pointserver->GetFirstPoint(); pointhandle; pointhandle = pointserver->GetNextPoint(pointhandle))
						{
							point = pointserver->GetPoint(pointhandle);

							// Skip points that don't match our criteria ...
							if(SDP_S_OK != point->State()->QueryState(Criteria))
								continue;

							// Apply the user function to the point ...
							if(!PointFunctor.HandlePoint(point))
								return false;
						}
				}
		}

	return true;
}

/// Applies a functor to every polygon in the given document
template<class Functor>
bool k3dForEachPolygon(k3dIDocument* Document, gulong Criteria, Functor& PolyFunctor)
{
	// Sanity checks ...
	g_assert(Document);

	// Declare some stuff up front for efficiency ...
	k3dIObjectContext* context;
	k3dIPolygonServer* polygonserver;
	K3DPOLYGONHANDLE polygonhandle;
	k3dIPolygon* polygon;

	// For each object in the document ...
	for(K3DOBJECTCONTEXTHANDLE objecthandle = Document->GetFirstObjectContext(); objecthandle; objecthandle = Document->GetNextObjectContext(objecthandle))
		{
			context = Document->GetObjectContext(objecthandle);

			// Skip it if it doesn't contain polygons ...
			polygonserver = sdpQueryInterface<k3dIPolygonServer>(context->Object());
			if(polygonserver)
				{
					// For each polygon ...
					for(polygonhandle = polygonserver->GetFirstPolygon(); polygonhandle; polygonhandle = polygonserver->GetNextPolygon(polygonhandle))
						{
							polygon = polygonserver->GetPolygon(polygonhandle);

							// Skip polygons that don't match our criteria ...
							if(SDP_S_OK != polygon->State()->QueryState(Criteria))
								continue;

							// Apply the user function to the polygon ...
							if(!PolyFunctor.HandlePolygon(polygon))
								return false;
						}
				}
		}

	return true;
}

#endif // K3DMISC_H
