#ifndef K3DNURBSCURVE_H
#define K3DNURBSCURVE_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the k3dNURBSCurve class, which encapsulates a NURBS curve of arbitrary degree and configuration
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include <k3dinurbscurve.h>
#include <k3dipolypointserver.h>
#include <k3dilocationserver.h>
#include <k3diknotserver.h>
#include <k3dstate.h>
#include <k3dactionclient.h>

#include <sdpmath.h>

// Forward class declarations
class k3dMesh;
class k3dPolyPoint;

///////////////////////////////////////////////////////////////////////////////////////////
// k3dNURBSCurve

/// Encapsulates a NURBS curve
class k3dNURBSCurve : public k3dINURBSCurve,
										public k3dIPolyPointServer,
										public k3dILocationServer,
										public k3dIKnotServer,
										public k3dState,
										public k3dActionClient
{
public:
	k3dNURBSCurve(k3dMesh* Mesh, k3dID ID);
	virtual ~k3dNURBSCurve();

	// k3dIUnknown overrides ...
	sdpResult QueryInterface(const sdpIID& InterfaceID, void** Interface) { return SDP_E_NO_INTERFACE; }
	gulong AddRef() { return 0; }
	gulong Release() { return 0; }

	// k3dINURBSCurve overrides ...
	k3dIMesh* Mesh();
	k3dIPolyPointServer* Points();
	k3dILocationServer* Locations();
	k3dIKnotServer* Knots();
	k3dIState* State();
	gulong Order();
	const k3dID ID();

	// k3dIPolyPointServer overrides ...
	gulong GetPointCount();
	K3DPOLYPOINTHANDLE GetFirstPolyPoint();
	K3DPOLYPOINTHANDLE GetNextPolyPoint(K3DPOLYPOINTHANDLE Point);
	k3dIPolyPoint* GetPolyPoint(K3DPOLYPOINTHANDLE Point);
	k3dIPolyPoint* AddPoint(k3dIPoint* Point);
	sdpResult RemovePoint(k3dIPoint* Point);
	sdpResult RemovePolyPoint(k3dIPolyPoint* PolyPoint);

	// k3dILocationServer overrides ...
	gulong GetLocationCount();
	K3DLOCATIONHANDLE GetFirstLocation();
	K3DLOCATIONHANDLE GetNextLocation(K3DLOCATIONHANDLE Location);
	k3dILocation* GetLocation(K3DLOCATIONHANDLE Location);

	// k3dIKnotServer overrides ...
	gulong GetKnotCount();
	K3DKNOTHANDLE GetFirstKnot();
	K3DKNOTHANDLE GetNextKnot(K3DKNOTHANDLE Knot);
	k3dIKnot* GetKnot(K3DKNOTHANDLE Knot);

	// k3dIState overrides ...
	sdpResult SetState(gulong Mask, bool State);

	// k3dIActionClient overrides ...
	sdpResult HandleStep(gulong ValueCode, gulong Value, gulong ValueSize);

	// k3dIPersist overrides ...
	sdpResult Save(sdpXmlIElement* Element, k3dIDependencies* Dependencies);
	sdpResult Load(sdpXmlIElement* Element, k3dMesh* Mesh);
	sdpResult LoadComplete();

	/// Draws a wireframe representation of the curve
	void OpenGLWireframe();
	/// Draws a wireframe representation of the curve with orientation
	void OpenGLWireframeOrientation(sdpVector3& FirstColor, sdpVector3& Color);

	/// Stores the curve points attached to this curve
	typedef std::vector<k3dPolyPoint*> PolyPoints;
	typedef PolyPoints::iterator PolyPointIterator;
	PolyPoints m_PolyPoints;

	/// Stores the parent mesh that owns this curve
	k3dMesh* m_Mesh;
	/// Stores the order (degree+1) of the curve
	gulong m_Order;
	/// Stores the unique (within the parent mesh) curve ID
	k3dID m_ID;
	
	/// Enumerates undo-able steps
	typedef enum
	{
		STEP_STATE,
	} StepType;
};

#endif K3DNURBSCURVE_H
