#ifndef K3DSHADERARGUMENTSPAGE_H
#define K3DSHADERARGUMENTSPAGE_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the k3dShaderArgumentsPage class, which provides a UI for editing a set of shader arguments
		\author Tim Shead (tshead@k-3d.com)
*/

#include <sdpgtkobjectcontainer.h>

#include <k3dobjectselector.h>
#include <k3dshaderinstance.h>
#include <k3dchannelbutton.h>

#include <list>

///////////////////////////////////////////////////////////////////////////
// k3dShaderArgumentsPage

/// Manages an instance of a shader and its arguments
class k3dShaderArgumentsPage : sdpGtkObjectContainer
{
public:
	/////////////////////////////////////////////////////////////////////////
	// ButtonBase

	/// Base class for a button that represents one shader argument
	class ButtonBase
	{
	public:
		/// Creates buttons based on argument type
		static ButtonBase* CreateButton(k3dIDocument* Document, k3dShaderInstance::ArgumentBase* Argument);

		ButtonBase(k3dIDocument* Document);
		virtual ~ButtonBase();

		/// Creates the appropriate GTK+ widget(s)
		virtual bool Create(GtkWidget* Parent) = 0;
		
	protected:
		k3dIDocument* m_Document;
	};

	/////////////////////////////////////////////////////////////////////////
	// ChannelButton

	/// Base class for a button that represents any value stored as a channel
	class ChannelButton : public ButtonBase
	{
	public:
		ChannelButton(k3dIDocument* Document, k3dIChannel* Channel);
		virtual ~ChannelButton();

		/// Creates the appropriate GTK+ widget
		bool Create(GtkWidget* Parent);

	protected:
		k3dChannelButton m_Button;
		k3dIChannel* m_Channel;
	};

	/////////////////////////////////////////////////////////////////////////
	// TextureButton

	/// Base class for a button that represents texture argument
	class TextureButton : public ButtonBase
	{
	public:
		TextureButton(k3dIDocument* Document, k3dObjectVariable<k3dITexture>* ObjectContainer);
		virtual ~TextureButton();

		/// Creates the appropriate GTK+ widget
		bool Create(GtkWidget* Widget);

	protected:
		k3dObjectSelector m_Selector;
		k3dObjectVariable<k3dITexture>* m_ObjectContainer;
	};


	
	k3dShaderArgumentsPage();
	~k3dShaderArgumentsPage();

	/// Creates all the widgets needed to edit the given shader's arguments, and attaches them to the parent widget
	bool Create(k3dIDocument* Document, k3dShaderInstance* ShaderInstance, GtkContainer* ParentWidget);
	/// Updates all UI controls to reflect the current shader
	void Update();

protected:
	/// Frees all allocations and cleans-up internal state for re-use
	void Release();

	/// Handles custom events
	void OnEvent(sdpGtkEvent* Event);
	
	/// Called when the GTK+ widgets are about to be destroyed
	void OnDestroy();

	/// Stores our parent document
	k3dIDocument* m_Document;
	/// Stores the shader instance we're attached to
	k3dShaderInstance* m_ShaderInstance;
	/// Stores the widget we're attached to
	GtkContainer* m_ParentWidget;
	/// Declares a list of argument buttons
	typedef std::list<ButtonBase*> Buttons;
	typedef Buttons::iterator ButtonIterator;
	/// Stores buttons for each argument
	Buttons m_Buttons;
};

#endif // K3DSHADERARGUMENTSPAGE_H
