#ifndef K3DTRIANGLEMESH_H
#define K3DTRIANGLEMESH_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the k3dTriangleMesh class, which efficiently encapsulates an implicit/procedural surface composed of triangles
		\author Timothy M. Shead (tshead@k-3d.com)
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include <k3dtransformobject.h>
#include <k3dlocationserver.h>
#include <k3dvariable.h>

#include <k3dimaterialserver.h>

#include <sdpgl.h>

/////////////////////////////////////////////////////////////////////////////////////
// k3dTriangleMesh

/// Efficiently encapsulates an implicit/procedural surface composed of triangles
class k3dTriangleMesh : public k3dTransformObject,
											public k3dLocationServer,
											public k3dIMaterialServer
{
public:
	k3dTriangleMesh();
	~k3dTriangleMesh();

	// k3dIEditorRender overrides ...
	sdpResult EditorRender(k3dIEditorEngine* Engine);

	// k3dIRiRender overrides ...
	sdpResult RiRender(k3dIRenderFrame* Frame, k3dIRiEngine* Engine);

	// k3dIMaterialServer overrides ...
	k3dIMaterial* GetMaterial();
	sdpResult SetMaterial(k3dIMaterial* Material);

protected:
	//////////////////////////////////////////////////////////////////////////////////////////
	// k3dTriangle
	
	class k3dTriangle
	{
	public:
		void SetPoints(gulong Point1, gulong Point2, gulong Point3)
		{
			m_Points[0] = Point1;
			m_Points[1] = Point2;
			m_Points[2] = Point3;
		}
	
		/// Store vertex indices
		gulong m_Points[3];
		
		/// Stores computed normal vector
		sdpVector3 m_Normal;
	};
		
	//////////////////////////////////////////////////////////////////////////////////////////
	// k3dTriangleMesh
	
	class k3dTriangleStrip
	{
	public:
		void AppendPoint(gulong Point)
		{
			m_Points.push_back(Point);
		}
		
		// Store vertex indices
		std::vector<gulong> m_Points;
	};
	
	// Declare a container for our triangles ...
	typedef std::vector<k3dTriangle> TriangleArray;
	typedef TriangleArray::iterator TriangleIterator;
	TriangleArray m_Triangles;
	
	// Triangle access ...
	void ReserveTriangles(gulong TriangleCount);
	k3dTriangle& Triangle(gulong Index) { return  m_Triangles[Index]; }

	// Declare a container for our triangle strips ...
	typedef std::vector<k3dTriangleStrip> TriangleStripArray;
	typedef TriangleStripArray::iterator TriangleStripIterator;
	TriangleStripArray m_TriangleStrips;
		
	// Triangle-strip access ...
	void ReserveTriangleStrips(gulong TriangleStripCount);
	k3dTriangleStrip& TriangleStrip(gulong Index) { return m_TriangleStrips[Index]; }

	/// Disable smooth-shaded triangles
	void SetFlatShading();
	
	/// Enable smooth-shaded triangles
	void SetSmoothShading();

private:	
	// Patch lookup ...
	static short m_RiBilinearPatchLookup[4];

	/// Stores the color for drawing selected geometry
	static sdpVector3 m_SelectedColor;
	/// Stores the default color for drawing unselected geometry
	static sdpVector3 m_DefaultColor;

	/// Stores the color for drawing unselected geometry
	k3dVariable<sdpVector3> m_Color;

	/// Stores our assigned material
	k3dObjectVariable<k3dIMaterial> m_Material;

	/// Editor rendering in wireframe mode
	void EditorRenderWireframe(k3dIEditorEngine* Engine);
	/// Editor rendering in solid mode
	void EditorRenderSolid(k3dIEditorEngine* Engine);
	/// Editor rendering in solid mode with flat shading
	void EditorRenderSolidFlat(k3dIEditorEngine* Engine);
	/// Editor rendering in solid mode with smooth shading
	void EditorRenderSolidSmooth(k3dIEditorEngine* Engine);
	/// Editor rendering in solid, textured mode
	void EditorRenderSolidTextured(k3dIEditorEngine* Engine);
	/// Editor rendering in solid, textured mode with flat shading
	void EditorRenderSolidTexturedFlat(k3dIEditorEngine* Engine);
	/// Editor rendering in solid, textured mode with smooth shading
	void EditorRenderSolidTexturedSmooth(k3dIEditorEngine* Engine);
	/// Editor rendering of points
	void EditorRenderPoints(k3dIEditorEngine* Engine);
	
	/// RI rendering
	void RiTriangle(k3dIRiEngine* Engine, k3dTriangle& Triangle);
	
	/// Use flat or smooth shading while rendering?
	bool m_SmoothShading;
	
	/// Calculate per-polygon normals for flat shading
	void CalculateTriangleNormals();
	/// Calculate per-location normals for smooth shading
	void CalculateLocationNormals();
};

#endif // K3DTRIANGLEMESH_H
