#ifndef K3DVARIABLE_H
#define K3DVARIABLE_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares objects for storing and manipulating serializable data in a standardized way
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dcom.h>
#include <k3dichannel.h>
#include <k3diobject.h>
#include <k3deventclient.h>

#include <sdpxmlutility.h>

// Forward class declarations ...
class sdpXmlIElement;
class k3dIDocument;
class k3dIDependencies;
class k3dIEventClient;

///////////////////////////////////////////////////////////
// k3dVariableBase

/// Pure virtual base class for variable objects that stores the variable name and can save/load the variable
class k3dVariableBase
{
public:
	k3dVariableBase() {}
	virtual ~k3dVariableBase() {}

	/// Returns the variable name
	SDPCSTRING Name() { return m_Name; }
	/// Sets the variable name
	void SetName(SDPCSTRING Name) { m_Name = Name; }

	/// Saves the variable to a K-3D XML document output stream
	virtual sdpResult Save(sdpXmlIElement* Element, k3dIDependencies* Dependencies) = 0;
	/// Loads the variable from a K-3D XML document
	virtual sdpResult Load(sdpXmlIElement* Element) = 0;

protected:
	/// Stores the human-readable name of this variable
	sdpString m_Name;
};

///////////////////////////////////////////////////////////
// k3dCustomVariable

/// Provides type-safe access to a variable that stores a custom type
/** Derive from this template and implement ConvertToString() & ConvertFromString()
		to create a variable object that stores your custom type
*/
template<class Type>
class k3dCustomVariable : public k3dVariableBase
{
public:
	k3dCustomVariable() {}

	/// Initializes the variable's name and value in one step
	k3dCustomVariable<Type>& operator()(SDPCSTRING Name, const Type& Value)
	{
		SetName(Name);
		SetValue(Value);
		return *this;
	}
	
	/// Sets the variable's contents to the given value
	void SetValue(const Type& Value) { m_Value = Value; }
	/// Returns the variable's contents by value
	Type Value() { return m_Value; }
	/// Accesses the variable's contents by reference
	Type& operator()() { return m_Value; }

	/// Saves the variable to a K-3D XMLdocument output stream
	sdpResult Save(sdpXmlIElement* Element, k3dIDependencies* Dependencies)
	{
		// Sanity checks ...
		g_assert(Element);
		g_assert(Dependencies);

		sdpXmlCreateElement(Element, "variable", 0, "name", m_Name, "value", ConvertToString());

		return SDP_S_OK;
	}

	/// Loads the variable from a K-3D XML document
	sdpResult Load(sdpXmlIElement* Element)
	{
		// Sanity checks ...
		g_assert(Element);
		g_assert(Element->Name() == sdpString("variable"));

		// Pull out the name ...
		sdpString name;
		g_return_val_if_fail(sdpXmlGetAttribute(Element, "name", name), SDP_E_FAIL);
		g_assert(name == m_Name);

		// Pull out the value ...
		sdpString value;
		g_return_val_if_fail(sdpXmlGetAttribute(Element, "value", value), SDP_E_FAIL);
	
		// Convert it ...
		g_return_val_if_fail(ConvertFromString(value), SDP_E_FAIL);

		return SDP_S_OK;
	}

	/// Converts the variable's contents to a string format for saving
	virtual sdpString ConvertToString() = 0;
	/// Loads the variable's contents from a string format for loading
	virtual bool ConvertFromString(SDPCSTRING String) = 0;

protected:
	/// Stores the contents of the variable
	Type m_Value;
};

///////////////////////////////////////////////////////////
// k3dVariable

/// Provides type-safe access to one of the standard variable types supported by sdpString
/** Use this template to create variable objects whose types can be converted to-and-from
		strings using sdpString's constructor and getvalue() syntax.
*/
template<class Type>
class k3dVariable : public k3dCustomVariable<Type>
{
public:
	/// Converts the variable contents to a string using and sdpString overloaded constructor
	virtual sdpString ConvertToString() { return sdpString(m_Value); }
	/// Converts the variable contents from a string using sdpString::getvalue()
	virtual bool ConvertFromString(SDPCSTRING String) { sdpString string(String); return string.getvalue(m_Value); }
};

/////////////////////////////////////////////////////////////
// k3dObjectVariableBase

/// A variable object that stores a reference to another K-3D object
class k3dObjectVariableBase : public k3dVariableBase,
															public k3dEventClient
{
public:
	k3dObjectVariableBase();
	~k3dObjectVariableBase();

	/// Overrides the default base-class behaviour for saving the variable
	virtual sdpResult Save(sdpXmlIElement* Element, k3dIDependencies* Dependencies);
	/// Overrides the default base-class behaviour for loading the variable
	virtual sdpResult Load(sdpXmlIElement* Element);

	/// Attaches an existing K-3D object to the variable
	virtual sdpResult AttachObject(sdpIUnknown* Object);
	/// Frees any existing references to a K-3D object
	virtual void ReleaseObject();

	/// Returns a k3dIObject interface to the attached object
	k3dIObject* Object() { return m_Object; }

	/// Returns the object ID of the attached object
	k3dID ObjectID() { return m_ObjectID; }

	/// Sets whether the attached object should be destroyed by the variable's destructor
	void SetAutoDestroy(bool AutoDestroy) { m_AutoDestroy = AutoDestroy; }
	/// Returns whether the attached object will be destroyed by the variable's destructor
	bool AutoDestroy() { return m_AutoDestroy; }

	/// Creates a new K-3D object using the variable's sdpCom ClassID and attaches it
	virtual bool Create(k3dIDocument* Document, sdpCLSID ClassID, SDPCSTRING Name, bool Edit, bool Show, bool Saveable);

protected:
	// sdpIUnknown overrides ...
	sdpResult QueryInterface(const sdpIID& InterfaceID, void** Interface) { return SDP_E_FAIL; }
	gulong AddRef() { return 1; }
	gulong Release() { return 0; }

	// k3dIEventClient overrides ...
	sdpResult OnEvent(k3dIObject* Object, SDPCSTRING Event, gulong ClientData);

	/// Stores the attached object's k3dIObject interface
	k3dIObject* m_Object;
	/// Stores the attached objects' ID
	k3dID m_ObjectID;
	/// Stores whether the attached object should be destroyed by the variable destructor
	bool m_AutoDestroy;
};

/////////////////////////////////////////////////////////////
// k3dObjectVariable

/// Variable object that references an object in a type-safe manner using the given interface
template<class InterfaceClass>
class k3dObjectVariable : public k3dObjectVariableBase
{
public:
	k3dObjectVariable() : k3dObjectVariableBase() { m_Interface = 0; }

	/// Initializes the variable's name and sdpCom ClassID in one step
	k3dObjectVariable<InterfaceClass>& operator()(SDPCSTRING Name)
	{
		SetName(Name);

		return *this;
	}

	/// Overrides the base-class Create() to obtain an interface pointer of the correct type to the new K-3D object
	bool Create(k3dIDocument* Document, sdpCLSID ClassID, SDPCSTRING Name, bool Edit, bool UserVisible, bool Saveable)
	{
		// Sanity checks ...
		g_assert(Document);

		g_return_val_if_fail(k3dObjectVariableBase::Create(Document, ClassID, Name, Edit, UserVisible, Saveable), false);

		m_Interface = sdpQueryInterface<InterfaceClass>(m_Object);
		g_return_val_if_fail(m_Interface, false);

		return true;
	}

	/// Overrides the base-class AttachObject() to obtain our extra interface pointer
	sdpResult AttachObject(sdpIUnknown* Object)
	{
		// Call the default implementation ...
		g_return_val_if_fail(sdpSucceeded(k3dObjectVariableBase::AttachObject(Object)), SDP_E_FAIL);

		// Get the desired interface ...
		if(m_Object)
			{
				m_Interface = sdpQueryInterface<InterfaceClass>(m_Object);
				g_return_val_if_fail(m_Interface, SDP_E_FAIL);
			}

		return SDP_S_OK;
	}

	/// Overrides the base-class ReleaseObject() so we can release our extra interface pointer
	void ReleaseObject()
	{
		// Call the default implementation ...
		k3dObjectVariableBase::ReleaseObject();

		// Release our extra interface ...
		if(m_Interface)
			{
				m_Interface->Release();
				m_Interface = 0;
			}
	}

	/// Returns the interface to the attached K-3D object
	InterfaceClass* Interface() { return m_Interface; }
	/// Returns the interface to the attached K-3D object (short form)
	InterfaceClass* operator->() { return m_Interface; }	

protected:
	/// Stores a pointer to the desired interface type on the underlying object
	InterfaceClass* m_Interface;
};

//////////////////////////////////////////////////////////////
// k3dChannelVariable

/// A convenience base-class for variables that store a K-3D channel object
class k3dChannelVariable : public k3dObjectVariable<k3dIChannel>
{
public:
	k3dChannelVariable();
};

///////////////////////////////////////////////////////////////
// k3dScalarChannel

/// A convenience class for variables that store a K-3D channel object storing scalar values
class k3dScalarChannel : public k3dChannelVariable
{
public:
	/// Initializes the variable name
	k3dScalarChannel& operator()(SDPCSTRING Name) { SetName(Name); return *this; }
	
	/// Creates the K-3D channel object and initializes it to store scalar values
	bool Create(k3dIDocument* Document, SDPCSTRING Name, double FirstValue, bool ShowTime, double MinY, double MaxY, SDPCSTRING XLabel, SDPCSTRING YLabel, k3dIChannel::DataType Data, bool Saveable, k3dIEventClient* EventClient);
};

///////////////////////////////////////////////////////////////
// k3dColorChannel

/// A convenience class for variables that store a K-3D channel object storing color values
class k3dColorChannel : public k3dChannelVariable
{
public:
	/// Initializes the variable name
	k3dColorChannel& operator()(SDPCSTRING Name) { SetName(Name); return *this; }

	/// Creates the K-3D channel object and initializes it to store color values
	bool Create(k3dIDocument* Document, SDPCSTRING Name, double FirstColor[3], bool ShowTime, SDPCSTRING XLabel, SDPCSTRING YLabel, bool Saveable, k3dIEventClient* EventClient);
};

#endif // K3DVARIABLE_H
