// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dChannelButton class, which provides a UI for representing objects that support k3dIChannel
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dchannelbutton.h>
#include <k3diobject.h>
#include <k3dmisc.h>

#include <sdpgtkevents.h>
#include <sdpgtkutility.h>
#include <sdpgtkcolorselector.h>

#include <sdpxmlutility.h>

#include <sdpmath.h>

#include <iostream>

//////////////////////////////////////////////////////////////////////
// k3dChannelButton

sdpVector3 k3dChannelButton::m_BackgroundColor(0.8, 0.8, 0.8);
sdpVector3 k3dChannelButton::m_NormalColor(0.0, 0.0, 0.0);

k3dChannelButton::k3dChannelButton()
{
//std::cout << "k3dChannelButton::k3dChannelButtion()" << std::endl;

	m_Channel = 0;
	m_ShowCurve = false;
}

k3dChannelButton::~k3dChannelButton()
{
//std::cout << "k3dChannelButton::~k3dChannelButtion()" << std::endl;
}

bool k3dChannelButton::Create(sdpGtkIObjectContainer* ObjectContainer, sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_assert(ObjectContainer);
	g_assert(Element);
	g_assert(Element->Name() == sdpString("k3dchannelbutton"));

	sdpGtkMarkAttribute(Element, "showcurve");
	sdpXmlGetAttribute(Element, "showcurve", m_ShowCurve);

	// Load our template ...
	g_return_val_if_fail(Create(), false);

	return true;
}

bool k3dChannelButton::Create(k3dIChannel* Channel, bool ShowCurve, GtkContainer* Parent)
{
	// Sanity checks ...
	g_assert(Channel);
	g_assert(!m_Channel);
	g_assert(Parent);

	// Store a reference to our channel ...
	m_Channel = Channel;

	// We want to know when the channel is modified ...
	sdpQueryInterface<k3dIObject>(Channel)->AddClient(this, 0, "channel-changed");
	
	// We force ShowCurve on if it's a color ...
	m_ShowCurve = k3dIChannel::DATA_COLOR == m_Channel->GetDatatype() ? true : ShowCurve;

	// Load our template ...
	g_return_val_if_fail(Create(), false);

	// Attach ourselves to the parent widget ...
	sdpGtkContainer(Parent).Attach(RootWidget());

	// Create the value widget ...
	g_return_val_if_fail(CreateValueWidget(), false); 

	// Update the display ...
	Update();

	return true;
}

bool k3dChannelButton::Create()
{
	// Create our UI template ...
	sdpString uitemplate;
	if(m_ShowCurve)
		uitemplate =
			"<gtkml>"
				"<hbox homogeneous=\"false\">"
					"<button expand=\"true\" fill=\"true\" name=\"quickedit\">"
						"<event signal=\"clicked\" name=\"onquickedit\"/>"
						"<eventbox name=\"value\" size=\"16 16\"/>"
					"</button>"
					"<button name=\"edit\">"
						"<event signal=\"clicked\" name=\"onedit\"/>"
						"<arrow direction=\"right\" shadowtype=\"in\"/>"
					"</button>"
				"</hbox>"
			"</gtkml>\n";
	else
		uitemplate =
			"<gtkml>"
				"<hbox homogeneous=\"false\">"
					"<frame shadowtype=\"out\" expand=\"true\" fill=\"true\">"
						"<entry name=\"value\" charactersize=\"8 1.5\">"
							"<event signal=\"focus-in-event\" name=\"onvaluefocusin\"/>"
							"<event signal=\"focus-out-event\" name=\"onvaluefocusout\"/>"
						"</entry>"
					"</frame>"
					"<button name=\"edit\">"
						"<event signal=\"clicked\" name=\"onedit\"/>"
						"<arrow direction=\"right\" shadowtype=\"in\"/>"
					"</button>"
				"</hbox>"
			"</gtkml>\n";

	// Load our template ...
	g_return_val_if_fail(Load(uitemplate, ""), false);
	
	return true;
}

bool k3dChannelButton::CreateValueWidget()
{
	// If the widget already exists, we're done ...
	if(m_ValueWidget.Attached())
		return true;

	// Create the value widget as needed ...
	if(m_ShowCurve)
		{
			sdpGtkContainer container(Container("value"));
			g_return_val_if_fail(m_ValueWidget.Create(container, true, true, 5, 6, 5, 0), false);

			MapEvent("configure-event", "onconfigurevalue", m_ValueWidget, true);
			MapEvent("expose-event", "onexposevalue", m_ValueWidget, true);
		}
		
	return true;
}

bool k3dChannelButton::Connect(k3dIChannel* Channel)
{
	// Sanity checks ...
	g_assert(Channel);
	g_assert(!m_Channel);

	// Store a reference to the channel ...
	m_Channel = Channel;

	// We want to know when the channel is modified ...
	sdpQueryInterface<k3dIObject>(Channel)->AddClient(this, 0, "channel-changed");

	// Create the value widget as needed ...
	g_return_val_if_fail(CreateValueWidget(), false);

	// Update our child controls ...
	Update();

	return true;
}

void k3dChannelButton::ReleaseChannel()
{
	// Don't notify us of any more events ...
	if(m_Channel)
		sdpQueryInterface<k3dIObject>(m_Channel)->RemoveClient(this, 0, "channel-changed");
	
	// Release the channel ...
	m_Channel = 0;
}

const sdpString k3dChannelButton::CustomType() const
{
	return sdpString("k3dChannelButton");
}

void k3dChannelButton::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "onconfigurevalue" || Event->Name() == "onexposevalue")
		OnValueDraw();
	else if(Event->Name() == "onvaluefocusin")
		OnValueFocusIn();
	else if(Event->Name() == "onvaluefocusout")
		OnValueFocusOut();
	else if(Event->Name() == "onedit")
		OnEdit();
	else if(Event->Name() == "onquickedit")
		OnQuickEdit();
	else if(Event->Name() == "ondestroy")
		OnDestroy();
	else
		sdpGtkObjectContainer::OnEvent(Event);
}

void k3dChannelButton::OnValueDraw()
{
	// Make sure we've got a channel, first!
	g_return_if_fail(m_Channel);

	m_ValueWidget.Begin();
	
	// Get the widget sizes ...
	gulong width = m_ValueWidget.Width();
	gulong height = m_ValueWidget.Height();
	glViewport(0, 0, width, height);

	if(k3dIChannel::DATA_COLOR == m_Channel->GetDatatype())
		{
			// Setup the viewing projection ...
			glMatrixMode(GL_PROJECTION);
			glLoadIdentity();
			glOrtho(0, 1, -1, 1, -1, 1);
	
			glMatrixMode(GL_MODELVIEW);
			glLoadIdentity();

			// Setup drawing options ...
			glDisable(GL_DEPTH_TEST);
			glDisable(GL_LIGHTING);
			glShadeModel(GL_FLAT);

			// Make it happen ...
			glBegin(GL_QUADS);
	
			sdpVector3 color;
			gdouble error = 0.5 / gdouble(width);
			gulong granularity = 1;

			gdouble x2;
			gdouble x1 = 0.0;		
			for(gulong i = 0; i < width+granularity; i += granularity)
				{
					x2 = gdouble(i) / gdouble(width);		
		
					m_Channel->GetValue3(x2, error, color);

					glColor3dv(color);
					glVertex2d(x1, -1);
					glVertex2d(x1, 1);
					glVertex2d(x2, 1);
					glVertex2d(x2, -1);
			
					x1 = x2;
				}

			glEnd();
		}
	else
		{
			// Clear the background ...
			glClearColor(m_BackgroundColor[0], m_BackgroundColor[1], m_BackgroundColor[2], 0.0);
			glClear(GL_COLOR_BUFFER_BIT);
			
			// Get the range of node positions from the channel ...
			gdouble min, max;
			m_Channel->GetCurveRange(&min, &max);
			
			// Setup the viewing projection ...
			glMatrixMode(GL_PROJECTION);
			glLoadIdentity();
			glOrtho(0, 1, min, max, -1, 1);
	
			glMatrixMode(GL_MODELVIEW);
			glLoadIdentity();

			// Setup drawing options ...
			glDisable(GL_DEPTH_TEST);
			glDisable(GL_LIGHTING);
			glShadeModel(GL_FLAT);
			glColor3dv(m_NormalColor);	

			// Make it happen ...
			glBegin(GL_LINE_STRIP);

			gdouble x;
			gdouble value;
			gdouble error = 0.5 / gdouble(width);
			gulong granularity = 1;

			for(gulong i = 0; i < width+granularity; i += granularity)
				{
					x = gdouble(i) / gdouble(width);		
					value = m_Channel->GetValue1(x, error);

					glVertex2d(x, value);
				}

			glEnd();
		}
	
	m_ValueWidget.SwapBuffers();
	m_ValueWidget.End();
}

void k3dChannelButton::OnValueFocusIn()
{
	// Make sure we've got a channel, first!
	g_return_if_fail(m_Channel);

	Editable("value").SelectRegion(0, -1);
	Editable("value").SetPosition(-1);
	Editable("value").Activate();
}

void k3dChannelButton::OnValueFocusOut()
{
	// Make sure we've got a channel, first!
	g_return_if_fail(m_Channel);

	// Get the new value from the control ...
	double newvalue = k3dConvertStringDouble(m_Channel->GetDatatype(), Editable("value").GetText(0, -1));

	// Reformat the new value and put it back into the control ...
	Entry("value").SetText(k3dConvertDoubleString(m_Channel->GetDatatype(), newvalue, 3));

	k3dChannelSetConstantValue(m_Channel, newvalue);
}

void k3dChannelButton::OnEdit()
{
	// Make sure we've got a channel, first!
	g_return_if_fail(m_Channel);

	sdpQueryInterface<k3dIObject>(m_Channel)->EditObject();
}

void k3dChannelButton::OnQuickEdit()
{
	// Make sure we've got a channel, first!
	g_return_if_fail(m_Channel);

	switch(m_Channel->GetDatatype())
		{
			case k3dIChannel::DATA_SCALAR:
			case k3dIChannel::DATA_TIME:
			case k3dIChannel::DATA_ANGLE:
			case k3dIChannel::DATA_DISTANCE:
			case k3dIChannel::DATA_AREA:
			case k3dIChannel::DATA_VOLUME:
			case k3dIChannel::DATA_MASS:
			case k3dIChannel::DATA_FORCE:
			case k3dIChannel::DATA_PRESSURE:
				{
					// Tell the channel to open its UI ...
					sdpQueryInterface<k3dIObject>(m_Channel)->EditObject();
					break;
				}

			case k3dIChannel::DATA_COLOR:
				{
					// Get the channel value ...
					sdpVector3 value;
					m_Channel->GetValue3(0.0, 0.1, value);

					// Prompt the user for a new color ...
					sdpGtkColorSelector selector("Select Color", value);
					selector.DoModal();
					if(selector.OK())
						{
							// If the user OK'd the new color, update the channel ...
							value = selector.Color();
							k3dChannelSetConstantValue(m_Channel, value);
							
							// Update our UI ...
							Update();
						}

					break;
				}

			case k3dIChannel::DATA_ROTATIONXYZ:
			case k3dIChannel::DATA_SCALEXYZ:
				g_assert_not_reached();
		}
}

void k3dChannelButton::OnDestroy()
{
	// Disconnect from the channel ...
	ReleaseChannel();

	// Disconnect our OpenGL context ...
	if(m_ValueWidget.Initialized())
		m_ValueWidget.Destroy();
		
	// No more events from this point forward!
	DisconnectAllEvents();

	// Clear out our storage ...
	Clear();
}

void k3dChannelButton::Update()
{
	// Make sure we've got a channel, first!
	g_return_if_fail(m_Channel);

	// If we're showing the curve on the face of the button, force an update ...
	if(m_ShowCurve)
		{
			Widget("value").QueueDraw();
			return;
		}

	// Otherwise, display the new value ...
	switch(m_Channel->GetDatatype())
		{
			case k3dIChannel::DATA_SCALAR:
			case k3dIChannel::DATA_TIME:
			case k3dIChannel::DATA_ANGLE:
			case k3dIChannel::DATA_DISTANCE:
			case k3dIChannel::DATA_AREA:
			case k3dIChannel::DATA_VOLUME:
			case k3dIChannel::DATA_MASS:
			case k3dIChannel::DATA_FORCE:
			case k3dIChannel::DATA_PRESSURE:
				{
					gdouble value = m_Channel->GetValue1(0.0, 0.1);
					Entry("value").SetText(k3dConvertDoubleString(m_Channel->GetDatatype(), value, 3));
					break;
				}

			case k3dIChannel::DATA_COLOR:
			case k3dIChannel::DATA_ROTATIONXYZ:
			case k3dIChannel::DATA_SCALEXYZ:
				g_return_if_fail(0);	// Should never happen!
		}
}

sdpResult k3dChannelButton::OnEvent(k3dIObject* Object, SDPCSTRING Event, gulong ClientData)
{
	if(Event == sdpString("channel-changed"))
		Update();
	else	
		k3dEventClient::OnEvent(Object, Event, ClientData);
		
	return SDP_S_OK;
}

