// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dDialog class, which provides a standard dialog implementation for K-3D objects
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3ddialog.h>
#include <k3duserobject.h>
#include <k3dmisc.h>
#include <k3dchannelbutton.h>
#include <k3dshaderselector.h>

#include <k3diapplication.h>
#include <k3didocument.h>

#include <sdpgtkevents.h>
#include <sdpfile.h>



k3dDialog::k3dDialog()
{
	m_Object = 0;
}

k3dDialog::~k3dDialog()
{
	// No more events from this point forward ...
	DisconnectAllEvents();
	
	// Get rid of any leftover GTK+ objects ...
	if(Root())
		RootWidget().Destroy();
		
	// Cleanup memory ...
	Clear();
}

// UI creation ...
bool k3dDialog::Create(k3dUserObject* Object, bool AutoDestroy, bool ShowObjectName, SDPCSTRING TemplateFile, SDPCSTRING HelpFile)
{
	// Sanity checks ...
	g_assert(Object);
	g_assert_string(TemplateFile);

	// Store data ...
	m_Object = Object;
	m_AutoDestroy = AutoDestroy;
	m_TemplateFile = sdpAppendPath(k3dLibraryPath(Object), TemplateFile);
	
	if(HelpFile)
		m_HelpFile = sdpAppendPath(k3dLibraryPath(Object), HelpFile);
	else
		m_HelpFile = "";

	// If we're already open, make ourselves visible, and we're done ...
	if(Root())
		{
			gdk_window_show(((GtkWidget*)RootWidget())->window);
			gdk_window_raise(((GtkWidget*)RootWidget())->window);
		
			return true;
		}

	// Load our template ...
	g_return_val_if_fail(LoadFile(m_TemplateFile), false);

	// Map some default events ...
	g_return_val_if_fail(MapEvent("delete-event", "ondelete", Root(), true), false);
	g_return_val_if_fail(MapEvent("destroy", "ondestroy", Root(), true), false);
	g_return_val_if_fail(MapEvent("focus-in-event", "onfocusin", Root(), true), false);

	// Create custom controls ...
	g_return_val_if_fail(InitializeControls(), false);
	
	// Initial update of controls ...
	UpdateControls();

	// Update the titlebar ...
	if(ShowObjectName)
		ObjectRenamed(Object->ObjectContext()->Name());

	// Make everything visible ...
	RootWidget().Show();

	return true;
}

bool k3dDialog::InitializeControls()
{
	return true;
}

void k3dDialog::UpdateControls()
{
}

void k3dDialog::FinalizeControls()
{
}

// We have an unfortunate clash with X ...
#ifdef RootWindow
#undef RootWindow
#endif // RootWindow

void k3dDialog::ObjectRenamed(SDPCSTRING NewName)
{
	RootWindow().SetTitle(NewName);
}

void k3dDialog::ObjectDeleted()
{
	// No more events from this point forward ...
	DisconnectAllEvents();
	
	// Cleanup controls
	FinalizeControls();

	// Get rid of any leftover GTK+ objects ...
	if(Root())
		RootWidget().Destroy();
		
	// Cleanup memory ...
	Clear();
}

// Custom event handling ...
void k3dDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "onok")
		OnOK();
	else if(Event->Name() == "oncancel")
		OnCancel();
	else if(Event->Name() == "onfocusin")
		OnSetFocus();
	else if(Event->Name() == "ondelete")
		OnDelete(Event);
	else if(Event->Name() == "ondestroy")
		OnDestroy();
	else if(Event->Name() == "onhelp")
		OnHelp();
	else
		sdpGtkObjectContainer::OnEvent(Event);
}

// Called when the user hits the standard "OK" button ...
void k3dDialog::OnOK()
{
	RootWidget().Destroy();
}

// Called when the user hits the standard "Cancel" button ...
void k3dDialog::OnCancel()
{
	RootWidget().Destroy();
}

// Called when our window receives the focus ...
void k3dDialog::OnSetFocus()
{
	// Sanity checks ...
	g_assert(m_Object);

	// Make our object the active object ...
	m_Object->Document()->SetActiveObjectContext(m_Object->ObjectContext());

}

// Called when the user hits the standard "Help" button ...
void k3dDialog::OnHelp()
{
	// Sanity checks ...
	g_assert(m_Object);

	// Open the help file in the user's browser ...
	m_Object->Application()->BrowserNavigate(HelpFile());
}

// Called when the user hits the WM "close" button ...
void k3dDialog::OnDelete(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	// Don't allow the window to close ... treat it like a cancel, instead ...
	((sdpGtkEventWidgetDeleteEvent*)Event)->SetResult(true);

	OnCancel();
}

// Called when we're going away ...
void k3dDialog::OnDestroy()
{
	// No more events from this point forward ...
	DisconnectAllEvents();

	// Cleanup controls
	FinalizeControls();

	// Get rid of our widget pointers ...
	Clear();

	// If AutoDestroy is enabled, kill our owning object ...
	if(m_AutoDestroy)
		k3dDeleteObject(m_Object);
}

///////////////////////////////////////////////////////////////////////////////////
// Simplified control updates ...

bool k3dDialog::UpdateControl(sdpGtkEntry Control, double& Storage, k3dIChannel::DataType DataType, gulong Precision)
{
	return UpdateControl(Control, Storage, DataType, -DBL_MAX, DBL_MAX, Precision);
}

bool k3dDialog::UpdateControl(sdpGtkEntry Control, double& Storage, k3dIChannel::DataType DataType, double MinValue, double MaxValue, gulong Precision)
{
	// Sanity checks ...
	g_assert(MinValue < MaxValue);

	// Get the new value from the control ...
	double newvalue = k3dConvertStringDouble(DataType, Control.GetText());

	// Clamp the value ...
	newvalue = sdpMax(newvalue, MinValue);
	newvalue = sdpMin(newvalue, MaxValue);

	// Reformat the new value and put it back into the control ...
	Control.SetText(k3dConvertDoubleString(DataType, newvalue, Precision));

	// If the value didn't change we're done ...
	if(newvalue == Storage)
		return false;

	// Store the new value ...
	Storage = newvalue;

	return true;
}

bool k3dDialog::UpdateControl(sdpGtkEntry Control, gulong& Storage, k3dIChannel::DataType DataType)
{
	return UpdateControl(Control, Storage, DataType, 0, ULONG_MAX);
}

bool k3dDialog::UpdateControl(sdpGtkEntry Control, gulong& Storage, k3dIChannel::DataType DataType, gulong MinValue, gulong MaxValue)
{
	// Sanity checks ...
	g_assert(MinValue < MaxValue);

	// Get the new value from the control ...
	double newvalue = k3dConvertStringDouble(DataType, Control.GetText());

	// Clamp the value ...
	newvalue = sdpMax(newvalue, double(MinValue));
	newvalue = sdpMin(newvalue, double(MaxValue));

	// Reformat the new value and put it back into the control ...
	Control.SetText(k3dConvertDoubleString(DataType, newvalue, 0));

	// If the value didn't change we're done ...
	if(newvalue == Storage)
		return false;

	// Store the new value ...
	Storage = gulong(newvalue);

	return true;
}

sdpGtkObject* k3dDialog::CreateCustomObject(sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_assert(Element);
	
	if(Element->Name() == sdpString("k3dchannelbutton"))
		return new k3dChannelButton();
	else if(Element->Name() == sdpString("k3dshaderselector"))
		return new k3dShaderSelector();
	
	return sdpGtkObjectContainer::CreateCustomObject(Element);
}

// Custom control access ...
k3dChannelButton* k3dDialog::ChannelButton(SDPCSTRING Name)
{
	// Sanity checks ...
	g_assert_string(Name);

	sdpGtkObject* object = CustomObject(Name);
	g_return_val_if_fail(object, 0);

	g_return_val_if_fail(object->CustomType() == "k3dChannelButton", 0);

	return static_cast<k3dChannelButton*>(object);
}

k3dShaderSelector* k3dDialog::ShaderSelector(SDPCSTRING Name)
{
	// Sanity checks ...
	g_assert_string(Name);

	sdpGtkObject* object = CustomObject(Name);
	g_return_val_if_fail(object, 0);

	g_return_val_if_fail(object->CustomType() == "k3dShaderSelector", 0);

	return static_cast<k3dShaderSelector*>(object);
}

void k3dDialog::ConnectChannelButton(k3dIChannel* Channel, SDPCSTRING Name)
{
	// Sanity checks ...
	g_assert(Channel);
	g_assert_string(Name);
	
	k3dChannelButton* channelbutton = ChannelButton(Name);
	g_return_if_fail(channelbutton);
	
	channelbutton->Connect(Channel);
}


