
// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dFrameRange class for working with ranges of numbered files
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include <k3dframerange.h>
#include <math.h>
#include <fstream>

k3dFrameRange::k3dFrameRange()
{
	m_FirstFrame = 0;
	m_LastFrame = 0;
	m_FrameDigits = 0;
}

sdpString k3dFrameRange::Frame(gulong FrameNumber)
{
	// Sanity checks ...
	g_assert(m_FrameDigits);

	if(m_Extension.empty())
		return sdpFormatString("%s%.*d%s", SDPCSTRING(m_PreDigits), m_FrameDigits, FrameNumber, SDPCSTRING(m_PostDigits));
	else
		return sdpFormatString("%s%.*d%s.%s", SDPCSTRING(m_PreDigits), m_FrameDigits, FrameNumber, SDPCSTRING(m_PostDigits), SDPCSTRING(m_Extension));
}

gulong k3dFrameRange::MaxFrame()
{
	return gulong(pow(10, m_FrameDigits) - 1);
}

bool k3dFrameRange::Initialize(SDPCSTRING BaseFileName)
{
	// Sanity checks ...
	g_return_val_if_string_fail(BaseFileName, false);

	sdpString buffer = BaseFileName;

	// Get the file extension ...
	m_Extension = buffer.right(buffer.size() - buffer.find_last_of(".") - 1);

	// Strip the file extension from the buffer ...
	buffer = buffer.left(buffer.size() - m_Extension.size() - 1);

	// Get the post-digits string ...
	m_PostDigits = buffer.right(buffer.size() - buffer.find_last_of("0123456789") - 1);

	// Strip the post digits string from the buffer ...
	buffer = buffer.left(buffer.size() - m_PostDigits.size());

	// Get the digits ...
	m_FrameDigits = buffer.size() - buffer.find_first_of("0123456789");

	// Strip the digits from the base filename ...
	m_PreDigits = buffer.left(buffer.size() - m_FrameDigits);

	// Find the first available frame ...
	gulong maxframe = MaxFrame();
	for(gulong i = 0; i < maxframe; i++)
		{
			std::ifstream file(Frame(i), std::ios::in);
			if(file.good())
				{
					m_FirstFrame = i;
					break;
				}
		}

	// Find the last available frame ...
	for(gulong j = m_FirstFrame + 1; j < maxframe; j++)
		{
			std::ifstream file(Frame(j), std::ios::in);
			if(!file.good())
				{
					m_LastFrame = j-1;
					break;
				}
		}

	return true;
}

bool k3dFrameRange::Initialize(SDPCSTRING File, gulong FirstFrame, gulong LastFrame, gulong FrameDigits)
{
	// Sanity checks ...
	g_return_val_if_string_fail(File, false);

	sdpString buffer = File;

	// Get the file extension ...
	sdpString extension = buffer.right(buffer.size() - buffer.find_last_of(".") - 1);

	// Strip the file extension from the buffer ...
	buffer = buffer.left(buffer.size() - extension.size() - 1);

	return Initialize(buffer, "", extension, FirstFrame, LastFrame, FrameDigits);
}

bool k3dFrameRange::Initialize(SDPCSTRING PreDigits, SDPCSTRING PostDigits, SDPCSTRING Extension, gulong FirstFrame, gulong LastFrame, gulong FrameDigits)
{
	// Sanity checks ...
	g_return_val_if_fail(FirstFrame <= LastFrame, false);
	g_return_val_if_fail(FrameDigits, false);

	m_PreDigits = PreDigits;
	m_PostDigits = PostDigits;
	m_Extension = Extension;

	m_FirstFrame = FirstFrame;
	m_LastFrame = LastFrame;
	m_FrameDigits = FrameDigits;

	return true;
}

bool k3dFrameRange::IsInitialized()
{
	return m_FrameDigits ? true : false;
}

