
// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements a kitchen-junk-drawer-ful of convenience functions, templates, and classes
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dilocation.h>	
#include <k3diobjectinfo.h>	
#include <k3diobjectlibrary.h>	
#include <k3dipolypointserver.h>	
#include <k3displinechannel.h>	
#include <k3distate.h>	
#include <k3ditransform.h>	
#include <k3dicamera.h>

#include <k3dmisc.h>

#include <sdpxmlidocument.h>
#include <sdpxmlielement.h>

#include <sdpgl.h>

#include <sdpmath.h>
#include <sdpfile.h>
#include <sdpstring.h>

#include <map>

///////////////////////////////////////////////////////////////////////
// k3dConversionUnit

class k3dConversionUnit
{
public:
	const char* m_Symbol;
	const char* m_Name;
	double m_ConversionFactor;
};

#define UnitCount(Array) sizeof(Array) / sizeof(k3dConversionUnit)

/// An array of angular units-of-measure
static const k3dConversionUnit AngleUnits[] =
{
	{"d", "Degree", 0.01745329252},
	{"deg", "Degree", 0.01745329252},
	{"degree", "Degree", 0.01745329252},
	{"arcmin", "Minute of arc", 0.00290888209},
	{"arcs", "Second of arc", 0.00000484814},
	{"g", "Grade", 0.01570796327},
	{"grad", "Grade", 0.01570796327},
	{"r", "Radian", 1.0},
	{"rad", "Radian", 1.0},
	{"radian", "Radian", 1.0},
};

static const SDPCSTRING AngleDisplay = "deg";

/// An array of time units-of-measure
static const k3dConversionUnit TimeUnits[] =
{
	{"s", "Second", 1.0},
	{"d", "Day", 86400.0},
	{"day", "Day", 86400.0},
	{"h", "Hour", 3600.0},
	{"hour", "Hour", 3600.0},
	{"ms", "Millisecond", 0.001},
	{"msec", "Millisecond", 0.001},
	{"min", "Minute", 60.0},
	{"minute", "Minute", 60.0},
	{"second", "Second", 1.0},
	{"yr", "Year", 31556925.9747},
	{"year", "Year", 31556925.9747},
};

static const SDPCSTRING TimeDisplay = "s";

/// An array of distance units-of-measure
static const k3dConversionUnit DistanceUnits[] =
{
	{"m", "Meter", 1.0},
	{"au", "Astronomical unit", 1.495979e11},
	{"A", "Angstrom", 0.0000000001},
	{"angstrom", "Angstrom", 0.0000000001},
	{"chain", "Chain", 20.1168402337},
	{"fath", "Fathom", 1.82880365761},
	{"fathom", "Fathom", 1.82880365761},
	{"fermi", "Fermi", 1e-15},
	{"ftintl", "International foot", 0.3048},
	{"ft", "Survey foot", 0.304800609601},
	{"foot", "Survey foot", 0.304800609601},
	{"ftUS", "Survey foot", 0.304800609601},
	{"in", "Inch", 0.0254},
	{"inch", "Inch", 0.0254},
	{"ly", "Light year", 9.46052840488e15},
	{"lyr", "Light year", 9.46052840488e15},
	{"lightyear", "Light year", 9.46052840488e15},
	{"Em", "Exameter", 1e18},
	{"Pm", "Petemeter", 1e15},
	{"Tm", "Terameter", 1e12},
	{"Gm", "Gigameter", 1e9},
	{"Mm", "Megameter", 1e6},
	{"Km", "Kilometer", 1e3},
	{"km", "Kilometer", 1e3},
	{"Hm", "Hectometer", 1e2},
	{"hm", "Hectometer", 1e2},
	{"Dm", "Dekameter", 1e1},
	{"meter", "Meter", 1.0},
	{"dm", "Decimeter", 1e-1},
	{"cm", "Centimeter", 1e-2},
	{"mm", "Millimeter", 1e-3},
	{"um", "Micrometer", 1e-6},
	{"nm", "Nanometer", 1e-9},
	{"pm", "Picometer", 1e-12},
	{"fm", "Femtometer", 1e-15},
	{"am", "Attometer", 1e-18},
	{"miIntl", "International mile", 1609.344},
	{"mil", "Mil", 0.0000254},
	{"mi", "US statute mile", 1609.34721869},
	{"mile", "US statute mile", 1609.34721869},
	{"mileUS", "US statute mile", 1609.34721869},
	{"nmi", "Nautical mile", 1852.0},
	{"pc", "Parsec", 3.08567818585e16},
	{"parsec", "Parsec", 3.08567818585e16},
	{"rd", "Rod", 5.02921005842},
	{"rod", "Rod", 5.02921005842},
	{"yd", "Yard", 0.9144},
	{"yard", "Yard", 0.9144},
};

static const SDPCSTRING DistanceDisplay = "m";

/// An array of area units-of-measure
static const k3dConversionUnit AreaUnits[] =
{
	{"m^2", "Square meter", 1.0},
	{"a", "Are", 100.0},
	{"acre", "Acre", 4046.87260987},
	{"b", "Barn", 1e-28},
};

static const SDPCSTRING AreaDisplay = "m^2";

/// An array of volume units-of-measure
static const k3dConversionUnit VolumeUnits[] =
{
	{"m^3", "Cubic meter", 1.0},
	{"bbl", "Barrel, oil", 0.158987294928},
	{"bu", "Bushel", 0.03523907},
	{"cu", "US cup", 2.365882365e-4},
	{"fbm", "Board foot", 0.002359737216},
	{"gal", "US gallon", 0.003785411784},
	{"galUS", "US gallon", 0.003785411784},
	{"galC", "Canadian gallon", 0.00454609},
	{"galUK", "UK gallon", 0.004546092},
	{"l", "Liter", 0.001},
	{"ozfl", "US fluid ounce", 2.95735295625e-5},
	{"ozUK", "UK fluid ounce", 0.000028413075},
	{"pk", "Peck", 0.0088097675},
	{"pt", "Pint", 0.000473176473},
	{"pint", "Pint", 0.000473176473},
	{"qt", "Quart", 0.000946352946},
	{"quart", "Quart", 0.000946352946},
	{"st", "Stere", 1.0},
	{"stere", "Stere", 1.0},
	{"tbsp", "Tablespoon", 1.47867647813e-5},
	{"tsp", "Teaspoon", 4.92892159375e-6},
};

static const SDPCSTRING VolumeDisplay = "m^3";

/// An array of mass units-of-measure
static const k3dConversionUnit MassUnits[] =
{
	{"Kg", "Kilogram", 1.0},
	{"ct", "Carat", 0.0002},
	{"g", "Gram", 0.001},
	{"gram", "Gram", 0.001},
	{"grain", "Grain", 0.00006479891},
	{"lb", "Avoirdupois pound", 0.45359237},
	{"lbt", "Troy pound", 0.3732417},
	{"oz", "Ounce", 0.028349523125},
	{"ounce", "Ounce", 0.028349523125},
	{"ozt", "Troy ounce", 0.031103475},
	{"slug", "Slug", 14.5939029372},
	{"t", "Metric ton", 1000.0},
	{"ton", "Short ton", 907.18474},
	{"tonUK", "Long ton", 1016.0469088},
	{"u", "Unified atomic mass", 1.66057e-27},
};

static const SDPCSTRING MassDisplay = "Kg";

/// An array of force units-of-measure
static const k3dConversionUnit ForceUnits[] =
{
	{"N", "Newton", 1.0},
	{"dyn", "Dyne", 0.00001},
	{"dyne", "Dyne", 0.00001},
	{"gf", "Gram-force", 0.00980665},
	{"kip", "Kilopound-force", 4448.22161526},
	{"lbf", "Point-force", 4.44822161526},
	{"Newton", "Newton", 1.0},
	{"pdl", "Poundal", 0.138254954376},
};

static const SDPCSTRING ForceDisplay = "N";

/// An array of pressure units-of-measure
static const k3dConversionUnit PressureUnits[] =
{
	{"Pa", "Pascal", 1.0},
	{"atm", "Atmosphere", 101325.0},
	{"bar", "Bar", 100000.0},
	{"inHG", "Inch of mercury", 3386.38815789},
	{"inH2O", "Inch of water", 248.84},
	{"mmHG", "Millimeter of mercury", 133.322368421},
	{"Pascal", "Pascal", 1.0},
	{"psi", "Pound per square inch", 6894.75729317},
	{"torr", "Torr", 133.322368421},
};

static const SDPCSTRING PressureDisplay = "Pa";

///////////////////////////////////////////////////////////////////////
// k3dConversionUnitDict

typedef std::map<sdpString, k3dConversionUnit> k3dConversionUnitMap;
typedef k3dConversionUnitMap::iterator k3dConversionUnitIterator;

class k3dConversionUnitDict : public k3dConversionUnitMap
{
public:
	k3dConversionUnitDict(const k3dConversionUnit Units[], gulong Quantity, SDPCSTRING Display);
};

k3dConversionUnitDict::k3dConversionUnitDict(const k3dConversionUnit Units[], gulong Quantity, SDPCSTRING Display)
{
	// Sanity checks ...
	g_assert_string(Display);

	// Add the units to the dictionary ...
	for(gulong i = 0; i < Quantity; i++)
		insert(k3dConversionUnitDict::value_type(Units[i].m_Symbol, Units[i]));

	// Insert an extra entry for the display unit of measure ...
	insert(k3dConversionUnitDict::value_type("display", (*this)[Display]));
}

/// Mapped dictionary of angular units-of-measure
static k3dConversionUnitDict AngleDict(AngleUnits, UnitCount(AngleUnits), AngleDisplay);
/// Mapped dictionary of time units-of-measure
static k3dConversionUnitDict TimeDict(TimeUnits, UnitCount(TimeUnits), TimeDisplay);
/// Mapped dictionary of distance units-of-measure
static k3dConversionUnitDict DistanceDict(DistanceUnits, UnitCount(DistanceUnits), DistanceDisplay);
/// Mapped dictionary of area units-of-measure
static k3dConversionUnitDict AreaDict(AreaUnits, UnitCount(AreaUnits), AreaDisplay);
/// Mapped dictionary of volume units-of-measure
static k3dConversionUnitDict VolumeDict(VolumeUnits, UnitCount(VolumeUnits), VolumeDisplay);
/// Mapped dictionary of mass units-of-measure
static k3dConversionUnitDict MassDict(MassUnits, UnitCount(MassUnits), MassDisplay);
/// Mapped dictionary of force units-of-measure
static k3dConversionUnitDict ForceDict(ForceUnits, UnitCount(ForceUnits), ForceDisplay);
/// Mapped dictionary of pressure units-of-measure
static k3dConversionUnitDict PressureDict(PressureUnits, UnitCount(PressureUnits), PressureDisplay);

static k3dConversionUnitDict& k3dConversionUnitDictionary(k3dIChannel::DataType Data)
{
	switch(Data)
		{
			case k3dIChannel::DATA_TIME:
				return TimeDict;
			case k3dIChannel::DATA_ANGLE:
				return AngleDict;
			case k3dIChannel::DATA_DISTANCE:
				return DistanceDict;
			case k3dIChannel::DATA_AREA:
				return AreaDict;
			case k3dIChannel::DATA_VOLUME:
				return VolumeDict;
			case k3dIChannel::DATA_MASS:
				return MassDict;
			case k3dIChannel::DATA_FORCE:
				return ForceDict;
			case k3dIChannel::DATA_PRESSURE:
				return PressureDict;
			default:
				g_assert_not_reached();
		}

	return DistanceDict;	// Can never be reached!
}

double k3dConvertStringDouble(k3dIChannel::DataType Data, SDPCSTRING Input)
{
	static const char* alpha = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";

	// If it's scalar data, we're done ...
	if(k3dIChannel::DATA_SCALAR == Data)
		return atof(Input);

	// Figure out which dictionary to use ...
	k3dConversionUnitDict& dictionary = k3dConversionUnitDictionary(Data);

	// Parse units ...
	sdpString input(Input);
	sdpString numberstring;
	sdpString unitstring;
	sdpString::size_type unitindex = input.find_first_of(alpha);
	if(unitindex != sdpString::npos)
		{
			numberstring = input.left(unitindex);
			unitstring = input.right(input.size() - unitindex);
		}
	else
		{
			numberstring = input;
		}

	// Get the input unit of measurement ...
	k3dConversionUnitIterator inputunits = dictionary.find(unitstring);

	// If there isn't any, go with the default ...
	if(inputunits == dictionary.end())
		{
			inputunits = dictionary.find("display");
			g_assert(inputunits != dictionary.end());	// Should never fail ...
		}

	// Calculate the result ...
	return atof(numberstring) * inputunits->second.m_ConversionFactor;
}

sdpString k3dConvertDoubleString(k3dIChannel::DataType Data, double InputValue, gulong Precision)
{
	if(k3dIChannel::DATA_SCALAR == Data)
		return sdpFormatString("%.*f", int(Precision), InputValue);

	// Figure out which dictionary to use ...
	k3dConversionUnitDict& dictionary = k3dConversionUnitDictionary(Data);
	
	// Create the output string ...
	k3dConversionUnit& outputunits = dictionary["display"];
	return sdpFormatString("%.*f %s", int(Precision), InputValue / outputunits.m_ConversionFactor, outputunits.m_Symbol);
}

sdpString k3dUpdateDoubleString(k3dIChannel::DataType Data, SDPCSTRING Input, double& OutputValue, gulong Precision)
{
	OutputValue = k3dConvertStringDouble(Data, Input);
	return k3dConvertDoubleString(Data, OutputValue, Precision);
}

// Return the path to an object's library ...
SDPCSTRING k3dLibraryPath(k3dIObject* Object)
{
	// Sanity checks ...
	g_assert(Object);

	return Object->ObjectContext()->ObjectInfo()->Library()->Path();
}

// Return the path to the global collection of dialog templates ...
sdpString k3dGlobalDialogTemplatePath(k3dIApplication* Application)
{
	// Sanity checks ...
	g_assert(Application);
	
	return sdpAppendPath(Application->BasePath(), "dialogtemplates");
}

// Return the object with the given ID (if it exists) ...
k3dIObject* k3dGetObject(k3dIDocument* Document, k3dID ObjectID)
{
	// Sanity checks ...
	g_assert(Document);
	g_assert(ObjectID);

	// For each object in the document ...
	k3dIObjectContext* context;
	for(K3DOBJECTCONTEXTHANDLE handle = Document->GetFirstObjectContext(); handle; handle = Document->GetNextObjectContext(handle))
		{
			context = Document->GetObjectContext(handle);

			// See if the IDs match ...
			if(context->ObjectID() == ObjectID)
				return context->Object();
		}

	// Couldn't find it!!!
	return 0;
}

// Return the object with the requested name (if it exists) ...
k3dIObject* k3dGetObject(k3dIDocument* Document, SDPCSTRING Name)
{
	// Sanity checks ...
	g_assert(Document);
	g_assert_string(Name);
	
	sdpString name(Name);
	
	// For each object in the document ...
	k3dIObjectContext* context;
	for(K3DOBJECTCONTEXTHANDLE handle = Document->GetFirstObjectContext(); handle; handle = Document->GetNextObjectContext(handle))
		{
			context = Document->GetObjectContext(handle);

			// See if the name matches ...
			if(context->Name() == name)
				return context->Object();
		}

	// Couldn't find it!!!
	return 0;
}

/// Returns a set of object infos (zero or one) that implement the requested class ID
void k3dGetObjectInfo(k3dIApplication* Application, const sdpCLSID& ClassID, k3dObjectInfoSet& Results)
{
	// Sanity checks ...
	g_assert(Application);
	g_assert(ClassID != sdpCLSIDNull);
	
	for(K3DOBJECTINFOHANDLE handle = Application->GetFirstObjectInfo(); handle; handle = Application->GetNextObjectInfo(handle))
		{
			k3dIObjectInfo* info = Application->GetObjectInfo(handle);
			if(info->ClassID() == ClassID)
				Results.insert(info);
		}
}

/// Returns a set of object infos that support the requested interface ID
void k3dGetObjectInfo(k3dIApplication* Application, const sdpIID& InterfaceID, k3dObjectInfoSet& Results)
{
	// Sanity checks ...
	g_assert(Application);
	g_assert(InterfaceID != sdpIIDNull);
	
	for(K3DOBJECTINFOHANDLE handle = Application->GetFirstObjectInfo(); handle; handle = Application->GetNextObjectInfo(handle))
		{
			k3dIObjectInfo* info = Application->GetObjectInfo(handle);
			if(SDP_S_OK == info->QueryObjectInterface(InterfaceID))
				Results.insert(info);
		}
}

/// Returns a set of object infos that support an object with the requested name
void k3dGetObjectInfo(k3dIApplication* Application, SDPCSTRING Name, k3dObjectInfoSet& Results)
{
	// Sanity checks ...
	g_assert(Application);
	g_assert_string(Name);

	sdpString name(Name);
			
	for(K3DOBJECTINFOHANDLE handle = Application->GetFirstObjectInfo(); handle; handle = Application->GetNextObjectInfo(handle))
		{
			k3dIObjectInfo* info = Application->GetObjectInfo(handle);
			if(info->Name() == name)
				Results.insert(info);
		}
}

/// Provides a container for building an array of objects
class k3dGetObjectByClass
{
public:
	k3dGetObjectByClass(const sdpCLSID& ClassID, k3dObjectSet& Results) : m_ClassID(ClassID), m_Results(Results) {}
	
	bool HandleObject(k3dIObjectContext* Context)
	{
		if(Context->ObjectInfo()->ClassID() == m_ClassID)
			m_Results.insert(Context->Object());
		return true;
	}
	
	const sdpCLSID& m_ClassID;
	k3dObjectSet& m_Results;
};

/// Returns the set of objects that are of the given class ...
void k3dGetObjects(k3dIDocument* Document, const sdpCLSID& ClassID, k3dObjectSet& Results)
{
	// Sanity checks ...
	g_assert(Document);
	g_assert(ClassID != sdpCLSIDNull);
	
	k3dGetObjectByClass functor(ClassID, Results);
	k3dForEachObject(Document, 0, functor);
}

/// Provides a container for building an array of objects
class k3dGetObjectByInterface
{
public:
	k3dGetObjectByInterface(const sdpIID& InterfaceID, k3dObjectSet& Results) : m_InterfaceID(InterfaceID), m_Results(Results) {}
	
	bool HandleObject(k3dIObjectContext* Context)
	{
		if(SDP_S_OK == Context->ObjectInfo()->QueryObjectInterface(m_InterfaceID))
			m_Results.insert(Context->Object());
		return true;
	}
	
	const sdpIID& m_InterfaceID;
	k3dObjectSet& m_Results;
};

/// Returns the set of objects that support the requested interface ...
void k3dGetObjects(k3dIDocument* Document, const sdpIID& InterfaceID, k3dObjectSet& Results)
{
	// Sanity checks ...
	g_assert(Document);
	g_assert(InterfaceID != sdpIIDNull);
	
	k3dGetObjectByInterface functor(InterfaceID, Results);
	k3dForEachObject(Document, 0, functor);
}

/// Provides a container for building an array of objects
class k3dGetObjectByState
{
public:
	k3dGetObjectByState(k3dObjectSet& Results) : m_Results(Results) {}
	
	bool HandleObject(k3dIObjectContext* Context)
	{
		m_Results.insert(Context->Object());
		return true;
	}
	
	k3dObjectSet& m_Results;
};

/// Returns a set of objects that fit the given state criteria ...
void k3dGetObjects(k3dIDocument* Document, const gulong Criteria, k3dObjectSet& Results)
{
	k3dGetObjectByState functor(Results);
	k3dForEachObject(Document, Criteria, functor);
}

/// Provides a container for building an array of meshes
class k3dGetMesh
{
public:
	k3dGetMesh(k3dMeshArray& Results) : m_Results(Results) {}
	
	bool HandleMesh(k3dIMesh* Mesh)
	{
		m_Results.push_back(Mesh);
		return true;
	}
	
	k3dMeshArray& m_Results;
};

/// Returns an array of meshes that fit the given state criteria ...
void k3dGetMeshes(k3dIDocument* Document, const gulong Criteria, k3dMeshArray& Results)
{
	// Sanity checks ...
	g_assert(Document);
	
	k3dGetMesh functor(Results);
	k3dForEachMesh(Document, Criteria, functor);
}

/// Provides a container for building an array of points
class k3dGetPoint
{
public:
	k3dGetPoint(k3dPointArray& Results) : m_Results(Results) {}
	
	bool HandlePoint(k3dIPoint* Point)
	{
		m_Results.push_back(Point);
		return true;
	}
	
	k3dPointArray& m_Results;
};

void k3dGetPoints(k3dIDocument* Document, const gulong Criteria, k3dPointArray& Results)
{
	// Sanity checks ...
	g_assert(Document);
	
	k3dGetPoint functor(Results);
	k3dForEachPoint(Document, Criteria, functor);
}

void k3dGetPoints(k3dIPolygon* Polygon, k3dPointArray& Results)
{
	// Sanity checks ...
	g_assert(Polygon);
	
	k3dIPolyPointServer* polypointserver = Polygon->Points();
	g_assert(polypointserver);
	
	for(K3DPOLYPOINTHANDLE polypoint = polypointserver->GetFirstPolyPoint(); polypoint; polypoint = polypointserver->GetNextPolyPoint(polypoint))
		Results.push_back(polypointserver->GetPolyPoint(polypoint)->Point());
}

void k3dGetPolyPoints(k3dIPolygon* Polygon, k3dPolyPointArray& Results)
{
	// Sanity checks ...
	g_assert(Polygon);
	
	k3dIPolyPointServer* polypointserver = Polygon->Points();
	g_assert(polypointserver);
	
	for(K3DPOLYPOINTHANDLE polypoint = polypointserver->GetFirstPolyPoint(); polypoint; polypoint = polypointserver->GetNextPolyPoint(polypoint))
		Results.push_back(polypointserver->GetPolyPoint(polypoint));
}

/// Provides a container for building an array of polygons
class k3dGetPolygon
{
public:
	k3dGetPolygon(k3dPolygonArray& Results) : m_Results(Results) {}
	
	bool HandlePolygon(k3dIPolygon* Polygon)
	{
		m_Results.push_back(Polygon);
		return true;
	}
	
	k3dPolygonArray& m_Results;
};

void k3dGetPolygons(k3dIDocument* Document, const gulong Criteria, k3dPolygonArray& Results)
{
	// Sanity checks ...
	g_assert(Document);
	
	k3dGetPolygon functor(Results);
	k3dForEachPolygon(Document, Criteria, functor);
}

/// Calculates a polygon normal
sdpVector3 k3dGetNormal(const sdpVector3Array& Points)
{
	// Handle the simplest case ...
	if(Points.size() < 3)
		return sdpVector3(0, 0, 0);
	
	sdpVector3 result;
	
	sdpVector3 segment1 = Points.back() - Points.front();
	for(gulong i = 1; i < Points.size() - 1; i++)
		result += (segment1) ^ (Points[i] - Points[0]);
	
	return result.Normalize();
}

/// Calculates a polygon normal
sdpVector3 k3dGetNormal(const k3dPointArray& Points)
{
	sdpVector3 coords;
	sdpVector3Array points;
	for(gulong i = 0; i < Points.size(); i++)
		{
			Points[i]->Location()->GetLocalXYZ(coords);
			points.push_back(coords);
		}
	
	return k3dGetNormal(points);
}

/// Calculates a polygon normal
sdpVector3 k3dGetNormal(const k3dPolyPointArray& PolyPoints)
{
	sdpVector3 coords;
	sdpVector3Array points;
	for(gulong i = 0; i < PolyPoints.size(); i++)
		{
			PolyPoints[i]->Location()->GetLocalXYZ(coords);
			points.push_back(coords);
		}
	
	return k3dGetNormal(points);
}

/// Calculates a polygon normal
sdpVector3 k3dGetNormal(k3dIPolygon* Polygon)
{
	// Sanity checks ...
	g_assert(Polygon);
	
	k3dPointArray points;
	k3dGetPoints(Polygon, points);
	
	return k3dGetNormal(points);
}

/// Destroys an object
sdpResult k3dDeleteObject(sdpIUnknown* Object)
{
	// Sanity checks ...
	g_assert(Object);
	
	// Make sure it's a K-3D object!
	k3dIObject* object = sdpQueryInterface<k3dIObject>(Object);
	g_return_val_if_fail(object, SDP_E_FAIL);
	
	// Zap it ...
	return object->ObjectContext()->Layer()->Document()->DeleteObject(object);
}

bool k3dSetPosition(sdpIUnknown* Object, sdpVector3 Position)
{
	// Sanity check ...
	g_assert(Object);

	// See if the object supports k3dITransform ...
	k3dITransform* transform = sdpQueryInterface<k3dITransform>(Object);
	if(0 == transform)
		return false;

	// If so, position it ...
	transform->SetPosition(k3dITransform::OLD_FRAME, Position);
	transform->SetPosition(k3dITransform::CURRENT_FRAME, Position);

	transform->SetPosition(k3dITransform::OLD_CHILD_FRAME, sdpVector3());
	transform->SetPosition(k3dITransform::CURRENT_CHILD_FRAME, sdpVector3());

	transform->Release();

	return true;
}

bool k3dSetOrientation(sdpIUnknown* Object, sdpEulerAngles Orientation)
{
	// Sanity check ...
	g_assert(Object);

	// See if the object supports k3dITransform ...
	k3dITransform* transform = sdpQueryInterface<k3dITransform>(Object);
	if(0 == transform)
		return false;

	// If so, orient it ...
	transform->SetOrientation(k3dITransform::OLD_FRAME, sdpAngleAxis(sdpQuaternion(Orientation)));
	transform->SetOrientation(k3dITransform::CURRENT_FRAME, sdpAngleAxis(sdpQuaternion(Orientation)));
	
	transform->SetOrientation(k3dITransform::OLD_CHILD_FRAME, sdpAngleAxis());
	transform->SetOrientation(k3dITransform::CURRENT_CHILD_FRAME, sdpAngleAxis());
	
	transform->Release();

	return true;
}

bool k3dSetScale(sdpIUnknown* Object, sdpVector3 Scale)
{
	// Sanity check ...
	g_assert(Object);

	// See if the object supports k3dITransform ...
	k3dITransform* transform = sdpQueryInterface<k3dITransform>(Object);
	if(0 == transform)
		return false;

	// If so, orient it ...
	transform->SetScale(k3dITransform::OLD_FRAME, Scale);
	transform->SetScale(k3dITransform::CURRENT_FRAME, Scale);

	transform->SetScale(k3dITransform::OLD_CHILD_FRAME, sdpVector3(1));
	transform->SetScale(k3dITransform::CURRENT_CHILD_FRAME, sdpVector3(1));

	transform->Release();

	return true;
}

// Set an object's name (if possible) ...
bool k3dSetName(sdpIUnknown* Object, SDPCSTRING ObjectName)
{
	// Sanity check ...
	g_assert(Object);

	// See if it's really a K-3D object ...
	k3dIObject* object = sdpQueryInterface<k3dIObject>(Object);
	g_return_val_if_fail(object, false);

	// If so, rename it ...
	object->ObjectContext()->SetName(ObjectName);
	object->Release();

	return true;
}

// Sets a channel's maximum range of values (if possible)
bool k3dChannelSetValueRange(sdpIUnknown* Channel, const gdouble MinValue, const gdouble MaxValue)
{
	// Sanity checks ...
	g_assert(Channel);
	
	k3dISplineChannel* splinechannel = sdpQueryInterface<k3dISplineChannel>(Channel);
	if(!splinechannel)
		return false;

	splinechannel->SetRange1(MinValue, MaxValue);
	splinechannel->Release();	
	
	return true;
}

// Sets a channel's maximum range of values (if possible)
bool k3dChannelSetValueRange(sdpIUnknown* Channel, const sdpVector3 MinValue, const sdpVector3 MaxValue)
{
	// Sanity checks ...
	g_assert(Channel);
	
	k3dISplineChannel* splinechannel = sdpQueryInterface<k3dISplineChannel>(Channel);
	if(!splinechannel)
		return false;

	splinechannel->SetRange3(const_cast<sdpVector3&>(MinValue), const_cast<sdpVector3&>(MaxValue));
	splinechannel->Release();	
	
	return true;
}

// Sets a channel to a constant value (if possible)
bool k3dChannelSetConstantValue(sdpIUnknown* Channel, const gdouble Value)
{
	// Sanity checks ...
	g_assert(Channel);

	k3dISplineChannel* splinechannel = sdpQueryInterface<k3dISplineChannel>(Channel);
	if(!splinechannel)
		return false;

	// Get rid of all segments ...
	for(K3DSPLINECHANNELSEGMENTHANDLE handle = splinechannel->GetFirstSegment(); handle; handle = splinechannel->GetFirstSegment())
		splinechannel->DestroySegment(splinechannel->GetSegment(handle));

	// Get the (1) remaining node ...
	k3dISplineChannelNode* firstnode = splinechannel->GetNode(splinechannel->GetFirstNode());
	g_return_val_if_fail(firstnode, false);
	
	firstnode->SetPosition(sdpVector2(0.0, Value));
	firstnode->SetValue1(Value);

	splinechannel->Release();	
	
	return true;
}

// Sets a channel to a constant value (if possible)
bool k3dChannelSetConstantValue(sdpIUnknown* Channel, const sdpVector3 Value)
{
	// Sanity checks ...
	g_assert(Channel);

	k3dISplineChannel* splinechannel = sdpQueryInterface<k3dISplineChannel>(Channel);
	if(!splinechannel)
		return false;

	// Get rid of all segments ...
	for(K3DSPLINECHANNELSEGMENTHANDLE handle = splinechannel->GetFirstSegment(); handle; handle = splinechannel->GetFirstSegment())
		splinechannel->DestroySegment(splinechannel->GetSegment(handle));

	// Get the (1) remaining node ...
	k3dISplineChannelNode* firstnode = splinechannel->GetNode(splinechannel->GetFirstNode());
	g_return_val_if_fail(firstnode, false);
	
	firstnode->SetPosition(sdpVector2(0.0, 0.0));
	firstnode->SetValue3(const_cast<sdpVector3&>(Value));

	splinechannel->Release();	
	
	return true;
}

// Sets a channel to a linear range of values (if possible)
bool k3dChannelSetLinearValue(sdpIUnknown* Channel, const gdouble Value1, const gdouble Value2)
{
	// Sanity checks ...
	g_assert(Channel);

	k3dISplineChannel* splinechannel = sdpQueryInterface<k3dISplineChannel>(Channel);
	if(!splinechannel)
		return false;

	// Get rid of all segments ...
	for(K3DSPLINECHANNELSEGMENTHANDLE handle = splinechannel->GetFirstSegment(); handle; handle = splinechannel->GetFirstSegment())
		splinechannel->DestroySegment(splinechannel->GetSegment(handle));

	// Add a single segment back ...
	splinechannel->CreateSegment();

	// Setup all four nodes ...
	K3DSPLINECHANNELNODEHANDLE nodehandle = 0;
	k3dISplineChannelNode* node = 0;
		
	nodehandle = splinechannel->GetFirstNode();
	node = splinechannel->GetNode(nodehandle);
	node->SetPosition(sdpVector2(0.0, Value1));
	node->SetValue1(Value1);
	
	nodehandle = splinechannel->GetNextNode(nodehandle);
	node = splinechannel->GetNode(nodehandle);
	node->SetPosition(sdpLerp(sdpVector2(0.0, Value1), sdpVector2(1.0, Value2), 0.2));
		
	nodehandle = splinechannel->GetNextNode(nodehandle);
	node = splinechannel->GetNode(nodehandle);
	node->SetPosition(sdpLerp(sdpVector2(0.0, Value1), sdpVector2(1.0, Value2), 0.8));
	
	nodehandle = splinechannel->GetNextNode(nodehandle);
	node = splinechannel->GetNode(nodehandle);
	node->SetPosition(sdpVector2(1.0, Value2));
	node->SetValue1(Value2);
	
	splinechannel->Release();	
	
	return true;
}

// Sets a channel to a linear range of values (if possible)
bool k3dChannelSetLinearValue(sdpIUnknown* Channel, const sdpVector3 Value1, const sdpVector3 Value2)
{
	// Sanity checks ...
	g_assert(Channel);

	k3dISplineChannel* splinechannel = sdpQueryInterface<k3dISplineChannel>(Channel);
	if(!splinechannel)
		return false;

	// Get rid of all segments ...
	for(K3DSPLINECHANNELSEGMENTHANDLE handle = splinechannel->GetFirstSegment(); handle; handle = splinechannel->GetFirstSegment())
		splinechannel->DestroySegment(splinechannel->GetSegment(handle));

	// Add a single segment back ...
	splinechannel->CreateSegment();

	// Setup all four nodes ...
	K3DSPLINECHANNELNODEHANDLE nodehandle = 0;
	k3dISplineChannelNode* node = 0;
		
	nodehandle = splinechannel->GetFirstNode();
	node = splinechannel->GetNode(nodehandle);
	node->SetPosition(sdpVector2(0.0, 0.0));
	node->SetValue3(const_cast<sdpVector3&>(Value1));
	
	nodehandle = splinechannel->GetNextNode(nodehandle);
	node = splinechannel->GetNode(nodehandle);
	node->SetPosition(sdpLerp(sdpVector2(0.0, 0.0), sdpVector2(1.0, 1.0), 0.2));
		
	nodehandle = splinechannel->GetNextNode(nodehandle);
	node = splinechannel->GetNode(nodehandle);
	node->SetPosition(sdpLerp(sdpVector2(0.0, 0.0), sdpVector2(1.0, 1.0), 0.8));
	
	nodehandle = splinechannel->GetNextNode(nodehandle);
	node = splinechannel->GetNode(nodehandle);
	node->SetPosition(sdpVector2(1.0, 1.0));
	node->SetValue3(const_cast<sdpVector3&>(Value2));
	
	splinechannel->Release();	
	
	return true;
}

// Return the min/max world coordinates for every point in the document that meets the given criteria ...
bool k3dWorldDimensions(k3dIDocument* Document, gulong Criteria, sdpVector3& Min, sdpVector3& Max)
{
g_assert_not_reached();
return false;
/*
	// Calculate world coordinates ...
	Document->SnapshotWorldCoordinates();

	// Reset our results ...
	Min = DBL_MAX;
	Max = -DBL_MAX;

	// Iterate all points ...
	sdpVector3 coords;
	bool foundsomething = false;

	for(k3dIPoint* point = Document->GetFirstPoint(); point; point = Document->GetNextPoint(point))
		{
			// Skip points that don't match our criteria ...
			if(SDP_S_OK != point->State()->QueryState(Criteria)) continue;

			// OK, we found something that fits the criteria ...
			foundsomething = true;

			// Get the point's world coordinates ...
			point->Location()->GetWorldXYZ(coords);

			// Compare against our min/max coordinates ...
			if(coords[0] < Min[0]) Min[0] = coords[0];
			if(coords[0] > Max[0]) Max[0] = coords[0];
			if(coords[1] < Min[1]) Min[1] = coords[1];
			if(coords[1] > Max[1]) Max[1] = coords[1];
			if(coords[2] < Min[2]) Min[2] = coords[2];
			if(coords[2] > Max[2]) Max[2] = coords[2];
		}

	return foundsomething;
*/
}

bool k3dObjectDimensions(k3dIObject* Object, gulong Criteria, sdpVector3& Min, sdpVector3& Max)
{
	// Sanity checks ...
	g_assert(Object);
	
	// If it's a mesh, ask the mesh for the results ...
	k3dIMesh* mesh = sdpQueryInterface<k3dIMesh>(Object);
	if(mesh)
		{
			mesh->GetDimensions(Criteria, Min, Max);
			mesh->Release();
			return true;
		}
		
	// If it's a location-server, call our specialized function for handling it ...
	k3dILocationServer* locationserver = sdpQueryInterface<k3dILocationServer>(Object);
	if(locationserver)
		{
			k3dObjectDimensions(locationserver, Criteria, Min, Max);
			locationserver->Release();
		}
		
	// Otherwise, give up!
	Min = 0.0;
	Max = 0.0;
	
	return false;
}

// Return the min/max model coordinates for every location in the given object that meets the given criteria ...
bool k3dObjectDimensions(k3dILocationServer* LocationServer, gulong Criteria, sdpVector3& Min, sdpVector3& Max)
{
	// Sanity checks ...
	g_assert(LocationServer);

	// Reset our results ...
	Min = DBL_MAX;
	Max = -DBL_MAX;

	// Iterate all locations ...
	sdpVector3 coords;
	k3dILocation* location;

	for(K3DLOCATIONHANDLE locationhandle = LocationServer->GetFirstLocation(); locationhandle; locationhandle = LocationServer->GetNextLocation(locationhandle))
		{
			location = LocationServer->GetLocation(locationhandle);

			// Skip locations that don't match our criteria ...
			if(SDP_S_OK != location->State()->QueryState(Criteria))
				continue;

			// Get the location's current coordinates ...
			location->GetAnimXYZ(coords);

			// Compare against our min/max coordinates ...
			if(coords[0] < Min[0])
				Min[0] = coords[0];
			if(coords[0] > Max[0])
				Max[0] = coords[0];
			if(coords[1] < Min[1])
				Min[1] = coords[1];
			if(coords[1] > Max[1])
				Max[1] = coords[1];
			if(coords[2] < Min[2])
				Min[2] = coords[2];
			if(coords[2] > Max[2])
				Max[2] = coords[2];
		}

	return true;
}

class k3dCalculateNormal
{
public:
	bool HandleMesh(k3dIMesh* Mesh)
	{
		Mesh->CalculateNormals();
		return true;
	}
};

void k3dCalculateNormals(k3dIDocument* Document)
{
	// Sanity checks ...
	g_assert(Document);

	k3dCalculateNormal functor;
	k3dForEachMesh(Document, 0, functor);
}

bool k3dIsInside(sdpVector2 Point, k3dIPolygon* Polygon)
{
	// Sanity check ...
	g_assert(Polygon);

	sdpVector2Array array;
	sdpVector3 screencoords;

	// For each point ...
	k3dIPolyPointServer* points = Polygon->Points();
	for(K3DPOLYPOINTHANDLE polypointhandle = points->GetFirstPolyPoint(); polypointhandle; polypointhandle = points->GetNextPolyPoint(polypointhandle))
		{
			points->GetPolyPoint(polypointhandle)->Point()->Location()->GetScreenXYZ(screencoords);
			array.push_back(sdpVector2(screencoords, 2));	// Lose the Z coordinate ...
		}

	return sdpIsInside(Point, array);
}

sdpMatrix4 k3dGetTransformMatrix(k3dITransform* Transform, const k3dITransform::ReferenceFrame Frame)
{
	// Sanity checks ...
	g_assert(Transform);
	
	sdpVector3 position;
	Transform->GetPosition(Frame, position);
	
	sdpAngleAxis orientation;
	Transform->GetOrientation(Frame, orientation);
	
	sdpVector3 scale;
	Transform->GetScale(Frame, scale);
	
	return sdpTranslation3D(position) * sdpRotation3D(orientation) * sdpScaling3D(scale);
}

sdpMatrix4 k3dGetObjectToWorldMatrix(k3dIObjectContext* Object)
{
	// Sanity checks ...
	g_assert(Object);

	// Start out with an identity matrix ...
	sdpMatrix4 result = sdpIdentity3D();
	
	// Create a stack of k3dITransform-enabled objects ...
	std::stack<k3dITransform*> transformstack;

	// Start with the given object and move to the root of the hierarchy ...
	k3dITransform* transform;
	for(; Object; Object = Object->Parent())
		{
			transform = sdpQueryInterface<k3dITransform>(Object->Object());
			if(transform)
				transformstack.push(transform);
		}

	// Concatenate parent self- and child-matrices in order from root -> leaf ...
	while(transformstack.size() > 1)	
		{
			result = result * k3dGetTransformMatrix(transformstack.top(), k3dITransform::ANIMATED_FRAME);
			result = result * k3dGetTransformMatrix(transformstack.top(), k3dITransform::ANIMATED_CHILD_FRAME);
			
			transformstack.top()->Release();
			transformstack.pop();
		}
	// Concatenate the given object's self-matrix only ...
	while(transformstack.size() > 0)
		{
			result = result * k3dGetTransformMatrix(transformstack.top(), k3dITransform::ANIMATED_FRAME);
			
			transformstack.top()->Release();
			transformstack.pop();
		}

	return result;
}

sdpMatrix4 k3dGetWorldToObjectMatrix(k3dIObjectContext* Object)
{
	// Sanity checks ...
	g_assert(Object);

	// Return theinverse of the object-to-world matrix ...
	return k3dGetObjectToWorldMatrix(Object).Inverse();
}

sdpMatrix4 k3dGetObjectToObjectMatrix(k3dIObjectContext* From, k3dIObjectContext* To)
{
	// Sanity checks ...
	g_assert(From);
	g_assert(To);
	
	// Get the matrix from object-to-world ...
	sdpMatrix4 from = k3dGetObjectToWorldMatrix(From);
	
	// Get the matrix from world-to-object ...
	sdpMatrix4 to = k3dGetWorldToObjectMatrix(To);
	
	// return the concatenated result ...
	return from * to;
}

sdpMatrix4 k3dGetWorldToCameraMatrix(k3dICamera* Camera)
{
	// Sanity checks ...
	g_assert(Camera);
	
	// Convert the camera to an object ...
	k3dIObject* object = sdpQueryInterface<k3dIObject>(Camera);
	g_assert(object);
	
	sdpMatrix4 result = k3dGetWorldToObjectMatrix(object->ObjectContext());
	object->Release();
	
	return result;
}

sdpMatrix4 k3dGetCameraToWorldMatrix(k3dICamera* Camera)
{
	// Sanity checks ...
	g_assert(Camera);
	
	// Convert the camera to an object ...
	k3dIObject* object = sdpQueryInterface<k3dIObject>(Camera);
	g_assert(object);
	
	sdpMatrix4 result = k3dGetObjectToWorldMatrix(object->ObjectContext());
	object->Release();
	
	return result;
}

sdpMatrix4 k3dGetObjectToCameraMatrix(k3dIObjectContext* Object, k3dICamera* Camera)
{
	// Sanity checks ...
	g_assert(Object);
	g_assert(Camera);
	
	// Convert the camera to an object ...
	k3dIObject* object = sdpQueryInterface<k3dIObject>(Camera);
	g_assert(object);
	
	sdpMatrix4 result = k3dGetObjectToObjectMatrix(Object, object->ObjectContext());
	object->Release();
	
	return result;
}

sdpMatrix4 k3dGetCameraToObjectMatrix(k3dICamera* Camera, k3dIObjectContext* Object)
{
	// Sanity checks ...
	g_assert(Camera);
	g_assert(Object);
	
	// Convert the camera to an object ...
	k3dIObject* object = sdpQueryInterface<k3dIObject>(Camera);
	g_assert(object);
	
	sdpMatrix4 result = k3dGetObjectToObjectMatrix(object->ObjectContext(), Object);
	object->Release();
	
	return result;
}

sdpVector3 k3dGetWorldPosition(k3dIObjectContext* Context)
{
	// Sanity checks ...
	g_assert(Context);
	
	return k3dGetObjectToWorldMatrix(Context) * sdpVector3(0, 0, 0);
}

void k3dPushOpenGLMatrix(const sdpMatrix4& Matrix)
{
	double glmatrix[16];
	Matrix.Transpose().CopyArray(glmatrix);
	glMultMatrixd(glmatrix);
}

void k3dPushRiMatrix(const sdpMatrix4& Matrix, k3dIRiEngine* Engine)
{
	// Sanity checks ...
	g_assert(Engine);

	sdpMatrix4 matrix = Matrix.Transpose();
	Engine->RiConcatTransform((gdouble*)matrix);
}

/// Fucntion for creating an sdpcomlibraryinfo document
sdpXmlIElement* k3dStartLibraryInfo(sdpXmlIDocument* Document, SDPCSTRING Name, SDPCSTRING Version)
{
	// Sanity checks ...
	g_assert(Document);
	g_assert_string(Name);
	g_assert_string(Version);
	
	sdpXmlIElement* result = Document->CreateRootElement("sdpcomlibraryinfo");
	result->CreateAttribute("name", Name);
	result->CreateAttribute("version", Version);

	return result;
}

/// Function for creating a standardized object entry in an sdpcomlibraryinfo document
sdpXmlIElement* k3dAddObject(sdpXmlIElement* Library, const sdpCLSID& ClassID, SDPCSTRING Name)
{
	// Sanity checks ...
	g_assert(Library);
	g_assert_string(Name);
	
	sdpXmlIElement* result = Library->CreateElement("object");
	result->CreateAttribute("clsid", sdpString(ClassID));
	result->CreateAttribute("name", Name);

	return result;
}

/// Function for adding a standardized object interface entry in an sdpcomlibraryinfo document
void k3dAddInterface(sdpXmlIElement* Object, const sdpIID& InterfaceID)
{
	// Sanity checks ...
	g_assert(Object);
	
	sdpXmlIElement* objectinterface = Object->CreateElement("interface");
	objectinterface->CreateAttribute("iid", sdpString(InterfaceID));
}

/// Function for adding a K-3D object bookmark entry in an sdpcomlibraryinfo document
void k3dAddBookmark(sdpXmlIElement* Object, SDPCSTRING Location)
{
	// Sanity checks ...
	g_assert(Object);
	g_assert_string(Location);
	
	sdpXmlIElement* bookmark = Object->CreateElement("bookmark");
	bookmark->CreateAttribute("location", Location);
}

sdpVector2 k3dTransformMouseCoords(k3dICamera* Camera, const sdpVector2 MouseCoords, gdouble Left, gdouble Right, gdouble Top, gdouble Bottom)
{
	// Sanity Checks ...
	g_assert(Camera);
	
	gdouble width = Camera->Width();
	gdouble height = Camera->Height();

	gdouble mousex = sdpClamp<gdouble>(0, width, MouseCoords[0]);
	gdouble mousey = sdpClamp<gdouble>(0, height, MouseCoords[1]);
	
	if(0 == width)
		width = 1;
		
	if(0 == height)
		height = 1;
		
	return sdpVector2(sdpLerp(Left, Right, mousex / width), sdpLerp(Top, Bottom, mousey / height));
}

sdpVector3 k3dMouseToCameraRay(k3dICamera* Camera, const sdpVector2 MouseCoords)
{
	// Sanity checks ...
	g_assert(Camera);

	// Get the camera frustum ...
	gdouble left, right, top, bottom, nearplane, farplane;
	Camera->GetViewingFrustum(&left, &right, &top, &bottom, &nearplane, &farplane);

	// Convert mouse coordinates (pixels) to screen coordinates ...
	sdpVector2 screencoords = k3dTransformMouseCoords(Camera, MouseCoords, left, right, top, bottom);

	sdpVector3 result(screencoords, nearplane);
	result.Normalize();

	return result;
}

void k3dMouseToWorldRay(k3dICamera* Camera, const sdpVector2 MouseCoords, sdpVector3& RayOrigin, sdpVector3& RayDirection)
{
	// Sanity checks ...
	g_assert(Camera);

	// Get the mouse ray in camera coordinates ...
	sdpVector3 ray = k3dMouseToCameraRay(Camera, MouseCoords);

	// Transform it to the world frame ...
	sdpMatrix4 cameratoworld = k3dGetCameraToWorldMatrix(Camera);
	RayOrigin = cameratoworld * sdpVector3(0, 0, 0);
	RayDirection = (cameratoworld * ray) - RayOrigin;
}

class k3dRedrawAllFunctor
{
public:
	k3dRedrawAllFunctor(bool Synchronous) : m_Synchronous(Synchronous) {}

	bool HandleObject(k3dIObjectContext* ObjectContext)
	{
		k3dICamera* camera = sdpQueryInterface<k3dICamera>(ObjectContext->Object());
		if(camera)
			{
				camera->Redraw(m_Synchronous);
				camera->Release();
			}
		
		return true;
	}
	
protected:
	bool m_Synchronous;
};

void k3dRedrawAll(k3dIDocument* Document, bool Synchronous)
{
	k3dRedrawAllFunctor functor(Synchronous);
	k3dForEachObject(Document, 0, functor);
}


