// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dMouseInput class, which converts mouse & button input into a set of standardized logical events
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dmouseinput.h>
#include <sdpgtkevents.h>

/////////////////////////////////////////////////////////////////////////////
// k3dMouseInput

k3dMouseInput::k3dMouseInput()
{
	m_Dragging = false;
}

k3dMouseInput::~k3dMouseInput()
{
}

void k3dMouseInput::OnMouseMove(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	sdpGtkEventWidgetMotionNotifyEvent* event = static_cast<sdpGtkEventWidgetMotionNotifyEvent*>(Event);
	
	// Get the pointer position to force the X server to send us the next motion event ...
	int x;
	int y;
	GdkModifierType modifiers;
	gdk_window_get_pointer(event->Event()->window, &x, &y, &modifiers);
	m_LastMouse = m_CurrentMouse;
	m_CurrentMouse = sdpVector2(gdouble(x), gdouble(y));

	// If we're already dragging, send drag events ...
	if(m_Dragging)
		{
			if(m_Buttons[1] && m_Buttons[3])
				OnLRButtonDrag(modifiers, m_CurrentMouse, m_LastMouse, m_StartMouse);
			else if(m_Buttons[1])
				OnLButtonDrag(modifiers, m_CurrentMouse, m_LastMouse, m_StartMouse);
			else if(m_Buttons[2])
				OnMButtonDrag(modifiers, m_CurrentMouse, m_LastMouse, m_StartMouse);
			else if(m_Buttons[3])
				OnRButtonDrag(modifiers, m_CurrentMouse, m_LastMouse, m_StartMouse);
		}

	// Otherwise, see if we should start dragging ...
	else
		{
			// Got any buttons down?
			if(m_Buttons[1] || m_Buttons[2] || m_Buttons[3])
				{
					// Has the mouse moved far away enough from the point where a button was clicked?
					if((m_CurrentMouse - m_StartMouse).Length() > 4.0)
						{
							// Grab mouse events ...
							gtk_grab_add(event->Widget());

							m_Dragging = true;

							if(m_Buttons[1] && m_Buttons[3])
								OnLRButtonStartDrag(modifiers, m_CurrentMouse);
							else if(m_Buttons[1])
								OnLButtonStartDrag(modifiers, m_CurrentMouse);
							else if(m_Buttons[2])
								OnMButtonStartDrag(modifiers, m_CurrentMouse);
							else if(m_Buttons[3])
								OnRButtonStartDrag(modifiers, m_CurrentMouse);
						}
				}
			// No, so it's just a plain mouse move event ...
			else
				{
					OnMouseMove(modifiers, m_CurrentMouse);
				}
		}
}

void k3dMouseInput::OnButtonDown(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);
	
	sdpGtkEventWidgetButtonPressEvent* event = static_cast<sdpGtkEventWidgetButtonPressEvent*>(Event);
	
	// Get the pointer position so we can send it along ...
	int x;
	int y;
	GdkModifierType modifiers;
	gdk_window_get_pointer(event->Event()->window, &x, &y, &modifiers);
	m_CurrentMouse = sdpVector2(gdouble(x), gdouble(y));

	// If it's a double-click, send the event and don't record anything
	// (because double-clicks don't get a button-up event) ...
	if(event->Event()->type == GDK_2BUTTON_PRESS)
		{
			switch(event->Event()->button)
				{
					case 1:
						OnLButtonDoubleClick(modifiers, m_CurrentMouse);
						break;
					case 2:
						OnMButtonDoubleClick(modifiers, m_CurrentMouse);
						break;
					case 3:
						OnRButtonDoubleClick(modifiers, m_CurrentMouse);
						break;
				}
			
			m_Dragging = false;
			m_Buttons[event->Event()->button] = false;	
			
			return;
		}

	// If there aren't any other buttons pressed, record our starting point ...
	if(!m_Buttons[1] && !m_Buttons[2] && !m_Buttons[3])
		{
			m_StartMouse = m_LastMouse = m_CurrentMouse;
		}

	// Record which button was pressed ...
	m_Buttons[event->Event()->button] = true;
	
	// Generate low-level events ...
	switch(event->Event()->button)
		{
			case 1:
				OnLButtonDown(modifiers, m_CurrentMouse);
				break;
			case 2:
				OnMButtonDown(modifiers, m_CurrentMouse);
				break;
			case 3:
				OnRButtonDown(modifiers, m_CurrentMouse);
				break;
		}
}

void k3dMouseInput::OnButtonUp(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);
	
	sdpGtkEventWidgetButtonReleaseEvent* event = static_cast<sdpGtkEventWidgetButtonReleaseEvent*>(Event);
	
	// Get the pointer position so we can send it along ...
	int x;
	int y;
	GdkModifierType modifiers;
	gdk_window_get_pointer(event->Event()->window, &x, &y, &modifiers);
	m_CurrentMouse = sdpVector2(gdouble(x), gdouble(y));

	// Generate low-level events ...
	switch(event->Event()->button)
		{
			case 1:
				OnLButtonUp(modifiers, m_CurrentMouse);
				break;
			case 2:
				OnMButtonUp(modifiers, m_CurrentMouse);
				break;
			case 3:
				OnRButtonUp(modifiers, m_CurrentMouse);
				break;
		}

	// If we were dragging ...
	if(m_Dragging)
		{
			m_Dragging = false;

			// Release mouse events ...
			gtk_grab_remove(event->Widget());

			if(m_Buttons[1] && m_Buttons[3])
				OnLRButtonEndDrag(modifiers, m_CurrentMouse, m_LastMouse, m_StartMouse);
			else if(m_Buttons[1])
				OnLButtonEndDrag(modifiers, m_CurrentMouse, m_LastMouse, m_StartMouse);
			else if(m_Buttons[2])
				OnMButtonEndDrag(modifiers, m_CurrentMouse, m_LastMouse, m_StartMouse);
			else if(m_Buttons[3])
				OnRButtonEndDrag(modifiers, m_CurrentMouse, m_LastMouse, m_StartMouse);
		}

	// ... otherwise, it's a click ...
	else
		{
			switch(event->Event()->button)
				{
					case 1:
						OnLButtonClick(modifiers, m_CurrentMouse);
						break;
					case 2:
						OnMButtonClick(modifiers, m_CurrentMouse);
						break;
					case 3:
						OnRButtonClick(modifiers, m_CurrentMouse);
						break;
				}
		}
		
	// Record which button was released ...
	m_Buttons[event->Event()->button] = false;
}

void k3dMouseInput::OnMouseMove(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
}

void k3dMouseInput::OnLButtonDown(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
}

void k3dMouseInput::OnLButtonUp(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
}

void k3dMouseInput::OnLButtonClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
}

void k3dMouseInput::OnLButtonDoubleClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
}

void k3dMouseInput::OnLButtonStartDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
}

void k3dMouseInput::OnLButtonDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
}

void k3dMouseInput::OnLButtonEndDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
}

void k3dMouseInput::OnMButtonDown(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
}

void k3dMouseInput::OnMButtonUp(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
}

void k3dMouseInput::OnMButtonClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
}

void k3dMouseInput::OnMButtonDoubleClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
}

void k3dMouseInput::OnMButtonStartDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
}

void k3dMouseInput::OnMButtonDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
}

void k3dMouseInput::OnMButtonEndDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
}

void k3dMouseInput::OnRButtonDown(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
}

void k3dMouseInput::OnRButtonUp(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
}

void k3dMouseInput::OnRButtonClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
}

void k3dMouseInput::OnRButtonDoubleClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
}

void k3dMouseInput::OnRButtonStartDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
}

void k3dMouseInput::OnRButtonDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
}

void k3dMouseInput::OnRButtonEndDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
}

void k3dMouseInput::OnLRButtonStartDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
}

void k3dMouseInput::OnLRButtonDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
}

void k3dMouseInput::OnLRButtonEndDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
}


