
// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements k3dObject, a default implementation of the k3dIObject interface, for use as a base class for K-3D objects
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dobject.h>

#include <k3dieventclient.h>

#include <algorithm>
#include <iostream>

////////////////////////////////////////////////////////////////////////////
// class k3dObject

k3dObject::k3dObject()
{
	// Set our sdpCOM reference count to zero ...
	m_ReferenceCount = 0;

	// Initialize interface pointers ...
	m_ObjectContext = 0;

	// Event queueing is off by default ...
	m_QueueEvents = false;
}

k3dObject::~k3dObject()
{
	g_return_if_fail(!m_QueueEvents);	// Somebody forgot to turn off event queueing!
}

bool k3dObject::Create()
{
	return true;
}

////////////////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult k3dObject::SetObjectContext(k3dIObjectContext* ObjectContext)
{
	// Sanity check ...
	g_return_val_if_fail(ObjectContext, SDP_E_FAIL);

	// We should never be called more than once ...
	g_return_val_if_fail(!m_ObjectContext, SDP_E_FAIL);

	// Cache interfaces ...
	m_ObjectContext = ObjectContext;

	// Initialization ...
	g_return_val_if_fail(Create(), SDP_E_FAIL);

	return SDP_S_OK;
}

k3dIObjectContext* k3dObject::ObjectContext()
{
	return m_ObjectContext;
}

sdpResult k3dObject::EditObject()
{
	// No user interface to open
	return SDP_S_OK;
}

sdpResult k3dObject::AddClient(k3dIEventClient* Client, gulong ClientData, SDPCSTRING Event)
{
	// Sanity checks ...
	g_return_val_if_fail(Client, SDP_E_FAIL);
	g_return_val_if_string_fail(Event, SDP_E_FAIL);

	// Add the new client ...
	m_Clients.push_back(k3dObjectClient(Client, ClientData, Event));

	// Increment the client's reference count ...
	Client->AddRef();

	return SDP_S_OK;
}

sdpResult k3dObject::RemoveClient(k3dIEventClient* Client, gulong ClientData, SDPCSTRING Event)
{
	// Sanity checks ...
	g_return_val_if_fail(Client, SDP_E_FAIL);
	g_return_val_if_string_fail(Event, SDP_E_FAIL);

	// Make sure we really own this client ...
	ClientIterator client;
	for(client = m_Clients.begin(); client != m_Clients.end(); client++)
		{
			if((client->m_Client == Client) && (client->m_ClientData == ClientData) && (client->m_Event == Event))
				break;
		}
	g_return_val_if_fail(client != m_Clients.end(), SDP_E_FAIL);

	// De-reference the client ...
	client->m_Client->Release();

	// Get rid of it ...
	m_Clients.erase(client);

	return SDP_S_OK;
}

sdpResult k3dObject::Notify(SDPCSTRING Event)
{
	// Sanity checks ...
	g_return_val_if_string_fail(Event, SDP_E_FAIL);

	// If event queueing is enabled, just record it ...
	if(m_QueueEvents)
		{
			// Prevent back-to-back duplicate events ...
			if(m_Events.size())
				{
					if(m_Events.back() != Event)
						m_Events.push_back(Event);
				}
			else
				{
					m_Events.push_back(Event);
				}
		}
	// Otherwise, signal the event ...
	else
		{
			RawNotify(Event);
		}
		
	return SDP_S_OK;
}

void k3dObject::RawNotify(SDPCSTRING Event)
{
	// Sanity checks ...
	g_assert_string(Event);

//	std::cout << "k3dObject::Notify() - \"" << Event << "\"" << std::endl;

	// For each client (the loop is a little convoluted since a client could remove itself in response to an event) ...
	ClientIterator nextclient;
	for(ClientIterator client = m_Clients.begin(); client != m_Clients.end(); client = nextclient)
		{
			nextclient = client;
			nextclient++;

			// Skip clients that aren't registered for this event ...
			if(client->m_Event != Event)
				continue;

			// Notify the client about the event ...
			client->m_Client->OnEvent(this, client->m_Event, client->m_ClientData);
		}
}

sdpResult k3dObject::StartQueueEvents()
{
	m_QueueEvents = true;

	return SDP_S_OK;
}

sdpResult k3dObject::EndQueueEvents()
{
	m_QueueEvents = false;
	
	for(EventIterator event = m_Events.begin(); event != m_Events.end(); event++)
		RawNotify(*event);

	m_Events.clear();

	return SDP_S_OK;
}

sdpResult k3dObject::QuerySelectObject()
{
	// By default, allow ourselves to be selected (override and return SDP_S_FALSE	 to prevent)
	return SDP_S_OK;
}

sdpResult k3dObject::ObjectSelected(bool State)
{
	// Notify clients that our selection state changed ...
	if(State)
		Notify("selected");
	else
		Notify("deselected");

	return SDP_S_OK;
}

sdpResult k3dObject::QueryRenameObject()
{
	// By default, allow ourselves to be renamed (override and return SDP_S_FALSE to prevent)
	return SDP_S_OK;
}

sdpResult k3dObject::ObjectRenamed(SDPCSTRING NewName)
{
	// Notify our clients that we got renamed ...
	Notify("renamed");

	return SDP_S_OK;
}

sdpResult k3dObject::QueryDeleteObject()
{
	// By default, allow ourselves to be deleted (override and return SDP_S_FALSE to prevent)
	return SDP_S_OK;
}

sdpResult k3dObject::ObjectDeleted()
{
	// Notify our clients that we're going away (don't allow it to be queued)
	RawNotify("deleted");
	
	return SDP_S_OK;
}


