// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dObjectSelector class, which provides a user interface for either creating new, or selecting existing, K-3D objects
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dobjectselector.h>
#include <k3dvariable.h>

#include <k3diapplication.h>
#include <k3diobjectcontext.h>
#include <k3diobjectinfo.h>

#include <sdpgtkevents.h>

k3dObjectSelector::k3dObjectSelector()
{
	m_Document = 0;
	m_ObjectContainer = 0;
}

k3dObjectSelector::~k3dObjectSelector()
{
}

// UI creation ...
bool k3dObjectSelector::Create(k3dIDocument* Document, k3dObjectVariableBase* ObjectContainer, bool AllowNew, bool AllowExisting, bool AllowNone, sdpCLSID SelectionClass, sdpIID SelectionInterface)
{
	// Sanity checks ...
	g_assert(Document);
	g_assert(ObjectContainer);
	// We require a selection class or interface, but not both
	g_assert(SelectionClass != sdpCLSIDNull || SelectionInterface != sdpIIDNull);
	g_assert(!(SelectionClass != sdpCLSIDNull && SelectionInterface != sdpIIDNull));

	// Get rid of any old stuff ...
	Release();

	// Store data ...
	m_Document = Document;
	m_ObjectContainer = ObjectContainer;
	m_AllowNew = AllowNew;
	m_AllowExisting = AllowExisting;
	m_AllowNone = AllowNone;
	m_SelectionClass = SelectionClass;
	m_SelectionInterface = SelectionInterface;

	// Create our UI template ...
	sdpString uitemplate =
	
		"<gtkml>"
			"<hbox homogeneous=\"false\">"
				"<optionmenu name=\"optionmenu\" expand=\"true\" fill=\"true\">"
					"<event signal=\"destroy\" name=\"ondestroy\"/>"
					"<menu name=\"menu\"/>"
				"</optionmenu>"
				"<button name=\"edit\">"
					"<event signal=\"clicked\" name=\"onedit\"/>"
					"<arrow direction=\"right\" shadowtype=\"in\"/>"
				"</button>"
			"</hbox>"
		"</gtkml>\n";

	// Load our template ...
	g_return_val_if_fail(Load(uitemplate, ""), false);

	// Get our menu object ...
	sdpGtkMenu menu = Menu("menu");
	menu.Show();
	
	// Keep track of the menu item index so we can select one ...
	guint index = 0;
	guint selecteditem = 0;

	// If we give the user the option of selecting nothing ...
	if(m_AllowNone)
		{
			// Create the "none" option ...
			sdpGtkMenuItem menuitem;
			menuitem.Create("--None--");
			menuitem.Show();
			menu.Append(menuitem);

			MapEvent("activate", "onnone", menuitem, true);
			
			index++;
		}
		
	// If we give the user the option of creating new objects ...
	if(m_AllowNew)
		{
			// For each type of available object ...
			for(K3DOBJECTINFOHANDLE handle = m_Document->Application()->GetFirstObjectInfo(); handle; handle = m_Document->Application()->GetNextObjectInfo(handle))
				{
					k3dIObjectInfo* info = m_Document->Application()->GetObjectInfo(handle);
					g_assert(info);
					
					// Skip objects that don't fit our criteria ...
					if(m_SelectionClass != sdpCLSIDNull)
						if(info->ClassID() != m_SelectionClass)
							continue;
							
					if(m_SelectionInterface != sdpIIDNull)
						if(SDP_S_OK != info->QueryObjectInterface(m_SelectionInterface))
							continue;
					
					// Create a "new" menu item ...
					sdpGtkMenuItem menuitem;
					menuitem.Create(sdpFormatString("--New %s--", info->Name()));
					menuitem.Show();
					menu.Append(menuitem);
	
					MapEvent("activate", sdpFormatString("new-%d", info), menuitem, true);
	
					index++;
				}
		}

	// If we give the user the option of selecting existing objects ...
	if(m_AllowExisting)
		{
			// For each object in the document ...
			for(K3DOBJECTCONTEXTHANDLE handle = m_Document->GetFirstObjectContext(); handle; handle = m_Document->GetNextObjectContext(handle))
				{
					k3dIObjectContext* context = m_Document->GetObjectContext(handle);

					// Skip objects that don't fit our criteria ...
					if(m_SelectionClass != sdpCLSIDNull)
						if(context->ObjectInfo()->ClassID() != m_SelectionClass)
							continue;
							
					if(m_SelectionInterface != sdpIIDNull)
						if(SDP_S_OK != context->ObjectInfo()->QueryObjectInterface(m_SelectionInterface))
							continue;

					// Create a menu item for the object ...
					sdpGtkMenuItem menuitem;
					menuitem.Create(context->Name());
					menuitem.Show();
					menu.Append(menuitem);			

					MapEvent("activate", sdpFormatString("select-%d", context), menuitem, true);

					if(context->ObjectID() == m_ObjectContainer->ObjectID())
						selecteditem = index;
	
					index++;
				}
		}
		
	// Add the menu to our optionmenu ...
	OptionMenu("optionmenu").SetMenu(menu);
	
	// Select the right item ...
	OptionMenu("optionmenu").SetHistory(selecteditem);

	// Map some default events ...
	g_return_val_if_fail(MapEvent("destroy", "ondestroy", Root(), true), false);

	return true;
}

bool k3dObjectSelector::Create(k3dIDocument* Document, k3dObjectVariableBase* ObjectContainer, bool AllowNew, bool AllowExisting, bool AllowNone, sdpCLSID SelectionClass, sdpIID SelectionInterface, GtkContainer* Parent)
{
	// Sanity checks ...
	g_assert(Parent);
	
	// Create the UI ...
	g_return_val_if_fail(Create(Document, ObjectContainer, AllowNew, AllowExisting, AllowNone, SelectionClass, SelectionInterface), false);
	
	// Attach ourselves to the parent widget ...
	sdpGtkContainer(Parent).Attach(RootWidget());
	
	return true;
}

void k3dObjectSelector::Release()
{
	// No more events from this point forward ...
	DisconnectAllEvents();

	// Delete all of the GTK+ widgets ...
	if(Root())
		RootWidget().Destroy();
		
	// Clear out our storage ...
	Clear();
}

// Custom event handling ...
void k3dObjectSelector::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "ondestroy")
		OnDestroy();
	else if(Event->Name() == "onnone")
		OnNone();
	else if(Event->Name() == "onedit")
		OnEdit();
	else if(Event->Name().find("select-") != sdpString::npos)
		OnSelect(Event->Name());
	else if(Event->Name().find("new-") != sdpString::npos)
		OnNew(Event->Name());
	else
		sdpGtkObjectContainer::OnEvent(Event);
}

// Called when we're going away ...
void k3dObjectSelector::OnDestroy()
{
	// No more events from this point forward ...
	DisconnectAllEvents();

	// Clear out our storage ...
	Clear();
}

// Called when the user chooses "new" ...
void k3dObjectSelector::OnNew(sdpString EventName)
{
	// Sanity checks ...
	g_return_if_fail(m_ObjectContainer);
	
	// Create a new object ...
	sdpString infotext = EventName.right(EventName.size() - 4);
	m_ObjectContainer->AttachObject(m_Document->CreateObject(reinterpret_cast<k3dIObjectInfo*>(atol(infotext))->ClassID(), true, true, true));
}

// Called when the user chooses "none" ...
void k3dObjectSelector::OnNone()
{
	// Sanity checks ...
	g_return_if_fail(m_ObjectContainer);
	
	m_ObjectContainer->ReleaseObject();
}

// Called when the user clicks the "edit" button ...
void k3dObjectSelector::OnEdit()
{
	// Sanity checks ...
	g_return_if_fail(m_ObjectContainer);
	
	if(m_ObjectContainer->Object())
		m_ObjectContainer->Object()->EditObject();
}

// Called when the user clicks on our list of existing objects ...
void k3dObjectSelector::OnSelect(sdpString EventName)
{
	// Sanity checks ...
	g_return_if_fail(m_ObjectContainer);
	
	sdpString contexttext = EventName.right(EventName.size() - 7);
	SetObject(reinterpret_cast<k3dIObjectContext*>(atol(contexttext)));
}

void k3dObjectSelector::SetObject(k3dIObjectContext* Context)
{
	// Sanity checks ...
	g_return_if_fail(m_ObjectContainer);
	
	// If the user chose the same object, we're done ...
	if(m_ObjectContainer->Object())
		if(Context == m_ObjectContainer->Object()->ObjectContext())
			return;
		
	// Set the new object ...
	m_ObjectContainer->AttachObject(Context->Object());
}



