
// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dPersistObject class, which encapsulates a K-3D object capable of automatically serializing itself
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dpersistobject.h>
#include <k3diobjectcontext.h>
#include <k3dvariable.h>
#include <k3dmisc.h>

#include <sdpxml.h>

//////////////////////////////////////////////////////////////////////////////////
// k3dPersistObject

k3dPersistObject::k3dPersistObject()
{
}


k3dPersistObject::~k3dPersistObject()
{
}


///////////////////////////////////////////////////////////////////////////////////
// k3dIPersist overrides

sdpResult k3dPersistObject::InitNew()
{
	return SDP_S_OK;
}

sdpResult k3dPersistObject::Load(sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_return_val_if_fail(Element, SDP_E_FAIL);

	// Load Objects
	g_return_val_if_fail(sdpSucceeded(LoadObjects(Element)), SDP_E_FAIL);

	// Load Variables
	g_return_val_if_fail(sdpSucceeded(LoadVariables(Element)), SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult k3dPersistObject::LoadComplete()
{
	// Reconnect objects ...
	g_return_val_if_fail(sdpSucceeded(BindObjects()), SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult k3dPersistObject::Save(sdpXmlIElement* Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert(Dependencies);

	// Save objects
	g_return_val_if_fail(sdpSucceeded(SaveObjects(Element, Dependencies)), SDP_E_FAIL);

	// Save variables
	g_return_val_if_fail(sdpSucceeded(SaveVariables(Element, Dependencies)), SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult k3dPersistObject::SaveComplete()
{
	return SDP_S_OK;
}


///////////////////////////////////////////////////////////////////////////////
// Helper functions

sdpResult k3dPersistObject::SaveObjects(sdpXmlIElement* Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert(Dependencies);

	// See if we have any objects to save ...
	if(!m_Objects.size())
		return SDP_S_FALSE;

	// Save objects ...
	sdpXmlIElement* objects = sdpXmlCreateElement(Element, "objects");

	for(ObjectIterator i = m_Objects.begin(); i != m_Objects.end(); i++)
		g_return_val_if_fail(sdpSucceeded((*i)->Save(objects, Dependencies)), SDP_E_FAIL);

	return SDP_S_OK;
}


sdpResult k3dPersistObject::LoadObjects(sdpXmlIElement* Element)
{
	// See if the element contains any objects ...
	sdpXmlIElement* objects = sdpXmlFindElement(Element, "objects");
	if(!objects)
		return SDP_S_FALSE;

	// For each object ...
	for(ObjectIterator object = m_Objects.begin(); object != m_Objects.end(); object++)
		{
			// See if it has an element ...
			sdpXmlIElement* objectelement = sdpXmlFindElement(objects, "object", "name", (*object)->Name());
			if(!objectelement)
				continue;

			// Load it up!
			(*object)->Load(objectelement);
		}

	return SDP_S_OK;
}


class k3dPersistObjectBindObject
{
public:
	k3dPersistObjectBindObject(k3dID ObjectID) { m_ObjectID = ObjectID; m_Context = 0; }

	bool HandleObject(k3dIObjectContext* Context)
	{
		// If this object has the same ID as the one we're looking for ...
		if(Context->ObjectID() == m_ObjectID)
			{
				// Store a pointer to it and break out of the enclosing k3dForEachObject() loop ...
				m_Context = Context;
				return false;
			}

		// Didn't find it, so keep looking ...
		return true;
	}

	k3dID m_ObjectID;
	k3dIObjectContext* m_Context;
};


sdpResult k3dPersistObject::BindObjects()
{
	// For each object we contain ...
	for(ObjectIterator object = m_Objects.begin(); object != m_Objects.end(); object++)
		{
			// Find a document object with the same ID ...
			k3dPersistObjectBindObject functor((*object)->ObjectID());
			k3dForEachObject(Document(), 0, functor);
			g_assert(functor.m_Context);

			// Bind it to our internal object ...
			(*object)->AttachObject(functor.m_Context->Object());
		}

	return SDP_S_OK;
}


sdpResult k3dPersistObject::SaveVariables(sdpXmlIElement* Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert(Dependencies);

	// See if we have any variables to save ...
	if(!m_Variables.size())
		return SDP_S_FALSE;

	// Save variables ...
	sdpXmlIElement* variables = sdpXmlCreateElement(Element, "variables");

	for(VariableIterator variable = m_Variables.begin(); variable != m_Variables.end(); variable++)
		(*variable)->Save(variables, Dependencies);

	return SDP_S_OK;
}



sdpResult k3dPersistObject::LoadVariables(sdpXmlIElement* Element)
{
	// See if the element contains any variables ...
	sdpXmlIElement* variables = sdpXmlFindElement(Element, "variables");
	if(!variables)
		return SDP_S_FALSE;

	// For each variable ...
	for(VariableIterator variable = m_Variables.begin(); variable != m_Variables.end(); variable++)
		{
			// See if it has an element ...
			sdpXmlIElement* variableelement = sdpXmlFindElement(variables, "variable", "name", (*variable)->Name());
			if(!variableelement)
				continue;

			// Load it up!
			(*variable)->Load(variableelement);
		}

	return SDP_S_OK;
}


