// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dPoint class, which implements a geometric point
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dpoint.h>
#include <k3dmesh.h>

#include <k3diactions.h>
#include <k3didocument.h>

#include <sdpxmlutility.h>

/////////////////////////////////////////////////////////////////////////
// k3dPoint

k3dPoint::k3dPoint(k3dMesh* Mesh, k3dID ID)
{
	m_Mesh = Mesh;
	m_ID = ID;

	SetState(k3dIState::VISIBLE | k3dIState::SELECTABLE | k3dIState::USERVIEWABLE | k3dIState::SAVEABLE, true);
}

k3dPoint::~k3dPoint()
{
/*
	// Reset active point
	if(m_Mesh->m_Layer->m_Document->m_ActivePoint == this)
		m_Mesh->m_Layer->m_Document->m_ActivePoint = 0;
*/
}

sdpResult k3dPoint::Save(sdpXmlIElement* Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert(Dependencies);

	sdpXmlIElement* point = sdpXmlCreateElement(Element, "point", 0, "position", m_OldXYZ, "id", m_ID);
	k3dState::Save(point, Dependencies);

	return SDP_S_OK;
}

sdpResult k3dPoint::Load(sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_assert(Element);

	// The point ID is required ...
	g_return_val_if_fail(sdpXmlGetAttribute(Element, "id", m_ID), SDP_E_FAIL);

	// Load the point coordinates ...
	sdpXmlGetAttribute(Element, "position", m_OldXYZ);
	m_LocalXYZ = m_OldXYZ;

	// Load our state ...
	k3dState::Load(Element);

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////////////////
// k3dIPoint overrides

k3dIState* k3dPoint::State()
{
	return this;
}

k3dILocation* k3dPoint::Location()
{
	return this;
}

k3dIMesh* k3dPoint::Mesh()
{
	return m_Mesh;
}

const k3dID k3dPoint::ID()
{
	return m_ID;
}

///////////////////////////////////////////////////////////////////////////
// k3dIState overrides ...

sdpResult k3dPoint::SetState(gulong Mask, bool State)
{
	// See if anything changed ...
	gulong changed = State ? ((Mask^m_StateFlags) & Mask) : (Mask & m_StateFlags);
	if(changed)
		{
			// Are we recording anything?
			if(SDP_S_OK == m_Mesh->Document()->Actions()->IsRecording())
				{
					// If so, store our old data ...
					gulong oldvalue(m_StateFlags);

					// Actually change our flags ...
					k3dState::SetState(Mask, State);

					// Record what happened ...
					m_Mesh->Document()->Actions()->RecordStep(this, STEP_STATE, oldvalue, 0, STEP_STATE, m_StateFlags, 0);
				}
			else
				{
					// Otherwise, just change the flags ...
					k3dState::SetState(Mask, State);
				}
				
			return SDP_S_OK;
		}

	// Nothing changed ...
	return SDP_S_FALSE;
}

///////////////////////////////////////////////////////////////////////////
// k3dIActionClient overrides ...

sdpResult k3dPoint::HandleStep(gulong ValueCode, gulong Value, gulong ValueSize)
{
	if(ValueCode == STEP_STATE)
		{
			m_StateFlags = Value;
			return SDP_S_OK;
		}
	else if(ValueCode == STEP_POSITION)
		{
			m_LocalXYZ = *reinterpret_cast<sdpVector3*>(Value);
			return SDP_S_OK;
		}
		
	return k3dActionClient::HandleStep(ValueCode, Value, ValueSize);
}

///////////////////////////////////////////////////////////////////////////
// k3dILocation overrides ...

/*
k3dIState* k3dPoint::State()
{
	return this;
}
*/

sdpResult k3dPoint::GetOldXYZ(double Coords[3])
{
	m_OldXYZ.CopyArray(Coords);
	return SDP_S_OK;
}

sdpResult k3dPoint::GetLocalXYZ(double Coords[3])
{
	m_LocalXYZ.CopyArray(Coords);
	return SDP_S_OK;
}

sdpResult k3dPoint::GetAnimXYZ(double Coords[3])
{
	m_AnimXYZ.CopyArray(Coords);
	return SDP_S_OK;
}

sdpResult k3dPoint::GetWorldXYZ(double Coords[3])
{
	m_WorldXYZ.CopyArray(Coords);
	return SDP_S_OK;
}

sdpResult k3dPoint::GetCameraXYZ(double Coords[3])
{
	m_CameraXYZ.CopyArray(Coords);
	return SDP_S_OK;
}

sdpResult k3dPoint::GetScreenXYZ(double Coords[3])
{
	m_ScreenXYZ.CopyArray(Coords);
	return SDP_S_OK;
}

sdpResult k3dPoint::GetNormalXYZ(double Coords[3])
{
	return SDP_S_FALSE;
}

sdpResult k3dPoint::GetTextureXYZ(double Coords[3])
{
	return SDP_S_FALSE;
}

sdpResult k3dPoint::GetImplicitTextureXYZ(double Coords[3])
{
	return SDP_S_FALSE;
}

sdpResult k3dPoint::SetOldXYZ(double Coords[3])
{
	m_OldXYZ = Coords;
	return SDP_S_OK;
}

sdpResult k3dPoint::SetLocalXYZ(double Coords[3])
{
	// Are we recording actions?
	if(SDP_S_OK == m_Mesh->Document()->Actions()->IsRecording())
		{
			// Store record what changed ...
			m_Mesh->Document()->Actions()->RecordStep(this, STEP_POSITION, reinterpret_cast<gulong>(new sdpVector3(m_OldXYZ)), sizeof(sdpVector3),
																										STEP_POSITION, reinterpret_cast<gulong>(new sdpVector3(Coords)), sizeof(sdpVector3));

			// Make the change ...
			m_LocalXYZ = Coords;
		}
	// We're not, so just do it!
	else
		{
			// Make the change ...
			m_LocalXYZ = Coords;
		}

	return SDP_S_OK;
}

sdpResult k3dPoint::SetAnimXYZ(double Coords[3])
{
	m_AnimXYZ = Coords;
	return SDP_S_OK;
}

sdpResult k3dPoint::SetWorldXYZ(double Coords[3])
{
	m_WorldXYZ = Coords;
	return SDP_S_OK;
}

sdpResult k3dPoint::SetCameraXYZ(double Coords[3])
{
	m_CameraXYZ = Coords;
	return SDP_S_OK;
}

sdpResult k3dPoint::SetScreenXYZ(double Coords[3])
{
	m_ScreenXYZ = Coords;
	return SDP_S_OK;
}

sdpResult k3dPoint::SetNormalXYZ(double Coords[3])
{
	return SDP_S_FALSE;
}

sdpResult k3dPoint::SetTextureXYZ(double Coords[3])
{
	return SDP_S_FALSE;
}

sdpResult k3dPoint::SetImplicitTextureXYZ(double Coords[3])
{
	return SDP_S_FALSE;
}

