// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dPolygon class, which encapsulates a geometric polygon
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include <k3dpolygon.h>
#include <k3dpolypoint.h>
#include <k3dpoint.h>
#include <k3dmesh.h>
#include <k3dmisc.h>

#include <k3diactions.h>
#include <k3didocument.h>
#include <k3dimaterial.h>

#include <sdpgl.h>
#include <sdpxmlutility.h>

#include <algorithm>

/////////////////////////////////////////////////////////////////////////
// k3dPolygon

k3dPolygon::k3dPolygon(k3dMesh* Mesh, k3dID ID)
{
	m_Mesh = Mesh;
	m_ID = ID;
	m_Material = 0;

	SetState(k3dIState::VISIBLE | k3dIState::SELECTABLE | k3dIState::USERVIEWABLE | k3dIState::SAVEABLE, true);
}


class k3dPolygonDeletePolyPoint
{
public:
	void operator()(k3dPolyPoint* PolyPoint) { delete PolyPoint; }
};

k3dPolygon::~k3dPolygon()
{
/*
	// Reset the active polygon
	if(m_Mesh->m_Layer->m_Document->m_ActivePolygon == this)
		m_Mesh->m_Layer->m_Document->m_ActivePolygon = 0;
*/

	// Delete polypoints ...
	k3dPolygonDeletePolyPoint deletepolypoint;
	std::for_each(m_PolyPoints.begin(), m_PolyPoints.end(), deletepolypoint);
}

void k3dPolygon::CalculateNormal()
{
	// Reset the normal to zero ...
	m_Normal = 0.0;

	// Skip degenerate polys ...
	gulong pointcount = m_PolyPoints.size();
	if(pointcount < 3)
		return;

	// Integrate the polygon normal from its' edge normals ...
	sdpVector3 normal1;
	sdpVector3 normal2;

	for(gulong i = 2; i < pointcount; i++)
		{
			normal1 = m_PolyPoints[i]->m_Point->m_AnimXYZ - m_PolyPoints[i-1]->m_Point->m_AnimXYZ;
			normal2 = m_PolyPoints[i-1]->m_Point->m_AnimXYZ - m_PolyPoints[i-2]->m_Point->m_AnimXYZ;

			m_Normal += normal1 ^ normal2;
		}

	normal1 = m_PolyPoints[0]->m_Point->m_AnimXYZ - m_PolyPoints[pointcount-1]->m_Point->m_AnimXYZ;
	normal2 = m_PolyPoints[pointcount-1]->m_Point->m_AnimXYZ - m_PolyPoints[pointcount-2]->m_Point->m_AnimXYZ;

	m_Normal += normal1 ^ normal2;
}

sdpResult k3dPolygon::Save(sdpXmlIElement* Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert(Dependencies);

	k3dID materialid = 0;
	if(m_Material)
		{
			k3dIObject* object = sdpQueryInterface<k3dIObject>(m_Material);
			g_assert(object);
			materialid = object->ObjectContext()->ObjectID();
		}

	sdpXmlIElement* polygon = sdpXmlCreateElement(Element, "polygon", 0, "material", materialid, "id", m_ID);

	// Write properties ...
	k3dState::Save(polygon, Dependencies);

	// Write the polypoints ...
	if(m_PolyPoints.size())
		{
			sdpXmlIElement* polypoints = sdpXmlCreateElement(polygon, "polypoints");

			for(PolyPointIterator polypoint = m_PolyPoints.begin(); polypoint != m_PolyPoints.end(); polypoint++)
				(*polypoint)->Save(polypoints, Dependencies);
		}

	return SDP_S_OK;
}

class k3dPolygonLoadPolyPoint
{
public:
	k3dPolygonLoadPolyPoint(k3dPolygon* Polygon) { m_Polygon = Polygon; }
	
	bool HandleElement(sdpXmlIElement* Element)
	{
		// Sanity checks ...
		g_assert(Element);

		// Create a new polypoint and push it into the polypoint list ...
		k3dPolyPoint* polypoint = new k3dPolyPoint(m_Polygon);
		m_Polygon->m_PolyPoints.push_back(polypoint);
		g_assert(polypoint);

		// Load the polypoint and attach it to its point ...
		polypoint->Load(Element);

		return true;
	}
protected:
	k3dPolygon* m_Polygon;
};

sdpResult k3dPolygon::Load(sdpXmlIElement* Element, k3dMesh* Mesh)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert(Mesh);

	// Read the ID (required) ...
	g_return_val_if_fail(sdpXmlGetAttribute(Element, "id", m_ID), SDP_E_FAIL);

	// Read properties ...
	k3dState::Load(Element);

	// Read polygon material ...
	k3dID materialid = 0;
	sdpXmlGetAttribute(Element, "material", materialid);

	// Store the material ID for binding later ...
	m_Material = reinterpret_cast<k3dIMaterial*>(materialid);

	// Load polypoints
	sdpXmlIElement* polypoints = sdpXmlFindElement(Element, "polypoints");
	if(polypoints)
		{
			k3dPolygonLoadPolyPoint functor(this);
			sdpXmlForEachElement(polypoints, "polypoint", functor);
		}

	return SDP_S_OK;
}

sdpResult k3dPolygon::LoadComplete()
{
	// Bind our material ID to a material ...
	k3dID materialid = reinterpret_cast<k3dID>(m_Material);
	if(materialid)
		{
			k3dIObject* materialobject = k3dGetObject(m_Mesh->Document(), materialid);
			if(materialobject)
				m_Material = sdpQueryInterface<k3dIMaterial>(materialobject);
			else
				m_Material = 0;
		}

	// Give polypoints a chance to bind to points ...
	for(PolyPointIterator polypoint = m_PolyPoints.begin(); polypoint != m_PolyPoints.end(); polypoint++)
		(*polypoint)->LoadComplete(m_Mesh);

	return SDP_S_OK;
}

//////////////////////////////////////////
// k3dIPolygon overrides
k3dIMesh* k3dPolygon::Mesh()
{
	return m_Mesh;
}

const k3dID k3dPolygon::ID()
{
	return m_ID;
}

k3dIPolyPointServer* k3dPolygon::Points()
{
	return this;
}

k3dILocationServer* k3dPolygon::Locations()
{
	return this;
}

k3dIMaterialServer* k3dPolygon::Material()
{
	return this;
}

k3dIState* k3dPolygon::State()
{
	return this;
}

sdpResult k3dPolygon::GetPolygonNormal(double Normal[3])
{
	m_Normal.CopyArray(Normal);
	return SDP_S_OK;
}

sdpResult k3dPolygon::SetPolygonNormal(double Normal[3])
{
	m_Normal = Normal;
	return SDP_S_OK;
}

/////////////////////////////////////////
// k3dIPolyPointServer overrides

gulong k3dPolygon::GetPointCount()
{
	return m_PolyPoints.size();
}

K3DPOLYPOINTHANDLE k3dPolygon::GetFirstPolyPoint()
{
	return K3DPOLYPOINTHANDLE(m_PolyPoints.size() ? 1 : 0);
}

K3DPOLYPOINTHANDLE k3dPolygon::GetNextPolyPoint(K3DPOLYPOINTHANDLE Point)
{
	// Sanity checks ...
	g_return_val_if_fail(Point, 0);
	g_return_val_if_fail(gulong(Point) <= m_PolyPoints.size(), 0);

	return K3DPOLYPOINTHANDLE(gulong(Point) < m_PolyPoints.size() ? gulong(Point)+1 : 0);
}

k3dIPolyPoint* k3dPolygon::GetPolyPoint(K3DPOLYPOINTHANDLE Point)
{
	// Sanity checks ...
	g_return_val_if_fail(Point, 0);
	g_return_val_if_fail(gulong(Point) <= m_PolyPoints.size(), 0);

	return m_PolyPoints[gulong(Point)-1];
}

k3dIPolyPoint* k3dPolygon::AddPoint(k3dIPoint* Point)
{
	// Sanity checks ...
	g_return_val_if_fail(Point, 0);

	k3dPolyPoint* polypoint = new k3dPolyPoint(this, static_cast<k3dPoint*>(Point), m_Mesh->GetNextPolyPointID());
	g_return_val_if_fail(polypoint, 0);
	m_PolyPoints.push_back(polypoint);

	return polypoint;
}

sdpResult k3dPolygon::RemovePoint(k3dIPoint* Point)
{
g_assert_not_reached();
return 0;
/*
	// Sanity check ...
	g_return_val_if_fail(Point, SDP_E_FAIL);

	k3dPolyPoint* next;
	for(k3dPolyPoint* polypoint = m_PolyPoints.front(); polypoint; polypoint = next)
		{
			next = m_PolyPoints[i]->m_Next;
			if(m_PolyPoints[i]->m_Point != (k3dPoint3D*)Point) continue;

			m_PolyPoints.erase(polypoint);
		}

	return SDP_S_OK;
*/
}

sdpResult k3dPolygon::RemovePolyPoint(k3dIPolyPoint* PolyPoint)
{
g_assert_not_reached();
return 0;
/*
	// Sanity check ...
	g_return_val_if_fail(PolyPoint, SDP_E_FAIL);

	for(k3dPolyPoint* polypoint = m_PolyPoints.front(); polypoint; polypoint = m_PolyPoints[i]->m_Next)
		{
			if(polypoint == (k3dPolyPoint*)PolyPoint)
				{
					m_PolyPoints.erase((k3dPolyPoint*)PolyPoint);
					return SDP_S_OK;
				}
		}

	g_return_val_if_fail(0, SDP_E_FAIL);
*/
}

/////////////////////////////////////////
// k3dILocationServer overrides

gulong k3dPolygon::GetLocationCount()
{
	return m_PolyPoints.size();
}

K3DLOCATIONHANDLE k3dPolygon::GetFirstLocation()
{
	return K3DLOCATIONHANDLE(m_PolyPoints.size() ? 1 : 0);
}

K3DLOCATIONHANDLE k3dPolygon::GetNextLocation(K3DLOCATIONHANDLE Location)
{
	// Sanity checks ...
	g_return_val_if_fail(Location, 0);
	g_return_val_if_fail(gulong(Location) <= m_PolyPoints.size(), 0);

	return K3DLOCATIONHANDLE(gulong(Location) < m_PolyPoints.size() ? gulong(Location)+1 : 0);
}

k3dILocation* k3dPolygon::GetLocation(K3DLOCATIONHANDLE Location)
{
	// Sanity checks ...
	g_return_val_if_fail(Location, 0);
	g_return_val_if_fail(gulong(Location) <= m_PolyPoints.size(), 0);

	return m_PolyPoints[gulong(Location)-1]->m_Point;
}

/////////////////////////////////////////
// k3dIMaterialServer overrides

k3dIMaterial* k3dPolygon::GetMaterial()
{
	return m_Material;
}

sdpResult k3dPolygon::SetMaterial(k3dIMaterial* Material)
{
	m_Material = Material;
	return SDP_S_OK;
}


/////////////////////////////////////////
// k3dIState overrides

class k3dPolygonSetState
{
public:
	k3dPolygonSetState(gulong Mask, bool State) { m_Mask = Mask; m_State = State; }

	void operator()(k3dPolyPoint* PolyPoint) { PolyPoint->State()->SetState(m_Mask, m_State); }

	gulong m_Mask;
	bool m_State;
};


sdpResult k3dPolygon::SetState(gulong Mask, bool State)
{
	// See if anything changed ...
	gulong changed = State ? ((Mask^m_StateFlags) & Mask) : (Mask & m_StateFlags);
	if(changed)
		{
			// Are we recording actions?
			if(SDP_S_OK == m_Mesh->Document()->Actions()->IsRecording())
				{
					// If so, store our old data ...
					gulong oldvalue(m_StateFlags);

					// Actually change our flags ...
					k3dState::SetState(Mask, State);

					// Record the new data ...
					m_Mesh->Document()->Actions()->RecordStep(this, STEP_STATE, oldvalue, 0,
																												STEP_STATE, m_StateFlags, 0);
				}
			// We're not, so just do it ...
			else
				{
					// Actually change our flags ...
					k3dState::SetState(Mask, State);
				}

			// Force all our points to the same state ...
			k3dPolygonSetState functor(Mask, State);
			std::for_each(m_PolyPoints.begin(), m_PolyPoints.end(), functor);
		}

	// Let the caller know whether anything happened or not ...
	return changed ? SDP_S_OK : SDP_S_FALSE;
}


///////////////////////////////////////////////////////////////////////////
// k3dIActionClient overrides ...

sdpResult k3dPolygon::HandleStep(gulong ValueCode, gulong Value, gulong ValueSize)
{
	if(ValueCode == STEP_STATE)
		{
			m_StateFlags = Value;
			return SDP_S_OK;
		}
		
	return k3dActionClient::HandleStep(ValueCode, Value, ValueSize);
}

////////////////////////////////////////////////////
// OpenGL rendering ...

void k3dPolygon::OpenGLWireframe()
{
	glBegin(GL_LINE_LOOP);

	for(gulong i = 0; i < m_PolyPoints.size(); i++)
		glVertex3dv(m_PolyPoints[i]->m_Point->m_AnimXYZ);

	glEnd();
}

void k3dPolygon::OpenGLWireframeOrientation(sdpVector3& FirstColor, sdpVector3& Color)
{
	glBegin(GL_LINE_STRIP);

	if(m_PolyPoints.size())
		{
			glColor3dv(FirstColor);
			glVertex3dv(m_PolyPoints[0]->m_Point->m_AnimXYZ);
			glColor3dv(Color);

			for(gulong i = 1; i < m_PolyPoints.size(); i++)
				glVertex3dv(m_PolyPoints[i]->m_Point->m_AnimXYZ);

			glColor3dv(Color);
			glVertex3dv(m_PolyPoints[0]->m_Point->m_AnimXYZ);
		}

	glEnd();
}

void k3dPolygon::OpenGLSolid()
{
	glBegin(GL_POLYGON);
	for(gulong i = 0; i < m_PolyPoints.size(); i++)
		{
			glNormal3dv(-m_PolyPoints[i]->m_NormalXYZ);
			glVertex3dv(m_PolyPoints[i]->m_Point->m_AnimXYZ);
		}
	glEnd();
}

void k3dPolygon::OpenGLSolidTextured()
{
	glBegin(GL_POLYGON);
	for(gulong i = 0; i < m_PolyPoints.size(); i++)
		{
			glTexCoord3dv(m_PolyPoints[i]->m_TextureXYZ);
			glNormal3dv(-m_PolyPoints[i]->m_NormalXYZ);
			glVertex3dv(m_PolyPoints[i]->m_Point->m_AnimXYZ);
		}
	glEnd();
}

