// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dPolyPoint class, which encapsulates the many-to-many relationship between k3dIPoint and k3dIPolygon/k3dINURBSCurve
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dpolypoint.h>
#include <k3dpoint.h>
#include <k3dmesh.h>

#include <k3diactions.h>
#include <k3didocument.h>

#include <sdpxmlutility.h>

/////////////////////////////////////////////////////////////////////////
// k3dPolyPoint

k3dPolyPoint::k3dPolyPoint()
{
	m_Polygon = 0;
	m_Point = 0;
	m_ID = 0;
}

k3dPolyPoint::k3dPolyPoint(k3dPolygon* Polygon)
{
	// Sanity checks ...
	g_assert(Polygon);
	
	m_Polygon = Polygon;
	m_Point = 0;
	m_ID = 0;
}

k3dPolyPoint::k3dPolyPoint(k3dPoint* Point, k3dID ID)
{
	// Sanity checks ...
	g_assert(Point);
	g_assert(ID);

	m_Polygon = 0;
	m_Point = Point;
	m_ID = ID;
}

k3dPolyPoint::k3dPolyPoint(k3dPolygon* Polygon, k3dPoint* Point, k3dID ID)
{
	// Sanity checks ...
	g_assert(Polygon);
	g_assert(Point);
	g_assert(ID);

	m_Polygon = Polygon;
	m_Point = Point;
	m_ID = ID;
}

k3dPolyPoint::~k3dPolyPoint()
{
}

///////////////////////////////////////////////////////
// k3dIPolyPoint overrides

k3dIPoint* k3dPolyPoint::Point()
{
	return m_Point;
}

k3dIState* k3dPolyPoint::State()
{
	return m_Point->State();
}

k3dILocation* k3dPolyPoint::Location()
{
	return this;
}

const k3dID k3dPolyPoint::ID()
{
	return m_ID;
}

///////////////////////////////////////////////
// k3dILocation overrides

/*
k3dIState* k3dPolyPoint::State()
{
	return m_Point->State();
}
*/

sdpResult k3dPolyPoint::GetOldXYZ(double Coords[3])
{
	return m_Point->GetOldXYZ(Coords);
}

sdpResult k3dPolyPoint::GetLocalXYZ(double Coords[3])
{
	return m_Point->GetLocalXYZ(Coords);
}

sdpResult k3dPolyPoint::GetAnimXYZ(double Coords[3])
{
	return m_Point->GetAnimXYZ(Coords);
}

sdpResult k3dPolyPoint::GetWorldXYZ(double Coords[3])
{
	return m_Point->GetWorldXYZ(Coords);
}

sdpResult k3dPolyPoint::GetCameraXYZ(double Coords[3])
{
	return m_Point->GetCameraXYZ(Coords);
}

sdpResult k3dPolyPoint::GetScreenXYZ(double Coords[3])
{
	return m_Point->GetScreenXYZ(Coords);
}

sdpResult k3dPolyPoint::GetNormalXYZ(double Coords[3])
{
	m_NormalXYZ.CopyArray(Coords);
	return SDP_S_OK;
}

sdpResult k3dPolyPoint::GetTextureXYZ(double Coords[3])
{
	m_TextureXYZ.CopyArray(Coords);
	return SDP_S_OK;
}

sdpResult k3dPolyPoint::GetImplicitTextureXYZ(double Coords[3])
{
	m_ImplicitTextureXYZ.CopyArray(Coords);
	return SDP_S_OK;
}

sdpResult k3dPolyPoint::SetOldXYZ(double Coords[3])
{
	return m_Point->SetOldXYZ(Coords);
}

sdpResult k3dPolyPoint::SetLocalXYZ(double Coords[3])
{
	return m_Point->SetLocalXYZ(Coords);
}

sdpResult k3dPolyPoint::SetAnimXYZ(double Coords[3])
{
	return m_Point->SetAnimXYZ(Coords);
}

sdpResult k3dPolyPoint::SetWorldXYZ(double Coords[3])
{
	return m_Point->SetWorldXYZ(Coords);
}

sdpResult k3dPolyPoint::SetCameraXYZ(double Coords[3])
{
	return m_Point->SetCameraXYZ(Coords);
}

sdpResult k3dPolyPoint::SetScreenXYZ(double Coords[3])
{
	return m_Point->SetScreenXYZ(Coords);
}

sdpResult k3dPolyPoint::SetNormalXYZ(double Coords[3])
{
	m_NormalXYZ = Coords;
	return SDP_S_OK;
}

sdpResult k3dPolyPoint::SetTextureXYZ(double Coords[3])
{
	// Are we recording actions?
	if(SDP_S_OK == m_Point->Mesh()->Document()->Actions()->IsRecording())
		{
			// Store record what changed ...
			m_Point->Mesh()->Document()->Actions()->RecordStep(this, STEP_TEXTUREXYZ, reinterpret_cast<gulong>(new sdpVector3(m_TextureXYZ)), sizeof(sdpVector3),
																										STEP_TEXTUREXYZ, reinterpret_cast<gulong>(new sdpVector3(Coords)), sizeof(sdpVector3));

			// Make the change ...
			m_TextureXYZ = Coords;
		}
	// We're not, so just do it!
	else
		{
			// Make the change ...
			m_TextureXYZ = Coords;
		}

	return SDP_S_OK;
}

sdpResult k3dPolyPoint::SetImplicitTextureXYZ(double Coords[3])
{
	m_ImplicitTextureXYZ = Coords;
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////////////////
// k3dIActionClient overrides ...

sdpResult k3dPolyPoint::HandleStep(gulong ValueCode, gulong Value, gulong ValueSize)
{
	if(ValueCode == STEP_TEXTUREXYZ)
		{
			m_TextureXYZ = *reinterpret_cast<sdpVector3*>(Value);
			return SDP_S_OK;
		}
		
	return k3dActionClient::HandleStep(ValueCode, Value, ValueSize);
}

sdpResult k3dPolyPoint::Save(sdpXmlIElement* Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert(Dependencies);

	sdpXmlCreateElement(Element, "polypoint", 0, "id", ID(), "pointid", m_Point->ID(), "texture", m_TextureXYZ, "itexture", m_ImplicitTextureXYZ);

	return SDP_S_OK;
}

sdpResult k3dPolyPoint::Load(sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_assert(Element);

	// Load the polypoint ID (required) ...
	m_ID = 0;
	g_return_val_if_fail(sdpXmlGetAttribute(Element, "id", m_ID), SDP_E_FAIL);
	g_return_val_if_fail(m_ID, SDP_E_FAIL);

	// Grab the point handle (required) and store it for binding later ...
	k3dID pointid = 0;
	g_return_val_if_fail(sdpXmlGetAttribute(Element, "pointid", pointid), SDP_E_FAIL);
	g_return_val_if_fail(pointid, SDP_E_FAIL);
	m_Point = reinterpret_cast<k3dPoint*>(pointid);

	// Load texture coordinates ...
	sdpXmlGetAttribute(Element, "texture", m_TextureXYZ);
	sdpXmlGetAttribute(Element, "itexture", m_ImplicitTextureXYZ);

	return SDP_S_OK;
}

sdpResult k3dPolyPoint::LoadComplete(k3dMesh* Mesh)
{
	// Sanity checks ...
	g_assert(Mesh);

	// Find the point that matches our pointid ...
	m_Point = Mesh->FindPoint(reinterpret_cast<k3dID>(m_Point));
	g_return_val_if_fail(m_Point, SDP_E_FAIL);

	return SDP_S_OK;
}



