// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dPrimTool class, which encapsulates a K-3D tool capable of generating new geometry
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dprimtool.h>

#include <k3dimesh.h>
#include <k3dilayer.h>
#include <k3dipoint.h>
#include <k3dipolygon.h>
#include <k3dipolypoint.h>
#include <k3distate.h>
#include <k3dipointserver.h>
#include <k3dipolygonserver.h>
#include <k3dipolypointserver.h>
#include <k3diobjectcontext.h>

#include <sdpgtkentry.h>
#include <sdpstring.h>

#include <k3dcom.h>
#include <k3dmisc.h>

sdpVector3 k3dPrimTool::m_NewColor(0.0, 0.0, 1.0);
sdpVector3 k3dPrimTool::m_NormalColor(0.0);

k3dPrimTool::k3dPrimTool() : k3dTool()
{
	m_CreateGeometry = true;
}

k3dPrimTool::~k3dPrimTool()
{
	// Get rid of "leftover" geometry
	DeleteAll();
}

// Delete the objects/geometry we created ...
void k3dPrimTool::DeleteAll()
{
	gulong i;
	gulong j;

	// Get rid of new points ...
	for(i = 0; i < m_Points.size(); i++)
		{
			for(j = 0; j < m_Points[i].size(); j++)
				m_Points[i][j]->Mesh()->DeletePoint(m_Points[i][j]);
		}
	m_Points.clear();

	// Get rid of new polygons ...
	for(i = 0; i < m_Polygons.size(); i++)
		{
			for(j = 0; j < m_Polygons[i].size(); j++)
				m_Polygons[i][j]->Mesh()->DeletePolygon(m_Polygons[i][j], false);
		}
	m_Polygons.clear();

	// Get rid of new meshes ...
	for(MeshIterator mesh = m_Meshes.begin(); mesh != m_Meshes.end(); mesh++)
		k3dDeleteObject(*mesh);
	m_Meshes.clear();

	// Get rid of new objects ...
	for(ObjectIterator object = m_Objects.begin(); object != m_Objects.end(); object++)
		k3dDeleteObject(*object);
	m_Objects.clear();
}

// Make the objects/geometry we created permanent ...
void k3dPrimTool::CreateAll()
{
	gulong i;
	gulong j;

	// Tidy up new points ...
	for(i = 0; i < m_Points.size(); i++)
		{
			for(j = 0; j < m_Points[i].size(); j++)
				m_Points[i][j]->State()->SetState(k3dIState::SELECTABLE | k3dIState::SAVEABLE | k3dIState::SELECTED, true);
		}
	m_Points.clear();

	// Tidy up new polygons ...
	for(i = 0; i < m_Polygons.size(); i++)
		{
			for(j = 0; j < m_Polygons[i].size(); j++)
				m_Polygons[i][j]->State()->SetState(k3dIState::SELECTABLE | k3dIState::SAVEABLE | k3dIState::SELECTED, true);
		}
	m_Polygons.clear();

	// Tidy up new meshes ...
	for(MeshIterator mesh = m_Meshes.begin(); mesh != m_Meshes.end(); mesh++)
		{
			k3dIObject* object = sdpQueryInterface<k3dIObject>(*mesh);
			object->ObjectContext()->State()->SetState(k3dIState::SELECTABLE | k3dIState::SAVEABLE | k3dIState::SELECTED, true);
			object->Release();
			
			(*mesh)->SetColor(m_NormalColor);
		}
	m_Meshes.clear();

	// Tidy up new objects ...
	for(ObjectIterator object = m_Objects.begin(); object != m_Objects.end(); object++)
		(*object)->ObjectContext()->State()->SetState(k3dIState::SELECTABLE | k3dIState::SAVEABLE | k3dIState::SELECTED, true);
	m_Objects.clear();
}


gulong k3dPrimTool::PointCount()
{
	gulong count = 0;

	for(gulong i = 0; i < m_Points.size(); i++)
		count += m_Points[i].size();

	return count;
}


gulong k3dPrimTool::PolyCount()
{
	gulong count = 0;

	for(gulong i = 0; i < m_Polygons.size(); i++)
		count += m_Polygons[i].size();

	return count;
}


gulong k3dPrimTool::MeshCount()
{
	return m_Meshes.size();
}


// Display information about objects / geometry ...
void k3dPrimTool::DisplayStatistics(sdpGtkEntry Entry)
{
	sdpString message;

	// Points ...
	gulong pointcount = PointCount();
	if(pointcount)
		{
			if(message.size())
				message += ", ";

			message += sdpFormatString("%u Point", pointcount);

			if(pointcount > 1)
				message += "s";
		}

	// Polys ...
	gulong polycount = PolyCount();
	if(polycount)
		{
			if(message.size())
				message += ", ";

			message += sdpFormatString("%u Poly", polycount);

			if(polycount > 1)
				message += "s";
		}

	// Meshes ...
	gulong meshcount = MeshCount();
	if(meshcount)
		{
			if(message.size())
				message += ", ";

			message += sdpFormatString("%u Mesh", meshcount);

			if(meshcount > 1)
				message += "s";
		}

	// Display it ...
	Entry.SetText(message);
}


/*
void k3dPrimTool::GetNewGeometryDimensions(sdpVector3& Min, sdpVector3& Max)
{
	Min = DBL_MAX;
	Max = -DBL_MAX;

	sdpVector3 coords;

	// Calculate dimensions for the new geometry ...
	for(gulong i = 0; i < m_Points.size(); i++)
		{
			g_assert(m_Points[i]);
			m_Points[i]->Location()->GetOldXYZ(coords);

			if(coords[0] < Min[0]) Min[0] = coords[0];
			if(coords[0] > Max[0]) Max[0] = coords[0];
			if(coords[1] < Min[1]) Min[1] = coords[1];
			if(coords[1] > Max[1]) Max[1] = coords[1];
			if(coords[2] < Min[2]) Min[2] = coords[2];
			if(coords[2] > Max[2]) Max[2] = coords[2];
		}
}


void k3dPrimTool::CenterNewGeometry()
{
	sdpVector3 min, max;
	GetNewGeometryDimensions(min, max);

	sdpVector3 center = (min + max) * 0.5;

	// Center points ...
	sdpVector3 coords;
	for(gulong i = 0; i < m_Points.size(); i++)
		{
			g_assert(m_Points[i]);

			m_Points[i]->Location()->GetOldXYZ(coords);
			coords -= center;
			m_Points[i]->Location()->SetOldXYZ(coords);
			m_Points[i]->Location()->SetLocalXYZ(coords);
		}
}
*/

// Caller wants to reserve space for "Columns" number of points, in a single row ...
void k3dPrimTool::ReservePoints(gulong Columns)
{
	// Sanity checks ...
	g_assert(Columns);

	ReservePoints(1, Columns);
}


// Caller wants to reserve space for "Columns" * "Rows" number of points, arranged as a grid ...
void k3dPrimTool::ReservePoints(gulong Rows, gulong Columns)
{
	// Sanity checks ...
	g_assert(Rows);
	g_assert(Columns);

	// Make space for the requested number of rows ...
	m_Points.resize(Rows);

	// For each row ...
	for(gulong row = 0; row < Rows; row++)
		{
			// Make space for the requested number of points ...
			m_Points[row].resize(Columns);
		}
}

// Create a point and append it to the first row ...
k3dIPoint* k3dPrimTool::CreatePoint(k3dIMesh* Mesh)
{
	// Sanity checks ...
	g_assert(Mesh);
	
	// Create a new slot to insert the created point into ...
	if(m_Points.size())
		ReservePoints(m_Points[0].size()+1);
	else
		ReservePoints(1);
	
	return CreatePoint(Mesh, m_Points[0].size()-1);
}

// Create a point at the given column in the first row ...
k3dIPoint* k3dPrimTool::CreatePoint(k3dIMesh* Mesh, gulong Column)
{
	// Sanity check ...
	g_assert(Mesh);

	return CreatePoint(Mesh, 0, Column);
}

// Create a point at the given row and column ...
k3dIPoint* k3dPrimTool::CreatePoint(k3dIMesh* Mesh, gulong Row, gulong Column)
{
	// Sanity checks ...
	g_assert(Mesh);
	g_assert(Row < m_Points.size());
	g_assert(Column < m_Points[Row].size());

	// Create the point ...
	k3dIPoint* point = Mesh->CreatePoint();
	g_assert(point);

	// Prevent the point from being selected or saved until we're done with it ...
	point->State()->SetState(k3dIState::SELECTABLE | k3dIState::SAVEABLE, false);

	// Store a reference to the point ...
	m_Points[Row][Column] = point;

	return point;
}


// Caller wants to reserve space for "Columns" number of polygons, in a single row ...
void k3dPrimTool::ReservePolygons(gulong Columns)
{
	// Sanity checks ...
	g_assert(Columns);

	ReservePolygons(1, Columns);
}


// Caller wants to reserve space for "Columns" * "Rows" number of polygons, arranged as a grid ...
void k3dPrimTool::ReservePolygons(gulong Rows, gulong Columns)
{
	// Sanity checks ...
	g_assert(Rows);
	g_assert(Columns);

	// Make space for the requested number of rows ...
	m_Polygons.resize(Rows);

	// For each row ...
	for(gulong row = 0; row < Rows; row++)
		{
			// Make space for the requested number of polygons ...
			m_Polygons[row].resize(Columns);
		}
}


// Create a polygon and append it to the first row ...
k3dIPolygon* k3dPrimTool::CreatePolygon(k3dIMesh* Mesh)
{
	// Sanity checks ...
	g_assert(Mesh);
	
	// Create a new slot to insert the created polygon into ...
	if(m_Polygons.size())
		ReservePolygons(m_Polygons[0].size()+1);
	else
		ReservePolygons(1);
	
	return CreatePolygon(Mesh, m_Polygons[0].size()-1);
}

// Create a polygon at the given column in the first row ...
k3dIPolygon* k3dPrimTool::CreatePolygon(k3dIMesh* Mesh, gulong Column)
{
	// Sanity check ...
	g_assert(Mesh);

	return CreatePolygon(Mesh, 0, Column);
}


// Create a polygon at the given row and column ...
k3dIPolygon* k3dPrimTool::CreatePolygon(k3dIMesh* Mesh, gulong Row, gulong Column)
{
	// Sanity checks ...
	g_assert(Mesh);
	g_assert(Row < m_Polygons.size());
	g_assert(Column < m_Polygons[Row].size());

	// Create the polygon ...
	k3dIPolygon* polygon = Mesh->CreatePolygon();
	g_assert(polygon);

	// Prevent the polygon from being selected or saved until we're done with it ...
	polygon->State()->SetState(k3dIState::SELECTABLE | k3dIState::SAVEABLE, false);

	// Store a reference to the polygon ...
	m_Polygons[Row][Column] = polygon;

	return polygon;
}


// Create a mesh ...
k3dIMesh* k3dPrimTool::CreateMesh(SDPCSTRING Name)
{
	// Sanity checks ...
	g_assert_string(Name);

	// Create the mesh ...
	k3dIObject* object = Document()->CreateObject(sdpMesh, false, true, false);
	g_assert(object);
	
	// Assign some properties ...
	object->ObjectContext()->State()->SetState(k3dIState::SELECTABLE | k3dIState::SAVEABLE, false);
	object->ObjectContext()->SetName(Name);

	// Get the mesh interface ...	
	k3dIMesh* mesh = sdpQueryInterface<k3dIMesh>(object);
	g_assert(mesh);

	mesh->SetColor(m_NewColor);

	// Store it ...
	m_Meshes.push_back(mesh);

	return mesh;
}


// Create a mesh with a single row of points and a single row of polygons ...
k3dIMesh* k3dPrimTool::CreateMesh(SDPCSTRING Name, gulong PointColumns, gulong PolygonColumns)
{
	// Sanity checks ...
	g_assert_string(Name);
	g_assert(PointColumns);
	g_assert(PolygonColumns);

	// Create the mesh ...
	k3dIMesh* mesh = CreateMesh(Name);
	g_assert(mesh);

	// Create points ...
	ReservePoints(PointColumns);
	for(gulong pointcol = 0; pointcol < PointColumns; pointcol++)
		CreatePoint(mesh, pointcol);

	// Create polygons ...
	ReservePolygons(PolygonColumns);
	for(gulong polycol = 0; polycol < PolygonColumns; polycol++)
		CreatePolygon(mesh, polycol);

	return mesh;
}


// Create a mesh with a multiple rows of points and multiple rows of polygons ...
k3dIMesh* k3dPrimTool::CreateMesh(SDPCSTRING Name, gulong PointRows, gulong PointColumns, gulong PolygonRows, gulong PolygonColumns)
{
	// Sanity checks ...
	g_assert_string(Name);
	g_assert(PointRows);
	g_assert(PointColumns);
	g_assert(PolygonRows);
	g_assert(PolygonColumns);

	// Create the mesh ...
	k3dIMesh* mesh = CreateMesh(Name);
	g_assert(mesh);

	// Create points ...
	ReservePoints(PointRows, PointColumns);
	for(gulong pointrow = 0; pointrow < PointRows; pointrow++)
		for(gulong pointcol = 0; pointcol < PointColumns; pointcol++)
			CreatePoint(mesh, pointrow, pointcol);

	// Create polygons ...
	ReservePolygons(PolygonRows, PolygonColumns);
	for(gulong polyrow = 0; polyrow < PolygonRows; polyrow++)
		for(gulong polycol = 0; polycol < PolygonColumns; polycol++)
			CreatePolygon(mesh, polyrow, polycol);

	return mesh;
}


gulong k3dPrimTool::MeshColumn(gulong Column, gulong Columns, bool StitchSide)
{
	return Column % (StitchSide ? Columns : Columns+1);
}


gulong k3dPrimTool::MeshRow(gulong Row, gulong Rows, bool StitchTop)
{
	return Row % (StitchTop ? Rows : Rows+1);
}


k3dIMesh* k3dPrimTool::CreateMesh(SDPCSTRING Name, gulong Rows, gulong Columns, bool StitchTop, bool StitchSide)
{
	// Sanity checks ...
	g_assert_string(Name);

	// Calculate the number of polygons along each axis ...
	gulong polyrows = Rows;
	gulong polycolumns = Columns;

	// Calculate the number of points along each axis ...
//	gulong pointrows = polyrows + 1;
//	gulong pointcolumns = polycolumns + 1;

	// Calculate the actual number of points that need to be created along each axis ...
	gulong newpointrows = polyrows + (StitchTop ? 0 : 1);
	gulong newpointcolumns = polycolumns + (StitchSide ? 0 : 1);

	// Create the mesh ...
	k3dIMesh* mesh = CreateMesh(Name, newpointrows, newpointcolumns, polyrows, polycolumns);
	g_assert(mesh);

	// For each polygon ...
	for(gulong polyrow = 0; polyrow < polyrows; polyrow++)
		{
			for(gulong polycol = 0; polycol < polycolumns; polycol++)
				{
					// Add four points in counterclockwise order ...
					k3dIPolygon* polygon = m_Polygons[polyrow][polycol];
					k3dIPolyPointServer* points = polygon->Points();
					k3dIPolyPoint* polypoint;
					sdpVector3 coords, texturecoords;

					// 1st point ...
					polypoint = points->AddPoint(m_Points[MeshRow(polyrow, polyrows, StitchTop)][MeshColumn(polycol+1, polycolumns, StitchSide)]);
					coords = sdpVector3((double(polycol+1) / double(polycolumns)) - 0.5, 0.5 - (double(polyrow) / double(polyrows)), 0.0);
					texturecoords = sdpVector3(double(polycol+1) / double(polycolumns), double(polyrow) / double(polyrows), 0.0);
					polypoint->Location()->SetOldXYZ(coords);
					polypoint->Location()->SetLocalXYZ(coords);
					polypoint->Location()->SetTextureXYZ(texturecoords);
					polypoint->Location()->SetImplicitTextureXYZ(texturecoords);

					// 2nd point ...
					polypoint = points->AddPoint(m_Points[MeshRow(polyrow, polyrows, StitchTop)][MeshColumn(polycol, polycolumns, StitchSide)]);
					coords = sdpVector3((double(polycol) / double(polycolumns)) - 0.5, 0.5 - (double(polyrow) / double(polyrows)), 0.0);
					texturecoords = sdpVector3(double(polycol) / double(polycolumns), double(polyrow) / double(polyrows), 0.0);
					polypoint->Location()->SetOldXYZ(coords);
					polypoint->Location()->SetLocalXYZ(coords);
					polypoint->Location()->SetTextureXYZ(texturecoords);
					polypoint->Location()->SetImplicitTextureXYZ(texturecoords);

					// 3rd point ...
					polypoint = points->AddPoint(m_Points[MeshRow(polyrow+1, polyrows, StitchTop)][MeshColumn(polycol, polycolumns, StitchSide)]);
					coords = sdpVector3((double(polycol) / double(polycolumns)) - 0.5, 0.5 - (double(polyrow+1) / double(polyrows)), 0.0);
					texturecoords = sdpVector3(double(polycol) / double(polycolumns), double(polyrow+1) / double(polyrows), 0.0);
					polypoint->Location()->SetOldXYZ(coords);
					polypoint->Location()->SetLocalXYZ(coords);
					polypoint->Location()->SetTextureXYZ(texturecoords);
					polypoint->Location()->SetImplicitTextureXYZ(texturecoords);

					// 4th point ...
					polypoint = points->AddPoint(m_Points[MeshRow(polyrow+1, polyrows, StitchTop)][MeshColumn(polycol+1, polycolumns, StitchSide)]);
					coords = sdpVector3((double(polycol+1) / double(polycolumns)) - 0.5, 0.5 - (double(polyrow+1) / double(polyrows)), 0.0);
					texturecoords = sdpVector3(double(polycol+1) / double(polycolumns), double(polyrow+1) / double(polyrows), 0.0);
					polypoint->Location()->SetOldXYZ(coords);
					polypoint->Location()->SetLocalXYZ(coords);
					polypoint->Location()->SetTextureXYZ(texturecoords);
					polypoint->Location()->SetImplicitTextureXYZ(texturecoords);
				}
		}

	return mesh;
}


/*
// Create a new sub-group ...
k3dIObject* k3dPrimTool::CreateGroup(k3dIMesh* Parent, SDPCSTRING Name, k3dIPointServer** PointServer, k3dIPolygonServer** PolygonServer)
{
	// Sanity checks ...
	g_assert(Parent);
	g_assert_string(Name);

	// Create a sub-mesh ...
	k3dIObject* object;
	g_return_val_if_fail(sdpSucceeded(Document()->CreateObject(sdpSubMesh, IID_k3dIObject, FALSE, TRUE, TRUE, (void**)&object)), 0);

	// Store a reference to it ...
	m_Objects.push_back(object);

	// Set the parent and name ...
	object->ObjectContext()->SetParent(Parent->ObjectContext());
	object->ObjectContext()->SetName(Name);

	// Return the requested interfaces ...
	if(PointServer)
		g_return_val_if_fail(sdpSucceeded(object->QueryInterface(IID_k3dIPointServer, (void**)PointServer)), 0);

	if(PolygonServer)
		g_return_val_if_fail(sdpSucceeded(object->QueryInterface(IID_k3dIPolygonServer, (void**)PolygonServer)), 0);

	return object;
}


// Create a new sub-group, and point it to the requested number of new points in its' parent ...
k3dIObject* k3dPrimTool::CreateGroup(k3dIMesh* Parent, SDPCSTRING Name, gulong Points, gulong Polygons)
{
	// Sanity checks ...
	g_assert(Parent);
	g_assert_string(Name);

	// Create the group ...	
	k3dIObject* object;
	k3dIPointServer* pointclient;
	k3dIPolygonServer* polygonclient;
	object = CreateSubMesh(Parent, Name, &pointclient, &polygonclient);

	// Create points and assign them to the group ...
	for(gulong i = 0; i < Points; i++)
		pointclient->AddPoint(CreatePoint(Parent));

	// Create polygons and assign them to the group ...
	for(gulong j = 0; j < Polygons; j++)
		polygonclient->AddPolygon(CreatePolygon(Parent));

	// Release interfaces ...
	pointclient->Release();
	polygonclient->Release();
	
	return object;
}
*/


