
// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements convenience classes for sending rendering parameters to a k3dIRiEngine interface
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3drishadersetup.h>

#include <iostream>

/////////////////////////////////////////////////////////////////////////////
// k3dRiParameterList

k3dRiParameterList::k3dRiParameterList()
{
}

k3dRiParameterList::~k3dRiParameterList()
{
	ClearAll();
}

// Self explanatory
void k3dRiParameterList::ClearAll()
{
	m_Tokens.clear();
	m_Parameters.clear();
	m_Types.clear();

	m_Ints.clear();
	m_Floats.clear();
	m_Points.clear();
	m_Colors.clear();
	m_Strings.clear();
	m_StringPointers.clear();
}

// Returns true if a token already exists in the token list
bool k3dRiParameterList::FindToken(SDPCSTRING Token)
{
	// Sanity checks ...
	g_assert_string(Token);

	for(gulong i = 0; i < m_Tokens.size(); i++)
		if(0 == strcmp(m_Tokens[i], Token))
			return true;

	return false;
}

// Declares parameters and their types ...
void k3dRiParameterList::DeclareParameters(k3dIRiEngine* Engine)
{
	// Sanity checks ...
	g_assert(Engine);

	// Declare each parameter ...
	for(gulong i = 0; i < m_Tokens.size(); i++)
		Engine->RiDeclare(m_Tokens[i], m_Types[i]);
}

void k3dRiParameterList::SetAttributes(SDPCSTRING Name, k3dIRiEngine* Engine)
{
	// Sanity checks ...
	g_assert_string(Name);
	g_assert(Engine);

	// If we don't have any parameters, we're done!
	if(!ParameterCount())
		return;

	// Set attributes ...
	Engine->RiAttributeV(const_cast<char*>(Name), ParameterCount(), Tokens(), Parameters());
}

void k3dRiParameterList::SetOptions(SDPCSTRING Name, k3dIRiEngine* Engine)
{
	// Sanity checks ...
	g_assert_string(Name);
	g_assert(Engine);

	// If we don't have any parameters, we're done!
	if(!ParameterCount())
		return;

	// Set options ...
	Engine->RiOptionV(const_cast<char*>(Name), ParameterCount(), Tokens(), Parameters());
}

void k3dRiParameterList::AddInteger(SDPCSTRING Token, const gint Arg)
{
	// Sanity checks ...
	g_assert_string(Token);

	// If this fails, you tried to add the same argument twice!
	if(FindToken(Token))
		std::cerr << "k3dRiParameterList::AddInteger(" << Token << ", " << Arg << ") - adding the same token twice!" << std::endl;

	// Store the token ...
	m_Tokens.push_back(Token);

	// Store the argument value ...
	m_Ints.push_back(Arg);

	// Store a pointer to the value ...
	m_Parameters.push_back(gpointer(&m_Ints.back()));

	// Store the type ...
	m_Types.push_back("integer");
}

void k3dRiParameterList::AddScalar(SDPCSTRING Token, const gdouble Arg)
{
	// Sanity checks ...
	g_assert_string(Token);

	// If this fails, you tried to add the same argument twice!
	if(FindToken(Token))
		std::cerr << "k3dRiParameterList::AddScalar(" << Token << ", " << Arg << ") - adding the same token twice!" << std::endl;

	// Store the token ...
	m_Tokens.push_back(Token);

	// Store the argument value ...
	m_Floats.push_back(Arg);

	// Store a pointer to the value ...
	m_Parameters.push_back(gpointer(&m_Floats.back()));

	// Store the type ...
	m_Types.push_back("float");
}

void k3dRiParameterList::AddPoint(SDPCSTRING Token, const sdpVector3& Arg)
{
	// Sanity checks ...
	g_assert_string(Token);

	// If this fails, you tried to add the same argument twice!
	if(FindToken(Token))
		std::cerr << "k3dRiParameterList::AddPoint(" << Token << ", " << sdpString(Arg) << ") - adding the same token twice!" << std::endl;

	// Store the token ...
	m_Tokens.push_back(Token);

	// Store the argument value ...
	m_Points.push_back(Arg);

	// Store a pointer to the value ...
	m_Parameters.push_back(gpointer(&m_Points.back()));

	// Store the type ...
	m_Types.push_back("point");
}

void k3dRiParameterList::AddColor(SDPCSTRING Token, const sdpVector3& Arg)
{
	// Sanity checks ...
	g_assert_string(Token);

	// If this fails, you tried to add the same argument twice!
	if(FindToken(Token))
		std::cerr << "k3dRiParameterList::AddColor(" << Token << ", " << sdpString(Arg) << ") - adding the same token twice!" << std::endl;

	// Store the token ...
	m_Tokens.push_back(Token);

	// Store the argument value ...
	m_Colors.push_back(Arg);

	// Store a pointer to the value ...
	m_Parameters.push_back(gpointer(&m_Colors.back()));

	// Store the type ...
	m_Types.push_back("color");
}

void k3dRiParameterList::AddString(SDPCSTRING Token, const SDPCSTRING Arg)
{
	// Sanity checks ...
	g_assert_string(Token);

	// If this fails, you tried to add the same argument twice!
	if(FindToken(Token))
		std::cerr << "k3dRiParameterList::AddString(" << Token << ", " << Arg << ") - adding the same token twice!" << std::endl;

	// Store the token ...
	m_Tokens.push_back(Token);

	// Store the argument value ...
	m_Strings.push_back(Arg);
	
	// Store a string pointer ...
	m_StringPointers.push_back(gpointer(m_Strings.back().c_str()));

	// Store a pointer to the string ...
	m_Parameters.push_back(gpointer(&m_StringPointers.back()));

	// Store the type ...
	m_Types.push_back("string");
}

void k3dRiParameterList::AddPointer(SDPCSTRING Token, const gpointer Arg)
{
	// Sanity checks ...
	g_assert_string(Token);

	// If this fails, you tried to add the same argument twice!
	if(FindToken(Token))
		std::cerr << "k3dRiParameterList::AddPointer(" << Token << ", " << Arg << ") - adding the same token twice!" << std::endl;

	// Store the token ...
	m_Tokens.push_back(Token);

	// Store the pointer ...
	m_Parameters.push_back(Arg);

	// Store the type ...
	m_Types.push_back("pointer");
}

/////////////////////////////////////////////////////////////////////////////
// k3dRiShader

// Constructor for use with a ShaderSelector ...
k3dRiShader::k3dRiShader()
{
}

// Constructor for stand-alone use ...
k3dRiShader::k3dRiShader(SDPCSTRING ShaderName)
{
	SetShader(ShaderName);
}

// Explicitly set the shader name ...
void k3dRiShader::SetShader(SDPCSTRING ShaderName)
{
	m_ShaderName = ShaderName;
}

//////////////////////////////////////////////////////////////////////
// k3dRiLightShader

k3dRiLightShader::k3dRiLightShader()
{
}

k3dRiLightShader::k3dRiLightShader(SDPCSTRING ShaderName)
	: k3dRiShader(ShaderName)
{
}

gulong k3dRiLightShader::SetupLight(k3dIRiEngine* Engine)
{
	DeclareParameters(Engine);
	return Engine->RiLightSourceV(m_ShaderName, m_Tokens.size(), &(m_Tokens.front()), &(m_Parameters.front()));
}

gulong k3dRiLightShader::SetupAreaLight(k3dIRiEngine* Engine)
{
	DeclareParameters(Engine);
	return Engine->RiAreaLightSourceV(m_ShaderName, m_Tokens.size(), &(m_Tokens.front()), &(m_Parameters.front()));
}

//////////////////////////////////////////////////////////////////////
// k3dRiVolumeShader

k3dRiVolumeShader::k3dRiVolumeShader()
{
}

k3dRiVolumeShader::k3dRiVolumeShader(SDPCSTRING ShaderName)
	: k3dRiShader(ShaderName)
{
}

sdpResult k3dRiVolumeShader::SetupInterior(k3dIRiEngine* Engine)
{
	DeclareParameters(Engine);
	return Engine->RiInteriorV(m_ShaderName, m_Tokens.size(), &(m_Tokens.front()), &(m_Parameters.front()));
}

sdpResult k3dRiVolumeShader::SetupExterior(k3dIRiEngine* Engine)
{
	DeclareParameters(Engine);
	return Engine->RiExteriorV(m_ShaderName, m_Tokens.size(), &(m_Tokens.front()), &(m_Parameters.front()));
}

sdpResult k3dRiVolumeShader::SetupAtmosphere(k3dIRiEngine* Engine)
{
	DeclareParameters(Engine);
	return Engine->RiAtmosphereV(m_ShaderName, m_Tokens.size(), &(m_Tokens.front()), &(m_Parameters.front()));
}

//////////////////////////////////////////////////////////////////////
// k3dRiTransformationShader

k3dRiTransformationShader::k3dRiTransformationShader()
{
}

k3dRiTransformationShader::k3dRiTransformationShader(SDPCSTRING ShaderName)
	: k3dRiShader(ShaderName)
{
}

sdpResult k3dRiTransformationShader::Setup(k3dIRiEngine* Engine)
{
	DeclareParameters(Engine);
	return Engine->RiDeformationV(m_ShaderName, m_Tokens.size(), &(m_Tokens.front()), &(m_Parameters.front()));
}

//////////////////////////////////////////////////////////////////////
// k3dRiDisplacementShader

k3dRiDisplacementShader::k3dRiDisplacementShader()
{
}

k3dRiDisplacementShader::k3dRiDisplacementShader(SDPCSTRING ShaderName)
	: k3dRiShader(ShaderName)
{
}

sdpResult k3dRiDisplacementShader::Setup(k3dIRiEngine* Engine)
{
	DeclareParameters(Engine);
	return Engine->RiDisplacementV(m_ShaderName, m_Tokens.size(), &(m_Tokens.front()), &(m_Parameters.front()));
}

//////////////////////////////////////////////////////////////////////
// k3dRiSurfaceShader

k3dRiSurfaceShader::k3dRiSurfaceShader()
{
}

k3dRiSurfaceShader::k3dRiSurfaceShader(SDPCSTRING ShaderName)
	: k3dRiShader(ShaderName)
{
}

sdpResult k3dRiSurfaceShader::Setup(k3dIRiEngine* Engine)
{
	DeclareParameters(Engine);
	return Engine->RiSurfaceV(m_ShaderName, m_Tokens.size(), &(m_Tokens.front()), &(m_Parameters.front()));
}

//////////////////////////////////////////////////////////////////////
// k3dRiImagerShader

k3dRiImagerShader::k3dRiImagerShader()
{
}

k3dRiImagerShader::k3dRiImagerShader(SDPCSTRING ShaderName)
	: k3dRiShader(ShaderName)
{
}

sdpResult k3dRiImagerShader::Setup(k3dIRiEngine* Engine)
{
	DeclareParameters(Engine);
	return Engine->RiImagerV(m_ShaderName, m_Tokens.size(), &(m_Tokens.front()), &(m_Parameters.front()));
}


