// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dShaderInstance class, which manages an instance of a Pixar Renderman shader and its arguments
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dshaderinstance.h>
#include <k3dcom.h>
#include <k3dmisc.h>
#include <k3diritexture.h>

#include <iostream>

//////////////////////////////////////////////////////////////////////
// ArgumentBase

k3dShaderInstance::ArgumentBase* k3dShaderInstance::ArgumentBase::CreateArgument(k3dIDocument* Document, SDPCSTRING Name, k3dIShaderInfoArgument::ArgumentType Type, SDPCSTRING Default)
{
	// Sanity checks ...
	g_assert(Document);
	g_assert_string(Name);

	switch(Type)
		{
			case k3dIShaderInfoArgument::UNKNOWN:
			case k3dIShaderInfoArgument::POINT:
			case k3dIShaderInfoArgument::VECTOR:
			case k3dIShaderInfoArgument::ORIENTATION:
			case k3dIShaderInfoArgument::SCALE:
			case k3dIShaderInfoArgument::NORMAL:
			case k3dIShaderInfoArgument::COORDINATESYSTEM:
			case k3dIShaderInfoArgument::BOOLEAN:
			case k3dIShaderInfoArgument::OPTIONLIST:
//				std::cout << "k3dShaderInstance::Argument::CreateArgument(): Unsupported shader argument type!!" << std::endl;
				break;
			case k3dIShaderInfoArgument::TEXTURE:
				return new k3dShaderInstance::TextureArgument(Document, Name, Type, Default);
			case k3dIShaderInfoArgument::COLOR:
			case k3dIShaderInfoArgument::SCALAR:
			case k3dIShaderInfoArgument::TIME:
			case k3dIShaderInfoArgument::ANGLE:
			case k3dIShaderInfoArgument::DISTANCE:
			case k3dIShaderInfoArgument::AREA:
			case k3dIShaderInfoArgument::VOLUME:
			case k3dIShaderInfoArgument::MASS:
			case k3dIShaderInfoArgument::FORCE:
			case k3dIShaderInfoArgument::PRESSURE:
				return new k3dShaderInstance::ChannelArgument(Document, Name, Type, Default);
			case k3dIShaderInfoArgument::STRING:
				return new k3dShaderInstance::Argument<sdpString>(Document, Name, Type, Default);
			default:
				g_assert_not_reached();	// Unknown argument type!!
		}

	return 0;
}

k3dShaderInstance::ArgumentBase* k3dShaderInstance::ArgumentBase::LoadArgument(k3dIDocument* Document, SDPCSTRING Name, k3dIShaderInfoArgument::ArgumentType Type, SDPCSTRING Value)
{
	// Sanity checks ...
	g_assert(Document);
	g_assert_string(Name);

	switch(Type)
		{
			case k3dIShaderInfoArgument::UNKNOWN:
			case k3dIShaderInfoArgument::POINT:
			case k3dIShaderInfoArgument::VECTOR:
			case k3dIShaderInfoArgument::ORIENTATION:
			case k3dIShaderInfoArgument::SCALE:
			case k3dIShaderInfoArgument::NORMAL:
			case k3dIShaderInfoArgument::COORDINATESYSTEM:
			case k3dIShaderInfoArgument::BOOLEAN:
			case k3dIShaderInfoArgument::OPTIONLIST:
//				std::cout << "k3dShaderInstance::ArgumentBase::LoadArgument(): Unsupported shader argument type!!" << std::endl;
				break;
			case k3dIShaderInfoArgument::TEXTURE:
				return new k3dShaderInstance::TextureArgument(Document, Name, Type, Value);
			case k3dIShaderInfoArgument::COLOR:
			case k3dIShaderInfoArgument::SCALAR:
			case k3dIShaderInfoArgument::TIME:
			case k3dIShaderInfoArgument::ANGLE:
			case k3dIShaderInfoArgument::DISTANCE:
			case k3dIShaderInfoArgument::AREA:
			case k3dIShaderInfoArgument::VOLUME:
			case k3dIShaderInfoArgument::MASS:
			case k3dIShaderInfoArgument::FORCE:
			case k3dIShaderInfoArgument::PRESSURE:
				return new k3dShaderInstance::ChannelArgument(Document, Name, Type, Value);
			case k3dIShaderInfoArgument::STRING:
				return new k3dShaderInstance::Argument<sdpString>(Document, Name, Type, Value);
			default:
				g_assert_not_reached();	// Unknown argument type!!
		}

	return 0;
}

k3dShaderInstance::ArgumentBase::ArgumentBase(k3dIDocument* Document, SDPCSTRING Name, k3dIShaderInfoArgument::ArgumentType Type, SDPCSTRING Default)
{
	// Sanity checks ...
	g_assert(Document);
	g_assert_string(Name);

	m_Document = Document;
	m_Name = Name;
	m_Type = Type;
	m_Default = Default;
}

k3dShaderInstance::ArgumentBase::~ArgumentBase()
{
}

//////////////////////////////////////////////////////////////////////
// ChannelArgument

k3dShaderInstance::ChannelArgument::ChannelArgument(k3dIDocument* Document, SDPCSTRING Name, k3dIShaderInfoArgument::ArgumentType Type, SDPCSTRING Default)
	: k3dShaderInstance::ArgumentBase(Document, Name, Type, Default)
{
}

bool k3dShaderInstance::ChannelArgument::Create()
{
	// Create the channel object ...
	g_return_val_if_fail(m_Channel.Create(m_Document, sdpChannel, m_Name, false, false, true), false);

	// Setup the channel datatype and value, based on our argument type ...
	switch(m_Type)
		{
			case k3dIShaderInfoArgument::UNKNOWN:
			case k3dIShaderInfoArgument::TIME:
			case k3dIShaderInfoArgument::POINT:
			case k3dIShaderInfoArgument::VECTOR:
			case k3dIShaderInfoArgument::ORIENTATION:
			case k3dIShaderInfoArgument::SCALE:
			case k3dIShaderInfoArgument::NORMAL:
				{
					std::cout << "k3dShaderInstance::ChannelArgument::Create(): Unsupported argument type!!" << std::endl;
					return false;
				}
			case k3dIShaderInfoArgument::COLOR:
				{
					sdpVector3 defaultvalue;
					m_Default.getvalue(defaultvalue);
					m_Channel->SetDatatype(k3dIChannel::DATA_COLOR);
					k3dChannelSetConstantValue(m_Channel.Interface(), defaultvalue);
					return true;
				}
			case k3dIShaderInfoArgument::SCALAR:
				{
					gdouble defaultvalue = 0.0;
					m_Default.getvalue(defaultvalue);
					m_Channel->SetDatatype(k3dIChannel::DATA_SCALAR);
					k3dChannelSetConstantValue(m_Channel.Interface(), defaultvalue);
					return true;
				}
			case k3dIShaderInfoArgument::ANGLE:
				{
					gdouble defaultvalue = 0.0;
					m_Default.getvalue(defaultvalue);
					m_Channel->SetDatatype(k3dIChannel::DATA_ANGLE);
					k3dChannelSetConstantValue(m_Channel.Interface(), defaultvalue);
					return true;
				}
			case k3dIShaderInfoArgument::DISTANCE:
				{
					gdouble defaultvalue = 0.0;
					m_Default.getvalue(defaultvalue);
					m_Channel->SetDatatype(k3dIChannel::DATA_DISTANCE);
					k3dChannelSetConstantValue(m_Channel.Interface(), defaultvalue);
					return true;
				}
			case k3dIShaderInfoArgument::AREA:
				{
					gdouble defaultvalue = 0.0;
					m_Default.getvalue(defaultvalue);
					m_Channel->SetDatatype(k3dIChannel::DATA_AREA);
					k3dChannelSetConstantValue(m_Channel.Interface(), defaultvalue);
					return true;
				}
			case k3dIShaderInfoArgument::VOLUME:
				{
					gdouble defaultvalue = 0.0;
					m_Default.getvalue(defaultvalue);
					m_Channel->SetDatatype(k3dIChannel::DATA_VOLUME);
					k3dChannelSetConstantValue(m_Channel.Interface(), defaultvalue);
					return true;
				}
			case k3dIShaderInfoArgument::MASS:
				{
					gdouble defaultvalue = 0.0;
					m_Default.getvalue(defaultvalue);
					m_Channel->SetDatatype(k3dIChannel::DATA_MASS);
					k3dChannelSetConstantValue(m_Channel.Interface(), defaultvalue);
					return true;
				}
			case k3dIShaderInfoArgument::FORCE:
				{
					gdouble defaultvalue = 0.0;
					m_Default.getvalue(defaultvalue);
					m_Channel->SetDatatype(k3dIChannel::DATA_FORCE);
					k3dChannelSetConstantValue(m_Channel.Interface(), defaultvalue);
					return true;
				}
			case k3dIShaderInfoArgument::PRESSURE:
				{
					gdouble defaultvalue = 0.0;
					m_Default.getvalue(defaultvalue);
					m_Channel->SetDatatype(k3dIChannel::DATA_PRESSURE);
					k3dChannelSetConstantValue(m_Channel.Interface(), defaultvalue);
					return true;
				}

			case k3dIShaderInfoArgument::COORDINATESYSTEM:
			case k3dIShaderInfoArgument::BOOLEAN:
			case k3dIShaderInfoArgument::OPTIONLIST:
			case k3dIShaderInfoArgument::TEXTURE:
			case k3dIShaderInfoArgument::STRING:
				g_assert_not_reached();	// These types never should have been encoded as a channel!

			default:
				g_assert_not_reached();	// Unknown argument type!!
		}

	return true;
}

bool k3dShaderInstance::ChannelArgument::LoadComplete()
{
	// Get our object by ID ...
	k3dIObject* channel = k3dGetObject(m_Document, atol(m_Default));
	g_return_val_if_fail(channel, false);

	// Attach the channel object ...
	g_return_val_if_fail(sdpSucceeded(m_Channel.AttachObject(channel)), false);

	return true;
}

sdpString k3dShaderInstance::ChannelArgument::ConvertToString(gdouble Time, gdouble DeltaTime)
{
	// Setup the channel datatype and value, based on our argument type ...
	switch(m_Type)
		{
			case k3dIShaderInfoArgument::UNKNOWN:
			case k3dIShaderInfoArgument::TIME:
			case k3dIShaderInfoArgument::POINT:
			case k3dIShaderInfoArgument::VECTOR:
			case k3dIShaderInfoArgument::ORIENTATION:
			case k3dIShaderInfoArgument::SCALE:
			case k3dIShaderInfoArgument::NORMAL:
				{
					std::cout << "k3dShaderInstance::ChannelArgument::ConvertToString(): Unsupported argument type!!" << std::endl;
					return sdpString();
				}
			case k3dIShaderInfoArgument::COLOR:
				{
					sdpVector3 result;
					m_Channel->GetValue3(Time, DeltaTime, result);
					return sdpString(result);
				}
			case k3dIShaderInfoArgument::SCALAR:
				{
					return sdpString(m_Channel->GetValue1(Time, DeltaTime));
				}
			case k3dIShaderInfoArgument::ANGLE:
				{
					return sdpString(m_Channel->GetValue1(Time, DeltaTime));
				}
			case k3dIShaderInfoArgument::DISTANCE:
				{
					return sdpString(m_Channel->GetValue1(Time, DeltaTime));
				}
			case k3dIShaderInfoArgument::AREA:
				{
					return sdpString(m_Channel->GetValue1(Time, DeltaTime));
				}
			case k3dIShaderInfoArgument::VOLUME:
				{
					return sdpString(m_Channel->GetValue1(Time, DeltaTime));
				}
			case k3dIShaderInfoArgument::MASS:
				{
					return sdpString(m_Channel->GetValue1(Time, DeltaTime));
				}
			case k3dIShaderInfoArgument::FORCE:
				{
					return sdpString(m_Channel->GetValue1(Time, DeltaTime));
				}
			case k3dIShaderInfoArgument::PRESSURE:
				{
					return sdpString(m_Channel->GetValue1(Time, DeltaTime));
				}

			case k3dIShaderInfoArgument::COORDINATESYSTEM:
			case k3dIShaderInfoArgument::BOOLEAN:
			case k3dIShaderInfoArgument::OPTIONLIST:
			case k3dIShaderInfoArgument::TEXTURE:
			case k3dIShaderInfoArgument::STRING:
				g_assert_not_reached();	// These types never should have been encoded as a channel!

			default:
				g_assert_not_reached();	// Unknown argument type!!
		}

	return sdpString();
}

//////////////////////////////////////////////////////////////////////
// TextureArgument

k3dShaderInstance::TextureArgument::TextureArgument(k3dIDocument* Document, SDPCSTRING Name, k3dIShaderInfoArgument::ArgumentType Type, SDPCSTRING Default)
	: k3dShaderInstance::ArgumentBase(Document, Name, Type, Default)
{
}

bool k3dShaderInstance::TextureArgument::Create()
{
	// Create the texture object ...
//g_return_val_if_fail(m_Texture.Create(m_Document, sdpBitmap, m_Name, false, false, true), false);
	return true;
}

bool k3dShaderInstance::TextureArgument::LoadComplete()
{
	// Get our object by ID ...
	k3dIObject* texture = k3dGetObject(m_Document, atol(m_Default));
	g_return_val_if_fail(texture, false);

	// Attach the texture object ...
	g_return_val_if_fail(sdpSucceeded(m_Texture.AttachObject(texture)), false);

	return true;
}

sdpString k3dShaderInstance::TextureArgument::ConvertToString(gdouble Time, gdouble DeltaTime)
{
	return sdpString();
}

//////////////////////////////////////////////////////////////////////
// k3dShaderInstance

k3dShaderInstance::k3dShaderInstance()
{
	m_Application = 0;
	m_Document = 0;
	m_PrimaryTexture = 0;

	m_ShaderHandle = 0;
}

k3dShaderInstance::~k3dShaderInstance()
{
	Release();
}

bool k3dShaderInstance::Create(k3dIDocument* Document)
{
	// Sanity checks ...
	g_assert(Document);

	m_Document = Document;
	m_Application = Document->Application();
	
	return true;
}

bool k3dShaderInstance::SetShader(SDPCSTRING Shader)
{
	// Sanity checks ...
	g_assert(m_Application);	// You have to call Create() before calling this function!
	g_assert(m_Document);

	// For each shader ...
	for(K3DSHADERINFOHANDLE handle = m_Application->GetFirstShaderInfo(); handle; handle = m_Application->GetNextShaderInfo(handle))
		{
			// Test to see if the shader's name matches ...
			k3dIShaderInfo* shader = m_Application->GetShaderInfo(handle);
			g_assert(shader);

			if(shader->Name() == sdpString(Shader))
				return SetShaderHandle(handle);
		}

	// Couldn't find a shader to match ...
	SetShaderHandle(0);
	std::cerr << "k3dShaderInstance::SetShader(): No shader named \"" << Shader << "\" available!" << std::endl;
	return false;
}

bool k3dShaderInstance::SetShaderHandle(K3DSHADERINFOHANDLE ShaderHandle)
{
	// Get rid of old data ...
	Release();

	// Set the shader handle (we allow nulls) ...
	m_ShaderHandle = ShaderHandle;

	// Allocate new data ...
	g_return_val_if_fail(Allocate(), false);

	return true;
}

k3dITexture* k3dShaderInstance::GetPrimaryTexture()
{
	if(m_PrimaryTexture)
		return m_PrimaryTexture->Texture();
	
	return 0;
}

SDPCSTRING k3dShaderInstance::ShaderName()
{
	// Sanity checks ...
	g_assert(m_Application);	// You have to call Create() before calling this function!
	g_assert(m_Document);

	// If we don't have a shader, we're done ...
	if(0 == m_ShaderHandle)
		return "";

	k3dIShaderInfo* shader = m_Application->GetShaderInfo(m_ShaderHandle);
	g_assert(shader);

	return shader->Name();
}

k3dShaderInstance::ArgumentBase* k3dShaderInstance::ArgumentAt(gulong Index)
{
	// Sanity checks ...
	g_assert(Index < ArgumentCount());

	ArgumentIterator argument = m_Arguments.begin();
	for(gulong i = 0; i < Index; i++)
		argument++;

	return *argument;
}

bool k3dShaderInstance::Allocate()
{
	// Sanity checks ...
	g_assert(m_Application);	// You have to call Create() before calling this function!
	g_assert(m_Document);

	// If we don't have a shader handle, we're done ...
	if(!m_ShaderHandle)
		return true;

	// Get the shader info ...
	k3dIShaderInfo* shader = m_Application->GetShaderInfo(m_ShaderHandle);
	g_return_val_if_fail(shader, false);

	// For each shader argument ...
	for(K3DSHADERINFOARGUMENTHANDLE handle = shader->GetFirstArgument(); handle; handle = shader->GetNextArgument(handle))
		{
			k3dIShaderInfoArgument* shaderargument = shader->GetArgument(handle);
			g_assert(shaderargument);

			// Create a record for this argument
			ArgumentBase* argument = ArgumentBase::CreateArgument(m_Document, shaderargument->Name(), shaderargument->Type(), shaderargument->Default());
			if(!argument)
				continue;

			// If this argument is a texture argument, store a reference to it for possible use in Editor/OpenGL rendering ...
			if(k3dIShaderInfoArgument::TEXTURE == argument->Type())
				m_PrimaryTexture = static_cast<TextureArgument*>(argument);

			// Initialize the argument ...
			g_assert(argument->Create());

			// Store the argument ...
			m_Arguments.push_back(argument);
		}

	return true;
}

void k3dShaderInstance::Release()
{
	// Get rid of the arguments ...
	for(ArgumentIterator arg = m_Arguments.begin(); arg != m_Arguments.end(); arg++)
		delete (*arg);

	// Empty our list ...
	m_Arguments.clear();
	
	m_PrimaryTexture = 0;
}

void k3dShaderInstance::Initialize(k3dIRenderFrame* Frame, k3dIRiEngine* Engine, k3dRiShader& Shader, gdouble Time, gdouble DeltaTime)
{
	// Sanity checks ...
	g_assert(Engine);

	// Set the shader name ...
	Shader.SetShader(ShaderName());

	// For each argument in this shader ...
	for(ArgumentIterator i = m_Arguments.begin(); i != m_Arguments.end(); i++)
		{
			switch((*i)->Type())
				{
					case k3dIShaderInfoArgument::UNKNOWN:
					case k3dIShaderInfoArgument::POINT:
					case k3dIShaderInfoArgument::VECTOR:
					case k3dIShaderInfoArgument::ORIENTATION:
					case k3dIShaderInfoArgument::SCALE:
					case k3dIShaderInfoArgument::NORMAL:
					case k3dIShaderInfoArgument::COORDINATESYSTEM:
					case k3dIShaderInfoArgument::BOOLEAN:
					case k3dIShaderInfoArgument::OPTIONLIST:
						{
							std::cout << "Unsupported argument type!!" << std::endl;
							break;
						}
					case k3dIShaderInfoArgument::TEXTURE:
						{
							TextureArgument* argument = static_cast<TextureArgument*>(*i);
							if(argument->Texture())
								{
									k3dIRiTexture* texture = sdpQueryInterface<k3dIRiTexture>(argument->Texture());
									if(texture)
										Shader.AddString(argument->Name(), texture->RiTexturePath(Frame, Engine));
									else
										Shader.AddString(argument->Name(), "");
								}
							else
								Shader.AddString(argument->Name(), "");
								
							break;
						}
					case k3dIShaderInfoArgument::COLOR:
						{
							ChannelArgument* argument = static_cast<ChannelArgument*>(*i);
							sdpVector3 value;
							argument->Channel()->GetValue3(Time, DeltaTime, value);
							Shader.AddColor(argument->Name(), value);
							break;
						}
					case k3dIShaderInfoArgument::SCALAR:
					case k3dIShaderInfoArgument::TIME:
					case k3dIShaderInfoArgument::ANGLE:
					case k3dIShaderInfoArgument::DISTANCE:
					case k3dIShaderInfoArgument::AREA:
					case k3dIShaderInfoArgument::VOLUME:
					case k3dIShaderInfoArgument::MASS:
					case k3dIShaderInfoArgument::FORCE:
					case k3dIShaderInfoArgument::PRESSURE:
						{
							ChannelArgument* argument = static_cast<ChannelArgument*>(*i);
							gdouble value = argument->Channel()->GetValue1(Time, DeltaTime);
							Shader.AddScalar(argument->Name(), value);
							break;
						}
					case k3dIShaderInfoArgument::STRING:
						{
							StringArgument* argument = static_cast<StringArgument*>(*i);
							Shader.AddString(argument->Name(), argument->Value());
							break;;
						}
					default:
						g_assert_not_reached();	// Unknown argument type!!
				}
		}
}

sdpString ToString(k3dIShaderInfoArgument::ArgumentType Type)
{
	switch(Type)
		{
			case k3dIShaderInfoArgument::UNKNOWN:
				return sdpString("unknown");
			case k3dIShaderInfoArgument::POINT:
				return sdpString("point");
			case k3dIShaderInfoArgument::VECTOR:
				return sdpString("vector");
			case k3dIShaderInfoArgument::ORIENTATION:
				return sdpString("orientation");
			case k3dIShaderInfoArgument::SCALE:
				return sdpString("scale");
			case k3dIShaderInfoArgument::NORMAL:
				return sdpString("normal");
			case k3dIShaderInfoArgument::COORDINATESYSTEM:
				return sdpString("coordinatesystem");
			case k3dIShaderInfoArgument::BOOLEAN:
				return sdpString("boolean");
			case k3dIShaderInfoArgument::OPTIONLIST:
				return sdpString("optionlist");
			case k3dIShaderInfoArgument::TEXTURE:
				return sdpString("texture");
			case k3dIShaderInfoArgument::COLOR:
				return sdpString("color");
			case k3dIShaderInfoArgument::SCALAR:
				return sdpString("scalar");
			case k3dIShaderInfoArgument::TIME:
				return sdpString("time");
			case k3dIShaderInfoArgument::ANGLE:
				return sdpString("angle");
			case k3dIShaderInfoArgument::DISTANCE:
				return sdpString("distance");
			case k3dIShaderInfoArgument::AREA:
				return sdpString("area");
			case k3dIShaderInfoArgument::VOLUME:
				return sdpString("volume");
			case k3dIShaderInfoArgument::MASS:
				return sdpString("mass");
			case k3dIShaderInfoArgument::FORCE:
				return sdpString("force");
			case k3dIShaderInfoArgument::PRESSURE:
				return sdpString("pressure");
			case k3dIShaderInfoArgument::STRING:
				return sdpString("string");
			default:
				g_assert_not_reached();	// did you add a new argument type without telling me?
		}
		
	return sdpString();	// Should never be called ...
}

k3dIShaderInfoArgument::ArgumentType ToArgumentType(SDPCSTRING String)
{
	sdpString string(String);
	
	if(string == "unknown")
		return k3dIShaderInfoArgument::UNKNOWN;
	else if(string == "point")
		return k3dIShaderInfoArgument::POINT;
	else if(string == "vector")
		return k3dIShaderInfoArgument::VECTOR;
	else if(string == "orientation")
		return k3dIShaderInfoArgument::ORIENTATION;
	else if(string == "scale")
		return k3dIShaderInfoArgument::SCALE;
	else if(string == "normal")
		return k3dIShaderInfoArgument::NORMAL;
	else if(string == "coordinatesystem")
		return k3dIShaderInfoArgument::COORDINATESYSTEM;
	else if(string == "boolean")
		return k3dIShaderInfoArgument::BOOLEAN;
	else if(string == "optionlist")
		return k3dIShaderInfoArgument::OPTIONLIST;
	else if(string == "texture")
		return k3dIShaderInfoArgument::TEXTURE;
	else if(string == "color")
		return k3dIShaderInfoArgument::COLOR;
	else if(string == "scalar")
		return k3dIShaderInfoArgument::SCALAR;
	else if(string == "time")
		return k3dIShaderInfoArgument::TIME;
	else if(string == "angle")
		return k3dIShaderInfoArgument::ANGLE;
	else if(string == "distance")
		return k3dIShaderInfoArgument::DISTANCE;
	else if(string == "area")
		return k3dIShaderInfoArgument::AREA;
	else if(string == "volume")
		return k3dIShaderInfoArgument::VOLUME;
	else if(string == "mass")
		return k3dIShaderInfoArgument::MASS;
	else if(string == "force")
		return k3dIShaderInfoArgument::FORCE;
	else if(string == "pressure")
		return k3dIShaderInfoArgument::PRESSURE;
	else if(string == "string")
		return k3dIShaderInfoArgument::STRING;
		
	return k3dIShaderInfoArgument::UNKNOWN;
}

sdpResult k3dShaderInstance::Save(sdpXmlIElement* Element, k3dIDependencies* Dependencies, SDPCSTRING Name)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert(Dependencies);
	g_assert_string(Name);

	sdpXmlIElement* shader = sdpXmlCreateElement(Element, "shader", 0, "name", Name, "shadername", ShaderName());
	
	for(ArgumentIterator argument = m_Arguments.begin(); argument != m_Arguments.end(); argument++)
		{
			switch((*argument)->Type())
				{
					case k3dIShaderInfoArgument::UNKNOWN:
					case k3dIShaderInfoArgument::POINT:
					case k3dIShaderInfoArgument::VECTOR:
					case k3dIShaderInfoArgument::ORIENTATION:
					case k3dIShaderInfoArgument::SCALE:
					case k3dIShaderInfoArgument::NORMAL:
					case k3dIShaderInfoArgument::COORDINATESYSTEM:
					case k3dIShaderInfoArgument::BOOLEAN:
					case k3dIShaderInfoArgument::OPTIONLIST:
					case k3dIShaderInfoArgument::TEXTURE:
						{
							std::cout << "k3dShaderInstance::Save() - unsupported argument type!!" << std::endl;
							break;
						}
					case k3dIShaderInfoArgument::COLOR:
					case k3dIShaderInfoArgument::SCALAR:
					case k3dIShaderInfoArgument::TIME:
					case k3dIShaderInfoArgument::ANGLE:
					case k3dIShaderInfoArgument::DISTANCE:
					case k3dIShaderInfoArgument::AREA:
					case k3dIShaderInfoArgument::VOLUME:
					case k3dIShaderInfoArgument::MASS:
					case k3dIShaderInfoArgument::FORCE:
					case k3dIShaderInfoArgument::PRESSURE:
						{
							ChannelArgument* channel = static_cast<ChannelArgument*>(*argument);
							sdpXmlCreateElement(shader, "argument", 0, "name", (*argument)->Name(), "type", ToString((*argument)->Type()), "value", channel->ID());
							break;
						}
					case k3dIShaderInfoArgument::STRING:
						{
							StringArgument* string = static_cast<StringArgument*>(*argument);
							sdpXmlCreateElement(shader, "argument", 0, "name", (*argument)->Name(), "type", ToString((*argument)->Type()), "value", string->Value());
							break;;
						}
					default:
						g_assert_not_reached();	// Unknown argument type!!
				}
		}
	
	return SDP_S_OK;
}

class k3dShaderInstanceLoadShaderInstanceArguments
{
public:
	k3dShaderInstanceLoadShaderInstanceArguments(k3dIDocument* Document, k3dShaderInstance* ShaderInstance)
	{
		m_Document = Document;
		m_ShaderInstance = ShaderInstance;
	}
	
	bool HandleElement(sdpXmlIElement* Element)
	{
		// Get the argument name ...
		sdpString name;
		sdpXmlGetAttribute(Element, "name", name);
		g_return_val_if_fail(name.size(), true);
		
		// Get the argument type ...
		sdpString type;
		sdpXmlGetAttribute(Element, "type", type);
		g_return_val_if_fail(type.size(), true);

		// Get the argument value ...
		sdpString value;
		sdpXmlGetAttribute(Element, "value", value);

		// Create a record for this argument, based on the argument type ...
		k3dShaderInstance::ArgumentBase* argument = k3dShaderInstance::ArgumentBase::LoadArgument(m_Document, name, ToArgumentType(type), value);
		g_return_val_if_fail(argument, true);

		// Store the argument ...
		m_ShaderInstance->m_Arguments.push_back(argument);
				
		return true;
	}
	
protected:
	k3dIDocument* m_Document;
	k3dShaderInstance* m_ShaderInstance;
};

sdpResult k3dShaderInstance::Load(sdpXmlIElement* Element, SDPCSTRING Name)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert_string(Name);

	// Try to find our element ...
	sdpXmlIElement* shaderelement = sdpXmlFindElement(Element, "shader", "name", Name);
	if(!shaderelement)
		return SDP_S_FALSE;

	// Load the shader name ...
	sdpString shadername;
	sdpXmlGetAttribute(shaderelement, "shadername", shadername);
	
	// No name, so we're done ...
	if("" == shadername)
		{
			SetShaderHandle(0);
			return SDP_S_FALSE;
		}

	// Turn the shader name into a handle ...
	K3DSHADERINFOHANDLE handle;
	for(handle = m_Application->GetFirstShaderInfo(); handle; handle = m_Application->GetNextShaderInfo(handle))
		{
			// Test to see if the shader's name matches ...
			k3dIShaderInfo* shader = m_Application->GetShaderInfo(handle);
			g_assert(shader);

			if(shader->Name() == shadername)
				break;
		}
		
	// No handle, so we're done ...
	if(0 == handle)
		{
			SetShaderHandle(0);
			return SDP_S_FALSE;
		}

	// Keep that handle!
	m_ShaderHandle = handle;

	// Load the shader arguments ...
	k3dShaderInstanceLoadShaderInstanceArguments functor(m_Document, this);
	sdpXmlForEachElement(shaderelement, "argument", functor);

	return SDP_S_OK;
}

sdpResult k3dShaderInstance::LoadComplete()
{
	for(ArgumentIterator argument = m_Arguments.begin(); argument != m_Arguments.end(); argument++)
		(*argument)->LoadComplete();
	
	return SDP_S_OK;
}



