// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dShaderSelector class, which provides a user interface for choosing Pixar Renderman shaders
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dshaderselector.h>
#include <sdpgtkevents.h>

k3dShaderSelector::k3dShaderSelector()
{
//	m_ShaderHandle = 0;
//	m_OK = false;
}

k3dShaderSelector::~k3dShaderSelector()
{
}

// UI creation ...
bool k3dShaderSelector::Create(k3dIApplication* Application, k3dShaderInstance* ShaderInstance, k3dShaderArgumentsPage* ShaderArgumentsPage, bool AllowNone, gulong ShaderTypes)
{
	// Sanity checks ...
	g_assert(Application);
	g_assert(ShaderInstance);
	g_assert(ShaderTypes);

	// Get rid of any old stuff ...
	Release();
	
	// Store data ...
	m_Application = Application;
	m_ShaderInstance = ShaderInstance;
	m_ShaderArgumentsPage = ShaderArgumentsPage;
	m_AllowNone = AllowNone;
	m_ShaderTypes = ShaderTypes;

	// Create our UI template ...
	sdpString uitemplate =
	
		"<gtkml>"
			"<optionmenu name=\"optionmenu\">"
				"<menu name=\"menu\">"
				"</menu>"
			"</optionmenu>"
		"</gtkml>\n";

	// Load our template ...
	g_return_val_if_fail(Load(uitemplate, ""), false);

	// Get our menu object ...
	sdpGtkMenu menu = Menu("menu");
	menu.Show();

	// Keep track of the menu item index so we can select one ...
	guint index = 0;
	guint selecteditem = 0;

	// If we give the user the option of selecting nothing ...
	if(m_AllowNone)
		{
			// Create the "none" option ...
			sdpGtkMenuItem menuitem;
			menuitem.Create("--None--");
			menuitem.Show();
			menu.Append(menuitem);

			MapEvent("activate", "onnone", menuitem, true);
			
			index++;
		}

	// For each available shader ...
	for(K3DSHADERINFOHANDLE handle = m_Application->GetFirstShaderInfo(); handle; handle = m_Application->GetNextShaderInfo(handle))
		{
			k3dIShaderInfo* shader = m_Application->GetShaderInfo(handle);

			// Skip shaders that don't match the requested types ...
			gulong test = gulong(shader->Type()) & m_ShaderTypes;
			if(!test)
				continue;

			// Create a menu item for the shader ...
			sdpGtkMenuItem menuitem;
			menuitem.Create(shader->Name());
			menuitem.Show();
			menu.Append(menuitem);			
			
			g_return_val_if_fail(MapEvent("activate", sdpFormatString("select-%d", handle), menuitem, true), false);
			
			if(handle == m_ShaderInstance->ShaderHandle())
				selecteditem = index;
				
			index++;
		}
		
	// Add the menu to our optionmenu ...
	OptionMenu("optionmenu").SetMenu(menu);
	
	// Select the right item ...
	OptionMenu("optionmenu").SetHistory(selecteditem);

	// Map some default events ...
	g_return_val_if_fail(MapEvent("destroy", "ondestroy", Root(), true), false);

	return true;
}

bool k3dShaderSelector::Create(k3dIApplication* Application, k3dShaderInstance* ShaderInstance, k3dShaderArgumentsPage* ShaderArgumentsPage, bool AllowNone, gulong ShaderTypes, GtkContainer* Parent)
{
	// Sanity checks ...
	g_assert(Parent);
	
	// Create the UI ...
	g_return_val_if_fail(Create(Application, ShaderInstance, ShaderArgumentsPage, AllowNone, ShaderTypes), false);

	// Attach ourselves to the parent widget ...
	sdpGtkContainer(Parent).Attach(RootWidget());
	
	return true;
}

void k3dShaderSelector::Release()
{
	// No more events from this point forward ...
	DisconnectAllEvents();
	
	// Delete all of the GTK+ widgets ...
	if(Root())
		RootWidget().Destroy();

	// Clear out our storage ...
	Clear();
}

// Custom event handling ...
void k3dShaderSelector::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "onnone")
		OnNone();
	else if(Event->Name().find("select-") != sdpString::npos)
		OnSelect(Event->Name());
	else if(Event->Name() == "ondestroy")
		OnDestroy();
	else
		sdpGtkObjectContainer::OnEvent(Event);
}

void k3dShaderSelector::OnNone()
{
	SetShader(0);
}

void k3dShaderSelector::OnSelect(sdpString EventName)
{
	sdpString handletext = EventName.right(EventName.size() - 7);
	
	SetShader(K3DSHADERINFOHANDLE(atol(handletext)));
}

void k3dShaderSelector::OnDestroy()
{
	// No more events from this point forward ...
	DisconnectAllEvents();
	
	// Clear out our storage ...
	Clear();
}

void k3dShaderSelector::SetShader(const K3DSHADERINFOHANDLE ShaderHandle)
{
	// If the user chose the same shader, we're done ...
	if(ShaderHandle == m_ShaderInstance->ShaderHandle())
		return;
		
	// Set the new shader handle ...
	m_ShaderInstance->SetShaderHandle(ShaderHandle);
	
	// Notify the shader page to update itself ...
	if(m_ShaderArgumentsPage)
		m_ShaderArgumentsPage->Update();
}

