
// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements objects for storing and manipulating serializable data in a standardized way
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dvariable.h>

#include <sdpxmlutility.h>

#include <k3didocument.h>
#include <k3diobject.h>
#include <k3diobjectinfo.h>
#include <k3dichannel.h>
#include <k3didependencies.h>

#include <k3dcom.h>
#include <k3dmisc.h>

////////////////////////////////////////////////////////////////////////////////////////////////////////
// k3dVariableBase

////////////////////////////////////////////////////////////////////////////////////////////////////////
// k3dObjectVariableBase

k3dObjectVariableBase::k3dObjectVariableBase()
{
	m_Object = 0;
	m_ObjectID = 0;
	m_AutoDestroy = false;
}

k3dObjectVariableBase::~k3dObjectVariableBase()
{
	if(m_AutoDestroy && m_Object)
		k3dDeleteObject(m_Object);

	m_Object = 0;
	m_ObjectID = 0;
}

bool k3dObjectVariableBase::Create(k3dIDocument* Document, sdpCLSID ClassID, SDPCSTRING Name, bool Edit, bool UserVisible, bool Saveable)
{
	// Sanity checks ...
	g_assert(Document);
	g_assert(ClassID != sdpCLSIDNull);

	// Create the object ...
	m_Object = Document->CreateObject(ClassID, Edit, UserVisible, Saveable);
	g_return_val_if_fail(m_Object, false);

	// Set the object name ...
	g_return_val_if_fail(sdpSucceeded(m_Object->ObjectContext()->SetName(Name)), false);

	return true;
}

sdpResult k3dObjectVariableBase::Save(sdpXmlIElement* Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert(Dependencies);

	k3dID id = m_Object ? m_Object->ObjectContext()->ObjectID() : 0;
	sdpXmlCreateElement(Element, "object", 0, "name", m_Name, "id", id);

	return SDP_S_OK;
}

sdpResult k3dObjectVariableBase::Load(sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert(Element->Name() == sdpString("object"));

	// Pull out the name ...
	sdpString name;
	g_return_val_if_fail(sdpXmlGetAttribute(Element, "name", name), SDP_E_FAIL);
	g_assert(name == m_Name);

	// Pull out the id ...
	sdpString id;
	g_return_val_if_fail(sdpXmlGetAttribute(Element, "id", id), SDP_E_FAIL);

	// Convert it ...
	m_ObjectID = atol(id);

	return SDP_S_OK;
}

sdpResult k3dObjectVariableBase::AttachObject(sdpIUnknown* Object)
{
	// Release any previous object ...
	ReleaseObject();

	// We allow the caller to assign a null object ...
	if(Object)
		{
			k3dIObject* object = sdpQueryInterface<k3dIObject>(Object);
			g_return_val_if_fail(object, SDP_E_FAIL);
		
			// Store the new object ...
			m_Object = object;

			// Get the new object's ID ...
			m_ObjectID = m_Object->ObjectContext()->ObjectID();
			
			// We want the new object to notify us if it goes away ...
			m_Object->AddClient(this, 0, "delete");
		}

	return SDP_S_OK;
}

void k3dObjectVariableBase::ReleaseObject()
{
	if(m_Object)
		{
			// Let the object know not to notify us anymore ...
			m_Object->RemoveClient(this, 0, "delete");
			m_Object->Release();
			m_Object = 0;
		}

	m_ObjectID = 0;
}

sdpResult k3dObjectVariableBase::OnEvent(k3dIObject* Object, SDPCSTRING Event, gulong ClientData)
{
	// Sanity checks ...
	g_return_val_if_fail(Object == m_Object, SDP_E_FAIL);

	// Release our object ...
	if(Event == sdpString("delete"))
		{
			ReleaseObject();
			return SDP_S_OK;
		}

	return k3dEventClient::OnEvent(Object, Event, ClientData);
}

/////////////////////////////////////////////////////////////
// k3dObjectVariable


//////////////////////////////////////////////////////////////
// k3dChannelVariable

k3dChannelVariable::k3dChannelVariable()
{
}

///////////////////////////////////////////////////////////////
// k3dScalarChannel

bool k3dScalarChannel::Create(k3dIDocument* Document, SDPCSTRING Name, double FirstValue, bool ShowTime, double MinY, double MaxY, SDPCSTRING XLabel, SDPCSTRING YLabel, k3dIChannel::DataType Data, bool Saveable, k3dIEventClient* EventClient)
{
	g_return_val_if_fail(k3dChannelVariable::Create(Document, sdpChannel, Name, false, false, Saveable), false);

	g_return_val_if_fail(sdpSucceeded(m_Interface->SetDatatype(Data)), false);
	
	k3dChannelSetConstantValue(m_Interface, FirstValue);
	k3dChannelSetValueRange(m_Interface, MinY, MaxY);
	
	if(EventClient)
		Object()->AddClient(EventClient, 0, "channel-changed");
	
//	g_return_val_if_fail(sdpSucceeded(m_Interface->SetShowTime(ShowTime)), false);
//	g_return_val_if_fail(sdpSucceeded(m_Interface->SetLabels(XLabel, YLabel)), false);

	return true;
}

///////////////////////////////////////////////////////////////
// k3dColorChannel

bool k3dColorChannel::Create(k3dIDocument* Document, SDPCSTRING Name, double FirstColor[3], bool ShowTime, SDPCSTRING XLabel, SDPCSTRING YLabel, bool Saveable, k3dIEventClient* EventClient)
{
	g_return_val_if_fail(k3dChannelVariable::Create(Document, sdpChannel, Name, false, false, Saveable), false);

	g_return_val_if_fail(sdpSucceeded(m_Interface->SetDatatype(k3dIChannel::DATA_COLOR)), false);

	k3dChannelSetConstantValue(m_Interface, FirstColor);
	
	if(EventClient)
		Object()->AddClient(EventClient, 0, "channel-changed");
	
//	g_return_val_if_fail(sdpSucceeded(m_Interface->SetShowTime(ShowTime)), false);
//	g_return_val_if_fail(sdpSucceeded(m_Interface->SetLabels(XLabel, YLabel)), false);

	return true;
}


