// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the renderframe application, which renders frames scheduled as part of a job with the virtual render farm
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3direnderframe.h>
#include <k3ddocument.h>

#include <sdpxmldocument.h>
#include <sdpxmlutility.h>
#include <sdpfile.h>
#include <sdputility.h>

#include <list>
#include <iostream>
#include <fstream>

#include <stdio.h>

#ifdef SDPUNIX
#include <unistd.h>
#endif // SDPUNIX

#include <time.h>

/// Returns the current time, formatted as an sdpString
sdpString Now()
{
	// Get the current time ...
	time_t now_t;
	time(&now_t);
	
	tm* now_tm = localtime(&now_t);
	
	// Format it ...
	sdpString result;
	strftime(result.buffer(64), 64, "%m/%d/%Y %H:%M:%S", now_tm);

	return result;
}

/// Parses the frame and global options files to obtain the appropriate rendering command
bool GetFrameCommand(sdpXmlDocument& GlobalOptions, sdpString& Command)
{
	// Load the frame control file ...
	const sdpString controlpath = k3dIRenderFrame::ControlFile;
	k3dDocument framecontrol;
	if(!framecontrol.LoadFile(controlpath))
		{
			std::cerr << "Error loading " << controlpath << " configuration file" << std::endl;
			return false;
		}

	// Get the frame ...
	sdpXmlIElement* frame = sdpXmlGetXPathElement(framecontrol.GetRootElement(), "frame");
	if(!frame)
		{
			std::cerr << controlpath << " does not contain frame data" << std::endl;
			return false;
		}
		
	// Load frame data ...
	bool requestedviewduringrender = false;
	sdpXmlGetAttribute(frame, "viewduringrender", requestedviewduringrender);
	
	// Get the requested engine ...
	sdpString requestedenginetype;
	sdpXmlGetAttribute(frame, "enginetype", requestedenginetype);
	if(0 == requestedenginetype.size())
		{
			std::cerr << "No enginetype specified" << std::endl;
			return false;
		}
	
	sdpString requestedengine;
	sdpXmlGetAttribute(frame, "engine", requestedengine);
	if(0 == requestedengine.size())
		{
			std::cerr << "No engine specified" << std::endl;
			return false;
		}

	// Poke through global options and look for a match ...
	sdpString xpath = "application/helpers/command[@type=\"render\" and @enginetype=\"" + requestedenginetype + "\" and @engine=\"" + requestedengine + "\" and @viewduringrender=\"" + sdpString(requestedviewduringrender) + "\"]";
	sdpXmlElementArray commands = sdpXmlXPathSelect(&GlobalOptions, xpath);
	if(0 == commands.size())
		{
			std::cerr << "Could not find requested render engine" << std::endl;
			return false;
		}
	else if(1 != commands.size())
		{
			std::cerr << "Too many matches for requested render engine" << std::endl;
			return false;
		}
	
	Command = commands[0]->Text();	
	if(0 == Command.size())
		{
			std::cerr << "Empty render engine command" << std::endl;
			return false;
		}

	return true;
}

/// Handles all processing for the given frame
bool RenderFrame(sdpXmlDocument& GlobalOptions, SDPCSTRING FrameDirectory)
{
	// Sanity checks ...
	g_assert_string(FrameDirectory);

	// Setup some standard paths ...
	const sdpString readypath = sdpAppendPath(FrameDirectory, k3dIRenderFrame::StatusReadyFile);
	const sdpString runningpath = sdpAppendPath(FrameDirectory, k3dIRenderFrame::StatusRunningFile);
	const sdpString errorpath = sdpAppendPath(FrameDirectory, k3dIRenderFrame::StatusErrorFile);
	const sdpString completepath = sdpAppendPath(FrameDirectory, k3dIRenderFrame::StatusCompleteFile);

	// Make sure the frame exists ...
	if(!sdpFileExists(FrameDirectory))
		{
			std::cerr << Now() << " - " << FrameDirectory << " not found" << std::endl;
			return false;
		}

	// Skip the frame if it's complete ...
	std::ifstream completefile(completepath, std::ios::in);
	if(completefile.good())
		return true;

	// Skip the frame if it errored out ...
	std::ifstream errorfile(errorpath, std::ios::in);
	if(errorfile.good())
		return true;

	// Skip the frame if it's running ...
	std::ifstream runningfile(runningpath, std::ios::in);
	if(runningfile.good())
		return true;

	// Make sure the frame is ready ...
	std::ifstream readyfile(readypath, std::ios::in);
	if(!readyfile.good())
		{
			std::cerr << Now() << " - " << FrameDirectory << " is not ready" << std::endl;
			return false;
		}

	// Standard logging ...
	std::cout << Now() << " - Running " << FrameDirectory << std::endl;

	// Switch the frame status to running ...
	rename(readypath, runningpath);
	
	// Setup our execution environment ...
#ifdef SDPWIN32
	SetCurrentDirectory(FrameDirectory);
#else // SDPWIN32
	chdir(FrameDirectory);
#endif // !SDPWIN32
			
	// Get the render command for this frame's options ...
	sdpString commandline;
	GetFrameCommand(GlobalOptions, commandline);
	
	// Log the render output ...
	commandline += " 2> ";
	commandline += k3dIRenderFrame::EngineErrorFile;
	commandline += " | tee ";
	commandline += k3dIRenderFrame::EngineLogFile;
	
	// Substitute the RIB file path ...
	commandline.formattedreplace('%', "p", "world.rib");
	
	// Execute the command ...
	std::cout << Now() << " - Executing " << commandline << std::endl;
	system(commandline);
			
	// Switch the frame status to complete ...
	rename(runningpath, completepath);
	
	// Standard logging ...
	std::cout << Now() << " - Complete" << std::endl;
	
	return true;
}

/// Prints usage info
void Usage()
{
	std::cout << "usage: renderframe [options] [directory ...]" << std::endl;
	std::cout << std::endl;
	std::cout << "  -h, --help               prints this help information and exits" << std::endl;
	std::cout << "      --optionspath [path] overrides the path for storing user options" << std::endl;
	std::cout << "      --version            prints program version information and exits" << std::endl;
	std::cout << std::endl;
}

/// Prints version info
void Version()
{
#include "../../k3d_version.h"

	std::cout << "K-3D RenderFrame Version " << VERSION_STRING << std::endl;
	std::cout << "Copyright (c) 1995-2000, Timothy M. Shead.  See the AUTHORS file for contributors." << std::endl;
	std::cout << "Licensed by the GNU General Public License.  See the COPYING file for details." << std::endl;
	std::cout << "K-3D Home Page: http://www.k-3d.com" << std::endl;
	std::cout << "Bug reports to tshead@k-3d.com" << std::endl;
}

/// Program main
int main(int argc, char* argv[])
{
	sdpString optionspath;

	// Convert our arguments into an array of strings ...
	typedef std::list<sdpString> ArgumentList;
	typedef ArgumentList::iterator ArgumentIterator;

	ArgumentList arguments;
	for(int i = 1; i < argc; i++)
		arguments.push_back(argv[i]);

	// Test our arguments ...
	if(0 == arguments.size())
		{
			Usage();
			return 1;
		}

	for(ArgumentIterator argument = arguments.begin(); argument != arguments.end(); )
		{
			if(*argument == "--help" || *argument == "-h")
				{
					Usage();
					return 0;
				}
			else if(*argument == "--version")
				{
					Version();
					return 0;
				}
			else if(*argument == "--optionspath")
				{
					ArgumentIterator name = argument++;
					ArgumentIterator value = argument++;
					optionspath = *(value);
					arguments.erase(name);
					arguments.erase(value);
				}
			else
				{
					argument++;
				}
		}

	// Load our global options file ...
	k3dDocument options;
	if(!options.LoadOptions(optionspath))
		{
			std::cerr << "*** renderframe: error loading configuration file " << options.OptionsFilePath(optionspath) << std::endl;
			return 1;
		}

	// Each remaining argument should be a frame path to render ...						
	int result = 0;
	for(ArgumentIterator frame = arguments.begin(); frame != arguments.end(); frame++)
		{
			if(!RenderFrame(options, *frame))
				result = 1;
		}

	return result;
}


