#ifndef FRACTALTERRAIN_H
#define FRACTALTERRAIN_H

// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the CFractalTerrain K-3D object, which creates fractal landscapes
		\author Romain Behar (romainbehar@yahoo.com)
		\author Timothy M. Shead (tshead@k-3d.com)
*/

// Forward class declarations ...
class CFractalTerrain;

#include <k3ddialog.h>
#include <k3dtrianglemesh.h>
#include <k3deditorrender.h>
#include <k3dvariable.h>

#include <k3dimaterial.h>

#include <complex>

extern const sdpCLSID sdpFractalTerrain;

/////////////////////////////////////////////////////////////////////////////
// CFractalTerrainDialog

class CFractalTerrainDialog : public k3dDialog
{
friend CFractalTerrain;
public:
	CFractalTerrainDialog();

	bool Create(CFractalTerrain* Object);

protected:
	CFractalTerrain* m_Object;

	// Event handling ...
	void OnEvent(sdpGtkEvent* Event);

	void OnXLength();
	void OnYLength();
	void OnZLength();
	void OnIterations();
	void OnFractalDimension();
	void OnRandomSeed();

	// Control updates ...
	void UpdateControls();
};


/////////////////////////////////////////////////////
// CFractalTerrain

class CFractalTerrain :	public k3dTriangleMesh
{
friend class CFractalTerrainDialog;
public:
	CFractalTerrain();
	~CFractalTerrain();

	// k3dIUnknown overrides ...
	sdpResult QueryInterface(const sdpIID& InterfaceID, void** Interface);
	gulong AddRef();
	gulong Release();

	// k3dIObject overrides ...
	sdpResult EditObject();
	sdpResult ObjectRenamed(SDPCSTRING NewName);
	sdpResult ObjectDeleted();

	// k3dIPersist overrides ...
	sdpResult InitNew();
	sdpResult LoadComplete();

	// Where it all happens ...
	void OnModify(bool SaveAction);

protected:
	CFractalTerrainDialog m_UserInterface;

	bool m_CreateGeometry;

	k3dVariable<gdouble> m_XLength;
	k3dVariable<gdouble> m_YLength;
	k3dVariable<gdouble> m_ZLength;
	k3dVariable<gulong> m_Iterations;
	k3dVariable<gdouble> m_FractalDimension;
	k3dVariable<gulong> m_RandomSeed;

	// Use 2D Fast Fourier Transform
	void FFTTerrain();

	class terrainSeed
	{
	public:
		/// Store four random numbers
		sdpVector4 m_Seeds;

		terrainSeed() {}
		terrainSeed(sdpVector4 seed) { m_Seeds = seed; }

		// A random variable
		gdouble drand()
			{
				gdouble	d = (gdouble)rand()/(gdouble)RAND_MAX;
				if(d < 1e-6) return 1e-6;
				return d;
			}

		// A normal random variable
		gdouble normaldrand(int k, int l, gdouble d)
			{
				gdouble dk = (gdouble)k, dl = (gdouble)l;
				gdouble sigma = 1/exp((4-d)/2*log(dk*dk + dl*dl));
				return sigma*sqrt(-2*log(m_Seeds.n[0]))*cos(2*sdpPi*m_Seeds.n[1]);
			}

		// A normal random complex variable
		std::complex<gdouble> cplxnormaldrand(int k, int l, gdouble d)
			{
				gdouble dk = (gdouble)k, dl = (gdouble)l;
				gdouble sigma = 1/exp((4-d)/2*log(dk*dk + dl*dl));
				return sigma*std::complex<gdouble>(sqrt(-2*log(m_Seeds.n[0]))*cos(2*sdpPi*m_Seeds.n[1]), sqrt(-2*log(m_Seeds.n[2]))*cos(2*sdpPi*m_Seeds.n[3]));
			}

		void Randomize()
			{
				for(gulong i = 0; i < 4; i++)
					m_Seeds.n[i] = drand();
			}
	};

	class fftVector
	{
	public:
		std::vector< std::complex<double> > P;

		fftVector();
		fftVector(int Size)
			{
				m_Size = Size;

				P.reserve(m_Size);
				Roots.reserve(m_Size);
				Temp.reserve(m_Size);


				std::complex<double>	Wj, Wt;

				Roots[0] = 1;
				Wt = std::polar(1.0, sdpPiTimes2 / m_Size);
				Wj = Wt;
				for(int i = m_Size-1; i > 0; i--)
					{
						Roots[i] = Wt;
						Wt *= Wj;
					}
			}

		~fftVector()
			{
				P.clear();
				Roots.clear();
				Temp.clear();
			}

		void InvFFT()
			{
				RealFFT(m_Size, 0);
			}

		void RealFFT(gulong d, gulong k)
			{
				if(d >= 2)
					{
						gulong i, j;

						for(i = 0; i < d/2; i++)
							{
								j = k + 2*i;
								Temp[i] = P[j];
								Temp[i+d/2] = P[j+1];
							}

						for(i = 0; i < d; i++)
							P[k+i] = Temp[i];

						RealFFT(d/2, k);
						RealFFT(d/2, k+d/2);

						j = m_Size / d;
						for(i = 0; i < d/2; i++)
							{
								std::complex<double> aux = Roots[i*j] * P[k+d/2+i];
								Temp[i] = P[k+i] + aux;
								Temp[i+d/2] = P[k+i] - aux;
							}

						for(i = 0; i < d; i++) P[k+i] = Temp[i];
					}
			}

	protected:
		int m_Size;

		std::vector< std::complex<double> > Roots;
		std::vector< std::complex<double> > Temp;
	};
	typedef fftVector *fvector;

	// Random seeds for a fractal terrain (built using rules)
	std::vector< std::vector<terrainSeed> > m_TerrainSeeds;

	// FFT fractal terrain
	fvector* m_Terrain;

	// Generate a fractal terrain using preset random values (m_TerrainSeeds)
	void Mountain(gulong N, gdouble d);
	void ClearMountain(gulong N);
};


#endif // FRACTALTERRAIN_H
