#ifndef PARTICLEBURST_H
#define PARTICLEBURST_H

// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the CParticleBurst K-3D object, which creates a parametric particle system
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dilocation.h>
#include <k3dstate.h>
#include <k3dtransformobject.h>
#include <k3dilocationserver.h>
#include <k3ditimebase.h>
#include <k3dvariable.h>
#include <k3ddialog.h>
#include <k3dchannelbutton.h>

#include <vector>

/////////////////////////////////////////////////////////////////////////////
// CParticleBurstDialog

/// Encapsulates the ParticleBurst UI
class CParticleBurstDialog : public k3dDialog
{
friend class CParticleBurst;
public:
	CParticleBurstDialog();

	/// Called to initialize the UI
	bool Create(CParticleBurst* Object);

protected:
	/// Stores a pointer to the owning object
	CParticleBurst* m_Object;

	/// Handles our custom events
	void OnEvent(sdpGtkEvent* Event);

	/// Called to initialize controls at creation-time
	bool InitializeControls();

	/// Called to update the state of any UI controls
	void UpdateControls();
	
	//	k3dShaderArgumentsPage m_ShaderPage;
};

////////////////////////////////////////////////////////////////
// CParticle

/// Encapsulates a single particle within a ParticleBurst particle system
class CParticle : public k3dILocation,
									public k3dState
{
public:
	CParticle();

	// k3dIUnknown overrides
	sdpResult QueryInterface(const sdpIID& InterfaceID, void** Interface) { return SDP_E_NO_INTERFACE; }
	gulong AddRef() { return 0; }
	gulong Release() { return 0; }

	// k3dILocation overrides ...
	k3dIState* State();
	sdpResult GetOldXYZ(double Coords[3]);
	sdpResult GetLocalXYZ(double Coords[3]);
	sdpResult GetAnimXYZ(double Coords[3]);
	sdpResult GetWorldXYZ(double Coords[3]);
	sdpResult GetCameraXYZ(double Coords[3]);
	sdpResult GetScreenXYZ(double Coords[3]);
	sdpResult GetNormalXYZ(double Coords[3]);
	sdpResult GetTextureXYZ(double Coords[3]);
	sdpResult GetImplicitTextureXYZ(double Coords[3]);
	sdpResult SetOldXYZ(double Coords[3]);
	sdpResult SetLocalXYZ(double Coords[3]);
	sdpResult SetAnimXYZ(double Coords[3]);
	sdpResult SetWorldXYZ(double Coords[3]);
	sdpResult SetCameraXYZ(double Coords[3]);
	sdpResult SetScreenXYZ(double Coords[3]);
	sdpResult SetNormalXYZ(double Coords[3]);
	sdpResult SetTextureXYZ(double Coords[3]);
	sdpResult SetImplicitTextureXYZ(double Coords[3]);

	sdpVector2 m_Time;
	sdpVector2 m_Distance;

	sdpVector3 m_DirectionXYZ;
	sdpVector3 m_TwistXYZ[2];
	sdpVector3 m_DriftXYZ[2];

	sdpVector2 m_Strength;
	sdpVector2 m_Size;
	sdpVector2 m_Width;
	sdpVector2 m_Height;
	sdpVector2 m_Angle;

	sdpVector3 m_Color;

	sdpVector3 m_BirthXYZ;
	sdpVector3 m_AnimXYZ;
	sdpVector3 m_WorldXYZ;
	sdpVector3 m_CameraXYZ;
	sdpVector3 m_ScreenXYZ;
};



/////////////////////////////////////////////////////////////
// CParticleBurst

class CParticleBurst : public k3dTransformObject,
											 public k3dILocationServer
{
friend class CParticleBurstDialog;
public:
	CParticleBurst();
	~CParticleBurst();

	bool Create();
	double DRand(gulong& Seed);

	// k3dIUnknown overrides
	sdpResult QueryInterface(const sdpIID& InterfaceID, void** Interface);
	gulong AddRef();
	gulong Release();

	// k3dIObject overrides
	sdpResult SetObjectContext(k3dIObjectContext* Context);
	sdpResult EditObject();
	sdpResult ObjectRenamed(SDPCSTRING NewName);
	sdpResult ObjectDeleted();

	// k3dIEventClient overrides
	sdpResult OnEvent(k3dIObject* Object, SDPCSTRING Event, gulong ClientData);

	// k3dIAnimate overrides
	sdpResult PopAnimate();

	// k3dIEditorRender overrides
	sdpResult EditorRender(k3dIEditorEngine* Engine);

	// k3dIRiRender overrides
	sdpResult RiRender(k3dIRenderFrame* Frame, k3dIRiEngine* Engine);

	// k3dILocationServer overrides
	gulong GetLocationCount();
	K3DLOCATIONHANDLE GetFirstLocation();
	K3DLOCATIONHANDLE GetNextLocation(K3DLOCATIONHANDLE Location);
	k3dILocation* GetLocation(K3DLOCATIONHANDLE Location);

	// k3dIPersist overrides
	sdpResult InitNew();
	sdpResult LoadComplete();

	// Helper functions
	void CreateParticles();
	void RecalculateParticles();
	void RecalculateEmitterPosition();

	void UpdateFull();
	void UpdateCreateParticles();
	void UpdateCalculateParticles();

	void OpenGLPreviewRenderPoints();

	/// Enumerates the types of entities controlled by this burst
	typedef enum
	{
		POINTS = 1,
		FLARES = 2,
		POLYGONS = 3,
		MESHES = 4,
	} ParticleType;

	class k3dVariableParticleType : public k3dCustomVariable<ParticleType>
	{
	protected:
		sdpString ConvertToString() { return sdpFormatString("%d", m_Value); }
		bool ConvertFromString(SDPCSTRING String) { m_Value = ParticleType(atoi(String)); return true; }
	};

	/// Enumerates how the particle colors are affected by the burst color channel
	typedef enum
	{
		RANDOM = 1,
		LIFETIME = 2,
		GLOBAL = 3,
	} ColorType;

	class k3dVariableColorType : public k3dCustomVariable<ColorType>
	{
	protected:
		sdpString ConvertToString() { return sdpFormatString("%d", m_Value); }
		bool ConvertFromString(SDPCSTRING String) { m_Value = ColorType(atoi(String)); return true; }
	};

	k3dVariableParticleType m_ParticleType;
	k3dVariableColorType m_ColorType;
	k3dVariable<gdouble> m_EmitterPositionSamples;
	k3dVariable<gdouble> m_BirthRateSamples;
	k3dVariable<gulong> m_RandomSeed;
	k3dVariable<bool> m_DirectionRandom;
	k3dVariable<bool> m_DirectionEmitter;
	k3dVariable<bool> m_TwistRandom;
	k3dVariable<bool> m_TwistRelative;
	k3dVariable<bool> m_TwistLifetime;
	k3dVariable<bool> m_WidthRandom;
	k3dVariable<bool> m_WidthRelative;
	k3dVariable<bool> m_WidthLifetime;
	k3dVariable<bool> m_HeightRandom;
	k3dVariable<bool> m_HeightRelative;
	k3dVariable<bool> m_HeightLifetime;
	k3dVariable<bool> m_AngleAlign;
	k3dVariable<bool> m_AngleRandom;
	k3dVariable<bool> m_AngleRelative;
	k3dVariable<bool> m_AngleLifetime;
	k3dVariable<bool> m_DistanceRandom;
	k3dVariable<bool> m_DistanceRelative;
	k3dVariable<bool> m_DistanceLifetime;
	k3dVariable<bool> m_DriftRandom;
	k3dVariable<bool> m_DriftRelative;
	k3dVariable<bool> m_DriftLifetime;
	k3dVariable<bool> m_StrengthRandom;
	k3dVariable<bool> m_StrengthRelative;
	k3dVariable<bool> m_StrengthLifetime;
	k3dVariable<bool> m_SizeRandom;
	k3dVariable<bool> m_SizeRelative;
	k3dVariable<bool> m_SizeLifetime;

	k3dScalarChannel m_Globaltime;
	k3dScalarChannel m_Birthrate;
	k3dScalarChannel m_Minlife;
	k3dScalarChannel m_Maxlife;
	k3dScalarChannel m_Globalstrength;

	k3dScalarChannel m_Minstrength;
	k3dScalarChannel m_Minstrengthend;
	k3dScalarChannel m_Maxstrength;
	k3dScalarChannel m_Maxstrengthend;
	k3dScalarChannel m_Strengthlife;
	
	k3dScalarChannel m_Minsize;
	k3dScalarChannel m_Minsizeend;
	k3dScalarChannel m_Maxsize;
	k3dScalarChannel m_Maxsizeend;
	k3dScalarChannel m_Sizelife;

	k3dScalarChannel m_Mindirectionx;
	k3dScalarChannel m_Maxdirectionx;
	k3dScalarChannel m_Mindirectiony;
	k3dScalarChannel m_Maxdirectiony;
	k3dScalarChannel m_Mindirectionz;
	k3dScalarChannel m_Maxdirectionz;

	k3dScalarChannel m_Mindistance;
	k3dScalarChannel m_Mindistanceend;
	k3dScalarChannel m_Maxdistance;
	k3dScalarChannel m_Maxdistanceend;
	k3dScalarChannel m_Distancelife;

	k3dScalarChannel m_Mindriftx;
	k3dScalarChannel m_Mindriftxend;
	k3dScalarChannel m_Maxdriftx;
	k3dScalarChannel m_Maxdriftxend;
	k3dScalarChannel m_Driftxlife;
	k3dScalarChannel m_Mindrifty;
	k3dScalarChannel m_Mindriftyend;
	k3dScalarChannel m_Maxdrifty;
	k3dScalarChannel m_Maxdriftyend;
	k3dScalarChannel m_Driftylife;
	k3dScalarChannel m_Mindriftz;
	k3dScalarChannel m_Mindriftzend;
	k3dScalarChannel m_Maxdriftz;
	k3dScalarChannel m_Maxdriftzend;
	k3dScalarChannel m_Driftzlife;

	k3dScalarChannel m_Mintwistx;
	k3dScalarChannel m_Mintwistxend;
	k3dScalarChannel m_Maxtwistx;
	k3dScalarChannel m_Maxtwistxend;
	k3dScalarChannel m_Twistxlife;
	k3dScalarChannel m_Mintwisty;
	k3dScalarChannel m_Mintwistyend;
	k3dScalarChannel m_Maxtwisty;
	k3dScalarChannel m_Maxtwistyend;
	k3dScalarChannel m_Twistylife;
	k3dScalarChannel m_Mintwistz;
	k3dScalarChannel m_Mintwistzend;
	k3dScalarChannel m_Maxtwistz;
	k3dScalarChannel m_Maxtwistzend;
	k3dScalarChannel m_Twistzlife;

	k3dScalarChannel m_Minwidth;
	k3dScalarChannel m_Minwidthend;
	k3dScalarChannel m_Maxwidth;
	k3dScalarChannel m_Maxwidthend;
	k3dScalarChannel m_Widthlife;
	k3dScalarChannel m_Minheight;
	k3dScalarChannel m_Minheightend;
	k3dScalarChannel m_Maxheight;
	k3dScalarChannel m_Maxheightend;
	k3dScalarChannel m_Heightlife;
	k3dScalarChannel m_Minangle;
	k3dScalarChannel m_Minangleend;
	k3dScalarChannel m_Maxangle;
	k3dScalarChannel m_Maxangleend;
	k3dScalarChannel m_Anglelife;
	k3dColorChannel m_Color;

	typedef std::vector<CParticle> Particles;
	typedef Particles::iterator ParticleIterator;
	Particles m_Particles;

	bool m_Animate;

	k3dITimeBase* m_TimeBase;

	CParticleBurstDialog m_UserInterface;
};

#endif // PARTICLEBURST_H
