// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CBooleanOperator K-3D object, which allows geometry to be rendered with boolean effects
		\author Tim Shead (tshead@k-3d.com)
*/

#include "booleanoperator.h"

#include <sdpfile.h>
#include <sdpgtkevent.h>

#include <k3diobjectcontext.h>
#include <k3diriengine.h>
#include <k3dicamera.h>
#include <k3dmisc.h>

CBooleanOperator::CBooleanOperator()
{
	AddVariable(&m_Type("type", BOOLEAN_INTERSECTION));
}

CBooleanOperator::~CBooleanOperator()
{
}

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CBooleanOperator::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIAnimate>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dIRiRender>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dIEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dIObject>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dIPersist>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dITimeServer>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dIUserEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIObject*>(this), InterfaceID, Interface)) return SDP_S_OK;											

	return SDP_E_NO_INTERFACE;
}

gulong CBooleanOperator::AddRef()
{
	return sdpComAddRef(m_ReferenceCount);
}

gulong CBooleanOperator::Release()
{
	return sdpComRelease(this, m_ReferenceCount);
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CBooleanOperator::EditObject()
{
	g_return_val_if_fail(m_UserInterface.Create(this), SDP_E_FAIL);
	return SDP_S_OK;
}

sdpResult CBooleanOperator::ObjectRenamed(SDPCSTRING NewName)
{
	m_UserInterface.ObjectRenamed(NewName);
	return k3dPersistObject::ObjectRenamed(NewName);
}

sdpResult CBooleanOperator::ObjectDeleted()
{
	m_UserInterface.ObjectDeleted();
	return k3dPersistObject::ObjectDeleted();
}

///////////////////////////////////////////////////////////////////////
// k3dIRiRender overrides

class CBooleanOperatorRenderSolid
{
public:
	bool HandleDescendantObject(k3dIObjectContext* Context)
	{
		k3dIRiRender* rirender = sdpQueryInterface<k3dIRiRender>(Context->Object());
		if(rirender)
			{
				rirender->RiSetSolid(true);
				rirender->Release();
			}
	
		return true;
	}
};

sdpResult CBooleanOperator::RiPushRender(k3dIRenderFrame* Frame, k3dIRiEngine* Engine)
{
	// Record the type of boolean ...
	switch(m_Type())
		{
			case BOOLEAN_INTERSECTION:
				Engine->RiSolidBegin(k3dIRiEngine::RI_INTERSECTION);
				break;

			case BOOLEAN_UNION:
				Engine->RiSolidBegin(k3dIRiEngine::RI_UNION);
				break;

			case BOOLEAN_DIFFERENCE:
				Engine->RiSolidBegin(k3dIRiEngine::RI_DIFFERENCE);
				break;

			default:
				g_assert_not_reached();
				break;
		}
		
	// Make sure our descendants know that they should render themselves as "solid" ...
	CBooleanOperatorRenderSolid functor;
	k3dForEachDescendant(Document(), ObjectContext(), functor);

	return SDP_S_OK;
}

sdpResult CBooleanOperator::RiPopRender(k3dIRenderFrame* Frame, k3dIRiEngine* Engine)
{
	Engine->RiSolidEnd();

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIPersist overrides

/////////////////////////////////////////////////////////////////////////////
// CBooleanOperatorDialog

CBooleanOperatorDialog::CBooleanOperatorDialog()
{
	m_Object = 0;
}

bool CBooleanOperatorDialog::Create(CBooleanOperator* Object)
{
	// Store our parent object ...
	m_Object = Object;

	// Load our template ...
	g_return_val_if_fail(k3dDialog::Create(Object, false, true, "booleanoperator.gtkml", "booleanoperator.html"), false);

	return true;
}

void CBooleanOperatorDialog::UpdateControls()
{
	ToggleButton("intersection").SetState(m_Object->m_Type() == CBooleanOperator::BOOLEAN_INTERSECTION);
	ToggleButton("union").SetState(m_Object->m_Type() == CBooleanOperator::BOOLEAN_UNION);
	ToggleButton("difference").SetState(m_Object->m_Type() == CBooleanOperator::BOOLEAN_DIFFERENCE);
}

void CBooleanOperatorDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "onintersection")
		OnIntersection();
	else if(Event->Name() == "onunion")
		OnUnion();
	else if(Event->Name() == "ondifference")
		OnDifference();
	else
		k3dDialog::OnEvent(Event);
}

void CBooleanOperatorDialog::OnIntersection()
{
	if(ToggleButton("intersection").GetState())
		m_Object->m_Type() = CBooleanOperator::BOOLEAN_INTERSECTION;
}

void CBooleanOperatorDialog::OnUnion()
{
	if(ToggleButton("union").GetState())
		m_Object->m_Type() = CBooleanOperator::BOOLEAN_UNION;
}

void CBooleanOperatorDialog::OnDifference()
{
	if(ToggleButton("difference").GetState())
		m_Object->m_Type() = CBooleanOperator::BOOLEAN_DIFFERENCE;
}


