// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CFractalTerrain K-3D object, which creates fractal landscapes
		\author Romain Behar (romainbehar@yahoo.com)
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include "fractalterrain.h"

#include <sdpfile.h>
#include <sdpgtkevents.h>

#include <k3diobjectcontext.h>
#include <k3dicamera.h>
#include <k3dmisc.h>

#include <sdpgl.h>

const sdpCLSID sdpFractalTerrain(0xF0000000, 0x00000000, 0x00000000, 0x00000001);

CFractalTerrain::CFractalTerrain()
{
	AddVariable(&m_XLength("xlength", 10.0));
	AddVariable(&m_YLength("ylength", 0.4));
	AddVariable(&m_ZLength("zlength", 10.0));
	AddVariable(&m_Iterations("iterations", 5));
	AddVariable(&m_FractalDimension("fractaldimension", 2.1));
	AddVariable(&m_RandomSeed("randomseed", 1));

	m_Terrain = 0;
	m_CreateGeometry = true;
}

CFractalTerrain::~CFractalTerrain()
{
	if(m_Terrain != 0)
		delete [] m_Terrain;
}

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CFractalTerrain::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIAnimate>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIEditorRender>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dILocationServer>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIMaterialServer>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIObject>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIPersist>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIRiRender>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dITarget>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dITimeServer>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dITransform>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIUserEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIObject*>(this), InterfaceID, Interface)) return SDP_S_OK;

	return SDP_E_NO_INTERFACE;
}

gulong CFractalTerrain::AddRef()
{
	return sdpComAddRef(m_ReferenceCount);
}

gulong CFractalTerrain::Release()
{
	return sdpComRelease(this, m_ReferenceCount);
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CFractalTerrain::EditObject()
{
	g_return_val_if_fail(m_UserInterface.Create(this), SDP_E_FAIL);
	return SDP_S_OK;
}

sdpResult CFractalTerrain::ObjectRenamed(SDPCSTRING NewName)
{
	m_UserInterface.ObjectRenamed(NewName);
	return k3dTransformObject::ObjectRenamed(NewName);
}

sdpResult CFractalTerrain::ObjectDeleted()
{
	m_UserInterface.ObjectDeleted();
	return k3dTransformObject::ObjectDeleted();
}

// Generate random terrain

// Compute a random terrain with FFT
void CFractalTerrain::Mountain(gulong N, gdouble d)
{
	gulong k, l;
	std::complex<gdouble> Tmp;


	m_Terrain = new fvector[N];
	for(k = 0; k < N; k++)
		m_Terrain[k] = new fftVector(N);

	// Set all values using a normal random variable (RNV),
	// Array[0][0] = 0, Array[0][N/2] = RNV, Array[N/2][0] = RNV,
	// Array[N/2][N/2]; fill up the array a normal random complex
	// variable (CRNV) following the rule : CRNVi = conj(CRNVj)
	// when CRNVi and CRNVj are symmetrically placed around one
	// of the three real values :

	//  0  C  C  C R1  C  C  C  (N = 8)
	//  C  C  C C1  C  C  C  C
	//  C  C  C  C  C  C  C  C
	// C3  C  C  C  C  C  C  C
	// R2  C  C  C R3  C  C  C  C3 = conj(C4), C1 = conj(C2)
	// C4  C  C  C  C  C  C  C
	//  C  C  C  C  C  C  C  C
	//  C  C  C  C  C C2  C  C


	// 0, R1, R2, R3

	m_Terrain[0]->P[0] = 0;
	m_Terrain[0]->P[N/2] = m_TerrainSeeds[0][N/2].normaldrand(0, N/2, d);
	m_Terrain[N/2]->P[0] = m_TerrainSeeds[N/2][0].normaldrand(N/2, 0, d);
	m_Terrain[N/2]->P[N/2] = m_TerrainSeeds[N/2][N/2].normaldrand(N/2, N/2, d);

	// Row[0] and Row[N/2]

	for(k = 1; k < N/2; k++)
	{
		Tmp = m_TerrainSeeds[0][k].cplxnormaldrand(0, k, d);
		m_Terrain[k]->P[0] = Tmp;
		m_Terrain[N-k]->P[0] = std::conj(Tmp);

		Tmp = m_TerrainSeeds[N/2][k].cplxnormaldrand(N/2, k, d);
		m_Terrain[k]->P[N/2] = Tmp;
		m_Terrain[N-k]->P[N/2] = std::conj(Tmp);
	}

	// Column[0] and Column[N/2]

	for(k = 1; k < N/2; k++)
	{
		Tmp = m_TerrainSeeds[k][0].cplxnormaldrand(k, 0, d);
		m_Terrain[0]->P[k] = Tmp;
		m_Terrain[0]->P[N-k] = std::conj(Tmp);

		Tmp = m_TerrainSeeds[k][N/2].cplxnormaldrand(k, N/2, d);
		m_Terrain[N/2]->P[k] = Tmp;
		m_Terrain[N/2]->P[N-k] = std::conj(Tmp);
	}

	// Fill up

	for(k = 1; k < N/2; k++)
		for(l = 1; l < N/2; l++)
		{
			Tmp = m_TerrainSeeds[k][l].cplxnormaldrand(k, l, d);
			// The square inside 0 R1 R2 R3 and symmtrical values
			m_Terrain[k]->P[l] = Tmp;
			m_Terrain[N-k]->P[N-l] = std::conj(Tmp);

			Tmp = m_TerrainSeeds[N-k][l].cplxnormaldrand(k, l, d);
			// Bottom left inside square and symmetrical values
			m_Terrain[N-k]->P[l] = Tmp;
			m_Terrain[k]->P[N-l] = std::conj(Tmp);
		}


	// 2D inverse FFT

	// Apply InvFFT on each row
	for(k = 0; k < N; k++) m_Terrain[k]->InvFFT();

	// Apply InvFFT on each column
	fftVector TmpFFT(N);
	for(l = 0; l < N; l++)
		{
			for(k = 0; k < N; k++) TmpFFT.P[k] = m_Terrain[k]->P[l];
			TmpFFT.InvFFT();
			for(k = 0; k < N; k++) m_Terrain[k]->P[l] = TmpFFT.P[k];
		}
}

void CFractalTerrain::ClearMountain(gulong N)
{
	for(gulong k = 0; k < N; k++)
		delete m_Terrain[k];

	delete m_Terrain;
}


// Use 2D Fast Fourier Transform
void CFractalTerrain::FFTTerrain()
{
	gulong	N = gulong(pow(2, m_Iterations()));

	if(m_CreateGeometry)
		{
			srand(m_RandomSeed());

			m_TerrainSeeds.resize(N);
			for(gulong i = 0; i < N; i++)
				{
					m_TerrainSeeds[i].resize(N);
					for(gulong j = 0; j < N; j++)
						m_TerrainSeeds[i][j].Randomize();
				}

			m_CreateGeometry = false;
		}

	// Build terrain using its fractal dimension and seeds
	Mountain(N, m_FractalDimension());

	// Setup terrain
	gdouble xVoxelSize = m_XLength() / gdouble(N);
	gdouble zVoxelSize = m_ZLength() / gdouble(N);

	ReserveLocations(N*N);
	gulong n = 0;
	for(gulong z = 0; z < N; z++)
		for(gulong x = 0; x < N; x++)
			{
				// Put the terrain in a grid according to the lengths
				gdouble px = (gdouble)x*xVoxelSize - m_XLength()/2;
				gdouble pz = (gdouble)z*zVoxelSize - m_ZLength()/2;

				gdouble py = std::real(m_Terrain[z]->P[x])*m_YLength();
				Location(n++).m_LocalXYZ = sdpVector3(px, py, pz);
			}

	// Render as triangles
	ReserveTriangles((N-1)*(N-1)*2);
	n = 0;
	for(gulong k = 0; k < N-1; k++)
		for(gulong l = 0; l < N-1; l++)
		{
			Triangle(n++).SetPoints(k*N + l, k*N + l+1, (k+1)*N + l+1);
			Triangle(n++).SetPoints(k*N + l, (k+1)*N + l+1, (k+1)*N + l);
		}

	// Delete temporary data
	ClearMountain(N);
}

// Where it all happens ...
void CFractalTerrain::OnModify(bool SaveAction)
{
	FFTTerrain();

	SetSmoothShading();

	// Update the display ...
	k3dTriangleMesh::OnModify(SaveAction);
}

///////////////////////////////////////////////////////////////
// k3dIPersist overrides

sdpResult CFractalTerrain::InitNew()
{
	g_return_val_if_fail(sdpSucceeded(k3dTriangleMesh::InitNew()), SDP_E_FAIL);

	OnModify(false);

	return SDP_S_OK;
}

sdpResult CFractalTerrain::LoadComplete()
{
	g_return_val_if_fail(sdpSucceeded(k3dTriangleMesh::LoadComplete()), SDP_E_FAIL);

	OnModify(false);

	return SDP_S_OK;
}

/////////////////////////////////////////////////////////////////////////////
// CFractalTerrainDialog

CFractalTerrainDialog::CFractalTerrainDialog()
{
	m_Object = 0;
}

bool CFractalTerrainDialog::Create(CFractalTerrain* Object)
{
	// Store our parent object ...
	m_Object = Object;

	// Load our template ...
	g_return_val_if_fail(k3dDialog::Create(Object, false, true, "fractalterrain.gtkml", "fractalterrain.html"), false);

	return true;
}

void CFractalTerrainDialog::UpdateControls()
{
	Entry("xlength").SetText(k3dConvertDoubleString(k3dIChannel::DATA_DISTANCE, m_Object->m_XLength(), 2));
	Entry("ylength").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_YLength(), 2));
	Entry("zlength").SetText(k3dConvertDoubleString(k3dIChannel::DATA_DISTANCE, m_Object->m_ZLength(), 2));
	Entry("iterations").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_Iterations(), 0));
	Entry("fractaldimension").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_FractalDimension(), 2));
	Entry("randomseed").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_RandomSeed(), 0));
}

void CFractalTerrainDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "onxlength")
		OnXLength();
	else if(Event->Name() == "onylength")
		OnYLength();
	else if(Event->Name() == "onzlength")
		OnZLength();
	else if(Event->Name() == "oniterations")
		OnIterations();
	else if(Event->Name() == "onfractaldimension")
		OnFractalDimension();
	else if(Event->Name() == "onrandomseed")
		OnRandomSeed();
	else
		k3dDialog::OnEvent(Event);
}

void CFractalTerrainDialog::OnXLength()
{
	if(UpdateControl(Entry("xlength"), m_Object->m_XLength(), k3dIChannel::DATA_DISTANCE, 2))
		m_Object->OnModify(false);
}

void CFractalTerrainDialog::OnYLength()
{
	if(UpdateControl(Entry("ylength"), m_Object->m_YLength(), k3dIChannel::DATA_DISTANCE, 2))
		m_Object->OnModify(false);
}

void CFractalTerrainDialog::OnZLength()
{
	if(UpdateControl(Entry("zlength"), m_Object->m_ZLength(), k3dIChannel::DATA_DISTANCE, 2))
		m_Object->OnModify(false);
}

void CFractalTerrainDialog::OnIterations()
{
	if(UpdateControl(Entry("iterations"), m_Object->m_Iterations(), k3dIChannel::DATA_SCALAR, 1, ULONG_MAX))
		{
			m_Object->m_CreateGeometry = true;
			m_Object->OnModify(false);
		}
}

void CFractalTerrainDialog::OnFractalDimension()
{
	if(UpdateControl(Entry("fractaldimension"), m_Object->m_FractalDimension(), k3dIChannel::DATA_SCALAR, 2))
		m_Object->OnModify(false);
}

void CFractalTerrainDialog::OnRandomSeed()
{
	if(UpdateControl(Entry("randomseed"), m_Object->m_RandomSeed(), k3dIChannel::DATA_SCALAR))
		{
			m_Object->m_CreateGeometry = true;
			m_Object->OnModify(true);
		}
}



