#ifndef EXTRUDE_H
#define EXTRUDE_H

// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares CExtrude, a K-3D tool for creating ducted solids from existing geometry
		\author Tim Shead (tshead@k-3d.com)
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include <k3dprimtool.h>
#include <k3dprimtooldialog.h>

#include <sdpgl.h>

// Forward class references
class CExtrude;

/////////////////////////////////////////////////////////////////////////////
// CExtrudeDialog

/// Provides the user interface for the Extrude tool
class CExtrudeDialog : public k3dPrimToolDialog
{
friend class CExtrude;
public:
	CExtrudeDialog();

	/// Called to create the UI
	bool Create(CExtrude* Object);

protected:
	bool InitializeControls();
	void UpdateControls();

	void OnEvent(sdpGtkEvent* Event);

	void OnPageSwitch(sdpGtkEvent* Event);

	void OnSegments();
	void OnConnectExterior();
	void OnConnectInterior();
	void OnCreateLayers();
	void OnCreateEndCaps();
	
	void OnOffsetX();
	void OnOffsetY();
	void OnOffsetZ();
	void OnScaleX();
	void OnScaleY();
	void OnScaleZ();
	void OnRotationX();
	void OnRotationY();
	void OnRotationZ();
	
	/// Stores a pointer to the owning Extrude object
	CExtrude* m_Object;
	
	/// Stores the currently active notebook page
	typedef enum
	{
		EXTRUDE,
		OFFSETS,
		SCALING,
		ROTATION
	} NotebookPage;
	
	NotebookPage m_CurrentPage;
};

/////////////////////////////////////////////////////////////////////////////
// Edges ...

class CExtrudeEdge
{
public:
	CExtrudeEdge(k3dIPoint* Point1, k3dIPoint* Point2);

	k3dIPoint* m_Point1;
	k3dIPoint* m_Point2;

	gulong m_ReferenceCount;
	bool m_Flipped;
};

inline bool operator < (const CExtrudeEdge& A, const CExtrudeEdge& B)
{
	if(A.m_Point1 == B.m_Point1)
		return A.m_Point2 < B.m_Point2;
	else
		return A.m_Point1 < B.m_Point1;
}

inline bool operator == (const CExtrudeEdge& A, const CExtrudeEdge& B) { return (A.m_Point1 == B.m_Point1) && (A.m_Point2 == B.m_Point2); }

typedef std::set<CExtrudeEdge> ExtrudeEdgeSet;
typedef ExtrudeEdgeSet::iterator ExtrudeEdgeSetIterator;

typedef std::vector<CExtrudeEdge> ExtrudeEdgeArray;
typedef ExtrudeEdgeArray::iterator ExtrudeEdgeArrayIterator;

/////////////////////////////////////////////////////////////////////////////
// Points ...

typedef std::map<k3dIPoint*, k3dPointArray> ExtrudePointMap;
typedef ExtrudePointMap::iterator ExtrudePointMapIterator;

/////////////////////////////////////////////////////////////////////////////
// Polygons ...

typedef std::map<k3dIPolygon*, k3dPolygonArray> ExtrudePolygonMap;
typedef ExtrudePolygonMap::iterator ExtrudePolygonMapIterator;

/////////////////////////////////////////////////////////////////////////////
// CExtrude

class CExtrude : public k3dPrimTool
{
friend class CExtrudeDialog;
public:
	CExtrude();

	// k3dIUnknown overrides
	sdpResult QueryInterface(const sdpIID& InterfaceID, void** Interface);
	gulong AddRef();
	gulong Release();

	// k3dTool overrides ...
	void OnModify(bool SaveAction);

	//k3dIObject overrides
	sdpResult EditObject();
	sdpResult ObjectDeleted();

	// k3dIUserEventClient overrides
	sdpResult OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY);

protected:
	CExtrudeDialog m_UserInterface;
	
	ExtrudePolygonMap m_Polygons;
	ExtrudeEdgeArray m_ExteriorEdges;
	ExtrudeEdgeArray m_InteriorEdges;
	ExtrudePointMap m_Points;
		
	gulong	m_Segments;

	bool	m_ConnectExterior;
	bool m_ConnectInterior;
	bool	m_CreateLayers;
	bool	m_CreateEndCaps;

	gdouble m_Sensitivity;

	sdpVector3 m_Offset;
	sdpVector3 m_Scale;
	sdpVector3 m_Rotation;

//	KObjectContainer<&sdpChannel, &IID_k3dIChannel, k3dIChannel> m_Xoffset;
//	KObjectContainer<&sdpChannel, &IID_k3dIChannel, k3dIChannel> m_Yoffset;
//	KObjectContainer<&sdpChannel, &IID_k3dIChannel, k3dIChannel> m_Zoffset;
//	KObjectContainer<&sdpChannel, &IID_k3dIChannel, k3dIChannel> m_Xscale;
//	KObjectContainer<&sdpChannel, &IID_k3dIChannel, k3dIChannel> m_Yscale;
//	KObjectContainer<&sdpChannel, &IID_k3dIChannel, k3dIChannel> m_Zscale;
};


#endif // EXTRUDE_H

