// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements CFreehand, a K-3D tool for creating freehand geometry
		\author Timothy M. Shead (tshead@k-3d.com)
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include "freehand.h"

#include <k3dmisc.h>
#include <k3dboundingbox.h>
#include <k3dprimtool.h>
#include <k3dprimtooldialog.h>

#include <k3diactions.h>
#include <k3dicamera.h>
#include <k3dipolypointserver.h>
#include <k3dicamera.h>

#include <sdpgtkevent.h>
#include <sdpfile.h>

#include <iostream>

/////////////////////////////////////////////////////////////////////////////
// CFreehand

CFreehand::CFreehand()
{
}

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CFreehand::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIEditorRender>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIObject>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIUserEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIObject*>(this), InterfaceID, Interface)) return SDP_S_OK;

	return SDP_E_NO_INTERFACE;
}

gulong CFreehand::AddRef()
{
	return sdpComAddRef(m_ReferenceCount);
}

gulong CFreehand::Release()
{
	return sdpComRelease(this, m_ReferenceCount);
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CFreehand::SetObjectContext(k3dIObjectContext* ObjectContext)
{
	if(sdpFailed(k3dPrimTool::SetObjectContext(ObjectContext)))
		return SDP_E_FAIL;

	// Make sure that we get called by the rendering engine ...
	ObjectContext->State()->SetState(k3dIState::VISIBLE, true);

	return SDP_S_OK;
}

// Called to open an object's UI
sdpResult CFreehand::EditObject()
{
	// Create a user interface ...
	g_return_val_if_fail(m_UserInterface.Create(this), SDP_E_FAIL);

	// Update the display ...
	OnModify(false);

	return SDP_S_OK;
}

sdpResult CFreehand::ObjectDeleted()
{
	m_UserInterface.ObjectDeleted();
	return k3dPrimTool::ObjectDeleted();
}

///////////////////////////////////////////////////////////////
// k3dIEditorRender overrides

sdpResult CFreehand::EditorPostRender(k3dIEditorEngine* Engine)
{
	// Sanity checks ...
	g_assert(Engine);

	glPushAttrib(GL_ALL_ATTRIB_BITS);
	glDisable(GL_LIGHTING);
	glDisable(GL_DEPTH_TEST);

	glColor3d(0.0, 0.0, 1.0);
	glPointSize(6.0f);

	// Draw a plane to help the user orient themselves ...
	
	// First, create a bounding box ...
	k3dBoundingBox3 box;
	box.Expand(m_MouseCursor);
	
	for(gulong i = 0; i < m_PolygonPoints.size(); i++)
		box.Expand(m_PolygonPoints[i]);

	// Put a border around it ...
	const gdouble border = 1.0;
	
	sdpVector3 min = box.Min();
	sdpVector3 max = box.Max();
	
	min += sdpVector3(-border, -border, -border);
	max += sdpVector3(border, border, border);

	// Draw it ...			
	glBegin(GL_LINE_LOOP);
	
	switch(Document()->GetActiveAxis())
		{
			case k3dIDocument::XAXIS:
				glVertex3d(0, min[1], max[2]);
				glVertex3d(0, max[1], max[2]);
				glVertex3d(0, max[1], min[2]);
				glVertex3d(0, min[1], min[2]);
				break;
				
			case k3dIDocument::YAXIS:
				glVertex3d(min[0], 0, max[2]);
				glVertex3d(max[0], 0, max[2]);
				glVertex3d(max[0], 0, min[2]);
				glVertex3d(min[0], 0, min[2]);
				break;
				
			case k3dIDocument::ZAXIS:
				glVertex3d(min[0], max[1], 0);
				glVertex3d(max[0], max[1], 0);
				glVertex3d(max[0], min[1], 0);
				glVertex3d(min[0], min[1], 0);
				break;
				
			default:
				g_assert_not_reached();	// Did you add a new axis type without telling me?
		}
	
	glEnd();

	// Draw the vertices selected by the user
	glBegin(GL_POINTS);
	for(gulong j = 0; j < m_PolygonPoints.size(); j++)
		glVertex3dv(m_PolygonPoints[j]);
	glEnd();

	// Draw the current mouse cursor
	glColor3d(1.0, 0.0, 0.0);
	glBegin(GL_POINTS);
		glVertex3dv(m_MouseCursor);
	glEnd();

	glPopAttrib();	// GL_ALL_ATTRIB_BITS

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIUserEventClient overrides

sdpResult CFreehand::OnMouseMove(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY)
{
	m_MouseCursor = ScreenToWorld(Camera, Modifiers, CurrentX, CurrentY);

	OnModify(false);

	return SDP_S_OK;
}

sdpResult CFreehand::OnLButtonClick(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY)
{
	m_MouseCursor = ScreenToWorld(Camera, Modifiers, CurrentX, CurrentY);

	// Save our point
	m_PolygonPoints.push_back(m_MouseCursor);

	// Update the display
	m_CreateGeometry = true;
	OnModify(false);

	return SDP_S_OK;
}

sdpResult CFreehand::OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	m_MouseCursor = ScreenToWorld(Camera, Modifiers, CurrentX, CurrentY);

	// Save our point
	m_PolygonPoints.push_back(m_MouseCursor);

	// Update the display
	m_CreateGeometry = true;
	OnModify(false);

	return SDP_S_OK;
}

sdpResult CFreehand::OnLButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	// Update the display
	OnModify(false);

	return SDP_S_OK;
}

/////////////////////////////////////////////////////////////////////////////
// CFreehand message handlers

void CFreehand::OnModify(bool SaveAction)
{
	// Create new geometry as needed ...
	if(m_CreateGeometry)
		{
			// Get rid of old geometry ...
			DeleteAll();

			// Record the action if requested ...
			if(SaveAction)
				Document()->Actions()->StartAction("Freehand");

			// Create a mesh ...
			k3dIMesh* mesh = CreateMesh("Freehand");
			g_assert(mesh);

			// Finish recording the action if requested ...
			if(SaveAction)
				{
					Document()->Actions()->FinishAction();
					Document()->FreezeChanges();
				}

			// Create a single poly ...
			k3dIPolygon* polygon = CreatePolygon(mesh);
			g_assert(polygon);

			// Create points and attach them to the polygon ...
			for(gulong i = 0; i < m_PolygonPoints.size(); i++)
				{
					k3dIPoint* point = CreatePoint(mesh);
					g_assert(point);
					
					polygon->Points()->AddPoint(point);
				}
				
			m_CreateGeometry = false;
		}

	// Update point locations ...		
	sdpVector3 coords;
	for(gulong i = 0; i < m_PolygonPoints.size(); i++)
		{
			coords = m_PolygonPoints[i];
			m_Points[0][i]->Location()->SetOldXYZ(coords);
			m_Points[0][i]->Location()->SetLocalXYZ(coords);
		}

	// Update statistics in our dialog ...
	DisplayStatistics(m_UserInterface.Entry("stats"));

	// Update the display
	k3dUserObject::OnModify(SaveAction);
}

sdpVector3 CFreehand::ScreenToWorld(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY)
{
	// Sanity checks ...
	g_assert(Camera);

	// Turn mouse coordinates into a vector ...
	sdpVector3 mouseorigin, mousedirection;
	k3dMouseToWorldRay(Camera, sdpVector2(CurrentX, CurrentY), mouseorigin, mousedirection);

	// Intersect the vector with the active plane ...
	sdpVector3 result;

	switch(Document()->GetActiveAxis())
		{
			case k3dIDocument::XAXIS:
				sdpPlaneLineIntersection(sdpPlane(sdpVector3(1, 0, 0), 0), mouseorigin, mousedirection, result);
				break;
			case k3dIDocument::YAXIS:
				sdpPlaneLineIntersection(sdpPlane(sdpVector3(0, 1, 0), 0), mouseorigin, mousedirection, result);
				break;
			case k3dIDocument::ZAXIS:
				sdpPlaneLineIntersection(sdpPlane(sdpVector3(0, 0, 1), 0), mouseorigin, mousedirection, result);
				break;
			default:
				g_assert_not_reached();	// Did you add a new axis type without telling me?
		}

	return result;
}

//////////////////////////////////////////////////////////////////////////////////////
// CFreehandDialog

CFreehandDialog::CFreehandDialog()
{
	m_Object = 0;
}

bool CFreehandDialog::Create(CFreehand* Object)
{
	// Store our parent object ...
	m_Object = Object;

	// Load our template ...
	g_return_val_if_fail(k3dPrimToolDialog::Create(Object, true, "freehand.gtkml", "freehand.html"), false);

	return true;
}

void CFreehandDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	k3dPrimToolDialog::OnEvent(Event);
}


