// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the functions necessary to export K-3D objects from the toollib module
		\author Timothy M. Shead (tshead@k-3d.com)
*/
		
#include <gmodule.h>

#include <strstream>
#include <sdpxml.h>

#include "extrude.h"
#include "freehand.h"
#include "lissacurve.h"
#include "mirror.h"
#include "origin.h"
#include "primitives.h"
#include "random.h"
#include "rotate.h"
#include "scale.h"
#include "slant.h"
#include "smooth.h"
#include "snap.h"
#include "sphereize.h"
#include "taper.h"
#include "translate.h"
#include "triangulate.h"
#include "twist.h"
#include "uvmap.h"
#include "wave.h"

#include <k3dcom.h>
#include <k3dmisc.h>

#include "../../k3d_version.h"

extern "C" {

/// Stores the XML document that describes this module and its exported objects
static std::strstream ClassInfo;

/// Called to create an XML stream object that describes this module and its exported objects
static void BuildClassInfo()
{
	// Generate an XML document ...
	sdpXmlDocument document;
	sdpXmlIElement* object;
	sdpXmlIElement*  library = k3dStartLibraryInfo(&document, "toollib", VERSION_STRING);

	// Extrude ...
	object = k3dAddObject(library, sdpExtrude, "Extrude");
	k3dAddBookmark(object, "Tools/Extrude");
	
	// Freehand ...
	object = k3dAddObject(library, sdpFreehand, "Freehand");
	k3dAddBookmark(object, "Tools/Freehand");
	
	// LissaCurve ...
	object = k3dAddObject(library, sdpLissaCurve, "LissaCurve");
	k3dAddBookmark(object, "Tools/LissaCurve");
	
	// Mirror ...
	object = k3dAddObject(library, sdpMirror, "Mirror");
	k3dAddBookmark(object, "Tools/Mirror");
	
	// Origin ...
	object = k3dAddObject(library, sdpOrigin, "Origin");
	k3dAddBookmark(object, "Tools/Origin");
	
	// Primitives ...
	object = k3dAddObject(library, sdpPrimitives, "Primitives");
	k3dAddBookmark(object, "Tools/Primitives");
	
	// Random ...
	object = k3dAddObject(library, sdpRandom, "Random");
	k3dAddBookmark(object, "Tools/Random");
	
	// Rotate ...
	object = k3dAddObject(library, sdpRotate, "Rotate");
	k3dAddBookmark(object, "Tools/Rotate");
	
	// Scale ...
	object = k3dAddObject(library, sdpScale, "Scale");
	k3dAddBookmark(object, "Tools/Scale");
	
	// Slant ...
	object = k3dAddObject(library, sdpSlant, "Slant");
	k3dAddBookmark(object, "Tools/Slant");
	
	// Snap ...
	object = k3dAddObject(library, sdpSnap, "Snap");
	k3dAddBookmark(object, "Tools/Snap");
	
	// Sphereize ...
	object = k3dAddObject(library, sdpSphereize, "Sphereize");
	k3dAddBookmark(object, "Tools/Sphereize");
	
	// Smooth ...
	object = k3dAddObject(library, sdpSmooth, "Smooth");
	k3dAddBookmark(object, "Tools/Smooth");
	
	// Taper ...
	object = k3dAddObject(library, sdpTaper, "Taper");
	k3dAddBookmark(object, "Tools/Taper");
	
	// Translate ...
	object = k3dAddObject(library, sdpTranslate, "Translate");
	k3dAddBookmark(object, "Tools/Translate");
	
	// Triangulate ...
	object = k3dAddObject(library, sdpTriangulate, "Triangulate");
	k3dAddBookmark(object, "Tools/Triangulate");
	
	// Twist ...
	object = k3dAddObject(library, sdpTwist, "Twist");
	k3dAddBookmark(object, "Tools/Twist");
	
	// UVMap ...
	object = k3dAddObject(library, sdpUVMap, "UVMap");
	k3dAddBookmark(object, "Tools/UVMap");
	
	// Wave ...
	object = k3dAddObject(library, sdpWave, "Wave");
	k3dAddBookmark(object, "Tools/Wave");

	ClassInfo << document << std::ends;
}

/// Called when the module is dynamically loaded
G_MODULE_EXPORT const gchar* g_module_check_init(GModule *module)
{
	// Sanity checks ...
	g_assert(module);

	// Build our XML sdpComClassInfo document ...
	BuildClassInfo();

	return 0;
}

/// Called when the module is about to be unloaded
G_MODULE_EXPORT void g_module_unload(GModule *module)
{
	// Unlock memory on our XML sdpComClassInfo document ...
	ClassInfo.rdbuf()->freeze(false);
}

/// Called to return an XML document describing this module and its exported objects
G_MODULE_EXPORT SDPCSTRING sdpComLibraryInfo()
{
	return ClassInfo.str();
}

/// Called to create an object of the given type, returning an interface of the given type
G_MODULE_EXPORT sdpResult sdpComClassProcedure(const sdpCLSID& ClassID, const sdpIID& InterfaceID, void** Interface)
{
	if(ClassID == sdpExtrude)
		return sdpCreateClassInstance<CExtrude>(InterfaceID, Interface) ? SDP_S_OK : SDP_E_FAIL;
	if(ClassID == sdpFreehand)
		return sdpCreateClassInstance<CFreehand>(InterfaceID, Interface) ? SDP_S_OK : SDP_E_FAIL;
	if(ClassID == sdpLissaCurve)
		return sdpCreateClassInstance<CLissaCurve>(InterfaceID, Interface) ? SDP_S_OK : SDP_E_FAIL;
	if(ClassID == sdpMirror)
		return sdpCreateClassInstance<CMirror>(InterfaceID, Interface) ? SDP_S_OK : SDP_E_FAIL;
	if(ClassID == sdpOrigin)
		return sdpCreateClassInstance<COrigin>(InterfaceID, Interface) ? SDP_S_OK : SDP_E_FAIL;
	if(ClassID == sdpPrimitives)
		return sdpCreateClassInstance<CPrimitives>(InterfaceID, Interface) ? SDP_S_OK : SDP_E_FAIL;
	if(ClassID == sdpRotate)
		return sdpCreateClassInstance<CRotate>(InterfaceID, Interface) ? SDP_S_OK : SDP_E_FAIL;
	if(ClassID == sdpRandom)
		return sdpCreateClassInstance<CRandom>(InterfaceID, Interface) ? SDP_S_OK : SDP_E_FAIL;
	if(ClassID == sdpScale)
		return sdpCreateClassInstance<CScale>(InterfaceID, Interface) ? SDP_S_OK : SDP_E_FAIL;
	if(ClassID == sdpSlant)
		return sdpCreateClassInstance<CSlant>(InterfaceID, Interface) ? SDP_S_OK : SDP_E_FAIL;
	if(ClassID == sdpSmooth)
		return sdpCreateClassInstance<CSmooth>(InterfaceID, Interface) ? SDP_S_OK : SDP_E_FAIL;
	if(ClassID == sdpSnap)
		return sdpCreateClassInstance<CSnap>(InterfaceID, Interface) ? SDP_S_OK : SDP_E_FAIL;
	if(ClassID == sdpSphereize)
		return sdpCreateClassInstance<CSphereize>(InterfaceID, Interface) ? SDP_S_OK : SDP_E_FAIL;
	if(ClassID == sdpTaper)
		return sdpCreateClassInstance<CTaper>(InterfaceID, Interface) ? SDP_S_OK : SDP_E_FAIL;
	if(ClassID == sdpTwist)
		return sdpCreateClassInstance<CTwist>(InterfaceID, Interface) ? SDP_S_OK : SDP_E_FAIL;
	if(ClassID == sdpTranslate)
		return sdpCreateClassInstance<CTranslate>(InterfaceID, Interface) ? SDP_S_OK : SDP_E_FAIL;
	if(ClassID == sdpTriangulate)
		return sdpCreateClassInstance<CTriangulate>(InterfaceID, Interface) ? SDP_S_OK : SDP_E_FAIL;
	if(ClassID == sdpUVMap)
		return sdpCreateClassInstance<CUVMap>(InterfaceID, Interface) ? SDP_S_OK : SDP_E_FAIL;
	if(ClassID == sdpWave)
		return sdpCreateClassInstance<CWave>(InterfaceID, Interface) ? SDP_S_OK : SDP_E_FAIL;

	return SDP_E_UNKNOWN_OBJECT;
}

}	// extern "C"

