// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CRandom K-3D tool, which randomly displaces geometry
		\author Tim Shead (tshead@k-3d.com)
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include "random.h"

#include <sdpfile.h>
#include <k3dmisc.h>

#include <sdpgtkevent.h>
#include <k3diactions.h>

/////////////////////////////////////////////////////////////////////////////
// CRandom

CRandom::CRandom()
{
	m_Amplitude = 0.0;

	m_Distance = 10.0;
	m_Seed = 123457;

	m_RandomType = LINEAR;
	m_DistanceType = INFINITEDISTANCE;
	m_MoveType = POINTS;
}

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CRandom::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIObject>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIUserEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIObject*>(this), InterfaceID, Interface)) return SDP_S_OK;

	return SDP_E_NO_INTERFACE;
}

gulong CRandom::AddRef()
{
	return sdpComAddRef(m_ReferenceCount);
}

gulong CRandom::Release()
{
	return sdpComRelease(this, m_ReferenceCount);
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

// Called to open an object's UI
sdpResult CRandom::EditObject()
{
	// Create a user interface ...
	g_return_val_if_fail(m_UserInterface.Create(this), SDP_E_FAIL);

	// Update the display ...
	OnModify(false);

	return SDP_S_OK;
}

sdpResult CRandom::ObjectDeleted()
{
	m_UserInterface.ObjectDeleted();
	return k3dTool::ObjectDeleted();
}

///////////////////////////////////////////////////////////////
// k3dIUserEventClient overrides

// Called while the user drags the left mouse button
sdpResult CRandom::OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	// Compute scaling factors for each mouse axis ...
	double xfactor = (CurrentX - LastX) * 0.1;
	double yfactor = (CurrentY - LastY) * 0.1;

	if(ShiftDown(Modifiers))
		{
			m_Amplitude[0] -= (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor);
			m_Amplitude[1] -= (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor);
			m_Amplitude[2] -= (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor);
		}
	else if(ControlDown(Modifiers))
		{
			switch(Document()->GetActiveAxis())
				{
					case k3dIDocument::XAXIS:
						m_Amplitude[0] -= (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor);
						break;
					case k3dIDocument::YAXIS:
						m_Amplitude[1] -= (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor);
						break;
					case k3dIDocument::ZAXIS:
						m_Amplitude[2] -= (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor);
						break;
					default:
						g_assert(0);
				}
		}
	else
		{
			switch(Document()->GetActiveAxis())
				{
					case k3dIDocument::XAXIS:
						m_Amplitude[2] -= xfactor;
						m_Amplitude[1] -= yfactor;
						break;
					case k3dIDocument::YAXIS:
						m_Amplitude[0] -= xfactor;
						m_Amplitude[2] -= yfactor;
						break;
					case k3dIDocument::ZAXIS:
						m_Amplitude[0] -= xfactor;
						m_Amplitude[1] -= yfactor;
						break;
					default:
						g_assert(0);
				}
		}


	// Update our dialog ...
	m_UserInterface.UpdateControls();

	// Update the display
	OnModify(false);

	return SDP_S_OK;
}

sdpResult CRandom::OnLButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	OnModify(true);

	return SDP_S_OK;
}

void CRandom::OnModify(bool SaveAction)
{
	// Record the action if requested ...
	if(SaveAction)
		Document()->Actions()->StartAction("Random");

	// Call the default implementation ...
	k3dTool::OnModify(SaveAction);

	// Finish recording the action if requested ...
	if(SaveAction)
		{
			Document()->Actions()->FinishAction();
			Document()->FreezeChanges();
		}
}

void CRandom::ModifyObject(k3dIObject* Object)
{
	// Sanity checks ...
	g_assert(Object);

	// Take a location-by-location approach to modifying the object ...
	ModifyLocationServer(Object);
}

void CRandom::ModifyLocationServer(k3dIObject* Object)
{
	// Sanity checks ...
	g_assert(Object);

	// Setup our seed ...
	m_CurrentSeed = m_Seed;

	// Call the default implementation ...
	k3dTool::ModifyLocationServer(Object);
}

void CRandom::ModifyLocation(k3dILocation* Location)
{
	sdpVector3 coords;
	double distance, newradius;

//	double amplitude;
//	sdpVector3 offset;

	double rangefactor = 1.0;
	sdpVector3 halfamplitude = m_Amplitude * 0.5;

	if(POINTS == m_MoveType)
		{
			Location->GetOldXYZ(coords);
			coords -= m_Origin;

			distance = coords.Length();

			if(FINITEDISTANCE == m_DistanceType)
				{
					rangefactor = distance ? (distance < m_Distance ? 1.0 - (distance / m_Distance) : 0.0) : 0.0;
				}

			if(LINEAR == m_RandomType)
				{
					coords[0] += ((m_Amplitude[0] * DRand(m_CurrentSeed)) - halfamplitude[0]) * rangefactor;
					coords[1] += ((m_Amplitude[1] * DRand(m_CurrentSeed)) - halfamplitude[1]) * rangefactor;
					coords[2] += ((m_Amplitude[2] * DRand(m_CurrentSeed)) - halfamplitude[2]) * rangefactor;
				}
			else
				{
					newradius = distance ? ((distance + (((m_Amplitude[0] * DRand(m_CurrentSeed)) - halfamplitude[0]) * rangefactor)) / distance) : 1.0;
					coords *= newradius;
				}

			coords += m_Origin;
			Location->SetLocalXYZ(coords);
		}
	else if(POLYGONS == m_MoveType)
		{
/*
			for(K3DPOINTHANDLE pointhandle = Mesh->Points()->GetFirstPoint(); pointhandle; pointhandle = Mesh->Points()->GetNextPoint(pointhandle))
				{
					if(sdpSucceeded(m_MeshPolygon->SetFirstPoint()))
						{
							m_MeshPolygonLocation->GetOldXYZ(coords);
							coords -= origin;

							distance = coords.Length();
							if(FINITEDISTANCE == m_DistanceType)
								{
									rangefactor = distance ? (distance < m_Distance ? 1.0 - (distance / m_Distance) : 0.0) : 0.0;
								}

							if(LINEAR == m_RandomType)
								{
									offset[0] = ((m_Amplitude[0] * DRand(m_CurrentSeed)) - halfamplitude[0]) * rangefactor;
									offset[1] = ((m_Amplitude[1] * DRand(m_CurrentSeed)) - halfamplitude[0]) * rangefactor;
									offset[2] = ((m_Amplitude[2] * DRand(m_CurrentSeed)) - halfamplitude[0]) * rangefactor;
								}
							else	// Radial
								{
									amplitude = ((m_Amplitude[0] * DRand(m_CurrentSeed)) - halfamplitude[0]) * rangefactor;
									if(distance)
										{
											offset = amplitude * coords / distance;
										}
									else
										{
											offset = 0.0;
										}
								}

							for(hr = m_MeshPolygon->SetFirstPoint(); SDP_S_OK == hr; hr = m_MeshPolygon->SetNextPoint())
								{
									m_MeshPolygonLocation->GetOldXYZ(coords);
									coords += offset;
									m_MeshPolygonLocation->SetLocalXYZ(coords);
								}
						}
				}
*/
		}
}

//////////////////////////////////////////////////////
// Random number generator

#define RANDMODULUS	2147483647
#define RANDMULTIPLIER	16807
#define RANDINCREMENT	0

double CRandom::DRand(gulong& Seed)
{
	Seed = ((RANDMULTIPLIER * Seed) + RANDINCREMENT) % RANDMODULUS;
	return((double)Seed / (double)RANDMODULUS);
}

////////////////////////////////////////////////////////////////////////////
// CRandomDialog

CRandomDialog::CRandomDialog()
{
	m_Object = 0;
}

bool CRandomDialog::Create(CRandom* Object)
{
	// Store our parent object ...
	m_Object = Object;

	// Load our template ...
	g_return_val_if_fail(k3dDialog::Create(Object, true, false, "random.gtkml", "random.html"), false);

	return true;
}

void CRandomDialog::UpdateControls()
{
	Entry("amplitudex").SetText(k3dConvertDoubleString(k3dIChannel::DATA_DISTANCE, m_Object->m_Amplitude[0], 3));
	Entry("amplitudey").SetText(k3dConvertDoubleString(k3dIChannel::DATA_DISTANCE, m_Object->m_Amplitude[1], 3));
	Entry("amplitudez").SetText(k3dConvertDoubleString(k3dIChannel::DATA_DISTANCE, m_Object->m_Amplitude[2], 3));
	Entry("distance").SetText(k3dConvertDoubleString(k3dIChannel::DATA_DISTANCE, m_Object->m_Distance, 3));
	Entry("randomseed").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_Seed, 0));
	
	switch(m_Object->m_RandomType)
		{
			case CRandom::LINEAR:
				ToggleButton("linear").SetState(true);
				break;
			case CRandom::RADIAL:
				ToggleButton("radial").SetState(true);
				break;
			default:
				g_assert_not_reached();
		}
	
	switch(m_Object->m_DistanceType)
		{
			case CRandom::FINITEDISTANCE:
				ToggleButton("finite").SetState(true);
				break;
			case CRandom::INFINITEDISTANCE:
				ToggleButton("infinite").SetState(true);
				break;
			default:
				g_assert_not_reached();
		}
}

void CRandomDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "onamplitudex")
		OnAmplitudeX();
	else if(Event->Name() == "onamplitudey")
		OnAmplitudeY();
	else if(Event->Name() == "onamplitudez")
		OnAmplitudeZ();
	else if(Event->Name() == "ondistance")
		OnDistance();
	else if(Event->Name() == "onrandomseed")
		OnRandomSeed();
	else if(Event->Name() == "onlinear")
		OnLinear();
	else if(Event->Name() == "onradial")
		OnRadial();
	else if(Event->Name() == "onfinite")
		OnFinite();
	else if(Event->Name() == "oninfinite")
		OnInfinite();
	else if(Event->Name() == "onpoints")
		OnPoints();
	else if(Event->Name() == "onpolygons")
		OnPolygons();
	else
		k3dDialog::OnEvent(Event);
}

void CRandomDialog::OnAmplitudeX()
{
	if(UpdateControl(Entry("amplitudex"), m_Object->m_Amplitude[0], k3dIChannel::DATA_DISTANCE, 3))
		m_Object->OnModify(true);
}

void CRandomDialog::OnAmplitudeY()
{
	if(UpdateControl(Entry("amplitudey"), m_Object->m_Amplitude[1], k3dIChannel::DATA_DISTANCE, 3))
		m_Object->OnModify(true);
}

void CRandomDialog::OnAmplitudeZ()
{
	if(UpdateControl(Entry("amplitudez"), m_Object->m_Amplitude[2], k3dIChannel::DATA_DISTANCE, 3))
		m_Object->OnModify(true);
}

void CRandomDialog::OnDistance()
{
	if(UpdateControl(Entry("distance"), m_Object->m_Distance, k3dIChannel::DATA_DISTANCE, 3))
		m_Object->OnModify(true);
}

void CRandomDialog::OnRandomSeed()
{
	if(UpdateControl(Entry("randomseed"), m_Object->m_Seed, k3dIChannel::DATA_SCALAR))
		m_Object->OnModify(true);
}

void CRandomDialog::OnFinite()
{
	if(ToggleButton("finite").GetState())
		{
			m_Object->m_DistanceType = CRandom::FINITEDISTANCE;
			m_Object->OnModify(false);
		}
}

void CRandomDialog::OnInfinite()
{
	if(ToggleButton("infinite").GetState())
		{
			m_Object->m_DistanceType = CRandom::INFINITEDISTANCE;
			m_Object->OnModify(false);
		}
}

void CRandomDialog::OnLinear()
{
	if(ToggleButton("linear").GetState())
		{
			m_Object->m_RandomType = CRandom::LINEAR;
			m_Object->OnModify(false);
		}
}

void CRandomDialog::OnRadial()
{
	if(ToggleButton("radial").GetState())
		{
			m_Object->m_RandomType = CRandom::RADIAL;
			m_Object->OnModify(false);
		}
}

void CRandomDialog::OnPoints()
{
/*
	if(ToggleButton("points").GetState())
		{
			m_Object->m_MoveType = CRandom::POINTS;
			m_Object->OnModify(false);
		}
*/
}

void CRandomDialog::OnPolygons()
{
/*
	if(ToggleButton("polygons").GetState())
		{
			m_Object->m_MoveType = CRandom::POLYGONS;
			m_Object->OnModify(false);
		}
*/
}

