// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CSphereize K-3D tool, which "sphereizes" objects by making them conform to the surface of a sphere
		\author Timothy M. Shead (tshead@k-3d.com)
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include "sphereize.h"

#include <sdpfile.h>
#include <k3dmisc.h>

#include <sdpgtkevent.h>
#include <k3diactions.h>

/////////////////////////////////////////////////////////////////////////////
// CSphereize

CSphereize::CSphereize()
{
	m_SphereizeFactor = 0.0;
}

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CSphereize::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIObject>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIUserEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIObject*>(this), InterfaceID, Interface)) return SDP_S_OK;

	return SDP_E_NO_INTERFACE;
}


gulong CSphereize::AddRef()
{
	return sdpComAddRef(m_ReferenceCount);
}


gulong CSphereize::Release()
{
	return sdpComRelease(this, m_ReferenceCount);
}



///////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CSphereize::EditObject()
{
	// Create a user interface ...
	g_return_val_if_fail(m_UserInterface.Create(this), SDP_E_FAIL);

	// Update the display ...
	OnModify(false);

	return SDP_S_OK;
}

sdpResult CSphereize::ObjectDeleted()
{
	m_UserInterface.ObjectDeleted();
	return k3dTool::ObjectDeleted();
}

///////////////////////////////////////////////////////////////
// k3dIUserEventClient overrides

// Called while the user drags the left mouse button
sdpResult CSphereize::OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	// Compute scaling factors for each mouse axis ...
	double xfactor = (CurrentX - LastX) * 0.01;
	double yfactor = (CurrentY - LastY) * 0.01;

	m_SphereizeFactor += fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor;

	// Update our dialog ...
	m_UserInterface.UpdateControls();

	// Update the display
	OnModify(false);

	return SDP_S_OK;
}


sdpResult CSphereize::OnLButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	OnModify(true);

	return SDP_S_OK;
}


/*
// Called when the user rolls the mouse wheel on an Intellimouse-compatible mouse
sdpResult CSphereize::OnViewMouseWheel(long Flags, long Delta, long X, long Y)
{
}
*/


void CSphereize::OnModify(bool SaveAction)
{
	// Record the action if requested ...
	if(SaveAction)
		Document()->Actions()->StartAction("Sphereize");

	// Call the default implementation ...
	k3dTool::OnModify(SaveAction);

	// Finish recording the action if requested ...
	if(SaveAction)
		{
			Document()->Actions()->FinishAction();
			Document()->FreezeChanges();
		}
}


void CSphereize::ModifyObject(k3dIObject* Object)
{
	// Sanity checks ...
	g_assert(Object);

	// Go to the default, location-by-location approach ...
	ModifyLocationServer(Object);
}


void CSphereize::ModifyLocationServer(k3dIObject* Object)
{
	// Sanity checks ...
	g_assert(Object);

	// Call the default implementation ...
	k3dTool::ModifyLocationServer(Object);
}


void CSphereize::ModifyLocation(k3dILocation* Location)
{
	sdpVector3 coords, coords2;
	double targetradius, radius;

	targetradius = m_OriginSize.Length();

	Location->GetOldXYZ(coords);
//	coords -= m_Origin;

	radius = coords.Length();
	if(radius)
		{
			coords2 = coords;
//			coords += coords2.Normalize() * (m_SphereizeFactor * (targetradius - radius)) + m_Origin;
			coords += coords2.Normalize() * (m_SphereizeFactor * (targetradius - radius));

			Location->SetLocalXYZ(coords);
		}
}


//////////////////////////////////////////////////////////////////////////////////////
// CSphereizeDialog

CSphereizeDialog::CSphereizeDialog()
{
	m_Object = 0;
}

bool CSphereizeDialog::Create(CSphereize* Object)
{
	// Store our parent object ...
	m_Object = Object;

	// Load our template ...
	g_return_val_if_fail(k3dDialog::Create(Object, true, false, "sphereize.gtkml", "sphereize.html"), false);

	return true;
}

void CSphereizeDialog::UpdateControls()
{
	Entry("spherefactor").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_SphereizeFactor, 2));
}

void CSphereizeDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "onspherefactor")
		OnSphereFactor();
	else
		k3dDialog::OnEvent(Event);
}


void CSphereizeDialog::OnSphereFactor()
{
	if(UpdateControl(Entry("spherefactor"), m_Object->m_SphereizeFactor, k3dIChannel::DATA_SCALAR, 3))
		m_Object->OnModify(true);
}

