// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements CTriangulate, a K-3D tool for tessellating polygons
		\author Tim Shead (tshead@k-3d.com)
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include "triangulate.h"

#include <k3dmisc.h>
#include <k3dprimtool.h>
#include <k3dprimtooldialog.h>
#include <k3diactions.h>
#include <k3dipolypointserver.h>

#include <sdpgtkevent.h>

#include <sdpmath.h>
#include <sdpfile.h>

#include <iostream>

/// A negative-aware modulo
gint sdpModulo(gint Argument, gint Modulo)
{
	if(0 == Modulo)
		return 0;
		
	while(Argument < 0)
		Argument += Modulo;
		
	while(Argument >= Modulo)
		Argument -= Modulo;
		
	return Argument;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
// k3dTriangulatePoint2

class k3dTriangulatePoint2
{
public:
	k3dTriangulatePoint2(const sdpVector2 Coords, void* UserData) : m_Coords(Coords), m_UserData(UserData) {}
	
	const sdpVector2& Coords() const { return m_Coords; }
	void* UserData() const { return m_UserData; }

	gdouble operator[](int Index) const { return m_Coords[Index]; }

protected:
	sdpVector2 m_Coords;
	void* m_UserData;
};

//////////////////////////////////////////////////////////////////////////////////////////////////////
// k3dTriangulatePoint2Array

typedef std::vector<k3dTriangulatePoint2> k3dTriangulatePoint2ArrayBase;

class k3dTriangulatePoint2Array : public k3dTriangulatePoint2ArrayBase
{
public:
	const k3dTriangulatePoint2& operator[](const gint Index) const { return k3dTriangulatePoint2ArrayBase::operator[](sdpModulo(Index, size())); }
};

// Calculate the area of a polygon ...
gdouble Area(const k3dTriangulatePoint2Array& Points)
{
	// Simplest case ... zero area
	if(Points.size() < 3)
		return 0.0;

	// Calculate the result by adding up determinants ...
	gdouble result = 0;

	gint pointsize = Points.size();
	for(gint i = 0; i < pointsize; i++)
		result += (Points[i][0] * Points[i+1][1]) - (Points[i+1][0] * Points[i][1]);
	
	return result * 0.5;
}

// Figure out whether an ear diagonal intersects any of the remaining polygon line segments ...
bool EarDiagonalIntersects(const k3dTriangulatePoint2Array& Points, const gint Index)
{
	gint pointsize = Points.size();
	for(gint i = 0; i < pointsize; i++)
		{
			// Don't test the ear- or adjacent segments ...
			gulong a = i;
			gulong b = sdpModulo(i+1, pointsize);
			gulong c = sdpModulo(Index-1, pointsize);
			gulong d = sdpModulo(Index+1, pointsize);
			
			if(a == c || a == d || b == c || b == d)
				continue;			

			// See if the segment intersects the ear diagonal (and if so, how) ...
			sdpLineSegmentIntersectionType intersection = sdpLineSegmentIntersection(Points[i].Coords(), Points[i+1].Coords(), Points[Index-1].Coords(), Points[Index+1].Coords());

			switch(intersection)
				{
					case SDP_INTERIORINTERSECTION:
					case SDP_INTERIORCOLINEARPOINTS:
					case SDP_INTERIOREXTERIORCOLINEARPOINTS:
					case SDP_EXTERIORCOLINEARPOINTS:
						return true;
					case SDP_NOINTERSECTION:
					case SDP_EXTERIORCOLINEARPOINT:
					case SDP_INTERIORCOLINEARPOINT:
					case SDP_EXTERIORINTERSECTION:
					case SDP_EXTERIORCOLINEARPOINTSNOINTERSECTION:
						break;
					case SDP_ERROR:
						g_return_val_if_fail(0, true);
						break;
				}
		}

	return false;	
}

// "Clip" an ear from the polygon ...
void ClipEar(k3dTriangulatePoint2Array& Points, const gint Index, k3dTriangulatePoint2Array& Results)
{
	Results.push_back(Points[Index-1]);
	Results.push_back(Points[Index]);
	Results.push_back(Points[Index+1]);
	
	Points.erase(Points.begin() + Index);
}

/// Apply an ear-clipping algorithm to the polygon contour ...
bool EarClipping(k3dTriangulatePoint2Array& Points, k3dTriangulatePoint2Array& Results)
{
	// Calculate the polygon's signed area ...
	gdouble area = Area(Points);
	// We use the sign of the area to determine what's "convex" ...
	const gdouble polygonconvex = sdpSign(Area(Points));
	// We store the polygon's unsigned area to test against later ...
	gdouble polygonarea = fabs(area);

	while(Points.size() > 3)
		{
			gint i;
			gint pointsize = Points.size();
			for(i = 0; i < pointsize; i++)
				{
					// Calculate the ear's signed area ...
					area = sdpArea(Points[i-1].Coords(), Points[i].Coords(), Points[i+1].Coords());
					// We use the sign of the ear's area to test the ear for convexity ...
					const gdouble earconvex = sdpSign(area);
					// We use the ear's unsigned area to test against the polygon's area ...
					const gdouble eararea = fabs(area);

					// If the "ear" isn't convex, it can't be used ...
					if(earconvex != polygonconvex)
						continue;

					// If the "ear" area is larger than the total polygon area, it obviously can't be used ...
					if(eararea > polygonarea)
						continue;

					// If the "ear" diagonal intersects any other line segments in the poly, it can't be used ...
					if(EarDiagonalIntersects(Points, i))
						continue;

					// Clip the ear!
					ClipEar(Points, i, Results);
					
					// Keep track of decreasing polygon area ...
					polygonarea -= eararea;

					break;
				}
				
			// If the loop actually ended and we get this far, we couldn't find an ear to clip!
			g_return_val_if_fail(i < pointsize, false);
		}
	ClipEar(Points, 1, Results);

	return true;
}

/// Calculate a triangulation of 2D points ...
bool k3dTriangulate(k3dTriangulatePoint2Array& Points, k3dTriangulatePoint2Array& Results)
{
	if(Points.size() == 3)
		{
			Results = Points;
		}
	else if(Points.size() > 3)
		{
			EarClipping(Points, Results);
		}

	return true;
}

bool k3dTriangulate(const k3dPolyPointArray& PolyPoints, k3dPolyPointArray& Results)
{
	// See which axis to drop ...
	int dropaxis;
	sdpVector3 normal = k3dGetNormal(PolyPoints).Abs();
	
	if(normal[0] >= normal[1] && normal[0] >= normal[2])
		dropaxis = 0;
	else if(normal[1] >= normal[0] && normal[1] >= normal[2])
		dropaxis = 1;
	else
		dropaxis = 2;

	sdpVector3 coords;
	k3dTriangulatePoint2Array points;
	for(gulong i = 0; i < PolyPoints.size(); i++)
		{
			PolyPoints[i]->Location()->GetLocalXYZ(coords);
			points.push_back(k3dTriangulatePoint2(sdpVector2(coords, dropaxis), PolyPoints[i]));
		}
	
	k3dTriangulatePoint2Array results;
	if(!k3dTriangulate(points, results))
		return false;
	
	for(gulong j = 0; j < results.size(); j++)
		Results.push_back(reinterpret_cast<k3dIPolyPoint*>(results[j].UserData()));
			
	return true;
}

bool k3dTriangulate(k3dIPolygon* Polygon, k3dPolyPointArray& Results)
{
	// Sanity checks ...
	g_assert(Polygon);

	k3dPolyPointArray polypoints;
	k3dGetPolyPoints(Polygon, polypoints);
	
	return k3dTriangulate(polypoints, Results);
}

/////////////////////////////////////////////////////////////////////////////
// CTriangulate

CTriangulate::CTriangulate()
{
}

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CTriangulate::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIObject>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIUserEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIObject*>(this), InterfaceID, Interface)) return SDP_S_OK;

	return SDP_E_NO_INTERFACE;
}

gulong CTriangulate::AddRef()
{
	return sdpComAddRef(m_ReferenceCount);
}

gulong CTriangulate::Release()
{
	return sdpComRelease(this, m_ReferenceCount);
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CTriangulate::EditObject()
{
	// Create a user interface ...
	g_return_val_if_fail(m_UserInterface.Create(this), SDP_E_FAIL);

	// Update the display ...
	OnModify(false);

	return SDP_S_OK;
}

sdpResult CTriangulate::ObjectDeleted()
{
	m_UserInterface.ObjectDeleted();
	return k3dPrimTool::ObjectDeleted();
}

void CTriangulate::OnModify(bool SaveAction)
{
	if(SaveAction)
		Document()->Actions()->StartAction("Triangulate");
		
	// Create the new triangles ...
	k3dPolygonArray polygons;
	k3dGetPolygons(Document(), k3dIState::VISIBLESELECTED, polygons);
	for(gulong i = 0; i < polygons.size(); i++)
		{
			k3dPolyPointArray results;
			k3dTriangulate(polygons[i], results);
			
			k3dIMesh* mesh = polygons[i]->Mesh();
			g_assert(mesh);
			
			// For each result vertex ...
			k3dIPolygon* triangle = 0;
			for(gulong j = 0; j < results.size(); j++)
				{
					// Allocate a new polygon every third vertex ...
					if(0 == j % 3)
						{
							triangle = CreatePolygon(mesh);
							g_assert(triangle);
						}
						
					k3dIPolyPoint* polypoint = results[j];
					g_assert(polypoint);
					
					// Add the vertex to the current triangle ...
					k3dIPolyPoint* newpolypoint = triangle->Points()->AddPoint(polypoint->Point());
					g_assert(newpolypoint);

					// Copy the texture coordinates ...
					sdpVector3 texture;
					polypoint->Location()->GetTextureXYZ(texture);
					newpolypoint->Location()->SetTextureXYZ(texture);
					
					sdpVector3 implicit;
					polypoint->Location()->GetImplicitTextureXYZ(implicit);
					newpolypoint->Location()->SetImplicitTextureXYZ(implicit);
				}
		}
		
	// If we're doing it for real, get rid of the originals ...
	if(SaveAction)
		for(gulong j = 0; j < polygons.size(); j++)
			polygons[j]->Mesh()->DeletePolygon(polygons[j], false);

	// Update statistics in our dialog ...
	DisplayStatistics(m_UserInterface.Entry("stats"));

	// Update the display
	k3dUserObject::OnModify(SaveAction);
	
	if(SaveAction)
		Document()->Actions()->FinishAction();
}

////////////////////////////////////////////////////////////////////////////
// CTriangulateDialog

CTriangulateDialog::CTriangulateDialog()
{
	m_Object = 0;
}

bool CTriangulateDialog::Create(CTriangulate* Object)
{
	// Store our parent object ...
	m_Object = Object;

	// Load our template ...
	g_return_val_if_fail(k3dPrimToolDialog::Create(Object, true, "triangulate.gtkml", "triangulate.html"), false);

	return true;
}


