// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CTwist tool, which applies a "twist" transformation to selected geometry
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include "twist.h"

#include <sdpfile.h>
#include <k3dmisc.h>

#include <sdpgtkevent.h>
#include <k3diactions.h>

/////////////////////////////////////////////////////////////////////////////
// CTwist

CTwist::CTwist()
{
	m_Twist = 0.0;
}

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CTwist::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIObject>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIUserEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIObject*>(this), InterfaceID, Interface)) return SDP_S_OK;

	return SDP_E_NO_INTERFACE;
}


gulong CTwist::AddRef()
{
	return sdpComAddRef(m_ReferenceCount);
}


gulong CTwist::Release()
{
	return sdpComRelease(this, m_ReferenceCount);
}



///////////////////////////////////////////////////////////////
// k3dIObject overrides

// Called to open an object's UI
sdpResult CTwist::EditObject()
{
	// Create a user interface ...
	g_return_val_if_fail(m_UserInterface.Create(this), SDP_E_FAIL);

	// Update the display ...
	OnModify(false);

	return SDP_S_OK;
}

sdpResult CTwist::ObjectDeleted()
{
	m_UserInterface.ObjectDeleted();
	return k3dTool::ObjectDeleted();
}

///////////////////////////////////////////////////////////////
// k3dIUserEventClient overrides

// Called while the user drags the left mouse button
sdpResult CTwist::OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	// Compute scaling factors for each mouse axis ...
	double xfactor = sdpRadians(gdouble(CurrentX - LastX));
	double yfactor = sdpRadians(gdouble(CurrentY - LastY));

	// If the user holds down either control key, only translate along the active axis ...
	if(ControlDown(Modifiers))
		{
			switch(Document()->GetActiveAxis())
				{
					case k3dIDocument::XAXIS:
						m_Twist[0] -= (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor);
						break;
					case k3dIDocument::YAXIS:
						m_Twist[1] -= (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor);
						break;
					case k3dIDocument::ZAXIS:
						m_Twist[2] -= (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor);
						break;
					default:
						g_assert(0);
				}
		}
	// Otherwise, translate along the two non-active axes ...
	else
		{
			switch(Document()->GetActiveAxis())
				{
					case k3dIDocument::XAXIS:
						m_Twist[2] -= xfactor;
						m_Twist[1] -= yfactor;
						break;
					case k3dIDocument::YAXIS:
						m_Twist[0] -= xfactor;
						m_Twist[2] -= yfactor;
						break;
					case k3dIDocument::ZAXIS:
						m_Twist[0] += xfactor;
						m_Twist[1] -= yfactor;
						break;
					default:
						g_assert(0);
				}
		}

	// Update our dialog ...
	m_UserInterface.UpdateControls();

	// Update the display
	OnModify(false);

	return SDP_S_OK;
}


sdpResult CTwist::OnLButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	OnModify(true);

	return SDP_S_OK;
}


/*
// Called when the user rolls the mouse wheel on an Intellimouse-compatible mouse
sdpResult CTwist::OnViewMouseWheel(long Flags, long Delta, long X, long Y)
{
	double factor;

	k3dTool::OnViewMouseWheel(Flags, Delta, X, Y);

	factor = (double)Delta / (double)WHEEL_DELTA * 0.05;

	// Twist along the active axis only
	switch(Document()->GetActiveAxis())
		{
			case k3dIDocument::XAXIS:
				m_Twist[0] += factor;
				break;
			case k3dIDocument::YAXIS:
				m_Twist[1] += factor;
				break;
			case k3dIDocument::ZAXIS:
				m_Twist[2] += factor;
				break;
			default:
				g_assert(0);
		}


	// Update the scaling factor edit controls
	k3dSetDouble(k3dIChannel::DATA_SCALAR, m_Sheet->m_TwistPage->GetDlgItem(IDC_SCALEX)->m_hWnd, m_Twist[0]);
	k3dSetDouble(k3dIChannel::DATA_SCALAR, m_Sheet->m_TwistPage->GetDlgItem(IDC_SCALEY)->m_hWnd, m_Twist[1]);
	k3dSetDouble(k3dIChannel::DATA_SCALAR, m_Sheet->m_TwistPage->GetDlgItem(IDC_SCALEZ)->m_hWnd, m_Twist[2]);

	// Update the display
	OnModify(false);

	return SDP_S_OK;
}
*/


void CTwist::OnModify(bool SaveAction)
{
	// Record the action if requested ...
	if(SaveAction)
		Document()->Actions()->StartAction("Twist");

	// Call the default implementation ...
	k3dTool::OnModify(SaveAction);

	// Finish recording the action if requested ...
	if(SaveAction)
		{
			Document()->Actions()->FinishAction();
			Document()->FreezeChanges();
		}
}


void CTwist::ModifyObject(k3dIObject* Object)
{
	// Sanity checks ...
	g_assert(Object);

	// Go to the default, location-by-location approach ...
	ModifyLocationServer(Object);
}


void CTwist::ModifyLocationServer(k3dIObject* Object)
{
	// Sanity checks ...
	g_assert(Object);

	// Call the default implementation ...
	k3dTool::ModifyLocationServer(Object);
}


void CTwist::ModifyLocation(k3dILocation* Location)
{
	double cosx, cosy, cosz;
	double sinx, siny, sinz;
	double delta, px, py;

	sdpVector3 coords;
	Location->GetOldXYZ(coords);
//	coords -= m_Origin;

	// Twist around X-Axis
	if(m_Twist[0] && m_Size[0])
		{
			delta = coords[0] / m_Size[0];
//			delta = sdpSign<double>(delta) * m_Xfalloff->GetValue1(fabs(delta), .0001);

			cosx = cos(m_Twist[0] * delta);
			sinx = sin(m_Twist[0] * delta);

			py = coords[1];
			coords[1] = (cosx * coords[1]) - (sinx * coords[2]);
			coords[2] = (sinx * py) + (cosx * coords[2]);
		}

	// Twist around Y-Axis
	if(m_Twist[1] && m_Size[1])
		{
			delta = coords[1] / m_Size[1];
//			delta = sdpSign<double>(delta) * m_Yfalloff->GetValue1(fabs(delta), .0001);

			cosy = cos(m_Twist[1] * delta);
			siny = sin(m_Twist[1] * delta);

			px = coords[0];
			coords[0] = (cosy * coords[0]) + (siny * coords[2]);
			coords[2] = (cosy * coords[2]) - (siny * px);
		}

	// Twist around Z-Axis
	if(m_Twist[2] && m_Size[2])
		{
			delta = coords[2] / m_Size[2];
//			delta = sdpSign<double>(delta) * m_Zfalloff->GetValue1(fabs(delta), .0001);

			cosz = cos(m_Twist[2] * delta);
			sinz = sin(m_Twist[2] * delta);

			px = coords[0];
			coords[0] = (cosz * coords[0]) - (sinz * coords[1]);
			coords[1] = (sinz * px) + (cosz * coords[1]);
		}

//	coords += m_Origin;
	Location->SetLocalXYZ(coords);
}


//////////////////////////////////////////////////////////////////////////////////////
// CTwistDialog

CTwistDialog::CTwistDialog()
{
	m_Object = 0;
}

bool CTwistDialog::Create(CTwist* Object)
{
	// Store our parent object ...
	m_Object = Object;

	// Load our template ...
	g_return_val_if_fail(k3dDialog::Create(Object, true, false, "twist.gtkml", "twist.html"), false);

	return true;
}

void CTwistDialog::UpdateControls()
{
	Entry("x").SetText(k3dConvertDoubleString(k3dIChannel::DATA_ANGLE, m_Object->m_Twist[0], 3));
	Entry("y").SetText(k3dConvertDoubleString(k3dIChannel::DATA_ANGLE, m_Object->m_Twist[1], 3));
	Entry("z").SetText(k3dConvertDoubleString(k3dIChannel::DATA_ANGLE, m_Object->m_Twist[2], 3));
}

void CTwistDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "onx")
		OnX();
	else if(Event->Name() == "ony")
		OnY();
	else if(Event->Name() == "onz")
		OnZ();
	else
		k3dDialog::OnEvent(Event);
}

void CTwistDialog::OnX()
{
	if(UpdateControl(Entry("x"), m_Object->m_Twist[0], k3dIChannel::DATA_ANGLE, 3))
		m_Object->OnModify(true);
}

void CTwistDialog::OnY()
{
	if(UpdateControl(Entry("y"), m_Object->m_Twist[1], k3dIChannel::DATA_ANGLE, 3))
		m_Object->OnModify(true);
}

void CTwistDialog::OnZ()
{
	if(UpdateControl(Entry("z"), m_Object->m_Twist[2], k3dIChannel::DATA_ANGLE, 3))
		m_Object->OnModify(true);
};

