// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CUVMap K-3D tool, which is used to modify the texture-mapping coordinates of geometry
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include "uvmap.h"

#include <k3diactions.h>
#include <k3dimesh.h>
#include <k3dipolygon.h>
#include <k3dipolypointserver.h>
#include <k3dipolypoint.h>
#include <k3dmisc.h>

#include <sdpgtkevent.h>

#include <sdpfile.h>

/////////////////////////////////////////////////////////////////////////////
// CUVMap

CUVMap::CUVMap()
{
	m_Type = PROJECTION;
}

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CUVMap::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIObject>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIUserEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIObject*>(this), InterfaceID, Interface)) return SDP_S_OK;

	return SDP_E_NO_INTERFACE;
}

gulong CUVMap::AddRef()
{
	return sdpComAddRef(m_ReferenceCount);
}

gulong CUVMap::Release()
{
	return sdpComRelease(this, m_ReferenceCount);
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

// Called to open an object's UI
sdpResult CUVMap::EditObject()
{
	// Create a user interface ...
	g_return_val_if_fail(m_UserInterface.Create(this), SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CUVMap::ObjectDeleted()
{
	m_UserInterface.ObjectDeleted();
	return k3dTool::ObjectDeleted();
}

class CUVMapProjectionTexture
{
public:
	CUVMapProjectionTexture(k3dIDocument* Document) { m_Axis = Document->GetActiveAxis(); }

	bool HandleMesh(k3dIMesh* Mesh)
	{
		Mesh->GetDimensions(k3dIState::VISIBLESELECTED, m_Min, m_Max);
		m_Size = m_Max - m_Min;
		
		return true;
	}

	bool HandleMeshPolygon(k3dIMesh* Mesh, k3dIPolygon* Polygon)
	{
		sdpVector3 coords, texturecoords;
		k3dIPolyPointServer* polypoints = Polygon->Points();
		for(K3DPOLYPOINTHANDLE polypointhandle = polypoints->GetFirstPolyPoint(); polypointhandle; polypointhandle = polypoints->GetNextPolyPoint(polypointhandle))
			{
				k3dIPolyPoint* point = polypoints->GetPolyPoint(polypointhandle);

				if(SDP_S_OK != point->State()->QueryState(k3dIState::VISIBLESELECTED)) continue;

				point->Location()->GetOldXYZ(coords);

				switch(m_Axis)
					{
						case k3dIDocument::XAXIS:
							texturecoords[0] = ((m_Max[2] - coords[2]) / m_Size[2]);
							texturecoords[1] = ((m_Max[1] - coords[1]) / m_Size[1]);
							texturecoords[2] = ((m_Max[0] - coords[0]) / m_Size[0]);
							break;

						case k3dIDocument::YAXIS:
							texturecoords[0] = ((coords[0] - m_Min[0]) / m_Size[0]);
							texturecoords[1] = ((m_Max[2] - coords[2]) / m_Size[2]);
							texturecoords[2] = ((m_Max[1] - coords[1]) / m_Size[1]);
							break;

						case k3dIDocument::ZAXIS:
							texturecoords[0] = ((coords[0] - m_Min[0]) / m_Size[0]);
							texturecoords[1] = ((m_Max[1] - coords[1]) / m_Size[1]);
							texturecoords[2] = ((m_Max[2] - coords[2]) / m_Size[2]);
							break;

						default:
							g_assert_not_reached();	// unknown axis!
					}

				point->Location()->SetTextureXYZ(texturecoords);
			}
		
		return true;
	}
	
protected:
	sdpVector3 m_Min;
	sdpVector3 m_Max;
	sdpVector3 m_Size;

	k3dIDocument::Axis m_Axis;
};

#define ONE_OVER_PI						.318309886184
#define ONE_OVER_TWOPI				.159154943092

#define PI										3.14159265359
#define TWOPI									6.28318530718

class CUVMapCylindricalTexture
{
public:
	CUVMapCylindricalTexture(k3dIDocument* Document) { m_Axis = Document->GetActiveAxis(); }

	bool HandleMesh(k3dIMesh* Mesh)
	{
		Mesh->GetDimensions(k3dIState::VISIBLESELECTED, m_Min, m_Max);
		m_Size = m_Max - m_Min;
		m_Center = (m_Min + m_Max) * 0.5;
		
		return true;
	}

	bool HandleMeshPolygon(k3dIMesh* Mesh, k3dIPolygon* Polygon)
	{
		sdpVector3 coords, texturecoords;
		double theta;

		// Record polygon quadrants
		gulong quadrants = 0;

		K3DPOLYPOINTHANDLE point;
		k3dIPolyPointServer* points = Polygon->Points();
		for(point = points->GetFirstPolyPoint(); point; point = points->GetNextPolyPoint(point))
			{
				points->GetPolyPoint(point)->Location()->GetOldXYZ(coords);
				coords -= m_Center;

				switch(m_Axis)
					{
						case k3dIDocument::XAXIS:
							quadrants |= ((coords[2] >= 0) ? (coords[1] >= 0 ? 0x02 : 0x01) : (coords[1] >= 0 ? 0x04 : 0x08));
							break;

						case k3dIDocument::YAXIS:
							quadrants |= ((coords[0] <= 0) ? (coords[2] <= 0 ? 0x02 : 0x01) : (coords[2] <= 0 ? 0x04 : 0x08));
							break;

						case k3dIDocument::ZAXIS:
							quadrants |= ((coords[0] <= 0) ? (coords[1] >= 0 ? 0x02 : 0x01) : (coords[1] >= 0 ? 0x04 : 0x08));
							break;
					}
			}

		// Calculate texture coordinates
		for(point = points->GetFirstPolyPoint(); point; point = points->GetNextPolyPoint(point))
			{
				points->GetPolyPoint(point)->Location()->GetOldXYZ(coords);
				coords -= m_Center;

				switch(m_Axis)
					{
						case k3dIDocument::XAXIS:
    					theta = atan2(coords[2], -coords[1]);
    					theta += TWOPI * (theta < 0.0);

    					texturecoords[0] = (theta * ONE_OVER_TWOPI);
    					texturecoords[1] = ((m_Max[0] - coords[0]) / m_Size[0]);
							texturecoords[2] = 0.0;

							if((theta <= PI) && ((quadrants & 0x09) == 0x09))
								{
									texturecoords[0] += 1.0;
								}

							break;

						case k3dIDocument::YAXIS:
    					theta = atan2(-coords[0], coords[2]);
    					theta += TWOPI * (theta < 0.0);

    					texturecoords[0] = (theta * ONE_OVER_TWOPI);
    					texturecoords[1] = ((m_Max[1] - coords[1]) / m_Size[1]);
							texturecoords[2] = 0.0;

							if((theta <= PI) && ((quadrants & 0x09) == 0x09))
								{
									texturecoords[0] += 1.0;
								}

							break;

						case k3dIDocument::ZAXIS:
    					theta = atan2(-coords[0], -coords[1]);
    					theta += TWOPI * (theta < 0.0);

    					texturecoords[0] = (theta * ONE_OVER_TWOPI);
    					texturecoords[1] = ((m_Max[2] - coords[2]) / m_Size[2]);
							texturecoords[2] = 0.0;

							if((theta <= PI) && ((quadrants & 0x09) == 0x09))
								{
									texturecoords[0] += 1.0;
								}

							break;
					}
				points->GetPolyPoint(point)->Location()->SetTextureXYZ(texturecoords);
			}
		
		return true;
	}
	
protected:
	sdpVector3 m_Min;
	sdpVector3 m_Max;
	sdpVector3 m_Size;
	sdpVector3 m_Center;

	k3dIDocument::Axis m_Axis;
};

class CUVMapSphericalTexture
{
public:
	CUVMapSphericalTexture(k3dIDocument* Document) { m_Axis = Document->GetActiveAxis(); }

	bool HandleMesh(k3dIMesh* Mesh)
	{
		Mesh->GetDimensions(k3dIState::VISIBLESELECTED, m_Min, m_Max);
		m_Size = m_Max - m_Min;
		m_Center = (m_Min + m_Max) * 0.5;
		
		return true;
	}

	bool HandleMeshPolygon(k3dIMesh* Mesh, k3dIPolygon* Polygon)
	{
		sdpVector3 coords, texturecoords;
		double theta, dist, phi;

		// Record polygon quadrants
		gulong quadrants = 0;

		K3DPOLYPOINTHANDLE point;
		k3dIPolyPointServer* points = Polygon->Points();
		for(point = points->GetFirstPolyPoint(); point; point = points->GetNextPolyPoint(point))
			{
				points->GetPolyPoint(point)->Location()->GetOldXYZ(coords);
				coords -= m_Center;

				switch(m_Axis)
					{
						case k3dIDocument::XAXIS:
							quadrants |= ((coords[2] >= 0) ? (coords[1] >= 0 ? 0x02 : 0x01) : (coords[1] >= 0 ? 0x04 : 0x08));
							break;

						case k3dIDocument::YAXIS:
							quadrants |= ((coords[0] <= 0) ? (coords[2] <= 0 ? 0x02 : 0x01) : (coords[2] <= 0 ? 0x04 : 0x08));
							break;

						case k3dIDocument::ZAXIS:
							quadrants |= ((coords[0] <= 0) ? (coords[1] >= 0 ? 0x02 : 0x01) : (coords[1] >= 0 ? 0x04 : 0x08));
							break;
					}
			}

		// Calculate texture coordinates
		for(point = points->GetFirstPolyPoint(); point; point = points->GetNextPolyPoint(point))
			{
				points->GetPolyPoint(point)->Location()->GetOldXYZ(coords);
				coords -= m_Center;

				switch(m_Axis)
					{
						case k3dIDocument::XAXIS:
    					theta = atan2(coords[2], -coords[1]);
    					theta += TWOPI * (theta < 0.0);

							dist = sdpVector2(coords[2], -coords[1]).Length();
							phi = atan2(dist, coords[0]);


    					texturecoords[0] = theta * ONE_OVER_TWOPI;
    					texturecoords[1] = phi * ONE_OVER_PI;
							texturecoords[2] = 0.0;

							if((theta <= PI) && ((quadrants & 0x09) == 0x09))
								{
									texturecoords[0] += 1.0;
								}

							break;

						case k3dIDocument::YAXIS:
    					theta = atan2(-coords[0], coords[2]);
    					theta += TWOPI * (theta < 0.0);

							dist = sdpVector2(coords[0], coords[2]).Length();
							phi = atan2(dist, coords[1]);

    					texturecoords[0] = theta * ONE_OVER_TWOPI;
    					texturecoords[1] = phi * ONE_OVER_PI;
							texturecoords[2] = 0.0;

							if((theta <= PI) && ((quadrants & 0x09) == 0x09))
								{
									texturecoords[0] += 1.0;
								}

							break;

						case k3dIDocument::ZAXIS:
    					theta = atan2(-coords[0], -coords[1]);
    					theta += TWOPI * (theta < 0.0);

							dist = sdpVector2(coords[0], -coords[1]).Length();
							phi = atan2(dist, coords[2]);

    					texturecoords[0] = theta * ONE_OVER_TWOPI;
    					texturecoords[1] = phi * ONE_OVER_PI;
							texturecoords[2] = 0.0;

							if((theta <= PI) && ((quadrants & 0x09) == 0x09))
								{
									texturecoords[0] += 1.0;
								}

							break;
					}

				points->GetPolyPoint(point)->Location()->SetTextureXYZ(texturecoords);
			}
		
		return true;
	}
	
protected:
	sdpVector3 m_Min;
	sdpVector3 m_Max;
	sdpVector3 m_Size;
	sdpVector3 m_Center;

	k3dIDocument::Axis m_Axis;
};

class CUVMapCubicTexture
{
public:
	bool HandleMesh(k3dIMesh* Mesh)
	{
		Mesh->GetDimensions(k3dIState::VISIBLESELECTED, m_Min, m_Max);
		m_Size = m_Max - m_Min;
		
		return true;
	}

	bool HandleMeshPolygon(k3dIMesh* Mesh, k3dIPolygon* Polygon)
	{
		if(Polygon->Points()->GetPointCount() < 3)
			return true;

		// Calculate which axis this polygon should be on
		sdpVector3 normal;
		Polygon->GetPolygonNormal(normal);
		normal.Normalize();

		Axis axis;
		if(fabs(normal[1]) >= fabs(normal[0]))
			{
				if(fabs(normal[1]) >= fabs(normal[2]))
					{
						axis = (normal[1] >= 0) ? YAXIS : NYAXIS;
					}
				else
					{
						axis = (normal[2] >= 0) ? ZAXIS : NZAXIS;
					}
			}
		else
			{
				if(fabs(normal[2]) >= fabs(normal[0]))
					{
						axis = (normal[2] >= 0) ? ZAXIS : NZAXIS;
					}
				else
					{
						axis = (normal[0] >= 0) ? XAXIS : NXAXIS;
					}
			}

		// For each point in the polygon ...
		sdpVector3 coords, texturecoords;
		k3dIPolyPointServer* points = Polygon->Points();
		for(K3DPOLYPOINTHANDLE point = points->GetFirstPolyPoint(); point; point = points->GetNextPolyPoint(point))
			{
				points->GetPolyPoint(point)->Location()->GetOldXYZ(coords);

				switch(axis)
					{
						case XAXIS:
							texturecoords[0] = ((m_Max[2] - coords[2]) / m_Size[2]);
							texturecoords[1] = ((m_Max[1] - coords[1]) / m_Size[1]);
							texturecoords[2] = ((m_Max[0] - coords[0]) / m_Size[0]);
							break;

						case NXAXIS:
							texturecoords[0] = ((coords[2] - m_Min[2]) / m_Size[2]);
							texturecoords[1] = ((m_Max[1] - coords[1]) / m_Size[1]);
							texturecoords[2] = ((m_Max[0] - coords[0]) / m_Size[0]);
							break;


						case YAXIS:
							texturecoords[0] = ((coords[0] - m_Min[0]) / m_Size[0]);
							texturecoords[1] = ((coords[2] - m_Min[2]) / m_Size[2]);
							texturecoords[2] = ((m_Max[1] - coords[1]) / m_Size[1]);
							break;

						case NYAXIS:
							texturecoords[0] = ((coords[0] - m_Min[0]) / m_Size[0]);
							texturecoords[1] = ((m_Max[2] - coords[2]) / m_Size[2]);
							texturecoords[2] = ((m_Max[1] - coords[1]) / m_Size[1]);
							break;

						case ZAXIS:
							texturecoords[0] = ((coords[0] - m_Min[0]) / m_Size[0]);
							texturecoords[1] = ((m_Max[1] - coords[1]) / m_Size[1]);
							texturecoords[2] = ((m_Max[2] - coords[2]) / m_Size[2]);
							break;

						case NZAXIS:
							texturecoords[0] = ((m_Max[0] - coords[0]) / m_Size[0]);
							texturecoords[1] = ((m_Max[1] - coords[1]) / m_Size[1]);
							texturecoords[2] = ((m_Max[2] - coords[2]) / m_Size[2]);
							break;

						default:
							g_assert(0);
							break;
					}

				points->GetPolyPoint(point)->Location()->SetTextureXYZ(texturecoords);
			}
		
		return true;
	}
	
protected:
	typedef enum
	{
		XAXIS,
		YAXIS,
		ZAXIS,
		NXAXIS,
		NYAXIS,
		NZAXIS
	} Axis;

	sdpVector3 m_Min;
	sdpVector3 m_Max;
	sdpVector3 m_Size;
};

class CUVMapPolygonTexture
{
public:
	bool HandleMesh(k3dIMesh* Mesh)
	{
		return true;
	}

	bool HandleMeshPolygon(k3dIMesh* Mesh, k3dIPolygon* Polygon)
	{
		static sdpVector3 texturecoords[4] =
		{
			sdpVector3(1.0, 0.0, 0.0),
			sdpVector3(0.0, 0.0, 0.0),
			sdpVector3(0.0, 1.0, 0.0),
			sdpVector3(1.0, 1.0, 0.0),
		};

		k3dIPolyPointServer* points = Polygon->Points();
		gulong pointcount = points->GetPointCount();
		if(pointcount < 3)
			return true;
		if(pointcount > 4)
			return true;

		gulong index = 0;
		for(K3DPOLYPOINTHANDLE polypointhandle = points->GetFirstPolyPoint(); polypointhandle; polypointhandle = points->GetNextPolyPoint(polypointhandle))
			{
				points->GetPolyPoint(polypointhandle)->Location()->SetTextureXYZ(texturecoords[index++]);
			}
		
		return true;
	}
};

class CUVMapImplicitTexture
{
public:
	bool HandleMesh(k3dIMesh* Mesh)
	{
		return true;
	}

	bool HandleMeshPolygon(k3dIMesh* Mesh, k3dIPolygon* Polygon)
	{
		sdpVector3 texturecoords;
		k3dILocation* location;
		k3dIPolyPointServer* points = Polygon->Points();
		for(K3DPOLYPOINTHANDLE point = points->GetFirstPolyPoint(); point; point = points->GetNextPolyPoint(point))
			{
				location = points->GetPolyPoint(point)->Location();
				location->GetImplicitTextureXYZ(texturecoords);
				location->SetTextureXYZ(texturecoords);
			}
			
		return true;
	}
};

void CUVMap::OnModify(bool SaveAction)
{
	// Record the action if requested ...
	if(SaveAction)
		Document()->Actions()->StartAction("UVMap");

	switch(m_Type)
		{
			case PROJECTION:
				{
					CUVMapProjectionTexture functor(Document());
					k3dForEachMeshPolygon(Document(), k3dIState::VISIBLESELECTED, functor);
					break;
				}
			case CYLINDRICAL:
				{
					CUVMapCylindricalTexture functor(Document());
					k3dForEachMeshPolygon(Document(), k3dIState::VISIBLESELECTED, functor);
					break;
				}
			case SPHERICAL:
				{
					CUVMapSphericalTexture functor(Document());
					k3dForEachMeshPolygon(Document(), k3dIState::VISIBLESELECTED, functor);
					break;
				}
			case CONIC:
				{
//					CUVMapConicTexture functor;
//					k3dForEachMeshPolygon(Document(), k3dIState::VISIBLESELECTED, functor);
g_assert_not_reached();
					break;
				}
			case CUBIC:
				{
					CUVMapCubicTexture functor;
					k3dForEachMeshPolygon(Document(), k3dIState::VISIBLESELECTED, functor);
					break;
				}
			case POLYGON:
				{
					CUVMapPolygonTexture functor;
					k3dForEachMeshPolygon(Document(), k3dIState::VISIBLESELECTED, functor);
					break;
				}
			case IMPLICIT:
				{
					CUVMapImplicitTexture functor;
					k3dForEachMeshPolygon(Document(), k3dIState::VISIBLESELECTED, functor);
					break;
				}
			default:
				g_assert_not_reached();	// Unknown texture type!
		}

	// Finish recording the action if requested ...
	if(SaveAction)
		{
			Document()->Actions()->FinishAction();
			Document()->FreezeChanges();
		}
		
	// Update the display ...
	k3dRedrawAll(Document(), false);
}

//////////////////////////////////////////////////////////////////////////////////////
// CUVMapDialog

CUVMapDialog::CUVMapDialog()
{
	m_Object = 0;
}

bool CUVMapDialog::Create(CUVMap* Object)
{
	// Store our parent object ...
	m_Object = Object;

	// Load our template ...
	g_return_val_if_fail(k3dDialog::Create(Object, true, false, "uvmap.gtkml", "uvmap.html"), false);

	return true;
}

void CUVMapDialog::UpdateControls()
{
	ToggleButton("projection").SetState(m_Object->m_Type == CUVMap::PROJECTION);
	ToggleButton("cylindrical").SetState(m_Object->m_Type == CUVMap::CYLINDRICAL);
	ToggleButton("spherical").SetState(m_Object->m_Type == CUVMap::SPHERICAL);
	ToggleButton("cubic").SetState(m_Object->m_Type == CUVMap::CUBIC);
//	ToggleButton("conic").SetState(m_Object->m_Type == CUVMap::CONIC);
	ToggleButton("polygon").SetState(m_Object->m_Type == CUVMap::POLYGON);
	ToggleButton("implicit").SetState(m_Object->m_Type == CUVMap::IMPLICIT);
}

void CUVMapDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "onprojection")
		OnProjection();
	else if(Event->Name() == "oncylindrical")
		OnCylindrical();
	else if(Event->Name() == "onspherical")
		OnSpherical();
	else if(Event->Name() == "oncubic")
		OnCubic();
	else if(Event->Name() == "onconic")
		OnConic();
	else if(Event->Name() == "onpolygon")
		OnPolygon();
	else if(Event->Name() == "onimplicit")
		OnImplicit();
	else if(Event->Name() == "onapply")
		OnApply();
	else
		k3dDialog::OnEvent(Event);
}

void CUVMapDialog::OnProjection()
{
	if(ToggleButton("projection").GetState())
		m_Object->m_Type = CUVMap::PROJECTION;
}

void CUVMapDialog::OnCylindrical()
{
	if(ToggleButton("cylindrical").GetState())
		m_Object->m_Type = CUVMap::CYLINDRICAL;
}

void CUVMapDialog::OnSpherical()
{
	if(ToggleButton("spherical").GetState())
		m_Object->m_Type = CUVMap::SPHERICAL;
}

void CUVMapDialog::OnCubic()
{
	if(ToggleButton("cubic").GetState())
		m_Object->m_Type = CUVMap::CUBIC;
}

void CUVMapDialog::OnConic()
{
	if(ToggleButton("conic").GetState())
		m_Object->m_Type = CUVMap::CONIC;
}

void CUVMapDialog::OnPolygon()
{
	if(ToggleButton("polygon").GetState())
		m_Object->m_Type = CUVMap::POLYGON;
}

void CUVMapDialog::OnImplicit()
{
	if(ToggleButton("implicit").GetState())
		m_Object->m_Type = CUVMap::IMPLICIT;
}

void CUVMapDialog::OnApply()
{
	m_Object->OnModify(true);
}


