// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CWave K-3D tool, which applies a "wave" transformation to geometry
		\author Timothy M. Shead (tshead@k-3d.com)
*/


#include "wave.h"


#include <sdpfile.h>
#include <k3dmisc.h>

#include <sdpgtkevent.h>
#include <k3diactions.h>

/////////////////////////////////////////////////////////////////////////////
// CWave

CWave::CWave()
{
	m_DisplaceX = false;
	m_DisplaceY = true;
	m_DisplaceZ = false;
	m_Amplitude = 0.0;
	m_WaveLength = 10.0;
	m_Phase = 0.0;
	m_CustomWave = false;
	m_FiniteRange = false;
	m_Range = 20.0;
}

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CWave::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIObject>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIUserEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIObject*>(this), InterfaceID, Interface)) return SDP_S_OK;

	return SDP_E_NO_INTERFACE;
}


gulong CWave::AddRef()
{
	return sdpComAddRef(m_ReferenceCount);
}


gulong CWave::Release()
{
	return sdpComRelease(this, m_ReferenceCount);
}



///////////////////////////////////////////////////////////////
// k3dIObject overrides

// Called to open an object's UI
sdpResult CWave::EditObject()
{
	// Create a user interface ...
	g_return_val_if_fail(m_UserInterface.Create(this), SDP_E_FAIL);

	// Update the display ...
	OnModify(false);

	return SDP_S_OK;
}

sdpResult CWave::ObjectDeleted()
{
	m_UserInterface.ObjectDeleted();
	return k3dTool::ObjectDeleted();
}

///////////////////////////////////////////////////////////////
// k3dIUserEventClient overrides

// Called while the user drags the left mouse button
sdpResult CWave::OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	// Compute factors for each mouse axis ...
	double xfactor = gdouble(CurrentX - LastX) * 0.1;
	double yfactor = gdouble(CurrentY - LastY) * 0.1;

	if(ShiftDown(Modifiers))
		m_WaveLength -= (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor);
	else if(ControlDown(Modifiers))
		m_Phase -= sdpRadians(fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor);
	else
		m_Amplitude -= (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor);

	// Update our dialog ...
	m_UserInterface.UpdateControls();

	// Update the display
	OnModify(false);

	return SDP_S_OK;
}


sdpResult CWave::OnLButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	OnModify(true);

	return SDP_S_OK;
}


void CWave::OnModify(bool SaveAction)
{
	// Record the action if requested ...
	if(SaveAction)
		Document()->Actions()->StartAction("Wave");

	// Call the default implementation ...
	k3dTool::OnModify(SaveAction);

	// Finish recording the action if requested ...
	if(SaveAction)
		{
			Document()->Actions()->FinishAction();
			Document()->FreezeChanges();
		}
}


void CWave::ModifyObject(k3dIObject* Object)
{
	// Sanity checks ...
	g_assert(Object);

	// Take a location-by-location approach to modifying the object ...
	ModifyLocationServer(Object);
}


void CWave::ModifyLocationServer(k3dIObject* Object)
{
	// Sanity checks ...
	g_assert(Object);

	// Call the default implementation ...
	k3dTool::ModifyLocationServer(Object);
}


void CWave::ModifyLocation(k3dILocation* Location)
{
	double phase = m_CustomWave ? 0.00277777778 * m_Phase : m_Phase;
	double frequency = m_WaveLength ? 1.0 / m_WaveLength : 0.0;

	sdpVector3 coords;
	Location->GetOldXYZ(coords);

//	coords -= m_Origin;
	double range = coords.Length();

	// Calculate displacement based on position in "wave" space ...
	double displacement;
	if(m_CustomWave)
		{
/*
			switch(m_ActiveAxis)
				{
					case k3dIDocument::XAXIS:
						displacement = (coords[0] * frequency) + phase;
						break;
					case k3dIDocument::YAXIS:
						displacement = (coords[1] * frequency) + phase;
						break;
					case k3dIDocument::ZAXIS:
						displacement = (coords[2] * frequency) + phase;
						break;
					default:
						g_assert(0);
						displacement = 0.0;
				}

			displacement -= floor(displacement);
			displacement = m_Wave->GetValue1(displacement, .0001) * m_Amplitude;
*/
		}
	else
		{
			switch(m_ActiveAxis)
				{
					case k3dIDocument::XAXIS:
						displacement = m_Amplitude * sin(sdpPiTimes2 * (coords[0] * frequency) + phase);
						break;
					case k3dIDocument::YAXIS:
						displacement = m_Amplitude * sin(sdpPiTimes2 * (coords[1] * frequency) + phase);
						break;
					case k3dIDocument::ZAXIS:
						displacement = m_Amplitude * sin(sdpPiTimes2 * (coords[2] * frequency) + phase);
						break;
					default:
						g_assert(0);
						displacement = 0.0;
				}
		}

	// Factor in the distance from origin
	if(m_FiniteRange && m_Range)
//		displacement *= (range <= m_Range) ? m_RangeChannel->GetValue1(range / m_Range, .0001) : 0.0;
		displacement *= (range <= m_Range) ? range / m_Range : 0.0;

	// Figure out which direction to move the points ...
	sdpVector3 normal;

	// Spherical waves ...
	if(m_DisplaceX && m_DisplaceY && m_DisplaceZ)
		normal = coords;

	// Cylindrical waves ...
	else if(m_DisplaceX && m_DisplaceY)
		normal = sdpVector3(coords[0], coords[1], 0.0);
	else if(m_DisplaceY && m_DisplaceZ)
		normal = sdpVector3(0.0, coords[1], coords[2]);
	else if(m_DisplaceX && m_DisplaceZ)
		normal = sdpVector3(coords[0], 0.0, coords[2]);

	// Linear waves ...
	else if(m_DisplaceX)
		normal = sdpVector3(1.0, 0.0, 0.0);
	else if(m_DisplaceY)
		normal = sdpVector3(0.0, 1.0, 0.0);
	else if(m_DisplaceZ)
		normal = sdpVector3(0.0, 0.0, 1.0);

	normal.Normalize();

	coords += displacement * normal;
	coords += m_Origin;
	Location->SetLocalXYZ(coords);
}


//////////////////////////////////////////////////////////////////////////////////////
// CWaveDialog

CWaveDialog::CWaveDialog()
{
	m_Object = 0;
}

bool CWaveDialog::Create(CWave* Object)
{
	// Store our parent object ...
	m_Object = Object;

	// Load our template ...
	g_return_val_if_fail(k3dDialog::Create(Object, true, false, "wave.gtkml", "wave.html"), false);

	return true;
}

void CWaveDialog::UpdateControls()
{
	CheckButton("displacex").SetState(m_Object->m_DisplaceX);
	CheckButton("displacey").SetState(m_Object->m_DisplaceY);
	CheckButton("displacez").SetState(m_Object->m_DisplaceZ);

	Entry("amplitude").SetText(k3dConvertDoubleString(k3dIChannel::DATA_DISTANCE, m_Object->m_Amplitude, 3));
	Entry("wavelength").SetText(k3dConvertDoubleString(k3dIChannel::DATA_DISTANCE, m_Object->m_WaveLength, 3));
	Entry("phase").SetText(k3dConvertDoubleString(k3dIChannel::DATA_ANGLE, m_Object->m_Phase, 3));
}

void CWaveDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "ondisplacex")
		OnDisplaceX();
	else if(Event->Name() == "ondisplacey")
		OnDisplaceY();
	else if(Event->Name() == "ondisplacez")
		OnDisplaceZ();
	else if(Event->Name() == "onamplitude")
		OnAmplitude();
	else if(Event->Name() == "onwavelength")
		OnWavelength();
	else if(Event->Name() == "onphase")
		OnPhase();
	else
		k3dDialog::OnEvent(Event);
}

void CWaveDialog::OnDisplaceX()
{
	m_Object->m_DisplaceX = CheckButton("displacex").GetState();
	m_Object->OnModify(false);

}

void CWaveDialog::OnDisplaceY()
{
	m_Object->m_DisplaceY = CheckButton("displacey").GetState();
	m_Object->OnModify(false);

}

void CWaveDialog::OnDisplaceZ()
{
	m_Object->m_DisplaceZ = CheckButton("displacez").GetState();
	m_Object->OnModify(false);
}

void CWaveDialog::OnAmplitude()
{
	if(UpdateControl(Entry("amplitude"), m_Object->m_Amplitude, k3dIChannel::DATA_DISTANCE, 3))
		m_Object->OnModify(true);
}

void CWaveDialog::OnWavelength()
{
	if(UpdateControl(Entry("wavelength"), m_Object->m_WaveLength, k3dIChannel::DATA_DISTANCE, 3))
		m_Object->OnModify(true);
}

void CWaveDialog::OnPhase()
{
	if(UpdateControl(Entry("phase"), m_Object->m_Phase, k3dIChannel::DATA_ANGLE, 3))
		m_Object->OnModify(true);
};


