// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CSlanter K-3D object, which applies an animated "slant" transformation to geometry
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include "slanter.h"

#include <sdpfile.h>
#include <sdpgtkevent.h>
#include <k3dmisc.h>
#include <k3dilocation.h>

CSlanter::CSlanter() : k3dPersistObject()
{
	AddVariable(&m_Displace("displace", X));
	AddVariable(&m_Along("along", Y));
	AddVariable(&m_Mirror("mirror", false));
	
	AddObject(&m_SlantFactor("slantfactor"));
	AddObject(&m_SlantFalloff("slantfalloff"));
}

CSlanter::~CSlanter()
{
}

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CSlanter::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIAnimate>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIObject>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIPersist>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dITimeServer>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIObject*>(this), InterfaceID, Interface)) return SDP_S_OK;										

	return SDP_E_NO_INTERFACE;
}

gulong CSlanter::AddRef()
{
	return sdpComAddRef(m_ReferenceCount);
}

gulong CSlanter::Release()
{
	return sdpComRelease(this, m_ReferenceCount);
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CSlanter::EditObject()
{
	g_return_val_if_fail(m_UserInterface.Create(this), SDP_E_FAIL);
	return SDP_S_OK;
}

sdpResult CSlanter::ObjectRenamed(SDPCSTRING NewName)
{
	m_UserInterface.ObjectRenamed(NewName);
	return k3dPersistObject::ObjectRenamed(NewName);
}

sdpResult CSlanter::ObjectDeleted()
{
	m_UserInterface.ObjectDeleted();
	return k3dPersistObject::ObjectDeleted();
}

///////////////////////////////////////////////////////////////
// k3dIAnimate overrides

class CSlanterSlantObject
{
public:
	CSlanterSlantObject(CSlanter::Axis Displace, CSlanter::Axis Along, bool Mirror, gdouble SlantFactor, k3dIChannel* SlantFalloff)
	{
		// Sanity checks ...
		g_assert(SlantFalloff);
		
		m_Displace = Displace;
		m_Along = Along;
		m_Mirror = Mirror;
		m_SlantFactor = SlantFactor;
		m_SlantFalloff = SlantFalloff;
	}

	bool HandleDescendantObject(k3dIObjectContext* Context)
	{
		// Get the object ...
		k3dIObject* object = Context->Object();
		
		// Get the object's dimensions ... if we can't, skip it ...
		sdpVector3 min;
		sdpVector3 max;
		if(!k3dObjectDimensions(object, 0, min, max))
			return true;

		// If the object doesn't support k3dILocationServer, we're done ...
		k3dILocationServer* locationserver = sdpQueryInterface<k3dILocationServer>(object);
		g_return_val_if_fail(locationserver, true);

		// Calculate the object origin ...
		sdpVector3 origin = (min + max) * 0.5;		

		// Calculate helper variables
		gdouble maxx = (max[0] - origin[0]) > (origin[0] - min[0]) ? max[0] - origin[0] : origin[0] - min[0];
		gdouble maxy = (max[1] - origin[1]) > (origin[1] - min[1]) ? max[1] - origin[1] : origin[1] - min[1];
		gdouble maxz = (max[2] - origin[2]) > (origin[2] - min[2]) ? max[2] - origin[2] : origin[2] - min[2];

		// Test to see that our slant axis isn't zero-length ...
		switch(m_Along)
			{
				case CSlanter::X:
					if(0 == maxx)
						return true;
					break;
				case CSlanter::Y:
					if(0 == maxy)
						return true;
					break;
				case CSlanter::Z:
					if(0 == maxz)
						return true;
					break;
				default:
					g_assert_not_reached();	// Unknown axis!!!
			}

		// For each location ...
		gdouble delta;
		sdpVector3 coords;
		k3dILocation* location;
		for(K3DLOCATIONHANDLE handle = locationserver->GetFirstLocation(); handle; handle = locationserver->GetNextLocation(handle))
			{
				// Get location coordinates ...
				location = locationserver->GetLocation(handle);
				location->GetAnimXYZ(coords);

				// Offset the origin ...
				coords -= origin;

				// Get our position along the slant axis ...
				switch(m_Along)
					{
						case CSlanter::X:
							delta = coords[0] / maxx;
							break;
						case CSlanter::Y:
							delta = coords[1] / maxy;
							break;
						case CSlanter::Z:
							delta = coords[2] / maxz;
							break;
						default:
							g_assert_not_reached();	// Unknown axis!!!
					}

				// Mirror it and handle falloff ...
				if(m_Mirror)
					delta = m_SlantFalloff->GetValue1(fabs(delta), 0.001);
				else
					delta = sdpSign(delta) * m_SlantFalloff->GetValue1(fabs(delta), 0.001);

				// Offset the coordinates ...
				switch(m_Displace)
					{
						case CSlanter::X:
							coords[0] += delta * m_SlantFactor;
							break;
						case CSlanter::Y:
							coords[1] += delta * m_SlantFactor;
							break;
						case CSlanter::Z:
							coords[2] += delta * m_SlantFactor;
							break;
						default:
							g_assert_not_reached();	// Unknown axis!!!
					}

				// Restore the origin ...
				coords += origin;
				location->SetAnimXYZ(coords);
			}
			
		locationserver->Release();									
						
		return true;
	}
	
protected:
	CSlanter::Axis m_Displace;
	CSlanter::Axis m_Along;
	bool m_Mirror;
	gdouble m_SlantFactor;
	k3dIChannel* m_SlantFalloff;
};

sdpResult CSlanter::PopAnimate()
{
	// Calculate Rotation values ...
	double slantfactor = m_SlantFactor->GetValue1(m_CurrentTime, m_DeltaTime);
	
	// Setup our slant functor ...
	CSlanterSlantObject functor(m_Displace(), m_Along(), m_Mirror(), slantfactor, m_SlantFalloff.Interface());
	
	// Apply it to each of our descendants ...
	k3dForEachDescendant(Document(), m_ObjectContext, functor);

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIPersist overrides

sdpResult CSlanter::InitNew()
{
	g_return_val_if_fail(sdpSucceeded(k3dPersistObject::InitNew()), SDP_E_FAIL);

	g_return_val_if_fail(m_SlantFactor.Create(Document(), "Slant Factor", 0.0, true, -1.0 * ULONG_MAX, ULONG_MAX, "Time", "Slant Factor", k3dIChannel::DATA_DISTANCE, true, this), SDP_E_FAIL);
	g_return_val_if_fail(m_SlantFalloff.Create(Document(), "Slant Falloff", 0.0, true, -1.0 * ULONG_MAX, ULONG_MAX, "Time", "Slanter Falloff", k3dIChannel::DATA_SCALAR, true, this), SDP_E_FAIL);
	k3dChannelSetLinearValue(m_SlantFalloff.Interface(), 0.0, 1.0);

	return SDP_S_OK;
}

/////////////////////////////////////////////////////////////////////////////
// CSlanterDialog

CSlanterDialog::CSlanterDialog()
{
	m_Object = 0;
}

bool CSlanterDialog::Create(CSlanter* Object)
{
	// Store our parent object ...
	m_Object = Object;

	// Load our template ...
	g_return_val_if_fail(k3dDialog::Create(Object, false, true, "slanter.gtkml", "slanter.html"), false);

	return true;
}

bool CSlanterDialog::InitializeControls()
{
	// Setup channel buttons ...
	ConnectChannelButton(m_Object->m_SlantFactor.Interface(), "slantfactor");
	ConnectChannelButton(m_Object->m_SlantFalloff.Interface(), "slantfalloff");

	return true;
}

void CSlanterDialog::UpdateControls()
{
	switch(m_Object->m_Displace())
		{
			case CSlanter::X:
				ToggleButton("displacex").SetState(true);
				break;
			case CSlanter::Y:
				ToggleButton("displacey").SetState(true);
				break;
			case CSlanter::Z:
				ToggleButton("displacez").SetState(true);
				break;
			default:
				g_assert_not_reached();
		}

	switch(m_Object->m_Along())
		{
			case CSlanter::X:
				ToggleButton("alongx").SetState(true);
				break;
			case CSlanter::Y:
				ToggleButton("alongy").SetState(true);
				break;
			case CSlanter::Z:
				ToggleButton("alongz").SetState(true);
				break;
			default:
				g_assert_not_reached();
		}
		
	ToggleButton("mirror").SetState(m_Object->m_Mirror());
}

void CSlanterDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "ondisplacex")
		OnDisplaceX();
	else if(Event->Name() == "ondisplacey")
		OnDisplaceY();
	else if(Event->Name() == "ondisplacez")
		OnDisplaceZ();
	else if(Event->Name() == "onalongx")
		OnAlongX();
	else if(Event->Name() == "onalongy")
		OnAlongY();
	else if(Event->Name() == "onalongz")
		OnAlongZ();
	else if(Event->Name() == "onmirror")
		OnMirror();
	else
		k3dDialog::OnEvent(Event);
}

void CSlanterDialog::OnDisplaceX()
{
	if(ToggleButton("displacex").GetState())
		{
			m_Object->m_Displace() = CSlanter::X;
			m_Object->OnModify(false);
		}
}

void CSlanterDialog::OnDisplaceY()
{
	if(ToggleButton("displacey").GetState())
		{
			m_Object->m_Displace() = CSlanter::Y;
			m_Object->OnModify(false);
		}
}

void CSlanterDialog::OnDisplaceZ()
{
	if(ToggleButton("displacez").GetState())
		{
			m_Object->m_Displace() = CSlanter::Z;
			m_Object->OnModify(false);
		}
}

void CSlanterDialog::OnAlongX()
{
	if(ToggleButton("alongx").GetState())
		{
			m_Object->m_Along() = CSlanter::X;
			m_Object->OnModify(false);
		}
}

void CSlanterDialog::OnAlongY()
{
	if(ToggleButton("alongy").GetState())
		{
			m_Object->m_Along() = CSlanter::Y;
			m_Object->OnModify(false);
		}
}

void CSlanterDialog::OnAlongZ()
{
	if(ToggleButton("alongz").GetState())
		{
			m_Object->m_Along() = CSlanter::Z;
			m_Object->OnModify(false);
		}
}

void CSlanterDialog::OnMirror()
{
	m_Object->m_Mirror() = ToggleButton("mirror").GetState();
	m_Object->OnModify(false);
}


