// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CSphereizer K-3D object, which applies an animated "sphereize" transformation to geometry
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include "sphereizer.h"

#include <sdpfile.h>
#include <k3dmisc.h>
#include <k3dilocation.h>

CSphereizer::CSphereizer() : k3dPersistObject()
{
	AddObject(&m_SphereizeFactor("sphereizefactor"));
}

CSphereizer::~CSphereizer()
{
}

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CSphereizer::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIAnimate>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIObject>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIPersist>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dITimeServer>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIObject*>(this), InterfaceID, Interface)) return SDP_S_OK;										

	return SDP_E_NO_INTERFACE;
}

gulong CSphereizer::AddRef()
{
	return sdpComAddRef(m_ReferenceCount);
}

gulong CSphereizer::Release()
{
	return sdpComRelease(this, m_ReferenceCount);
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CSphereizer::EditObject()
{
	g_return_val_if_fail(m_UserInterface.Create(this), SDP_E_FAIL);
	return SDP_S_OK;
}

sdpResult CSphereizer::ObjectRenamed(SDPCSTRING NewName)
{
	m_UserInterface.ObjectRenamed(NewName);
	return k3dPersistObject::ObjectRenamed(NewName);
}

sdpResult CSphereizer::ObjectDeleted()
{
	m_UserInterface.ObjectDeleted();
	return k3dPersistObject::ObjectDeleted();
}

///////////////////////////////////////////////////////////////
// k3dIAnimate overrides

class CSphereizerSphereizeObject
{
public:
	CSphereizerSphereizeObject(gdouble SphereizeFactor)
	{
		m_SphereizeFactor = SphereizeFactor;
	}

	bool HandleDescendantObject(k3dIObjectContext* Context)
	{
		// Get the object ...
		k3dIObject* object = Context->Object();
		
		// Get the object's dimensions ... if we can't, skip it ...
		sdpVector3 min;
		sdpVector3 max;
		if(!k3dObjectDimensions(object, 0, min, max))
			return true;

		// If the object doesn't support k3dILocationServer, we're done ...
		k3dILocationServer* locationserver = sdpQueryInterface<k3dILocationServer>(object);
		g_return_val_if_fail(locationserver, true);

		// Calculate the object origin ...
		sdpVector3 origin = (min + max) * 0.5;		

		// Calculate helper variables
		gdouble targetradius = (max - origin).Length();

		// For each location ...
		sdpVector3 coords, coords2;
		k3dILocation* location;
		gdouble radius;
		
		for(K3DLOCATIONHANDLE handle = locationserver->GetFirstLocation(); handle; handle = locationserver->GetNextLocation(handle))
			{
				// Get location coordinates ...
				location = locationserver->GetLocation(handle);
				location->GetAnimXYZ(coords);

				// Offset the origin ...
				coords -= origin;

				radius = coords.Length();
				if(radius)
					{
						coords2 = coords;
						coords += coords2.Normalize() * (m_SphereizeFactor * (targetradius - radius));
					}

				// Restore the origin ...
				coords += origin;
				location->SetAnimXYZ(coords);
			}
			
		locationserver->Release();									
						
		return true;
	}
	
protected:
	gdouble m_SphereizeFactor;
};

sdpResult CSphereizer::PopAnimate()
{
	// Calculate Rotation values ...
	gdouble factor = m_SphereizeFactor->GetValue1(m_CurrentTime, m_DeltaTime);
	
	// Setup our sphereiz functor ...
	CSphereizerSphereizeObject functor(factor);
	
	// Apply it to each of our descendants ...
	k3dForEachDescendant(Document(), m_ObjectContext, functor);

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIPersist overrides

sdpResult CSphereizer::InitNew()
{
	g_return_val_if_fail(sdpSucceeded(k3dPersistObject::InitNew()), SDP_E_FAIL);

	g_return_val_if_fail(m_SphereizeFactor.Create(Document(), "Sphereize Factor", 0.0, true, -1.0 * ULONG_MAX, ULONG_MAX, "Time", "Sphereize Factor", k3dIChannel::DATA_SCALAR, true, this), SDP_E_FAIL);

	return SDP_S_OK;
}

/////////////////////////////////////////////////////////////////////////////
// CSphereizerDialog

CSphereizerDialog::CSphereizerDialog()
{
	m_Object = 0;
}

bool CSphereizerDialog::Create(CSphereizer* Object)
{
	// Store our parent object ...
	m_Object = Object;

	// Load our template ...
	g_return_val_if_fail(k3dDialog::Create(Object, false, true, "sphereizer.gtkml", "sphereizer.html"), false);

	return true;
}

bool CSphereizerDialog::InitializeControls()
{
	// Setup channel buttons ...
	ConnectChannelButton(m_Object->m_SphereizeFactor.Interface(), "sphereizefactor");

	return true;
}

void CSphereizerDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

//	if(Event->Name() == "oncastshadows")
//		OnCastShadows();
//	else
		k3dDialog::OnEvent(Event);
}


