// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CTaperer K-3D object, which applies an animated "taper" transformation to geometry
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include "taperer.h"

#include <sdpfile.h>
#include <sdpgtkevent.h>
#include <k3dmisc.h>
#include <k3dilocation.h>

CTaperer::CTaperer()
{
	AddVariable(&m_DisplaceX("displacex", false));
	AddVariable(&m_DisplaceY("displacey", true));
	AddVariable(&m_DisplaceZ("displacez", true));
	AddVariable(&m_Along("along", X));

	AddObject(&m_TaperFactor("taperfactor"));
	AddObject(&m_TaperFalloff("taperfalloff"));
}

CTaperer::~CTaperer()
{
}

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CTaperer::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIAnimate>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIObject>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIPersist>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dITimeServer>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIObject*>(this), InterfaceID, Interface)) return SDP_S_OK;										

	return SDP_E_NO_INTERFACE;
}

gulong CTaperer::AddRef()
{
	return sdpComAddRef(m_ReferenceCount);
}

gulong CTaperer::Release()
{
	return sdpComRelease(this, m_ReferenceCount);
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CTaperer::EditObject()
{
	g_return_val_if_fail(m_UserInterface.Create(this), SDP_E_FAIL);
	return SDP_S_OK;
}

sdpResult CTaperer::ObjectRenamed(SDPCSTRING NewName)
{
	m_UserInterface.ObjectRenamed(NewName);
	return k3dPersistObject::ObjectRenamed(NewName);
}

sdpResult CTaperer::ObjectDeleted()
{
	m_UserInterface.ObjectDeleted();
	return k3dPersistObject::ObjectDeleted();
}

///////////////////////////////////////////////////////////////
// k3dIAnimate overrides

class CTapererTaperObject
{
public:
	CTapererTaperObject(bool DisplaceX, bool DisplaceY, bool DisplaceZ, CTaperer::Axis Along, gdouble TaperFactor, k3dIChannel* TaperFalloff)
	{
		// Sanity checks ...
		g_assert(TaperFalloff);
		
		m_DisplaceX = DisplaceX;
		m_DisplaceY = DisplaceY;
		m_DisplaceZ = DisplaceZ;
		m_Along = Along;
		m_TaperFactor = TaperFactor;
		m_TaperFalloff = TaperFalloff;
	}

	bool HandleDescendantObject(k3dIObjectContext* Context)
	{
		// Get the object ...
		k3dIObject* object = Context->Object();
		
		// Get the object's dimensions ... if we can't, skip it ...
		sdpVector3 min;
		sdpVector3 max;
		if(!k3dObjectDimensions(object, 0, min, max))
			return true;

		// If the object doesn't support k3dILocationServer, we're done ...
		k3dILocationServer* locationserver = sdpQueryInterface<k3dILocationServer>(object);
		g_return_val_if_fail(locationserver, true);

		// Calculate the object origin ...
		sdpVector3 origin = (min + max) * 0.5;		

		// Calculate helper variables
		gdouble maxx = (max[0] - origin[0]) > (origin[0] - min[0]) ? max[0] - origin[0] : origin[0] - min[0];
		gdouble maxy = (max[1] - origin[1]) > (origin[1] - min[1]) ? max[1] - origin[1] : origin[1] - min[1];
		gdouble maxz = (max[2] - origin[2]) > (origin[2] - min[2]) ? max[2] - origin[2] : origin[2] - min[2];
		
		gdouble omtaperfactor = 1.0 - m_TaperFactor;

		// Test to see that our taper axis isn't zero-length ...
		switch(m_Along)
			{
				case CTaperer::X:
					if(0 == maxx)
						return true;
					break;
				case CTaperer::Y:
					if(0 == maxy)
						return true;
					break;
				case CTaperer::Z:
					if(0 == maxz)
						return true;
					break;
				default:
					g_assert_not_reached();	// Unknown axis!!!
			}

		// For each location ...
		gdouble delta;
		gdouble scale;
		sdpVector3 coords;
		k3dILocation* location;
		for(K3DLOCATIONHANDLE handle = locationserver->GetFirstLocation(); handle; handle = locationserver->GetNextLocation(handle))
			{
				// Get location coordinates ...
				location = locationserver->GetLocation(handle);
				location->GetAnimXYZ(coords);

				// Offset the origin ...
				coords -= origin;

				// Get our position along the slant axis ...
				switch(m_Along)
					{
						case CTaperer::X:
							delta = coords[0] / maxx;
							break;
						case CTaperer::Y:
							delta = coords[1] / maxy;
							break;
						case CTaperer::Z:
							delta = coords[2] / maxz;
							break;
						default:
							g_assert_not_reached();	// Unknown axis!!!
					}

				// Handle falloff ...
				delta = m_TaperFalloff->GetValue1(fabs(delta), 0.001);
				scale = delta * omtaperfactor + m_TaperFactor;

				// Offset the coordinates ...
				if(m_DisplaceX)
					coords[0] *= scale;
				if(m_DisplaceY)
					coords[1] *= scale;
				if(m_DisplaceZ)
					coords[2] *= scale;

				// Restore the origin ...
				coords += origin;
				location->SetAnimXYZ(coords);
			}
			
		locationserver->Release();									
						
		return true;
	}
	
protected:
	bool m_DisplaceX;
	bool m_DisplaceY;
	bool m_DisplaceZ;
	CTaperer::Axis m_Along;
	gdouble m_TaperFactor;
	k3dIChannel* m_TaperFalloff;
};

sdpResult CTaperer::PopAnimate()
{
	// Calculate Rotation values ...
	gdouble taperfactor = m_TaperFactor->GetValue1(m_CurrentTime, m_DeltaTime);
	
	// Setup our taper functor ...
	CTapererTaperObject functor(m_DisplaceX(), m_DisplaceY(), m_DisplaceZ(), m_Along(), taperfactor, m_TaperFalloff.Interface());
	
	// Apply it to each of our descendants ...
	k3dForEachDescendant(Document(), m_ObjectContext, functor);

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIPersist overrides

sdpResult CTaperer::InitNew()
{
	g_return_val_if_fail(sdpSucceeded(k3dPersistObject::InitNew()), SDP_E_FAIL);

	g_return_val_if_fail(m_TaperFactor.Create(Document(), "Taper Factor", 0.0, true, -1.0 * ULONG_MAX, ULONG_MAX, "Time", "Taper Factor", k3dIChannel::DATA_SCALAR, true, this), SDP_E_FAIL);
	g_return_val_if_fail(m_TaperFalloff.Create(Document(), "Taper Falloff", 0.0, true, -1.0 * ULONG_MAX, ULONG_MAX, "Time", "Taper Falloff", k3dIChannel::DATA_SCALAR, true, this), SDP_E_FAIL);
	k3dChannelSetLinearValue(m_TaperFalloff.Interface(), 1.0, 0.0);
	
	return SDP_S_OK;
}

/////////////////////////////////////////////////////////////////////////////
// CTapererDialog

CTapererDialog::CTapererDialog()
{
	m_Object = 0;
}

bool CTapererDialog::Create(CTaperer* Object)
{
	// Store our parent object ...
	m_Object = Object;

	// Load our template ...
	g_return_val_if_fail(k3dDialog::Create(Object, false, true, "taperer.gtkml", "taperer.html"), false);

	return true;
}

bool CTapererDialog::InitializeControls()
{
	// Setup channel buttons ...
	ConnectChannelButton(m_Object->m_TaperFactor.Interface(), "taperfactor");
	ConnectChannelButton(m_Object->m_TaperFalloff.Interface(), "taperfalloff");

	return true;
}

void CTapererDialog::UpdateControls()
{
	CheckButton("displacex").SetState(m_Object->m_DisplaceX());
	CheckButton("displacey").SetState(m_Object->m_DisplaceY());
	CheckButton("displacez").SetState(m_Object->m_DisplaceZ());

	switch(m_Object->m_Along())
		{
			case CTaperer::X:
				ToggleButton("alongx").SetState(true);
				break;
			case CTaperer::Y:
				ToggleButton("alongy").SetState(true);
				break;
			case CTaperer::Z:
				ToggleButton("alongz").SetState(true);
				break;
			default:
				g_assert_not_reached();
		}
}

void CTapererDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "ondisplacex")
		OnDisplaceX();
	else if(Event->Name() == "ondisplacey")
		OnDisplaceY();
	else if(Event->Name() == "ondisplacez")
		OnDisplaceZ();
	else if(Event->Name() == "onalongx")
		OnAlongX();
	else if(Event->Name() == "onalongy")
		OnAlongY();
	else if(Event->Name() == "onalongz")
		OnAlongZ();
	else
		k3dDialog::OnEvent(Event);
}

void CTapererDialog::OnDisplaceX()
{
	m_Object->m_DisplaceX() = ToggleButton("displacex").GetState();
	m_Object->OnModify(false);
}

void CTapererDialog::OnDisplaceY()
{
	m_Object->m_DisplaceY() = ToggleButton("displacey").GetState();
	m_Object->OnModify(false);
}

void CTapererDialog::OnDisplaceZ()
{
	m_Object->m_DisplaceZ() = ToggleButton("displacez").GetState();
	m_Object->OnModify(false);
}

void CTapererDialog::OnAlongX()
{
	if(ToggleButton("alongx").GetState())
		{
			m_Object->m_Along() = CTaperer::X;
			m_Object->OnModify(false);
		}
}

void CTapererDialog::OnAlongY()
{
	if(ToggleButton("alongy").GetState())
		{
			m_Object->m_Along() = CTaperer::Y;
			m_Object->OnModify(false);
		}
}

void CTapererDialog::OnAlongZ()
{
	if(ToggleButton("alongz").GetState())
		{
			m_Object->m_Along() = CTaperer::Z;
			m_Object->OnModify(false);
		}
}


