// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CTransformer class, which is used to animate the position, orientation, and scale of objects
		\author Timothy M. Shead (tshead@k-3d.com)
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include "transformer.h"

#include <k3dmisc.h>
#include <k3dilocation.h>
#include <sdpgtkevent.h>

CTransformer::CTransformer()
{
	// Owned variables ...
	AddVariable(&m_GlobalCycles("globalcycles", 1.0));
	AddVariable(&m_GlobalPhase("globalphase", 0.0));

	AddVariable(&m_EnableTranslation("enabletranslation", true));
	AddVariable(&m_AbsolutePath("absolutepath", true));
	AddVariable(&m_TranslateLast("translatelast", true));
	AddVariable(&m_TranslateType("translatetype", LINEAR));
	
	// Owned objects ...
	AddObject(&m_Translation("translation"));
	AddObject(&m_XRotation("xrotation"));
	AddObject(&m_YRotation("yrotation"));
	AddObject(&m_ZRotation("zrotation"));
	AddObject(&m_XScale("xscale"));
	AddObject(&m_YScale("yscale"));
	AddObject(&m_ZScale("zscale"));
}

CTransformer::~CTransformer()
{
}

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CTransformer::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIAnimate>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIEditorRender>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dILocationServer>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIMesh>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIObject>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIPersist>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIPointServer>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIPolygonServer>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dINURBSCurveServer>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIRiRender>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dITimeServer>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dITransform>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIUserEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIObject*>(this), InterfaceID, Interface)) return SDP_S_OK;										

	return SDP_E_NO_INTERFACE;
}

gulong CTransformer::AddRef()
{
	return sdpComAddRef(m_ReferenceCount);
}

gulong CTransformer::Release()
{
	return sdpComRelease(this, m_ReferenceCount);
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CTransformer::EditObject()
{
	g_return_val_if_fail(m_UserInterface.Create(this), SDP_E_FAIL);
	return SDP_S_OK;
}

sdpResult CTransformer::ObjectRenamed(SDPCSTRING NewName)
{
	m_UserInterface.ObjectRenamed(NewName);
	return k3dMesh::ObjectRenamed(NewName);
}

sdpResult CTransformer::ObjectDeleted()
{
	m_UserInterface.ObjectDeleted();
	return k3dMesh::ObjectDeleted();
}

///////////////////////////////////////////////////////////////
// k3dIEditorRender overrides

sdpResult CTransformer::EditorRender(k3dIEditorEngine* Engine)
{
	// Default rendering ...
	g_return_val_if_fail(sdpSucceeded(k3dMesh::EditorRender(Engine)), SDP_E_FAIL);
	
	// Calculate our global time ...
	gdouble globaltime = (m_CurrentTime * m_GlobalCycles()) + m_GlobalPhase();
	globaltime -= floor(globaltime);

	// Calculate transformations for our children ...
	SetPosition(k3dITransform::ANIMATED_CHILD_FRAME, GetPosition(globaltime, m_DeltaTime));
	SetOrientation(k3dITransform::ANIMATED_CHILD_FRAME, sdpAngleAxis(sdpQuaternion(sdpEulerAngles(GetOrientation(globaltime, m_DeltaTime), sdpEulerAngles::ZXYstatic))));
	SetScale(k3dITransform::ANIMATED_CHILD_FRAME, GetScale(globaltime, m_DeltaTime));
		
	return SDP_S_OK;
}

sdpVector3 CTransformer::GetPointPosition(gdouble PathPercent)
{
	gulong pointcount = GetPointCount();
	if(0 == pointcount)
		return sdpVector3();

	gulong pointindex = 0;
	for(K3DPOINTHANDLE pointhandle = GetFirstPoint(); pointhandle; pointhandle = GetNextPoint(pointhandle))
		{
			if(gdouble(++pointindex) / gdouble(pointcount) < PathPercent)
				continue;
				
			sdpVector3 result;
			GetPoint(pointhandle)->Location()->GetAnimXYZ(result);
			return result;
		}

	g_return_val_if_fail(0, sdpVector3());
}

sdpVector3 CTransformer::GetLinearPosition(gdouble PathPercent)
{
	if(0 == GetPointCount())
		return sdpVector3();
	
	// Get the first point ...
	sdpVector3 firstpoint;
	GetPoint(GetFirstPoint())->Location()->GetAnimXYZ(firstpoint);
	if(1 == GetPointCount())
		return firstpoint;
	
	// Calculate the total distance around the path
	sdpVector3 point;
	sdpVector3 lastpoint = firstpoint;
	gdouble pathlength = 0;
	for(K3DPOINTHANDLE pointhandle1 = GetNextPoint(GetFirstPoint()); pointhandle1; pointhandle1 = GetNextPoint(pointhandle1))
		{
			GetPoint(pointhandle1)->Location()->GetAnimXYZ(point);

			pathlength += (point - lastpoint).Length();
			lastpoint = point;
		}

	if(m_TranslateLast())
		{
			pathlength += (firstpoint - lastpoint).Length();
		}

	// Calculate the desired current distance
	gdouble pathdistance = PathPercent * pathlength;

	// Find the location in space ...
	lastpoint = firstpoint;
	for(K3DPOINTHANDLE pointhandle2 = GetNextPoint(GetFirstPoint()); pointhandle2; pointhandle2 = GetNextPoint(pointhandle2))
		{
			sdpVector3 point;
			GetPoint(pointhandle2)->Location()->GetAnimXYZ(point);

			gdouble segmentlength = (point - lastpoint).Length();

			if((pathdistance - segmentlength) < 0.0)
				return sdpLerp(lastpoint, point, pathdistance / segmentlength);

			pathdistance -= segmentlength;
			lastpoint = point;
		}
	
	gdouble segmentlength = (firstpoint - lastpoint).Length();
	return sdpLerp(lastpoint, firstpoint, pathdistance / segmentlength);
}

sdpVector3 CTransformer::GetNURBSPosition(gdouble PathPercent)
{
	return sdpVector3();
}

sdpVector3 CTransformer::GetPosition(gdouble Time, gdouble Delta)
{
	if(m_EnableTranslation())
		{
			gdouble pathpercent = m_Translation->GetValue1(Time, Delta);
			
			switch(m_TranslateType())
				{
					case POINT:
						return GetPointPosition(pathpercent);
					case LINEAR:
						return GetLinearPosition(pathpercent);
					case NURBS:
						return GetNURBSPosition(pathpercent);
					default:
						g_assert_not_reached();	// Did you add a new translate type without telling me?
				}
		}

	return sdpVector3();
}

sdpVector3 CTransformer::GetOrientation(gdouble Time, gdouble Delta)
{
	return sdpVector3(m_XRotation->GetValue1(Time, Delta), m_YRotation->GetValue1(Time, Delta), m_ZRotation->GetValue1(Time, Delta));
}

sdpVector3 CTransformer::GetScale(gdouble Time, gdouble Delta)
{
	return sdpVector3(m_XScale->GetValue1(Time, Delta), m_YScale->GetValue1(Time, Delta), m_ZScale->GetValue1(Time, Delta));
}

///////////////////////////////////////////////////////////////
// k3dIPersist overrides

sdpResult CTransformer::InitNew()
{
	g_return_val_if_fail(sdpSucceeded(k3dMesh::InitNew()), SDP_E_FAIL);

	g_return_val_if_fail(m_XRotation.Create(Document(), "X Rotation", 0.0, true, -DBL_MAX, DBL_MAX, "Time", "X Rotation", k3dIChannel::DATA_ANGLE, true, this), SDP_E_FAIL);
	g_return_val_if_fail(m_YRotation.Create(Document(), "Y Rotation", 0.0, true, -DBL_MAX, DBL_MAX, "Time", "Y Rotation", k3dIChannel::DATA_ANGLE, true, this), SDP_E_FAIL);
	g_return_val_if_fail(m_ZRotation.Create(Document(), "Z Rotation", 0.0, true, -DBL_MAX, DBL_MAX, "Time", "Z Rotation", k3dIChannel::DATA_ANGLE, true, this), SDP_E_FAIL);
	g_return_val_if_fail(m_XScale.Create(Document(), "X Scale", 1.0, true, -DBL_MAX, DBL_MAX, "Time", "X Scale", k3dIChannel::DATA_SCALAR, true, this), SDP_E_FAIL);
	g_return_val_if_fail(m_YScale.Create(Document(), "Y Scale", 1.0, true, -DBL_MAX, DBL_MAX, "Time", "Y Scale", k3dIChannel::DATA_SCALAR, true, this), SDP_E_FAIL);
	g_return_val_if_fail(m_ZScale.Create(Document(), "Z Scale", 1.0, true, -DBL_MAX, DBL_MAX, "Time", "Z Scale", k3dIChannel::DATA_SCALAR, true, this), SDP_E_FAIL);
	g_return_val_if_fail(m_Translation.Create(Document(), "Translation", 0.0, true, 0.0, 1.0, "Time", "Translation", k3dIChannel::DATA_SCALAR, true, this), SDP_E_FAIL);
	k3dChannelSetLinearValue(m_Translation.Interface(), 0.0, 1.0);

	
	k3dIPoint* point;
	point = CreatePoint();
	point->Location()->SetLocalXYZ(sdpVector3(0, 0, 0));
	point->Location()->SetOldXYZ(sdpVector3(0, 0, 0));
	point = CreatePoint();
	point->Location()->SetLocalXYZ(sdpVector3(10, 1, 0));
	point->Location()->SetOldXYZ(sdpVector3(10, 1, 0));
	point = CreatePoint();
	point->Location()->SetLocalXYZ(sdpVector3(-5, 6, 0));
	point->Location()->SetOldXYZ(sdpVector3(-5, 6, 0));
		
	return SDP_S_OK;
}

/////////////////////////////////////////////////////////////////////////////
// CTransformerDialog

CTransformerDialog::CTransformerDialog()
{
	m_Object = 0;
}

bool CTransformerDialog::Create(CTransformer* Object)
{
	// Store our parent object ...
	m_Object = Object;

	// Load our template ...
	g_return_val_if_fail(k3dDialog::Create(Object, false, true, "transformer.gtkml", "transformer.html"), false);

	return true;
}

bool CTransformerDialog::InitializeControls()
{
	// Setup channel buttons ...
	ConnectChannelButton(m_Object->m_Translation.Interface(), "translation");

	ConnectChannelButton(m_Object->m_XRotation.Interface(), "xrotation");
	ConnectChannelButton(m_Object->m_YRotation.Interface(), "yrotation");
	ConnectChannelButton(m_Object->m_ZRotation.Interface(), "zrotation");

	ConnectChannelButton(m_Object->m_XScale.Interface(), "xscale");
	ConnectChannelButton(m_Object->m_YScale.Interface(), "yscale");
	ConnectChannelButton(m_Object->m_ZScale.Interface(), "zscale");

	// Setup translate type combo ...
	sdpStringArray translatetypes;
	translatetypes.push_back("Point");
	translatetypes.push_back("Linear");
	translatetypes.push_back("NURBS");
	Combo("translatetype").SetList(translatetypes);
	
	MapEvent("hide", "ontranslatetype", GTK_OBJECT(GTK_COMBO(Combo("translatetype").Object())->popwin), true);

	return true;
}

void CTransformerDialog::UpdateControls()
{
	ToggleButton("translatelast").SetState(m_Object->m_TranslateLast());
}

void CTransformerDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "ontranslatelast")
		OnTranslateLast();
	else if(Event->Name() == "ontranslatetype")
		OnTranslateType();
	else
		k3dDialog::OnEvent(Event);
}

void CTransformerDialog::OnTranslateLast()
{
	m_Object->m_TranslateLast() = ToggleButton("translatelast").GetState();
	m_Object->OnModify(false);
}

void CTransformerDialog::OnTranslateType()
{
	sdpString translatetype = Combo("translatetype").GetEntryText();
	if(translatetype == "Point")
		m_Object->m_TranslateType() = CTransformer::POINT;
	else if(translatetype == "Linear")
		m_Object->m_TranslateType() = CTransformer::LINEAR;
	else if(translatetype == "NURBS")
		m_Object->m_TranslateType() = CTransformer::NURBS;
	else
		g_assert_not_reached();	 // Did you add a new type without telling me?
	
	m_Object->OnModify(false);
}

