#include <sdpxmldocument.h>
#include <sdpxmlutility.h>

#include <map>
#include <utility>
#include <iostream>
#include <fstream>

static const sdpString GTKMLVersion = "0.1.7";

typedef std::multimap<sdpString, sdpXmlIElement*> GTKMLElementMap;

void StartHTMLPage(SDPCSTRING Title, std::ostream& Stream)
{
	// Sanity checks ...
	g_assert_string(Title);
	g_assert(Stream.good());
	
	Stream << "<html>" << std::endl;
	Stream << "<head>" << std::endl;
	Stream << "<title>" << Title << "</title>" << std::endl;
	Stream << "<link rel=\"stylesheet\" type=\"text/css\" href=\"gtkml.css\">" << std::endl;
	Stream << "</head>" << std::endl;
	Stream << "<body>" << std::endl << std::endl;
}

void FinishHTMLPage(std::ostream& Stream)
{
	// Sanity checks ...
	g_assert(Stream.good());
	
	Stream << "</body>" << std::endl;
	Stream << "</html>" << std::endl << std::endl;
}

void Header(std::ostream& Stream)
{
	// Sanity checks ...
	g_assert(Stream.good());
	
	Stream << "<h4 align=\"right\">GTKML " << GTKMLVersion << " Reference</h4>" << std::endl;
}

void Footer(std::ostream& Stream)
{
	// Sanity checks ...
	g_assert(Stream.good());
	
	Stream << "<h4 align=\"right\">GTKML " << GTKMLVersion << " Reference</h4>" << std::endl;
}

void GetGTKMLElements(sdpXmlIElement* Parent, GTKMLElementMap& Results)
{
	// Sanity checks ...
	g_assert(Parent);

	if(Parent->Name() == sdpString("element"))
		{
			sdpString elementname;
			sdpXmlGetAttribute(Parent, "name", elementname);

			if(elementname.size())			
				Results.insert(std::make_pair(elementname, Parent));
		}
			
	for(SDPXMLELEMENTHANDLE e = Parent->GetFirstElement(); e; e = Parent->GetNextElement(e))
		GetGTKMLElements(Parent->GetElement(e), Results);
}

void GetGTKMLAttributes(sdpXmlIElement* Parent, GTKMLElementMap& Results)
{
	// Sanity checks ...
	g_assert(Parent);
	
	for(SDPXMLELEMENTHANDLE e = Parent->GetFirstElement(); e; e = Parent->GetNextElement(e))
		{
			sdpXmlIElement* element = Parent->GetElement(e);
			
			if(element->Name() == sdpString("attribute"))
				{				
					sdpString attributename;
					sdpXmlGetAttribute(element, "name", attributename);
					g_return_if_fail(attributename.size());
					Results.insert(std::make_pair(attributename, element));
				}
			else
				{
					GetGTKMLAttributes(element, Results);
				}
		}
}

void GetInheritedGTKMLAttributes(sdpXmlIElement* Element, GTKMLElementMap& Results)
{
	// Sanity checks ...
	g_assert(Element);
	
	for(SDPXMLELEMENTHANDLE e = Element->GetFirstElement(); e; e = Element->GetNextElement(e))
		{
			sdpXmlIElement* element = Element->GetElement(e);
			
			if(element->Name() == sdpString("attribute"))
				{				
					sdpString attributename;
					sdpXmlGetAttribute(element, "name", attributename);
					g_return_if_fail(attributename.size());
					
					Results.insert(std::make_pair(attributename, element));
				}
		}
		
	if(Element->ParentElement())
		GetInheritedGTKMLAttributes(Element->ParentElement(), Results);
}

sdpString QualifiedAttributeName(sdpXmlIElement* Attribute)
{
	// Sanity checks ...
	g_assert(Attribute);
	
	sdpString name;
	sdpXmlGetAttribute(Attribute, "name", name);
	
	sdpString ownerclass;
	sdpXmlGetAttribute(Attribute->ParentElement(), "sdpgtk", ownerclass);
	
	sdpString ownerelement;
	sdpXmlGetAttribute(Attribute->ParentElement(), "name", ownerelement);
	
	if(ownerclass.size())
		return name + "_" + ownerclass;
		
	g_assert(ownerelement.size());
	
	return name + "_" + ownerelement;
}

void CreateElementsIndex(sdpXmlIDocument* Document)
{
	// Sanity checks ...
	g_assert(Document);
	
	std::ofstream stream("elements.html");
	g_return_if_fail(stream.good());

	GTKMLElementMap elements;
	GetGTKMLElements(Document->GetRootElement(), elements);
			
	StartHTMLPage("GTKML Version " + GTKMLVersion + " Elements", stream);
	Header(stream);
	
	stream << "<h1>GTKML Elements</h1>" << std::endl;

	stream << "<p>GTKML currently includes the following " << elements.size() << " elements:</p>" << std::endl;

	stream << "<table>" << std::endl;
	stream << "<tr><td class=\"BOX\">GTKML Element</td><td class=\"BOX\">Description</td>" << std::endl;
	
	for(GTKMLElementMap::iterator element = elements.begin(); element != elements.end(); element++)
		{
			sdpString name = element->first;
			
			sdpString description;
			sdpXmlGetAttribute(element->second, "description", description);
			
			stream << "<tr>";
			stream << "<td><b><a href=\"element_" << name << ".html\">" << "&lt;" << name << "&gt;" << "</a></b></td>";
			stream << "<td>" << description << "</td>";
			stream << "</tr>" << std::endl;
		}
	
	stream << "</table>" << std::endl;
	
	Footer(stream);	
	FinishHTMLPage(stream);
}

void CreateElements(sdpXmlIDocument* Document)
{
	// Sanity checks ...
	g_assert(Document);
	
	GTKMLElementMap elements;
	GetGTKMLElements(Document->GetRootElement(), elements);
	
	for(GTKMLElementMap::iterator element = elements.begin(); element != elements.end(); element++)
		{
			sdpString name = element->first;

			sdpString description;
			sdpXmlGetAttribute(element->second, "description", description);
			
			sdpString example;
			sdpXmlGetAttribute(element->second, "example", example);
			
			sdpString children;
			sdpXmlGetAttribute(element->second, "children", children);
			
			sdpString text = "ignored";
			sdpXmlGetAttribute(element->second, "text", text);
			
			sdpString gtk = "none";
			sdpXmlGetAttribute(element->second, "gtk", gtk);
			
			sdpString sdpgtk = "none";
			sdpXmlGetAttribute(element->second, "sdpgtk", sdpgtk);
			
			GTKMLElementMap attributes;
			GetInheritedGTKMLAttributes(element->second, attributes);
			
			std::ofstream stream("element_" + name + ".html");
			g_return_if_fail(stream.good());
			
			StartHTMLPage("GTKML Version " + GTKMLVersion + " &lt;" + name + "&gt; Element", stream);
			Header(stream);
			
			stream << "<h1>&lt;" << name << "&gt; element</h1>" << std::endl;
			stream << "<h3>description:</h3><p>" << description << "</p>" << std::endl;
//			stream << "<h3>example:</h3><pre>" << example << "</pre>" << std::endl;
			stream << "<h3>children:</h3><p>" << children << "</p>" << std::endl;
			stream << "<h3>element text:</h3><p>" << text << "</p>" << std::endl;
			stream << "<h3>attributes:</h3>" << std::endl;

			if(attributes.size())
				{
					stream << "<table cellpadding=\"4\">" << std::endl;
					stream << "<tr><td class=\"BOX\">name</td><td class=\"BOX\">allowed values</td><td class=\"BOX\">default value</td><td class=\"BOX\">required?</td></tr>" << std::endl;

					bool row = true;	
					for(GTKMLElementMap::iterator attribute = attributes.begin(); attribute != attributes.end(); attribute++)
						{
							sdpString name;
							sdpXmlGetAttribute(attribute->second, "name", name);
							
							sdpString qualifiedname = QualifiedAttributeName(attribute->second);
							
							sdpString values;
							sdpXmlGetAttribute(attribute->second, "values", values);
							
							sdpString defaultvalue;
							sdpXmlGetAttribute(attribute->second, "default", defaultvalue);
							if(0 == defaultvalue.size())
								defaultvalue = "&nbsp;";
							
							sdpString required;
							sdpXmlGetAttribute(attribute->second, "required", required);
							
							sdpString rowcolor = row ? "#FFFFFF" : "#EEEEEE";
							row = !row;
	
							stream << "<tr>";
							stream << "<td bgcolor=\"" << rowcolor << "\">" << "<a href=\"attribute_" << qualifiedname << ".html\">" << name << "</a></td>";
							stream << "<td bgcolor=\"" << rowcolor << "\">" << values << "</td>";
							stream << "<td bgcolor=\"" << rowcolor << "\">" << defaultvalue << "</td>";
							stream << "<td bgcolor=\"" << rowcolor << "\">" << required << "</td>";
							stream << "</tr>" << std::endl;
						}

					stream << "</table>" << std::endl;
				}
			else
				{
					stream << "none" << std::endl;
				}

			stream << "<h3>GTK class:</h3><p>" << gtk << "</p>" << std::endl;
			stream << "<h3>SDPGTK class:</h3><p>" << sdpgtk << "</p>" << std::endl;
	
			Footer(stream);											
			FinishHTMLPage(stream);
		}	
}

void CreateAttributesIndex(sdpXmlIDocument* Document)
{
	// Sanity checks ...
	g_assert(Document);
	
	std::ofstream stream("attributes.html");
	g_return_if_fail(stream.good());

	GTKMLElementMap attributes;
	GetGTKMLAttributes(Document->GetRootElement(), attributes);
			
	StartHTMLPage("GTKML Version " + GTKMLVersion + " Attributes", stream);
	Header(stream);
	
	stream << "<h1>GTKML Attributes</h1>" << std::endl;

	stream << "<p>GTKML currently includes the following " << attributes.size() << " attributes:</p>" << std::endl;

	stream << "<table>" << std::endl;
	stream << "<tr><td class=\"BOX\">GTKML Attribute</td><td class=\"BOX\">Applies To</td>" << std::endl;
	
	for(GTKMLElementMap::iterator attribute = attributes.begin(); attribute != attributes.end(); attribute++)
		{
			sdpString name = attribute->first;

			sdpString qualifiedname = QualifiedAttributeName(attribute->second);

			GTKMLElementMap appliesto;
			GetGTKMLElements(attribute->second->ParentElement(), appliesto);
			
			stream << "<tr>";
			stream << "<td><b><a href=\"attribute_" << qualifiedname << ".html\">" << "&lt;" << name << "&gt;" << "</a></b></td>";
			
			stream << "<td>";
			for(GTKMLElementMap::iterator element = appliesto.begin(); element != appliesto.end(); element++)
				{
					if(element != appliesto.begin())
						stream << ", ";
						
					stream << "<a href=\"element_" << element->first << ".html\">" << element->first << "</a>";
				}
			stream << "</td>";

			stream << "</tr>" << std::endl;
		}
	
	stream << "</table>" << std::endl;
	
	Footer(stream);
	FinishHTMLPage(stream);
}

void CreateAttributes(sdpXmlIDocument* Document)
{
	// Sanity checks ...
	g_assert(Document);
	
	GTKMLElementMap attributes;
	GetGTKMLAttributes(Document->GetRootElement(), attributes);
	
	for(GTKMLElementMap::iterator attribute = attributes.begin(); attribute != attributes.end(); attribute++)
		{
			sdpString name = attribute->first;
			sdpString qualifiedname = QualifiedAttributeName(attribute->second);

			sdpString description;
			sdpXmlGetAttribute(attribute->second, "description", description);
			
			sdpString example;
			sdpXmlGetAttribute(attribute->second, "example", example);
			
			sdpString values;
			sdpXmlGetAttribute(attribute->second, "values", values);
			
			sdpString defaultvalue;
			sdpXmlGetAttribute(attribute->second, "default", defaultvalue);
			
			sdpString required;
			sdpXmlGetAttribute(attribute->second, "required", required);
			
			GTKMLElementMap appliesto;
			GetGTKMLElements(attribute->second->ParentElement(), appliesto);
			
			std::ofstream stream("attribute_" + qualifiedname + ".html");
			g_return_if_fail(stream.good());
			
			StartHTMLPage("GTKML Version " + GTKMLVersion + " " + name + " Attribute", stream);
			Header(stream);

			stream << "<h1>\"" << name << "\" attribute</h1>" << std::endl;

			stream << "<h3>applies to:</h3>" << std::endl;
			stream << "<p>";
			for(GTKMLElementMap::iterator element = appliesto.begin(); element != appliesto.end(); element++)
				{
					if(element != appliesto.begin())
						stream << ", ";
						
					stream << "<a href=\"element_" << element->first << ".html\">" << element->first << "</a>";
				}
			stream << "</p>";
			
			stream << "<h3>description:</h3><p>" << description << "</p>" << std::endl;
//			stream << "<h3>example:</h3><pre>" << example << "</pre>" << std::endl;
			stream << "<h3>allowed values:</h3><p>" << values << "</p>" << std::endl;
			stream << "<h3>default value:</h3><p>" << defaultvalue << "</p>" << std::endl;
			stream << "<h3>required?:</h3><p>" << required << "</p>" << std::endl;

			Footer(stream);
			FinishHTMLPage(stream);
		}	
}

int main(int argc, char argv[])
{
	sdpXmlDocument doc;
	g_return_val_if_fail(doc.LoadFile("gtkmldoc.xml"), 1);
	g_return_val_if_fail(doc.GetRootElement(), 2);
	g_return_val_if_fail(doc.GetRootElement()->Name() == sdpString("gtkmldoc"), 3);

	CreateElementsIndex(&doc);
	CreateElements(&doc);
	
	CreateAttributesIndex(&doc);
	CreateAttributes(&doc);
			
	return 0;
}
