// SDPGTK
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the nurbsample application, which can be used to explore NURBS curve configurations
		\author Tim Shead (tshead@k-3d.com)
*/

#include <sdpgtkobjectcontainer.h>
#include <sdpgtkopengldrawingarea.h>
#include <sdpgtkevent.h>
#include <sdpmath.h>

#include <iostream>
#include <vector>

// sdpGtkObjectContainer is the main SDPGTK class - without alteration,
// it encapsulates all the functionality required to load GTKML documents
// and instantiate all their widgets.
//
// We derive our own class from it so we can override OnEvent() and
// handle user events.

//////////////////////////////////////////////////////////////////////////
// SampleWindow

#ifndef CALLBACK
#define CALLBACK
#endif // !CALLBACK

#define MAXORDER 8
#define MAXCONTROLPOINTCOUNT 8
#define MAXKNOTCOUNT (MAXORDER + MAXCONTROLPOINTCOUNT)

/// The main window for the application
class SampleWindow : public sdpGtkObjectContainer
{
public:
	SampleWindow();

	/// Loads the GTKML template
	void Load();

	gulong Order() { return m_Order; }
	gulong ControlPointCount() { return m_ControlPointCount; }
	gulong KnotCount() { return m_Order + m_ControlPointCount; }

protected:
	/// Updates controls
	void UpdateControls();

	/// Handles events for this template
	void OnEvent(sdpGtkEvent* Event);
	/// Called to exit the program
	void Exit();
	/// Called when the contents of the window need to be redrawn
	void OnDraw();
	/// Called when the user changes the sample tolerance
	void OnTolerance();
	/// Called when the user changes the curve order
	void OnOrder();
	/// Called when the user changes the number of curve control points
	void OnControlPointCount();
	/// Called when the user wants to make the curve periodic-uniform
	void OnPeriodicUniform();
	/// Called when the user wants to make the curve open-uniform
	void OnOpenUniform();
	/// Called when the user wants to reduce the curve to a Bezier basis
	void OnBezier();
	/// Called when the user adjusts the position of one of the control points
	void OnPosition();
	/// Called when the user adjust the value of one of the rational control points
	void OnRational();
	/// Called when the user changes a knot value
	void OnKnot(const sdpString& Knot);
	/// Schedules a redraw of the window
	void Redraw();

	/// Makes the curve periodic-uniform
	void MakePeriodicUniform();
	/// Makes the curve open-uniform
	void MakeOpenUniform();
	/// Makes the curve use a Bezier basis
	void MakeBezier();
	
	/// OpenGL context for the main window
	sdpGtkOpenGLDrawingArea m_OpenGLWidget;
	
	/// GLU NURBS sample tolerance
	gdouble m_Tolerance;
	
	/// NURBS order
	gulong m_Order;
	/// NURBS control point count
	gulong m_ControlPointCount; 
	
	/// NURBS knot vector
	static GLfloat m_Knots[MAXKNOTCOUNT];
	/// Control points
	static GLfloat m_ControlPoints[MAXCONTROLPOINTCOUNT][4];
};

GLfloat SampleWindow::m_Knots[MAXKNOTCOUNT];

GLfloat SampleWindow::m_ControlPoints[MAXCONTROLPOINTCOUNT][4] =
{
	{ 0, 2, 0, 1 },
	{ 1, 5, 0, 1 },
	{ 3, 4, 0, 1 },
	{ 3, 1, 0, 1 },
	{ 5, 0, 0, 1 },
	{ 7, 1, 0, 1 },
	{ 8, 4, 0, 1 },
	{ 10, 4, 0, 1 },
};

SampleWindow::SampleWindow()
{
	m_Tolerance = 15.0;
	m_Order = 4;
	m_ControlPointCount = 8;
	
	MakeOpenUniform();
};

void SampleWindow::Load()
{
	// Load the GTKML document ...
	sdpGtkObjectContainer::LoadFile("mainwindow.gtkml");

	// Create our OpenGL drawing area ...
	sdpGtkContainer container = Container("drawingarea");
	g_return_if_fail(m_OpenGLWidget.Create(container, true, true, 5, 6, 5, 16));

	MapEvent("configure-event", "onconfigureevent", &m_OpenGLWidget, true);
	MapEvent("expose-event", "onexposeevent", &m_OpenGLWidget, true);

	UpdateControls();

	// Make everything visible ...
	RootWidget().Show();
}

void SampleWindow::UpdateControls()
{
	Entry("tolerance").SetText(sdpString(m_Tolerance));
	Entry("order").SetText(sdpString(m_Order));
	Entry("controlpointcount").SetText(sdpString(m_ControlPointCount));

	for(gulong i = 0; i < MAXKNOTCOUNT; i++)
		{
			Entry(sdpString(i)).SetText(sdpString(m_Knots[i]));
			Widget(sdpString(i)).SetSensitive(i < KnotCount());
		}
}

void SampleWindow::OnEvent(sdpGtkEvent* Event)
{
	// Handle events ...
	if(Event->Name() == "ondestroy" || Event->Name() == "onexit")
		Exit();
	else if(Event->Name() == "onconfigureevent" || Event->Name() == "onexposeevent")
		OnDraw();
	else if(Event->Name() == "ontolerance")
		OnTolerance();
	else if(Event->Name() == "onorder")
		OnOrder();
	else if(Event->Name() == "oncontrolpointcount")
		OnControlPointCount();
	else if(Event->Name() == "onperiodicuniform")
		OnPeriodicUniform();
	else if(Event->Name() == "onopenuniform")
		OnOpenUniform();
	else if(Event->Name() == "onbezier")
		OnBezier();
	else if(Event->Name() == "onposition")
		OnPosition();
	else if(Event->Name() == "onrational")
		OnRational();
	else if(Event->Name().isnumeric())
		OnKnot(Event->Name());
	else
		sdpGtkObjectContainer::OnEvent(Event);
}

void SampleWindow::OnTolerance()
{
	m_Tolerance = atof(Entry("tolerance").GetText());
	Redraw();
}

void SampleWindow::OnOrder()
{
	m_Order = sdpMin<gulong>(atoi(Entry("order").GetText()), ControlPointCount());
	MakeOpenUniform();
	UpdateControls();
	Redraw();	
}

void SampleWindow::OnControlPointCount()
{
	m_ControlPointCount = atoi(Entry("controlpointcount").GetText());
	m_Order = sdpMin(m_Order, m_ControlPointCount);
	MakeOpenUniform();
	UpdateControls();
	Redraw();	
}

void SampleWindow::OnPeriodicUniform()
{
	MakePeriodicUniform();
	UpdateControls();
	Redraw();
}

void SampleWindow::OnOpenUniform()
{
	MakeOpenUniform();
	UpdateControls();
	Redraw();
}

void SampleWindow::OnBezier()
{
	MakeBezier();
	UpdateControls();
	Redraw();
}

void SampleWindow::MakePeriodicUniform()
{
	for(gulong i = 0; i < MAXKNOTCOUNT; i++)
		m_Knots[i] = i;
}

void SampleWindow::MakeOpenUniform()
{
	gulong index = 0;
	gulong value = 0;
	
	while(index < Order())
		m_Knots[index++] = value;

	value++;
	while(index < KnotCount() - Order())
		m_Knots[index++] = value++;
		
	while(index < KnotCount())
		m_Knots[index++] = value;
}

void SampleWindow::MakeBezier()
{
	m_Order = ControlPointCount();
	MakeOpenUniform();
}

void SampleWindow::OnPosition()
{
	m_ControlPoints[4][1] = 5 * sin(sdpRadians<gdouble>(90 - Adjustment("position").Value()));
	OnDraw();
}

void SampleWindow::OnRational()
{
	for(gulong i = 0; i < MAXCONTROLPOINTCOUNT; i++)
		m_ControlPoints[i][3] = Adjustment("rational").Value();
	OnDraw();
}

void SampleWindow::OnKnot(const sdpString& Knot)
{
	m_Knots[atoi(Knot)] = atof(Entry(Knot).GetText());
	Redraw();
}

void SampleWindow::Exit()
{
	DisconnectAllEvents();
	gtk_main_quit();
}

void nurbsError(GLenum errorCode)
{
	std::cerr << "NURBS error: " << gluErrorString(errorCode) << std::endl;
}

void SampleWindow::OnDraw()
{
	// If we haven't been properly initialized, yet, we're done ...
	if(!m_OpenGLWidget.Initialized())
		return;

	// Start OpenGL rendering ...
	m_OpenGLWidget.Begin();
	
	// Set the viewport to the full size size of the window ...
	GLint width = m_OpenGLWidget.Width();
	GLint height = m_OpenGLWidget.Height();
	glViewport(0, 0, width, height);

	// Setup a perspective projection ...
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(-1, 12, -1, 6, -1, 1);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	// Clear the buffer ...
	glClearDepth(1.0);
	glClearColor(0.0, 0.0, 0.0, 0.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// Setup odds-and-ends ...
	glDisable(GL_LIGHTING);
	glDisable(GL_DEPTH_TEST);
	glEnable(GL_MAP1_VERTEX_4);
	
	// Draw the axes ...
	glColor3d(0.5, 0.0, 0.0);

	glBegin(GL_LINES);
	glVertex2d(-1000, 0);
	glVertex2d(1000, 0);
	glVertex2d(0, -1000);
	glVertex2d(0, 1000);
	glEnd();

	// Draw the control point hull ...
	glColor3d(0.5, 0.5, 0);
	glBegin(GL_LINE_STRIP);
	for(gulong i = 0; i < m_ControlPointCount; i++)
		glVertex3fv(m_ControlPoints[i]);
	glEnd();

	// Draw control points ...
	glColor3d(1, 1, 0);
	glPointSize(5.0);
	
	glBegin(GL_POINTS);
	for(gulong j = 0; j < m_ControlPointCount; j++)
		glVertex3fv(m_ControlPoints[j]);
	glEnd();

	// Draw the curve with a Bezier evalutator ...
	glColor3d(0, 1, 0);
	
	glMapGrid1f(30, 0, 1);
	glMap1f(GL_MAP1_VERTEX_4, 0, 1, 4, m_ControlPointCount, &m_ControlPoints[0][0]);
	glEvalMesh1(GL_LINE, 0, 30);	

	// Draw the curve using NURBS ...
	glColor3d(1, 1, 1);

	GLUnurbs* nurb = gluNewNurbsRenderer();
//	gluNurbsProperty(nurb, GLU_SAMPLING_METHOD, GLU_PARAMETRIC_ERROR);
	gluNurbsProperty(nurb, GLenum(GLU_SAMPLING_TOLERANCE), m_Tolerance);
	gluNurbsCallback(nurb, GLenum(GLU_ERROR), (GLvoid (CALLBACK*) ()) nurbsError);

	gluBeginCurve(nurb);
	gluNurbsCurve(nurb, KnotCount(), &m_Knots[0], 4, &m_ControlPoints[0][0], m_Order, GL_MAP1_VERTEX_4);
	gluEndCurve(nurb);

	gluDeleteNurbsRenderer(nurb);

	glFlush();

	// Finish OpenGL rendering ...
	m_OpenGLWidget.SwapBuffers();
	m_OpenGLWidget.End();
}

void SampleWindow::Redraw()
{
	m_OpenGLWidget.QueueDraw();
}

//////////////////////////////////////////////////////////////////////////////////
// main

/// Initializes GTK+ and loads the main window
gint main(gint argc, gchar* argv[])
{
	if(!sdpglExists(std::cerr))
		return 1;
	
	gtk_init(&argc, &argv);
	
	SampleWindow window;
	window.Load();

	gtk_main();

	return 0;
}


