// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the SDPGTK sample application, which demonstrates the SDPGTK widgets
		\author Tim Shead (tshead@k-3d.com)
*/

#include <sdpgtkobjectcontainer.h>
#include <sdpgtkevents.h>

#include <sdpmath.h>
#include <sdputility.h>

#include <iostream>
#include <fstream>
#include <stdlib.h>
#include <vector>

#include <stdio.h>

#if defined SDPWIN32
#include <gdk/win32/gdkwin32.h>
#else // SDPWIN32
#include <gdk/gdkx.h>
#include <X11/X.h>
#endif // !SDPWIN32

#ifdef RootWindow
#undef RootWindow
#endif // RootWindow

// Some constants ...
const int ImageWidth = 100;
const int ImageHeight = 100;

// sdpGtkObjectContainer is the main SDPGTK class - without alteration,
// it encapsulates all the functionality required to load GTKML documents.
//
// We derive our own classes from it so we can override OnEvent() and
// handle our SDPGTK events.

////////////////////////////////////////////////////////////////////////
// AboutBox

/// Creates a modal "about" box
class AboutBox : public sdpGtkObjectContainer
{
protected:
	void OnEvent(sdpGtkEvent* Event);
};

/// Overrides sdpGtkObjectContainer::OnEvent() to handle our own events
void AboutBox::OnEvent(sdpGtkEvent* Event)
{
	// Handle events ...
	if(Event->Name() == "ondestroy")
		CancelModal();
	else if(Event->Name() == "onok")
		RootWidget().Destroy();
	else
		sdpGtkObjectContainer::OnEvent(Event);
}

/////////////////////////////////////////////////////////////////////////
// ModalDialog

/// A general-purpose modal dialog with event-handling for an "OK" and a "Cancel" button
class ModalDialog : public sdpGtkObjectContainer
{
public:
	ModalDialog();

	/// Loads the dialog template and starts the modal loop
	void DoModal(GtkWindow* Window = 0);

	/// Returns true if the user closed the dialog with the "OK" button
	bool OK();

	/// Returns true if the user closed the dialog with the "Cancel" or window manager "Close" button
	bool Cancelled();

protected:
	/// Override this in derived classes to load/generate the dialog template
	virtual bool Load() = 0;

	/// Overrides sdpGtkObjectContainer::OnEvent() to handle our built-in events
	void OnEvent(sdpGtkEvent* Event);

	/// Called when the GTK+ object is about to be destroyed
	virtual void OnDestroy();
	/// Called when the user hits the "OK" button
	virtual void OnOK();
	/// Called when the user hits the "Cancel" button
	virtual void OnCancel();

	/// Records whether the user hit "OK" or "Cancel"
	bool m_Result;
};

ModalDialog::ModalDialog() : sdpGtkObjectContainer()
{
	m_Result = false;
}

bool ModalDialog::OK()
{
	return m_Result;
}

bool ModalDialog::Cancelled()
{
	return !m_Result;
}

void ModalDialog::DoModal(GtkWindow* Window)
{
	g_return_if_fail(Load());
	
	if(Window)
		sdpGtkObjectContainer::DoModal(Window);
	else
		sdpGtkObjectContainer::DoModal();
}

void ModalDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "ondestroy")
		OnDestroy();
	else if(Event->Name() == "onok")
		OnOK();
	else if(Event->Name() == "oncancel")
		OnCancel();
	else
		sdpGtkObjectContainer::OnEvent(Event);
};

void ModalDialog::OnDestroy()
{
	CancelModal();
}

void ModalDialog::OnOK()
{
	m_Result = true;

	RootWidget().Destroy();
}

void ModalDialog::OnCancel()
{
	m_Result = false;

	RootWidget().Destroy();
}


//////////////////////////////////////////////////////////////////////////
// ColorSelector

/// A modal color selection dialog
class ColorSelector : public ModalDialog
{
public:
	/// Returns the color chosen by the user
	const sdpVector4& Color() { return m_Color; }
	
protected:
	/// Loads a GtkColorSelectionDialog using a static GTKML template and runtime event mappings
	bool Load();
	/// Records the chosen color when the user hits the "OK" button
	void OnOK();
	
	/// Stores the color the user chose as R,G,B,A
	sdpVector4 m_Color;
};

bool ColorSelector::Load()
{
	// Load the GTKML document ...
	sdpString dialogtemplate="<gtkml><colorselectiondialog/></gtkml>\n";
	g_return_val_if_fail(sdpGtkObjectContainer::Load(dialogtemplate, ""), false);

	// Map some events ...
	MapEvent("destroy", "ondestroy", RootObject(), true);
	MapEvent("clicked", "onok", RootColorSelectionDialog().OKButton(), true);
	MapEvent("clicked", "oncancel", RootColorSelectionDialog().CancelButton(), true);

	return true;
}

void ColorSelector::OnOK()
{
	RootColorSelectionDialog().Selection().GetColor(m_Color);
	ModalDialog::OnOK();
}

//////////////////////////////////////////////////////////////////////////
// FileSelector

/// A modal file selection dialog
class FileSelector : public ModalDialog
{
public:
	/// Returns the filepath chosen by the user
	SDPCSTRING FilePath() { return m_FilePath; }
	
protected:
	/// Loads a GtkFileSelection dialog using a static GTKML template and runtime event mappings
	bool Load();

	/// Records the chosen filepath when the user hits the "OK" button
	void OnOK();
	
	/// Stores the filepath the user chose
	sdpString m_FilePath;
};

bool FileSelector::Load()
{
	// Load the GTKML document ...
	sdpString dialogtemplate="<gtkml><fileselection/></gtkml>\n";
	g_return_val_if_fail(sdpGtkObjectContainer::Load(dialogtemplate, ""), false);

	// Map some events ...
	MapEvent("destroy", "ondestroy", RootObject(), true);
	MapEvent("clicked", "onok", RootFileSelection().OKButton(), true);
	MapEvent("clicked", "oncancel", RootFileSelection().CancelButton(), true);

	return true;
}

void FileSelector::OnOK()
{
	m_FilePath = RootFileSelection().GetFilePath();
	ModalDialog::OnOK();
}

//////////////////////////////////////////////////////////////////////////
// FontSelector

/// A modal font selection dialog
class FontSelector : public ModalDialog
{
public:
	/// Returns the font string chosen by the user
	SDPCSTRING Font() { return m_Font; }
	
protected:
	/// Creates a GtkFontSelectionDialog using a static GTKML template and runtime event mappings
	bool Load();
	void OnOK();
	
	sdpString m_Font;
};

bool FontSelector::Load()
{
	// Load the GTKML document ...
	sdpString dialogtemplate="<gtkml><fontselectiondialog/></gtkml>\n";
	g_return_val_if_fail(sdpGtkObjectContainer::Load(dialogtemplate, ""), false);

	// Map some events ...
	MapEvent("destroy", "ondestroy", RootObject(), true);
	MapEvent("clicked", "onok", RootFontSelectionDialog().OKButton(), true);
	MapEvent("clicked", "oncancel", RootFontSelectionDialog().CancelButton(), true);

	return true;
}

void FontSelector::OnOK()
{
	m_Font = RootFontSelectionDialog().Selection().GetFontName();
	ModalDialog::OnOK();
}

//////////////////////////////////////////////////////////////////////////
// MessageBox

/// A modal dialog for displaying short messages
class MessageBox : public ModalDialog
{
public:
	MessageBox(SDPCSTRING Title, SDPCSTRING Message);
protected:
	/// Creates a top-level GtkWindow using a static GTKML template with built-in event handling
	bool Load();

	/// Stores the message box titlebar string
	sdpString m_Title;
	/// Stores the message string
	sdpString m_Message;
};

MessageBox::MessageBox(SDPCSTRING Title, SDPCSTRING Message)
{
	m_Title = Title;
	m_Message = Message;
}

bool MessageBox::Load()
{
	// Load the GTKML document ...
	sdpString dialogtemplate=
	"<gtkml><window name=\"window\" show=\"false\"><event signal=\"destroy\" name=\"ondestroy\"/><vbox homogeneous=\"false\">"
		"<label expand=\"true\" fill=\"true\" name=\"label\">Message</label>"
		"<button>OK<event signal=\"clicked\" name=\"onok\"/></button>"
	"</vbox></window></gtkml>\n";

	g_return_val_if_fail(sdpGtkObjectContainer::Load(dialogtemplate, ""), false);

	Window("window").SetTitle(m_Title);
	Label("label").SetText(m_Message);

	RootWidget().Show();

	return true;
}


//////////////////////////////////////////////////////////////////////////
// SampleWindow

/// The main window for the application
class SampleWindow : public sdpGtkObjectContainer
{
public:
	SampleWindow();

	/// Loads the GTKML template
	void Load();

protected:
	/// Handles events for this template
	void OnEvent(sdpGtkEvent* Event);

	/// The timer was signalled
	void OnTimer();
	/// User requested an "about" box
	void OnAbout();
	/// Demonstrates how and sdpGtkObjectContainer can be "reused" by reloading a GTKML template
	void OnReload();
	/// Displays a modal color selection dialog
	void OnColorSelector();
	/// Displays a modal file selection dialog
	void OnFileSelector();
	/// Displays a modal font selection dialog
	void OnFontSelector();
	/// Uses a GtkTipsQuery widget to provide interactive help
	void OnTipsQuery();
	/// Tests functions for warping the pointer (yes, this is evil)
	void OnPointerWarp();
	/// Tests functions for simulating a button press (also evil)
	void OnButtonPress();
	/// Try running a client application, to test the plug/socket implementation
	void RunClient();
	/// Try capturing an existing X window
	void Capture();
	
	/// Records whether the template is being reloaded
	bool m_Reloading;
	/// A GtkStatusBar context for displaying messages in the statusbar
	guint m_StatusbarContext;
	/// Holds the state of our flashing label (responds to timer events)
	bool m_ShowLabel;
};

SampleWindow::SampleWindow()
{
	m_Reloading = false;
	m_StatusbarContext = 0;
	m_ShowLabel = false;
}

void SampleWindow::Load()
{
	// Load the GTKML document ...
	sdpGtkObjectContainer::LoadFile("mainwindow.gtkml");

	gtk_widget_set_events(RootWidget(), GDK_ALL_EVENTS_MASK);
	gtk_widget_set_events(Text("text"), GDK_ALL_EVENTS_MASK);

	// Put some items into our ComboBox ...
	sdpGtkCombo combo = Combo("combo1");
	sdpStringArray list;
	for(gint i = 0; i < 10; i++)
		list.push_back(sdpFormatString("%d", i));
	combo.SetList(list);
	combo.SetEntryText("Howdy, World!");

	// Put some stuff into our CList ...
	sdpGtkCList clist = CList("clist");

	clist.SetColumnTitle(0, "ID");
	clist.SetColumnTitle(1, "Name");
	clist.SetColumnTitle(2, "Address");
	clist.ShowColumnTitles();

	SDPCSTRING clistdata[] = {"0123", "Eric", "123 Main Street"};
	for(gint j = 0; j < 15; j++)
		clist.Append(clistdata);

	// Put some stuff into our CTree ...
	sdpGtkCTree ctree = CTree("ctree");

	ctree.SetColumnTitle(0, "Score");
	ctree.SetColumnTitle(1, "Sequence");
	ctree.SetColumnTitle(2, "Family");
	ctree.ShowColumnTitles();

	SDPCSTRING ctreedata1[] = {"1e-10", "gattacagacttcaatg", "gcr"};
	GtkCTreeNode* node = ctree.InsertNode(0, 0, ctreedata1);

	SDPCSTRING ctreedata2[] = {"1e-20", "atacattagtaca", "gcrp"};
	ctree.InsertNode(node, 0, ctreedata2);

	SDPCSTRING ctreedata3[] = {"1e-30", "tagctagttaccgga", "gcrp"};
	ctree.InsertNode(0, 0, ctreedata3);

	// Put our document into the status-bar ...
	Statusbar("statusbar").Push(Statusbar("statusbar").GetContextID("sample"), "Loaded mainwindow.gtkml");

	// Assign the edit/cut menu item to our GtkAccelLabel ...
//	AccelLabel("accellabel").SetAcceleratorWidget(Widget("cutmenuitem"));
//	AccelLabel("accellabel").RefetchAccelerator();

	// Create a GtkImage and stuff it into our container ...
	GdkImage* gdkimage = gdk_image_new(GDK_IMAGE_FASTEST, gdk_visual_get_best(), ImageWidth, ImageHeight);
	
	sdpGtkImage image;
	image.Create(gdkimage, 0);
	image.Show();
	
	Frame("imagecontainer").Attach(image);
	
	MapObject("image", image);
	
	// Make everything visible ...
	RootWidget().Show();
}

void SampleWindow::OnEvent(sdpGtkEvent* Event)
{
	// Handle events ...
	if(Event->Name() == "ondestroy" && !m_Reloading)
		gtk_main_quit();
	else if(Event->Name() == "onexit")
		gtk_main_quit();
	else if(Event->Name() == "onabout")
		OnAbout();
	else if(Event->Name() == "onreload")
		{
			OnReload();
			return;
		}
	else if(Event->Name() == "oncolorselector")
		OnColorSelector();
	else if(Event->Name() == "onfileselector")
		OnFileSelector();
	else if(Event->Name() == "onfontselector")
		OnFontSelector();
	else if(Event->Name() == "ontimer")
		{
			OnTimer();
			return;
		}
	else if(Event->Name() == "ontipsquery")
		OnTipsQuery();
	else if(Event->Name() == "onpointerwarp")
		{
			OnPointerWarp();
			return;
		}
	else if(Event->Name() == "onbuttonpress")
		{
			OnButtonPress();
			return;
		}
	else if(Event->Name() == "runclient")
		RunClient();
	else if(Event->Name() == "capture")
		Capture();

	// Put the event in the statusbar ...
	Statusbar("statusbar").Push(Statusbar("statusbar").GetContextID("sample"), sdpFormatString("Event: %s", SDPCSTRING(Event->Name())));
}

void SampleWindow::OnAbout()
{
	AboutBox aboutbox;
	aboutbox.LoadFile("about.gtkml");
	aboutbox.DoModal();
}

void SampleWindow::OnReload()
{
	m_Reloading = true;

	// Kill all of our events/timers ...
	DisconnectAllEvents();

	// Destroy the GTK widget hierarchy ...
	RootWidget().Destroy();

	// Reload the GTKML file (will automatically clear out existing SDPGTK structures) ...
	Load();

	m_Reloading = false;
}

void SampleWindow::OnColorSelector()
{
	ColorSelector selector;
	selector.DoModal();
	if(selector.OK())
		{
			MessageBox message("Color:", sdpString(selector.Color()));
			message.DoModal();
		}
}

void SampleWindow::OnFileSelector()
{
	FileSelector selector;
	selector.DoModal((GtkWindow*)RootWindow());
	if(selector.OK())
		{
			MessageBox message("File:", selector.FilePath());
			message.DoModal();
		}
}

void SampleWindow::OnFontSelector()
{
	FontSelector selector;
	selector.DoModal();
	if(selector.OK())
		{
			MessageBox message("Font:", selector.Font());
			message.DoModal();
		}
}

void SampleWindow::OnTimer()
{
	// Make our welcome page blink real friendly-like ...
	sdpGtkLabel label = Label("howdylabel");

	if(m_ShowLabel)
		label.Show();
	else
		label.Hide();

	m_ShowLabel = !m_ShowLabel;
	
	// Add random junk to our image to demonstrate that it's client-side ...
	GdkImage* image = 0;
	GdkBitmap* mask = 0;
	Image("image").GetImage(&image, &mask);

	for(int i = 0; i < 50; i++)	
		gdk_image_put_pixel(image, rand() % ImageWidth, rand() % ImageHeight, rand());
}

void SampleWindow::OnTipsQuery()
{
	TipsQuery("tipsquery").SetCaller(Button("tipsquerybutton"));
	TipsQuery("tipsquery").StartQuery();
}

void SampleWindow::OnPointerWarp()
{
	// Get the X display ...
	Display* display = (Display*)GDK_DISPLAY();
	g_return_if_fail(display);

	for(gdouble theta = 0; theta < sdpPiTimes2; theta += 0.1)
		{
			XWarpPointer(display, None, None, 0, 0, 0, 0, int(10*cos(theta)), int(10*sin(theta)));
			XFlush(display);
			sdpSleep(10);
		}
}

#include <sdpgtkutility.h>
void SampleWindow::OnButtonPress()
{
	Button("testbutton").InteractiveWarpPointer();
	Button("testbutton").FakeInteractiveActivate();
	
	MenuItem("menu_edit").InteractiveWarpPointer();
	MenuItem("menu_edit").InteractiveSelect();
	
	MenuItem("menu_edit_copy").InteractiveWarpPointer(0.25, 0.5);
	MenuItem("menu_edit_copy").InteractiveActivate();
	MenuItem("menu_edit_copy").Deselect();
		
	MenuItem("menu_edit").Deselect();
}

void SampleWindow::RunClient()
{
#ifdef SDPWIN32
#define popen _popen
#endif // SDPWIN32

	guint socketid = Socket("socket").SocketID();
	popen(sdpFormatString("./client %d", socketid), "r");
}

void SampleWindow::Capture()
{
	guint xid = atol(Entry("xid").GetText());
	Socket("socket").Steal(xid);
}

//////////////////////////////////////////////////////////////////////////////////
// main

/// Initializes GTK+ and loads the main window
gint main(gint argc, gchar* argv[])
{
	gtk_init(&argc, &argv);

	SampleWindow window;
	window.Load();

	gtk_main();

	return 0;
}


