
// SDPGTK Library
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <sdpgtkwidgets.h>
#include <sdpgtkutility.h>
#include <sdpxmlutility.h>

#include <iostream>

///////////////////////////////////////////////////////////////////////////////////
// sdpGtkObject

sdpGtkObject::sdpGtkObject()
{
	m_Object = 0;
}

sdpGtkObject::sdpGtkObject(GtkObject* Object)
{
	m_Object = Object;
}

sdpGtkObject::~sdpGtkObject()
{
}

bool sdpGtkObject::Attached() const
{
	return m_Object ? true : false;
}

const sdpString sdpGtkObject::CustomType() const
{
	return sdpString("sdpGtkObject");
}

bool sdpGtkObject::Create(sdpGtkIObjectContainer* ObjectContainer, sdpXmlIElement* Element)
{
	// This base-class implementation should never be called ...
	std::cerr << "sdpGtkObject::Create() should never be called!" << std::endl;
	return true;
}

bool sdpGtkObject::CreateOptions(sdpGtkIObjectContainer* ObjectContainer, sdpXmlIElement* Element)
{
	return true;
}

static void sdpGtkLoadObject(sdpGtkObject& ParentObject, sdpGtkIObjectContainer* ObjectContainer, sdpXmlIElement* Element, sdpGtkObject& NewObject)
{
	// Sanity checks ...
	g_assert(ObjectContainer);
	g_assert(Element);

	// Handle control options ...
	NewObject.CreateOptions(ObjectContainer, Element);

	// Notify the object container about the new control ...
	ObjectContainer->NotifyObject(NewObject.Object());

	// If the control has a name, tell the object container to create a mapping ...
	sdpGtkMarkAttribute(Element, "name");
	sdpXmlIAttribute* name = sdpXmlFindAttribute(Element, "name");
	if(name)
		ObjectContainer->MapObject(name->Value(), NewObject.Object());

	// Recursively create children ...
	NewObject.CreateChildren(ObjectContainer, Element);
	
	// Attach the object to its parent ...
	ParentObject.AttachChild(ObjectContainer, Element, &NewObject);
}

static void sdpGtkLoadCustomObject(sdpGtkObject& ParentObject, sdpGtkIObjectContainer* ObjectContainer, sdpXmlIElement* Element, sdpGtkObject& NewObject)
{
	// Sanity checks ...
	g_assert(ObjectContainer);
	g_assert(Element);

	// Handle control options ...
	NewObject.CreateOptions(ObjectContainer, Element);

	// Notify the object container about the new control ...
	ObjectContainer->NotifyObject(NewObject.Object());

	// If the control has a name, tell the object container to create a mapping ...
	sdpGtkMarkAttribute(Element, "name");
	sdpXmlIAttribute* name = sdpXmlFindAttribute(Element, "name");
	if(name)
		ObjectContainer->MapCustomObject(name->Value(), &NewObject);

	// Recursively create children ...
	NewObject.CreateChildren(ObjectContainer, Element);
	
	// Attach the object to its parent ...
	ParentObject.AttachChild(ObjectContainer, Element, &NewObject);
}

// Once again, the prototype for this template function was *supposed* to be
// "static void", but the !@#$!$ MSVC++ 6 compiler can't seem to properly instantiate
// a template function unless the prototype contains the template arguments somewhere,
// even when you explicitly specify the template arguments in the function call (see below) ...
template<class Type>
static Type sdpGtkCreateObject(sdpGtkObject& ParentObject, sdpGtkIObjectContainer* ObjectContainer, sdpXmlIElement* Element)
{
	// Create the object ...
	Type object;
	if(!object.Create(ObjectContainer, Element))
		{
			std::cout << "sdpGtk: Unable to create <" << Element->Name() << ">" << std::endl;
			g_return_val_if_fail(0, object);
		}
	
	// Give it a chance to load itself and its children ...
	sdpGtkLoadObject(ParentObject, ObjectContainer, Element, object);

	return object;
}

static void sdpGtkCreateAccelerator(sdpGtkIObjectContainer* ObjectContainer, sdpXmlIElement* Element, GtkObject* Object)
{
	// Sanity checks ...
	g_assert(ObjectContainer);
	g_assert(Element);
	g_assert(Object);

	// Get the accelerator signal ...
	sdpString signal;
	sdpGtkMarkAttribute(Element, "signal");
	sdpXmlGetAttribute(Element, "signal", signal);

	// Get the accelerator key ...
	sdpString key;
	sdpGtkMarkAttribute(Element, "key");
	sdpXmlGetAttribute(Element, "key", key);

	// Get the accelerator modifiers ...
	sdpString modifierstring;
	sdpGtkMarkAttribute(Element, "modifiers");
	sdpXmlGetAttribute(Element, "modifiers", modifierstring);

	guint modifiers = 0;
	if(modifierstring.find("shift") != sdpString::npos)
		modifiers |= GDK_SHIFT_MASK;
	if(modifierstring.find("lock") != sdpString::npos)
		modifiers |= GDK_LOCK_MASK;
	if(modifierstring.find("control") != sdpString::npos)
		modifiers |= GDK_CONTROL_MASK;
	if(modifierstring.find("mod1") != sdpString::npos)
		modifiers |= GDK_MOD1_MASK;
	if(modifierstring.find("mod2") != sdpString::npos)
		modifiers |= GDK_MOD2_MASK;
	if(modifierstring.find("mod3") != sdpString::npos)
		modifiers |= GDK_MOD3_MASK;
	if(modifierstring.find("mod4") != sdpString::npos)
		modifiers |= GDK_MOD4_MASK;
	if(modifierstring.find("mod5") != sdpString::npos)
		modifiers |= GDK_MOD5_MASK;
	if(modifierstring.find("button1") != sdpString::npos)
		modifiers |= GDK_BUTTON1_MASK;
	if(modifierstring.find("button2") != sdpString::npos)
		modifiers |= GDK_BUTTON2_MASK;
	if(modifierstring.find("button3") != sdpString::npos)
		modifiers |= GDK_BUTTON3_MASK;
	if(modifierstring.find("button4") != sdpString::npos)
		modifiers |= GDK_BUTTON4_MASK;
	if(modifierstring.find("button5") != sdpString::npos)
		modifiers |= GDK_BUTTON5_MASK;
	if(modifierstring.find("release") != sdpString::npos)
		modifiers |= GDK_RELEASE_MASK;

	// Get the accelerator flags ...
	bool visible = true;
	sdpGtkMarkAttribute(Element, "visible");
	sdpXmlGetAttribute(Element, "visible", visible);

	GtkAccelFlags flags = GtkAccelFlags(visible ? GTK_ACCEL_VISIBLE : 0);

	// Get the current accelerator group ...
	GtkAccelGroup* group = ObjectContainer->GetAcceleratorGroup();
	g_assert(group);

	gtk_widget_add_accelerator(GTK_WIDGET(Object), signal, group, key[0], modifiers, flags);
}

static void sdpGtkCreateResource(sdpGtkIObjectContainer* ObjectContainer, sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_assert(ObjectContainer);
	g_assert(Element);

	sdpString resourcename;
	sdpXmlGetAttribute(Element, "name", resourcename);
	sdpGtkMarkAttribute(Element, "name");
	if(0 == resourcename.size())
		{
			std::cerr << "sdpGtk: Anonymous resources not allowed" << std::endl;
			return;
		}

	sdpString resourcevalue;
	sdpXmlGetAttribute(Element, "value", resourcevalue);
	sdpGtkMarkAttribute(Element, "value");
				
	ObjectContainer->MapResource(resourcename, resourcevalue);
}

bool sdpGtkObject::CreateChildren(sdpGtkIObjectContainer* ObjectContainer, sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_assert(ObjectContainer);
	g_assert(Element);

	// For each child element ...
	for(SDPXMLELEMENTHANDLE e = Element->GetFirstElement(); e; e = Element->GetNextElement(e))
		{
			// Get the element name ...
			sdpXmlIElement* childelement = Element->GetElement(e);
			sdpString name = childelement->Name();

			// If it's a control, create it ...
			if(name == "accellabel") { sdpGtkCreateObject<sdpGtkAccelLabel>(*this, ObjectContainer, childelement); continue; }
			if(name == "adjustment") { sdpGtkCreateObject<sdpGtkAdjustment>(*this, ObjectContainer, childelement); continue; }
			if(name == "alignment") { sdpGtkCreateObject<sdpGtkAlignment>(*this, ObjectContainer, childelement); continue; }
			if(name == "arrow") { sdpGtkCreateObject<sdpGtkArrow>(*this, ObjectContainer, childelement); continue; }
			if(name == "aspectframe") { sdpGtkCreateObject<sdpGtkAspectFrame>(*this, ObjectContainer, childelement); continue; }
			if(name == "button") { sdpGtkCreateObject<sdpGtkButton>(*this, ObjectContainer, childelement); continue; }
			if(name == "calendar") { sdpGtkCreateObject<sdpGtkCalendar>(*this, ObjectContainer, childelement); continue; }
			if(name == "checkbutton") { sdpGtkCreateObject<sdpGtkCheckButton>(*this, ObjectContainer, childelement); continue; }
			if(name == "checkmenuitem") { sdpGtkCreateObject<sdpGtkCheckMenuItem>(*this, ObjectContainer, childelement); continue; }
			if(name == "clist") { sdpGtkCreateObject<sdpGtkCList>(*this, ObjectContainer, childelement); continue; }
			if(name == "colorselection") { sdpGtkCreateObject<sdpGtkColorSelection>(*this, ObjectContainer, childelement); continue; }
			if(name == "colorselectiondialog") { sdpGtkCreateObject<sdpGtkColorSelectionDialog>(*this, ObjectContainer, childelement); continue; }
			if(name == "combo") { sdpGtkCreateObject<sdpGtkCombo>(*this, ObjectContainer, childelement); continue; }
			if(name == "ctree") { sdpGtkCreateObject<sdpGtkCTree>(*this, ObjectContainer, childelement); continue; }
			if(name == "curve") { sdpGtkCreateObject<sdpGtkCurve>(*this, ObjectContainer, childelement); continue; }
			if(name == "dialog") { sdpGtkCreateObject<sdpGtkDialog>(*this, ObjectContainer, childelement); continue; }
			if(name == "drawingarea") { sdpGtkCreateObject<sdpGtkDrawingArea>(*this, ObjectContainer, childelement); continue; }
			if(name == "entry") { sdpGtkCreateObject<sdpGtkEntry>(*this, ObjectContainer, childelement); continue; }
			if(name == "eventbox") { sdpGtkCreateObject<sdpGtkEventBox>(*this, ObjectContainer, childelement); continue; }
			if(name == "fileselection") { sdpGtkCreateObject<sdpGtkFileSelection>(*this, ObjectContainer, childelement); continue; }
			if(name == "fixed") { sdpGtkCreateObject<sdpGtkFixed>(*this, ObjectContainer, childelement); continue; }
			if(name == "fontselection") { sdpGtkCreateObject<sdpGtkFontSelection>(*this, ObjectContainer, childelement); continue; }
			if(name == "fontselectiondialog") { sdpGtkCreateObject<sdpGtkFontSelectionDialog>(*this, ObjectContainer, childelement); continue; }
			if(name == "frame") { sdpGtkCreateObject<sdpGtkFrame>(*this, ObjectContainer, childelement); continue; }
			if(name == "gammacurve") { sdpGtkCreateObject<sdpGtkGammaCurve>(*this, ObjectContainer, childelement); continue; }
			if(name == "handlebox") { sdpGtkCreateObject<sdpGtkHandleBox>(*this, ObjectContainer, childelement); continue; }
			if(name == "hbox") { sdpGtkCreateObject<sdpGtkHBox>(*this, ObjectContainer, childelement); continue; }
			if(name == "hbuttonbox") { sdpGtkCreateObject<sdpGtkHButtonBox>(*this, ObjectContainer, childelement); continue; }
			if(name == "hpaned") { sdpGtkCreateObject<sdpGtkHPaned>(*this, ObjectContainer, childelement); continue; }
			if(name == "hruler") { sdpGtkCreateObject<sdpGtkHRuler>(*this, ObjectContainer, childelement); continue; }
			if(name == "hscale") { sdpGtkCreateObject<sdpGtkHScale>(*this, ObjectContainer, childelement); continue; }
			if(name == "hscrollbar") { sdpGtkCreateObject<sdpGtkHScrollbar>(*this, ObjectContainer, childelement); continue; }
			if(name == "hseparator") { sdpGtkCreateObject<sdpGtkHSeparator>(*this, ObjectContainer, childelement); continue; }
//			if(name == "image") { sdpGtkCreateObject<sdpGtkImage>(*this, ObjectContainer, childelement); continue; }
			if(name == "inputdialog") { sdpGtkCreateObject<sdpGtkInputDialog>(*this, ObjectContainer, childelement); continue; }
			if(name == "label") { sdpGtkCreateObject<sdpGtkLabel>(*this, ObjectContainer, childelement); continue; }
			if(name == "list") { sdpGtkCreateObject<sdpGtkList>(*this, ObjectContainer, childelement); continue; }
			if(name == "listitem") { sdpGtkCreateObject<sdpGtkListItem>(*this, ObjectContainer, childelement); continue; }
			if(name == "menu") { sdpGtkCreateObject<sdpGtkMenu>(*this, ObjectContainer, childelement); continue; }
			if(name == "menubar") { sdpGtkCreateObject<sdpGtkMenuBar>(*this, ObjectContainer, childelement); continue; }
			if(name == "menuitem") { sdpGtkCreateObject<sdpGtkMenuItem>(*this, ObjectContainer, childelement); continue; }
			if(name == "notebook") { sdpGtkCreateObject<sdpGtkNotebook>(*this, ObjectContainer, childelement); continue; }
			if(name == "optionmenu") { sdpGtkCreateObject<sdpGtkOptionMenu>(*this, ObjectContainer, childelement); continue; }
			if(name == "pixmap") { sdpGtkCreateObject<sdpGtkPixmap>(*this, ObjectContainer, childelement); continue; }
			if(name == "radiobutton") { sdpGtkCreateObject<sdpGtkRadioButton>(*this, ObjectContainer, childelement); continue; }
			if(name == "radiomenuitem") { sdpGtkCreateObject<sdpGtkRadioMenuItem>(*this, ObjectContainer, childelement); continue; }
			if(name == "scrolledwindow") { sdpGtkCreateObject<sdpGtkScrolledWindow>(*this, ObjectContainer, childelement); continue; }
			if(name == "socket") { sdpGtkCreateObject<sdpGtkSocket>(*this, ObjectContainer, childelement); continue; }
			if(name == "spinbutton") { sdpGtkCreateObject<sdpGtkSpinButton>(*this, ObjectContainer, childelement); continue; }
			if(name == "statusbar") { sdpGtkCreateObject<sdpGtkStatusbar>(*this, ObjectContainer, childelement); continue; }
			if(name == "table") { sdpGtkCreateObject<sdpGtkTable>(*this, ObjectContainer, childelement); continue; }
			if(name == "tearoffmenuitem") { sdpGtkCreateObject<sdpGtkTearoffMenuItem>(*this, ObjectContainer, childelement); continue; }
			if(name == "text") { sdpGtkCreateObject<sdpGtkText>(*this, ObjectContainer, childelement); continue; }
			if(name == "tipsquery") { sdpGtkCreateObject<sdpGtkTipsQuery>(*this, ObjectContainer, childelement); continue; }
			if(name == "togglebutton") { sdpGtkCreateObject<sdpGtkToggleButton>(*this, ObjectContainer, childelement); continue; }
			if(name == "toolbar") { sdpGtkCreateObject<sdpGtkToolbar>(*this, ObjectContainer, childelement); continue; }
			if(name == "tooltips") { sdpGtkCreateObject<sdpGtkTooltips>(*this, ObjectContainer, childelement); continue; }
			if(name == "tree") { sdpGtkCreateObject<sdpGtkTree>(*this, ObjectContainer, childelement); continue; }
			if(name == "treeitem") { sdpGtkCreateObject<sdpGtkTreeItem>(*this, ObjectContainer, childelement); continue; }
			if(name == "vbox") { sdpGtkCreateObject<sdpGtkVBox>(*this, ObjectContainer, childelement); continue; }
			if(name == "vbuttonbox") { sdpGtkCreateObject<sdpGtkVButtonBox>(*this, ObjectContainer, childelement); continue; }
			if(name == "vpaned") { sdpGtkCreateObject<sdpGtkVPaned>(*this, ObjectContainer, childelement); continue; }
			if(name == "vruler") { sdpGtkCreateObject<sdpGtkVRuler>(*this, ObjectContainer, childelement); continue; }
			if(name == "vscale") { sdpGtkCreateObject<sdpGtkVScale>(*this, ObjectContainer, childelement); continue; }
			if(name == "vscrollbar") { sdpGtkCreateObject<sdpGtkVScrollbar>(*this, ObjectContainer, childelement); continue; }
			if(name == "vseparator") { sdpGtkCreateObject<sdpGtkVSeparator>(*this, ObjectContainer, childelement); continue; }
			if(name == "viewport") { sdpGtkCreateObject<sdpGtkViewport>(*this, ObjectContainer, childelement); continue; }
			if(name == "window") { sdpGtkCreateObject<sdpGtkWindow>(*this, ObjectContainer, childelement); continue; }

			// Handle events ...
			if(name == "event") { ObjectContainer->MapEvent(childelement, Object(), false); continue; }

			// Handle accelerators ...
			if(name == "accelerator") { sdpGtkCreateAccelerator(ObjectContainer, childelement, Object()); continue; }
				
			// Handle resources ...
			if(name == "resource") { sdpGtkCreateResource(ObjectContainer, childelement); continue; }

			// Handle custom objects ...
			sdpGtkObject* customobject = ObjectContainer->CreateCustomObject(childelement);
			if(customobject)
				{
					customobject->Create(ObjectContainer, childelement);
					sdpGtkLoadCustomObject(*this, ObjectContainer, childelement, *customobject);
					continue;
				}

			// If we made it this far, it's an unknown widget type!
			std::cout << Element->Document()->FilePath() << ": " << childelement->Line() << ": " << "Unknown widget type <" << name << "> and its' children will not be loaded" << std::endl;
			continue;
		}

	return true;
}

bool sdpGtkObject::AttachChild(sdpGtkIObjectContainer* ObjectContainer, sdpXmlIElement* ChildElement, sdpGtkObject* Child)
{
	return true;
}

