// SDPGTK
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the xmleditor sample app, which uses SDPGTK to create a simple XML editor
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include <sdpgtkfileselector.h>
#include <sdpgtkevents.h>

#include <sdpxmldocument.h>
#include <sdpfile.h>

#include <iostream>
#include <fstream>

/////////////////////////////////////////////////////////////////////////////////////
// CXMLEditor

/// The main window for the application
class CXMLEditor : public sdpGtkObjectContainer
{
public:
	CXMLEditor(SDPCSTRING ApplicationPath);

	/// Loads the GTKML template
	void LoadUI();

	void ClearData();
	
	void LoadDocument(SDPCSTRING FilePath);
	void SaveDocument(SDPCSTRING FilePath);

	void UpdateTitlebar();
	void UpdateTree();
	void UpdateControls();
	void UpdateCurrentRow();

protected:
	sdpGtkCTree Tree();

	/// Handles events for this template
	void OnEvent(sdpGtkEvent* Event);

	void OnFileNew();
	void OnFileOpen();
	void OnFileSave();
	void OnFileSaveAs();
	void OnFileClose();
	void OnEditNewElement();
	void OnEditNewCData();
	void OnEditNewAttribute();
	void OnEditDelete();
	void OnTreeClicked(sdpGtkEvent* Event);
	void OnMarkupName();
	void OnMarkupValue();
	void OnDelete();

	GtkCTreeNode* InsertRow(sdpGtkCTree Tree, GtkCTreeNode* ParentNode, sdpXmlIElement* Element);
	GtkCTreeNode* InsertRow(sdpGtkCTree Tree, GtkCTreeNode* ParentNode, sdpXmlIAttribute* Attribute);
	void InsertChildren(sdpGtkCTree Tree, GtkCTreeNode* ParentNode, sdpXmlIElement* Element);
		
	sdpString m_ApplicationPath;
	sdpXmlDocument m_Document;
	sdpString m_FilePath;
	
	GdkPixmap* m_AttributeIcon;
	GdkBitmap* m_AttributeIconMask;
	GdkPixmap* m_CDataIcon;
	GdkBitmap* m_CDataIconMask;
	GdkPixmap* m_ElementIcon;
	GdkBitmap* m_ElementIconMask;
	
	/// A discriminated union for storing pointers to elements / attributes in the tree
	class Markup
	{
	public:
		Markup(sdpXmlIElement* Element)
		{
			// Sanity checks ...
			g_assert(Element);
		
			Type = ELEMENT;
			Object.Element = Element;
		}
		
		Markup(sdpXmlIAttribute* Attribute)
		{
			// Sanity checks ...
			g_assert(Attribute);
			
			Type = ATTRIBUTE;
			Object.Attribute = Attribute;
		}

		bool IsElement() { return Type == ELEMENT; }
		bool IsAttribute() { return Type == ATTRIBUTE; }
		
		sdpXmlIElement* Element() { return Object.Element; }
		sdpXmlIAttribute* Attribute() { return Object.Attribute; }

	protected:
		typedef enum
		{
			ELEMENT,
			ATTRIBUTE,
		} XMLType;
	
		XMLType Type;
	
		typedef union
		{
			sdpXmlIElement* Element;
			sdpXmlIAttribute* Attribute;
		} XMLObject;
		
		XMLObject Object;
	};
	
	/// Stores a list of markup objects
	typedef std::list<Markup> Markups;
	typedef Markups::iterator MarkupIterator;
	Markups m_Markups;

	/// Stores the currently selected tree row (if any)	
	GtkCTreeNode* m_CurrentRow;
	Markup* CurrentMarkup();

	/// Returns the collection of currently selected markup (if any)
	typedef std::map<GtkCTreeNode*, Markup*> SelectionMap;
	typedef SelectionMap::iterator SelectionIterator;
	SelectionMap CurrentSelection();
};

CXMLEditor::CXMLEditor(SDPCSTRING ApplicationPath)
{
	// Sanity checks ...
	g_assert_string(ApplicationPath);
	m_ApplicationPath = ApplicationPath;

	m_AttributeIcon = 0;
	m_AttributeIconMask =0;
	m_CDataIcon = 0;
	m_CDataIconMask = 0;
	m_ElementIcon = 0;
	m_ElementIconMask = 0;
	
	m_CurrentRow = 0;
}

void CXMLEditor::LoadUI()
{
	// Load the GTKML document ...
	sdpGtkObjectContainer::LoadFile(sdpAppendPath(m_ApplicationPath, "xmleditor.gtkml"));

	// Load our auxiliary pixmaps ...
	sdpGtkPixmap::CreatePixmap(sdpAppendPath(m_ApplicationPath, "attribute.xpm"), m_AttributeIcon, m_AttributeIconMask);
	sdpGtkPixmap::CreatePixmap(sdpAppendPath(m_ApplicationPath, "cdata.xpm"), m_CDataIcon, m_CDataIconMask);
	sdpGtkPixmap::CreatePixmap(sdpAppendPath(m_ApplicationPath, "element.xpm"), m_ElementIcon, m_ElementIconMask);

	// Update our UI ...
	UpdateTitlebar();
	UpdateTree();
	UpdateControls();

	// Make everything visible ...
	RootWidget().Show();
}

void CXMLEditor::ClearData()
{
	m_FilePath.clear();
	m_Document.Clear();
	m_Markups.clear();
	m_CurrentRow = 0;
}

void CXMLEditor::LoadDocument(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_if_string_fail(FilePath);
	
	// Load the XML document ...
	m_Document.LoadFile(FilePath);	
	
	// Store the path for later ...
	m_FilePath = FilePath;
}

void CXMLEditor::SaveDocument(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_if_string_fail(FilePath);
	
	// Try opening the file ...
	std::ofstream file(FilePath);
	g_return_if_fail(file.good());
	
	// Save the document ...
	file << m_Document;
	
	// Store the path for later ...
	m_FilePath = FilePath;
}

void CXMLEditor::UpdateTitlebar()
{
	if(m_FilePath.size())
		RootWindow().SetTitle("SDPGTK XML Editor: " + m_FilePath);
	else
		RootWindow().SetTitle("SDPGTK XML Editor: Untitled Document");
}

void CXMLEditor::UpdateTree()
{
	// Get the tree ...
	sdpGtkCTree tree = Tree();

	// Start out fresh ...
	tree.Clear();
	m_Markups.clear();
	m_CurrentRow = 0;

	// Freeze updates ...
	tree.Freeze();

	// Start adding elements to the tree ...
	if(m_Document.GetRootElement())
		{
			// Handle children recursively ...
			GtkCTreeNode* node = InsertRow(tree, 0, m_Document.GetRootElement());
			InsertChildren(tree, node, m_Document.GetRootElement());
		}

	// Setup our columns ...
	tree.ShowColumnTitles();
	tree.SetColumnTitle(0, "Markup");
	tree.SetColumnTitle(1, "Value");
	for(int column = 0; column < 2; column++)
		{
			tree.SetColumnVisibility(column, true);
			tree.SetColumnWidth(column, tree.GetOptimalColumnWidth(column));
		}

	// Restore normal tree updates ...
	tree.Thaw();
}

GtkCTreeNode* CXMLEditor::InsertRow(sdpGtkCTree Tree, GtkCTreeNode* ParentNode, sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_assert(Element);

	GtkCTreeNode* node = 0;

	if(Element->Name() == sdpString("<![CDATA["))
		{
			SDPCSTRING text[] = 
			{
				"CDATA",
				Element->Text()
			};

			node = Tree.InsertNode(ParentNode, 0, text, 0, m_CDataIcon, m_CDataIconMask, m_CDataIcon, m_CDataIconMask);
		}
	else
		{
			SDPCSTRING text[] = 
			{
				Element->Name(),
				Element->Text()
			};

			node = Tree.InsertNode(ParentNode, 0, text, 0, m_ElementIcon, m_ElementIconMask, m_ElementIcon, m_ElementIconMask);
		}

	m_Markups.push_back(Markup(Element));
	gtk_ctree_node_set_row_data(Tree, node, gpointer(&m_Markups.back()));
	return node;
}

GtkCTreeNode* CXMLEditor::InsertRow(sdpGtkCTree Tree, GtkCTreeNode* ParentNode, sdpXmlIAttribute* Attribute)
{
	// Sanity checks ...
	g_assert(ParentNode);
	g_assert(Attribute);

	// Create the root element ...
	SDPCSTRING text[] = 
	{
		Attribute->Name(),
		Attribute->Value()
	};

	GtkCTreeNode* node = Tree.InsertNode(ParentNode, 0, text, 0, m_AttributeIcon, m_AttributeIconMask, m_AttributeIcon, m_AttributeIconMask);
	m_Markups.push_back(Markup(Attribute));
	gtk_ctree_node_set_row_data(Tree, node, gpointer(&m_Markups.back()));

	return node;
}

void CXMLEditor::InsertChildren(sdpGtkCTree Tree, GtkCTreeNode* ParentNode, sdpXmlIElement* ParentElement)
{
	// Sanity checks ...
	g_assert(ParentNode);
	g_assert(ParentElement);

	// For each attribute ...
	for(SDPXMLATTRIBUTEHANDLE attributehandle = ParentElement->GetFirstAttribute(); attributehandle; attributehandle = ParentElement->GetNextAttribute(attributehandle))
		{
			sdpXmlIAttribute* attribute = ParentElement->GetAttribute(attributehandle);
			InsertRow(Tree, ParentNode, attribute);			
		}

	// For each child element ...
	for(SDPXMLELEMENTHANDLE elementhandle = ParentElement->GetFirstElement(); elementhandle; elementhandle = ParentElement->GetNextElement(elementhandle))
		{
			sdpXmlIElement* element = ParentElement->GetElement(elementhandle);
			GtkCTreeNode* node = InsertRow(Tree, ParentNode, element);

			// Handle children recursively ...
			InsertChildren(Tree, node, element);
		}
}

void CXMLEditor::UpdateControls()
{
	// If the user hasn't selected anything ...
	if(!CurrentMarkup())
		{
			// Get rid of leftover data ...
			Label("markuptype").SetText("None Selected");
			Entry("markupname").SetText("");
			Entry("markupvalue").SetText("");
			Text("markuptextvalue").DeleteText(0, -1);
		
			// Disable all our controls ...
			Widget("markuptype").SetSensitive(false);
			Widget("markupname").SetSensitive(false);
			Widget("markupvalue").SetSensitive(false);
			Widget("markuptextvalue").SetSensitive(false);
			return;
		}

	// Enable all our controls ...
	Widget("markuptype").SetSensitive(true);
	Widget("markupname").SetSensitive(true);
	Widget("markupvalue").SetSensitive(true);
	Widget("markuptextvalue").SetSensitive(false);
		
	if(CurrentMarkup()->IsElement())
		{
			Label("markuptype").SetText("Element");
			Entry("markupname").SetText(CurrentMarkup()->Element()->Name());
			
			int position = 0;
			sdpString text = CurrentMarkup()->Element()->Text();
			Entry("markupvalue").SetText(text);
			Text("markuptextvalue").DeleteText(0, -1);
			Text("markuptextvalue").InsertText(text, text.size(), &position);
		}
	else if(CurrentMarkup()->IsAttribute())
		{
			Label("markuptype").SetText("Attribute");
			Entry("markupname").SetText(CurrentMarkup()->Attribute()->Name());
			
			int position = 0;
			sdpString text = CurrentMarkup()->Attribute()->Value();
			Entry("markupvalue").SetText(text);
			Text("markuptextvalue").DeleteText(0, -1);
			Text("markuptextvalue").InsertText(text, text.size(), &position);
		}
}

void CXMLEditor::UpdateCurrentRow()
{
	// Sanity checks ...
	g_assert(m_CurrentRow);
	g_assert(CurrentMarkup());
	
	sdpGtkCTree tree = Tree();
	
	tree.Freeze();
	
	if(CurrentMarkup()->IsElement())
		{
			if(CurrentMarkup()->Element()->Name() == sdpString("<![CDATA["))
				{
					tree.SetPixmapText(m_CurrentRow, 0, "CDATA", 0, m_CDataIcon, m_CDataIconMask);
					tree.SetText(m_CurrentRow, 1, CurrentMarkup()->Element()->Text());
				}
			else
				{
					tree.SetPixmapText(m_CurrentRow, 0, CurrentMarkup()->Element()->Name(), 0, m_ElementIcon, m_ElementIconMask);
					tree.SetText(m_CurrentRow, 1, CurrentMarkup()->Element()->Text());
				}
		}
	else if(CurrentMarkup()->IsAttribute())
		{
			tree.SetPixmapText(m_CurrentRow, 0, CurrentMarkup()->Attribute()->Name(), 0, m_AttributeIcon, m_AttributeIconMask);
			tree.SetText(m_CurrentRow, 1, CurrentMarkup()->Attribute()->Value());
		}
		
	tree.Thaw();
}

sdpGtkCTree CXMLEditor::Tree()
{
	return CTree("tree");
}

void CXMLEditor::OnEvent(sdpGtkEvent* Event)
{
	// Handle events ...
	if(Event->Name() == "onfilenew")
		OnFileNew();
	else if(Event->Name() == "onfileopen")
		OnFileOpen();
	else if(Event->Name() == "onfilesave")
		OnFileSave();
	else if(Event->Name() == "onfilesaveas")
		OnFileSaveAs();
	else if(Event->Name() == "onfileclose")
		OnFileClose();
	else if(Event->Name() == "oneditnewelement")
		OnEditNewElement();
	else if(Event->Name() == "oneditnewcdata")
		OnEditNewCData();
	else if(Event->Name() == "oneditnewattribute")
		OnEditNewAttribute();
	else if(Event->Name() == "oneditdelete")
		OnEditDelete();
	else if(Event->Name() == "ontreeclicked")
		OnTreeClicked(Event);
	else if(Event->Name() == "onmarkupname")
		OnMarkupName();
	else if(Event->Name() == "onmarkupvalue")
		OnMarkupValue();
	else if(Event->Name() == "ondelete")
		OnDelete();
	else
		sdpGtkObjectContainer::OnEvent(Event);
}

void CXMLEditor::OnFileNew()
{
	ClearData();
	Tree().Clear();
		
	UpdateTitlebar();
	
	OnEditNewElement();

	UpdateTree();	
	UpdateControls();
}

void CXMLEditor::OnFileOpen()
{
	// Get a filepath from the user ...
	sdpGtkFileSelector selector("Open XML Document:", "");
	selector.DoModal(RootWindow());

	// If the user cancelled, we're done ...
	if(!selector.OK())
		return;

	// Clear out the old ...
	ClearData();

	// Load the document ...
	LoadDocument(selector.FilePath());
	
	// Update the UI ...
	UpdateTitlebar();
	UpdateTree();
	UpdateControls();
}

void CXMLEditor::OnFileSave()
{
	// If we don't have a filepath, prompt the user ...
	if(m_FilePath.size())
		SaveDocument(m_FilePath);
	else
		OnFileSaveAs();
}

void CXMLEditor::OnFileSaveAs()
{
	// Prompt the user for a filepath ...
	sdpGtkFileSelector selector("Save XML Document As:", "");
	selector.DoModal(RootWindow());
	
	// If the user cancelled, we're done ...
	if(!selector.OK())
		return;

	// Save the document ...
	SaveDocument(selector.FilePath());
	
	// Update our titlebar ...
	UpdateTitlebar();
}

void CXMLEditor::OnFileClose()
{
	DisconnectAllEvents();
	gtk_main_quit();
}

void CXMLEditor::OnEditNewElement()
{
	SelectionMap selection = CurrentSelection();
	if(selection.size() > 1)
		return;

	sdpXmlIElement* newelement = 0;
	GtkCTreeNode* newnode = 0;
	
	if(1 == selection.size())
		{					
			// Create a new element ...
			newelement = selection.begin()->second->Element()->CreateElement("unnamed");
			// Insert it into the tree ...
			newnode = InsertRow(Tree(), selection.begin()->first, newelement);
		}
	else if(0 == m_Document.GetRootElement())
		{
			// Create a new element ...
			newelement = m_Document.CreateRootElement("unnamed");
			// Insert it into the tree ...
			newnode = InsertRow(Tree(), 0, newelement);
		}
		
	// Select it ...
	Tree().Select(newnode);
	
	// Set it up for editing ...
	m_CurrentRow = newnode;
	UpdateControls();
}

void CXMLEditor::OnEditNewCData()
{
	SelectionMap selection = CurrentSelection();
	if(1 != selection.size())
		return;

	if(!selection.begin()->second->IsElement())
		return;

	// Create a new element ...
	sdpXmlIElement* newelement = selection.begin()->second->Element()->CreateElement("<![CDATA[");
	// Insert it into the tree ...
	GtkCTreeNode* node = InsertRow(Tree(), selection.begin()->first, newelement);
	// Select it ...
	Tree().Select(node);
	
	// Set it up for editing ...
	m_CurrentRow = node;
	UpdateControls();
}

void CXMLEditor::OnEditNewAttribute()
{
	SelectionMap selection = CurrentSelection();
	if(1 != selection.size())
		return;

	if(!selection.begin()->second->IsElement())
		return;

	// Create a new attribute ...
	sdpXmlIAttribute* newattribute = selection.begin()->second->Element()->CreateAttribute("unnamed", "");
	// Insert it into the tree ...
	GtkCTreeNode* node = InsertRow(Tree(), selection.begin()->first, newattribute);
	// Select it ...
	Tree().Select(node);
	
	// Set it up for editing ...
	m_CurrentRow = node;
	UpdateControls();
}

void CXMLEditor::OnEditDelete()
{
	SelectionMap selection = CurrentSelection();
	for(SelectionIterator i = selection.begin(); i != selection.end(); i++)
		{
			if(i->second->IsElement())
				{
					if(i->second->Element()->ParentElement())
						i->second->Element()->ParentElement()->DestroyElement(i->second->Element());
				}
			else if(i->second->IsAttribute())
				{
					i->second->Attribute()->ParentElement()->DestroyAttribute(i->second->Attribute());
				}
		}

	UpdateTree();
	UpdateControls();
}

void CXMLEditor::OnTreeClicked(sdpGtkEvent*Event)
{
	// Sanity checks ...
	g_assert(Event);

	sdpGtkEventWidgetButtonPressEvent* event = (sdpGtkEventWidgetButtonPressEvent*)Event;

	// Get the row we clicked on ...
	int row;
	int column;
	sdpGtkCTree tree = CTree("tree");
	tree.GetHitInfo(event->Event()->x, event->Event()->y, &row, &column);

	// Get the node for the given row ...
	m_CurrentRow = tree.Node(row);
	
	// If the user clicked outside any existing rows, we get a null node, so exit quietly ...
	if(!m_CurrentRow)
		return;

	// Update our editing controls ...
	UpdateControls();
}

void CXMLEditor::OnMarkupName()
{
	// If nothing's selected, ignore it ...
	if(!CurrentMarkup())
		return;
		
	if(CurrentMarkup()->IsElement())
		{
			CurrentMarkup()->Element()->SetName(Entry("markupname").GetText());
			UpdateCurrentRow();
		}
	else if(CurrentMarkup()->IsAttribute())
		{
			CurrentMarkup()->Attribute()->SetName(Entry("markupname").GetText());
			UpdateCurrentRow();
		}
}

void CXMLEditor::OnMarkupValue()
{
	// If nothing's selected, ignore it ...
	if(!CurrentMarkup())
		return;

	if(CurrentMarkup()->IsElement())
		{
			CurrentMarkup()->Element()->SetText(Entry("markupvalue").GetText());
			UpdateCurrentRow();
		}
	else if(CurrentMarkup()->IsAttribute())
		{
			CurrentMarkup()->Attribute()->SetValue(Entry("markupvalue").GetText());
			UpdateCurrentRow();
		}
}

void CXMLEditor::OnDelete()
{
	DisconnectAllEvents();
	gtk_main_quit();
}

CXMLEditor::Markup* CXMLEditor::CurrentMarkup()
{
	if(0 == m_CurrentRow)
		return 0;
		
	return reinterpret_cast<Markup*>(Tree().GetRowData(m_CurrentRow));
}

CXMLEditor::SelectionMap CXMLEditor::CurrentSelection()
{
	SelectionMap result;

	GtkCTree* tree = static_cast<GtkCTree*>(Tree());
	for(GList* selection = tree->clist.selection; selection; selection = selection->next)
		{
			GtkCTreeNode* node = reinterpret_cast<GtkCTreeNode*>(selection->data);
			Markup* markup = reinterpret_cast<Markup*>(Tree().GetRowData(node));
			
			result[node] = markup;
		}
		
	return result;
}

//////////////////////////////////////////////////////////////////////////////////
// main

/// Initializes GTK+ and loads the main window
gint main(gint argc, gchar* argv[])
{
	// Standard GTK+ initialization ...
	gtk_init(&argc, &argv);

	// Find the directory we were run from ...
	sdpString applicationpath = sdpGetProcessPath(argv[0]);

	// Create our editor object ...
	CXMLEditor editor(applicationpath);

	// See if we should open a file right away ...
	if(argc == 2)
		editor.LoadDocument(argv[1]);
	
	// Load the UI ...
	editor.LoadUI();
	
	// Main event loop ...
	gtk_main();

	return 0;
}
