#ifndef SDPCOM_H
#define SDPCOM_H

// SDPTypes Library
// Copyright  2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
// 
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares classes, functions, and definitions for sdpCOM, a simple Component Object Model for objects that communicate via binary interfaces
		\author Tim Shead (tshead@k-3d.com)
**/

#include <memory.h>
#include <sdpsimpletypes.h>

/// The preferred result for interface functions
typedef glong sdpResult;

/// Tests an sdpResult for success (non-negative numbers indicate success)
inline bool sdpSucceeded(const sdpResult Result) { return Result >= 0; }

/// Tests an sdpResult for failure (positive numbers indicate failure)
inline bool sdpFailed(const sdpResult Result) { return Result < 0; }

/// The method completed completely successfully (also indicates boolean TRUE)
extern const sdpResult SDP_S_OK;

/// The method completed successfully, and returned a boolean FALSE
extern const sdpResult SDP_S_FALSE;

/// The method failed
extern const sdpResult SDP_E_FAIL;

/// The requested interface could not be returned
extern const sdpResult SDP_E_NO_INTERFACE;

/// The requested object is an unknown type
extern const sdpResult SDP_E_UNKNOWN_OBJECT;

/// The method failed due to an out-of-memory condition
extern const sdpResult SDP_E_OUT_OF_MEMORY;

/// Globally unique identifier used to identify sdpCOM classes and interfaces
class sdpGUID
{
public:
	sdpGUID()
	{
		m_Data1 = m_Data2 = m_Data3 = m_Data4 = 0;
	}
	
	sdpGUID(const gulong Data1, const gulong Data2, const gulong Data3, const gulong Data4)
	{
		m_Data1 = Data1;
		m_Data2 = Data2;
		m_Data3 = Data3;
		m_Data4 = Data4;
	}

	const gulong& Data1() const { return m_Data1; }
	const gulong& Data2() const { return m_Data2; }
	const gulong& Data3() const { return m_Data3; }
	const gulong& Data4() const { return m_Data4; }

	void SetData1(const gulong Data1) { m_Data1 = Data1; }
	void SetData2(const gulong Data2) { m_Data2 = Data2; }
	void SetData3(const gulong Data3) { m_Data3 = Data3; }
	void SetData4(const gulong Data4) { m_Data4 = Data4; }

	bool LessThan(const sdpGUID& B) const
	{
		if(m_Data1 == B.m_Data1)
			{
				if(m_Data2 == B.m_Data2)
					{
						if(m_Data3 == B.m_Data3)
							{
								return m_Data4 < B.m_Data4;
							}
						else
							return m_Data3 < B.m_Data3;
					}
				else
					return m_Data2 < B.m_Data2;
			}
		else
			return m_Data1 < B.m_Data1;
	}
	
	bool Equals(const sdpGUID& B) const
	{
		return !memcmp(this, &B, sizeof(sdpGUID));
	}

protected:
	gulong m_Data1;
	gulong m_Data2;
	gulong m_Data3;
	gulong m_Data4;
};

/// Orders two sdpGUIDs
inline bool operator<(const sdpGUID& A, const sdpGUID& B)
{
	return A.LessThan(B);
} 

/// Tests two sdpGUIDs for equality
inline bool operator==(const sdpGUID& A, const sdpGUID& B)
{
	return A.Equals(B);
}

/// Tests two sdpGUIDs for inequality
inline bool operator!=(const sdpGUID& A, const sdpGUID& B)
{
	return !A.Equals(B);
}

/// A class identifier
class sdpCLSID : public sdpGUID
{
public:
	sdpCLSID() : sdpGUID() {}
	sdpCLSID(const gulong Data1, const gulong Data2, const gulong Data3, const gulong Data4) : sdpGUID(Data1, Data2, Data3, Data4) {}
};

/// Orders two sdpCLSIDs
inline bool operator<(const sdpCLSID& A, const sdpCLSID& B)
{
	return A.LessThan(B);
} 

/// Tests two sdpCLSIDs for equality
inline bool operator==(const sdpCLSID& A, const sdpCLSID& B)
{
	return A.Equals(B);
}

/// Tests two sdpCLSIDs for inequality
inline bool operator!=(const sdpCLSID& A, const sdpCLSID& B)
{
	return !A.Equals(B);
}

/// An interface identifier
class sdpIID : public sdpGUID
{
public:
	sdpIID() : sdpGUID() {}
	sdpIID(const gulong Data1, const gulong Data2, const gulong Data3, const gulong Data4) : sdpGUID(Data1, Data2, Data3, Data4) {}
};

/// Orders two sdpIIDs
inline bool operator<(const sdpIID& A, const sdpIID& B)
{
	return A.LessThan(B);
} 

/// Tests two sdpIIDs for equality
inline bool operator==(const sdpIID& A, const sdpIID& B)
{
	return A.Equals(B);
}

/// Tests two sdpIIDs for inequality
inline bool operator!=(const sdpIID& A, const sdpIID& B)
{
	return !A.Equals(B);
}

/// The NULL GUID
extern const sdpGUID sdpGUIDNull;

/// The NULL IID
extern const sdpIID sdpIIDNull;

/// The NULL CLSID
extern const sdpCLSID sdpCLSIDNull;

/// The base class for all other sdpCOM interfaces
class sdpIUnknown
{
public:
	const static sdpIID iid;

	/// Queries the underlying object at run-time for an interface of the requested type
	virtual sdpResult QueryInterface(const sdpIID& InterfaceID, void** Interface) = 0;
	/// Increments the reference count of the underlying object
	virtual gulong AddRef() = 0;
	/// Decrements the reference count of the underlying object
	virtual gulong Release() = 0;
};

/// If possible, converts an interface of one type into another using QueryInterface()
template<class Interface>
Interface* sdpQueryInterface(sdpIUnknown* Source)
{
	// Sanity checks ...
	g_assert(Source);

	// Our result interface pointer ...
	Interface* result = 0;

	// Try and get it ...
	if(SDP_S_OK != Source->QueryInterface(Interface::iid, (void**)&result))
		return 0;

	return result;
}

/// Handles QueryInterface() requests in a standardized way
template<class Interface>
bool sdpSupportInterface(Interface* Server, const sdpIID& Requested, void** Result)
{
	// Sanity checks ...
	g_assert(Server);
	g_assert(Result);

	const sdpIID interfaceid = Interface::iid;

	if(interfaceid == Requested)
		{
			*Result = (void*)Server;
			Server->AddRef();

			return true;
		}

	return false;
}

/*

This is how I wanted this function to work, but those wacky folks at M$ have some bizarre problems
with template instantiation under MSVC++ 6.0 ...

/// Handles requests to create a class instance in a standardized way
template<class ObjectClass>
sdpResult sdpCreateClassInstance(const sdpIID& RequestInterfaceID, void** RequestInterface)
{
	// Sanity checks ...
	g_return_val_if_fail(RequestInterface, SDP_E_FAIL);

	ObjectClass* object = new ObjectClass();
	g_return_val_if_fail(object, SDP_E_OUT_OF_MEMORY);

	*RequestInterface = 0;
	sdpResult result = object->QueryInterface(RequestInterfaceID, RequestInterface);
	if(sdpFailed(result))
		delete object;

	return result;
}
*/

/// Handles requests to create a class instance in a standardized way
template<class ObjectClass>
ObjectClass* sdpCreateClassInstance(const sdpIID& RequestInterfaceID, void** RequestInterface)
{
	// Sanity checks ...
	g_return_val_if_fail(RequestInterface, 0);

	ObjectClass* object = new ObjectClass();
	g_return_val_if_fail(object, 0);

	*RequestInterface = 0;
	sdpResult result = object->QueryInterface(RequestInterfaceID, RequestInterface);
	if(sdpFailed(result))
		{
			delete object;
			object = 0;
		}

	return object;
}

/// Handles AddRef() in a standardized way
inline gulong sdpComAddRef(gulong& Count)
{
	return ++Count;
}

/// Handles Release() in a standardized way
template<class Type>
gulong sdpComRelease(const Type* Owner, gulong& Count)
{
	gulong newcount = --Count;
	if(0 == newcount)
		delete Owner;
		
	return newcount;
}

/// Required function for sdpCom libraries that returns an XML document describing the library and the sdpCOM classes it supports
typedef SDPCSTRING (*sdpComLibraryInfoProc)();

/// Required function for sdpCom libraries that creates an object and returns an interface to it of the given type
typedef sdpResult (*sdpComClassProcedureProc)(const sdpCLSID& ClassID, const sdpIID& InterfaceID, void** Interface);


#endif SDPCOM_H
