#ifndef SDPMATH_H
#define SDPMATH_H

// SDPTypes Library
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
// 
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


/** \file
		\brief Declares a wide variety of math functions, constants, and vector, matrix and quaternion classes useful in 3D graphics work
		\author Tim Shead (tshead@k-3d.com)
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include <sdpsimpletypes.h>

#include <vector>
#include <algorithm>

#include <math.h>

/// Pi
extern const gdouble sdpPi;
/// Pi divided-by two
extern const gdouble sdpPiOver2;
/// Pi times two
extern const gdouble sdpPiTimes2;

/// Rounds a floating-point value to the closest integer
gdouble sdpRound(const gdouble Value);

/// Returns the maximum of two values
template<class Type> Type sdpMax(const Type& A, const Type& B)
{ return A >= B ? A : B; }

/// Returns the minimum of two values
template<class Type> Type sdpMin(const Type& A, const Type& B)
{ return A <= B ? A : B; }

/// Clamps a value within the specified range
template<class Type> Type sdpClamp(const Type& Min, const Type& Value, const Type& Max)
{ return sdpMax(Min, sdpMin(Max, Value)); }

/// Returns the linear interpolation of two values
template<class Type> Type sdpLerp(const Type& A, const Type& B, const gdouble Mix)
{	return A + ((B - A) * Mix); }

/// Returns the sign of an argument: -1 if negative, 0 if zero, +1 if positive
template<class Type> gdouble sdpSign(const Type& Arg)
{
	if(Arg > 0.0) return 1.0;
	if(Arg < 0.0) return -1.0;

	return 0.0;
}

/// Converts degrees to radians
template<class Type> Type sdpRadians(const Type& Degrees)
{	return Degrees * 0.01745329252; }

/// Converts radians to degrees
template<class Type> Type sdpDegrees(const Type& Radians)
{	return Radians * 57.2957795131; }

/// Computes N!
gdouble sdpFactorial(gdouble N);

/// Returns a Bezier / Bernstein basis for the given order, control point number, and parameter value
gdouble sdpBernsteinBasis(const gulong Order, const gulong ControlPoint, const gdouble Parameter);

/// Computes a Bezier curve value with given order,  control points, and parameter value
template<class Type>
Type sdpBezier(const std::vector<Type>& ControlPoints, const gdouble Parameter)
{
	// Sanity checks ...
	g_assert(ControlPoints.size() > 1);
	
	Type result = 0.0;
	
	for(gulong i = 0; i < ControlPoints.size(); i++)
		result += sdpBernsteinBasis(ControlPoints.size(), i, Parameter) * ControlPoints[i];
	
	return result;
}

/****************************************************************
*
* C++ Vector and Matrix Algebra routines
* Author: Jean-Francois DOUE
* Version 3.1 --- October 1993
*
****************************************************************/

//
//	From "Graphics Gems IV / Edited by Paul S. Heckbert
//	Academic Press, 1994, ISBN 0-12-336156-9
//	"You are free to use and modify this code in any way
//	you like." (p. xv)
//
//	Modified by J. Nagle, March 1997
//	-	All functions are inline.
//	-	All functions are const-correct.
//	-	All checking is via the standard "g_assert" macro.
//

// Modified by Tim Shead for use with K-3D, January 1998

// Forward class declarations ...
class sdpVector2;
class sdpVector3;
class sdpVector4;
class sdpRectangle;
class sdpMatrix3;
class sdpMatrix4;
class sdpQuaternion;
class sdpAngleAxis;
class sdpEulerAngles;
class sdpIPolygon;
class sdpIPolyPoint;

/// Axes array indices
enum {VX, VY, VZ, VW};
/// Plane array indices
enum {PA, PB, PC, PD};


/////////////////////////////////////////////////////////////////
// sdpVector2

/// A two-dimensional vector
class sdpVector2
{
public:
	/// Stores the vector values
	gdouble n[2];

	// Constructors
	sdpVector2();
	sdpVector2(const gdouble x, const gdouble y);
	sdpVector2(const gdouble d);
	sdpVector2(const gdouble[2]);
	/// Copy constructor
	sdpVector2(const sdpVector2& v);
	/// Casts an sdpVector3 to an sdpVector2, dropping the third dimension
	sdpVector2(const sdpVector3& v);
	/// Casts an sdpVector3 to an sdpVector2, dropping the given dimension
	sdpVector2(const sdpVector3& v, gint dropAxis);
	/// Assigns an sdpVector2
	sdpVector2& operator	= ( const sdpVector2& v );
	/// Assigns a C/C++ style array
	sdpVector2& operator  = ( const gdouble[2]);
	/// Adds an sdpVector2
	sdpVector2& operator += ( const sdpVector2& v );
	/// Subtracts an sdpVector2
	sdpVector2& operator -= ( const sdpVector2& v );
	/// Multiplication by a constant
	sdpVector2& operator *= ( const gdouble d );
	/// Division by a constant
	sdpVector2& operator /= ( const gdouble d );
	/// Returns an indexed dimension by reference
	gdouble& operator [] ( gint i);
	/// Returns an indexed dimension by value
	gdouble sdpVector2::operator [] ( gint i) const;
	/// Returns the vector length
	gdouble Length() const;
	/// Returns the square of the vector length
	gdouble Length2() const;
	/// Normalizes the vector in place
	sdpVector2& Normalize();
	/// Returns the perpendicular to this vector
	sdpVector2 Perpendicular() const;
	/// Returns the quadrant containing this vector
	glong Quadrant();
	/// Returns the angle of this vector with respect to the positive X axis
	gdouble Angle();
	/// Casting to a C/C++ style array pointer
	operator gdouble*() { return &n[0]; }
	/// Copies the vector into a C/C++ style array
	void CopyArray(gfloat f[2]);
	/// Copies the vector into a C/C++ style array
	void CopyArray(gdouble d[2]);
};

/// Negation
sdpVector2 operator - (const sdpVector2& v);
/// Addition
sdpVector2 operator + (const sdpVector2& a, const sdpVector2& b);
/// Subtraction
sdpVector2 operator - (const sdpVector2& a, const sdpVector2& b);
/// Multiplication by a constant
sdpVector2 operator * (const sdpVector2& a, const gdouble d);
/// Multiplication by a constant
sdpVector2 operator * (const gdouble d, const sdpVector2& a);
/// Multiplication by a matrix
sdpVector2 operator * (const sdpMatrix3& a, const sdpVector2& v);
/// Multiplication by a matrix
sdpVector2 operator * (const sdpVector2& v, const sdpMatrix3& a);
/// Returns the dot product of two vectors
gdouble operator * (const sdpVector2& a, const sdpVector2& b);
/// Division by a constant
sdpVector2 operator / (const sdpVector2& a, const gdouble d);
/// Returns the cross product of two vectors
sdpVector3 operator ^ (const sdpVector2& a, const sdpVector2& b);
/// Tests for equality
gint operator == (const sdpVector2& a, const sdpVector2& b);
/// Tests for non-equality
gint operator != (const sdpVector2& a, const sdpVector2& b);
/// Returns the term-by-term minimum of two vectors
sdpVector2 sdpVectorMin(const sdpVector2& a, const sdpVector2& b);
/// Returns the term-by-term maximum of two vectors
sdpVector2 sdpVectorMax(const sdpVector2& a, const sdpVector2& b);
/// Returns the term-by-term product of two vectors
sdpVector2 sdpProduct(const sdpVector2& a, const sdpVector2& b);

/// An array of two-dimensional vectors
typedef std::vector<sdpVector2> sdpVector2Array;
typedef sdpVector2Array::iterator sdpVector2Iterator;
/// Tests to see if the point is "within" the area defined by the array of points
bool sdpIsInside(sdpVector2& Point, sdpVector2Array& Array);
/// Returns the array point closest to the given coordinate
sdpVector2 sdpClosestPoint(sdpVector2& Point, sdpVector2Array& Array);

/////////////////////////////////////////////////////////////////
// sdpVector3

/// A three-dimensional vector
class sdpVector3
{
public:
	/// Stores the vector values
	gdouble n[3];

	// Constructors
	sdpVector3();
	sdpVector3(const gdouble x, const gdouble y, const gdouble z);
	sdpVector3(const gdouble d);
	sdpVector3(const gdouble d[3]);
	/// Copy constructor
	sdpVector3(const sdpVector3& v);
	/// Casts an sdpVector2 to an sdpVector3 (with zero third dimension)
	sdpVector3(const sdpVector2& v);
	/// Casts an sdpVector2 to an sdpVector3 with the given third dimension
	sdpVector3(const sdpVector2& v, gdouble d);
	/// Casts an sdpVector4 to an sdpVector3, dropping the fourth dimension
	sdpVector3(const sdpVector4& v);
	/// Casts an sdpVector4 to an sdpVector3, dropping the given dimension
	sdpVector3(const sdpVector4& v, gint dropAxis);
	/// Assignment of an sdpVector3
	sdpVector3& operator	= ( const sdpVector3& v );
	/// Assignment of a C/C++ style array
	sdpVector3& operator  = ( const gdouble d[3]);
	/// Addition
	sdpVector3& operator += ( const sdpVector3& v );
	/// Subtraction
	sdpVector3& operator -= ( const sdpVector3& v );
	/// Multiplication by a constant
	sdpVector3& operator *= ( const gdouble d );
	/// Division by a constant
	sdpVector3& operator /= ( const gdouble d );
	/// Returns the given dimension by reference
	gdouble& operator [] ( gint i);
	/// Returns the given dimension by value
	gdouble operator[] (gint i) const;
	/// Returns the vector length
	gdouble Length() const;
	/// Returns the squared vector length
	gdouble Length2() const;
	/// Normalizes the vector in place
	sdpVector3& Normalize();
	/// Converts Cartesian coordinates to spherical coordinates in place
	sdpVector3& CartesianToSpherical();
	/// Converts an RGB color to HSV in place
	sdpVector3& RGBtoHSV();
	/// Converts an HSV color to RGB in place
	sdpVector3& HSVtoRGB();
	/// Makes an RGB color TV safe by reducing either chroma or saturation as needed
	sdpVector3& VideoSafe(bool NTSC, bool ReduceSat);
	/// Sets each dimension to its absolute value in place
	sdpVector3& Abs();
	/// Casting to a C/C++ style array pointer
	operator gdouble*() { return &n[0]; }
	/// Copies the vector into a C/C++ style array
	void CopyArray(gfloat f[3]);
	/// Copies the vector into a C/C++ style array
	void CopyArray(gdouble d[3]);
};

/// Negation
sdpVector3 operator - (const sdpVector3& v);
/// Addition
sdpVector3 operator + (const sdpVector3& a, const sdpVector3& b);
/// Subtraction
sdpVector3 operator - (const sdpVector3& a, const sdpVector3& b);
/// Multiplication by a constant
sdpVector3 operator * (const sdpVector3& a, const gdouble d);
/// Multiplication by a constant
sdpVector3 operator * (const gdouble d, const sdpVector3& a);
/// Matrix multiplication
sdpVector3 operator * (const sdpMatrix4& a, const sdpVector3& v);
/// Matrix multiplication
sdpVector3 operator * (const sdpVector3& v, const sdpMatrix4& a);
/// Returns the dot product of two vectors
gdouble operator * (const sdpVector3& a, const sdpVector3& b);
/// Division by a constant
sdpVector3 operator / (const sdpVector3& a, const gdouble d);
/// Returns the cross product of two vectors
sdpVector3 operator ^ (const sdpVector3& a, const sdpVector3& b);
/// Equality
gint operator == (const sdpVector3& a, const sdpVector3& b);
/// Non-equality
gint operator != (const sdpVector3& a, const sdpVector3& b);
/// Returns the term-by-term minimum of two vectors
sdpVector3 sdpVectorMin(const sdpVector3& a, const sdpVector3& b);
/// Returns the term-by-term maximum of two vectors
sdpVector3 sdpVectorMax(const sdpVector3& a, const sdpVector3& b);
/// Returns the term-by-term product of two vectors
sdpVector3 sdpProduct(const sdpVector3& a, const sdpVector3& b);
/// Returns a linear transformation
sdpVector2 operator * (const sdpMatrix3& a, const sdpVector2& v);
/// Returns a matrix product
sdpMatrix3 operator * (const sdpMatrix3& a, const sdpMatrix3& b);

/// An array of three-dimensional vectors
typedef std::vector<sdpVector3> sdpVector3Array;
typedef sdpVector3Array::iterator sdpVector3Iterator;


////////////////////////////////////////////////////////////////////////////////////////////
// sdpVector4

/// A four-dimensional vector
class sdpVector4
{
public:
	/// Stores the vector values
	gdouble n[4];

	// Constructors
	sdpVector4();
	sdpVector4(const gdouble x, const gdouble y, const gdouble z, const gdouble w);
	sdpVector4(const gdouble d);
	sdpVector4(const gdouble d[4]);
	/// Copy Constructor
	sdpVector4(const sdpVector4& v);
	/// Casts an sdpVector3 to an sdpVector4 with zero fourth dimension
	sdpVector4(const sdpVector3& v);
	/// Casts an sdpVector3 to an sdpVector4 with the given fourth dimension
	sdpVector4(const sdpVector3& v, const gdouble d);
	/// Assignment of an sdpVector4
	sdpVector4& operator	= ( const sdpVector4& v );
	/// Assignment of a C/C++ array
	sdpVector4& operator  = ( const gdouble d[4]);
	/// Addition
	sdpVector4& operator += ( const sdpVector4& v );
	/// Subtraction
	sdpVector4& operator -= ( const sdpVector4& v );
	/// Multiplication by a constant
	sdpVector4& operator *= ( const gdouble d );
	/// Division by a constant
	sdpVector4& operator /= ( const gdouble d );
	/// Returns a vector element by reference
	gdouble& operator [] ( gint i);
	/// Returns a vector element by value
	gdouble operator[] (gint i) const;
	/// Returns the length of a vector
	gdouble Length() const;
	/// Returns the square of the vector length
	gdouble Length2() const;
	/// Normalizes the vector in place
	sdpVector4& Normalize();
	/// Casts the contents of the vector to a C/C++ array
	operator gdouble*() { return &n[0]; }
};

/// Negation
sdpVector4 operator - (const sdpVector4& v);
/// Addition
sdpVector4 operator + (const sdpVector4& a, const sdpVector4& b);
/// Subtraction
sdpVector4 operator - (const sdpVector4& a, const sdpVector4& b);
/// Multiplication by a constant
sdpVector4 operator * (const sdpVector4& a, const gdouble d);
/// Multiplication by a constant
sdpVector4 operator * (const gdouble d, const sdpVector4& a);
/// Matrix multiplication
sdpVector4 operator * (const sdpMatrix4& a, const sdpVector4& v);
/// Matrix multiplication
sdpVector4 operator * (const sdpVector4& v, const sdpMatrix4& a);
/// Returns the dot product of two vectors
gdouble operator * (const sdpVector4& a, const sdpVector4& b);
/// Division by a constant
sdpVector4 operator / (const sdpVector4& a, const gdouble d);
/// Equality
gint operator == (const sdpVector4& a, const sdpVector4& b);
/// Non-equality
gint operator != (const sdpVector4& a, const sdpVector4& b);
/// Returns the element-by-element minimum of two vectors
sdpVector4 sdpVectorMin(const sdpVector4& a, const sdpVector4& b);
/// Returns the element-by-element maximum of two vectors
sdpVector4 sdpVectorMax(const sdpVector4& a, const sdpVector4& b);
/// Returns the element-by-element product of two vectors
sdpVector4 sdpProduct(const sdpVector4& a, const sdpVector4& b);
/// Linear transformation
sdpVector3 operator * (const sdpMatrix4& a, const sdpVector3& v);
/// Matrix multiplication
sdpMatrix4 operator * (const sdpMatrix4& a, const sdpMatrix4& b);

/// An array of four-dimensional vectors
typedef std::vector<sdpVector2> sdpVector4Array;
typedef sdpVector4Array::iterator sdpVector4Iterator;


////////////////////////////////////////////////////////////////////////////////////////////
// sdpRectangle

/// A two-dimensional rectangle with left, right, top, bottom
class sdpRectangle
{
public:
	/// Stores the rectangle's left edge
	gdouble left;
	/// Stores the rectangle's right edge
	gdouble right;
	/// Stores the rectangle's top edge
	gdouble top;
	/// Stores the rectangle's bottom edge
	gdouble bottom;

	// Constructors
	sdpRectangle();
	sdpRectangle(const gdouble Left, const gdouble Right, const gdouble Top, const gdouble Bottom);
	/// Copy Constructor
	sdpRectangle(const sdpRectangle& v);
	/// Assignment of an sdpRectangle
	sdpRectangle& operator	= ( const sdpRectangle& v );
	/// Returns the left coordinate
	gdouble Left();
	/// Returns the right coordinate
	gdouble Right();
	/// Returns the top coordinate
	gdouble Top();
	/// returns the bottom coordinate
	gdouble Bottom();
	/// Returns the rectangle width
	gdouble Width();
	/// Returns the rectangle height
	gdouble Height();
	/// Applies horizontal and vertical offsets to the rectangle coordinates
	void Offset(const gdouble Horizontal, const gdouble Vertical);
	/// Returns a larger rectangle
	void Grow(const gdouble Thickness);
	/// Returns a larger rectangle
	void Grow(const gdouble Horizontal, const gdouble Vertical);
	/// Returns a smaller rectangle
	void Shrink(const gdouble Thickness);
	/// Returns a smaller rectangle
	void Shrink(const gdouble Horizontal, const gdouble Vertical);
	/// Ensures that the rectangle coordinates are ordered properly, Left < Right, Top < Bottom
	void Normalize();
	/// Returns the center of the rectangle
	sdpVector2 Center();
	/// Tests a point to see if it is contained within the rectangle
	bool Contains(const sdpVector2& Point);
};

/// Returns the source rectangle, proportionately constrained to fit within the destination
sdpRectangle sdpConstrainProportions(const sdpRectangle& Source, const sdpRectangle& Destination);

/// An array of rectangles
typedef std::vector<sdpRectangle> sdpRectangleArray;
typedef sdpRectangleArray::iterator sdpRectangleIterator;


////////////////////////////////////////////////////////////////
// sdpMatrix3

/// A 3x3 matrix
class sdpMatrix3
{
public:
	/// Stores the matrix data
	sdpVector3 v[3];
	// Constructors
	sdpMatrix3();
	/// Creates the matrix from three vectors
	sdpMatrix3(const sdpVector3& v0, const sdpVector3& v1, const sdpVector3& v2);
	/// Creates a matrix of identical elements
	sdpMatrix3(const gdouble d);
	/// Copy constructor
	sdpMatrix3(const sdpMatrix3& m);
	/// Assignment
	sdpMatrix3& operator	= ( const sdpMatrix3& m );
	/// Addition
	sdpMatrix3& operator += ( const sdpMatrix3& m );
	/// Subtraction
	sdpMatrix3& operator -= ( const sdpMatrix3& m );
	/// Multiplication by a constant
	sdpMatrix3& operator *= ( const gdouble d );
	/// Division by a constant
	sdpMatrix3& operator /= ( const gdouble d );
	/// Returns a vector reference by index
	sdpVector3& operator [] ( gint i);
	/// Returns a vector reference by index
	const sdpVector3& operator [] ( gint i) const;
	/// Transposes the matrix in place
	sdpMatrix3 Transpose() const;
	/// Inverts the matrix in place
	sdpMatrix3 Inverse() const;
};

/// Negation
sdpMatrix3 operator - (const sdpMatrix3& a);
/// Addition
sdpMatrix3 operator + (const sdpMatrix3& a, const sdpMatrix3& b);
/// Subtraction
sdpMatrix3 operator - (const sdpMatrix3& a, const sdpMatrix3& b);
/// Matrix multiplication
sdpMatrix3 operator * (const sdpMatrix3& a, const sdpMatrix3& b);
/// Multiplication by a constant
sdpMatrix3 operator * (const sdpMatrix3& a, const gdouble d);
/// Multiplication by a constant
sdpMatrix3 operator * (const gdouble d, const sdpMatrix3& a);
/// Division by a constant
sdpMatrix3 operator / (const sdpMatrix3& a, const gdouble d);
/// Equality
gint operator == (const sdpMatrix3& a, const sdpMatrix3& b);
/// Non-equality
gint operator != (const sdpMatrix3& a, const sdpMatrix3& b);
/// Linear transformation
sdpVector3 operator * (const sdpMatrix3& a, const sdpVector3& v);
/// Linear transformation
sdpVector2 operator * (const sdpMatrix3& a, const sdpVector2& v);


/////////////////////////////////////////////////////////////////
// sdpMatrix4

/// A 4x4 matrix
class sdpMatrix4
{
public:
	/// Stores the matrix elements
	sdpVector4 v[4];
	// Constructors
	sdpMatrix4();
	sdpMatrix4(const sdpVector4& v0, const sdpVector4& v1, const sdpVector4& v2, const sdpVector4& v3);
	sdpMatrix4(const gdouble d);
	sdpMatrix4(sdpEulerAngles Angles);
	/// Copy constructor
	sdpMatrix4(const sdpMatrix4& m);
	/// Assignment of an sdpMatrix4
	sdpMatrix4& operator	= ( const sdpMatrix4& m );
	/// Assignment of a C/C++ array
	sdpMatrix4& operator  = ( const gdouble d[4][4]);
	/// Assignment of a C/C++ array
	sdpMatrix4& operator  = ( const gdouble d[16]);
	/// Addition
	sdpMatrix4& operator += ( const sdpMatrix4& m );
	/// Subtraction
	sdpMatrix4& operator -= ( const sdpMatrix4& m );
	/// Multiplication by a constant
	sdpMatrix4& operator *= ( const gdouble d );
	/// Division by a constant
	sdpMatrix4& operator /= ( const gdouble d );
	/// Returns a vector by index
	sdpVector4& operator [] ( gint i);
	/// Returns a vector by index
	const sdpVector4& operator [] ( gint i) const;
	/// Transposes the matrix in place
	sdpMatrix4 Transpose() const;
	/// Inverts the matrix in place
	sdpMatrix4 Inverse() const;
	/// Copies the matrix contents into a C/C++ style array
	void CopyArray(gfloat m[16]);
	/// Copies the matrix contents into a C/C++ style array
	void CopyArray(gdouble m[16]);
	/// Casts the matrix to a C/C++ style array
	operator gdouble*() { return &v[0][0]; }
};

/// Negation
sdpMatrix4 operator - (const sdpMatrix4& a);
/// Addition
sdpMatrix4 operator + (const sdpMatrix4& a, const sdpMatrix4& b);
/// Subtraction
sdpMatrix4 operator - (const sdpMatrix4& a, const sdpMatrix4& b);
/// Matrix multiplication
sdpMatrix4 operator * (const sdpMatrix4& a, const sdpMatrix4& b);
/// Multiplication by a constant
sdpMatrix4 operator * (const sdpMatrix4& a, const gdouble d);
/// Multiplication by a constant
sdpMatrix4 operator * (const gdouble d, const sdpMatrix4& a);
/// Division by a constant
sdpMatrix4 operator / (const sdpMatrix4& a, const gdouble d);
/// Equality
gint operator == (const sdpMatrix4& a, const sdpMatrix4& b);
/// Non-equality
gint operator != (const sdpMatrix4& a, const sdpMatrix4& b);
/// Linear transformation
sdpVector4 operator * (const sdpMatrix4& a, const sdpVector4& v);
/// Linear transformation
sdpVector3 operator * (const sdpMatrix4& a, const sdpVector3& v);


////////////////////////////////////////////////////////////////////
// sdpQuaternion

/// Encapsulates a quaternion
class sdpQuaternion
{
public:
	// The scalar+vector representation is used
	gdouble w;
	sdpVector3 v;

	// Constructors
	sdpQuaternion();
	sdpQuaternion(const gdouble u, const sdpVector3& t);
	sdpQuaternion(const gdouble u, const gdouble x, const gdouble y, const gdouble z);
	sdpQuaternion(const sdpAngleAxis& AngleAxis);
	sdpQuaternion(sdpEulerAngles Angles);
	/// Assignment of a quaternion
	sdpQuaternion& operator = (const sdpQuaternion& q);
	/// Assignment of an axis/angle
	sdpQuaternion& operator = (const sdpAngleAxis& AngleAxis);
	/// Multiplication by a constant
	sdpQuaternion& operator *= (const gdouble f);
	/// Division by a constant
	sdpQuaternion& operator /= (const gdouble f);
	/// Addition
	sdpQuaternion& operator += (const sdpQuaternion& q);
	/// Substraction
	sdpQuaternion& operator -= (const sdpQuaternion& q);
	/// Multiplication
	sdpQuaternion& operator *= (const sdpQuaternion& q);
	/// Division
	sdpQuaternion& operator /= (const sdpQuaternion& q);
	/// Magnitude
	gdouble Magnitude() const;
	/// Normalization
	sdpQuaternion& Normalize();
	/// Conjugate
	sdpQuaternion& Conjugate();
	/// Inversion in place
	sdpQuaternion& Inverse();
	/// Squares the quaternion in place
	sdpQuaternion& Square();
	/// Axis rotation
	sdpQuaternion& AxisRotation(const gdouble phi);
};

/// Negation
sdpQuaternion operator - (sdpQuaternion& q);
/// Addition
sdpQuaternion operator + (const sdpQuaternion& q, const sdpQuaternion& r);
/// Substraction
sdpQuaternion operator - (const sdpQuaternion& q, const sdpQuaternion& r);
/// Multiplication by a quaternion
sdpQuaternion operator * (const sdpQuaternion& q, const sdpQuaternion& r);
/// Multiplication by a constant
sdpQuaternion operator * (const sdpQuaternion& q, const double f);
/// Multiplication by a constant
sdpQuaternion operator * (double f, const sdpQuaternion& q);
/// Division by a quaternion
sdpQuaternion operator / (const sdpQuaternion& q, const sdpQuaternion& r);
/// Division by a constant
sdpQuaternion operator / (const sdpQuaternion& q, const double f);
/// Equality test
gint operator == (const sdpQuaternion& q, const sdpQuaternion& r);
/// Non-equality test
gint operator != (const sdpQuaternion& q, const sdpQuaternion& r);


////////////////////////////////////////////////////////////////////
// sdpAngleAxis

/// Encapsulates an angle/axis representation of an object's orientation
class sdpAngleAxis
{
public:
	gdouble angle;
	sdpVector3 axis;

	// Constructors ...
	sdpAngleAxis();
	sdpAngleAxis(const gdouble Angle, const sdpVector3& Axis);
	sdpAngleAxis(const gdouble Angle, const gdouble X, const gdouble Y, const gdouble Z);
	sdpAngleAxis(const gdouble AngleAxis[4]);
	sdpAngleAxis(const sdpQuaternion& Quaternion);
	/// Copy constructor
	sdpAngleAxis(const sdpAngleAxis& AngleAxis);
	/// Assignment of an sdpAngleAxis
	sdpAngleAxis& operator = ( const sdpAngleAxis& AngleAxis );
	/// Assignment of a quaternion
	sdpAngleAxis& operator = ( const sdpQuaternion& Quaternion );
	/// Assignment of a C/C++ style array
	sdpAngleAxis& operator = ( const gdouble AngleAxis[4]);
	/// Casting to a C/C++ style array pointer
	operator gdouble*() { return &angle; }
	/// Copies the vector into a C/C++ style array
	void CopyArray(gfloat AngleAxis[4]);
	/// Copies the vector into a C/C++ style array
	void CopyArray(gdouble AngleAxis[4]);
};

#define SDPEULERANGLEORDER(initialaxis, parity, repetition, frame) ((((((initialaxis << 1) + parity) << 1) + repetition) << 1) + frame)

/// Encapsulates a set of Euler angles, including the order in which they should be applied
class sdpEulerAngles
{
public:

	/// Enumerates what gets rotated - the axes themselves, or the frame they represent
	typedef enum
	{
		StaticFrame = 0,
		RotatingFrame = 1,
	} OrderFrame;

	/// Enumerates whether the last axis will be the same as the first
	typedef enum
	{
		NoRepetition = 0,
		Repeats = 1,
	} OrderRepetition;

	/// Enumerates which axes will become the second axis
	typedef enum
	{
		EvenParity = 0,
		OddParity = 1,
	} OrderParity;

	/// Enumerates all possible axis order permutations
	typedef enum
	{
		XYZstatic = SDPEULERANGLEORDER(0, EvenParity, NoRepetition, StaticFrame),
		XYXstatic = SDPEULERANGLEORDER(0, EvenParity, Repeats, StaticFrame),
		XZYstatic = SDPEULERANGLEORDER(0, OddParity, NoRepetition, StaticFrame),
		XZXstatic = SDPEULERANGLEORDER(0, OddParity, Repeats, StaticFrame),
		YZXstatic = SDPEULERANGLEORDER(1, EvenParity, NoRepetition, StaticFrame),
		YZYstatic = SDPEULERANGLEORDER(1, EvenParity, Repeats, StaticFrame),
		YXZstatic = SDPEULERANGLEORDER(1, OddParity, NoRepetition, StaticFrame),
		YXYstatic = SDPEULERANGLEORDER(1, OddParity, Repeats, StaticFrame),
		ZXYstatic = SDPEULERANGLEORDER(2, EvenParity, NoRepetition, StaticFrame),
		ZXZstatic = SDPEULERANGLEORDER(2, EvenParity, Repeats, StaticFrame),
		ZYXstatic = SDPEULERANGLEORDER(2, OddParity, NoRepetition, StaticFrame),
		ZYZstatic = SDPEULERANGLEORDER(2, OddParity, Repeats, StaticFrame),
		ZYXrotating = SDPEULERANGLEORDER(0, EvenParity, NoRepetition, RotatingFrame),
		XYXrotating = SDPEULERANGLEORDER(0, EvenParity, Repeats, RotatingFrame),
		YZXrotating = SDPEULERANGLEORDER(0, OddParity, NoRepetition, RotatingFrame),
		XZXrotating = SDPEULERANGLEORDER(0, OddParity, Repeats, RotatingFrame),
		XZYrotating = SDPEULERANGLEORDER(1, EvenParity, NoRepetition, RotatingFrame),
		YZYrotating = SDPEULERANGLEORDER(1, EvenParity, Repeats, RotatingFrame),
		ZXYrotating = SDPEULERANGLEORDER(1, OddParity, NoRepetition, RotatingFrame),
		YXYrotating = SDPEULERANGLEORDER(1, OddParity, Repeats, RotatingFrame),
		YXZrotating = SDPEULERANGLEORDER(2, EvenParity, NoRepetition, RotatingFrame),
		ZXZrotating = SDPEULERANGLEORDER(2, EvenParity, Repeats, RotatingFrame),
		XYZrotating = SDPEULERANGLEORDER(2, OddParity, NoRepetition, RotatingFrame),
		ZYZrotating = SDPEULERANGLEORDER(2, OddParity, Repeats, RotatingFrame),
	} AngleOrder;

	/// Stores the three axis angles
	gdouble n[3];
	/// Stores the axis order in packed format
	AngleOrder order;

	sdpEulerAngles();
	sdpEulerAngles(sdpVector3 Angles, AngleOrder Order);
	sdpEulerAngles(gdouble X, gdouble Y, gdouble Z, AngleOrder Order);
	/// Conversion from a quaternion
	sdpEulerAngles(sdpQuaternion& Quaternion, AngleOrder Order);
	/// Conversion from a 4x4 matrix
	sdpEulerAngles(sdpMatrix4& Matrix, AngleOrder Order);

	/// Returns the frame type from a packed order representation	
	static OrderFrame Frame(AngleOrder& Order) { return OrderFrame(Order & 1); }
	/// Returns the repetition type from a packed order representation
	static OrderRepetition Repetition(AngleOrder& Order) { return OrderRepetition((Order >> 1) & 1); }
	/// Returns the parity type from a packed order representation
	static OrderParity Parity(AngleOrder& Order) { return OrderParity((Order >> 2) & 1); }
	/// Returns the axes in order, from a packed order representation
	static void Axes(AngleOrder& Order, gint& i, gint& j, gint& k);
	
	/// Returns the frame type of this instance
	OrderFrame Frame() { return Frame(order); }
	/// Returns the repetition type of this instance
	OrderRepetition Repetition() { return Repetition(order); }
	/// Returns the parity type of this instance
	OrderParity Parity() { return Parity(order); }
	/// Returns the axes in order, for this instance
	void Axes(gint& i, gint& j, gint& k) { Axes(order, i, j, k); }
	
	/// Returns the given angle by reference
	gdouble& operator [] ( gint i);
	/// Returns the given angle by value
	gdouble operator[] (gint i) const;
};

/// Encapsulates a 3D plane surface
class sdpPlane
{
public:
	/// Constructs a plane from its normal and distance from the origin
	sdpPlane(const sdpVector3& Normal, const gdouble Distance);
	/// Constructs a plane from its normal and one point located within the plane
	sdpPlane(const sdpVector3& Normal, const sdpVector3& ContainedPoint);
	/// Constructs a plane from three points contained within the plane
	sdpPlane(const sdpVector3& ContainedPointA, const sdpVector3& ContainedPointB, const sdpVector3& ContainedPointC);

	/// Returns the plane normal vector
	sdpVector3 Normal() const;
	/// Returns the plane's distance from the origin
	gdouble Distance() const;

	/// Sets the plane normal vector
	void SetNormal(const sdpVector3& Normal);
	/// Sets the plane's distance from the origin
	void SetDistance(gdouble Distance);

protected:
	sdpVector3 m_Normal;
	gdouble m_Distance;
};

//////////////////////////////////////////////////////////////////
// 2D and 3D functions

/// Returns a two-dimensional identity matrix
sdpMatrix3 sdpIdentity2D();
/// Returns a three-dimensional identity matrix
sdpMatrix4 sdpIdentity3D();
/// Returns a two-dimensional translation matrix
sdpMatrix3 sdpTranslation2D(const sdpVector2& v);
/// Returns a three-dimensional translation matrix
sdpMatrix4 sdpTranslation3D(const sdpVector3& v);
/// Returns a two-dimensional rotation matrix
sdpMatrix3 sdpRotation2D(const sdpVector2& Center, const gdouble Angle);
/// Returns a three-dimensional rotation matrix about an arbitrary axis
sdpMatrix4 sdpRotation3D(const gdouble Angle, sdpVector3 Axis);
/// Returns a three-dimensional rotation matrix about an arbitrary axis
sdpMatrix4 sdpRotation3D(const sdpAngleAxis& AngleAxis);
/// Returns a three-dimensional rotation matrix based on a set of Euler angles
sdpMatrix4 sdpRotation3D(const sdpVector3& EulerAngles);
/// Returns a three-dimensional rotation matrix based on a quaternion
sdpMatrix4 sdpRotation3D(const sdpQuaternion& Quaternion);
/// Converts a 3x3 rotation matrix into 4x4
sdpMatrix4 sdpRotation3D(const sdpMatrix3& m);
/// Returns a two-dimensional scaling matrix
sdpMatrix3 sdpScaling2D(const sdpVector2& scaleVector);
/// Returns a three-dimensional scaling matrix
sdpMatrix4 sdpScaling3D(const sdpVector3& scaleVector);
/// Returns a three-dimensional perspective matrix
sdpMatrix4 sdpPerspective3D(const gdouble d);
/// Extracts rotation from a three-dimensional matrix
sdpMatrix3 sdpExtractRotation(const sdpMatrix4& pose);
/// Extracts translation from a three-dimensional matrix
sdpVector3 sdpExtractTranslation(const sdpMatrix4& pose);
/// Extracts scaling from a three-dimensional matrix
sdpVector3 sdpExtractScaling(const sdpMatrix4& pose);



////////////////////////////////////////////////////////////////////
// sdpVector2 implementation

inline sdpVector2::sdpVector2()
{ n[VX] = n[VY] = 0.0; }

inline sdpVector2::sdpVector2(const gdouble x, const gdouble y)
{ n[VX] = x; n[VY] = y; }

inline sdpVector2::sdpVector2(const gdouble d)
{ n[VX] = n[VY] = d; }

inline sdpVector2::sdpVector2(const gdouble d[2])
{ n[VX] = d[0]; n[VY] = d[1]; }

inline sdpVector2::sdpVector2(const sdpVector2& v)
{ n[VX] = v.n[VX]; n[VY] = v.n[VY]; }

inline sdpVector2::sdpVector2(const sdpVector3& v) // it is up to caller to avoid divide-by-zero
{ n[VX] = v.n[VX]/v.n[VZ]; n[VY] = v.n[VY]/v.n[VZ]; }

inline sdpVector2::sdpVector2(const sdpVector3& v, gint dropAxis) {
 switch (dropAxis) {
	case VX: n[VX] = v.n[VY]; n[VY] = v.n[VZ]; break;
	case VY: n[VX] = v.n[VX]; n[VY] = v.n[VZ]; break;
	default: n[VX] = v.n[VX]; n[VY] = v.n[VY]; break;
 }
}

inline sdpVector2& sdpVector2::operator = (const sdpVector2& v)
{ n[VX] = v.n[VX]; n[VY] = v.n[VY]; return *this; }

inline sdpVector2& sdpVector2::operator = (const gdouble d[2])
{ n[VX] = d[0]; n[VY] = d[1]; return *this; }

inline sdpVector2& sdpVector2::operator += ( const sdpVector2& v )
{ n[VX] += v.n[VX]; n[VY] += v.n[VY]; return *this; }

inline sdpVector2& sdpVector2::operator -= ( const sdpVector2& v )
{ n[VX] -= v.n[VX]; n[VY] -= v.n[VY]; return *this; }

inline sdpVector2& sdpVector2::operator *= ( const gdouble d )
{ n[VX] *= d; n[VY] *= d; return *this; }

inline sdpVector2& sdpVector2::operator /= ( const gdouble d )
{ gdouble d_inv = 1./d; n[VX] *= d_inv; n[VY] *= d_inv; return *this; }

inline gdouble& sdpVector2::operator [] ( gint i)
{
 g_assert(!(i < VX || i > VY));
 return n[i];
}

inline gdouble sdpVector2::operator [] ( gint i) const
{
 g_assert(!(i < VX || i > VY));
 return n[i];
}

inline gdouble sdpVector2::Length() const
{ return sqrt(Length2()); }

inline gdouble sdpVector2::Length2() const
{ return n[VX]*n[VX] + n[VY]*n[VY]; }

inline sdpVector2& sdpVector2::Normalize()
{
	gdouble length = Length();
	if(length)
		*this /= length;
	
	return *this;
}

inline sdpVector2 sdpVector2::Perpendicular() const
{
	return sdpVector2(-n[1], n[0]);
}

inline glong sdpVector2::Quadrant()
{
	if(n[VX] < 0.0)
		return n[VY] < 0.0 ? 2 : 1;

	return n[VY] < 0.0 ? 3 : 0;
}

inline gdouble sdpVector2::Angle()
{
	// atan2() returns a result between -180.0 and 180.0 degrees ...
	return atan2(n[VY], n[VX]);
}

inline void sdpVector2::CopyArray(gfloat f[2])
{	f[0] = gfloat(n[0]); f[1] = gfloat(n[1]); }

inline void sdpVector2::CopyArray(gdouble d[2])
{	d[0] = n[0]; d[1] = n[1]; }

inline bool sdpIsInside(sdpVector2& Point, sdpVector2Array& Array)
{
	// Sanity check ...
	g_return_val_if_fail(Array.size(), false);

	glong winding = 0;

	// Close the set defined by this list ...
	Array.push_back(Array[0]);

	// Get the first point ...
	sdpVector2 testpoint(Array[0] - Point);
	glong lastquadrant = testpoint.Quadrant();

	// For each remaining point ...
	for(gulong i = 1; i < Array.size(); i++)
		{
			testpoint = sdpVector2(Array[i] - Point);
			glong quadrant = testpoint.Quadrant();
			if(quadrant == lastquadrant) continue;

			if(((lastquadrant + 1) & 3) == quadrant)
				{
					winding++;
				}
			else if(((quadrant + 1) & 3) == lastquadrant)
				{
					winding--;
				}
			else
				{
					gdouble right = Array[i-1][0] - Array[i][0];
					gdouble left = (Array[i-1][1] - Array[i][1]) * ((Point)[0] - Array[i-1][0]) + (Array[i-1][1] * right);
					winding += (left >= right * (Point)[1]) ? 2 : -2;
				}

			lastquadrant = quadrant;
		}

	return winding ? true : false;
}

inline sdpVector2 sdpClosestPoint(sdpVector2& Point, sdpVector2Array& Array)
{
	sdpVector2 result;
	gdouble distance2;
	gdouble distance = DBL_MAX;

	for(gulong i = 0; i < Array.size(); i++)
		{
			distance2 = (Array[i] - Point).Length2();
			if(distance2 <= distance)
				{
					distance = distance2;
					result = Array[i];
				}
		}

	return result;
}

inline sdpVector2 operator - (const sdpVector2& a)
{ return sdpVector2(-a.n[VX],-a.n[VY]); }

inline sdpVector2 operator + (const sdpVector2& a, const sdpVector2& b)
{ return sdpVector2(a.n[VX]+ b.n[VX], a.n[VY] + b.n[VY]); }

inline sdpVector2 operator - (const sdpVector2& a, const sdpVector2& b)
{ return sdpVector2(a.n[VX]-b.n[VX], a.n[VY]-b.n[VY]); }

inline sdpVector2 operator * (const sdpVector2& a, const gdouble d)
{ return sdpVector2(d*a.n[VX], d*a.n[VY]); }

inline sdpVector2 operator * (const gdouble d, const sdpVector2& a)
{ return a*d; }

inline sdpVector2 operator * (const sdpMatrix3& a, const sdpVector2& v) {
 sdpVector3 av;

 av.n[VX] = a.v[0].n[VX]*v.n[VX] + a.v[0].n[VY]*v.n[VY] + a.v[0].n[VZ];
 av.n[VY] = a.v[1].n[VX]*v.n[VX] + a.v[1].n[VY]*v.n[VY] + a.v[1].n[VZ];
 av.n[VZ] = a.v[2].n[VX]*v.n[VX] + a.v[2].n[VY]*v.n[VY] + a.v[2].n[VZ];
 return av;
}

inline sdpVector2 operator * (const sdpVector2& v, const sdpMatrix3& a)
{ return a.Transpose() * v; }

inline gdouble operator * (const sdpVector2& a, const sdpVector2& b)
{ return (a.n[VX]*b.n[VX] + a.n[VY]*b.n[VY]); }

inline sdpVector2 operator / (const sdpVector2& a, const gdouble d)
{ gdouble d_inv = 1./d; return sdpVector2(a.n[VX]*d_inv, a.n[VY]*d_inv); }

inline sdpVector3 operator ^ (const sdpVector2& a, const sdpVector2& b)
{ return sdpVector3(0.0, 0.0, a.n[VX] * b.n[VY] - b.n[VX] * a.n[VY]); }

inline gint operator == (const sdpVector2& a, const sdpVector2& b)
{ return (a.n[VX] == b.n[VX]) && (a.n[VY] == b.n[VY]); }

inline gint operator != (const sdpVector2& a, const sdpVector2& b)
{ return !(a == b); }

inline sdpVector2 sdpVectorMin(const sdpVector2& a, const sdpVector2& b)
{ return sdpVector2(sdpMin(a.n[VX], b.n[VX]), sdpMin(a.n[VY], b.n[VY])); }

inline sdpVector2 sdpVectorMax(const sdpVector2& a, const sdpVector2& b)
{ return sdpVector2(sdpMax(a.n[VX], b.n[VX]), sdpMax(a.n[VY], b.n[VY])); }

inline sdpVector2 sdpProduct(const sdpVector2& a, const sdpVector2& b)
{ return sdpVector2(a.n[VX] * b.n[VX], a.n[VY] * b.n[VY]); }


//////////////////////////////////////////////////////////////////////////////////////
// sdpVector3 implementation

inline sdpVector3::sdpVector3()
{ n[VX] = n[VY] = n[VZ] = 0.0; }

inline sdpVector3::sdpVector3(const gdouble x, const gdouble y, const gdouble z)
{ n[VX] = x; n[VY] = y; n[VZ] = z; }

inline sdpVector3::sdpVector3(const gdouble d)
{ n[VX] = n[VY] = n[VZ] = d; }

inline sdpVector3::sdpVector3(const gdouble d[3])
{ n[VX] = d[0]; n[VY] = d[1]; n[VZ] = d[2]; }

inline sdpVector3::sdpVector3(const sdpVector3& v)
{ n[VX] = v.n[VX]; n[VY] = v.n[VY]; n[VZ] = v.n[VZ]; }

inline sdpVector3::sdpVector3(const sdpVector2& v)
{ n[VX] = v.n[VX]; n[VY] = v.n[VY]; n[VZ] = 1.0; }

inline sdpVector3::sdpVector3(const sdpVector2& v, gdouble d)
{ n[VX] = v.n[VX]; n[VY] = v.n[VY]; n[VZ] = d; }

inline sdpVector3::sdpVector3(const sdpVector4& v) // it is up to caller to avoid divide-by-zero
{ n[VX] = v.n[VX] / v.n[VW]; n[VY] = v.n[VY] / v.n[VW];
 n[VZ] = v.n[VZ] / v.n[VW]; }

inline sdpVector3::sdpVector3(const sdpVector4& v, gint dropAxis) {
 switch (dropAxis) {
	case VX: n[VX] = v.n[VY]; n[VY] = v.n[VZ]; n[VZ] = v.n[VW]; break;
	case VY: n[VX] = v.n[VX]; n[VY] = v.n[VZ]; n[VZ] = v.n[VW]; break;
	case VZ: n[VX] = v.n[VX]; n[VY] = v.n[VY]; n[VZ] = v.n[VW]; break;
	default: n[VX] = v.n[VX]; n[VY] = v.n[VY]; n[VZ] = v.n[VZ]; break;
 }
}

inline sdpVector3& sdpVector3::operator = (const sdpVector3& v)
{ n[VX] = v.n[VX]; n[VY] = v.n[VY]; n[VZ] = v.n[VZ]; return *this; }

inline sdpVector3& sdpVector3::operator = (const gdouble d[3])
{ n[VX] = d[0]; n[VY] = d[1]; n[VZ] = d[2]; return *this; }

inline sdpVector3& sdpVector3::operator += ( const sdpVector3& v )
{ n[VX] += v.n[VX]; n[VY] += v.n[VY]; n[VZ] += v.n[VZ]; return *this; }

inline sdpVector3& sdpVector3::operator -= ( const sdpVector3& v )
{ n[VX] -= v.n[VX]; n[VY] -= v.n[VY]; n[VZ] -= v.n[VZ]; return *this; }

inline sdpVector3& sdpVector3::operator *= ( const gdouble d )
{ n[VX] *= d; n[VY] *= d; n[VZ] *= d; return *this; }

inline sdpVector3& sdpVector3::operator /= ( const gdouble d )
{ gdouble d_inv = 1./d; n[VX] *= d_inv; n[VY] *= d_inv; n[VZ] *= d_inv;
 return *this; }

inline gdouble& sdpVector3::operator [] ( gint i) {
 g_assert(! (i < VX || i > VZ));
 return n[i];
}

inline gdouble sdpVector3::operator [] ( gint i) const {
 g_assert(! (i < VX || i > VZ));
 return n[i];
}

inline gdouble sdpVector3::Length() const
{ return sqrt(Length2()); }

inline gdouble sdpVector3::Length2() const
{ return n[VX]*n[VX] + n[VY]*n[VY] + n[VZ]*n[VZ]; }

inline sdpVector3& sdpVector3::Normalize()
{
	gdouble length = Length();
	if(length)
		*this /= length;
	
	return *this;
}

inline void sdpVector3::CopyArray(gfloat f[3])
{	f[0] = (gfloat)n[0];	f[1] = (gfloat)n[1];	f[2] = (gfloat)n[2]; }

inline void sdpVector3::CopyArray(gdouble d[3])
{	d[0] = n[0]; d[1] = n[1]; d[2] = n[2]; }

inline sdpVector3& sdpVector3::CartesianToSpherical()
{
	// Handle the singularity at the poles
	if(!n[VX] && !n[VZ])
		{
			n[VZ] = Length();
			n[VX] = n[VY] >= 0.0 ? 90.0 : -90.0;
			n[VY] = 0.0;
		}
	else
		{
			sdpVector3 temp = *this;

			n[VZ] = temp.Length();
			n[VX] = -sdpDegrees(atan2(temp[VY], sqrt(temp[VX] * temp[VX] + temp[VZ] * temp[VZ])));
			n[VY] = sdpDegrees(atan2(temp[VX], temp[VZ]));
		}

	return *this;
}

inline sdpVector3& sdpVector3::RGBtoHSV()
{
	gdouble maxcomponent = sdpMax(sdpMax(n[0], n[1]), n[2]);
	gdouble mincomponent = sdpMin(sdpMin(n[0], n[1]), n[2]);
	gdouble difference = maxcomponent - mincomponent;

	sdpVector3 result;

	// Set value equal to the largest RGB component
	result[2] = maxcomponent;

	// Set saturation
	result[1] = (0.0 == maxcomponent) ? 0.0 : difference / maxcomponent;

	// If saturation is zero, assign Hue a reasonable default ...
	if(0.0 == result[1])
		{
			result[0] = 0.0;
		}
	else
		{
			gdouble reddistance = (maxcomponent - n[0]) / difference;
			gdouble greendistance = (maxcomponent - n[1]) / difference;
			gdouble bluedistance = (maxcomponent - n[2]) / difference;

			if(n[0] == maxcomponent)
				{
					result[0] = bluedistance - greendistance;
				}
			else if(n[1] == maxcomponent)
				{
					result[0] = 2 + reddistance - bluedistance;
				}
			else
				{
					result[0] = 4 + greendistance - reddistance;
				}

			result[0] *= 60.0;
			if(result[0] < 0.0) result[0] += 360.0;
		}

	*this = result;
	return *this;
}

inline sdpVector3& sdpVector3::HSVtoRGB()
{
	sdpVector3 result;

	if(n[1] == 0.0)	// If saturation is zero, the result is greyscale
		{
			result = n[2];
		}
	else
		{
			if(360.0 == n[0]) n[0] = 0.0;

			gdouble h = n[0] / 60.0;
			gdouble i = floor(h);
			gdouble f = h - i;
			gdouble p = n[2] * (1 - n[1]);
			gdouble q = n[2] * (1 - (n[1] * f));
			gdouble t = n[2] * (1 - (n[1] * (1 - f)));

			if(0.0 == i)
				{
					result[0] = n[2];
					result[1] = t;
					result[2] = p;
				}
			else if(1.0 == i)
				{
					result[0] = q;
					result[1] = n[2];
					result[2] = p;
				}
			else if(2.0 == i)
				{
					result[0] = p;
					result[1] = n[2];
					result[2] = t;
				}
			else if(3.0 == i)
				{
					result[0] = p;
					result[1] = q;
					result[2] = n[2];
				}
			else if(4.0 == i)
				{
					result[0] = t;
					result[1] = p;
					result[2] = n[2];
				}
			else if(5.0 == i)
				{
					result[0] = n[2];
					result[1] = p;
					result[2] = q;
				}
		}

	*this = result;
	return *this;
}

inline sdpVector3& sdpVector3::VideoSafe(bool NTSC, bool ReduceSat)
{
	const gdouble chroma_limit = 50.0;			// chroma amplitude limit
	const gdouble composite_limit = 110.0;	// max IRE amplitude

	const gdouble ntsc_pedestal = 7.5;			// 7.5 IRE black pedestal
	const gdouble ntsc_gamma = 2.2;

	const gdouble pal_pedestal = 0.0;				// no pedestal in PAL
	const gdouble pal_gamma = 2.8;

	gdouble y, y2, i, q, c, c2, py, t, scale;
	gdouble chroma_lim, chroma_lim2, compos_lim;
	gdouble pedestal, gamma;

	if(NTSC)
		{
			pedestal = ntsc_pedestal;
			gamma = ntsc_gamma;

			y = 0.2989 * n[0] + 0.5866 * n[1] + 0.1144 * n[2];
			i = 0.5959 * n[0] + -0.2741 * n[1] + -0.3218 * n[2];
			q = 0.2113 * n[0] + -0.5227 * n[1] + 0.3113 * n[2];
		}
	else
		{
			pedestal = pal_pedestal;
			gamma = pal_gamma;

			y = 0.2989 * n[0] + 0.5866 * n[1] + 0.1144 * n[2];
			i = -0.1473 * n[0] + -0.2891 * n[1] + 0.4364 * n[2];
			q = 0.6149 * n[0] + -0.5145 * n[1] + -0.1004 * n[2];
		}

	chroma_lim = (gdouble)chroma_limit / (100.0 - pedestal);
	chroma_lim2 = chroma_lim * chroma_lim;
	compos_lim = ((gdouble)composite_limit - pedestal) / (100.0 - pedestal);

	c2 = i * i + q * q;
	y2 = compos_lim - y;
	y2 *= y2;
	if(c2 <= chroma_lim2 && c2 <= y2)	// no problems
		{
			return *this;
		}

	c = sqrt(c2);

	if(ReduceSat)	// Reduce saturation to bring within limits
		{
			scale = chroma_lim / c;
			t = (compos_lim - y) / c;
			if (t < scale) scale = t;

			n[0] = pow(n[0], 1.0 / gamma);
			n[1] = pow(n[1], 1.0 / gamma);
			n[2] = pow(n[2], 1.0 / gamma);

			py = n[0] * 0.2989 + n[1] * 0.5866 + n[2] * 0.1144;

			n[0] = pow((py + scale * (n[0] - py)), gamma);
			n[1] = pow((py + scale * (n[1] - py)), gamma);
			n[2] = pow((py + scale * (n[2] - py)), gamma);
		}
	else // Reduce value
		{
			scale = chroma_lim / c;
			t = compos_lim / (y + c);
			if (t < scale) scale = t;
			scale = pow(scale, gamma);

			n[0] = scale * n[0];
			n[1] = scale * n[1];
			n[2] = scale * n[2];
		}

	return *this;
}

inline sdpVector3& sdpVector3::Abs()
{	n[0] = fabs(n[0]); n[1] = fabs(n[1]); n[2] = fabs(n[2]); return *this; }

inline sdpVector3 operator - (const sdpVector3& a)
{ return sdpVector3(-a.n[VX],-a.n[VY],-a.n[VZ]); }

inline sdpVector3 operator + (const sdpVector3& a, const sdpVector3& b)
{ return sdpVector3(a.n[VX]+ b.n[VX], a.n[VY] + b.n[VY], a.n[VZ] + b.n[VZ]); }

inline sdpVector3 operator - (const sdpVector3& a, const sdpVector3& b)
{ return sdpVector3(a.n[VX]-b.n[VX], a.n[VY]-b.n[VY], a.n[VZ]-b.n[VZ]); }

inline sdpVector3 operator * (const sdpVector3& a, const gdouble d)
{ return sdpVector3(d*a.n[VX], d*a.n[VY], d*a.n[VZ]); }

inline sdpVector3 operator * (const gdouble d, const sdpVector3& a)
{ return a*d; }

inline sdpVector3 operator * (const sdpMatrix4& a, const sdpVector3& v)
{ return a * sdpVector4(v); }

inline sdpVector3 operator * (const sdpVector3& v, const sdpMatrix4& a)
{ return a.Transpose() * v; }

inline gdouble operator * (const sdpVector3& a, const sdpVector3& b)
{ return (a.n[VX]*b.n[VX] + a.n[VY]*b.n[VY] + a.n[VZ]*b.n[VZ]); }

inline sdpVector3 operator / (const sdpVector3& a, const gdouble d)
{ gdouble d_inv = 1./d; return sdpVector3(a.n[VX]*d_inv, a.n[VY]*d_inv,
 a.n[VZ]*d_inv); }

inline sdpVector3 operator ^ (const sdpVector3& a, const sdpVector3& b) {
 return sdpVector3(a.n[VY]*b.n[VZ] - a.n[VZ]*b.n[VY],
		a.n[VZ]*b.n[VX] - a.n[VX]*b.n[VZ],
		a.n[VX]*b.n[VY] - a.n[VY]*b.n[VX]);
}

inline gint operator == (const sdpVector3& a, const sdpVector3& b)
{ return (a.n[VX] == b.n[VX]) && (a.n[VY] == b.n[VY]) && (a.n[VZ] == b.n[VZ]);
}

inline gint operator != (const sdpVector3& a, const sdpVector3& b)
{ return !(a == b); }

inline sdpVector3 sdpVectorMin(const sdpVector3& a, const sdpVector3& b)
{ return sdpVector3(sdpMin(a.n[VX], b.n[VX]), sdpMin(a.n[VY], b.n[VY]), sdpMin(a.n[VZ],
 b.n[VZ])); }

inline sdpVector3 sdpVectorMax(const sdpVector3& a, const sdpVector3& b)
{ return sdpVector3(sdpMax(a.n[VX], b.n[VX]), sdpMax(a.n[VY], b.n[VY]), sdpMax(a.n[VZ],
 b.n[VZ])); }

inline sdpVector3 sdpProduct(const sdpVector3& a, const sdpVector3& b)
{ return sdpVector3(a.n[VX] * b.n[VX], a.n[VY] * b.n[VY], a.n[VZ] * b.n[VZ]); }


////////////////////////////////////////////////////////////////////
// sdpVector4 implementation

inline sdpVector4::sdpVector4()
{ n[VX] = n[VY] = n[VZ] = n[VW] = 0.0; }

inline sdpVector4::sdpVector4(const gdouble x, const gdouble y, const gdouble z, const gdouble w)
{ n[VX] = x; n[VY] = y; n[VZ] = z; n[VW] = w; }

inline sdpVector4::sdpVector4(const gdouble d)
{ n[VX] = n[VY] = n[VZ] = n[VW] = d; }

inline sdpVector4::sdpVector4(const gdouble d[4])
{ n[VX] = d[0]; n[VY] = d[1]; n[VZ] = d[2]; n[VW] = d[3]; }

inline sdpVector4::sdpVector4(const sdpVector4& v)
{ n[VX] = v.n[VX]; n[VY] = v.n[VY]; n[VZ] = v.n[VZ]; n[VW] = v.n[VW]; }

inline sdpVector4::sdpVector4(const sdpVector3& v)
{ n[VX] = v.n[VX]; n[VY] = v.n[VY]; n[VZ] = v.n[VZ]; n[VW] = 1.0; }

inline sdpVector4::sdpVector4(const sdpVector3& v, const gdouble d)
{ n[VX] = v.n[VX]; n[VY] = v.n[VY]; n[VZ] = v.n[VZ]; n[VW] = d; }

inline sdpVector4& sdpVector4::operator = (const sdpVector4& v)
{ n[VX] = v.n[VX]; n[VY] = v.n[VY]; n[VZ] = v.n[VZ]; n[VW] = v.n[VW];
return *this; }

inline sdpVector4& sdpVector4::operator = (const gdouble d[4])
{ n[VX] = d[0]; n[VY] = d[1]; n[VZ] = d[2]; n[VW] = d[3]; return *this; }

inline sdpVector4& sdpVector4::operator += ( const sdpVector4& v )
{ n[VX] += v.n[VX]; n[VY] += v.n[VY]; n[VZ] += v.n[VZ]; n[VW] += v.n[VW];
return *this; }

inline sdpVector4& sdpVector4::operator -= ( const sdpVector4& v )
{ n[VX] -= v.n[VX]; n[VY] -= v.n[VY]; n[VZ] -= v.n[VZ]; n[VW] -= v.n[VW];
return *this; }

inline sdpVector4& sdpVector4::operator *= ( const gdouble d )
{ n[VX] *= d; n[VY] *= d; n[VZ] *= d; n[VW] *= d; return *this; }

inline sdpVector4& sdpVector4::operator /= ( const gdouble d )
{ gdouble d_inv = 1./d; n[VX] *= d_inv; n[VY] *= d_inv; n[VZ] *= d_inv;
 n[VW] *= d_inv; return *this; }

inline gdouble& sdpVector4::operator [] ( gint i) {
 g_assert(! (i < VX || i > VW));
 return n[i];
}

inline gdouble sdpVector4::operator [] ( gint i) const {
 g_assert(! (i < VX || i > VW));
 return n[i];
}

inline gdouble sdpVector4::Length() const
{ return sqrt(Length2()); }

inline gdouble sdpVector4::Length2() const
{ return n[VX]*n[VX] + n[VY]*n[VY] + n[VZ]*n[VZ] + n[VW]*n[VW]; }

inline sdpVector4& sdpVector4::Normalize()
{
	gdouble length = Length();
	if(length)
		*this /= length;
	
	return *this;
}

inline sdpVector4 operator - (const sdpVector4& a)
{ return sdpVector4(-a.n[VX],-a.n[VY],-a.n[VZ],-a.n[VW]); }

inline sdpVector4 operator + (const sdpVector4& a, const sdpVector4& b)
{ return sdpVector4(a.n[VX] + b.n[VX], a.n[VY] + b.n[VY], a.n[VZ] + b.n[VZ],
 a.n[VW] + b.n[VW]); }

inline sdpVector4 operator - (const sdpVector4& a, const sdpVector4& b)
{ return sdpVector4(a.n[VX] - b.n[VX], a.n[VY] - b.n[VY], a.n[VZ] - b.n[VZ],
 a.n[VW] - b.n[VW]); }

inline sdpVector4 operator * (const sdpVector4& a, const gdouble d)
{ return sdpVector4(d*a.n[VX], d*a.n[VY], d*a.n[VZ], d*a.n[VW] ); }

inline sdpVector4 operator * (const gdouble d, const sdpVector4& a)
{ return a*d; }

inline sdpVector4 operator * (const sdpMatrix4& a, const sdpVector4& v)
{
#define ROWCOL(i) a.v[i].n[0]*v.n[VX] + a.v[i].n[1]*v.n[VY] + a.v[i].n[2]*v.n[VZ] + a.v[i].n[3]*v.n[VW]
 return sdpVector4(ROWCOL(0), ROWCOL(1), ROWCOL(2), ROWCOL(3));
#undef ROWCOL // (i)
}

inline sdpVector4 operator * (const sdpVector4& v, const sdpMatrix4& a)
{ return a.Transpose() * v; }

inline gdouble operator * (const sdpVector4& a, const sdpVector4& b)
{ return (a.n[VX]*b.n[VX] + a.n[VY]*b.n[VY] + a.n[VZ]*b.n[VZ] +
 a.n[VW]*b.n[VW]); }

inline sdpVector4 operator / (const sdpVector4& a, const gdouble d)
{ gdouble d_inv = 1./d; return sdpVector4(a.n[VX]*d_inv, a.n[VY]*d_inv, a.n[VZ]*d_inv,
 a.n[VW]*d_inv); }

inline gint operator == (const sdpVector4& a, const sdpVector4& b)
{ return (a.n[VX] == b.n[VX]) && (a.n[VY] == b.n[VY]) && (a.n[VZ] == b.n[VZ])
 && (a.n[VW] == b.n[VW]); }

inline gint operator != (const sdpVector4& a, const sdpVector4& b)
{ return !(a == b); }

inline sdpVector4 sdpVectorMin(const sdpVector4& a, const sdpVector4& b)
{ return sdpVector4(sdpMin(a.n[VX], b.n[VX]), sdpMin(a.n[VY], b.n[VY]), sdpMin(a.n[VZ],
 b.n[VZ]), sdpMin(a.n[VW], b.n[VW])); }

inline sdpVector4 sdpVectorMax(const sdpVector4& a, const sdpVector4& b)
{ return sdpVector4(sdpMax(a.n[VX], b.n[VX]), sdpMax(a.n[VY], b.n[VY]), sdpMax(a.n[VZ],
 b.n[VZ]), sdpMax(a.n[VW], b.n[VW])); }

inline sdpVector4 sdpProduct(const sdpVector4& a, const sdpVector4& b)
{ return sdpVector4(a.n[VX] * b.n[VX], a.n[VY] * b.n[VY], a.n[VZ] * b.n[VZ],
 a.n[VW] * b.n[VW]); }


////////////////////////////////////////////////////////////////////
// sdpRectangle implementation

inline sdpRectangle::sdpRectangle()
{ left = right = top = bottom = 0.0; }

inline sdpRectangle::sdpRectangle(const gdouble Left, const gdouble Right, const gdouble Top, const gdouble Bottom)
{ left = Left; right = Right; top = Top; bottom = Bottom; }

inline sdpRectangle::sdpRectangle(const sdpRectangle& v)
{ left = v.left; right = v.right; top = v.top; bottom = v.bottom; }

inline sdpRectangle& sdpRectangle::operator = (const sdpRectangle& v)
{
	left = v.left; right = v.right; top = v.top; bottom = v.bottom;
	return *this;
}

inline gdouble sdpRectangle::Left() { return left; }
inline gdouble sdpRectangle::Right() { return right; }
inline gdouble sdpRectangle::Top() { return top; }
inline gdouble sdpRectangle::Bottom() { return bottom; }
inline gdouble sdpRectangle::Width() { return fabs(right - left); }
inline gdouble sdpRectangle::Height() { return fabs(bottom - top); }

inline void sdpRectangle::Offset(const gdouble Horizontal, const gdouble Vertical)
{
	left += Horizontal;
	right += Horizontal;
	top += Vertical;
	bottom += Vertical;
}

inline void sdpRectangle::Grow(const gdouble Thickness)
{
	Grow(Thickness, Thickness);
}

inline void sdpRectangle::Grow(const gdouble Horizontal, const gdouble Vertical)
{
	left -= Horizontal;
	right += Horizontal;
	top -= Vertical;
	bottom += Vertical;
}

inline void sdpRectangle::Shrink(const gdouble Thickness)
{
	Shrink(Thickness, Thickness);
}

inline void sdpRectangle::Shrink(const gdouble Horizontal, const gdouble Vertical)
{
	Grow(-Horizontal, -Vertical);
}

inline void sdpRectangle::Normalize()
{
	if(left > right)
		std::swap(left, right);

	if(top > bottom)
		std::swap(top, bottom);
}

inline sdpVector2 sdpRectangle::Center()
{
	return sdpVector2(left+right, top+bottom) * 0.5;
}

inline bool sdpRectangle::Contains(const sdpVector2& Point)
{
	return (left <= Point[0] && Point[0] <= right && top <= Point[1] && Point[1] <= bottom) ? true : false;
}

inline sdpRectangle sdpConstrainProportions(const sdpRectangle& Source, const sdpRectangle& Destination)
{
	sdpRectangle source(Source);
	sdpRectangle destination(Destination);

	source.Normalize();
	destination.Normalize();

	source.Offset(-source.Left(), -source.Top());

	if(source.Width() != destination.Width())
		{
			source.bottom *= destination.Width() / source.Width();
			source.right *= destination.Width() / source.Width();
		}

	if(source.Height() > destination.Height())
		{
			source.right *= destination.Height() / source.Height();
			source.bottom *= destination.Height() / source.Height();
		}

	source.Offset(destination.Left() + (0.5 * (destination.Width() - source.Width())),
								destination.Top() + (0.5 * (destination.Height() - source.Height())));

	return source;
}

////////////////////////////////////////////////////////////////////
// sdpMatrix3 implementation

inline sdpMatrix3::sdpMatrix3() {}

inline sdpMatrix3::sdpMatrix3(const sdpVector3& v0, const sdpVector3& v1, const sdpVector3& v2)
{ v[0] = v0; v[1] = v1; v[2] = v2; }

inline sdpMatrix3::sdpMatrix3(const gdouble d)
{ v[0] = v[1] = v[2] = sdpVector3(d); }

inline sdpMatrix3::sdpMatrix3(const sdpMatrix3& m)
{ v[0] = m.v[0]; v[1] = m.v[1]; v[2] = m.v[2]; }

inline sdpMatrix3& sdpMatrix3::operator = ( const sdpMatrix3& m )
{ v[0] = m.v[0]; v[1] = m.v[1]; v[2] = m.v[2]; return *this; }

inline sdpMatrix3& sdpMatrix3::operator += ( const sdpMatrix3& m )
{ v[0] += m.v[0]; v[1] += m.v[1]; v[2] += m.v[2]; return *this; }

inline sdpMatrix3& sdpMatrix3::operator -= ( const sdpMatrix3& m )
{ v[0] -= m.v[0]; v[1] -= m.v[1]; v[2] -= m.v[2]; return *this; }

inline sdpMatrix3& sdpMatrix3::operator *= ( const gdouble d )
{ v[0] *= d; v[1] *= d; v[2] *= d; return *this; }

inline sdpMatrix3& sdpMatrix3::operator /= ( const gdouble d )
{ v[0] /= d; v[1] /= d; v[2] /= d; return *this; }

inline sdpVector3& sdpMatrix3::operator [] ( gint i) {
 g_assert(! (i < VX || i > VZ));
 return v[i];
}

inline const sdpVector3& sdpMatrix3::operator [] ( gint i) const {
 g_assert(!(i < VX || i > VZ));
 return v[i];
}

inline sdpMatrix3 sdpMatrix3::Transpose() const {
 return sdpMatrix3(sdpVector3(v[0][0], v[1][0], v[2][0]),
		sdpVector3(v[0][1], v[1][1], v[2][1]),
		sdpVector3(v[0][2], v[1][2], v[2][2]));
}

inline sdpMatrix3 sdpMatrix3::Inverse()	const // Gauss-Jordan elimination with partial pivoting
 {
 sdpMatrix3 a(*this),	 // As a evolves from original mat into identity
	 b(sdpIdentity2D()); // b evolves from identity into Inverse(a)
 gint	 i, j, i1;

 // Loop over cols of a from left to right, eliminating above and below diag
 for (j=0; j<3; j++) { // Find largest pivot in column j among rows j..2
 i1 = j;		 // Row with largest pivot candidate
 for (i=j+1; i<3; i++)
	if (fabs(a.v[i].n[j]) > fabs(a.v[i1].n[j]))
	 i1 = i;

 // sdpSwap rows i1 and j in a and b to put pivot on diagonal
 std::swap(a.v[i1], a.v[j]);
 std::swap(b.v[i1], b.v[j]);

 // Scale row j to have a unit diagonal
 g_return_val_if_fail(a.v[j].n[j], b);	// Singular matrix - can't invert

 b.v[j] /= a.v[j].n[j];
 a.v[j] /= a.v[j].n[j];

 // Eliminate off-diagonal elems in col j of a, doing identical ops to b
 for (i=0; i<3; i++)
	if (i!=j) {
	b.v[i] -= a.v[i].n[j]*b.v[j];
	a.v[i] -= a.v[i].n[j]*a.v[j];
	}
 }
 return b;
}

inline sdpMatrix3 operator - (const sdpMatrix3& a)
{ return sdpMatrix3(-a.v[0], -a.v[1], -a.v[2]); }


inline sdpMatrix3 operator + (const sdpMatrix3& a, const sdpMatrix3& b)
{ return sdpMatrix3(a.v[0] + b.v[0], a.v[1] + b.v[1], a.v[2] + b.v[2]); }

inline sdpMatrix3 operator - (const sdpMatrix3& a, const sdpMatrix3& b)
{ return sdpMatrix3(a.v[0] - b.v[0], a.v[1] - b.v[1], a.v[2] - b.v[2]); }

inline sdpMatrix3 operator * (const sdpMatrix3& a, const sdpMatrix3& b)
{
#define ROWCOL(i, j) a.v[i].n[0]*b.v[0][j] + a.v[i].n[1]*b.v[1][j] + a.v[i].n[2]*b.v[2][j]
	return sdpMatrix3(sdpVector3(ROWCOL(0,0), ROWCOL(0,1), ROWCOL(0,2)),
					sdpVector3(ROWCOL(1,0), ROWCOL(1,1), ROWCOL(1,2)),
					sdpVector3(ROWCOL(2,0), ROWCOL(2,1), ROWCOL(2,2)));
#undef ROWCOL // (i, j)
}

inline sdpMatrix3 operator * (const sdpMatrix3& a, const gdouble d)
{ return sdpMatrix3(a.v[0] * d, a.v[1] * d, a.v[2] * d); }

inline sdpMatrix3 operator * (const gdouble d, const sdpMatrix3& a)
{ return a*d; }

inline sdpMatrix3 operator / (const sdpMatrix3& a, const gdouble d)
{ return sdpMatrix3(a.v[0] / d, a.v[1] / d, a.v[2] / d); }

inline gint operator == (const sdpMatrix3& a, const sdpMatrix3& b)
{ return (a.v[0] == b.v[0]) && (a.v[1] == b.v[1]) && (a.v[2] == b.v[2]); }

inline gint operator != (const sdpMatrix3& a, const sdpMatrix3& b)
{ return !(a == b); }

////////////////////////////////////////////////////////////////////
// sdpMatrix4 implementation

inline sdpMatrix4::sdpMatrix4() {}

inline sdpMatrix4::sdpMatrix4(const sdpVector4& v0, const sdpVector4& v1, const sdpVector4& v2, const sdpVector4& v3)
{ v[0] = v0; v[1] = v1; v[2] = v2; v[3] = v3; }

inline sdpMatrix4::sdpMatrix4(const gdouble d)
{ v[0] = v[1] = v[2] = v[3] = sdpVector4(d); }

inline sdpMatrix4::sdpMatrix4(const sdpMatrix4& m)
{ v[0] = m.v[0]; v[1] = m.v[1]; v[2] = m.v[2]; v[3] = m.v[3]; }

inline sdpMatrix4::sdpMatrix4(sdpEulerAngles Angles)
{
	sdpEulerAngles::OrderFrame frame = Angles.Frame();
	sdpEulerAngles::OrderParity parity = Angles.Parity();
	sdpEulerAngles::OrderRepetition repetition = Angles.Repetition();
	int i, j, k;
	Angles.Axes(i, j, k);

	if(frame == sdpEulerAngles::RotatingFrame)
		std::swap(Angles.n[0], Angles.n[2]);

	if(parity == sdpEulerAngles::OddParity)
		{
			Angles.n[0] = -Angles.n[0];
			Angles.n[1] = -Angles.n[1];
			Angles.n[2] = -Angles.n[2];
		}

	double ti = Angles.n[0], tj = Angles.n[1], th = Angles.n[2];
	double ci = cos(ti), cj = cos(tj), ch = cos(th);
	double si = sin(ti), sj = sin(tj), sh = sin(th);
	double cc = ci*ch, cs = ci*sh;
	double sc = si*ch, ss = si*sh;

	sdpMatrix3 m;
	if(repetition == sdpEulerAngles::Repeats)
		{
			m[i][i] =  cj;		m[i][j] =  sj*si;	m[i][k] =  sj*ci;
			m[j][i] =  sj*sh;	m[j][j] = -cj*ss+cc;	m[j][k] = -cj*cs-sc;
			m[k][i] = -sj*ch;	m[k][j] =  cj*sc+cs;	m[k][k] =  cj*cc-ss;
		}
		else
		{
			m[i][i] =  cj*ch;	m[i][j] = sj*sc-cs;	m[i][k] = sj*cc+ss;
			m[j][i] =  cj*sh;	m[j][j] = sj*ss+cc;	m[j][k] = sj*cs-sc;
			m[k][i] = -sj;		m[k][j] = cj*si;	m[k][k] = cj*ci;
		}

	*this = sdpRotation3D(m);
}

inline sdpMatrix4& sdpMatrix4::operator = ( const sdpMatrix4& m )
{ v[0] = m.v[0]; v[1] = m.v[1]; v[2] = m.v[2]; v[3] = m.v[3];
return *this; }

inline sdpMatrix4& sdpMatrix4::operator = (const gdouble d[4][4])
{ v[0] = d[0]; v[1] = d[1]; v[2] = d[2]; v[3] = d[3]; return *this; }

inline sdpMatrix4& sdpMatrix4::operator = (const gdouble d[16])
{ memcpy(&v[0][0], &d[0], 16 * sizeof(gdouble)); return *this; }

inline sdpMatrix4& sdpMatrix4::operator += ( const sdpMatrix4& m )
{ v[0] += m.v[0]; v[1] += m.v[1]; v[2] += m.v[2]; v[3] += m.v[3];
return *this; }

inline sdpMatrix4& sdpMatrix4::operator -= ( const sdpMatrix4& m )
{ v[0] -= m.v[0]; v[1] -= m.v[1]; v[2] -= m.v[2]; v[3] -= m.v[3];
return *this; }

inline sdpMatrix4& sdpMatrix4::operator *= ( const gdouble d )
{ v[0] *= d; v[1] *= d; v[2] *= d; v[3] *= d; return *this; }

inline sdpMatrix4& sdpMatrix4::operator /= ( const gdouble d )
{ v[0] /= d; v[1] /= d; v[2] /= d; v[3] /= d; return *this; }

inline sdpVector4& sdpMatrix4::operator [] ( gint i) {
 g_assert(! (i < VX || i > VW));
 return v[i];
}

inline const sdpVector4& sdpMatrix4::operator [] ( gint i) const {
 g_assert(! (i < VX || i > VW));
 return v[i];
}

inline sdpMatrix4 sdpMatrix4::Transpose() const{
 return sdpMatrix4(sdpVector4(v[0][0], v[1][0], v[2][0], v[3][0]),
		sdpVector4(v[0][1], v[1][1], v[2][1], v[3][1]),
		sdpVector4(v[0][2], v[1][2], v[2][2], v[3][2]),
		sdpVector4(v[0][3], v[1][3], v[2][3], v[3][3]));
}

inline sdpMatrix4 sdpMatrix4::Inverse()	const // Gauss-Jordan elimination with partial pivoting
{
 sdpMatrix4 a(*this),	 // As a evolves from original mat into identity
	 b(sdpIdentity3D()); // b evolves from identity into Inverse(a)
 gint i, j, i1;

 // Loop over cols of a from left to right, eliminating above and below diag
 for (j=0; j<4; j++) { // Find largest pivot in column j among rows j..3
 i1 = j;		 // Row with largest pivot candidate
 for (i=j+1; i<4; i++)
	if (fabs(a.v[i].n[j]) > fabs(a.v[i1].n[j]))
	i1 = i;

 // sdpSwap rows i1 and j in a and b to put pivot on diagonal
 std::swap(a.v[i1], a.v[j]);
 std::swap(b.v[i1], b.v[j]);

 // Scale row j to have a unit diagonal
 g_return_val_if_fail(a.v[j].n[j], b);	// Singular matrix - can't invert!
 b.v[j] /= a.v[j].n[j];
 a.v[j] /= a.v[j].n[j];

 // Eliminate off-diagonal elems in col j of a, doing identical ops to b
 for (i=0; i<4; i++)
	if (i!=j) {
	b.v[i] -= a.v[i].n[j]*b.v[j];
	a.v[i] -= a.v[i].n[j]*a.v[j];
	}
 }
 return b;
}

inline void sdpMatrix4::CopyArray(gfloat m[16])
{	gdouble* t = &v[0][0];	for(glong i = 0; i < 16; i++) m[i] = (gfloat)t[i]; }

inline void sdpMatrix4::CopyArray(gdouble m[16])
{	memcpy(&m[0], &v[0][0], 16 * sizeof(gdouble)); }

inline sdpMatrix4 operator - (const sdpMatrix4& a)
{ return sdpMatrix4(-a.v[0], -a.v[1], -a.v[2], -a.v[3]); }

inline sdpMatrix4 operator + (const sdpMatrix4& a, const sdpMatrix4& b)
{ return sdpMatrix4(a.v[0] + b.v[0], a.v[1] + b.v[1], a.v[2] + b.v[2],
 a.v[3] + b.v[3]);
}

inline sdpMatrix4 operator - (const sdpMatrix4& a, const sdpMatrix4& b)
{ return sdpMatrix4(a.v[0] - b.v[0], a.v[1] - b.v[1], a.v[2] - b.v[2], a.v[3] - b.v[3]); }

inline sdpMatrix4 operator * (const sdpMatrix4& a, const sdpMatrix4& b)
{
#define ROWCOL(i, j) a.v[i].n[0]*b.v[0][j] + a.v[i].n[1]*b.v[1][j] + a.v[i].n[2]*b.v[2][j] + a.v[i].n[3]*b.v[3][j]
	return sdpMatrix4(sdpVector4(ROWCOL(0,0), ROWCOL(0,1), ROWCOL(0,2), ROWCOL(0,3)),
									sdpVector4(ROWCOL(1,0), ROWCOL(1,1), ROWCOL(1,2), ROWCOL(1,3)),
									sdpVector4(ROWCOL(2,0), ROWCOL(2,1), ROWCOL(2,2), ROWCOL(2,3)),
									sdpVector4(ROWCOL(3,0), ROWCOL(3,1), ROWCOL(3,2), ROWCOL(3,3)));
}

inline sdpMatrix4 operator * (const sdpMatrix4& a, const gdouble d)
{ return sdpMatrix4(a.v[0] * d, a.v[1] * d, a.v[2] * d, a.v[3] * d); }

inline sdpMatrix4 operator * (const gdouble d, const sdpMatrix4& a)
{ return a*d; }

inline sdpMatrix4 operator / (const sdpMatrix4& a, const gdouble d)
{ return sdpMatrix4(a.v[0] / d, a.v[1] / d, a.v[2] / d, a.v[3] / d); }

inline gint operator == (const sdpMatrix4& a, const sdpMatrix4& b)
{ return ((a.v[0] == b.v[0]) && (a.v[1] == b.v[1]) && (a.v[2] == b.v[2]) &&
 (a.v[3] == b.v[3])); }

inline gint operator != (const sdpMatrix4& a, const sdpMatrix4& b)
{ return !(a == b); }

////////////////////////////////////////////////////////////////////
// sdpQuaternion implementation

inline sdpQuaternion::sdpQuaternion()
{ w = 1.0; v = sdpVector3(0.0, 0.0, 0.0); }

inline sdpQuaternion::sdpQuaternion(const gdouble u, const sdpVector3& t)
{ w = u; v = t; }

inline sdpQuaternion::sdpQuaternion(const gdouble u, const gdouble x, const gdouble y, const gdouble z)
{ w = u; v = sdpVector3(x, y, z); }

inline sdpQuaternion::sdpQuaternion(const sdpAngleAxis& AngleAxis)
{ w = cos(AngleAxis.angle * 0.5); sdpVector3 axis = AngleAxis.axis; axis.Normalize(); v = sin(AngleAxis.angle * 0.5) * axis; }

inline sdpQuaternion::sdpQuaternion(sdpEulerAngles Angles)
{
	sdpEulerAngles::OrderFrame frame = Angles.Frame();
	sdpEulerAngles::OrderParity parity = Angles.Parity();
	sdpEulerAngles::OrderRepetition repetition = Angles.Repetition();
	gint i, j, k;
	Angles.Axes(i, j, k);

	if(frame == sdpEulerAngles::RotatingFrame)
		std::swap(Angles.n[0], Angles.n[2]);

	if(parity == sdpEulerAngles::OddParity)
		Angles.n[1] = -Angles.n[1];

	double ti = Angles.n[0]*0.5, tj = Angles.n[1]*0.5, th = Angles.n[2]*0.5;
	double ci = cos(ti), cj = cos(tj), ch = cos(th);
	double si = sin(ti), sj = sin(tj), sh = sin(th);
	double cc = ci*ch, cs = ci*sh;
	double sc = si*ch, ss = si*sh;

	if(repetition == sdpEulerAngles::Repeats)
		{
			v[i] = cj*(cs + sc);
			v[j] = sj*(cc + ss);
			v[k] = sj*(cs - sc);
			w = cj*(cc - ss);
		}
	else
		{
			v[i] = cj*sc - sj*cs;
			v[j] = cj*ss + sj*cc;
			v[k] = cj*cs - sj*sc;
			w = cj*cc + sj*ss;
		}

	if(parity == sdpEulerAngles::OddParity)
		v[j] = -v[j];
}

inline sdpQuaternion& sdpQuaternion::operator = (const sdpQuaternion& q)
{ w = q.w; v = q.v; return *this; }

inline sdpQuaternion& sdpQuaternion::operator = (const sdpAngleAxis& AngleAxis)
{ w = cos(AngleAxis.angle * 0.5); sdpVector3 axis = AngleAxis.axis; axis.Normalize(); v = sin(AngleAxis.angle * 0.5) * axis; return *this; }

inline sdpQuaternion& sdpQuaternion::operator *= (const gdouble f)
{ w *= f; v *= f; return *this; }

inline sdpQuaternion& sdpQuaternion::operator /= (const gdouble f)
{ if(f != 0.0) { w /= f; v /= f; } return *this; }

inline sdpQuaternion& sdpQuaternion::operator += (const sdpQuaternion& q)
{ w += q.w; v += q.v; return *this; }

inline sdpQuaternion& sdpQuaternion::operator -= (const sdpQuaternion& q)
{ w -= q.w; v -= q.v; return *this; }

inline sdpQuaternion& sdpQuaternion::operator *= (const sdpQuaternion& q)
{ w = w*q.w - (v*q.v); v = q.w*v + w*q.v + (v^q.v); return *this; }

inline sdpQuaternion& sdpQuaternion::operator /= (const sdpQuaternion& q)
{ sdpQuaternion t = q; (*this) *= t.Inverse(); return *this; }

inline gdouble sdpQuaternion::Magnitude() const
{ return (w*w + v.Length2()); }

inline sdpQuaternion& sdpQuaternion::Normalize()
{
	gdouble magnitude = Magnitude();
	if(magnitude)
		*this /= magnitude;
	
	return *this;
}

inline sdpQuaternion& sdpQuaternion::Conjugate()
{ v = -v; return *this; }

inline sdpQuaternion& sdpQuaternion::Inverse()
{ if(Magnitude() != 0.0) { double i = 1.0 / Magnitude(); w *= i; v *= -i; } return *this; }

inline sdpQuaternion& sdpQuaternion::Square()
{ double t = 2*w; w = w*w - v.Length2(); v *= t; return *this; }

inline sdpQuaternion& sdpQuaternion::AxisRotation(const gdouble phi)
{ v.Normalize(); v *= sin(phi/2.0); w = cos(phi/2); return *this; }

inline sdpQuaternion operator - (const sdpQuaternion& q)
{ return sdpQuaternion(-q.w,-q.v); }

inline sdpQuaternion operator + (const sdpQuaternion& q, const sdpQuaternion& r)
{ return sdpQuaternion(q.w + r.w, q.v + r.v); }

inline sdpQuaternion operator - (const sdpQuaternion& q, const sdpQuaternion& r)
{ return sdpQuaternion(q.w - r.w, q.v - r.v); }

inline sdpQuaternion operator * (const sdpQuaternion& q, const sdpQuaternion& r)
{ return sdpQuaternion(q.w*r.w - (q.v*r.v), q.w*r.v + r.w*q.v + (q.v^r.v)); }

inline sdpQuaternion operator * (const sdpQuaternion& q, const double f)
{ return sdpQuaternion(q.w*f, q.v*f); }

inline sdpQuaternion operator * (const double f, const sdpQuaternion& q)
{ return sdpQuaternion(q.w*f, q.v*f); }

inline sdpQuaternion operator / (const sdpQuaternion& q, const sdpQuaternion& r)
{ sdpQuaternion t = r; return q*t.Inverse(); }

inline sdpQuaternion operator / (const sdpQuaternion& q, const double f)
{ if( f == 0.0 ) return sdpQuaternion(); else return sdpQuaternion(q.w/f, q.v/f); }

inline gint operator == (const sdpQuaternion& q, const sdpQuaternion& r)
{ return ((q.w == r.w) && (q.v == r.v)); }

inline gint operator != (const sdpQuaternion& q, const sdpQuaternion& r)
{ return !(q == r); }


////////////////////////////////////////////////////////////////////////////////
// sdpAngleAxis implementation ...

inline sdpAngleAxis::sdpAngleAxis()
{ angle = 0.0; axis = sdpVector3(0.0, 0.0, 1.0); }

inline sdpAngleAxis::sdpAngleAxis(const gdouble Angle, const sdpVector3& Axis)
{ angle = Angle; axis = Axis; axis.Normalize(); }

inline sdpAngleAxis::sdpAngleAxis(const gdouble Angle, const gdouble X, const gdouble Y, const gdouble Z)
{ angle = Angle; axis[0] = X; axis[1] = Y; axis[2] = Z; axis.Normalize(); }

inline sdpAngleAxis::sdpAngleAxis(const gdouble AngleAxis[4])
{ angle = AngleAxis[0]; axis[0] = AngleAxis[1]; axis[1] = AngleAxis[2]; axis[2] = AngleAxis[3]; axis.Normalize(); }

inline sdpAngleAxis::sdpAngleAxis(const sdpQuaternion& Quaternion)
{
	gdouble halftheta = acos(Quaternion.w);
	gdouble sinhalftheta = sin(halftheta);

	angle = 2.0 * halftheta;
	if(sinhalftheta != 0.0)
		axis = Quaternion.v / sinhalftheta;
	else
		axis = sdpVector3(0.0, 0.0, 1.0);
}

inline sdpAngleAxis::sdpAngleAxis(const sdpAngleAxis& AngleAxis)
{ angle = AngleAxis.angle; axis = AngleAxis.axis; }

inline sdpAngleAxis& sdpAngleAxis::operator = (const sdpAngleAxis& AngleAxis)
{ angle = AngleAxis.angle; axis = AngleAxis.axis; return *this; }

inline sdpAngleAxis& sdpAngleAxis::operator = (const sdpQuaternion& Quaternion)
{
	gdouble halftheta = acos(Quaternion.w);
	gdouble sinhalftheta = sin(halftheta);

	angle = 2.0 * halftheta;
	if(sinhalftheta != 0.0)
		axis = Quaternion.v / sinhalftheta;
	else
		axis = sdpVector3(0.0, 0.0, 1.0);

	return *this;
}

inline sdpAngleAxis& sdpAngleAxis::operator = (const gdouble AngleAxis[4])
{ angle = AngleAxis[0]; axis[0] = AngleAxis[1]; axis[1] = AngleAxis[2]; axis[2] = AngleAxis[3]; return *this; }

inline void sdpAngleAxis::CopyArray(gfloat AngleAxis[4])
{	AngleAxis[0] = gfloat(angle); AngleAxis[1] = gfloat(axis[0]); AngleAxis[2] = gfloat(axis[1]); AngleAxis[3] = gfloat(axis[2]); }

inline void sdpAngleAxis::CopyArray(gdouble AngleAxis[4])
{	AngleAxis[0] = angle; AngleAxis[1] = axis[0]; AngleAxis[2] = axis[1]; AngleAxis[3] = axis[2]; }


////////////////////////////////////////////////////////////////////
// 2D and 3D function implementations

inline sdpMatrix3 sdpIdentity2D()
{
	return sdpMatrix3(sdpVector3(1.0, 0.0, 0.0), sdpVector3(0.0, 1.0, 0.0), sdpVector3(0.0, 0.0, 1.0));
}

inline sdpMatrix3 sdpTranslation2D(const sdpVector2& v)
{
	return sdpMatrix3(sdpVector3(1.0, 0.0, v[VX]), sdpVector3(0.0, 1.0, v[VY]), sdpVector3(0.0, 0.0, 1.0));
}

inline sdpMatrix3 sdpRotation2D(const sdpVector2& Center, const gdouble Angle)
{
	gdouble c = cos(Angle);
	gdouble s = sin(Angle);

	return sdpMatrix3(sdpVector3(c, -s, Center[VX] * (1.0-c) + Center[VY] * s),
										sdpVector3(s, c, Center[VY] * (1.0-c) - Center[VX] * s),
										sdpVector3(0.0, 0.0, 1.0));
}

inline sdpMatrix3 sdpScaling2D(const sdpVector2& scaleVector)
{
	return sdpMatrix3(sdpVector3(scaleVector[VX], 0.0, 0.0), sdpVector3(0.0, scaleVector[VY], 0.0), sdpVector3(0.0, 0.0, 1.0));
}

inline sdpMatrix4 sdpIdentity3D()
{
	return sdpMatrix4(sdpVector4(1.0, 0.0, 0.0, 0.0),
										sdpVector4(0.0, 1.0, 0.0, 0.0),
										sdpVector4(0.0, 0.0, 1.0, 0.0),
										sdpVector4(0.0, 0.0, 0.0, 1.0));
}

inline sdpMatrix4 sdpTranslation3D(const sdpVector3& v)
{
	return sdpMatrix4(sdpVector4(1.0, 0.0, 0.0, v[VX]),
										sdpVector4(0.0, 1.0, 0.0, v[VY]),
										sdpVector4(0.0, 0.0, 1.0, v[VZ]),
										sdpVector4(0.0, 0.0, 0.0, 1.0));
}

inline sdpMatrix4 sdpRotation3D(const gdouble Angle, sdpVector3 Axis)
{
	const gdouble c = cos(Angle), s = sin(Angle), t = 1.0 - c;

	Axis.Normalize();
	return sdpMatrix4(sdpVector4(t * Axis[VX] * Axis[VX] + c, t * Axis[VX] * Axis[VY] - s * Axis[VZ], t * Axis[VX] * Axis[VZ] + s * Axis[VY], 0.0),
										sdpVector4(t * Axis[VX] * Axis[VY] + s * Axis[VZ], t * Axis[VY] * Axis[VY] + c, t * Axis[VY] * Axis[VZ] - s * Axis[VX], 0.0),
										sdpVector4(t * Axis[VX] * Axis[VZ] - s * Axis[VY], t * Axis[VY] * Axis[VZ] + s * Axis[VX], t * Axis[VZ] * Axis[VZ] + c, 0.0),
										sdpVector4(0.0, 0.0, 0.0, 1.0));
}

inline sdpMatrix4 sdpRotation3D(const sdpAngleAxis& AngleAxis)
{
	return sdpRotation3D(AngleAxis.angle, AngleAxis.axis);
}

inline sdpMatrix4 sdpRotation3D(const sdpVector3& EulerAngles)
{
	sdpMatrix4 matrix = sdpIdentity3D();
	matrix = matrix * sdpRotation3D(EulerAngles[1], sdpVector3(0.0, 1.0, 0.0));
	matrix = matrix * sdpRotation3D(EulerAngles[0], sdpVector3(1.0, 0.0, 0.0));
	matrix = matrix * sdpRotation3D(EulerAngles[2], sdpVector3(0.0, 0.0, 1.0));

	return matrix;
}

inline sdpMatrix4 sdpRotation3D(const sdpQuaternion& Quaternion)
{
	return sdpRotation3D(sdpAngleAxis(Quaternion));
}

inline sdpMatrix4 sdpScaling3D(const sdpVector3& scaleVector)
{
	return sdpMatrix4(sdpVector4(scaleVector[VX], 0.0, 0.0, 0.0),
										sdpVector4(0.0, scaleVector[VY], 0.0, 0.0),
										sdpVector4(0.0, 0.0, scaleVector[VZ], 0.0),
										sdpVector4(0.0, 0.0, 0.0, 1.0));
}

inline sdpMatrix4 sdpPerspective3D(const gdouble d)
{
	return sdpMatrix4(sdpVector4(1.0, 0.0, 0.0, 0.0),
										sdpVector4(0.0, 1.0, 0.0, 0.0),
										sdpVector4(0.0, 0.0, 1.0, 0.0),
										sdpVector4(0.0, 0.0, 1.0/d, 0.0));
}

inline sdpMatrix4 sdpRotation3D(const sdpMatrix3& m)
{
//	return(sdpMatrix4(sdpVector4(m[0][0], m[0][1], m[0][2], 0.0),
//									sdpVector4(m[1][0], m[1][1], m[1][2], 0.0),
//									sdpVector4(m[2][0], m[2][1], m[2][2], 0.0),
//									sdpVector4(0.0, 0.0, 0.0, 1.0)));

	// This slightly-more-verbose version eliminates a strange parse error in egcs ...
	sdpVector4 a(m[0][0], m[0][1], m[0][2], 0.0);
	sdpVector4 b(m[1][0], m[1][1], m[1][2], 0.0);
	sdpVector4 c(m[2][0], m[2][1], m[2][2], 0.0);
	sdpVector4 d(0.0, 0.0, 0.0, 1.0);

	return sdpMatrix4(a, b, c, d);
}

//
//	Matrix utilities for affine matrices.
//
//	The input matrix must be affine, but need not be orthogonal.
//	In other words, it may contain scaling, rotation, and translation,
//	but not perspective projection.
//

//	Ref. Foley and Van Dam, 2nd ed, p. 217.
/// Extracts the translation vector from a matrix
inline sdpVector3 sdpExtractTranslation(const sdpMatrix4& m)
{
	return(sdpVector3(m[0][3], m[1][3], m[2][3]));
}

//	Per Valerie Demers, Softimage.
//	Transposed for sdpMatrix4 matrices
/// Extracts the scaling vector from a matrix
inline sdpVector3 sdpExtractScaling(const sdpMatrix4& m)
{
	return(sdpVector3(sqrt(m[0][0] * m[0][0] + m[1][0] * m[1][0] + m[2][0] * m[2][0]),
									sqrt(m[0][1] * m[0][1] + m[1][1] * m[1][1] + m[2][1] * m[2][1]),
									sqrt(m[0][2] * m[0][2] + m[1][2] * m[1][2] + m[2][2] * m[2][2])));
}

/// Extracts a rotation matrix from a matrix
inline sdpMatrix3 sdpExtractRotation(const sdpMatrix4& m)
{
	// Get scaling ...
	sdpVector3 scale(sdpExtractScaling(m));
	// Compute the inverse of scaling ...
	sdpVector3 invscale(1.0 / scale[0], 1.0 / scale[1], 1.0 / scale[2]);
	// Apply Inverse of scaling as a transformation, to get unit scale ...
	sdpMatrix4 unscaled(m * sdpScaling3D(invscale));

	// Piece-together the rotation matrix ...
	sdpVector3 a(unscaled[0], VW);
	sdpVector3 b(unscaled[1], VW);
	sdpVector3 c(unscaled[2], VW);

	return sdpMatrix3(a, b, c);
}

//
//	Unit quaternion utilities.
//
//	Those special functions deal with rotations, thus require unit quaternions.
//

//	Euler/Matrix/Quaternion conversion functions are based on Ken Shoemake code,
//	from Graphic Gems IV

/// Returns the spherical linear interpolation between two quaternions
inline sdpQuaternion sdpSlerp(const sdpQuaternion& q1, const sdpQuaternion& q2, double t)
{
	// Calculate the angle between the two quaternions ...
	double cosangle = q1.w * q2.w + q1.v * q2.v;

	if(cosangle < -1.0)
		cosangle = -1.0;
	if(cosangle > 1.0)
		cosangle = 1.0;

	// If the angle is reasonably large, do the spherical interpolation
	if(1.0 - cosangle > 16 * FLT_EPSILON)
		{
			double angle = acos(cosangle);
			return (sin((1.0 - t)*angle)*q1 + sin(t*angle)*q2) / sin(angle);
		}

	// The angle is too close to zero - do a linear interpolation
	return sdpLerp(q1, q2, t);
}


///////////////////////////////////////////////////////////////////////////////////////////////////////////
// sdpEulerAngles implementation

inline sdpEulerAngles::sdpEulerAngles()
{
	n[0] = n[1] = n[2] = 0.0;
	order = YXZstatic;
}

inline sdpEulerAngles::sdpEulerAngles(sdpVector3 Angles, AngleOrder Order)
{
	n[0] = Angles[0];
	n[1] = Angles[1];
	n[2] = Angles[2];
	order = Order;
}

inline sdpEulerAngles::sdpEulerAngles(gdouble X, gdouble Y, gdouble Z, AngleOrder Order)
{
	n[0] = X;
	n[1] = Y;
	n[2] = Z;
	order = Order;
}

inline sdpEulerAngles::sdpEulerAngles(sdpQuaternion& Quaternion, AngleOrder Order)
{
	double NQuaternion = Quaternion.Magnitude();
	double s = (NQuaternion > 0.0) ? (2.0 / NQuaternion) : 0.0;

	double xs = Quaternion.v[0]*s, ys = Quaternion.v[1]*s, zs = Quaternion.v[2]*s;
	double wx = Quaternion.w*xs, wy = Quaternion.w*ys, wz = Quaternion.w*zs;
	double xx = Quaternion.v[0]*xs, xy = Quaternion.v[0]*ys, xz = Quaternion.v[0]*zs;
	double yy = Quaternion.v[1]*ys, yz = Quaternion.v[1]*zs, zz = Quaternion.v[2]*zs;

	sdpMatrix3 matrix;
	matrix[0][0] = 1.0 - (yy + zz);
	matrix[0][1] = xy - wz;
	matrix[0][2] = xz + wy;
	matrix[1][0] = xy + wz;
	matrix[1][1] = 1.0 - (xx + zz);
	matrix[1][2] = yz - wx;
	matrix[2][0] = xz - wy;
	matrix[2][1] = yz + wx;
	matrix[2][2] = 1.0 - (xx + yy);

	sdpMatrix4 rotmatrix = sdpRotation3D(matrix);
	*this = sdpEulerAngles(rotmatrix, Order);
}

inline sdpEulerAngles::sdpEulerAngles(sdpMatrix4& Matrix, AngleOrder Order)
{
	OrderRepetition repetition = Repetition(Order);
	OrderParity parity = Parity(Order);
	OrderFrame frame = Frame(Order);
	gint i, j, k;
	Axes(Order, i, j, k);
	

	if(repetition == sdpEulerAngles::Repeats )
		{
			double sy = sqrt(Matrix[i][j]*Matrix[i][j] + Matrix[i][k]*Matrix[i][k]);
			if( sy > 16*FLT_EPSILON )
				{
					n[0] = atan2(Matrix[i][j], Matrix[i][k]);
					n[1] = atan2(sy, Matrix[i][i]);
					n[2] = atan2(Matrix[j][i], -Matrix[k][i]);
				}
			else
				{
					n[0] = atan2(-Matrix[j][k], Matrix[j][j]);
					n[1] = atan2(sy, Matrix[i][i]);
					n[2] = 0;
				}
		}
	else
		{
			double cy = sqrt(Matrix[i][i]*Matrix[i][i] + Matrix[j][i]*Matrix[j][i]);
			if( cy > 16*FLT_EPSILON )
				{
					n[0] = atan2(Matrix[k][j], Matrix[k][k]);
					n[1] = atan2(-Matrix[k][i], cy);
					n[2] = atan2(Matrix[j][i], Matrix[i][i]);
				}
			else
				{
					n[0] = atan2(-Matrix[j][k], Matrix[j][j]);
					n[1] = atan2(-Matrix[k][i], cy);
					n[2] = 0;
				}
		}

	if(parity == sdpEulerAngles::OddParity )
		{
			n[0] = -n[0];
			n[1] = -n[1];
			n[2] = -n[2];
		}

	if(frame == sdpEulerAngles::RotatingFrame )
		std::swap(n[0], n[2]);

	order = Order;
}

inline void sdpEulerAngles::Axes(AngleOrder& Order, gint& i, gint& j, gint& k)
{
	static const gint safe[4] = {0, 1, 2, 0};
	static const gint next[4] = {1, 2, 0, 1};

	i = safe[(Order >> 3) & 3];
	j = next[i + Parity(Order)];
	k = next[i + 1 - Parity(Order)];
}

inline gdouble& sdpEulerAngles::operator [] ( gint i)
{
	g_assert(! (i < 0 || i > 2));
	return n[i];
}

inline gdouble sdpEulerAngles::operator [] ( gint i) const
{
	g_assert(! (i < 0 || i > 2));
	return n[i];
}

////////////////////////////////////////////////////////////////////////////////////////
// sdpPlane implementation

inline sdpPlane::sdpPlane(const sdpVector3& Normal, const gdouble Distance)
{
	m_Normal = Normal;
	m_Normal.Normalize();
	
	m_Distance = Distance;
}

inline sdpPlane::sdpPlane(const sdpVector3& Normal, const sdpVector3& ContainedPoint)
{
	m_Normal = Normal;
	m_Normal.Normalize();
	
	m_Distance = -ContainedPoint * m_Normal;
}

inline sdpPlane::sdpPlane(const sdpVector3& ContainedPointA, const sdpVector3& ContainedPointB, const sdpVector3& ContainedPointC)
{
	m_Normal = ((ContainedPointA - ContainedPointB) ^ (ContainedPointC - ContainedPointB)).Normalize();
	m_Distance = -ContainedPointA * m_Normal;
}

inline sdpVector3 sdpPlane::Normal() const
{
	return m_Normal;
}

inline gdouble sdpPlane::Distance() const
{
	return m_Distance;
}

inline void sdpPlane::SetNormal(const sdpVector3& Normal)
{
	m_Normal = Normal;
}

inline void sdpPlane::SetDistance(const gdouble Distance)
{
	m_Distance = Distance;
}

/// Enumerates ways in which two line segments might (or might not) intersect
typedef enum
{
	SDP_NOINTERSECTION,
	SDP_EXTERIORINTERSECTION,
	SDP_INTERIORINTERSECTION,

	SDP_EXTERIORCOLINEARPOINT,
	SDP_INTERIORCOLINEARPOINT,

	SDP_INTERIORCOLINEARPOINTS,
	SDP_INTERIOREXTERIORCOLINEARPOINTS,
	SDP_EXTERIORCOLINEARPOINTSNOINTERSECTION,
	SDP_EXTERIORCOLINEARPOINTS,

	SDP_ERROR
} sdpLineSegmentIntersectionType;

/// Determines whether two line segments intersect, and if so, how
sdpLineSegmentIntersectionType sdpLineSegmentIntersection(const sdpVector2& A1, const sdpVector2& A2, const sdpVector2& B1, const sdpVector2& B2);
/// Calculates the intersection of a plane with a line
bool sdpPlaneLineIntersection(const sdpPlane& Plane, const sdpVector3 LineOrigin, const sdpVector3 LineDirection, sdpVector3& Result);
/// Calculates the intersection of two lines, specified by two contained points each
bool sdpLineIntersection(const sdpVector2& A1, const sdpVector2& A2, const sdpVector2& B1, const sdpVector2& B2, sdpVector2& Result);

/// Calculates the area of a triangle
gdouble sdpArea(const sdpVector2& Point1, const sdpVector2& Point2, const sdpVector2& Point3);


#endif SDPMATH_H
