// SDPTypes Library
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
// 
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements an extended string class, sdpString, derived from std::string
		\author Tim Shead (tshead@k-3d.com)
*/

#include <sdpstring.h>
#include <sdpmath.h>

#include <stdio.h>
#include <stdarg.h>
#include <malloc.h>

#include <iomanip>
#include <strstream>
#include <algorithm>
#include <functional>

#if defined SDPWIN32
#define vsnprintf _vsnprintf
#endif	// SDPWIN32


// Value constructors ...
sdpString::sdpString(const bool& Value) { *this = Value ? "true" : "false"; }
sdpString::sdpString(const gshort& Value) { format("%d", Value); }
sdpString::sdpString(const gushort& Value) { format("%u", Value); }
sdpString::sdpString(const gint& Value) { format("%d", Value); }
sdpString::sdpString(const guint& Value) { format("%u", Value); }
sdpString::sdpString(const glong& Value) { format("%d", Value); }
sdpString::sdpString(const gulong& Value) { format("%u", Value); }
sdpString::sdpString(const gfloat& Value) { format("%.12g", Value); }
sdpString::sdpString(const gdouble& Value) { format("%.12g", Value); }
sdpString::sdpString(const sdpVector2& Value) { format("%.12g %.12g", Value[0], Value[1]); }
sdpString::sdpString(const sdpVector3& Value) { format("%.12g %.12g %.12g", Value[0], Value[1], Value[2]); }
sdpString::sdpString(const sdpVector4& Value) { format("%.12g %.12g %.12g %.12g", Value[0], Value[1], Value[2], Value[3]); }
sdpString::sdpString(const sdpAngleAxis& Value) { format("%.12g %.12g %.12g %.12g", Value.angle, Value.axis[0], Value.axis[1], Value.axis[2]); }
sdpString::sdpString(const sdpQuaternion& Value) { format("%.12g %.12g %.12g %.12g", Value.w, Value.v[0], Value.v[1], Value.v[2]); }
sdpString::sdpString(const sdpEulerAngles& Value) { format("%.12g %.12g %.12g %d", Value[0], Value[1], Value[2], Value.order); }
sdpString::sdpString(const sdpGUID& Value) { format("{%.8x %.8x %.8x %.8x}", Value.Data1(), Value.Data2(), Value.Data3(), Value.Data4()); }
sdpString::sdpString(const void*& Value) { format("%.8x", Value); }

SDPSTRING sdpString::buffer(size_type BufferSize)
{
	resize(BufferSize);
	return const_cast<SDPSTRING>(c_str());
}

void sdpString::format(SDPCSTRING FormatString, ...)
{
	va_list arglist;
	va_start(arglist, FormatString);

	formatv(FormatString, arglist);

	va_end(arglist);
}

void sdpString::formatv(SDPCSTRING FormatString, va_list ArgList)
{
	// Initial size of the results buffer
	int buffersize = 128;

	// Keep looping until we succeed ...
	while(1)
		{
			// Allocate a buffer on the stack to hold the formatted characters and zero it out ...
			gchar* buffer = new gchar[buffersize];
			memset(buffer, 0, buffersize);

			// Print into the buffer ...
			int required = vsnprintf(buffer, buffersize, FormatString, ArgList);
			
			// If it succeeded, copy the buffer into ourselves and quit ...
			if(required < buffersize)
				{
					*this = buffer;
					delete[] buffer;
					break;
				}
				
			// Otherwise, set ourselves up with a larger buffer ...
			delete[] buffer;
			buffersize = required+1;
		}
}

void sdpString::formattedreplace(const char Delimiter, SDPCSTRING Specifier, SDPCSTRING Replacement)
{
	// Sanity checks ...
	g_assert(Delimiter);
	g_assert_string(Specifier);

	sdpString specifier = Delimiter + sdpString(Specifier);
	sdpString replacement(Replacement);
	
	// For each copy of the specifier we find ...
	size_type index = find(specifier);
	while(index != npos)
		{
			// Make sure the user didn't want a literal ...
			if(index)
				if(at(index-1) == Delimiter)
					{
						index = find(specifier, index + 1);
						continue;
					}
			
			// Make the replacement!
			replace(index, specifier.size(), replacement);
			index = find(specifier, index + replacement.size() + 1);
		}
}

sdpString& sdpString::trim()
{
	trimleft();
	trimright();
	return *this;
}

sdpString& sdpString::trimleft()
{
	// Easiest case ...
	if(!size())
		return *this;

	size_type i;
	for(i = 0; i < size(); i++)
		if(!isspace(operator[](i)))
			break;

	// No whitespace, so we're done ...
	if(i == 0)
		return *this;

	// All whitespace, so clear the string ...
	if(i == size())
		{
			clear();
			return *this;
		}

	sdpString temp = right(size()-i);
	assign(temp);

	return *this;
}

sdpString& sdpString::trimright()
{
	// Easiest case ...
	if(!size())
		return *this;

	size_type i;
	for(i = size()-1; i >= 0; i--)
		if(!isspace(operator[](i)))
			break;

	// No whitespace, so we're done ...
	if(i == (size()-1))
		return *this;

	// All whitespace, so clear the string ...
	if(i == size_type(-1))
		{
			clear();
			return *this;
		}

	sdpString temp = left(i+1);
	assign(temp);

	return *this;
}

sdpString& sdpString::makeupper()
{
	std::transform(begin(), end(), begin(), toupper);	// portable, slow way of doing it
//	strupr(const_cast<SDPSTRING>(data()));									// unportable, fast way of doing it
	return *this;
}

sdpString& sdpString::makelower()
{
	std::transform(begin(), end(), begin(), tolower);	// portable, slow way of doing it
//	strlwr(const_cast<SDPSTRING>(data()));									// unportable, fast way of doing it
	return *this;
}

// Generate an sdpString with printf() style formatting in one step ...
sdpString sdpFormatString(SDPCSTRING Format, ...)
{
	// Get ahold of our argument list ...
	va_list arglist;
	va_start(arglist, Format);

	// Generate the formatted string ...
	sdpString result = sdpFormatStringV(Format, arglist);

	// Cleanup the argument list ...
	va_end(arglist);

	return result;
}

// Generate an sdpString with printf() style formatting in one step ...
sdpString sdpFormatStringV(SDPCSTRING Format, va_list ArgList)
{
	// Generate the formatted string ...
	sdpString result;
	result.formatv(Format, ArgList);

	return result;
}

sdpString& sdpString::clear()
{
	erase();
	return *this;
}

sdpString sdpString::left(size_type Count) const
{
	g_assert(Count <= size());
	return sdpString(begin(), begin() + Count);
}

sdpString sdpString::mid(size_type Start, size_type Count) const
{
	g_assert(Start + Count <= size());
	return sdpString(begin() + Start, begin() + Start + Count);
}

sdpString sdpString::right(size_type Count) const
{
	g_assert(Count <= size());
	return sdpString(end() - Count, end());
}

const bool sdpString::isalphanumeric() const
{
	for(const_iterator i = begin(); i != end(); i++)
		{
			if(!isalnum((gint)*i)) return false;
		}

	return true;
}

const bool sdpString::isalphabetic() const
{
	for(const_iterator i = begin(); i != end(); i++)
		if(!isalpha((gint)*i)) return false;

	return true;
}

const bool sdpString::isnumeric() const
{
	for(const_iterator i = begin(); i != end(); i++)
		if(!isdigit((gint)*i)) return false;

	return true;
}

const bool sdpString::islowercase() const
{
	for(const_iterator i = begin(); i != end(); i++)
		if(!islower((gint)*i)) return false;

	return true;
}

const bool sdpString::isprintable() const
{
	for(const_iterator i = begin(); i != end(); i++)
		if(!isprint((gint)*i)) return false;

	return true;
}

const bool sdpString::ispunctuation() const
{
	for(const_iterator i = begin(); i != end(); i++)
		if(!ispunct((gint)*i)) return false;

	return true;
}

const bool sdpString::iswhitespace() const
{
	for(const_iterator i = begin(); i != end(); i++)
		if(!isspace((gint)*i)) return false;

	return true;
}

const bool sdpString::isuppercase() const
{
	for(const_iterator i = begin(); i != end(); i++)
		if(!isupper((gint)*i)) return false;

	return true;
}

const bool sdpString::getvalue(bool& Result) const
{
	sdpString temp = *this;
	temp.makelower();
	temp.trim();

	if(!temp.size())
		return false;

	if(temp[0] == 't' || temp[0] == 'y')
		{
			Result = true;
			return true;
		}
	else if(temp[0] == 'f' || temp[0] == 'n')
		{
			Result = false;
			return true;
		}

	g_return_val_if_fail(0, false);
}

const bool sdpString::getvalue(gshort& Result) const
{
	Result = atol(c_str());
	return true;
}

const bool sdpString::getvalue(gushort& Result) const
{
	Result = atol(c_str());
	return true;
}

const bool sdpString::getvalue(gint& Result) const
{
	Result = atol(c_str());
	return true;
}

const bool sdpString::getvalue(guint& Result) const
{
	Result = atol(c_str());
	return true;
}

const bool sdpString::getvalue(glong& Result) const
{
	Result = atol(c_str());
	return true;
}

const bool sdpString::getvalue(gulong& Result) const
{
	Result = atol(c_str());
	return true;
}

const bool sdpString::getvalue(gfloat& Result) const
{
	Result = gfloat(atof(c_str()));
	return true;
}

const bool sdpString::getvalue(gdouble& Result) const
{
	Result = atof(c_str());
	return true;
}

const bool sdpString::getvalue(sdpVector2& Result) const
{
	sdpString temp = vectorcleanup();
	std::istrstream stream(temp);
	stream >> Result[0];
	Result[1] = Result[0];
	stream >> Result[1];

	return true;
}

const bool sdpString::getvalue(sdpVector3& Result) const
{
	sdpString temp = vectorcleanup();
	std::istrstream stream(temp);
	stream >> Result[0];
	Result[1] = Result[2] = Result[0];
	stream >> Result[1] >> Result[2];

	return true;
}

const bool sdpString::getvalue(sdpVector4& Result) const
{
	sdpString temp = vectorcleanup();
	std::istrstream stream(temp);
	stream >> Result[0];
	Result[1] = Result[2] = Result[3] = Result[0];
	stream >> Result[1] >> Result[2] >> Result[3];

	return true;
}

const bool sdpString::getvalue(sdpAngleAxis& Result) const
{
	sdpString temp = vectorcleanup();
	std::istrstream stream(temp);
	stream >> Result.angle >> Result.axis[0] >> Result.axis[1] >> Result.axis[2];

	return true;
}

const bool sdpString::getvalue(sdpQuaternion& Result) const
{
	sdpString temp = vectorcleanup();
	std::istrstream stream(temp);
	stream >> Result.w >> Result.v[0] >> Result.v[1] >> Result.v[2];

	return true;
}

const bool sdpString::getvalue(sdpEulerAngles& Result) const
{
	sdpString temp = vectorcleanup();
	std::istrstream stream(temp);
	gint order;
	stream >> Result[0] >> Result[1] >> Result[2] >> order;

	Result.order = sdpEulerAngles::AngleOrder(order);
	
	return true;
}

const bool sdpString::getvalue(sdpGUID& Result) const
{
	std::istrstream stream(c_str());
	stream >> std::hex;

	gchar c;
	stream >> c;
	g_return_val_if_fail(c == '{', false);

	gulong temp;
	stream >> temp; Result.SetData1(temp);
	stream >> temp; Result.SetData2(temp);
	stream >> temp; Result.SetData3(temp);
	stream >> temp; Result.SetData4(temp);

	stream >> c;
	g_return_val_if_fail(c == '}', false);

	return true;
}

const bool sdpString::getvalue(void*& Result) const
{
	std::istrstream stream(c_str());
	gulong result;
	stream >> std::hex >> result;

	Result = (void*)result;

	return true;
}

const bool sdpString::getvalue(sdpString& Result) const
{
	Result = *this;
	return true;
}

void sdpString::replaceall(SDPCSTRING Target, SDPCSTRING Replacement)
{
	// Sanity checks ...
	g_assert_string(Target);
//	g_assert(Replacement);	// Notice - the replacement can be empty, but not NULL ...

	size_type length = strlen(Target);

	for(size_type index = find(Target); index != npos; index = find(Target, index+1))
		replace(index, length, Replacement);
}

void sdpString::replaceall(const char Target, const char Replacement)
{
	for(size_type index = find(Target); index != npos; index = find(Target))
		(*this)[index] = Replacement;
}

sdpString sdpString::vectorcleanup() const
{
	sdpString result = *this;
	
	// Get rid of commas ...
	result.replaceall(',', ' ');
				
	// Get rid of parentheses ...
	result.replaceall('(', ' ');
	result.replaceall(')', ' ');
	
	result.trim();
	
	return result;
}

sdpString& sdpString::makereverse()
{
	sdpString temp = *this;

	for(size_type i = 0; i < size(); i++)
		(*this)[i] = temp[gint(size()-i-1)];

	return *this;
}

sdpString sdpString::spanincluding(SDPCSTRING Set)
{
	size_type index = find_first_not_of(Set);
	if(npos == index) index = 0;

	return left(index);
}

sdpString sdpString::spanexcluding(SDPCSTRING Set)
{
	size_type index = find_first_of(Set);
	if(npos == index) return *this;
	if(0 == index) return sdpString();

	return left(index);
}


