// SDPXML Library
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements utility functions for use with XML documents
		\author Tim Shead (tshead@k-3d.com)
*/

#include <sdpxmlutility.h>
#include <sdpregex.h>

#include <algorithm>
#include <iostream>

////////////////////////////////////////////////////////////////////////////////
// Handy-dandy helper functions ...

// Private helper for creating elements ...
static void xmlinitializev(sdpXmlIElement* Element, SDPCSTRING ElementText, va_list Arguments)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert(Arguments);
	
	// Set the text ...
	Element->SetText(ElementText);

	// Loop through our varargs ...
	while(1)
		{
			char* attributename = va_arg(Arguments, char*);
			if(0 == attributename)
				break;
				
			char* attributevalue = va_arg(Arguments, char*);
			g_return_if_fail(attributevalue);
			
			Element->CreateAttribute(attributename, attributevalue);
		}
}

// Private helpers for XPath selection ...
static sdpXmlElementArray xpathpredicatelastelement(sdpXmlIElement* Anchor, const sdpString& PathElement)
{
	// Our empty result set ...
	sdpXmlElementArray results;
	
	// Sanity checks ...
	g_return_val_if_fail(Anchor, results);
	g_return_val_if_fail(PathElement.size(), results);
	
	// For each element ...
	sdpXmlIElement* element = 0;
	for(SDPXMLELEMENTHANDLE e = Anchor->GetFirstElement(); e; e = Anchor->GetNextElement(e))
		{
			// Skip elements that don't match ...
			if(PathElement != Anchor->GetElement(e)->Name())
				continue;
				
			element = Anchor->GetElement(e);
		}
			
	if(element)
		results.push_back(element);
	
	return results;
}

static sdpXmlElementArray xpathpredicatenthelement(sdpXmlIElement* Anchor, const sdpString& PathElement, const gulong N)
{
	// Our empty result set ...
	sdpXmlElementArray results;
	
	// Sanity checks ...
	g_return_val_if_fail(Anchor, results);
	g_return_val_if_fail(PathElement.size(), results);
	g_return_val_if_fail(N, results);
	
	gulong count = 1;
	for(SDPXMLELEMENTHANDLE e = Anchor->GetFirstElement(); e; e = Anchor->GetNextElement(e))
		{
			// Skip elements that don't match ...
			sdpXmlIElement* element = Anchor->GetElement(e);
			if(PathElement != element->Name())
				continue;
				
			if(count++ != N)
				continue;
			
			results.push_back(element);
		}
		
	return results;
}

static sdpXmlElementArray xpathpredicateattribute(sdpXmlIElement* Anchor, const sdpString& PathElement, const sdpString& AttributeName)
{
	// Our empty result set ...
	sdpXmlElementArray results;
	
	// Sanity checks ...
	g_return_val_if_fail(Anchor, results);
	g_return_val_if_fail(PathElement.size(), results);
	g_return_val_if_fail(AttributeName.size(), results);

	// For each element ...						
	for(SDPXMLELEMENTHANDLE e = Anchor->GetFirstElement(); e; e = Anchor->GetNextElement(e))
		{
			// Skip elements that don't match ...
			sdpXmlIElement* element = Anchor->GetElement(e);
			if(PathElement != element->Name())
				continue;
				
			// For each attribute ...
			for(SDPXMLATTRIBUTEHANDLE a = element->GetFirstAttribute(); a; a = element->GetNextAttribute(a))
				{
					// Skip attributes that don't match ...
					sdpXmlIAttribute* attribute = element->GetAttribute(a);
					if(AttributeName != attribute->Name())
						continue;
						
					results.push_back(element);
				}
		}
	
	return results;
}

static sdpXmlElementArray xpathpredicateattributevalue(sdpXmlIElement* Anchor, const sdpString& PathElement, const sdpString& AttributeName, const sdpString& AttributeValue)
{
	// Our empty result set ...
	sdpXmlElementArray results;
	
	// Sanity checks ...
	g_return_val_if_fail(Anchor, results);
	g_return_val_if_fail(PathElement.size(), results);
	g_return_val_if_fail(AttributeName.size(), results);
	g_return_val_if_fail(AttributeValue.size(), results);
	
	// For each child element ...
	for(SDPXMLELEMENTHANDLE e = Anchor->GetFirstElement(); e; e = Anchor->GetNextElement(e))
		{
			// Skip elements that don't match ...
			sdpXmlIElement* element = Anchor->GetElement(e);
			if(PathElement != element->Name())
				continue;

			// For each attribute ...										
			for(SDPXMLATTRIBUTEHANDLE a = element->GetFirstAttribute(); a; a = element->GetNextAttribute(a))
				{
					// Skip attributes that don't match ...
					sdpXmlIAttribute* attribute = element->GetAttribute(a);
					if(AttributeName != attribute->Name())
						continue;
					
					if(AttributeValue != attribute->Value())
						continue;
						
					results.push_back(element);
				}
		}
	
	return results;
}

static sdpXmlElementArray xpathpredicatelogic(sdpXmlIElement* Anchor, SDPCSTRING PathElement, SDPCSTRING Predicate)
{
	// Initialize our regular expression search engines ...
	const static sdpRegEx orpredicate("^(or)([[:space:]]+)");
	const static sdpRegEx andpredicate("^(and)([[:space:]]+)");
	const static sdpRegEx lastpredicate("^(last\\(\\))([[:space:]]*)");
	const static sdpRegEx numericpredicate("^([[:digit:]]+)([[:space:]]*)");
	const static sdpRegEx attributepredicate("^@([[:alpha:]]*)([[:space:]]+|$)");
	const static sdpRegEx attributevaluepredicate("^@([[:alpha:]][[:alnum:]]*)=\"([^\"]*)\\\"([[:space:]]*)");

	// Our empty result set ...
	sdpXmlElementArray results;
	
	// Results from each predicate clause ...
	sdpXmlElementArray predicateresults;
	
	// Sanity checks ...
	g_return_val_if_fail(Anchor, results);
	g_return_val_if_string_fail(PathElement, results);
	g_return_val_if_string_fail(Predicate, results);
	
	// Get rid of unwanted whitespace ...
	sdpString pathelement(PathElement);
	pathelement.trim();
	
	sdpString predicate(Predicate);
	predicate.trim();

	sdpString boolean;
	sdpRegExMatchArray matches;
	
	while(predicate.size())
		{
			// Look for the last element whose name matches
			if(0 == lastpredicate.Search(predicate, matches))
				{
					predicate = predicate.right(predicate.size() - matches[0].second);

//std::cout << "last predicate: " << predicate << std::endl;		
					predicateresults = xpathpredicatelastelement(Anchor, pathelement);
				}
			// Look for the Nth element whose name matches
			else if(0 == numericpredicate.Search(predicate, matches))
				{
					gulong number = atol(predicate.left(matches[1].second));
					predicate = predicate.right(predicate.size() - matches[0].second);

//std::cout << "numeric predicate: " << predicate << " - " << number << std::endl;		
					predicateresults = xpathpredicatenthelement(Anchor, pathelement, number);
				}
			// Look for elements that have the given attribute
			else if(0 == attributepredicate.Search(predicate, matches))
				{
					sdpString attributename = predicate.mid(matches[1].first, matches[1].second - matches[1].first);
					predicate = predicate.right(predicate.size() - matches[0].second);

//std::cout << "attribute predicate: " << predicate << " - " << attributename << std::endl;
					predicateresults = xpathpredicateattribute(Anchor, pathelement, attributename);
				}
			// Look for elements that have the given attribute value
			else if(0 == attributevaluepredicate.Search(predicate, matches))
				{
					sdpString attributename = predicate.mid(matches[1].first, matches[1].second - matches[1].first);
					sdpString attributevalue = predicate.mid(matches[2].first, matches[2].second - matches[2].first);
					predicate = predicate.right(predicate.size() - matches[0].second);

//std::cout << "attribute value predicate: " << predicate << " - " << attributename << " = " << attributevalue << std::endl;
					predicateresults = xpathpredicateattributevalue(Anchor, PathElement, attributename, attributevalue);
				}
			// Look for an "or" clause ...
			else if(0 == orpredicate.Search(predicate, matches))
				{
					boolean = "or";
					predicate = predicate.right(predicate.size() - matches[0].second);

//std::cout << "or predicate: " << predicate << std::endl;
				}
			// Look for an "and" clause ...
			else if(0 == andpredicate.Search(predicate, matches))
				{
					boolean = "and";
					predicate = predicate.right(predicate.size() - matches[0].second);

//std::cout << "and predicate: " << predicate << std::endl;
				}
			else
				{
					std::cerr << "xpredicatelogic(): Unknown predicate " << Predicate << std::endl;
					break;
				}
				
			if(boolean == "and")
				{
					std::sort(results.begin(), results.end());
					std::sort(predicateresults.begin(), predicateresults.end());
					
					sdpXmlElementArray temp;
					std::set_intersection(results.begin(), results.end(), predicateresults.begin(), predicateresults.end(), std::back_inserter(temp));
					
					results = temp;
				}
			else if(boolean == "or")
				{
					std::sort(results.begin(), results.end());
					std::sort(predicateresults.begin(), predicateresults.end());
					
					sdpXmlElementArray temp;
					std::set_union(results.begin(), results.end(), predicateresults.begin(), predicateresults.end(), std::back_inserter(temp));
					
					results = temp;
				}
			else
				{
					results = predicateresults;
				}
		}
		
	return results;
}

static sdpXmlElementArray xpathelementlogic(sdpXmlIElement* Anchor, SDPCSTRING PathElement)
{
	// Our empty result set ...
	sdpXmlElementArray results;

	// Sanity checks ...
	g_return_val_if_string_fail(PathElement, results);

	// Get rid of unwanted whitespace ...
	sdpString pathelement(PathElement);
	pathelement.trim();

	// Match the current element ...
	if(pathelement == ".")
		{
			results.push_back(Anchor);
			return results;
		}
		
	// Match the parent element ...
	if(pathelement == "..")
		{
			if(Anchor->ParentElement())
				results.push_back(Anchor->ParentElement());
			return results;
		}

	// Match every child element ...
	if(pathelement == "*")
		{
			for(SDPXMLELEMENTHANDLE e = Anchor->GetFirstElement(); e; e = Anchor->GetNextElement(e))
				results.push_back(Anchor->GetElement(e));
			return results;
		}

	// Look for elements just by name (no predicate) ...
	sdpString::size_type index = pathelement.find("[");
	if(sdpString::npos == index)
		{
			// For each element ...
			for(SDPXMLELEMENTHANDLE e = Anchor->GetFirstElement(); e; e = Anchor->GetNextElement(e))
				{
					// Skip elements that don't match ...
					sdpXmlIElement* element = Anchor->GetElement(e);
					if(pathelement != element->Name())
						continue;
						
					results.push_back(element);
				}
				
			return results;
		}
	
	// We better have a closing character for the predicate!
	g_return_val_if_fail(pathelement[pathelement.size()-1] == ']', results);

	// Parse out the predicate ...
	sdpString predicate = pathelement.mid(index + 1, pathelement.size() - index - 2);
	pathelement = pathelement.left(index);

	// Add the predicate's matches to our results ...
	sdpXmlElementArray predicatematches = xpathpredicatelogic(Anchor, pathelement, predicate);
	results.insert(results.end(), predicatematches.begin(), predicatematches.end());
		
	return results;
}

static void xpathselect(sdpXmlIElement* Element, SDPCSTRING Path, sdpXmlElementArray& Results)
{
	// Sanity checks ...
	g_assert(Element);
	
	// Get rid of unwanted whitespace in the selection ...
	sdpString path(Path);
	path.trim();
	
	// First, figure out which node to begin with ... start with the given element ...
	sdpXmlIElement* anchor = Element;

	// Check for an absolute path and adjust the anchor if necessary ...
	if(path[0] == '/')
		{
			path = path.right(path.size() - 1);
			anchor = Element->Document()->GetRootElement();
		}

	// Get the next element in the path ...
	sdpString::size_type index = path.find("/");
	if(index != sdpString::npos)
		{
			sdpString subpath = path.left(index);
			path = path.right(path.size() - (index + 1));
			
			// Find every element that matches this criteria ...
			sdpXmlElementArray array = xpathelementlogic(anchor, subpath);
			
			// Call ourselves recursively ...
			for(sdpXmlElementArray::size_type i = 0; i < array.size(); i++)
				xpathselect(array[i], path, Results);
				
			return;
		}

	// We've reached the last element in the path ... find every element that
	// matches this final criteria and add it to the results
	sdpXmlElementArray found = xpathelementlogic(anchor, path);
	Results.insert(Results.end(), found.begin(), found.end());
}

sdpXmlElementArray sdpXmlXPathSelect(sdpXmlIDocument* Document, SDPCSTRING Path)
{
	// Sanity checks ...
	g_assert(Document);
	g_assert_string(Path);
	
	// See if the document contains anything ...
	if(Document->GetRootElement())
		return sdpXmlXPathSelect(Document->GetRootElement(), Path);
			
	// If not, return an empty set ...
	return sdpXmlElementArray();
}

sdpXmlElementArray sdpXmlXPathSelect(sdpXmlIElement* Element, SDPCSTRING Path)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert_string(Path);

	sdpXmlElementArray results;
	xpathselect(Element, Path, results);
	
	return results;	
}

sdpXmlIElement* sdpXmlGetXPathElement(sdpXmlIElement* Parent, SDPCSTRING Path)
{
	// Sanity checks ...
	g_assert(Parent);
	g_assert_string(Path);

	// Look for the requested element ...
	sdpXmlElementArray results = sdpXmlXPathSelect(Parent, Path);
	if(results.size())
		{
			if(1 != results.size())
				std::cout << "sdpXmlGetXPathElement() - <" << Parent->Name() << "> contains multiple <" << Path << "> elements.  Extras will be ignored." << std::endl;
				
			return results.front();
		}
		
	return 0;
}

sdpXmlIElement* sdpXmlGetSafeXPathElement(sdpXmlIElement* Parent, SDPCSTRING Path, SDPCSTRING ElementName, SDPCSTRING ElementText, ...)
{
	// Sanity checks ...
	g_assert(Parent);
	g_assert_string(Path);
	g_assert_string(ElementName);

	// Try to get an element that matches the request ...
	sdpXmlIElement* result = sdpXmlGetXPathElement(Parent, Path);
	if(result)
		return result;
		
	// Couldn't find anything, so create it ...
	result = Parent->CreateElement(ElementName);
	g_assert(result);
	
	va_list arguments;
	va_start(arguments, ElementText);
	xmlinitializev(result, ElementText, arguments);
	va_end(arguments);
	
	return result;
}

sdpXmlIElement* sdpXmlCreateElementV(sdpXmlIDocument* Document, SDPCSTRING ElementName, SDPCSTRING ElementText, ...)
{
	// Sanity checks ...
	g_assert(Document);
	g_assert_string(ElementName);
	
	// Create the element ...
	sdpXmlIElement* result = Document->CreateRootElement(ElementName);
	g_assert(result);

	// Initialize it using our varargs ...
	va_list arguments;
	va_start(arguments, ElementText);
	xmlinitializev(result, ElementText, arguments);
	va_end(arguments);
	
	return result;
}

sdpXmlIElement* sdpXmlCreateElementV(sdpXmlIElement* Parent, SDPCSTRING ElementName, SDPCSTRING ElementText, ...)
{
	// Sanity checks ...
	g_assert(Parent);
	g_assert_string(ElementName);
	
	// Create the element ...
	sdpXmlIElement* result = Parent->CreateElement(ElementName);
	g_assert(result);

	// Initialize it using our varargs ...
	va_list arguments;
	va_start(arguments, ElementText);
	xmlinitializev(result, ElementText, arguments);
	va_end(arguments);
	
	return result;
}

sdpXmlIElement* sdpXmlCreateElement(sdpXmlIElement* Parent, SDPCSTRING ElementName)
{
	// Sanity checks ...
	g_assert(Parent);
	g_assert_string(ElementName);

	return sdpXmlCreateElementV(Parent, ElementName, 0, 0);
}

sdpXmlIElement* sdpXmlCreateElement(sdpXmlIElement* Parent, SDPCSTRING ElementName, SDPCSTRING ElementText)
{
	// Sanity checks ...
	g_assert(Parent);
	g_assert_string(ElementName);
	
	return sdpXmlCreateElementV(Parent, ElementName, ElementText, 0);
}

sdpXmlIElement* sdpXmlFindElement(sdpXmlIElement* Parent, SDPCSTRING ElementName)
{
	// Sanity checks ...
	g_assert(Parent);
	g_assert_string(ElementName);

	sdpString name = ElementName;

	// For each child element ...
	for(SDPXMLELEMENTHANDLE e = Parent->GetFirstElement(); e; e = Parent->GetNextElement(e))
		if(name == Parent->GetElement(e)->Name())
			return Parent->GetElement(e);

	return 0;
}

sdpXmlIElement* sdpXmlFindElement(sdpXmlIElement* Parent, SDPCSTRING ElementName, SDPCSTRING AttributeName)
{
	// Sanity checks ...
	g_assert(Parent);
	g_assert_string(ElementName);
	g_assert_string(AttributeName);

	sdpString elementname = ElementName;

	// For each child element ...
	for(SDPXMLELEMENTHANDLE e = Parent->GetFirstElement(); e; e = Parent->GetNextElement(e))
		{
			sdpXmlIElement* element = Parent->GetElement(e);
			if(elementname != element->Name())
				continue;

			sdpXmlIAttribute* attribute = sdpXmlFindAttribute(element, AttributeName);
			if(attribute)
				return element;
		}

	return 0;
}

sdpXmlIElement* sdpXmlFindElement(sdpXmlIElement* Parent, SDPCSTRING ElementName, SDPCSTRING AttributeName, SDPCSTRING AttributeValue)
{
	// Sanity checks ...
	g_assert(Parent);
	g_assert_string(ElementName);
	g_assert_string(AttributeName);
	g_assert_string(AttributeValue);

	sdpString elementname = ElementName;

	// For each child element ...
	for(SDPXMLELEMENTHANDLE e = Parent->GetFirstElement(); e; e = Parent->GetNextElement(e))
		{
			sdpXmlIElement* element = Parent->GetElement(e);
			if(elementname != element->Name())
				continue;

			sdpXmlIAttribute* attribute = sdpXmlFindAttribute(element, AttributeName);
			if(attribute)
				if(sdpString(attribute->Value()) == sdpString(AttributeValue))
					return element;
		}

	return 0;
}

sdpXmlIAttribute* sdpXmlFindAttribute(sdpXmlIElement* Element, SDPCSTRING Name)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert_string(Name);

	sdpString name = Name;

	// For each child element ...
	for(SDPXMLATTRIBUTEHANDLE a = Element->GetFirstAttribute(); a; a = Element->GetNextAttribute(a))
		if(name == Element->GetAttribute(a)->Name())
			return Element->GetAttribute(a);

	return 0;
}
 
/// Functor for use with sdpXmlForEachElement() to count an element's children
class sdpXmlChildElementCounter
{
public:
	sdpXmlChildElementCounter() { m_Count = 0; }

	gulong Count() { return m_Count; }
	bool HandleElement(sdpXmlIElement* Element) { m_Count++; return true; }

protected:
	gulong m_Count;
};

gulong sdpXmlCountElements(sdpXmlIElement* Parent, SDPCSTRING ChildName)
{
	// Sanity checks ...
	g_assert(Parent);
	g_assert_string(ChildName);

	sdpXmlChildElementCounter functor;
	sdpXmlForEachElement(Parent, ChildName, functor);

	return functor.Count();
}

sdpString sdpXmlEncodeString(SDPCSTRING PlainText)
{
	sdpString plaintext(PlainText);
	
	return sdpXmlEncodeString(plaintext);
}

sdpString sdpXmlEncodeString(const sdpString& PlainText)
{
	sdpString result = PlainText;
	
	result.replaceall("&", "&amp;");	// This one's gotta be first!!!
	result.replaceall("\"", "&quot;");
	result.replaceall("<", "&lt;");
	result.replaceall(">", "&gt;");
	
	return result;
}

sdpString sdpXmlDecodeString(SDPCSTRING EncodedText)
{
	sdpString encodedtext(EncodedText);
	
	return sdpXmlDecodeString(encodedtext);
}

sdpString sdpXmlDecodeString(const sdpString& EncodedText)
{
	sdpString result = EncodedText;
	
	result.replaceall("&lt;", "<");
	result.replaceall("&gt;", ">");
	result.replaceall("&quot;", "\"");
	result.replaceall("&amp;", "&");
	
	return result;
}


